Package: openPrimeR
Title: Multiplex PCR Primer Design and Analysis
Version: 1.33.0
Authors@R: c(person("Matthias", "Döring", 
                    email = "matthias-doering@gmx.de", 
                    role = c("aut", "cre")),
            person("Nico", "Pfeifer", 
                    email = "pfeifer@informatik.uni-tuebingen.de", 
                    role = c("aut"))
            )
Description: An implementation of methods for designing, evaluating,
    and comparing primer sets for multiplex PCR.
    Primers are designed by solving a set cover problem such that
    the number of covered template sequences is maximized with
    the smallest possible set of primers.
    To guarantee that high-quality primers are generated,
    only primers fulfilling constraints on their physicochemical properties 
    are selected. A Shiny app providing a user interface for the functionalities
    of this package is provided by the 'openPrimeRui' package.
Depends: R (>= 4.0.0)
License: GPL-2
Encoding: UTF-8
RoxygenNote: 7.3.2
Imports: Biostrings (>= 2.38.4), pwalign, XML (>= 3.98-1.4), scales (>=
        0.4.0), reshape2 (>= 1.4.1), seqinr (>= 3.3-3), IRanges (>=
        2.4.8), GenomicRanges (>= 1.22.4), ggplot2 (>= 2.1.0), plyr (>=
        1.8.4), dplyr (>= 0.5.0), stringdist (>= 0.9.4.1), stringr (>=
        1.0.0), RColorBrewer (>= 1.1-2), DECIPHER (>= 1.16.1),
        lpSolveAPI (>= 5.5.2.0-17), digest (>= 0.6.9), Hmisc (>=
        3.17-4), ape (>= 3.5), BiocGenerics (>= 0.16.1), S4Vectors (>=
        0.8.11), foreach (>= 1.4.3), magrittr (>= 1.5), uniqtag (>=
        1.0), openxlsx (>= 4.0.17), grid (>= 3.1.0), grDevices (>=
        3.1.0), stats (>= 3.1.0), utils (>= 3.1.0), methods (>= 3.1.0)
Suggests: testthat (>= 1.0.2), knitr (>= 1.13), rmarkdown (>= 1.0),
        devtools (>= 1.12.0), doParallel (>= 1.0.10), pander (>=
        0.6.0), learnr (>= 0.9)
SystemRequirements: MAFFT (>= 7.305), OligoArrayAux (>= 3.8), ViennaRNA
        (>= 2.4.1), MELTING (>= 5.1.1), Pandoc (>= 1.12.3)
biocViews: Software, Technology, Coverage, MultipleComparison
VignetteBuilder: knitr
Collate: 'AnalysisStats.R' 'Comparison.R' 'Data.R' 'templates.R'
        'primers.R' 'IO.R' 'IO_view.R' 'Input.R' 'Ippolito.R'
        'Output.R' 'Plots.R' 'PrimerDesign.R' 'PrimerEvaluation.R'
        'RefCoverage.R' 'Scoring.R' 'SettingsDoc.R' 'TemplatesDoc.R'
        'Tiller.R' 'ambiguity.R' 'check_stop_codons.R'
        'con_annealing_temperature.R' 'con_dimerization.R'
        'con_gc_clamp.R' 'con_gc_ratio.R' 'con_melting_temperature.R'
        'con_primer_coverage.R' 'con_primer_efficiency.R'
        'con_primer_secondary_structures.R' 'con_repeats.R'
        'con_runs.R' 'con_template_secondary_structures.R'
        'constraints.R' 'constraints_eval.R' 'errors.R' 'filters.R'
        'helper_functions.R' 'initialize_primers.R'
        'initialize_primers_tree.R' 'openPrimeR.R' 'optimization_ILP.R'
        'optimization_algo.R' 'optimization_global.R'
        'optimization_greedy.R' 'plots_comparison.R' 'settings.R'
        'plots_constraints.R' 'plots_coverage.R' 'plots_filtering.R'
        'primer_significance.R' 'startApp.R' 'zzz.R'
git_url: https://git.bioconductor.org/packages/openPrimeR
git_branch: devel
git_last_commit: 9ac7cf4
git_last_commit_date: 2025-10-29
Repository: Bioconductor 3.23
Date/Publication: 2025-10-31
NeedsCompilation: no
Packaged: 2025-10-31 23:16:08 UTC; biocbuild
Author: Matthias Döring [aut, cre],
  Nico Pfeifer [aut]
Maintainer: Matthias Döring <matthias-doering@gmx.de>

# Generated by roxygen2: do not edit by hand

S3method(assign_binding_regions,character)
S3method(assign_binding_regions,numeric)
S3method(cbind,Primers)
S3method(cbind,Templates)
S3method(cbind,fill)
S3method(rbind,Primers)
S3method(rbind,Templates)
export(ConstraintOptions)
export(ConstraintSettings)
export(CoverageConstraints)
export(PCR_Conditions)
export(Primers)
export(Templates)
export(adjust_binding_regions)
export(assign_binding_regions)
export(check_constraints)
export(check_restriction_sites)
export(classify_design_problem)
export(create_coverage_xls)
export(create_report)
export(design_primers)
export(filter_primers)
export(get_comparison_table)
export(get_cvg_ratio)
export(get_cvg_stats)
export(get_cvg_stats_primer)
export(get_initial_primers)
export(parallel_setup)
export(plot_conservation)
export(plot_constraint)
export(plot_constraint_deviation)
export(plot_constraint_fulfillment)
export(plot_cvg_constraints)
export(plot_cvg_vs_set_size)
export(plot_penalty_vs_set_size)
export(plot_primer)
export(plot_primer_binding_regions)
export(plot_primer_cvg)
export(plot_primer_subsets)
export(plot_template_cvg)
export(primer_significance)
export(read_primers)
export(read_settings)
export(read_templates)
export(runTutorial)
export(score_conservation)
export(score_degen)
export(score_primers)
export(select_regions_by_conservation)
export(subset_primer_set)
export(update_template_cvg)
export(write_primers)
export(write_settings)
export(write_templates)
exportClasses(ConstraintOptions)
exportClasses(ConstraintSettings)
exportClasses(CoverageConstraints)
exportClasses(DesignSettings)
exportClasses(PCR_Conditions)
exportClasses(Primers)
exportClasses(Templates)
exportMethods("$<-")
exportMethods("PCR<-")
exportMethods("[")
exportMethods("conOptions<-")
exportMethods("constraintLimits<-")
exportMethods("constraints<-")
exportMethods("cvg_constraints<-")
exportMethods(PCR)
exportMethods(cbind2)
exportMethods(conOptions)
exportMethods(constraintLimits)
exportMethods(constraints)
exportMethods(cvg_constraints)
exportMethods(rbind2)
import(ggplot2)
import(lpSolveAPI)
import(methods)
importFrom(BiocGenerics,end)
importFrom(BiocGenerics,start)
importFrom(BiocGenerics,unlist)
importFrom(Biostrings,DNAStringSet)
importFrom(Biostrings,DNAStringSetList)
importFrom(Biostrings,DNA_BASES)
importFrom(Biostrings,IUPAC_CODE_MAP)
importFrom(Biostrings,extractAt)
importFrom(Biostrings,matchPattern)
importFrom(Biostrings,mergeIUPACLetters)
importFrom(Biostrings,reverseComplement)
importFrom(Biostrings,vmatchPattern)
importFrom(Biostrings,width)
importFrom(GenomicRanges,GRanges)
importFrom(IRanges,CharacterList)
importFrom(IRanges,IRanges)
importFrom(IRanges,Views)
importFrom(IRanges,as.matrix)
importFrom(IRanges,findOverlaps)
importFrom(IRanges,overlapsAny)
importFrom(RColorBrewer,brewer.pal)
importFrom(S4Vectors,"metadata<-")
importFrom(S4Vectors,metadata)
importFrom(foreach,"%dopar%")
importFrom(foreach,foreach)
importFrom(foreach,getDoParRegistered)
importFrom(foreach,getDoParWorkers)
importFrom(grDevices,colorRampPalette)
importFrom(magrittr,"%>%")
importFrom(plyr,.)
importFrom(plyr,arrange)
importFrom(plyr,catcolwise)
importFrom(plyr,ddply)
importFrom(plyr,here)
importFrom(plyr,numcolwise)
importFrom(plyr,rbind.fill)
importFrom(plyr,summarize)
importFrom(pwalign,compareStrings)
importFrom(pwalign,mismatch)
importFrom(pwalign,nucleotideSubstitutionMatrix)
importFrom(pwalign,pairwiseAlignment)
importFrom(reshape2,dcast)
importFrom(reshape2,melt)
importFrom(stats,as.formula)
importFrom(stats,ave)
importFrom(stats,fisher.test)
importFrom(stats,hclust)
importFrom(stats,na.omit)
importFrom(stats,p.adjust)
importFrom(stats,predict)
importFrom(stats,qnorm)
importFrom(stats,quantile)
importFrom(stats,reshape)
importFrom(stats,sd)
importFrom(utils,head)
importFrom(utils,read.csv)
importFrom(utils,read.delim)
importFrom(utils,setTxtProgressBar)
importFrom(utils,tail)
importFrom(utils,txtProgressBar)
importFrom(utils,write.csv)
importFrom(utils,write.table)

[![Build Status](https://travis-ci.org/matdoering/openPrimeR.svg?branch=master)](https://travis-ci.org/matdoering/openPrimeR)[![codecov](https://codecov.io/gh/matdoering/openPrimeR/branch/master/graph/badge.svg)](https://codecov.io/gh/matdoering/openPrimeR) 

# openPrimeR

## Synopsis
openPrimeR is an R package providing methods for designing, evaluating,and
comparing primer sets for multiplex polymerase chain reaction (PCR). The package provides a primer design
function that generates novel primer setes by solving a
set cover problem such that the number of covered template sequences is
maximized with the smallest possible set of primers. Moreover, existing primer sets can be evaluated
according to their coverage and their fulfillment of constraints on the
PCR-relevant physicochemical properties. For PCR tasks for which multiple
possible primer sets exist, openPrimeR can facilitate the selection of the
most suitable set by performing comparative analyses. The R package includes a Shiny application that
provides a comprehensive and intuitive user interface for the core functionalites of the package.

## More information
There is a [Conda repository for openPrimeR available](https://anaconda.org/bioconda/bioconductor-openprimer/badges).
For more information on how to install the third-party dependencies for openPrimeR, we refer to the corresponding [user-space repository](https://github.com/matdoering/openPrimeR-User), which provides installation routines.

### Docker Instructions

Our docker image is available at [dockerhub](https://hub.docker.com/r/mdoering88/openprimer/). In order to use the docker image, you need to [install docker](https://www.docker.com/) on your system and activate the docker daemon. 

After logging in with your docker account, in a console, enter 

>**docker pull mdoering88/openprimer**

to retrieve the latest docker image of the tool. Since the image is quite large the download (~4 GB) may take some time, especially if your internet connection is not very fast.

To run the image, enter

>**docker run -p 3838:3838 --rm mdoering88/openprimer**

After this, the tool is available by accessing **localhost:3838** in your web browser.

In case you want to have more control of the image you are running or you want to study the output of the tool, you can execute

>**docker run --rm -p 3838:3838 -v /tmp/logs/:/var/log/shiny-server/ mdoering88/openprimer:latest**

With this call, the tool's log file is stored in the */tmp/logs/* folder on your system with a filename starting with *shiny-shiny*). During the session, you can use 

>**tail -f shiny-shiny-X.txt**

to retrieve the current status messages. 

Moreover, in the above call to Docker, we have specified a tag, namely *:latest*, which means that we have started the most recent version of the tool. In case that no tag is provided, the latest available Docker image is used.

You can also work with the openPrimeR API through the container by executing

```
docker run -rm -it mdoering88/openprimer bash
```

and then starting an R session and loading the `openPrimeR` package.

## Changelog

Take a look at the [CHANGELOG](inst/NEWS) to view recent changes to the project.

##### FUNCTIONS DEALING WITH AMBIGUITIES

#' Sequence complement
#'
#' Complements the input sequence (re-write of seqinr comp function for gap support)
#' 
#' @param seq Input char vector.
#' @param forceToLower if TRUE the input is transformed to lower case.
#' @param ambiguous if TRUE ambiguous IUPAC nucleotides are complemented.
#' @return The complement of \code{seq}.
#' @keywords internal
comp <- function(seq, forceToLower = TRUE, ambiguous = FALSE) {
      
    if (all(seq %in% LETTERS)) {
        isUpper <- TRUE
    } else {
        isUpper <- FALSE
    }
    seq <- tolower(seq)
    result <- as.vector(seqinr::n2s((3 - seqinr::s2n(seq))))
    if (ambiguous) {
        result[which(seq == "b")] <- "v"
        result[which(seq == "d")] <- "h"
        result[which(seq == "h")] <- "d"
        result[which(seq == "k")] <- "m"
        result[which(seq == "m")] <- "k"
        result[which(seq == "s")] <- "s"
        result[which(seq == "v")] <- "b"
        result[which(seq == "w")] <- "w"
        result[which(seq == "n")] <- "n"
        result[which(seq == "y")] <- "r"
        result[which(seq == "r")] <- "y"
    }
    result[which(seq == "n")] <- "n"
    result[which(seq == "-")] <- "-"
    if (isUpper && !forceToLower) {
        result <- toupper(result)
    }
    return(result)
}

#' Conversion from IUPAC nucleotides
#' 
#' Convert sequences with IUPAC ambiguity codes to all possible sequences without ambiguities.
#'
#' @param seqs A vector of strings.
#'
#' @return A list containing the disambiguated input sequences.
#' @keywords internal
convert.from.iupac <- function(seqs) {
    if (length(seqs) == 0) {
        return(NULL)
    }
    degen.seqs <- DECIPHER::Disambiguate(DNAStringSet(seqs))
    result <- lapply(degen.seqs, function(x) tolower(as.character(x)))
    return(result)
}

#' Merge sequences.
#'
#' Merges the input sequences to one sequence containing IUPAC ambiguity codes.
#'
#' @param seqs Vector of strings.
#'
#' @return Consensus sequence of seqs.
#' @keywords internal
convert.to.iupac <- function(seqs) {
     
    if (length(unique(nchar(seqs))) != 1) {
        warning("Cannot convert to IUPAC. Sequences have different lengths.")
        return(seqs[1])
    }
    N <- nchar(seqs)[1]
    s <- strsplit(toupper(seqs), split = "")
    seq <- sapply(seq_len(N), function(x) paste(sapply(seq_along(s), function(y) s[[y]][x]), 
        collapse = ""))
    res <- paste(tolower(mergeIUPACLetters(seq)), collapse = "")
    return(res)
}

#' Sequence complement
#'
#' Computes the complement of the input sequence.
#'
#' @param seq The input sequence strings.
#'
#' @return The complements of the input sequences.
#' @keywords internal
complement.sequence <- function(seq) {
    idx <- which(seq != "")
    s.orig <- strsplit(seq[idx], split = "")
    complements <- lapply(s.orig, function(x) comp(x, ambiguous = TRUE))
    na.idx <- sapply(complements, function(x) which(is.na(x)))
    for (i in seq_along(na.idx)) {
        j <- na.idx[[i]]
        if (length(j) != 0) {
            s <- complements[[i]]
            s[j] <- s.orig[[i]][j]
            complements[[i]] <- s
        }
    }
    s <- sapply(complements, function(x) paste(x, collapse = ""))
    ret <- rep("", length(seq))
    ret[idx] <- s
    return(ret)
}
#' Reversion of a sequence
#'
#' Reverses the input sequences.
#'
#' @param seq the input sequence.
#'
#' @return The input sequence in reverse order.
#' @keywords internal
rev.sequence <- function(seq) {
      
    s <- sapply(strsplit(seq, split = ""), function(x) paste(rev(x), collapse = ""))
    return(s)
}

#' Reverse complement of a sequence
#'
#' Computes the reverse complement of the input sequences.
#'
#' @param seq the input strings
#'
#' @return The reverse complement of the input sequences.
#' @keywords internal
rev.comp.sequence <- function(seq) {
       s <- sapply(strsplit(seq, split = ""), function(x) paste(rev(comp(x, ambiguous = TRUE)), 
        collapse = ""))
    return(s)
}

#' Indices for merging sequences
#' 
#' Identifies the indices of similar input sequences to be merged.
#'
#' @param seqs The input sequence strings.
#' @param max.degeneracy The maximal allowed degeneracy of a merged seq.
#' 
#'@return A list of lists containing the indices of seqs to be merged. 
#'   For example [[1,2,3]] would indicate to merge primers 1, 2, and 3.
#' @keywords internal
get.merge.idx <- function(seqs, max.degeneracy) {
    s <- convert.from.iupac(seqs)
    merge.list <- NULL
    fw.l <- nchar(seqs)
    fw.lu <- unique(fw.l)
    if (0 %in% fw.lu) {
        # don't merge 0 length entries
        fw.lu <- fw.lu[-which(fw.lu == 0)]
    }
    len.idx <- lapply(seq_along(fw.lu), function(x) which(fw.l == fw.lu[x]))
    names(len.idx) <- fw.lu  # len.idx: all indices of primers of a given length
    merge.indices <- NULL
    for (i in seq_along(len.idx)) {
        # for every primer length
        idx <- len.idx[[i]]  # all primers of current len
        if (length(idx) == 1) {
            next
        }
        rnames <- idx[unlist(lapply(seq_along(s[idx]), function(x) rep(x, length(s[[idx[x]]]))))]  # idx of original primer in df
        p.matrix <- ape::as.DNAbin(ape::as.alignment(strsplit(unlist(s[idx]), split = "")))
        # set rownames of p.matrix
        rownames(p.matrix) <- seq_along(rnames)
        tree <- hclust.tree(p.matrix)
        if (length(tree) != 0) {
            t.seqs <- get.tree.seqs(tree, max.degeneracy, p.matrix)
            if (length(t.seqs) != 0) {
                # determine all possible merges of primers
                merge.idx <- lapply(t.seqs$Merge_Idx, function(x) unique(rnames[as.numeric(strsplit(x, 
                  ",")[[1]])]))
                merge.idx <- merge.idx[sapply(merge.idx, function(x) length(x) > 
                  1)]
                merge.sel <- merge.select(merge.idx)  # select only possible merges
                merge.indices <- c(merge.indices, merge.sel)
            }
        }
    }
    return(merge.indices)
}
#' Select merge indices
#'
#' Greedily identifies the smallest number of possible sequences merges that can be performed.
#'
#' @param merge.idx list of lists containing the indices of possible merges
#'
#' @return The smallest number of possible merge operations as an index list.
#' @keywords internal
merge.select <- function(merge.idx) {
      merge.len <- sapply(merge.idx, length)
    merge.idx <- merge.idx[order(merge.len, decreasing = TRUE)]  # sort by size of merge
    merge.out <- NULL  # vector of lists
    for (i in seq_along(merge.idx)) {
        idx <- merge.idx[[i]]
        if (any(idx %in% unlist(merge.out))) {
            # can't merge these primers
            next
        }
        merge.out <- c(merge.out, list(idx))
    }
    return(merge.out)
}
#' Merge similar primers
#'
#' Merges similar primers contained in the input primer data frame.
#'
#' @param primer.df Primer data frame.
#' @param mode.directionality Analysis direction.
#' @param max.degeneracy Maximal degeneracy of merged primers.
#'
#' @return A primer data frame where similar primers are merged into one entry. 
#' @keywords internal
merge.ambig.primers <- function(primer.df, mode.directionality = c("fw", "rev", "both"), max.degeneracy) {
    if (length(mode.directionality) == 0) {
        stop("'mode.directionality' not supplied.")
    }
    mode.directionality <- match.arg(mode.directionality) 
    if (mode.directionality == "fw") {
        merge.idx <- get.merge.idx(primer.df$Forward, max.degeneracy)
    } else if (mode.directionality == "rev") {
        merge.idx <- get.merge.idx(primer.df$Reverse, max.degeneracy)
    } else {
        # mode both merge only if we can merge both, fw, and rev, or if one of them is
        # not present.
        merge.idx.fw <- get.merge.idx(primer.df$Forward, max.degeneracy)
        merge.idx.rev <- get.merge.idx(primer.df$Reverse, max.degeneracy)
        idx.fw <- which(primer.df$Forward != "")
        idx.rev <- which(primer.df$Reverse != "")
        both <- intersect(idx.fw, idx.rev)
        m.fw <- merge.idx.fw[sapply(merge.idx.fw, function(x) !any(x %in% both))]  # merges of primers ONLY fw
        m.rev <- merge.idx.rev[sapply(merge.idx.rev, function(x) !any(x %in% both))]  # merges of primers ONLY rev
        # dont allow merges of primers containing both directions with primers containing
        # only one direction.  we might miss some merges among primers with both
        # directions, but that's ok.
        b.merges <- NULL # merges for pairs of primers
        if (length(merge.idx.fw) != 0 && length(merge.idx.rev) != 0) {
            both.merges.fw <- which(sapply(merge.idx.fw, function(x) all(x %in% both)))
            both.merges.rev <- which(sapply(merge.idx.rev, function(x) all(x %in% 
                both)))
            # if there's primers to be merged of one direction, check whether they perfectly agree with another pair
            if (length(both.merges.fw) != 0) {
                both.merges <- which(sapply(merge.idx.fw[both.merges.fw], function(x) any(sapply(merge.idx.rev[both.merges.rev], 
                  function(y) all(x == y)))))
                b.merges <- merge.idx.fw[both.merges.fw[both.merges]]  # merges of primers for both directions
            }
        } 
        merge.idx <- c(b.merges, m.fw, m.rev)
    }
    result <- merge.primer.entries(primer.df, merge.idx, mode.directionality)
    return(result)
}
#' Merge similar primers
#'
#' Merges the entries of similar entries in the input primer data frame, given a list with merge indices.
#'
#' @param opti.result Input primer data frame.
#' @param merge.idx List of lists with merge indices (get.merge.idx).
#' @param mode.directionality Direction of primers.
#'
#' @return A primer data frame where entries of similar primers are merged.
#' @keywords internal
merge.primer.entries <- function(opti.result, merge.idx, mode.directionality = c("fw", "rev","both")) {
    if (length(mode.directionality) == 0) {
        stop("'mode.directionality' not supplied.")
    }
    mode.directionality <- match.arg(mode.directionality)   
    if (length(opti.result) == 0 || nrow(opti.result) == 0 || length(merge.idx) == 
        0) {
        # nothing to merge
        return(opti.result)
    }
    new.seqs.fw <- NULL
    new.seqs.rev <- NULL
    if (mode.directionality == "fw") {
        new.seqs.fw <- merge.primer.entries.single(opti.result$Forward, merge.idx)
    } else if (mode.directionality == "rev") {
        new.seqs.rev <- merge.primer.entries.single(opti.result$Reverse, merge.idx)
    } else {
        # both
        new.seqs.fw <- merge.primer.entries.single(opti.result$Forward, merge.idx)
        new.seqs.rev <- merge.primer.entries.single(opti.result$Reverse, merge.idx)
    }
    R.idx <- NULL  # idx of merged rows to be removed after the for loop
    for (i in seq_along(merge.idx)) {
        # idea:retain the first primer.df entry per merge group -> repl.idx
        cur.idx <- merge.idx[[i]]
        repl.idx <- cur.idx[1]
        rm.idx <- cur.idx[2:length(cur.idx)]
        # replace the sequences of forward/reverse
        if (length(new.seqs.fw) != 0) {
            opti.result$Forward[repl.idx] <- new.seqs.fw[i]
        }
        if (length(new.seqs.rev) != 0) {
            opti.result$Reverse[repl.idx] <- new.seqs.rev[i]
        }
        # update ID
        opti.result$ID[repl.idx] <- paste(opti.result$ID[cur.idx], collapse = ";", 
            sep = "")
        # remove the merged rows
        R.idx <- c(R.idx, rm.idx)
    }
    if (length(R.idx) != 0) {
        opti.result <- opti.result[-R.idx, ]
    }
    return(opti.result)
}

#' Merge input sequences
#'
#' Merges the input sequences given a list with merge indices.
#'
#' @param seqs The input sequences.
#' @param merge.idx List of list with merge indices.
#'
#' @return Merged input sequences according to the input merge indices.
#' @keywords internal
merge.primer.entries.single <- function(seqs, merge.idx) {
    new.seqs <- rep("", length(merge.idx))
    for (i in seq_along(merge.idx)) {
        idx <- merge.idx[[i]]
        merge.seqs <- seqs[idx]
        if (any(merge.seqs == "")) {
            # merge not indended for this direction
            next
        }
        s <- convert.to.iupac(merge.seqs)
        new.seqs[i] <- s
    }
    return(as.character(new.seqs))
}

#' Disambiguation of Sequences.
#'
#' Disambiguates the input sequences, but does not disambiguate highly generate sequences.
#' @param template.seqs A \code{DNAStringSet} object with sequences to disambiguate.
#' @param gap.char The character indicating gaps in alignments.
#' @param degen.cutoff The maximal degeneration of sequences to be disambiguated.
#' @return A \code{DNAStringSetList} object with disambiguated sequences.
#' @keywords internal
my.disambiguate <- function(template.seqs, gap.char = "-", degen.cutoff = 2^10) {
    # consider ambiguous templates up to a cutoff of degeneration
    degen <- score_degen(strsplit(tolower(as.character(template.seqs)), split = ""), gap.char = gap.char)
    degen.idx <- which(degen <= degen.cutoff) # cutoff for degeneration
    degen.seqs <- DECIPHER::Disambiguate(template.seqs[degen.idx])
    seqs <- DNAStringSetList(as.list(as.character(template.seqs)))
    seqs[degen.idx] <- degen.seqs
    return(seqs)
}

context("Primer Design") 

test_that("primer_filtering_constraints", {
    # check that the filtering procedure excludes the correct primers
    data(Ippolito)
    constraints(settings)$primer_length <- c("min" = 18, "max" = 22)
    constraints(settings)$gc_clamp <- c("min" = 1, "max" = 2)
    constraints(settings)$gc_ratio <- c("min" = 0.4, "max" = 0.6)
    constraints(settings)$no_runs <- c("min" = 0, "max" = 3)
    constraints(settings)$no_repeats <- c("min" = 0, "max" = 3)
    constraints(settings)$melting_temp_range <- c("min" = 55, "max" = 65)
    constraints(settings)$melting_temp_diff <- c("min" = 0, "max" = 5)
    constraints(settings)$primer_coverage <- c("min" = 1)
    constraints(settings)$primer_specificity <- c("min" = 0.9, "max" = 1)
    conOptions(settings)$allowed_mismatches <- 5
    cvg_constraints(settings) <- list("annealing_DeltaG" = c("max" = -5))
    # evaluate constraints first to have reference values
    active.constraints <- c("primer_length", "gc_clamp", "gc_ratio", 
                            "no_runs", "no_repeats", 
                            "primer_coverage", "primer_specificity")
    filter.result <- suppressWarnings(cascaded.filter.quick(primer.df, template.df, settings,
                        to.compute.constraints = active.constraints))
    excluded.df <- filter.result$excluded
    # filtered by primer length:
    filter.reasons <- rep(NA, nrow(primer.df))
    filter.reasons[1:2] <- "primer_length"
    filter.reasons[3:7] <- "gc_clamp"
    filter.reasons[8] <- "primer_specificity"
    expect_equal(filter.reasons, excluded.df$Exclusion_Reason)
})
test_that("primer_degeneration", {
    primers <- c("ctccaaggt", "ktccaaggt",
                 "ntccaaggt", "nktccaggt",
                 "nkwsccaggt", "nkw-svhdb")
    degens <- c(1, 2, 4, 4 * 2, 4*2*2*2, 4*2*2*2*3*3*3*3)
    score <- score_degen(strsplit(primers, split = ""), gap.char = "-")
    expect_equal(score, degens)
})
test_that("primer_initialization_naive", {
    data(Ippolito)
    allowed.fw <- c(20,50)
    allowed.rev <- c(10,40)
    template.df <- assign_binding_regions(template.df, fw = allowed.fw, rev = allowed.rev)
    # nb: position check may fail by accident if primer binds earlier/later in some sequence by chance than intended ..
    # -> don't change selected templates!
    template.df <- template.df[c(1, 3),]
    # naive primer initialization fw
    primer.lengths <- 18:20
    primers <- create.initial.primer.set(template.df, primer.lengths, 
                mode.directionality = "fw", "test", allowed.region.definition = "within",
                init.algo = "naive", 16, 1)
    # ensure that all primers are unique
    expect_equal(length(unique(primers)), length(primers))
    # ensure that primer identifiers are unique
    expect_equal(length(unique(names(primers))), length(names(primers)))
    # check that primers match templates WITHIN the target region
    for (i in seq_along(primers)) {
        x <- primers[i] 
        hits <- regexpr(x, template.df$Sequence)
        match.len <- attr(hits, "match.length")
        hit.len.idx <- which(match.len != -1)
        # require at least one hit per primer:
        expect_gte(length(hit.len.idx), 1) 
        # check length of complementary region
        expect_lte(max(match.len[hit.len.idx]), max(primer.lengths))
        expect_gte(min(match.len[hit.len.idx]), min(primer.lengths))
        # check whether primer binds in allowed region
        binding.pos.s <- hits[hit.len.idx] # start of binding
        binding.pos.e <- hits[hit.len.idx] + (match.len[hit.len.idx] - 1) # end of binding
        expect_gte(min(binding.pos.s), allowed.fw[1])
        expect_lte(max(binding.pos.e), allowed.fw[2])
    }
    # check for "any" positions
    primers <- create.initial.primer.set(template.df, primer.lengths, 
                mode.directionality = "fw", "test", allowed.region.definition = "any",
                init.algo = "naive", 16, 1)
    min.pos.s <- 1000
    max.pos.e <- -1
    for (i in seq_along(primers)) {
        x <- primers[i] 
        hits <- regexpr(x, template.df$Sequence)
        match.len <- attr(hits, "match.length")
        hit.len.idx <- which(match.len != -1)
        # require at least one hit per primer:
        expect_gte(length(hit.len.idx), 1) 
        # check length of complementary region
        expect_lte(max(match.len[hit.len.idx]), max(primer.lengths))
        expect_gte(min(match.len[hit.len.idx]), min(primer.lengths))
        # check whether primer intersects with allowed region
        binding.pos.s <- hits[hit.len.idx] # start of binding
        binding.pos.e <- hits[hit.len.idx] + (match.len[hit.len.idx] - 1) # end of binding
        expect_lte(max(binding.pos.s), allowed.fw[2]) # don't overextend
        if (min(binding.pos.s) < min.pos.s) {
            min.pos.s <- min(binding.pos.s)
        }
        if (max(binding.pos.e) > max.pos.e) {
            max.pos.e <- max(binding.pos.e)
        }
    }
    expect_lt(min.pos.s, allowed.fw[1]) # there should be at least one primer binding earlier
    expect_gt(max.pos.e, allowed.fw[2]) # there should be at least one primer binding later
    # naive rev
    primers <- create.initial.primer.set(template.df, primer.lengths, 
                mode.directionality = "rev", "test", allowed.region.definition = "within",
                init.algo = "naive", 16, 1)
    for (i in seq_along(primers)) {
        x <- primers[i] 
        hits <- regexpr(x, rev.comp.sequence(template.df$Sequence))
        match.len <- attr(hits, "match.length")
        hit.len.idx <- which(match.len != -1)
        # require at least one hit per primer:
        expect_gte(length(hit.len.idx), 1) 
        # check length of complementary region
        expect_lte(max(match.len[hit.len.idx]), max(primer.lengths))
        expect_gte(min(match.len[hit.len.idx]), min(primer.lengths))
        # check whether primer binds in allowed region
        binding.pos.s <- hits[hit.len.idx] # start of binding
        binding.pos.e <- hits[hit.len.idx] + (match.len[hit.len.idx] - 1) # end of binding
        expect_gte(min(binding.pos.s), allowed.rev[1])
        expect_lte(max(binding.pos.e), allowed.rev[2])
    }
    # check for MAFFT:
    if (!check.tool.function()["MAFFT"]) {
        # cannot test without OligoArrayAux
        skip("MAFFT not available.")
    }
    # test tree primer creation
    primer.lengths <- 30
    data(Ippolito)
    template.df <- assign_binding_regions(template.df, fw = allowed.fw, rev = allowed.rev)
    max.degen <- 1
    idx <- 50:52
})


test_that("full_design_function", {
    skip_on_bioc() # design tests require a long time even for small sets
    data(Ippolito)
    template.df <- template.df[1:5,]
    constraints(settings)$primer_length[2] <- 18
    # test design without melting temp range but with melting temp diff:
    constraints(settings) <- constraints(settings)[names(constraints(settings)) != "melting_temp_range"]
    optimal.primers.greedy <- design_primers(template.df, "both", settings)
    # check that melting temp diff and coverage of the optimal set are ok
    opti.set <- optimal.primers.greedy$opti
    cvg.ratio <- as.numeric(get_cvg_ratio(opti.set, template.df))
    expect_equal(cvg.ratio, 1.0) # 100% cvg?
    # check Tm diff, should be small ...
    expect_lt(max(opti.set$melting_temp_diff), 5)
    # test without any melting temp
    constraints(settings) <- constraints(settings)[names(constraints(settings)) != "melting_temp_diff"]
    optimal.primers.greedy <- design_primers(template.df, "both", settings)
    # coverage ok?
    opti.set <- optimal.primers.greedy$opti
    cvg.ratio <- as.numeric(get_cvg_ratio(opti.set, template.df))
    expect_equal(cvg.ratio, 1.0) # 100% cvg?
    # only one result?
    expect_equal(length(optimal.primers.greedy$all_results), 1)
    # design with no constraints
    constraints(settings) <- list()
    optimal.primers.greedy <- try(design_primers(template.df, "both", settings), silent = TRUE)
    # shouldn't be possible to design without any constraints
    expect_that(class(optimal.primers.greedy), equals("try-error"))
    # but it should be possible with just primer length & coverage
    constraints(settings) <- list("primer_coverage" = c("min" = 1),
                                  "primer_length" = c("min" = 18, "max" = 18))
    optimal.primers.greedy <- design_primers(template.df, "both", settings)
    expect_equal(length(optimal.primers.greedy$all_results), 1.0)
    opti.set <- optimal.primers.greedy$opti
    cvg.ratio <- as.numeric(get_cvg_ratio(opti.set, template.df))
    expect_equal(cvg.ratio, 1.0) # 100% cvg?
    # test for ILP: use required.cvg arg
    data(Ippolito)
    template.df <- template.df[1:5,]
    # ramp up the constraints a bit to ensure we don't get a too high cvg
    constraints(settings)$no_runs <- c("min" = 2, "max" = 2)
    constraints(settings)$no_repeats <- c("min" = 2, "max" = 2)
    constraints(settings)$gc_clamp <- c("min" = 1, "max" = 1)
    # test with ILP
    optimal.primers.ILP <- design_primers(template.df, "both", settings,
                              required.cvg = 0.2, opti.algo = "ILP")
    opti.set <- optimal.primers.ILP$opti
    expect_gte(as.numeric(get_cvg_ratio(opti.set, template.df)), 0.2)
})

#' Identification of File extension.
#'
#' Identifies the file extension of \code{x}.
#' @param x A string for a filename.
#'
#' @return The extension of \code{x}.
#' @keywords internal
get.extension <- function(x) {
    pos <- regexpr("\\.([[:alnum:]]+)$", x)
    return(ifelse(pos > -1L, substring(x, pos + 1L), ""))
}
################ 
#' Copy Directories.
#'
#' Copies a directory to another location.
#' 
#' @param src.dir The directory to be copied.
#' @param dest.dir The target directory.
#' @param overwrite Overwrite existing files in \code{dest.dir}.
#' @return TRUE if copying was successful, FALSE otherwise.
#' @keywords internal
dir.copy <- function(src.dir, dest.dir, overwrite) {
    file.names <- dir(src.dir)
    suppressWarnings(dir.create(dest.dir))
    return(file.copy(from = file.path(src.dir, file.names), to = dest.dir, overwrite = overwrite, 
        recursive = TRUE))
}
#' Index for Unlisting.
#'
#' Determines indices for unlisting.
#'
#' @param primer.start Numeric vector.
#' @param primer.data.idx Selection indices.
#'
#' @return Indices.
#' @keywords internal
get.unlist.idx <- function(primer.start, primer.data.idx) {
    un <- unique(primer.data.idx)
    t <- rep(1, length(un))
    cat.count <- table(primer.data.idx)
    cat <- as.numeric(names(cat.count))
    names(t) <- un
    result <- rep(NA, length(primer.data.idx))
    prev.count <- 0
    for (i in seq_along(primer.data.idx)) {
        idx <- t[as.character(primer.data.idx[i])]
        cat.c <- sum(cat.count[which(cat < primer.data.idx[i])])  # nbr of list entries appearing before
        result[i] <- idx + cat.c
        t[as.character(primer.data.idx[i])] <- t[as.character(primer.data.idx[i])] + 
            1
    }
    return(result)
}

#' Rbind for Primer Data Frames.
#'
#' Merges all primer data frames in \code{primer.data} into one data frame.
#'
#' @param primer.data List with primer data frames.
#' @return A data frame containing all data in \code{primer.data}.
#' @keywords internal
rbind.primer.data <- function(primer.data) {
    if (length(primer.data) == 0 || all(sapply(primer.data, function(x) nrow(x) == 
        0))) {
        return(NULL)
    }
    # select non-zero primer sets
    idx <- sapply(primer.data, function(x) nrow(x) != 0)
    run.names <- get.run.names(primer.data)[idx]
    unique.names <- make.unique(run.names)
    idx <- which(idx)
    names(primer.data)[idx] <- unique.names  # list names need to be unique for my_rbind!
    data <- my_rbind(primer.data[idx])
    return(data)
}
#' Format Strings
#'
#' Changes the representation of the comma-separated string input.
#'
#' @param values A comma-separated string with values.
#' @return A percentage-formatted representation of the input string.
#' @keywords internal
string.list.format.total <- function(values) {
    # format a list of values to percentages for the whole data set
    na.idx <- which(is.na(values))
    if (length(na.idx) != 0) {
        values[na.idx] <- ""
    }
    v <- unlist(strsplit(values, split = ","))
    t <- table(v)
    t <- t[order(t, decreasing = TRUE)]
    res <- paste(names(t), "(", round((t/sum(t)) * 100, 2), "%)", collapse = ",", 
        sep = "")
    return(res)
}
#############
# helper functions for plotting
#############
#' Plot Extent
#'
#' Returns the extent of a plot.
#'
#' @param N Number of observations to plot.
#' @param px.per.n Pixels required per observations.
#' @param min.size Minimal extent of plot in pixels.
#' @param max.size Maximal extent of plot in pixels.
#' @return The extent of the plot.
#' @keywords internal
get.plot.height <- function(N, px.per.n = 50, min.size = 300, 
    max.size = 1500000) {

    height <- min(N * px.per.n + min.size, max.size)
    return(height)
}
#' Smartbind preserving classes.
#'
#' Rbind allowing for column mismatch, retains the classes of the data frames.
#' Motivation: smartbind/rbind.fill only keep the data.frame class
#' but not additional classes.
#'
#' @param ... Data frames.
#' @return A data frame resulting from row binding of \code{...}.
#' @keywords internal
my_rbind <- function(...) {
    df <- rbind.fill(...)
    args <- list(...)
    classes <- sapply(args, class)
    if (length(args) == 0) {
        return(NULL)
    }
    if (any(classes == "Templates")) {
       df <- Templates(df) 
    } 
    if (any(classes == "Primers")) {
        df <- Primers(df)
    }
    return(df)
}
#' Conversion of Positions to Ranges.
#' 
#' Converts two numeric values to a range.
#'
#' @param pos1 The first value.
#' @param pos2 The second value.
#' @return A character vector range.
#' @keywords internal
pos.to.range <- function(pos1, pos2) {
    s1 <- strsplit(pos1, split = ",")
    s2 <- strsplit(pos2, split = ",")
    res <- unlist(lapply(seq_along(s1), function(x) if (length(s1[[x]]) != 0 && length(s2[[x]]) != 
        0) 
        paste(s1[[x]], " to ", s2[[x]], collapse = ",", sep = "") else ""))
    return(res)
}

########
# Functions for formatted output.
###########

#' View the Input Primers.
#'
#' Creates a formatted primers table.
#'
#' @param primer.df A \code{Primers} object.
#' @param mode.directionality The direction of the primers.
#' @param for.shiny Whether output is intended for Shiny.
#' @return A formatted primer table.
#' @keywords internal
view.input.primers <- function(primer.df, mode.directionality, for.shiny = TRUE) {
    if (nrow(primer.df) == 0 || length(primer.df) == 0) {
        return(NULL)
    }
    view.df <- asS3(primer.df) # modifying columns here -> cannot guarantee object validity
    # remove all columns where all values are missing
    excl.idx <- which(unlist(lapply(seq_len(ncol(view.df)), function(x) 
                all(view.df[,x] == "" | is.na(view.df[,x])))))
    # additionally remove columns depending on run mode
    excl <- NULL
    if (mode.directionality == "fw") {
        excl <- grep("_rev", colnames(view.df))
    } else if (mode.directionality == "rev") {
        excl <- grep("_fw", colnames(view.df))
    }
    excl.idx <- unique(c(excl.idx, excl))
    if (length(excl.idx) != 0) {
        view.df <- view.df[, -excl.idx]
    }
    excl.col <- c("Identifier", "Run")
    view.df <- exclude.cols(excl.col, view.df)
    view.df <- modify.col.rep(view.df, for.shiny)
    return(view.df)
}
#' View the Evaluated Primers.
#'
#' Creates a formatted primers table.
#'
#' @param primer.df A \code{Primers} object.
#' @param template.df A \code{Templates} object.
#' @param mode.directionality The direction of the primers.
#' @param view.cvg.individual Whether information on individual coverage events should be retrained.
#' @param for.shiny Whether the table is intended for Shiny (HTML) or not.
#' @return A formatted primer table.
#' @keywords internal
view.cvg.primers <- function(primer.df, template.df, mode.directionality, 
                            view.cvg.individual = c("active", "inactive"), 
                            for.shiny = TRUE) {
    if (length(primer.df) == 0) {
        return(NULL)
    } else if (nrow(primer.df) == 0) {
        return(primer.df)
    }
    if (length(view.cvg.individual) == 0) {
        stop("Please supply the 'view.cvg.individual' arg.")
    }
    view.cvg.individual <- match.arg(view.cvg.individual)
    view.df <- asS3(primer.df)
    # convert binding positions to range
    view.df$Binding_Position_Start_fw <- pos.to.range(view.df$Binding_Position_Start_fw, 
        view.df$Binding_Position_End_fw)
    view.df$Binding_Position_Start_rev <- pos.to.range(view.df$Binding_Position_Start_rev, 
        view.df$Binding_Position_End_rev)
    view.df$Relative_Forward_Binding_Position_Start_fw <- pos.to.range(view.df$Relative_Forward_Binding_Position_Start_fw, 
        view.df$Relative_Forward_Binding_Position_End_fw)
    view.df$Relative_Forward_Binding_Position_Start_rev <- pos.to.range(view.df$Relative_Forward_Binding_Position_Start_rev, 
        view.df$Relative_Forward_Binding_Position_End_rev)
    view.df$Relative_Reverse_Binding_Position_Start_fw <- pos.to.range(view.df$Relative_Reverse_Binding_Position_Start_fw, 
        view.df$Relative_Reverse_Binding_Position_End_fw)
    view.df$Relative_Reverse_Binding_Position_Start_rev <- pos.to.range(view.df$Relative_Reverse_Binding_Position_End_rev, 
        view.df$Relative_Reverse_Binding_Position_Start_rev)
    # reorder columns:
    col.order <- c("ID", "Forward", "Reverse", "Coverage_Ratio",
                   "Nbr_of_mismatches_fw", "Nbr_of_mismatches_rev", 
                   "Relative_Forward_Binding_Position_Start_fw", 
                   "Relative_Reverse_Binding_Position_Start_rev",
                  "melting_temp", "melting_temp_diff", "Self_Dimer_DeltaG",
                  "Cross_Dimer_DeltaG", "Structure_deltaG", 
                  "primer_specificity", "mean_primer_efficiency",
                  "gc_ratio_fw", "gc_ratio_rev", "gc_clamp_fw",
                  "gc_clamp_rev", "no_repeats_fw", "no_repeats_rev", 
                  "no_runs_fw", "no_runs_rev",
                  "Covered_Seqs",  # primer coverage not used here
                  "Self_Dimer_Structure", "Cross_Dimer_Structure",
                  "Structure_fw", "Structure_rev", 
                  "Degeneracy_fw", "Degeneracy_rev",
                  "primer_length_fw", "primer_length_rev", 
                  "Direction")
    cur.order <- col.order[col.order %in% colnames(view.df)]
    other.cols <- setdiff(colnames(view.df), cur.order)  # don't care about these cols in ordering
    view.df <- view.df[, c(cur.order, other.cols)]
    # don't order!
    #view.df <- view.df[order(view.df$primer_coverage, decreasing = TRUE), ]
    # remove unimportant cols from table
    view.df <- view.df[, colnames(view.df) %in% col.order]
    # percent format some columns:
    percent.format.cols <- c("Coverage_Ratio", "gc_ratio_fw", "gc_ratio_rev",
                             "mean_primer_efficiency", "primer_specificity")
    for (i in seq_along(percent.format.cols)) {
        col <- percent.format.cols[i]
        if (col %in% colnames(view.df)) {
            view.df[, col] <- paste0(round(view.df[, col], 3) * 100, "%")
        }
    }
        # convert covered template seqs to group representation
    cvd <- covered.seqs.to.ID.string(as.character(view.df$Covered_Seqs), template.df)
    if (length(unique(template.df$Group)) >= 2 && view.cvg.individual == "inactive") {
        # show gene group instead of identifiers
        idx <- covered.seqs.to.idx(as.character(view.df$Covered_Seqs), template.df)
        cvd <- string.list.format(sapply(seq_along(idx), function(x) paste(template.df[idx[[x]], 
            "Group"], collapse = ",")))
    }
    view.df$Covered_Seqs <- unlist(cvd)
    format.cols <- c("Binding_Position_Start_fw", "Binding_Position_Start_rev", 
        "Relative_Forward_Binding_Position_Start_fw", "Relative_Reverse_Binding_Position_Start_rev", "Nbr_of_mismatches_fw", "Nbr_of_mismatches_rev", 
        "Binding_Region_Allowed", "Binding_Region_Allowed_fw", "Binding_Region_Allowed_rev", 
        "T_EVAL_primer_efficiency", "constraints_passed_T")
    if (view.cvg.individual == "inactive") {
        for (i in seq_along(format.cols)) {
            if (format.cols[i] %in% colnames(view.df)) {
                if (format.cols[i] %in% c("Nbr_of_mismatches_fw", "Nbr_of_mismatches_rev")) {
                    view.df[, format.cols[i]] <- string.list.format(view.df[, format.cols[i]], order.mode = "value")
                } else {
                    view.df[, format.cols[i]] <- string.list.format(view.df[, format.cols[i]])
                }
            }
        }
        mismatch.cols <- c("Mismatch_pos_fw", "Mismatch_pos_rev")
        for (j in seq_along(mismatch.cols)) {
            col <- mismatch.cols[j]
            if (col %in% colnames(view.df)) {
                view.df[, col] <- string.list.format(gsub("[()]", "", 
                    gsub(",()", "", 
                        gsub("(),", "", view.df[, col], fixed = TRUE), 
                    fixed = TRUE)))
            }
        }
    }
    # html representation of dimer structures:
    structure.cols <- c("Self_Dimer_Structure", "Cross_Dimer_Structure")
    for (j in seq_along(structure.cols)) {
        col <- structure.cols[j]
        if (col %in% colnames(view.df)) {
            view.df[, col] <- html.format.structure(as.character(view.df[, col]))
        }
    }
    #####
    # if we don't have primers of both directions, merge fw and rev annotations
    both.mode <- any(view.df$Forward != "" & view.df$Reverse != "")
    fw.idx <- which(view.df$Forward != "")
    rev.idx <- which(view.df$Reverse != "")
    if (!both.mode) {
        # merge constraint columns to single column
        cols <- c("gc_ratio", "gc_clamp", "Nbr_of_mismatches",
                  "no_repeats", "no_runs", "primer_length",
                  "Structure", "Sequence", "Relative_Binding_Position",
                  "Degeneracy")
        special.cols <- list("Sequence" = c("Forward", "Reverse"),
                            "Relative_Binding_Position" = 
                            c("Relative_Forward_Binding_Position_Start_fw",
                            "Relative_Reverse_Binding_Position_Start_rev"))
        for (i in seq_along(cols)) {
            col <- cols[i]
            if (col %in% names(special.cols)) {
                col.dir <- special.cols[[col]]
            } else {
                col.dir <- paste0(col, c("_fw", "_rev"))
            }
            if (all(col.dir %in% colnames(view.df))) {
                merge.col <- unlist(lapply(seq_len(nrow(view.df)), function(x) 
                                if (x %in% fw.idx) {
                                    view.df[x, col.dir[1]] 
                                } else {
                                    view.df[x, col.dir[2]]
                                }))
                view.df[, col.dir[1]] <- merge.col
                # rename col
                colnames(view.df)[colnames(view.df) == col.dir[1]] <- col
                # remove old columns
                view.df <- view.df[,-which(colnames(view.df) == col.dir[2])]
            }
        }
    }
    ########
    excl.col <- c("Binding_Position_End_fw", "Binding_Position_End_rev", "Relative_Forward_Binding_Position_Start_rev", 
        "Relative_Reverse_Binding_Position_Start_fw", "Relative_Forward_Binding_Position_End_fw", 
        "Relative_Reverse_Binding_Position_End_rev", "Relative_Reverse_Binding_Position_End_fw", 
        "Relative_Forward_Binding_Position_End_rev")
    view.df <- exclude.cols(excl.col, view.df)

    # round all numeric columns:
    nums <- vapply(view.df, is.numeric, FUN.VALUE = logical(1))
    view.df[,nums] <- round(view.df[,nums], 2)
    ############
    # rename columns for frontend (format/units)
    name.change.cols <- c("melting_temp", "melting_temp_diff", "Self_Dimer_DeltaG", 
                          "Cross_Dimer_DeltaG", "Structure_deltaG",
                          "Coverage_Ratio", 
                          "gc_ratio", "gc_ratio_fw", "gc_ratio_rev", 
                          "gc_clamp", "gc_clamp_fw", "gc_clamp_rev", 
                          "Nbr_of_mismatches", 
                          "Nbr_of_mismatches_fw", "Nbr_of_mismatches_rev", 
                          "no_repeats", "no_repeats_fw", "no_Repeats_rev",
                          "no_runs", "no_runs_fw", "no_runs_rev", 
                          "Structure", "Structure_fw", "Structure_rev",
                          "mean_primer_efficiency", "primer_specificity",
                          "Binding_Position_Start",
                          "Binding_Position_Start_fw", 
                          "Binding_Position_Start_rev", 
                          "Relative_Binding_Position",
                          "Relative_Forward_Binding_Position_Start_fw", 
                          "Relative_Reverse_Binding_Position_Start_rev")
    if (for.shiny) {
        new.names <- c("T<sub>m</sub> [&deg;C]", "&Delta;T<sub>m</sub> [&deg;C]",
                    "Self Dimer &Delta;G[<sup>kcal</sup>&frasl;<sub>mol</sub>]", 
                    "Cross Dimer &Delta;G[<sup>kcal</sup>&frasl;<sub>mol</sub>]",
                    "Structure &Delta;G[<sup>kcal</sup>&frasl;<sub>mol</sub>]",
                    "Coverage", "GC_ratio", "GC_ratio_fw", "GC_ratio_rev", 
                    "GC_clamp", "GC_clamp_fw", "GC_clamp_rev", 
                    "Mismatches", "Mismatches_fw", "Mismatches_rev", 
                    "Repeat count", "Repeat count_fw", "Repeat count_rev", 
                    "Run count", "Run count_fw", "Run count_rev", 
                    "Secondary structure", "Secondary structure_fw",
                    "Secondary structure_rev", 
                    "efficiency", "specificity",
                    "Template_Binding_Range", 
                    "Template_Binding_Range_fw", "Template_Binding_Range_rev", 
                    "Relative_Binding_Range",
                    "Relative_fw_Binding_Range", "Relative_rev_Binding_Range")
    } else {
        new.names <- c("Tm [C]", "Delta Tm [C]",
                    "Self Dimer", "Cross Dimer", "Structure",
                    "Coverage", "GC_ratio", "GC_ratio_fw", "GC_ratio_rev", 
                    "GC_clamp", "GC_clamp_fw", "GC_clamp_rev", 
                    "Mismatches", "Mismatches_fw", "Mismatches_rev", 
                    "Repeat count", "Repeat count_fw", "Repeat count_rev", 
                    "Run count", "Run count_fw", "Run count_rev", 
                    "Secondary structure", "Secondary structure_fw",
                    "Secondary structure_rev", 
                    "efficiency", "specificity",
                    "Template_Binding_Range", 
                    "Template_Binding_Range_fw", "Template_Binding_Range_rev", 
                    "Relative_Binding_Range",
                    "Relative_fw_Binding_Range", "Relative_rev_Binding_Range")
    }
    idx <- match(colnames(view.df), name.change.cols)
    colnames(view.df)[!is.na(idx)] <- new.names[idx[!is.na(idx)]]
    view.df <- view.input.primers(view.df, mode.directionality, for.shiny)
    return(view.df)
}

#' View the Evaluated Primers.
#'
#' Creates a formatted primers table.
#'
#' @param primer.df A \code{Primers} object.
#' @param template.df A \code{Templates} object.
#' @return A formatted primer table.
#' @keywords internal
view.primers <- function(primer.df, template.df) {
    mode.directionality <- get.analysis.mode(primer.df)
    # pair primers for a more compact representation
    primer.df <- pair_primers(primer.df, template.df)
    out.df <- view.cvg.primers(primer.df, template.df, 
                            mode.directionality, view.cvg.individual = "inactive", 
                            for.shiny = FALSE)
    # format relative binding range: select the majority binding region
    range.cols <- c("Relative Binding Range", "Relative Binding Range (fw)", "Relative Binding Range (rev)")
    for (i in seq_along(range.cols)) {
        col <- range.cols[i]
        if (col %in% colnames(out.df)) {
            out.df[, col] <- unlist(lapply(strsplit(out.df[, col], split = ","),
          function(x) {
            str <- ifelse(length(x) != 0, x[[1]], "")
            out <- ""
            if (str != "") {
                # the relative binding range: e.g. 0 to 30
                conv <- stringr::str_extract_all(str,"\\(?[0-9.%-]+\\)?")[[1]]
                if (as.numeric(conv[2]) > 0) { # end of binding
                    conv[2] <- paste0("+", conv[2])
                }
                if (as.numeric(conv[1]) > 0) { # start of binding
                    conv[1] <- paste0("+", conv[1])
                }
                out <- paste0(paste0(conv[seq_len(2)], collapse = " to "), " ", conv[3], collapse = " ")
            }
            return(out)
            }))
        }
    }
    # abbreviate IDs:
    out.df$ID <- abbreviate(out.df$ID, getOption("openPrimeR.plot_abbrev"))
    return(out.df)
}
#' Format a Sequence for LateX output.
#'
#' Formats a sequence for LateX report output.
#'
#' @param seqs Character vector of sequences.
#' @return Formatted sequences.
#' @keywords internal
format.seqs.tex <- function(seqs) {
    seqs <- seqs
    # italicize ambiguous bases 
    ambig.bases <- tolower(setdiff(names(IUPAC_CODE_MAP), DNA_BASES))
    replacement <- paste0("\\\\textit{", ambig.bases, "}")
    names(replacement) <- ambig.bases
    seqs <- stringr::str_replace_all(seqs, replacement)
    return(seqs)
}
#' View the Evaluated Primers in the Report.
#'
#' Creates a formatted primers table for the report PDF.
#'
#' @param primer.df A \code{Primers} object.
#' @param template.df A \code{Templates} object.
#' @return A formatted primer table.
#' @keywords internal
view.primers.report <- function(primer.df, template.df) {
    out.df <- view.primers(primer.df, template.df)
    # determine nbr of columns
    if ("Sequence" %in% colnames(out.df)) {
        # single direction: more space available
        out.df <- out.df[,seq_len(6)]
    } else {
        # pairs of primers: less space available
        out.df <- out.df[,seq_len(4)]
    }
    col.names <- colnames(out.df) 
    # change mismatch representation to IQR representation:
    mm.values <- sapply(seq_len(nrow(primer.df)), function(x) paste0(primer.df[x, "Nbr_of_mismatches_fw"], 
                                                                primer.df[x, "Nbr_of_mismatches_rev"], collapse = ","))
    #print(mm.values)
    #print("MM values:")
    mm.rep <- string.to.IQR(mm.values)
    if ("Sequence" %in% colnames(out.df)) { 
        # TODO: Integrate value columns when merging primers in 'view.primers()'
        # for both, the size of the primer data frame is reduced
        # .. and we haven't defined the values of the merged columns ..
        # -> number of mismatches is NA at the moment, also, we don't show more than 4 columns in the report anyway.
        out.df$Mismatches <- mm.rep
    }
    # tidy up the column names that are displayed only (the first couple of columns):
    col.names[col.names == "Tm [C]"] <- "$\\text{T}_\\text{m} [^{\\circ}\\text{C}]$"
    col.names[col.names == "Delta Tm [C]"] <- "$\\Delta\\text{T}_\\text{m} [^{\\circ}\\text{C}]$"
    col.names[col.names == "Relative Binding Range"] <- "Position"
    col.names[col.names == "Relative Binding Range (fw)"] <- "Position (rev)"
    col.names[col.names == "Relative Binding Range (rev)"] <- "Position (fw)"
    col.names[col.names == "Self Dimer"] <- "$\\text{Self Dimer} \\Delta\\text{G}$"
    col.names[col.names == "Cross Dimer"] <- "$\\text{Cross Dimer } \\Delta\\text{G}$"
    # nb: if only one direction was analyzed, the primer sequence resides
    # in 'Sequence'. Otherwise we still have the 'Forward' and 'Reverse' columns.
    out.df[col.names == "Sequence"] <- format.seqs.tex(out.df$Sequence)
    out.df[col.names == "Forward"] <- format.seqs.tex(out.df$Forward)
    out.df[col.names == "Reverse"] <- format.seqs.tex(out.df$Reverse)
    colnames(out.df) <- col.names
    return(out.df)
}

#' Exclusion of Columns
#'
#' Removes columns from a data frame.
#'
#' @param excl.col Names of columns in \code{template.df} to be removed.
#' @param template.df Data frame for which columns in \code{excl.col} should be removed.
#' @return \code{template.df} with removed columns as specified in \code{excl.col}.
#' @keywords internal
exclude.cols <- function(excl.col, template.df) {
    m <- match(excl.col, colnames(template.df))
    if (any(!is.na(m))) {
        view.df <- template.df[, -m[!is.na(m)]]
    } else {
        view.df <- template.df
    }
    return(view.df)
}

#' Modification of Column Names.
#'
#' Modifies column names for frontend output.
#' 
#' @param template.df The data frame whose column names are to be modified.
#' @param for.shiny Whether formatting should be for shiny.
#' @return \code{template.df} with modified column names.
#' @keywords internal
modify.col.rep <- function(template.df, for.shiny = TRUE) {
    # identicate direction of constraint with brackets:
    colnames(template.df) <- gsub("_fw", " (fw)", colnames(template.df))
    colnames(template.df) <- gsub("_rev", " (rev)", colnames(template.df))
    # replace underscores in columns:
    colnames(template.df) <- gsub("_", " ", colnames(template.df))
    # upper case first letter
    colnames(template.df) <- Hmisc::capitalize(colnames(template.df))
    return(template.df)
}
#' Format a String List.
#'
#' Formats a string list, summarizing values with percentages.
#'
#' @param values The string list to format.
#' @param order.mode How the result should be ordered.
#' For "percentage", the strings are ordered by their percentages, while
#' for "value", the strings are ordered by their values.
#' @return A formatted string with percentage annotations.
#' @keywords internal
string.list.format <- function(values, order.mode = c("percentage", "value")) {
    # values in comma separated string will be formatted
    order.mode <- match.arg(order.mode)
    v <- strsplit(values, split = ",")
    n.idx <- which(sapply(v, function(x) length(x)) == 0)
    # for every list element determine its frequency distribution
    dist <- lapply(v, function(x) table(x))
    if (order.mode == "percentage") {
        # order by largest percentage
        dist <- lapply(dist, function(x) x[order(x, decreasing = TRUE)])
    } else if (order.mode == "value") {
        values <- unique(unlist(lapply(dist, function(x) as.numeric(names(x)))))
        values <- as.character(values[order(values)])
        dist <- lapply(dist, function(x) {
                        x <- x[values]
                        x[is.na(x)] <- 0
                        names(x) <- values
                        return(x)
                       })
    } else {
        stop("Unknown order mode.")
    }
    f <- unlist(lapply(dist, function(x) paste(names(x), 
        " (", round((x/sum(x, na.rm = TRUE)), 
        2) * 100, "%)", collapse = ",", sep = "")))
    if (length(n.idx) != 0) {
        f[n.idx] <- ""
    }
    return(f)
}

#########
# Constraint: GC ratio
#########

#' GC ratio
#'
#' Computes the ratio of G/Cs in a sequence. 
#'
#' In case of ambiguities, the mean GC ratio of all possible sequences is computed.
#' 
#' @param x Input sequence.
#' @return The fraction of G/Cs in \code{x}.
#' @keywords internal
compute.gc.ratio <- function(x) {
    x.iupac <- convert.from.iupac(tolower(x))
    result <- rep(0, length(x))
    idx <- which(x != "")
    s <- lapply(seq_along(idx), function(x) strsplit(x.iupac[[idx[x]]], split = ""))
    # check each disambiguated seq
    hits <- lapply(seq_along(s), function(x) sapply(s[[x]], 
                                    function(y) length(which(y == "g" | y == "c")))) 
    if (length(idx) != 0) {
        # compute mean ratio for ambig seqs
        ratios <- sapply(seq_along(idx), function(y) 
                                            mean(hits[[y]]/nchar(x[idx[y]])))
        result[idx] <- ratios
    }
    return(result)
}

context("GC clamp and GC ratio")

test_that("GC Clamp", {
    # test primers for GC clamp
    expect_that(evaluate.GC.clamp("CAGAC"), equals(1))
    expect_that(evaluate.GC.clamp("CAGACC"), equals(2))
    expect_that(evaluate.GC.clamp("CAGACCG"), equals(3))
    expect_that(evaluate.GC.clamp("CAGACCGA"), equals(0))
}) 

test_that("GC Ratio", {
    # test GC ratios of sequences
    expect_that(compute.gc.ratio("CGACTACG"), equals(5/8))
    expect_that(compute.gc.ratio("ACG"), equals(2/3))
    expect_that(compute.gc.ratio("ATT"), equals(0))
    # test ambiguities: mean of disambiguated seqs
    seq <- "CGAGTAGTMC" # M -> A/C
    expect <- (5/10 + 6/10) / 2
    expect_that(compute.gc.ratio(seq), equals(expect))
})

####################
# Constraint plots
####################
#' Plot of Template Folding Energies.
#'
#' Plots the DeltaDeltaG of template folding, which is 
#' the difference between the free energy change of
#' the unconstrained folding and the free energy change
#' of the constrained folding.
#'
#' @param fold.df A data frame with free energies for the
#' template regions.
#' @return A plot of DeltaDeltaG.
#' @keywords internal
plot_template_structure <- function(fold.df) {
    if (length(fold.df) == 0) {
        # nothing to plot
        return(NULL)
    }
    # discretize regions:
    interval.df <- fold.df[, c("Constraint_Region_Start", "Constraint_Region_End")]
    intervals <- unlist(lapply(seq_len(nrow(interval.df)), function(x) paste0("[", interval.df[x, 1], ",", interval.df[x, 2], "]")))
    # introduce an interval order:
    o.intervals <- order(interval.df$Constraint_Region_Start)
    intervals <- factor(intervals, levels = unique(intervals[o.intervals]))
    fold.df$Region <- intervals
    ylab <- expression(paste(Delta, " G (no structure) - ", Delta, " G (structure) [kcal/mol]", sep = ""))
    p <- ggplot(fold.df) + 
        geom_boxplot(aes(x = .data[["Region"]], y = .data[["Delta_Delta_G"]], fill = .data[["Group"]])) +
        facet_wrap(~Direction, scales = "free_x") +
        ggtitle("Template secondary structures") + 
        ylab(ylab)
    return(p)
}
#' Retrieve data for Constraint Deviations.
#'
#' @param constraint.df An evaluated object of class \code{Primers}.
#' @param constraint.settings A list with settings for the constraints
#' that are to be evalated.
#' @return A data frame providing primer-specific 
#' information on deviations of primer properties from
#' the desired properties.
#' @keywords internal
get_constraint_deviation_data <- function(constraint.df, constraint.settings) {
    if (length(constraint.df) == 0 || nrow(constraint.df) == 0) {
        return(NULL)
    }
    if (!is(constraint.df, "Primers")) {
        stop("Please input a valid primer data frame.")
    }
    constraints <- names(constraint.settings)
    if (length(constraints) == 0) {
        return(NULL)
    }
    con.idx <- get.constraint.value.idx(constraints, constraint.df)
    con.cols <- lapply(con.idx, function(x) names(constraint.df)[x])
    boundaries <- constraint.settings[constraints]
    constraints <- constraints_to_unit(constraints)
    ######
    plot.df <- melt(constraint.df, id.vars = c("ID"), 
        measure.vars = unlist(con.cols),
        variable.name = "Constraint", 
        value.name = "Value")
    # annotate plot.df with 'real constraint names'
    # Select only fw values for fw primers etc.
    fw.idx <- which(constraint.df$Forward != "")
    rev.idx <- which(constraint.df$Reverse != "")
    # determine whether an entry in plot df concerns forward or reverse primers or any type:
    plot.df$Direction <- ifelse(grepl("_fw", plot.df$Constraint), "Forward", 
                            ifelse(grepl("_rev", plot.df$Constraint), "Reverse", "Overall"))
    plot.df.fw <- plot.df[plot.df$Direction == "Forward" & plot.df$ID %in% constraint.df$ID[fw.idx],]
    plot.df.rev <- plot.df[plot.df$Direction == "Reverse" & plot.df$ID %in% constraint.df$ID[rev.idx],]
    plot.df <- rbind(plot.df.fw, plot.df.rev, plot.df[plot.df$Direction == "Overall",])
    con.names <- unlist(lapply(plot.df$Constraint, function(x) names(con.cols)[sapply(seq_along(con.cols), function(y) any(x %in% con.cols[[y]]))]))
    plot.df$Constraint <- factor(con.names, levels = unique(con.names)[order(unique(con.names))])
    # convert to percentage deviations
    myfun <- function(value, con.name, constraint.settings) {
        con.name <- as.character(con.name)
        setting <- constraint.settings[[unique(con.name)]]
        newValue = value
        if ("min" %in% names(setting) && !"max" %in% names(setting)) {
            newValue[value < setting["min"]] = (value - setting["min"]) / abs(setting["min"])
            newValue[value >= setting["min"]] = 0
        } else if ("max" %in% names(setting) && !"min" %in% names(setting)) {
            newValue[value > setting["max"]] = (value - setting["max"]) / abs(setting["max"])
            newValue[value <= setting["max"]] = 0
        } else if ("min" %in% names(setting) && "max" %in% names(setting)) {
            newValue[value < setting["min"]] = (value - setting["min"]) / abs(setting["min"])
            newValue[value > setting["max"]] = (value - setting["max"]) / abs(setting["max"])
            newValue[value >= setting["min"] & value <= setting["max"]] = 0
        } else {
            warning("Could not transform values to deviations because min/max was missing.")
        }
        return(newValue)
    }
    df <- ddply(plot.df, c("ID", "Constraint", "Value"), here(summarize),            
                Deviation = myfun(substitute(Value), substitute(Constraint), 
                              constraint.settings))
    df$Run <- constraint.df$Run[1]
    return(df)
}
#' Histogram of Efficiencies
#'
#' Plots a histogram of primer efficiencies.
#'
#' @param primer.df Primer data frame.
#' @param opti.constraints List with constraint settings.
#' @return A plot of primer efficiencies.
#' @keywords internal
plot_constraint.histogram.primer.efficiencies <- function(primer.df, opti.constraints) {
    con.cols <- list("primer_efficiency" = "primer_efficiency")
    con.identifier <- "Primer efficiency"
    if (!"primer_efficiency" %in% colnames(primer.df)) {
        return(NULL)
    }
    eff <- strsplit(primer.df$primer_efficiency, split = ",")
    df <- do.call(rbind, lapply(seq_along(eff), function(x) primer.df[rep(x, length(eff[[x]])), 
        ]))
    df$primer_efficiency <- as.numeric(unlist(eff))
    boundaries <- opti.constraints["primer_efficiency"]
    p <- plot_constraint.histogram(df, con.cols, con.identifier, boundaries)
    return(p)
}
#' Histogram of Number of Mismatches.
#'
#' Plots a histogram of mismatches.
#'
#' @param primer.df Primer data frame.
#' @param allowed.mismatches Number of allowed mismatches.
#' @return A plot of the number of primer mismatches.
#' @keywords internal
plot_constraint.histogram.nbr.mismatches <- function(primer.df, allowed.mismatches) {
    if (length(primer.df) == 0 || nrow(primer.df) == 0) {
        return(NULL)
    }
    con.cols <- c("Nbr_of_mismatches_fw", "Nbr_of_mismatches_rev")
    con.cols <- list("allowed_mismatches" = con.cols)
    con.identifier <- "Number of mismatches"
    if (any(!"Nbr_of_mismatches_fw" %in% colnames(primer.df))) {
        return(NULL)
    }
    mm.fw <- strsplit(as.character(primer.df$Nbr_of_mismatches_fw), split = ",")
    mm.rev <- strsplit(as.character(primer.df$Nbr_of_mismatches_rev), split = ",")
    # could keep identifiers for new primer.df
    r.fw <- do.call(rbind, lapply(seq_along(mm.fw), function(x) primer.df[rep(x, 
        length(mm.fw[[x]])), ]))
    r.fw$Nbr_of_mismatches_fw <- as.numeric(unlist(mm.fw))
    r.fw$Nbr_of_mismatches_rev <- rep(NA, nrow(r.fw))
    r.rev <- do.call(rbind, lapply(seq_along(mm.rev), function(x) primer.df[rep(x, 
        length(mm.rev[[x]])), ]))
    r.rev$Nbr_of_mismatches_rev <- as.numeric(unlist(mm.rev))
    r.rev$Nbr_of_mismatches_fw <- rep(NA, nrow(r.rev))
    df <- rbind(r.fw, r.rev)
    boundaries <- list("allowed_mismatches" = allowed.mismatches)
    p <- plot_constraint.histogram(df, con.cols, con.identifier, boundaries)
    return(p)
}
#' Histogram of Constraints.
#'
#' Plots a histogram of constraint values.
#'
#' @param primer.df Primer data frame, not necessarily a \code{Primers} object.
#' @param con.cols Constraint identifiers in \code{primer.df} to plot.
#' @param con.identifier Name of the constraint to plot.
#' @param boundaries List with constraint settings.
#' @param x.limits Interval limiting the extent of the x-axis.
#' @return A constraint histogram plot.
#' @keywords internal
plot_constraint.histogram <- function(primer.df, con.cols, con.identifier, boundaries = NULL, 
    x.limits = NULL) {

    if (length(primer.df) == 0 || length(con.cols) == 0 || length(con.identifier) == 0) {
        return(NULL)
    }
    if (any(unlist(lapply(con.cols, function(x) !x %in% colnames(primer.df))))) {
        return(NULL)
    }
    plot.df <- melt(primer.df, id.vars = c("ID"), 
        measure.vars = unlist(con.cols),
        variable.name = "Constraint", 
        value.name = "Value")
    # Select only fw values for fw primers etc.
    fw.idx <- which(primer.df$Forward != "")
    rev.idx <- which(primer.df$Reverse != "")
    # determine whether an entry in plot df concerns forward or reverse primers or any type:
    plot.df$Direction <- ifelse(grepl("_fw", plot.df$Constraint), "Forward", 
                            ifelse(grepl("_rev", plot.df$Constraint), "Reverse", "Overall"))
    plot.df.fw <- plot.df[plot.df$Direction == "Forward" & plot.df$ID %in% primer.df$ID[fw.idx],]
    plot.df.rev <- plot.df[plot.df$Direction == "Reverse" & plot.df$ID %in% primer.df$ID[rev.idx],]
    plot.df <- rbind(plot.df.fw, plot.df.rev, plot.df[plot.df$Direction == "Overall",])
    con.names <- unlist(lapply(plot.df$Constraint, function(x) names(con.cols)[sapply(seq_along(con.cols), function(y) any(x %in% con.cols[[y]]))]))
    con.levels <- unique(con.names)[order(unique(con.names))]
    constraint.names <- unlist(constraints_to_unit(con.levels, TRUE))
    plot.df$Constraint <- factor(con.names, levels = con.levels,
                                labels = sapply(constraint.names, deparse))
    plot.colors <- c(Forward = brewer.pal(8, "Accent")[5], Reverse = "#0B4D46", Overall = brewer.pal(8, 
        "Accent")[2])
    boundary.data <- NULL # store plot boundaries
    cons <- as.character(levels(plot.df$Constraint))
    for (i in seq_along(cons)) {
        con <- cons[i] 
        con.name <- con.levels[i]
        bound <- boundaries[[con.name]] 
        if (length(bound) == 0) {
            warning(paste("Bound for ", con.name, " not available."))
            next
        }
        cur.bound <- data.frame(Constraint = rep(con, length(bound)),
                                Z = bound)
        boundary.data <- rbind(boundary.data, cur.bound)
    }
    p <- ggplot(plot.df, aes(x = .data[["Value"]], fill = .data[["Direction"]])) +
        xlab("Value") + 
        ylab("Count") + ggtitle("Constraints") + 
        scale_fill_manual(values = plot.colors) + 
        geom_histogram()
    if (length(con.identifier) > 1) {
        # create multiple facets
        p <- p + facet_wrap(as.formula("~Constraint"), scales = "free_x", ncol = 2,
                            labeller = ggplot2::label_parsed)

    } else {
        # change y axis label to con.identifier
        p <- p + xlab(con.identifier[[1]])
    }
    if (length(boundary.data) != 0) {
        p <- p + geom_vline(data = boundary.data, 
                    aes(xintercept = .data[["Z"]]),
                    size = 0.25, colour = "red", linetype = 2)
    }
    if (length(x.limits) != 0) {
        # don't lose bars outside of limits
        p <- p + coord_cartesian(xlim = x.limits)  
    }
    return(p)
}

#' Retrieval of Evaluation Columns.
#'
#' Retrieves the evaluation columns by intersecting
#' the already evaluated constraints in \code{primer.data}
#' as well as the constraints specified in input constraint settings.
#'
#' @param primer.data A list with \code{Primers} objects.
#' @param constraint.settings A list with constraint settings.
#' @return A character vector with EVAL-columns.
#' @keywords internal
get.eval.cols <- function(primer.data, constraint.settings) {
    available.cols <- unique(unlist(lapply(primer.data, function(x) 
                            colnames(x)[grep("^EVAL_", colnames(x))])))
    eval.names <- gsub("^EVAL_", "", available.cols)
    eval.cols <- intersect(eval.names, names(constraint.settings))
    if (length(eval.cols) != 0) {
        eval.cols <- paste0("EVAL_", eval.cols)
    }
    return(eval.cols)
}

#' Primer Set Statistics
#'
#' Creates an overview of all primer set constraint values.
#'
#' @param primer.df Primer data frame.
#' @param mode.directionality Direction of primers.
#' @param lex.seq Template data frame.
#' @return A data frame with statistics.
#' @keywords internal
primer.set.parameter.stats <- function(primer.df, mode.directionality, lex.seq) {
    # compute summary stats of an evaluated primer set check which stats are
    # available
    if (length(primer.df) == 0 || nrow(primer.df) == 0) {
        return(NULL)
    }
    used.measure <- "IQR"  # IQR (inter-quartile range) or CI (confidence interval)
    eval.col.names <- colnames(primer.df)[grep("^EVAL_", colnames(primer.df))]
    eval.cols <- gsub("EVAL_", "", eval.col.names)
    col.idx <- get.constraint.value.idx(eval.cols, primer.df)
    if (length(col.idx) == 0) {
        warning("No properties available to display stats for: ", primer.df$Run[1])
        return(NULL)
    }
    val.name <- lapply(col.idx, function(x) colnames(primer.df)[x])
    # use frontend identifiers for columns:
    names(val.name) <- constraints_to_unit(names(val.name), use.unit = FALSE)
    single.idx <- which(sapply(val.name, function(x) length(x) == 1))
    double.idx <- which(sapply(val.name, function(x) length(x) == 2))
    # select which direction features to show:
    sel.cols <- val.name[single.idx] # always show features for both directions
    # select by directionality
    if (mode.directionality == "fw") {
        fw.cols <- lapply(seq_along(double.idx), function(x) val.name[[double.idx[x]]][1])
        names(fw.cols) <- names(double.idx)
        sel.cols <- c(sel.cols, fw.cols)
    } else if (mode.directionality == "rev") {
        rev.cols <- lapply(seq_along(double.idx), function(x) val.name[[double.idx[x]]][2])
        names(rev.cols) <- names(double.idx)
        sel.cols <- c(sel.cols, rev.cols) 
    } else { # all columns
        sel.cols <- val.name
    }
    sel.cols <- unlist(sel.cols)  # columns where we can extract the values / create statistics
    names(sel.cols) <- gsub(" ", "_", names(sel.cols)) # gaps are ugly in data frame ..
    sel.cols <- sel.cols[sel.cols %in% colnames(primer.df)]
    # remove "Basic_*" columns
    sel.cols <- sel.cols[!grepl("Basic_", sel.cols)]
    nbr.primers <- nrow(primer.df)
    if (used.measure == "CI") {
        # CI not so appropriate here imo
        means <- ddply(asS3(primer.df)[, sel.cols], .(), numcolwise(mean, na.rm = TRUE))[-1]
        sds <- ddply(asS3(primer.df)[, sel.cols], .(), numcolwise(sd))[-1]  # numcolwise leads to problems with read.csv with NA entries when field is not converted to numeric by read.csv function
        na.idx <- which(is.na(sds))
        if (length(na.idx) != 0) {
            sds[na.idx] <- 0
        }
        error <- qnorm(0.975) * sds/sqrt(nbr.primers)
        conf.left <- round(means - error, 2)
        conf.right <- round(means + error, 2)
        entries <- paste("[", conf.left, ",", conf.right, "]", sep = "")
    } else if (used.measure == "IQR") {
        IQR.data <- ddply(asS3(primer.df)[, sel.cols], .(), numcolwise(quantile, na.rm = TRUE))[-1]
        IQR.l <- IQR.data[2, ]  # 1st quantile
        IQR.r <- IQR.data[4, ]  # 3rd quantile
        entries <- paste("[", round(IQR.l, 2), ",", round(IQR.r, 2), "]", sep = "")
    } else {
        stop("Unknown measure for creating primer stats!")
    }
    #num.cols <- sapply(sel.cols, function(x) is.numeric(primer.df[, x]))
    entry.df <- do.call(cbind, as.list(entries))
    colnames(entry.df) <- names(sel.cols)
    if (length(lex.seq) != 0) {
        cvg.info <- get_cvg_ratio(primer.df, lex.seq)
        cvg <- paste0(round(cvg.info, 4) * 100, "%")
        nbr.total <- nrow(lex.seq)
        nbr.covered <- attr(cvg.info, "no_covered")
        cvg <- paste(nbr.covered, " of ", nbr.total, " (", cvg, ")", sep = "")
    } else {
        cvg <- NA
    }
    # other features from strings: cvg constraints (efficiency, annealing), mismatches, binding
    #######
    # binding range:
    #####
    if (any(primer.df$Relative_Forward_Binding_Position_Start_fw != "")) {
        binding.pos.fw <- quantile(c(as.numeric(unlist(strsplit(primer.df$Relative_Forward_Binding_Position_Start_fw, split = ","))), 
                        as.numeric(unlist(strsplit(primer.df$Relative_Forward_Binding_Position_End_fw, split = ",")))))
        binding.pos.fw <- c(binding.pos.fw[2], binding.pos.fw[4]) # IQR
        binding.pos.fw <- paste0("[", paste0(ifelse(binding.pos.fw <= 0, as.character(binding.pos.fw), paste0("+", binding.pos.fw)), collapse = ","), "]")
    } else {
        binding.pos.fw <- NA
    }
    if (any(primer.df$Relative_Reverse_Binding_Position_Start_rev != "")) {
        binding.pos.rev <- quantile(c(as.numeric(unlist(strsplit(primer.df$Relative_Reverse_Binding_Position_End_rev, split = ","))), 
                         as.numeric(unlist(strsplit(primer.df$Relative_Reverse_Binding_Position_Start_rev, split = ",")))))
        binding.pos.rev <- c(binding.pos.rev[2], binding.pos.rev[4]) # IQR
        binding.pos.rev <- paste0("[", paste0(ifelse(binding.pos.rev <= 0, as.character(binding.pos.rev), paste0("+", binding.pos.rev)), collapse = ","), "]")
    } else {
        binding.pos.rev <- NA
    }
    binding.df <- data.frame(Binding_Range_fw = binding.pos.fw, Binding_Range_rev = binding.pos.rev,
                             stringsAsFactors = FALSE)
                            
    string.constraints <- c("primer_efficiency", "annealing_DeltaG", "coverage_model")
    if (mode.directionality == "both") {
        string.constraints <- c("Nbr_of_mismatches_fw", "Nbr_of_mismatches_rev", string.constraints)
    } else if (mode.directionality == "fw") {
        string.constraints <- c("Nbr_of_mismatches_fw", string.constraints)
    } else {
        string.constraints <- c("Nbr_of_mismatches_rev", string.constraints)
    }
    string.constraints <- string.constraints[string.constraints %in% colnames(primer.df)]
    string.res <- vector("list", length(string.constraints))
    for (i in seq_along(string.constraints)) {
        con <- string.constraints[i]
        entries <- string.to.IQR(paste(primer.df[, con], collapse = ","))
        string.res[[i]] <- entries
    }
    string.df <- do.call(cbind, string.res)
    colnames(string.df) <- string.constraints
    all.entries <- data.frame(cbind(Template_Coverage = cvg, 
                                  Nbr_Primers = nbr.primers, 
                                  binding.df,
                                  string.df, 
                                  entry.df, stringsAsFactors=FALSE), 
                    stringsAsFactors = FALSE)
    # select only non-NA columns
    sel.cols <- apply(all.entries, 2, function(x) any(!is.na(x)))
    all.entries  <- all.entries[, sel.cols]
    #print("set stats:")
    #print(all.entries)
    return(all.entries)
}
#' Conversion of Comma-Separated String to IQR String
#'
#' @param string.values A vector of comma-separated numeric strings.
#' @return The IQR corresponding to the input string.
#' @keywords internal
string.to.IQR <- function(string.values) {
    quant <- lapply(strsplit(string.values, split = ","), function(x) quantile(as.numeric(x), na.rm = TRUE))
    IQR <- lapply(quant, function(x) c(x[2], x[4]))
    entries <- unlist(lapply(IQR, function(x) if (all(is.na(x))) {""} else {paste0("[", paste0(round(x, 2), collapse = ","), "]")}))
    return(entries)
}
#' Plot Dimer DeltaG
#'
#' Plot the distribution of dimerization free energies.
#'
#' @param dimer.data Data frame with dimerization information.
#' @param deltaG.cutoff Free energy cutoff for dimerization.
#' @return A plot of dimerization free energies.
#' @keywords internal
plot.dimer.dist <- function(dimer.data, deltaG.cutoff) {
    if (length(dimer.data) == 0 || nrow(dimer.data) == 0) {
        return(NULL)
    }
    dimer.df <- dimer.data
    colors <- brewer.pal(8, "Pastel1")[c(1, 2)]
    dimer.df$Decision <- "No Dimer"
    dimer.df$Decision[dimer.data$DeltaG < deltaG.cutoff] <- "Dimer"
    dimer.df$Decision <- factor(dimer.df$Decision)
    if (all(dimer.df$Decision == "Dimer")) {
        colors <- colors[1]
    } else if (all(dimer.df$Decision == "No Dimer")) {
        colors <- colors[2]
    }
    p <- ggplot(dimer.df, aes(x = .data[["DeltaG"]])) + geom_histogram(aes(fill = .data[["Decision"]]), 
        colour = "grey") + geom_vline(xintercept = deltaG.cutoff, colour = "red") + 
        ggtitle(expression(paste("Dimerization ", Delta, " G values", 
            sep = ""))) + xlab(expression(paste(Delta, " G", sep = ""))) + scale_fill_manual(values = colors) + 
        theme(axis.text.x = element_text(hjust = 1, vjust = 0.5))
    return(p) 
}
#' Dimerization Table.
#'
#' Summarizes how often individual primers dimerize according to the \code{deltaG.cutoff}.
#'
#' @param dimer.data Data frame with dimerization data.
#' @param deltaG.cutoff Free energy cutoff for dimerization.
#' @param dimer.type String identifying whether \code{dimer.data} refers to cross-dimers or self-dimers?
#' @return Data frame with dimer counts.
#' @keywords internal
dimerization.table <- function(dimer.data, deltaG.cutoff, 
            dimer.type = c("Self-Dimerization", "Cross-Dimerization")) {
    # no need to compute if there's no data and no need to plot for self-dimerization as
    # well (maximum count is 1)
    if (length(dimer.type) == 0) {
        stop("Please provide a dimerization type.")
    }
    dimer.type <- match.arg(dimer.type)
    if (length(dimer.data) == 0 || dimer.type == "Self-Dimerization") {
        return(NULL)
    }
    ID.col <- c("Primer_1", "Primer_2") # only for cross-dimers ..
    dimer.idx <- which(dimer.data$DeltaG < deltaG.cutoff)
    dimer.IDs <- unlist(dimer.data[dimer.idx, ID.col])
    dimer.counts <- data.frame(table(factor(dimer.IDs)))
    if (nrow(dimer.counts) == 0) {
        return(data.frame())
    }
    colnames(dimer.counts)[1] <- "ID"
    dimer.mapping <- lapply(dimer.counts[, 1], function(x) unlist(lapply(ID.col, 
        function(y) which(dimer.data[, y] == x))))
    energy <- sapply(seq_along(dimer.mapping), function(x) mean(dimer.data[dimer.mapping[[x]], "DeltaG"]))
    dimer.counts$Mean_Delta_G <- energy
    o <- order(dimer.counts[, "Freq"], decreasing = TRUE)
    dimer.counts <- dimer.counts[o, ]
    colnames(dimer.counts) <- c("Primer", "Dimer Count", "Mean Free Energy")
    return(dimer.counts)
}

#' @rdname Plots
#' @name Plots
#' @aliases plot_constraint
#' @return \code{plot_constraint} returns a plot showing the distribution of primer properties.
#' @export
#' @include primers.R settings.R
#' @examples
#' 
#' # Plot histogram of constraints for a single primer set
#' data(Ippolito)
#' p <- plot_constraint(primer.df, settings, 
#'                      active.constraints = c("gc_clamp", "gc_ratio"))
#' # Compare constraints across multiple primer sets
#' data(Comparison)
#' p.cmp <- plot_constraint(primer.data[1:3], settings, 
#'                          active.constraints = c("gc_clamp", "gc_ratio"))
setGeneric("plot_constraint", 
    function(primers, settings, active.constraints = names(constraints(settings)), ...) {
        if (is(settings, "DesignSettings")) {
            # use only the constraints that are defined in the settings
            active.constraints <- active.constraints[active.constraints %in% names(constraints(settings))]
        } else {
            # no settings available -> trust that the input is ok for now
        }
        standardGeneric("plot_constraint")
})
#' Histogram of Constraint Values.
#'
#' Plots a histogram of constraint values.
#'
#' @param primers An evaluated object of class \code{Primers}.
#' @param settings A \code{DesignSettings} object containing the
#' settings for the constraints to be plotted. 
#' @param active.constraints Identifiers of constraints to be plotted.
#' provided settings are used to visualize the desired ranges of constraints.
#' If \code{active.constraints} is not provided, the plotting method
#' will automatically try to plot all constraints defined in \code{settings}.
#' @return A histogram of constraint values for the properties specified by
#' \code{constraints}.
#' @keywords internal
setMethod("plot_constraint", 
    signature(primers = "Primers"), 
    function(primers, settings, active.constraints) {

    if (length(primers) == 0 || nrow(primers) == 0 || length(active.constraints) == 0) {
        return(NULL)
    }
    if (!is(primers, "Primers")) {
        stop("Please input a 'Primers' object for 'primers'.")
    }
    con.idx <- get.constraint.value.idx(active.constraints, primers)
    con.cols <- lapply(con.idx, function(x) names(primers)[x])
    if (length(settings) != 0 && is(settings, "DesignSettings")) {
        boundaries <- constraints(settings)[active.constraints]
    } else {
        boundaries <- NULL
    }
    constraints <- constraints_to_unit(active.constraints)
    p <- plot_constraint.histogram(primers, con.cols, constraints, boundaries)
    return(p)
})

#' Boxplot for Comparing Constraints.
#'
#' Creates a boxplot visualizing the physicochemical properties
#' of multiple primer sets. 
#'
#' @param primers List with evaluated objects of class \code{Primers}. 
#' Each list element corresponds to a single primer set.
#' @param settings A \code{DesignSettings} object containing
#' the constraints to be plotted.
#' @param active.constraints The names of the constraints to be plotted.
#' @param constraint.settings List with settings for each constraint.
#' @param highlight.set Identifiers of primer sets to be highlighted.
#' @param nfacets A numeric providing the number of facet columns to show.
#' By default \code{nfacets} is \code{NULL} such that the number of 
#' facet columns is chosen automatically.
#' @return Boxplot comparing the values of the properties specified 
#' by \code{constraints}.
#' @keywords internal
setMethod("plot_constraint", 
    signature(primers = "list"), 
    function(primers, settings, active.constraints,
             highlight.set = NULL, nfacets = NULL) {
    if (length(primers) == 0 || length(active.constraints) == 0) {
        return(NULL)
    }
    #print("nfacets plot_constraint()")
    #print(nfacets)
    # check type of primers
    primer.classes <- sapply(primers, function(x) class(x))
    if (any(primer.classes != "Primers")) {
        stop("Please check whether all supplied primers have the right type.")
    }
    if (length(settings) != 0 && is(settings, "DesignSettings")) {
        boundaries <- constraints(settings)[active.constraints]
    } else {
        boundaries <- NULL
    }
    # consider all constraints that are contained at least once in any of the primer data
    con.cols <- lapply(seq_along(primers), function(x) {
        idx <- get.constraint.value.idx(active.constraints, primers[[x]])
        lapply(idx, function(i) colnames(primers[[x]])[i])
    })
    con.cols <- unlist(con.cols, recursive = FALSE)
    con.cols <- con.cols[!duplicated(names(con.cols))]
    constraints <- constraints_to_unit(active.constraints)
    p <- plot_primer.comparison.box(primers, constraints, 
            con.cols, boundaries, highlight.set = highlight.set,
            nfacets = nfacets)
    return(p)
})

#' @rdname Plots
#' @name Plots
#' @aliases plot_constraint_fulfillment
#' @return \code{plot_constraint_fulfillment} returns a plot indicating the constraints that are fulfilled by the input primers.
#' @export
#' @include primers.R settings.R
#' @examples
#' 
#' # Plot fulfillment for a single primer set:
#' data(Ippolito)
#' p <- plot_constraint_fulfillment(primer.df, settings)
#' # Plot fulfillment for multiple primer sets:
#' data(Comparison)
#' p.cmp <- plot_constraint_fulfillment(primer.data[1:5], settings)
setGeneric("plot_constraint_fulfillment", 
    function(primers, settings, active.constraints = names(constraints(settings)), 
             plot.p.vals = FALSE, ...) {
        standardGeneric("plot_constraint_fulfillment")
})
#' Overview of Constraint Fulfillment.
#'
#' Plots an overview of which primers passed the filtering constraints and which
#' primers did not.
#'
#' @param primers A \code{Primers} object.
#' @param settings A \code{DesignSettings} object.
#' @param active.constraints The identifiers of constarints to be plotted for fulfillment.
#' @param plot.p.vals Show p-value from Fisher's exact test 
#' for the significance
#' of primer constraint fulfillment in comparison to reference
#' primer sets.
#' @return A data frame with statistics on fulfilled constraints.
#' @keywords internal
setMethod("plot_constraint_fulfillment", 
    signature(primers = "Primers"),
    function(primers, settings, active.constraints, plot.p.vals = TRUE) {
    if (length(primers) == 0 || nrow(primers) == 0) {
        return(NULL)
    }
    if (!is(primers, "Primers")) {
        stop("Please input a valid primer data frame.")
    }
    if (!is(settings, "DesignSettings")) {
        stop("Please provide a 'DesignSettings' object for 'settings'.")
    }
    constraint.settings <- constraints(settings)[names(constraints(settings)) %in% active.constraints]
    constraint.df <- asS3(primers)
    # use available EVAL column data
    eval.cols <- get.eval.cols(list(constraint.df), constraint.settings)
    if (length(eval.cols) == 0) {
        # nothing to plot
        return(NULL)
    }
    active.constraints <- names(constraint.settings)
    # re-evaluate primer.data and use constraint.settings constraints
    mode.directionality <- get.analysis.mode(constraint.df)
    eval.df <- eval.constraints(constraint.df, 
                        constraint.settings, active.constraints,
                        mode.directionality, constraint.df)
    constraint.df <- update.constraint.values(constraint.df, eval.df)
    title <- "Constraint fulfillment"
    if (plot.p.vals) {
        # compute p-vals
        p.data <- primer_significance(constraint.df, active.constraints = active.constraints)
        if (length(p.data) != 0) {
            p.val <- as.numeric(p.data)
            title <- paste0(title, " (p-value: ", format(p.val, scientific = TRUE, digits = 2), ")")
        } 
    }
    color.set <- c("#3e6ebc", "#c62f1d")
    names(color.set) <- c("Passed", "Failed")
    eval.df <- data.frame(constraint.df[, eval.cols, drop = FALSE])
    if (all(do.call(c, eval.df), na.rm = TRUE)) {
        # all TRUE?
        colors <- color.set[1]
    } else if (!any(do.call(c, eval.df), na.rm = TRUE)) {
        # all FALSE?
        colors <- color.set[2]
    } else {
        colors <- color.set
    }
    passed.name <- "Passed"
    failed.name <- "Failed"
    eval.df[eval.df == TRUE] <- passed.name  # blue
    eval.df[eval.df == FALSE] <- failed.name  # grey
    eval.df <- cbind(Identifier = constraint.df$Identifier, ID = constraint.df$ID, 
                    eval.df, stringsAsFactors = TRUE)
    colnames(eval.df) <- gsub("EVAL_", "", colnames(eval.df))
    eval.m <- melt(eval.df, c("ID", "Identifier"), variable.name = "Constraint", 
        value.name = "Outcome")
    # sort columns by name
    levels <- unique(eval.m$Constraint)[order(as.character(unique(eval.m$Constraint)))]
    eval.m$Constraint <- factor(eval.m$Constraint, 
                         levels = levels)
    constraint.names <- unlist(constraints_to_unit(levels(eval.m$Constraint), FALSE))
    xlab <- "Constraint"
    ylab <- "Primer"
    nbr.constraints <- length(unique(eval.m$Constraint))
    eval.m$ID <- factor(unique(eval.m$ID))
    levels(eval.m$ID) = abbreviate(levels(eval.m$ID), getOption("openPrimeR.plot_abbrev"))
    p <- ggplot(eval.m, aes(x = .data[["Constraint"]], y = .data[["ID"]])) + geom_tile(aes(fill = .data[["Outcome"]]),
        colour = "black") + scale_fill_manual(values = colors) +
        theme(axis.text.x = element_text(angle = 60,
                hjust = 1)) +
        ggtitle(title) + xlab(xlab) + ylab(ylab) + 
        theme(legend.title = element_text(face = "bold")) +
        scale_x_discrete(labels = constraint.names) +
        scale_y_discrete(limits = rev(levels(eval.m$ID))) # top primer should be the first entry in the plot
    return(p)
})
#' Comparison of Evaluation Results.
#'
#' Plots the percentage of primers fulfilling the specified
#' constraints for multiple primer sets.
#'
#' @param primers A list of \code{Primers} objects.
#' @param settings A \code{DesignSettings} object.
#' @param active.constraints The identifiers of constarints to be plotted for fulfillment.
#' @param plot.p.vals Whether p-values from Fisher's exact test
#' should be annotated for every primer set.
#' @param ncol The number of columns for facet wrap.
#' @param highlight.set Identifiers of primer sets to be highlighted.
#' @return Plot indicating the ratio of primers
#' fulfilling the constraints specified in \code{constraint.settings}
#' for each primer set in \code{primers}. 
#' @keywords internal
setMethod("plot_constraint_fulfillment", 
    signature(primers = "list"),
    function(primers, settings, active.constraints, 
             plot.p.vals = FALSE, ncol = 2, highlight.set = NULL) {

    constraint.settings <- constraints(settings)[names(constraints(settings)) %in% active.constraints]
    new.df <- prepare.constraint.plot(primers, constraint.settings, plot.p.vals) 
    if (length(new.df) == 0) {
        # no constraint to plot
        return(NULL)
    }
    if (length(highlight.set) != 0) {
        # check whether highlight set is specified correctly
        m <- match(highlight.set, new.df$Run)
        na.idx <- which(is.na(m))
        if (length(na.idx) != 0) {
            msg <- paste("Highlight set not available in data:",
                paste(highlight.set[na.idx], collapse = ","))
            warning(msg)
            highlight.set <- highlight.set[!is.na(m)]
        }
    }
    constraint.names <- unlist(constraints_to_unit(levels(new.df$Constraint), FALSE))
    pal <- getOption("openPrimeR.plot_colors")["Constraint"] # the RColorBrewer palette to use
    colors <- colorRampPalette(brewer.pal(8, pal))(length(unique(new.df[, "Constraint"])))
    # plot
    p <- ggplot(new.df) + 
            geom_bar(mapping = aes(x = .data[["Constraint"]], y = .data[["Ratio"]], 
            fill = .data[["Constraint"]]), position = "dodge", stat = "identity") +
            scale_y_continuous(labels = scales::percent) + ylab("Fulfillment ratio") + 
            xlab("Constraint") + ggtitle("Evaluation of constraints") + 
            theme(
            axis.text.x = element_text(angle = 90, hjust = 1)) +
            facet_wrap(~Run, ncol = ncol) + 
            guides(fill = "none") + # remove legend: is redundant here
            scale_x_discrete(labels = constraint.names) + 
           scale_fill_manual(labels = constraint.names, values = colors)

    if (length(highlight.set) != 0) {
        # highlight selected sets (facet part can't be colored!!)
        highlights <- data.frame(Run = highlight.set)
        p <- p + geom_rect(data=highlights,aes(xmin=-Inf, xmax=Inf, ymin=-Inf, ymax=Inf), fill='red', alpha=0.1)
    }
    return(p)
})
#' @rdname Plots
#' @name Plots
#' @aliases plot_cvg_constraints
#' @return \code{plot_cvg_constraints} returns a plot showing the distribution of the coverage constraint values.
#' @export
#' @include primers.R settings.R
#' @examples
#' 
#' # Plot coverage constraints of a single primer set
#' data(Ippolito)
#' p <- plot_cvg_constraints(primer.df, settings)
#' # Plot coverage constraints for mulitple primer sets
#' data(Comparison)
#' p.cmp <- plot_cvg_constraints(primer.data[1:2], settings)
setGeneric("plot_cvg_constraints", 
    function(primers, settings, active.constraints = names(cvg_constraints(settings)), ...) {
        if (!is(settings, "DesignSettings")) {
            stop("'settings' should be of class 'DesignSettings'.")
        }
        standardGeneric("plot_cvg_constraints")
})

#' Histogram of Coverage Constraints.
#'
#' Plots a histogram of coverage constraint values.
#'
#' @param primers A \code{Primers} object.
#' @param settings A \code{DesignSettings} object.
#' @param active.constraints Names of coverage constraints to be plotted.
#' @return A plot of coverage constraints.
#' @keywords internal
setMethod("plot_cvg_constraints", 
    signature(primers = "Primers"),
    function(primers, settings, active.constraints) {

    if (!is(primers, "Primers")) {
        stop("'primers' should be of class 'Primers'.")
    }
    if (nrow(primers) == 0) {
        # nothing to plot
        return(NULL)
    }
    cvg.constraints <- cvg_constraints(settings)
    # select only the constraints from the settings that should be plotted
    con.columns <- names(cvg.constraints)[names(cvg.constraints) %in% active.constraints]
    # select only the available constraints
    con.columns <- con.columns[con.columns %in% colnames(primers)]
    if (length(con.columns) == 0) {
        warning("No coverage constraint data available for plotting.")
        return(NULL)
    }
    con.identifier <- constraints_to_unit(con.columns, use.unit = FALSE)
    boundaries <- cvg.constraints[con.columns]
    # prepare every cvg constraint for plotting:
    dfs <- vector("list", length(con.columns))
    for (i in seq_along(con.columns)) {
        con.cols <- con.columns[i]
        res <- strsplit(primers[, con.cols] , split = ",")
        res <- data.frame(as.numeric(unlist(res)))
        colnames(res) <- con.cols
        dfs[[i]] <- res
    }
    # output: data frame with number of rows <=> nbr of total cvg events
    out.df <- asS3(do.call(rbind, lapply(seq_len(nrow(primers)), function(x) primers[rep(x, primers$primer_coverage[x])])))
    con.df <- do.call(cbind, dfs)
    out.df[, con.columns] <- con.df
    out.con.columns <- as.list(con.columns)
    names(out.con.columns) <- con.columns
    p <- plot_constraint.histogram(out.df, out.con.columns, con.identifier, boundaries)
    return(p)
})
#' Plot for Comparing Primer Coverage Constraints.
#'
#' @param primers List with objects of class \code{Primers}. 
#' @param settings A \code{DesignSettings} object.
#' @param active.constraints Names of the coverage constraints to be plotted.
#' @param highlight.set Primer sets to highlight in the plot.
#' @return Plot of primer coverage constraints for multiple sets.
#' @keywords internal
setMethod("plot_cvg_constraints", 
    signature(primers = "list"),
    function(primers, settings, active.constraints, highlight.set = NULL) {

    cvg.constraints <- cvg_constraints(settings)
    # select only the active constraints
    con.columns <- names(cvg.constraints)[names(cvg.constraints) %in% active.constraints]
    if (length(con.columns) == 0) {
        warning("No coverage constraint data available for plotting.")
        return(NULL)
    }
    con.identifier <- constraints_to_unit(con.columns, use.unit = FALSE)
    boundaries <- cvg.constraints[con.columns]
    new.data <- primers
    # transform string columns to numeric by replication:
    for (i in seq_along(primers)) {
        primer.df <- primers[[i]]
        sel.cols <- con.columns[con.columns %in% colnames(primer.df)]
        missing.cols <- con.columns[!con.columns %in% colnames(primer.df)]
        df <- do.call(rbind, lapply(seq_len(nrow(primer.df)), function(x) asS3(primer.df[rep(x, primer.df$primer_coverage[x]), ])))
        if (length(df) != 0 && nrow(df) != 0) {
            # constraint is present in data set
            for (j in (seq_along(sel.cols))) {
                col <- sel.cols[j]
                val <- as.numeric(unlist(strsplit(as.character(primer.df[, col]), split = ",")))
                df[, col] <- val
            }
        } else {
            # constraint is not present
            df <- data.frame()
        }
        df[, missing.cols] <- NA
        new.data[[i]] <- df
    }
    out.con.columns <- as.list(con.columns)
    names(out.con.columns) <- con.columns
    p <- plot_primer.comparison.box(new.data, con.identifier, out.con.columns, 
        boundaries, show.points = FALSE, highlight.set = highlight.set)
    return(p)
})

#' @rdname Plots
#' @name Plots
#' @aliases plot_constraint_deviation
#' @details
#' The deviations for \code{plot_constraint_deviation} are computed in the following way. Let the
#' minimum and maximum allowed constraint values be given by
#' the interval \eqn{[s, e]} and the observed value be \eqn{p}. Then,
#' if \eqn{p < s}, we output \eqn{-p/|s|}, if \eqn{p > e} we output \eqn{p/|e|},
#' and otherwise, i.e. if \eqn{s <= p <= e}, we output 0.
#'
#' @return \code{plot_constraint_deviation} returns a plot showing the deviations of the primer properties from the target constraints.
#' @export
#' @include primers.R settings.R
#' @examples
#' 
#' # Deviations for a single primer set
#' data(Ippolito)
#' p.dev <- plot_constraint_deviation(primer.df, settings)
#' # Deviations for multiple primer sets
#' data(Comparison)
#' p.dev.cmp <- plot_constraint_deviation(primer.data, settings)
setGeneric("plot_constraint_deviation", 
    function(primers, settings, active.constraints = names(constraints(settings)), ...) {
        if (is(settings, "DesignSettings")) {
            # use only the constraints that are defined in the settings
            active.constraints <- active.constraints[active.constraints %in% names(constraints(settings))]
        } else {
            # no settings available -> trust that the input is ok for now
        }
        standardGeneric("plot_constraint_deviation")
})

#' Plot of Constraint Deviations for a Single Primer Set.
#'
#' Plots a box plot of deviations of 
#' primer properties from the target ranges.
#'
#' @param primers An evaluated object of class \code{Primers}.
#' @param settings A \code{DesignSettings} object
#' containing the target ranges for the primer properties.
#' @param active.constraints Constraint identifiers to be plotted.
#' @return A boxplot of deviations
#' @keywords internal
setMethod("plot_constraint_deviation", 
    signature(primers = "Primers"),
    function(primers, settings, active.constraints) {
        constraint.settings <- constraints(settings)
        constraint.settings <- constraint.settings[names(constraint.settings) %in% active.constraints]
        df <- get_constraint_deviation_data(primers, constraint.settings)
        if (length(df) == 0) {
            return(NULL)
        }
        # modify constraint names 
        constraint.names <- unlist(constraints_to_unit(levels(df$Constraint), TRUE))
        title <- "Constraint deviations"
        abs.total.deviation <- mean(abs(df$Deviation)) * 100
        abs.total.deviation <- paste0("mean |deviation| = ", 
                                round(abs.total.deviation, 0), "%")
        title <- paste0(title, " (", abs.total.deviation, ")")
        # set up colors
        pal <- getOption("openPrimeR.plot_colors")["Constraint"] # the RColorBrewer palette to use
        colors <- colorRampPalette(brewer.pal(8, pal))(length(unique(df$Constraint)))
        p <- ggplot(df, aes(x = .data[["Constraint"]], y = .data[["Deviation"]], colour = .data[["Constraint"]])) + 
            theme(axis.text.x = element_text(angle = 60, hjust = 1)) +
            ggtitle(title) +
            ylab("Deviation from target") +
            geom_boxplot(outlier.shape = NA) +
            geom_point(alpha = 0.4, position = position_jitter(width = 0.25, 
                height = 0)) + 
            scale_y_continuous(labels = scales::percent) +
            scale_x_discrete(labels = constraint.names) +
            guides(colour = "none") + 
            scale_colour_manual(values = colors)
        return(p)
})
#' Plot of Constraint Deviations for Multiple Primer Sets.
#'
#' Plots a box plot of the absolute mean deviation of each primer 
#' for comparing multiple primer sets.
#'
#' @param constraint.df An evaluated object of class \code{Primers}.
#' @param settings A \code{DesignSettings} object
#' containing the target ranges for the primer properties.
#' @param active.constraints Constraint identifiers to be plotted.
#' @param deviation.per.primer Whether to show the deviation per primer or per constraint.
#' @return A boxplot of deviations
#' @keywords internal
setMethod("plot_constraint_deviation", 
    signature(primers = "list"),
    function(primers, settings, active.constraints, deviation.per.primer = FALSE) {
        # check class of primer.data entries
        c <- sapply(primers, class)
        if (!all(c == "Primers")) {
            stop("Please ensure that all 'primers' objects are of class 'Primers'.")
        }
        # select active.constraints
        constraint.settings <- constraints(settings)
        constraint.settings <- constraint.settings[names(constraint.settings) %in% active.constraints]
        if (length(constraint.settings) == 0) {
            warning("No active constraint settings available.")
            return(NULL)
        }
        plot.data <- vector("list", length(primers))
        for (i in seq_along(primers)) {
            primer.df <- primers[[i]]
            df <- get_constraint_deviation_data(primer.df, constraint.settings)
            plot.data[[i]] <- df
        }
        plot.data <- do.call(rbind, plot.data)
        if (length(plot.data) == 0) {
            warning("No constraint data available for plotting.") 
            return(NULL)
        }
        # transform to mean absolute deviation 
        ylab <- "Absolute deviation"
        if (deviation.per.primer) {
            # show deviation per primer across all constraints
            plot.data <- ddply(plot.data, c("ID", "Run"), summarize, Mean_Deviation = mean(abs(substitute(Deviation))))
            col.id <- "Run"
            title <- "Deviations per primer and set"
        } else {
            # show deviation per considered constraint across all primers
            plot.data <- ddply(plot.data, c("Constraint", "Run"), summarize, Mean_Deviation = abs(substitute(Deviation)))
            col.id <- "Constraint"
            title <- "Deviations per constraint and set"
        }
        # set up colors
        pal <- getOption("openPrimeR.plot_colors")[col.id] # the RColorBrewer palette to use
        colors <- colorRampPalette(brewer.pal(8, pal))(length(unique(df[, col.id])))
        p <- ggplot(plot.data, aes(x = .data[["Run"]], y = .data[["Mean_Deviation"]], colour = .data[[col.id]])) + 
            theme(axis.text.x = element_text(angle = 60, hjust = 1)) +
            ggtitle(title) +
            ylab(ylab) + 
            # don't show individual points too large...
            geom_boxplot(outlier.size = 0.75) + 
            scale_y_continuous(labels = scales::percent)

        if (deviation.per.primer) {
            # show deviation per primer -> don't show legend for each primer
            p <- p +
                guides(colour = "none") 
        } else {
            constraint.names <- constraints_to_unit(levels(plot.data$Constraint), FALSE)
            # change names of constraints
            p <- p + 
                scale_colour_manual(labels = constraint.names, values = colors)
        }
        return(p)
})

#########
# Create the feature matrix for supervised learning using the experimentally determined coverage
#######

cellColor <- function(style) {
    fg  <- style$getFillForegroundXSSFColor()
    #fg <- style$getFillForegroundColor() # TODO wrong
    rgb <- tryCatch(fg$getRgb(), error = function(e) NULL)
    rgb <- paste(rgb, collapse = "")
    return(rgb)
}
get.ref.df <- function(color.df) {
    library(stringr)
    cnam <- colnames(color.df)
    counts <- str_count(cnam, "\\.")
    idx.1 <- which(counts == 1)
    idx.2 <- which(counts == 2)
    idx.3 <- which(counts == 3)
    cnam[idx.1] <- sub("\\.", "-", cnam[idx.1])
    cnam[idx.2] <- sub("\\.", "*", sub("\\.", "-", cnam[idx.2]))
    cnam[idx.3] <- sub("\\.", "*", sub("\\.", "-", sub("\\.", "-", cnam[idx.3])))
    m <- match(cnam, template.df$ID)
    # manual check:
    cbind(cnam, template.df$ID[m])
    color.df <- data.frame(color.df[,1], color.df[, !is.na(m)])
    colnames(color.df) <- c("Primer", cnam[!is.na(m)])
    return(color.df)
}
get.color.df <- function(xls.file, sheet.idx, nbr.rows) {
###############
# NEED TO RUN THIS ON WINDOWS!!!
###############
	library(xlsx)
	wb <- loadWorkbook(xls.file)
	sheet1 <- getSheets(wb)[[sheet.idx]] # first single primer evaluation, 18
	# get all rows
	rows  <- getRows(sheet1, seq_len(nbr.rows)) # if rows are not specified, too many are extracted
	cells <- getCells(rows)
	nbr.cols <- as.numeric(tail(strsplit(names(cells), split = "\\."), n =1)[[1]][2])
	styles <- vector("list", length(rows))
	for (i in seq_along(rows)) {
		cell.id <- paste0(i, ".", 1:nbr.cols)
		row.data <- cells[cell.id]
		cur.styles <- sapply(unlist(row.data), getCellStyle)
		styles[[i]] <- cur.styles
	}
	fill.colors <- lapply(styles, function(x) sapply(unlist(x), cellColor)) # if this is not the same length, add Identifier column leftmost
	fill.matrix <- do.call(rbind, fill.colors[2:length(fill.colors)]) # ignore header row
	col.df <- readColumns(sheet1, 1, nbr.cols, 1, nrow(fill.matrix) + 1) # doesn't read cell styles, nrow+1 to load header row
	cnames <- colnames(col.df)
	col.df[, 2:ncol(col.df)] <- fill.matrix[, 2:ncol(fill.matrix)] # first column has primer identifers -> don't overwrite!
	colnames(col.df) <- cnames
	return(col.df)
}
get_ref_data <- function(cvg.folders) {
    #############################
    consensus.fun <- function(x) {
        count1 <- length(which(x == 1))
        count0 <- length(which(x == 0))
        if (count1 > count0) {
            return(1)
        } else {
            return(0)
        }
    }
    agreement.rate <- function(x) {
        count1 <- length(which(x == 1))
        count0 <- length(which(x == 0))
        majority <- ifelse(count1 > count0, 1, 0)
        agreement.rate <- length(which(x == majority)) /length(x)
        return(agreement.rate)
    }
    library(plyr)
    library(reshape2)
    library(ggplot2)
    ignore.cols <- c("Evaluator", "experiment")
    ref.calls <- vector("list", length(cvg.folders))
    for (i in seq_along(cvg.folders)) { # primer sets
        csvs <- list.files(cvg.folders[i], full.names = TRUE)
        set.name <- basename(cvg.folders[i])
        ref.results <- vector("list", length(csvs))
        for (j in seq_along(csvs)) { # reviewer results
            csv <- read.csv(csvs[j], check.names = FALSE)
            csv.j <- csv[, !colnames(csv) %in% ignore.cols]
            # determine consensus across individual experiments
            res <- ddply(csv.j, c("Primer"), numcolwise(consensus.fun))
            res <- cbind("Evaluator" = unique(csv$Evaluator), res)
            ref.results[[j]] <- res
        }
        cvg.matrix <- do.call(rbind, ref.results)
        cvg.matrix.j <- cvg.matrix[, !colnames(cvg.matrix) %in% ignore.cols]
        # compute agreement rate per primer template for all the reviewers from this
        agree.df <- ddply(cvg.matrix.j, c("Primer"), numcolwise(agreement.rate))
        plot.df <- melt(agree.df, variable.name = "Template", value.name = "Agreement")
        # plot 
        p <- ggplot() + geom_boxplot(data = plot.df, aes(x = Primer, y = Agreement))
        ggsave(paste0("agreement_", set.name, ".png"), p)
        ref.summary <- ddply(cvg.matrix.j, c("Primer"), numcolwise(consensus.fun))
        ref.calls[[i]] <- ref.summary
    }
    # gather results for Tiller and openPrimeR in a data frame
    result.df <- do.call(rbind, ref.calls)
    # convert to machine-readable format
    res <- melt(result.df, variable.name = "Template", value.name = "Experimental_Coverage")
    res$Experimental_Coverage <- factor(ifelse(res$Experimental_Coverage == 1, "Amplified", "Unamplified"), levels = c("Unamplified", "Amplified"))
    return(res)
}
get_tool_primer_data <- function(primer.location, template.df, cur.settings) {
    # load primers
    primer.df <- openPrimeR::read_primers(primer.location, "_fw", "_rev")
    primer.df$ID <- factor(gsub("_fw", "", as.character(primer.df$ID)))
    primer.df <- openPrimeR::check_constraints(primer.df, template.df, cur.settings)
    template.df <- openPrimeR::update_template_cvg(template.df, primer.df)
    out <- list("Primers" = primer.df, "Templates" = template.df)
    return(out)
}
get_ref_data_old <- function(xls.file) {
    message("This function should be run on Windows for correct xls parsing ...")
    # turn color xls to labeled R data frame with colors indicating amplification (red: no, green: yes)
    tiller.ref.df <- get.color.df(xls.file, 1,13)
    tiller.ref.df <- get.ref.df(tiller.ref.df)
    open.ref.df <- get.color.df(xls.file, 2, 49)
    open.ref.df <- get.ref.df(open.ref.df)
    ref.data <- list(tiller.ref.df, open.ref.df)
    names(ref.data) <- c("Tiller2008", "openPrimeR")
    # compute tool cvg for tiller and openPrimeR
    primer.file.t <- system.file("extdata", "IMGT_data", "primers", 
                               "IGHV", "Tiller2008_1st.fasta",
                               package = "openPrimeR")
    primer.file.o <- system.file("extdata", "IMGT_data", "primers", 
                               "IGHV", "openPrimeR2017.fasta",
                               package = "openPrimeR")
    primer.files <- c(primer.file.t, primer.file.o)
    for (i in seq_along(primer.files)) {
        primer.df <- read_primers(primer.files[[i]])
        ref.data[[i]]$Primer <- unlist(lapply(primer.df$ID, function(x) rep(x, 3))) # annotate primers with right IDs (order MUST be the same as the provided 'primer.df' for correct cycling!) -> 3 because of triplicates
    }
    return(ref.data)
}
get_learning_matrix <- function(ref.data, template.df, tiller.location, open.location) {
    # primer set to use for calibration:
    mode.directionality <- "fw"
    # create tool data for Tiller / openPrimeR:
    filename <- system.file("extdata", "settings", 
                "C_Taq_PCR_high_stringency.xml", package = "openPrimeR")
    settings <- openPrimeR::read_settings(filename)
    # prevent other binding events:
    conOptions(settings)$allowed_other_binding_ratio <- 0.0
    # allow input nbr of mismatches
    conOptions(settings)$allowed_mismatches <- 12
    # require only intersection with target region
    conOptions(settings)$allowed_region_definition <- "any"
    # compute data
    # activate all 'interesting' cvg constraints as features, but don't filter! (set boundaries high)
    cvg_constraints(settings) <- list("annealing_DeltaG" = c("max" = 0), "primer_efficiency" = c("max" = 1),
                                      "terminal_mismatch_pos" = c("min" = 0))
    tiller.settings <- settings
    PCR(tiller.settings)$annealing_temp <- 57
    tiller.data <- get_tool_primer_data(tiller.location, template.df, tiller.settings)
    open.settings <- settings
    PCR(open.settings)$annealing_temp <- 55
    open.data <- get_tool_primer_data(open.location, template.df, open.settings)
    tool.data <- list("Tiller" = tiller.data, "openPrimeR" = open.data)
    feature.data <- vector("list", length(tool.data))
    for (i in seq_along(tool.data)) {
        ident <- names(tool.data)[i]
        primer.df <- tool.data[[i]]$Primers
        template.df <- tool.data[[i]]$Templates
        data.matrix <- prepare_learning_data(primer.df, template.df, ref.data, settings, cvg.type = "basic")
        feature.data[[i]] <- data.matrix
    }
    feature.matrix <- do.call(rbind, feature.data)
    # encode mismatches locally: hexamer encoding
    pos.mod <- feature.matrix$Position_3terminus 
    pos.mod[is.na(pos.mod) | pos.mod >= 7] <- 7 # no mismatch in 3' hexamer
    pos.mod <- abs(pos.mod - 7)
    feature.matrix$Position_3terminusLocal <- pos.mod
    #feature.matrix$Position_3terminus[is.na(feature.matrix$Position_3terminus)] <- max(feature.matrix$Position_3terminus, na.rm = TRUE) + 1
    feature.matrix$Number_of_mismatches_hexamer <- feature.matrix$Mismatch_pos_1 +  feature.matrix$Mismatch_pos_2 +  
                                                    feature.matrix$Mismatch_pos_3 +  feature.matrix$Mismatch_pos_4 +  
                                                    feature.matrix$Mismatch_pos_5 +  feature.matrix$Mismatch_pos_6
    result <- list("feature_matrix" = feature.matrix, "primer_data" = tool.data)
    return(result)
}

get_learning_matrix_old <- function(ref.data) {
    data(Ippolito) # load templates
    # primer set to use for calibration:
    mode.directionality <- "fw"
    # create tool data for Tiller / openPrimeR:
    # selected reduced template set, for which experimental annotation exists
    template.file <- fasta.file <- system.file("extdata", "IMGT_data", "templates", 
                                                "Homo_sapiens_IGH_functional_exon_exp.fasta", 
                                                package = "openPrimeR")
    leader.file <- system.file("extdata", "IMGT_data", "templates",
                              "Homo_sapiens_IGH_functional_leader_exp.fasta", 
                              package = "openPrimeR")
    hdr.structure <- c("ACCESSION", "GROUP", "SPECIES", "FUNCTION")
    # load templates
    template.df <- openPrimeR::read_templates(template.file, hdr.structure, "|", "GROUP", rm.keywords = "partial")
    # assign leader:
    template.df <- assign_binding_regions(template.df, leader.file, NULL)
    # extend forward binding region by one position for overlap cvg starting at V-gene
    template.df <- adjust_binding_regions(template.df, c(-max(template.df$Allowed_End_fw_initial - template.df$Allowed_Start_fw_initial), 0), c(-50,1)) 
    m <- match(colnames(ref.data[[1]]), template.df$ID)
    my.templates <- template.df[m[!is.na(m)],]
    tiller.location <- system.file("extdata", "IMGT_data", "primers", "IGHV", 
                            "Tiller2008_1st.fasta", package = "openPrimeR")
    open.location <- system.file("extdata", "IMGT_data", "primers", "IGHV", 
                            "openPrimeR2017.fasta", package = "openPrimeR")
    # create settings for individual tool data generation:
    filename <- system.file("extdata", "settings", 
                "C_Taq_PCR_high_stringency.xml", package = "openPrimeR")
    settings <- openPrimeR::read_settings(filename)
    # prevent other binding events:
    conOptions(settings)$allowed_other_binding_ratio <- 0.0
    # allow input nbr of mismatches
    conOptions(settings)$allowed_mismatches <- 12
    # require only intersection with target region
    conOptions(settings)$allowed_region_definition <- "any"
    # compute data
    # activate all 'interesting' cvg constraints as features, but don't filter! (set boundaries high)
    cvg_constraints(settings) <- list("annealing_DeltaG" = c("max" = 0), "primer_efficiency" = c("max" = 1),
                                      "terminal_mismatch_pos" = c("min" = 0))
    tiller.settings <- settings
    PCR(tiller.settings)$annealing_temp <- 57
    tiller.data <- get_tool_primer_data(tiller.location, my.templates, tiller.settings)
    open.settings <- settings
    PCR(open.settings)$annealing_temp <- 55
    open.data <- get_tool_primer_data(open.location, my.templates, open.settings)
    tool.data <- list("Tiller" = tiller.data, "openPrimeR" = open.data)
    feature.data <- vector("list", length(tool.data))
    for (i in seq_along(tool.data)) {
        ident <- names(tool.data)[i]
        primer.df <- tool.data[[i]]$Primers
        template.df <- tool.data[[i]]$Templates
        ref.cvg <- ref.data[[i]]
        data.matrix <- prepare_learning_data(primer.df, template.df, ref.cvg, settings, cvg.type = "basic")
        feature.data[[i]] <- data.matrix
    }
    feature.matrix <- do.call(rbind, feature.data)
    # important: set position of 3' terminus mismatch to primer length if NA (no mismatch)
    pos.mod <- feature.matrix$Position_3terminus 
    # add prior knowledge to positions representation (only consider 3' hexamer)
    pos.mod[is.na(pos.mod) | pos.mod >= 7] <- 7 # no mismatch in 3' hexamer
    pos.mode <- abs(pos.mod - 7)
    feature.matrix$Position_3terminusLocal <- pos.mode
    feature.matrix$Position_3terminus[is.na(feature.matrix$Position_3terminus)] <- max(feature.matrix$Position_3terminus, na.rm = TRUE) + 1
    feature.matrix$Number_of_mismatches_hexamer <- feature.matrix$Mismatch_pos_1 +  feature.matrix$Mismatch_pos_2 +  
                                                    feature.matrix$Mismatch_pos_3 +  feature.matrix$Mismatch_pos_4 +  
                                                    feature.matrix$Mismatch_pos_5 +  feature.matrix$Mismatch_pos_6
    return(feature.matrix)
}
color.to.class <- function(ref.df) {
    cnam <- colnames(ref.df)
    red <- "ff0000"
    orange <- "ed7d31"
    fail.colors  <- c(red, orange)
    conv <- function(color) {
        out <- rep("Amplified", length(color))
        idx <- which(color %in% fail.colors) # red-colored -> not amplified
        if (length(idx) != 0) {
            out[idx] <- "Unamplified"
        }
        return(out)
    }
    df <- data.frame(apply(ref.df, 2, conv))
    colnames(df) <- cnam
    return(df)
}

prepare_learning_data <- function(primer.df, template.df, ref.data, settings, cvg.type = c("constrained", "basic")) {
    #######
    # prepare tool results for learning:
    #######
    cvg.type <- match.arg(cvg.type)
    full.df <- prepare_mm_plot(primer.df, template.df)
    full.df <- full.df[full.df$Coverage_Type == cvg.type,]
    # select unique binding event for all mismatch contacts:
    df <- plyr::ddply(full.df, c("Primer", "Template", "Group"), plyr::summarize,
                        Number_of_mismatches = unique(substitute(Number_of_mismatches)),
                        primer_efficiency = unique(substitute(primer_efficiency)),
                        annealing_DeltaG = unique(substitute(annealing_DeltaG)),
                        Position_3terminus = min(substitute(Position_3terminus)),
                        All_mismatches = paste(substitute(Position_3prime), collapse = ","),# mismatches relative to the 3' end (pos 1 indicates 3' terminus)
                        Worst_Case_MismatchLocal = paste(substitute(Position_3terminusLocal), collapse = ","))  # worst case mismatches where 1 -> first pos in 3' hexamer, 6 -> 3' terminus
                        #Binding_Pos_Start = unique(substitute(Binding_Position_Start_fw)))
    # add some primer features
    m <- match(df$Primer, primer.df$ID)
    df$Primer_Sequence <- primer.df$Forward[m]
    df$Terminal_Dinucleotide <- substr(primer.df$Forward[m], nchar(primer.df$Forward[m]) - 1, nchar(primer.df$Forward[m]))
    # add some template features; TODO: need to adjust prepare_mm_plot to output the binding stretch or binding position in the templates
    m <- match(df$Template, template.df$ID)
    template.seqs <- template.df$Sequence[m]
    # add 3' hexamer mismatch features: 1 if mismatch, 0 if not, per position in the 3' hexamer: 
    # pos 1 -> first position in 3' hexamer, pos6: last position in 3' hexamer (<=> 3' terminus)
    hexa.df <- data.frame(matrix(rep(0, 6 * nrow(df)), nrow = nrow(df), ncol = 6))
    colnames(hexa.df) <- paste0("Mismatch_pos_", seq(1, 6)) # changed labels for considering position starting in the sequence direction 
    mm <- lapply(strsplit(df$All_mismatches, split = ","), function(x) ifelse(x == "NA", NA, as.numeric(x)))
    for (i in seq_along(mm)) {
        cur.mm <- mm[[i]]
        if (!is.na(cur.mm[1])) {
            sel <- which(cur.mm <= 6)
            if (length(sel) != 0) {
                hexa.mm <- abs(cur.mm[sel] - 7)
                hexa.df[i, hexa.mm] <- 1
            }
        }
    }
    # determine count of hexamer mismatches
    hexa.counts <- apply(hexa.df, 1, sum)
    df <- cbind(df, hexa.df, "Hexamer_Mismatch_Count" = hexa.counts)
    #######
    # prepare experimental data 
    #########
    # assign labels to the learning data
    combi.df <- merge(df, ref.data, by=c("Primer","Template")) # NA's match
    if (nrow(combi.df) != nrow(df)) {
        warning("combi.df excluded some obsevations; probably because number of allowed mismatches was too low when evaluating the primer coverage -> set it higher for creating a full training set.")
        print(paste("Obs. in tool data frame: ", nrow(df)))
        print(paste("Obs. in merged data frame: ", nrow(combi.df)))
    }
    # assign features from evaluation of primers:
    sel.features <- names(constraints(settings))
    excl.features <- c("cross_dimerization", "melting_temp_diff") # not single-primer based constraints
    sel.features <- setdiff(sel.features, excl.features)
    # rename some features for matching:
    sel.features[sel.features == "melting_temp_range"] <- "melting_temp"
    sel.features[sel.features == "secondary_structure"] <- "Structure_deltaG"
    sel.features[sel.features == "self_dimerization"] <- "Self_Dimer_DeltaG"
    feature.idx <- sapply(sel.features, function(x) grep(paste0("^",x), colnames(primer.df))[1])
    feature.df <- cbind(Primer = primer.df$ID, primer.df[, feature.idx])
    # extract features and assign to the combined data frame
    augmented.df <- merge(combi.df, feature.df, by = "Primer")
    # store info about primer set in rownames
    rownames(augmented.df) <- paste0(rownames(augmented.df), "_", unique(primer.df$Run))
    augmented.df$Experimental_Coverage <- factor(augmented.df$Experimental_Coverage, levels = c("Unamplified", "Amplified"))
    # add run
    augmented.df$Run <- unique(primer.df$Run)
    return(augmented.df)
}
devtools::load_all("src/openPrimeR") # load openPrimeR package
# get reference coverage for primer sets
cvg.folder <- system.file("data-raw", "coverage_data", package = "openPrimeR")
cvg.folders <- c(file.path(cvg.folder, "Tiller"), file.path(cvg.folder, "openPrimeR"))
ref.data <- get_ref_data(cvg.folders)
#save(ref.data, file = out.loc, compress = "xz")
# load templates
template.file <- file.path(cvg.folder, "IGHV_templates.fasta") 
# define 5' UTR + leader as allowed binding region
allowed.file <- file.path(cvg.folder, "IGHV_allowed_region.fasta")
template.df <- read_templates(template.file, c("GROUP"), delim = "|") 
template.df <- assign_binding_regions(template.df, fw = allowed.file, rev = NULL)
# extend forward binding region by one position in order to set coverage definition where any overlap with binding region is tolerated (ensure that primers bind at the start position at the latest)
template.df <- adjust_binding_regions(template.df, c(-max(template.df$Allowed_End_fw_initial - template.df$Allowed_Start_fw_initial), 0), c(-50,1)) 
# load primers from adjusted paths due to adapted filenames fitting for reference primer IDs
tiller.location <- file.path(cvg.folder, "Tiller2008_1st.fasta")
open.location <- file.path(cvg.folder,"openPrimeR2017.fasta")
######
# get feature matrix for supervised learning
#####
feature.matrix.data <- get_learning_matrix(ref.data, template.df, tiller.location, open.location)
feature.matrix <- feature.matrix.data$feature_matrix
primer.data <- feature.matrix.data$primer_data
out.loc <- file.path(system.file("data",  package = "openPrimeR"), "RefCoverage.rda")
#########
# Store the reference data, feature matrix:
#########
save(ref.data, feature.matrix, primer.data, file = out.loc, compress = "xz")
############
if (FALSE) {
    # old code:
    xls.file <- system.file("data-raw", "PCR_ref_data.xlsx", package = "openPrimeR")
    hdr.structure <- c("ACCESSION", "GROUP", "SPECIES", "FUNCTION")
    template.df <- read_templates(system.file("extdata", "IMGT_data", "templates", 
                                                    "Homo_sapiens_IGH_functional_exon_exp.fasta", 
                                                    package = "openPrimeR"), hdr.structure, "|")
    template.df <- assign_binding_regions(template.df, fw = system.file("inst", "extdata", "IMGT_data", "templates", "Homo_sapiens_IGH_functional_leader_exp.fasta", package = "openPrimeR"), rev = NULL)
    ref.data <- get_ref_data (xls.file)
    feature.matrix <- get_learning_matrix(ref.data)
}

#######
# Secondary structures (viennaRNA)
########

#' Read a Secondary Structure
#'
#' Reads the secondary structure output of ViennaRNA.
#'
#' @param fw.out Path to a ViennaRNA secondary structure output file.
#' @return Data frame with information on secondary structures.
#' @keywords internal
read.secondary.structure.raw <- function(fw.out) {
	if (!file.exists(fw.out)) {
		stop(paste("The input file:", fw.out,
					"does not exist."))
	}
    con <- file(fw.out)
	lines <- readLines(con)
    #lines <- try(suppressWarnings(readLines(con)), silent = TRUE)
    close(con)
    if (is(lines, "try-error") || length(lines) == 0) {
        return(data.frame(Structure = character(0), deltaG = numeric(0), 
            stringsAsFactors = FALSE))
    }
    seq.idx <- seq(1, length(lines), 2)
    seqs <- lines[seq.idx]
    str.idx <- seq.idx + 1
    str <- lines[str.idx]
    str.s <- strsplit(str, split = "  \\(  ")
    expr <- "([\\.\\(\\)|x><]+)[[:space:]]\\([[[:space:]]]*(.[^\\)^[[:space:]]]*)"
    m <- stringr::str_match(str, expr)
    delta.G <- as.numeric(m[, 3])
    structure <- m[, 2]
    result <- data.frame(Structure = structure, deltaG = delta.G, stringsAsFactors = FALSE)
    # message(result)
    return(result)
}


#' Computation of Secondary Structures with ViennaRNA.
#'
#' Computes secondary structures using ViennaRNA.
#'
#' @param seqs The input sequences for which structures shall be computed.
#' @param annealing.temperature The temperature in degree Celsius 
#' at which to compute secondary structures.
#' @param constraints If provide
#' @param id An identifier for storing the files
#' @param folding.constraints Character vector specifying the folding
#' conditions for every input sequence.
#' For example the constraint \code{xxxxxx...} would forbid 
#' folding in the first 6 bases and allow folding in the last 3 bases.
#' @return A data frame with secondary structures.
#' @keywords internal
compute.structure.vienna <- function(seqs, annealing.temperature, 
                                    folding.constraints = NULL, id = "") {
    if (length(annealing.temperature) != 1) {
        stop("Can only call viennaRNA with a single annealing temperature!")
    }
    if (length(seqs) == 0) {
        return(NULL)
    }
	runID <- digest::digest(seqs, "md5")
    id <- paste0("viennaRNA", id, runID)
    #message("Computing secondary structures @ annealing temp: ", annealing.temperature)
    rna.fold <- Sys.which("RNAfold")
    if (rna.fold == "") {
        stop("ViennaRNA is not available on your system. Cannot compute secondary structures.")
    }
    # set parameters for the call to RNAfold:
    # we removed --noLP here, too slow now?
    param.string <- paste("--temp", annealing.temperature, "--noconv", "--noPS")
	input.file <- tempfile(pattern = id, fileext = ".txt")
    out.file <- tempfile(pattern = paste0("out_", id), fileext = ".txt")
	# clean up result files (prevents viennaRNA possibly appending to existing files)
	on.exit({
	    if (file.exists(out.file)) { # remove the .fold file
			file.remove(out.file)
		}
        if (file.exists(input.file)) { # remove the .txt file
            file.remove(input.file) 
        }
	})
	file.string <- paste0("--infile=", input.file, " >", out.file)
    seqs <- toupper(seqs)
    if (length(folding.constraints) != 0) {
        # add constraints to input file for viennaRNA 
        seq.records <- sapply(seq_along(seqs), function(x) paste(seqs[x], 
            folding.constraints[x], sep = "\n"))
        writeLines(seq.records, input.file)
    } else {
        # temporary file containing the input sequences for viennaRNA
        writeLines(seqs, input.file)
    }
    if (!file.exists(input.file)) {
        stop(paste("IO error for secondary structure computation. ",
                "Could not write input file for viennaRNA at:",
                input.file))
    }
    if (length(folding.constraints) != 0) {
        # -C: necessary for having constraints
        param.string <- paste(param.string, "-C", "--batch")
    }

	cmd.status <- system2(rna.fold, args = c(param.string, file.string), wait = TRUE)
    if (cmd.status != 0) {
        call <- paste0(rna.fold, " ", param.string, " ", file.string)
        stop("The RNAfold call: '", call, "' has failed with error code: '", cmd.status, ". Please check whether you have an appropriate version of 'ViennaRNA' installed.")
    }
	##################
    # viennaRNA used to append '.fold' to result files (version before 2.4.1)
	# NB: this code is still necessary (wait time before write-out is complete?) for MacOS:
    out.file.ori <- out.file
    out.file <- paste0(out.file, ".fold")
    if (file.exists(out.file.ori)) {
        out.file <- out.file.ori
    }
	#########################
	if (file.exists(out.file)) {
		  result <- read.secondary.structure.raw(out.file)
    } else {
        stop("ViennaRNA (RNAfold) did not create the expected output file: '", out.file, "'")
	}
    return(result)
}

#' Secondary Structure Computations.
#'
#' Computes the secondary structures of the input primers using ViennaRNA.
#' 
#' @param primer.df Primer data frame.
#' @param mode.directionality Direction of primers.
#' @param annealing.temperature Temperatures at which to compute secondary structures for every primer
#' @return Data frame with secondary structure information.
#' @keywords internal
#'
#' @references Lorenz, Ronny and Bernhart, Stephan H. and Höner zu Siederdissen, 
#' Christian and Tafer, Hakim and Flamm, Christoph and Stadler, Peter F. and Hofacker, Ivo L.
#' ViennaRNA Package 2.0
#' Algorithms for Molecular Biology, 6:1 26, 2011, doi:10.1186/1748-7188-6-26
compute.secondary.structures <- function(primer.df, mode.directionality = c("fw", "rev", "both"), annealing.temperature) {
	if (!is(primer.df, "Primers")) {
		stop("Please supply a valid set of primers.")
	}
	if (length(annealing.temperature) == 0) {
		stop("Please supply an annealing temperature.")
	}
    if (length(mode.directionality) == 0) {
        stop("Please supply the 'mode.directionality' argument.")
    }
    mode.directionality <- match.arg(mode.directionality)
    fw.idx <- which(primer.df$Forward != "")
    rev.idx <- which(primer.df$Reverse != "")
    # create groups of annealing temperatures:
    # parallelize -> split into batches
    if (length(annealing.temperature) == 1)  {
        # split up seqs independent of annealing temp
        batches <- batchify(seq_len(nrow(primer.df)))
        annealing.temperature <- rep(annealing.temperature, length(batches))
    } else {
        # split up seqs dependent on annealing temp
        batches <- batchify(seq_len(nrow(primer.df)), annealing.temperature)
	    annealing.temperature <- as.numeric(names(batches))
    }
    # compute structures with ViennaRNA:
	i <- NULL
    str.fw <- foreach(i = seq_along(batches), .combine = "rbind") %dopar% {
        batch.idx <- batches[[i]]
        cur.temp <- annealing.temperature[i]
        str.fw <- compute.structure.vienna(primer.df$Forward[intersect(fw.idx, batch.idx)], cur.temp, id = "fw")
        #str.rev <- compute.structure.vienna(primer.df$Reverse[rev.idx], annealing.temperature, id = "rev")
    }
    str.rev <- foreach(i = seq_along(batches), .combine = "rbind") %dopar% {
        batch.idx <- batches[[i]]
        cur.temp <- annealing.temperature[i]
        str.rev <- compute.structure.vienna(primer.df$Reverse[intersect(rev.idx, batch.idx)], cur.temp, id = "rev")
    }

    # create order of indices for retrieval of results
    fw.idx.out <- intersect(unlist(batches), fw.idx)
    rev.idx.out <- intersect(unlist(batches), rev.idx)
    str.fw <- str.fw[match(fw.idx, fw.idx.out), ]
    str.rev <- str.rev[match(rev.idx, rev.idx.out), ]
    out <- data.frame(Structure_fw = rep(NA, nrow(primer.df)), Structure_rev = rep(NA, 
                nrow(primer.df)), Structure_deltaG_fw = rep(0, nrow(primer.df)), Structure_deltaG_rev = rep(0, 
                nrow(primer.df)), stringsAsFactors = FALSE)
    if (mode.directionality == "both") {
        # add info for missing primers
        res.fw <- str.fw[FALSE, ]
        res.rev <- str.rev[FALSE, ]
        res.fw[seq_len(nrow(primer.df)), ] <- NA
        res.rev[seq_len(nrow(primer.df)), ] <- NA
        res.fw[fw.idx, ] <- str.fw
        res.rev[rev.idx, ] <- str.rev
        out$Structure_fw <- res.fw$Structure
        out$Structure_deltaG_fw <- res.fw$deltaG
        out$Structure_rev <- res.rev$Structure
        out$Structure_deltaG_rev <- res.rev$deltaG
    } else if (mode.directionality == "fw") {
        out$Structure_fw[fw.idx] <- str.fw$Structure
        out$Structure_deltaG_fw[fw.idx] <- str.fw$deltaG
    } else {
        out$Structure_rev[rev.idx] <- str.rev$Structure
        out$Structure_deltaG_rev[rev.idx] <- str.rev$deltaG
    }
    if (length(out) != 0) {
        out$Structure_deltaG <- sapply(seq_len(nrow(out)), 
            function(x) min(out$Structure_deltaG_fw[x], 
                        out$Structure_deltaG_rev[x], na.rm = TRUE))
    }
    return(out)
}

####
# Actions to perform on loading/attaching the package
######
#' @import ggplot2 lpSolveAPI methods
#' @importFrom reshape2 melt dcast
#' @importFrom plyr ddply summarize arrange . here catcolwise rbind.fill numcolwise
#' @importFrom foreach foreach %dopar% getDoParRegistered getDoParWorkers
#' @importFrom IRanges IRanges as.matrix Views findOverlaps overlapsAny CharacterList
#' @importFrom Biostrings DNAStringSet IUPAC_CODE_MAP extractAt reverseComplement mergeIUPACLetters DNAStringSetList vmatchPattern matchPattern width DNA_BASES
#' @importFrom pwalign nucleotideSubstitutionMatrix pairwiseAlignment mismatch compareStrings
#' @importFrom RColorBrewer brewer.pal 
#' @importFrom grDevices colorRampPalette
#' @importFrom S4Vectors metadata metadata<-
#' @importFrom BiocGenerics unlist start end
#' @importFrom magrittr %>%
#' @importFrom stats na.omit qnorm quantile sd ave fisher.test p.adjust as.formula reshape predict hclust
#' @importFrom utils head read.csv read.delim setTxtProgressBar tail txtProgressBar write.csv write.table
#' @importFrom GenomicRanges GRanges
NULL # need to have some evaluated code here

#' Determination if Selenium is installed.
#'
#' Checks whether selenium module for python is installed on the system.
#'
#' @return \code{TRUE} is selenium for python is available,
#' \code{FALSE} otherwise.
#'
#' @keywords internal
selenium.installed <- function() {
	if (Sys.which("python") == "") {
		# python not available
		return(FALSE)
	}
    cmd <- "python -c 'import selenium'"
    ret <- system(cmd, ignore.stdout = TRUE, ignore.stderr = TRUE)
    if (ret == 0) {
        # installed
        return(TRUE)
    } else {
        return(FALSE)
    }
}
#' Check Tool Installation
#'
#' Checks whether all required tools are installed.
#'
#' @param frontend Whether tool installation shall be checked for the frontend.
#' If \code{TRUE}, dependencies that are required only by the frontend are considered additionally.
#' @return \code{TRUE} for each installed tool, \code{FALSE} otherwise.
#' @keywords internal
check.tool.installation <- function(frontend = FALSE) {
    available.tools <- NULL
    # for melting temperatures
    available.tools["MELTING"] <- Sys.which("melting-batch") != ""
    # for secondary structures
    available.tools["ViennaRNA"] <- Sys.which("RNAfold") != ""
    # for DECIPHER
    available.tools["OligoArrayAux"] <- Sys.which("hybrid-min") != ""
     # for multiple sequence alignments
    available.tools["MAFFT"] <-  Sys.which("mafft") != "" 
    available.tools["Pandoc"] <- Sys.which("pandoc") != ""
    ## for IMGT data retrieval in frontend
    if (frontend) {
        available.tools["Selenium"] <- selenium.installed()
        available.tools["PhantomJS"] <- Sys.which("phantomjs") != ""
    }
    return(available.tools)
}
#' Check Functionality of Third-Party Tools.
#'
#' Checks whether all required tools should work.
#'
#' @param frontend Whether tool functionality shall be checked for the frontend.
#' @return \code{TRUE} for each functioning tool, \code{FALSE} for non-functioning tools.
#' @keywords internal
check.tool.function <- function(frontend = FALSE) {
    available.tools <- check.tool.installation(frontend)
    out <- NULL
    # check for oligoArrayAux
    if (available.tools["OligoArrayAux"]) {
        try(out <- system("hybrid-min -n DNA -t 50 -T 50 -N 0.05 -E -q ACAGGTGCCCACTCCCAGGTGCAG CTGCACCTGGGAGTGGGCACCTGT", 
                    intern = FALSE, ignore.stdout = TRUE))
       if (out != 0) {
            # there was an error
            warning("oligoArrayAux failed checks: disabled. Do you have the UNAFOLDDAT environment variable set?")
            available.tools["OligoArrayAux"] <- FALSE
        }
    }
    # check for Pandoc/Latex
    if (available.tools["Pandoc"] && Sys.which("pdflatex") == "") {
        # don't warn here, otherwise too many warnings are generated
        #warning("Cannot create reports with pandoc since LateX is missing.")
        available.tools["Pandoc"] <- FALSE
    }
    # check for ViennaRNA version
    if (available.tools["ViennaRNA"]) {
        # need to require a specific version (2.4.1) of viennaRNA for support of the used commands
	version <- NULL
    	try (version <- system2("RNAfold", "--version", stdout = TRUE, stderr = TRUE))
        if (!is(version, "try-error")) {
            # the command worked -> check the version string
            v <- strsplit(version, " ")[[1]]
            if (length(v) >= 2) {
                nbr <- try(as.numeric(gsub("\\.", "", v[2])))
                if (!is(nbr, "try-error")) {
                    if (nbr < 241) { # version smaller than 2.4.1
                        warning("ViennaRNA had version < 2.4.1. Disabling ViennaRNA.")
                        available.tools["ViennaRNA"] <- FALSE
                    }
                } else {
                    warning("ViennaRNA version unknown. Disabling.")
                    available.tools["ViennaRNA"] <- FALSE # unknown version
                }
            } else {
                warning("ViennaRNA version unknown. Disabling")
                available.tools["ViennaRNA"] <- FALSE # unknown version
            }
        } else {
            warning("ViennaRNA version unknown. Disabling")
            available.tools["ViennaRNA"] <- FALSE # could not get version info
        }
    }
    return(available.tools)
}
#' Copy MELTING Config File
#'
#' Copies modified MELTING tandem mismatch file to the MELTING data folder.
#'
#' @return TRUE if the file is available in the MELTING folder, FALSE otherwise.
#' @keywords internal
copy.melt.config <- function(melt.bin = NULL) {
    print("DEPRECATED")
	if (length(melt.bin) == 0) {
		melt.bin <- Sys.which("melting-batch")[1]
	}
    tandem.mm.file <- system.file("extdata", 
                        "AllawiSantaluciaPeyret1997_1998_1999tanmm_mod.xml",
                        package = "openPrimeR")
    if (tandem.mm.file == "") {
        warning("The MELTING config file is not present in the openPrimeR package.")
        return(FALSE)
    }
    if (melt.bin != "" ) {
        melt.config.file <- file.path(dirname(melt.bin),
                                "..", "Data", basename(tandem.mm.file))
        if (!file.exists(melt.config.file)) {
			message("Copying MELTING config to: ", melt.config.file)
            s <- file.copy(tandem.mm.file, melt.config.file)
            if (any(!s)) {
                warning("Could not copy MELTING config file to destination.")
            }
            return(all(s))
        } else {
            # file is available
            return(TRUE)
        }
    } else {
        return(FALSE)
    }
}
# actions to be performed when loading the package namespace
.onLoad <- function(libname, pkgname) {
    ################
    # Define package options
    #######
    # order in which constraints are computed (least runtime to highest)
    con.order <- c("primer_length", "gc_clamp", "gc_ratio", "no_runs", "no_repeats", 
               "melting_temp_range", "self_dimerization", "secondary_structure", 
               "primer_coverage", "primer_specificity", 
               "melting_temp_diff", "cross_dimerization")
    # order in which constraints are relaxed (start with least important constraint)
    relax.order <- c("primer_length", "primer_coverage", "no_repeats", "no_runs", "gc_clamp", "primer_specificity", "secondary_structure", "self_dimerization", "cross_dimerization", "gc_ratio", "melting_temp_range", "melting_temp_diff")
    available.constraints <- select.constraints(con.order) # select constraints that can be computed by installed software
    # message("Available constraints:", available.constraints)
    con.order <- con.order[con.order %in% available.constraints]
    relax.order <- relax.order[relax.order %in% available.constraints]
    plot.colors <- c("Constraint" = "Set1", "Group" = "Set2", 
                     "Run" = "Set3", "Primer" = "Accent")
    op <- options()
    op.openPrimeR <- list(
        openPrimeR.constraint_order = con.order,
        openPrimeR.relax_order = relax.order,
        openPrimeR.plot_colors = plot.colors,
        openPrimeR.plot_abbrev = 15 # limit label extent for plots
    )
    # only set options once:
    toset <- !(names(op.openPrimeR) %in% names(op))
    if (any(toset)) {
        options(op.openPrimeR[toset])
    }
    # do not provide any output, even if no var is assigned to this function: 
    invisible()
}

# actions to be performed when attaching the package
.onAttach <- function(libname, pkgname) {
    # add start up message
    available.tools <- check.tool.function()
    if (any(!available.tools)) {
        tool.df <- build.tool.overview(available.tools)
        out <- paste0("There are missing/non-functioning external tools.\n",
                "To use the full potential of openPrimeR, please make sure\n",
                "that the required versions of the speciied tools are\n 
                installed and that they are functional:\n")
        idx <- which(!available.tools)
        tools <- paste0("o ", names(available.tools)[idx], 
              " (", tool.df$URL[match(names(available.tools[idx]), tool.df$Tool)], ")", 
              collapse = "\n")
        out <- paste(out, tools, sep = "")
        packageStartupMessage(out)
		# special warning for Pandoc (latex dependency)
		if (Sys.which("pdflatex") == "") {
			warning("'Pandoc' is non-functional, since 'pdflatex' is not installed on your system.")
		}
    }
    # set the default number of cores to use
    foreach.available <- requireNamespace("foreach", quietly = TRUE)
    # register parallel backend if not registered already
    if (foreach.available && !getDoParRegistered()) { 
        default.nbr.cores <- 2
        if (Sys.info()["sysname"] != "Windows") {
            # do not call parallel_setup for windows machines
            # it seems that registering parallel workers leads to
            # firewall issues -> TIMEOUT when loading the package
		    parallel_setup(default.nbr.cores)
        }
    }
}

context("Dimerization")

test_that("Self Dimerization", {
    # test self dimerization
    if (!check.tool.function()["OligoArrayAux"]) {
        # cannot test without oligo
        skip("OligoArrayAux not available.")
    }
    data(Ippolito)
    # select a primer for testing:
    # fix the annealing temp
    PCR(settings)$annealing_temp <- 57
    sel <- primer.df$Forward == "caggtgcagctgcaggagtcsg"
    constraint.df <- check_constraints(primer.df[sel,], template.df, 
                 settings, active.constraints = "self_dimerization")
    # worst-case binding mode: 12 maches, 6 mismatches
    #        5'-CAGGTGCAGCTGCAGGAGTCSG-3'
    #           |XXX||||||||||XXX|
    #    3'-GSCTGAGGACGTCGACGTGGAC-5'
    # free energy
    expect_that(constraint.df[,"Self_Dimer_DeltaG"], 
                equals(-6.996, tolerance = 0.5))
    # add a reverse primer and check worst-case binding again
    primer.df[sel, "Reverse"] <- "ctcctgcagctgcaggagtcsg"
    # new binding mode:
    #     5'-CTCCTGCAGCTGCAGGAGTCSG-3'</b><br>
    #        ||||||||||||||||||
    # 3'-GSCTGAGGACGTCGACGTCCTC-5'</b>
    constraint.df <- check_constraints(primer.df[sel,], template.df, 
                    settings, active.constraints = "self_dimerization")
    # match score should increase by 6 since we introduced 3 matches 
    # and prevented 3 mismatches
    # free energy
    expect_that(constraint.df[,"Self_Dimer_DeltaG"], 
                equals(-13.028, tolerance = 0.5))
}) 

test_that("Cross Dimerization", {
    # test cross dimerization
    if (!check.tool.function()["OligoArrayAux"]) {
        # cannot test without oligo
        skip("OligoArrayAux not available.")
    }
    data(Ippolito)
    constraint.df <- check_constraints(primer.df, template.df, 
                 settings, active.constraints = "cross_dimerization")
    sel <- "gaggtgcagctgktggagwcy"
    test.df <- constraint.df[which(constraint.df$Forward == sel),]
    #                 5'-GAGGTGCAGCTGKTGGAGWCY-3'
    #                    X||X||||X|||
    #       3'-GSCTGAGGACGTCGACGTGGAC-5'
    expect_that(test.df[, "Cross_Dimer_DeltaG"], 
                equals(-6.775, tolerance = 0.5))
    prev.G <- test.df[, "Cross_Dimer_DeltaG"]
    # modify cross-dimer partner reverse
    constraint.df[which(constraint.df$ID == test.df$Cross_Dimer_Idx1), "Reverse"] <-
                "cagctgcagctgcaggagtcsg"
    constraint.df <- check_constraints(constraint.df, template.df, 
                 settings, active.constraints = "cross_dimerization")
    test.df <- constraint.df[which(constraint.df$Forward == sel),]
    # energy should be smaller since this conformation has fewer mismatches
    expect_lte(test.df[, "Cross_Dimer_DeltaG"], prev.G)
})

context("Primer coverage")

test_that("Stop codon filter", {
    # test coverage stop codon filter
    data(Ippolito)
    template.df <- template.df[1:3,]
    # reading frame: aga|tac|...
    # note: third seq has already a stop codon at 'taa' -> coverage events
    # with 'taa' are ok!
    template.df$InputSequence <- c("agatacacccccccccccccc",
                                   "agatacagattacagattaca",
                                   "agataaagattacagattaca") 
    template.df <- assign_binding_regions(template.df, fw = c(1,14), rev = c(1,5))
    stop.codons <- c("TAG", "TAA", "TGA")
    # first primer has gaTAGa: should bind to 1st, 2nd, 3rd with stop codon
    # second primer has gaTAAa: should bind to 1st, 2nd, 3rd with stop codon
    # third primer has gaTGA: should bind to the third template with stop codon
    seqs <- c("gataga", "gataaa", "gatgaa")
    fname <- tempfile("test.fasta")
    seqinr::write.fasta(as.list(seqs), as.list(c("testPrimer1_fw", "testPrimer2_fw", "testPrimer3_fw")), fname)
    primer.df <- read_primers(fname)
    # allow a single mismatch to introduce stop codons
    conOptions(settings)$allowed_mismatches <- 1
    # do not use any additional cvg constraints:
    cvg_constraints(settings) <- list()
    constraint.df <- check_constraints(primer.df, template.df, settings,
                        active.constraints = "primer_coverage")
#######
    #mutation.types <- c("stop_codon")
    #mutation.check <- mismatch.mutation.check(constraint.df, template.df, mutation.types)
###
    expect_that(constraint.df$primer_coverage,
                equals(c(3, 3, 1), tolerance = 0.0))
    expect_that(constraint.df$primer_specificity,
                equals(c(1, 1, 1), tolerance = 0.01))
    # now, add stop codon constraint to filter out stop codon events
    cvg_constraints(settings) <- list("stop_codon" = c("min" = 0, "max" = 0))
    constraint.df <- suppressWarnings(check_constraints(primer.df, template.df, settings,
                        active.constraints = "primer_coverage"))
    # all primers can bind the third template with 1 mismatch:
    # binding is ok since 'taa' is already a stop codon in the template
    # and no additional stop codon is introduced!
    expect_that(constraint.df$primer_coverage,
                equals(c(1, 1, 1), tolerance = 0.0))
}) 
test_that("Free energy filter", {
    # check whether binding events are filtered correctly according to their free energies
    if (!check.tool.function()["OligoArrayAux"]) {
        # cannot test without OligoArrayAux
        skip("OligoArrayAux not available.")
    }
    data(Ippolito)
    primer.df <- primer.df[1:2,]
    # get baseline results if we don't filter
    cvg_constraints(settings) <- list("annealing_DeltaG" = c("max" = 0)) # no filtering
    constraint.df <- check_constraints(primer.df, template.df, settings,
                        active.constraints = "primer_coverage")
    # decide for a cutoff:
    deltaG.cutoff <- -15 # everything greater than -15 should be filtered
    # retrieve observed values
    deltaG.values <- lapply(strsplit(constraint.df$annealing_DeltaG, split = ","), as.numeric)
    sel.idx <- lapply(deltaG.values, function(x) which(x <= deltaG.cutoff))
    filtered.cvd.seqs <- sapply(seq_along(sel.idx), function(x) paste0(strsplit(constraint.df$Covered_Seqs[x], split = ",")[[1]][sel.idx[[x]]], collapse = ","))
    filtered.cvg <- sapply(sel.idx, length)
    # determine results from tool:
    cvg_constraints(settings) <- list("annealing_DeltaG" = c("max" = deltaG.cutoff)) # filter!
    constraint.df <- check_constraints(primer.df, template.df, settings,
                        active.constraints = "primer_coverage")
    expect_equal(filtered.cvg, constraint.df$primer_coverage) 
    expect_equal(filtered.cvd.seqs, constraint.df$Covered_Seqs)
})

test_that("Efficiency Filter", {
    # check whether binding events are filtered correctly according to their efficiencies
    if (!check.tool.function()["OligoArrayAux"]) {
        # cannot test without OligoArrayAux
        skip("OligoArrayAux not available.")
    }
    data(Ippolito)
    primer.df <- primer.df[1:2,]
    # get baseline results if we don't filter
    cvg_constraints(settings) <- list("primer_efficiency" = c("min" = 0)) # no filtering
    constraint.df <- check_constraints(primer.df, template.df, settings,
                        active.constraints = "primer_coverage")
    # decide for a cutoff:
    eff.cutoff <- 0.1 # everything smalelr than 0.1 efficiency should be filtered
    # retrieve observed values
    eff.values <- lapply(strsplit(constraint.df$primer_efficiency, split = ","), as.numeric)
    sel.idx <- lapply(eff.values, function(x) which(x >= eff.cutoff))
    filtered.cvd.seqs <- sapply(seq_along(sel.idx), function(x) paste0(strsplit(constraint.df$Covered_Seqs[x], split = ",")[[1]][sel.idx[[x]]], collapse = ","))
    filtered.cvg <- sapply(sel.idx, length)
    # determine results from tool:
    cvg_constraints(settings) <- list("primer_efficiency" = c("min" = eff.cutoff)) # filter
    constraint.df <- check_constraints(primer.df, template.df, settings,
                        active.constraints = "primer_coverage")
    expect_equal(filtered.cvg, constraint.df$primer_coverage) 
    expect_equal(filtered.cvd.seqs, constraint.df$Covered_Seqs)
})

test_that("3prime mismatches", {
    # check filtering for mismatches at the 3' ends
    data(Ippolito)
    primer.df <- primer.df[1:2,]
    # get baseline results if we don't filter
    cvg_constraints(settings) <- list("terminal_mismatch_pos" = c("min" = 0)) # no filtering
    conOptions(settings)$allowed_mismatches <- 3
    constraint.df <- check_constraints(primer.df, template.df, settings,
                        active.constraints = "primer_coverage")
    # decide for a cutoff:
    terminal.cutoff <- 6 # everything smalelr than 0.1 efficiency should be filtered
    # retrieve observed values
    terminal.pos <- lapply(strsplit(constraint.df$terminal_mismatch_pos, split = ","), as.numeric)
    sel.idx <- lapply(terminal.pos, function(x) which(x >= terminal.cutoff))
    filtered.cvd.seqs <- sapply(seq_along(sel.idx), function(x) paste0(strsplit(constraint.df$Covered_Seqs[x], split = ",")[[1]][sel.idx[[x]]], collapse = ","))
    filtered.cvg <- sapply(sel.idx, length)
    # determine results from tool:
    cvg_constraints(settings) <- list("terminal_mismatch_pos" = c("min" = terminal.cutoff)) # filter
    constraint.df <- check_constraints(primer.df, template.df, settings,
                        active.constraints = "primer_coverage")
    expect_equal(filtered.cvg, constraint.df$primer_coverage) 
    expect_equal(filtered.cvd.seqs, constraint.df$Covered_Seqs)
})

test_that("basic_coverage", {
    # check that basic reported coverage is correct:
        # position of binding
        # mismatch positions
        # number of mismatches
        # covered seqs
        # coverage count
    data(Ippolito)
    primer.df <- primer.df[1:2,]
    template.df <- template.df[template.df$Identifier %in% c(4,5,8,9)]
    cvg_constraints(settings) <- list() # no advanced cvg requirements
    conOptions(settings)$allowed_mismatches <- 0
    constraint.df <- check_constraints(primer.df, template.df, settings,
                        active.constraints = "primer_coverage")
    expect_equal(constraint.df$primer_coverage, c(2,2))
    expect_equal(constraint.df$Coverage_Ratio, c(0.5,0.5))
    expect_equal(constraint.df$Nbr_of_mismatches_fw, c("0,0", "0,0"))
    expect_equal(constraint.df$Binding_Position_Start_fw, c("58,58", "58,58"))
    expect_equal(constraint.df$Binding_Position_End_fw, c("80,80", "80,80"))
    cvg_constraints(settings) <- list("terminal_mismatch_pos" = c(min = 6))
})

context("Primer specificity")

test_that("primer_specificity", {
    data(Ippolito)
    template.df <- template.df[1:3,]
    # 1:13 binding, 14:... off
    template.df$InputSequence <- c("gatacacccccccccccccc",
                                   "gatacagattacagattaca",
                                   "gatacagattacagattaca")
    template.df <- assign_binding_regions(template.df, fw = c(1,13), rev = c(1,5))
    # create a new primer for testing
    seqs <- c("gataca", "gattaca", "ttaagaaaagga")
    fname <- tempfile("test.fasta")
    seqinr::write.fasta(as.list(seqs), as.list(c("testPrimer1_fw", "testPrimer2_fw", "testPrimer3_fw")), fname)
    primer.df <- read_primers(fname)
    conOptions(settings)$allowed_mismatches <- 0
    cvg_constraints(settings) <- list()
    constraint.df <- check_constraints(primer.df, template.df, settings,
                        active.constraints = "primer_coverage")
    expect_that(constraint.df$primer_coverage,
                equals(c(3, 2, 0), tolerance = 0.0))
    expect_that(constraint.df$primer_specificity,
                equals(c(1, 0.5, 0), tolerance = 0.01))
    # check coverage/specificity with mismatch binding
    seqs <- c("gaccca", "gacccca", "ttaagaaaagga") # 2 and 3 mismatches for binding required
    fname <- tempfile("test.fasta")
    seqinr::write.fasta(as.list(seqs), as.list(c("testPrimer1_fw", "testPrimer2_fw", "testPrimer3_fw")), fname)
    primer.df <- read_primers(fname)
    conOptions(settings)$allowed_mismatches <- 3
    constraint.df <- check_constraints(primer.df, template.df, settings,
                        active.constraints = "primer_coverage")
    expect_that(constraint.df$primer_coverage,
                equals(c(3, 3, 0), tolerance = 0.0))
    expect_that(constraint.df$primer_specificity,
                equals(c(3/(3+3), 3/(3+3), 0), tolerance = 0.01))
    # test filtering of cvg wrt specificity:
    # this part requires efficiency/annealing computations with oligoArrayAux
    if (!check.tool.function()["OligoArrayAux"]) {
        # cannot test without oligo
        skip("OligoArrayAux not available.")
    }
    data(Ippolito)
    conOptions(settings)$allowed_mismatches = 5
    seqs <- c("gatacaga")
    fname <- tempfile("test.fasta")
    seqinr::write.fasta(as.list(seqs), as.list(c("testPrimer1_fw")), fname)
    cvg_constraints(settings) <- list("annealing_DeltaG" = c("max" = (-10))) # select only bona-fide events
    primer.df <- read_primers(fname)
    constraint.df <- check_constraints(primer.df, template.df, settings,
                        active.constraints = "primer_coverage")
    low.spec <- constraint.df$primer_specificity
    expect_equal(low.spec, 1)
    expect_lt(length(unlist(strsplit(constraint.df$Off_Covered_Seqs, split = ","))), 
              length(unlist(strsplit(constraint.df$Basic_Off_Covered_Seqs, split = ","))))
}) 

context("Secondary Structures")

test_that("Secondary Structure", {
    # test self dimerization
    if (!check.tool.function()["ViennaRNA"]) {
        # cannot test without viennaRNA
        skip("Secondary structure tests require ViennaRNA.")
    }
    data(Ippolito)
    # select a primer for testing:
    # fix the annealing temp
    # test whether automatic setting of Ta works: structure energy should be different for the higher automatic Ta than the one we will set next:
    sel <- primer.df$Forward == "caggtgcagctgcaggagtcsg"
    constraint.df <- check_constraints(primer.df[sel,], template.df, 
                 settings, active.constraints = "secondary_structure")
    expect_that(constraint.df[, "Structure_deltaG"], 
                equals(0.0, tolerance = 0.01))
    # let's use a fixed Ta for the next computations: structure energy should be different
    PCR(settings)$annealing_temp <- 57
    sel <- primer.df$Forward == "caggtgcagctgcaggagtcsg"
    constraint.df <- check_constraints(primer.df[sel,], template.df, 
                 settings, active.constraints = "secondary_structure")
    expect_that(constraint.df[, "Structure_deltaG"], 
                equals(-0.12, tolerance = 0.01))
    expect_that(constraint.df[, "Structure_deltaG_fw"], 
                equals(constraint.df[, "Structure_deltaG"],
                tolerance = 0.01))

    # add a reverse primer with considerable structure formation
    primer.df[sel, "Reverse"] <- "cccccccaaaaagggggtttttttttt"
    constraint.df <- check_constraints(primer.df[sel,], template.df, 
                    settings, active.constraints = "secondary_structure")
    expect_that(constraint.df[,"Structure_deltaG"], 
                equals(-6.77, tolerance = 0.1))
    expect_that(constraint.df[,"Structure_deltaG"], 
                equals(constraint.df[, "Structure_deltaG_rev"], 
                tolerance = 0.1))
}) 

###################
# Number of repeats
###################
#' Number of Repeats
#' 
#' Computes the number of dinucleotide repeats in the input sequences.
#'
#' @param x Input sequence strings.
#' @return The maximal number of dinucleotide repeats in \code{x}.
#' @keywords internal
nbr.of.repeats <- function(x) {
    # check for dinucleotide repeats, e.g. AT AT AT AT (max nbr of repeats: 4) x:
    # vector of primers (character)
    if (length(x) == 0 || !is(x, "character")) {
        return(NULL)
    }
    nts <- c("a", "c", "g", "t")
    options <- expand.grid(nts, nts, stringsAsFactors = FALSE)  # all dinucleotides
    colnames <- c("ID", "Dinucleotide", "nbr_repeats")
    result <- data.frame(matrix(rep(NA, length(colnames) * nrow(options) * length(x)), 
        nrow = length(x) * nrow(options), ncol = length(colnames)), stringsAsFactors = FALSE)
    colnames(result) <- colnames
    # disambiguate primers to deal with amibuity codes
    x.iupac <- convert.from.iupac(x)
    i <- NULL
    results <- foreach(i = seq_len(nrow(options)), .combine = "rbind") %dopar% {
        # for every dinuclotide combination, count the number of occurrences in the seqs
        dinuc <- paste(as.character(options[i, ]), collapse = "")
        expr <- paste("(", dinuc, ")+", sep = "")
        hits <- gregexpr(expr, x.iupac)
        nbr.di <- unlist(lapply(hits, function(y) max(attr(y, "match.length"))))
        nbr.di <- sapply(nbr.di, function(x) if (x == -1) {
            0
        } else {
            x/2  # match length divided by 2 because we want to output the count of dinucleotides, not that of single nucleotides
        })
        t.s <- (i - 1) * length(x) + 1
        t.e <- t.s + length(x) - 1
        di.result <- data.frame(ID = seq_along(x), Dinucleotide = dinuc, nbr_repeats = nbr.di, 
            stringsAsFactors = FALSE)
    }
    repeat.table <- ddply(results, c("ID"), summarize, max_repeat = max(substitute(nbr_repeats)))  # max dinucleotide repeat length for every sequence
    return(repeat.table$max_repeat)
}

context("Repeats and Runs")

test_that("Number of repeats", {
    # test dinucleotide repeats
    expect_that(nbr.of.repeats("ATATATAT"), equals(4))
    expect_that(nbr.of.repeats("TCGCGCG"), equals(3))
    expect_that(nbr.of.repeats("T"), equals(0))
    expect_that(nbr.of.repeats("TACG"), equals(1))
    expect_that(nbr.of.repeats(5), equals(NULL))
}) 

test_that("Number of runs", {
    # test single nucleotide runs
    expect_that(nbr.of.runs("TAAAAGG"), equals(4))
    expect_that(nbr.of.runs("ACTG"), equals(1))
    expect_that(nbr.of.runs(numeric(0)), equals(NULL))
})

# Store Tiller's primers for IGH, the templates, and the settings.
fasta.file <- system.file("extdata", "IMGT_data", "templates", 
            "Homo_sapiens_IGH_functional_exon.fasta", 
            package = "openPrimeR")
hdr.structure <- c("ACCESSION", "GROUP", "SPECIES", "FUNCTION")
# load templates
tiller.template.df <- openPrimeR::read_templates(fasta.file, hdr.structure, "|", "GROUP", rm.keywords = "partial")
# assign leader:
leader.file <- system.file("extdata", "IMGT_data", "templates",
            "Homo_sapiens_IGH_functional_leader.fasta", package = "openPrimeR")
tiller.template.df <- assign_binding_regions(tiller.template.df, leader.file, NULL)
# load primers
primer.location <- system.file("extdata", "IMGT_data", "primers", "IGHV", 
                        "Tiller2008_1st.fasta", package = "openPrimeR")
tiller.primer.df <- openPrimeR::read_primers(primer.location, "_fw", "_rev")
# load settings for analysis
filename <- system.file("extdata", "settings", 
            "C_Taq_PCR_high_stringency.xml", package = "openPrimeR")
tiller.settings <- openPrimeR::read_settings(filename)
# evaluate constraints
# allow also other binding events:
conOptions(tiller.settings)$allowed_other_binding_ratio <- 1.0
conOptions(tiller.settings)$allowed_mismatches <- 5
tiller.primer.df <- openPrimeR::check_constraints(tiller.primer.df, tiller.template.df, tiller.settings)
tiller.template.df <- openPrimeR::update_template_cvg(tiller.template.df, tiller.primer.df)
out.loc <- file.path(system.file("data",  package = "openPrimeR"), "Tiller.rda")
# compression level: 'xz' ensures small size
save(tiller.primer.df, tiller.template.df, tiller.settings, file = out.loc, compress = "xz")

##################
# Primer efficiency
###################
#' Primer Efficiency.
#'
#' Computes the efficiency of primer binding events for Taq polymerase.
#'
#' This function uses DECIPHER's \link[DECIPHER]{CalculateEfficiencyPCR}.
#'
#' @param fw.primers Primer sequence strings.
#' @param fw.start Binding position (start).
#' @param fw.end Binding position (end).
#' @param covered List of covered template indices per primer.
#' @param taqEfficiency Whether the efficiency shall be computed
#' using a mismatch-model developed for Taq polymerases. The default setting
#' is \code{TRUE}. Set \code{taqEfficiency} to \code{FALSE} if you are using
#' another polymerase than Taq.
#' @param annealing.temp Annealing temperature for which to evaluate efficiency.
#' @param primer_conc Primer concentration.
#' @param sodium.eq.concentration The sodium-equivalent concentration of ions.
#' @param mode.directionality Primer directionality.
#' @param seqs Template sequence strings.
#' @return The efficiencies of primer binding events.
#' @keywords internal
#' @references Wright, Erik S., et al. 
#' "Exploiting extension bias in polymerase chain reaction to improve 
#' primer specificity in ensembles of nearly identical DNA templates." 
#' Environmental microbiology 16.5 (2014): 1354-1365.
# note: since we compute 'efficiency' for individual primers only, this means that the model for primer-primer formation does not do anything, except computing something relating to self dimers. this is ok, since openPrimeR offers its own methods for cross dimerization.
compute.efficiency <- function(fw.primers, fw.start, fw.end, covered, 
                               taqEfficiency, annealing.temp, primer_conc, 
                               sodium.eq.concentration, 
                               mode.directionality = c("fw", "rev"), seqs) {
    if (length(mode.directionality) == 0) {
        stop("Please supply the 'mode.directionality' argument.")
    }
    mode.directionality <- match.arg(mode.directionality)
    ori.primers <- fw.primers
    # select only viable sequences for efficiency considerations
    sel.idx <- which(fw.primers != "")
    fw.primers <- my.disambiguate(DNAStringSet(fw.primers[sel.idx]))
    fw.effs <- vector("list", length(fw.primers))
    # parallelize along primers:
    i <- NULL
    #for (i in seq_along(fw.primers)) { # for debug ..
    fw.effs <- foreach (i = seq_along(fw.primers), .combine = c) %dopar% {
        #print(i)
        Ta <- annealing.temp[i]
        starts <- fw.start[[sel.idx[i]]]
        ends <- fw.end[[sel.idx[i]]]
        covered.seq.idx <- covered[[sel.idx[i]]]
        covered.seqs <- seqs[covered.seq.idx]
        cur.data <- data.frame(Primers = rep(NA, length(covered.seqs)), Templates = rep(NA, length(covered.seqs)))
        for (j in seq_along(starts)) {
            # identify template regions that are covered
            targets <- extractAt(covered.seqs[[j]], at = IRanges(start = starts[j], 
                end = ends[j]))  # fw ambig target sequences
            combis <- expand.grid(as.character(unlist(targets)), as.character(fw.primers[[i]]), stringsAsFactors = FALSE)  # possible binding modes for ambig sequences
            # select only the best (i.e. the actual) binding mode -> the one with smallest distance
            d <- stringdist::stringdist(combis[,1], combis[,2], nthread = 1)
            conformation.idx <- which.min(d)
            cur.data$Primers[j] <- combis[conformation.idx, 2]
            cur.data$Templates[j] <- combis[conformation.idx, 1]
        }
        # select only unique combinations of primers and templates
        dup.data <- unlist(lapply(seq_len(nrow(cur.data)), function(x) paste(unlist(cur.data[x,]), collapse = "")))
        # create an index to compute efficiency only for unique combinations of primers&templates
        idx <- which(!duplicated(dup.data))
        m <- match(dup.data, dup.data[idx]) # idx of corresponding representatitives
        if (nrow(cur.data) == 0) { # no templates covered
            cur.eff <- 0 # primer has 0 efficiency
        } else if (mode.directionality == "fw") {
            # reverse complement the binding sequence
            cur.eff <- DECIPHER::CalculateEfficiencyPCR(DNAStringSet(cur.data$Primers[idx]), 
                    reverseComplement(
                        DNAStringSet(cur.data$Templates[idx])
                    ), Ta, primer_conc, 
                    sodium.eq.concentration, 
                    taqEfficiency = taqEfficiency, processors = 1)
        } else if (mode.directionality == "rev") {
            cur.eff <- DECIPHER::CalculateEfficiencyPCR(DNAStringSet(cur.data$Primers[idx]), 
                        DNAStringSet(cur.data$Templates[idx]), Ta, 
                        primer_conc, sodium.eq.concentration, 
                        taqEfficiency = taqEfficiency, processors = 1)
        }
        eff <- rep(NA, nrow(cur.data))
        eff <- cur.eff[m]
        list(eff)
    }
    out.effs <- vector("list", length(ori.primers))
    out.effs[sel.idx] <- fw.effs
    return(out.effs)
}
#' Primer Efficiency.
#'
#' Computes the efficiency of primer binding events for Taq polymerase.
#'
#' This function uses DECIPHER's \code{\link[DECIPHER]{CalculateEfficiencyPCR}}.
#'
#' @param primer.df Primer data frame.
#' @param template.df Template data frame.
#' @param annealing.temp Annealing temperature for which to evaluate efficiency.
#' @param taqEfficiency Whether the efficiency shall be computed
#' using a mismatch-model developed for Taq polymerases. The default setting
#' is \code{TRUE}. Set \code{taqEfficiency} to \code{FALSE} if you are using
#' another polymerase than Taq.
#' @param primer_conc Primer concentration.
#' @param na_salt_conc Sodium ion concentration.
#' @param mg_salt_conc Magensium ion concentration.
#' @param k_salt_conc Potassium ion concentration.
#' @param tris_salt_conc Tris ion concentration.
#' @param mode Compute efficiencies for on-target coverage events (\code{on_target})
#' or off-target coverage events (\code{off_target}).
#' @return A list with the efficiency of every primer binding event.
#' @keywords internal
#' @examples
#' data(Ippolito)
#' p <- PCR(settings)
#' # Requires OligoArrayAux software:
#' \dontrun{
#' eff.df <- compute.primer.efficiencies(primer.df, template.df, 55, 
#'              p$primer_concentration, p$Na_concentration,
#'              p$Mg_concentration, p$K_concentration, p$Tris_concentration)
#' } 
compute.primer.efficiencies <- function(primer.df, template.df, annealing.temp, 
    taqEfficiency, primer_conc, na_salt_conc, mg_salt_conc, k_salt_conc, tris_salt_conc, 
    mode = c("on_target", "off_target")) {

    if (!"primer_coverage" %in% colnames(primer.df)) {
        stop("Efficiency computations require primer coverage.")
    }
    if (!check.tool.function()["OligoArrayAux"]) {
        # cannot test without OligoArrayAux
        stop("Cannot compute efficiency: OligoArrayAux not available.")
    }
    if (length(which(!is.na(annealing.temp))) != nrow(primer.df)) {
        stop("Annealing temperatures not provided for all primers.")
    }
    #message("Computing efficiency @ annealing temp: ", annealing.temp)
    #time <- Sys.time()
    if (mode == "on_target") {
        fw.start <- lapply(primer.df$Binding_Position_Start_fw, function(x) as.numeric(unlist(strsplit(x, 
            split = ","))))
        fw.end <- lapply(primer.df$Binding_Position_End_fw, function(x) as.numeric(unlist(strsplit(x, 
            split = ","))))
        rev.start <- lapply(primer.df$Binding_Position_Start_rev, function(x) as.numeric(unlist(strsplit(x, 
            split = ","))))
        rev.end <- lapply(primer.df$Binding_Position_End_rev, function(x) as.numeric(unlist(strsplit(x, 
            split = ","))))
        covered <- lapply(primer.df$Covered_Seqs, function(x) match(as.numeric(unlist(strsplit(x, 
            split = ","))), template.df$Identifier))
    } else {
        fw.start <- lapply(primer.df$Off_Binding_Position_Start_fw, function(x) as.numeric(unlist(strsplit(x, 
            split = ","))))
        fw.end <- lapply(primer.df$Off_Binding_Position_End_fw, function(x) as.numeric(unlist(strsplit(x, 
            split = ","))))
        rev.start <- lapply(primer.df$Off_Binding_Position_Start_rev, function(x) as.numeric(unlist(strsplit(x, 
            split = ","))))
        rev.end <- lapply(primer.df$Off_Binding_Position_End_rev, function(x) as.numeric(unlist(strsplit(x, 
            split = ","))))
        covered <- lapply(primer.df$Off_Covered_Seqs, function(x) match(as.numeric(unlist(strsplit(x, 
            split = ","))), template.df$Identifier))
    }
    seqs <- my.disambiguate(DNAStringSet(template.df$Sequence))
    sodium.eq.concentration <- compute.sodium.equivalent.conc(na_salt_conc, 
                               mg_salt_conc, k_salt_conc, tris_salt_conc)
    fw.effs <- compute.efficiency(primer.df$Forward, fw.start, fw.end, covered, 
                                  taqEfficiency, annealing.temp, primer_conc, 
                                  sodium.eq.concentration,"fw", seqs)
    rev.effs <- compute.efficiency(primer.df$Reverse, rev.start, rev.end, 
                                   covered, taqEfficiency, annealing.temp, primer_conc, 
                                   sodium.eq.concentration, "rev", seqs)
    #message("Runtime: ", Sys.time() - time) 
    # combine efficiciencies from fw and rev primers if necessary
    fw.idx <- which(primer.df$Forward != "")
    rev.idx <- which(primer.df$Reverse != "")
    effs <- lapply(seq_len(nrow(primer.df)), function(j) {
        if (j %in% fw.idx && j %in% rev.idx) {
            # compute the geometric mean of forward and reverse efficiencies
            sqrt(fw.effs[[j]] * rev.effs[[j]])
        } else if (j %in% fw.idx) {
            fw.effs[[j]]
        } else if (j %in% rev.idx) {
            rev.effs[[j]]
        }
    })
    return(effs)
}

#' Settings Functionalities.
#'
#' @rdname Settings
#' @name Settings
#' @description
#' \describe{
#' \item{\code{DesignSettings}}{The \code{DesignSettings} class encapsulates all settings 
#' for designing and evaluating primer sets. 
#' Upon loading an XML file, the \code{DesignSettings} class checks whether
#' the defined constraints can be applied by identifying whether 
#' the requirements for external programs are fulfilled. 
#' If the requirements are not fulfilled, the affected constraints 
#' are removed from the loaded \code{DesignSettings} object
#' and a warning is issued.
#' The loaded constraints are automatically ordered according to
#' the option \code{openPrimeR.constraint_order} such that 
#' the runtime of the \code{\link{design_primers}} and \code{\link{filter_primers}}
#' functions is optimized.}
#' \item{\code{constraints}}{Gets the active constraints of the provided
#' \code{DesignSettings} object.}
#' \item{\code{constraints<-}}{Sets the active constraints of the provided
#' \code{DesignSettings} object.}
#' \item{\code{cvg_constraints}}{Gets the coverage constraints of the provided
#' \code{DesignSettings} object.}
#' \item{\code{cvg_constraints<-}}{Sets the coverage constraints of the provided
#' \code{DesignSettings} object.}
#' \item{\code{conOptions}}{Gets the constraint settings of the provided 
#' \code{DesignSettings} object.}
#' \item{\code{conOptions<-}}{Sets the constraint settings of the provided
#' \code{DesignSettings} object.}
#' \item{\code{constraintLimits}}{Gets the constraint limits that are defined in the provided 
#' \code{DesignSettings} object.}
#' \item{\code{constraintLimits<-}}{Sets the constraint limits of the provided 
#' \code{DesignSettings} object.}
#' \item{\code{PCR}}{Gets the PCR conditions that are defined in the provided
#' \code{DesignSettings} object.}
#' \item{\code{PCR<-}}{Sets the PCR conditions that are defined in the provided
#' \code{DesignSettings} object.}
#' \item{\code{ConstraintSettings}}{The \code{ConstraintSettings} class encapsulates the constraints
#' on the physicochemical properties of primers.}
#' \item{\code{CoverageConstraints}}{The \code{CoverageConstraints} class encapsulates the conditions
#' under which the coverage of primers is evaluated.}
#' \item{\code{PCR_Conditions}}{The \code{PCR_Conditions} class encapsulates the PCR conditions
#' for the computation of primer properties.}
#' \item{\code{ConstraintOptions}}{The \code{ConstraintOptions} class encapsulates the options
#' for constraint computations.}
#' \item{\code{parallel_setup}}{Registers the specified number of cores
#' with the parallel backend.}
#' }
#' @slot Input_Constraints A \code{\link{ConstraintSettings}} object specifying the 
#' desired target value ranges for primer properties. 
#' @slot Input_Constraint_Boundaries A \code{\link{ConstraintSettings}} object specifying
#' the limits for relaxing the constraints during the primer design procedure.
#' This slot may contain the same fields as the \code{Input_Constraints} slot,
#' but the specified desired ranges should be at least as general as
#' those specified in the \code{Input_Constraints} slot.
#' @slot Coverage_Constraints A \code{\link{CoverageConstraints}} object specifying
#' the constraints for computing the primer coverage. 
#' @slot PCR_conditions A \code{\link{PCR_Conditions}} object specifying the PCR-related settings. 
#' @slot constraint_settings A \code{\link{ConstraintSettings}} object providing options 
#' for the computation of individual physicochemical properties.
#' 
#' @slot status Named boolean vector indicating 
#' which of the possible constraints are active (\code{TRUE})
#' and which are not (\code{FALSE}).
#' @slot settings 
#' For \code{ConstraintSettings}, a named list containing the settings for
#' the active constraints.
#' The list may contain the following fields:
#' \describe{
#' \item{primer_coverage:}{The required number of covered template sequences per primer.}
#' \item{primer_specificity:}{The required required specificity of 
#'                            primers in terms of a ratio in the interval [0,1].}
#' \item{primer_length:}{The required lengths of primer sequences.}
#' \item{gc_clamp:}{The desired number of GCs at primer 3' termini.}
#' \item{gc_ratio:}{The desired ratio of GCs in primers 
#'                  in terms of numbers in the interval [0,1].}
#' \item{no_runs:}{The accepted length homopolymer runs in a primer.}
#' \item{no_repeats:}{The accepted length of dinucleotide repeats in a primer.}
#' \item{self_dimerization:}{The lowest acceptable free energy [kcal/mol] for the 
#'                           interaction of a primer with itself. The identification
#'                            of self dimers requires the software \emph{OligoArrayAux} (see notes).}
#' \item{melting_temp_range:}{The desired melting temperature (Celsius) of primers.
#'                           The accurate computation of melting temperatures requires the software \emph{MELTING} (see notes).}
#' \item{melting_temp_diff:}{The maximal allowed difference between the melting temperatures (Celsius)
#'                           of primers contained in the same set. The accurate computation of melting temperatures
#'                           requires the software \emph{MELTING} (see notes).}
#' \item{cross_dimerization:}{The lowest acceptable free energy [kcal/mol] for the
#'                            interaction of a primer with another primer. The identification
#'                            of cross dimers requires the software \emph{OligoArrayAux} (see notes).}
#' \item{secondary_structure:}{The lowest acceptable free energy [kcal/mol] for the
#'                             formation of primer secondary structures. Secondary structures are determined
#'                             using the software \emph{ViennaRNA} (see notes).}
#' }
#' 
#' For \code{PCR_Conditions}, a named list with PCR conditions.
#' The following fields are possible: 
#' \describe{
#' \item{\code{use_taq_polymerase}:}{A logical identifying whether you are performing PCR with a Taq polymerase (\code{TRUE}) or not (\code{FALSE}).}
#' \item{\code{annealing_temp}:}{The annealing temperature in Celsius that is to be used for evaluating the
#' constraints defined in the \code{\link{ConstraintSettings}} object.
#' If the annealing temperature field
#' is not provided, a suitable annealing temperature is automatically computed using a rule of thumb (i.e. subtracting 5 from the melting temperature).}
#' \item{\code{Na_concentration}:}{The molar concentration of monovalent sodium ions.}
#' \item{\code{Mg_concentration}:}{The molar concentration of divalent magnesium ions.}
#' \item{\code{K_concentration}:}{The molar concentration of monovalent potassium ions.}
#' \item{\code{Tris_concentration}:}{The molar concentration of the Tris(hydroxymethyl)-aminomethan buffer.}
#' \item{\code{primer_concentration}:}{The molar concentration of the PCR primers.}
#' \item{\code{template_concentration}:}{The molar concentration of the PCR templates.}
#' }
#' 
#' For \code{CoverageConstraints}, a named list with constraint options. Each 
#' list entry should have an entry \code{min} and/or \code{max}
#' in order to indicate the minimal and maximal allowed values,
#' respectively. 
#' The following identifiers can be used as coverage constraints:
#' \describe{
#' \item{\code{primer_efficiency}:}{The desired efficiencies of primer-template amplification events 
#' in order to be considered as \emph{covered}. \code{primer_efficiency} provides a value in the interval [0,1],
#' which is based on \pkg{DECIPHER}'s thermodynamic model, which considers the impact of 3' terminal mismatches.}
#' \item{\code{annealing_DeltaG}:}{The desired free energies of annealing for putative
#' coverage events between primers and templates. Typically, one would 
#' limit the maximally allowed free energy.}
#' \item{\code{stop_codon}:}{Whether coverage events introducing
#' stop codons into the amplicons should be allowed or discarded. 
#' Here, a value of \code{1} indicates coverage events that induce stop codons.
#' As such, setting both minimum and maximium to zero will disregard
#' coverage events inducing stop codons, while setting the minimum to zero
#' and the maximum to 1 will allow coverage events that induce stop codons.}
#' \item{\code{substitution}:}{Whether coverage events introducing
#' substitutions into the amino acid sequence are considered or discarded.
#' The same encoding as for \code{stop_codon} is used, that is,
#' the value \code{1} indicates coverage events
#' inducing substitutions. Hence, to prevent substitutions,
#' the maximal value of \code{substitution} can be set to zero.}
#' \item{\code{terminal_mismatch_pos}:}{The position relative to 
#' the primer 3' terminal end for which mismatch binding events should be allowed,
#' where the last base in a primer is indicated by position \code{1}.
#' For example, setting the minimal value of \code{terminal_mismatch_pos} 
#' to \code{7} means that only coverage events that do not have a terminal mismatch
#' within the last 6 bases of the primer are allowed.}
#' \item{\code{coverage_model}:}{Use a logistic regression model combining the free energy of annealing and 3' terminal mismatch positions
#' to determine the expected rate of false positive coverage calls. 
#' Using \code{coverage_model}, you can specify the allowed ratio of falsely predicted coverage events.
#' Typically, one would limit the maximal allowed rate of false positives. Note that setting a
#' small false positive rate will reduce the sensitivity of the coverage calls (i.e. true positives will be missed).}
#' }
#' 
#' For \code{ConstraintOptions}, a named list with constraint options.
#' The following fields are permissible:
#' \describe{
#' \item{allowed_mismatches:}{The maximal number of allowed mismatches between
#' a primer and a template sequence. If the number of mismatches of a primer
#' with a template exceeds the specified value, the primer is not considered
#' to cover the corresponding template when the coverage is being computed.}
#' \item{allowed_other_binding_ratio:}{Ratio of allowed binding events
#' outside the target binding ratio. This value should be in the interval
#' [0,1]. If the specified value is greater than zero, all coverage events
#' outside the primer binding region are reported. If, however, the
#' identified ratio of off-target events should exceed the allowed ratio,
#' a warning is issued. If \code{allowed_other_binding_ratio} is set to \code{0},
#' only on-target primer binding events are reported.
#' The setting of \code{allowed_other_binding_ratio} is ignored when designing primers, 
#' which always uses a value of 0.}
#' \item{allowed_region_definition:}{The definition of the target
#' binding regions that is used for evaluating the coverage.
#' In case that \code{allowed_region_definition} is \code{within}, primers have to lie within the allowed binding region.
#' If \code{allowed_region_definition} is \code{any}, primers only have to overlap with the target binding region.}
#' \item{hexamer_coverage:}{If \code{hexamer_coverage} is set to "active", primers whose 3' hexamer (the last 6 bases) is fully complementary to the corresponding
#' template region are automatically considered to cover the template. 
#' If \code{hexamer_coverage} is set to \code{inactive}, 
#' hexamer complementarity does not guarantee template coverage.}
#' }
#'
#' @param x A \code{DesignSettings} object.
#' @param value 
#' An object to be used in one of the setters.
#' For \code{constraints<-} and \code{constraintLimits<-}, a list with constraint settings or boundaries. Each list entry
#' should have a permissible name and consist of at most two
#' values providing the minimal and/or maximal allowed values, which
#' have to be denominated via \code{min} and \code{max}.
#' 
#' For \code{conOptions<-}, a list with constraint options. The permissible
#' fields of the list and their types are documented in the 
#' \code{\link{ConstraintOptions}} class.
#'
#' For \code{cvg_constraints<-}, a list with coverage constraints. Each
#' list entry must have a permissible name and
#' contain a numeric vector with at most two components
#' describing the minimal and/or maximal required values that
#' are to be indicated via \code{min} and \code{max}.
#' The permissible contraint identifiers are documented in the
#' \code{\link{CoverageConstraints}} class.
#' 
#' For \code{PCR<-}, a named list providing PCR conditions 
#' The permissible fields of the list and their types
#' are documented in the \code{\link{PCR_Conditions}} class.
#' @param cores A numeric providing the number of cores to use. The default is \code{NULL}
#' such that half the number of available cores are used.
NULL

#' Primer Design Functionalities.
#'
#' @rdname PrimerDesign
#' @name PrimerDesign
#'
#' @description
#' \describe{
#' \item{\code{design_primers}}{Designs a primer set maximizing the number
#' of covered templates using the smallest possible number of primers. 
#' The algorithm tries to ensure that the designed set of primers 
#' achieves a coverage ratio not lower than \code{required.cvg}.
#' To this end, the constraints for designing primers may be relaxed.}
#' \item{\code{get_initial_primers}}{Creates a set of primer candidates 
#' based on the input template sequences. This set of primers can 
#' be used to create custom primer design algorithms.}
#' }
#' 
#' @param template.df A \code{Templates} object containing the template
#' sequences with annotated primer target binding regions.
#' @param mode.directionality The template strand for which primers shall be designed.
#' Primers can be designed either for forward strands ("fw"), 
#' for reverse strands ("rev"), or for both strands ("both"). The default setting
#' is "both".
#' @param settings A \code{DesignSettings} object specifying the constraint settings for designing primers.
#' @param init.algo The algorithm to be used for initializing primers.
#' If \code{init.algo} is "naive", then primers are constructed from substrings of the input template sequences.
#' If \code{init.algo} is "tree", phylogenetic trees are used to form degenerate primers whose degeneracy is bounded by \code{max.degen}.
#' This option requires an installation of MAFFT (see notes). The default \code{init.algo} is "naive".
#' @param opti.algo The algorithm to be used for solving the primer set covering problem. 
#' If \code{opti.algo} is "Greedy" a greedy algorithm is used to solve the 
#' set cover problem. If \code{opti.algo} is "ILP" an integer linear 
#' programming formulation is used. The default \code{opti.algo} is "Greedy".
#' @param required.cvg The desired ratio of of covered template sequences. 
#' If the target coverage ratio cannot be reached, the constraint settings
#' are relaxed according to the the constraint limits in order to reach the target coverage. 
#' The default \code{required.cvg} is set to 1, indicating that 100\% of the templates are to be covered.
#' @param timeout Timeout in seconds. Only applicable when \code{opti.algo} is "ILP".
#' The default is \code{Inf}, which does not limit the runtime.
#' @param max.degen The maximal degeneracy of primer candidates. This setting is particularly
#' relevant when \code{init.algo} is set to "tree". The default setting is \code{16}, which means
#' that at most 4 maximally degenerate positions are allowed per primer.
#' @param conservation Restrict the percentile of considered regions according to their conservation.
#' Only applicable for the tree-based primer initialization. At the 
#' default of 1, all available binding regions are considered.
#' @param sample.name An identifier for the primer design task. The default setting is
#' \code{NULL}, which means that the run identifier provided in \code{template.df} is used.
#' @param cur.results.loc Directory for storing the results of the primer design procedure.
#' The default setting is \code{NULL} such that no output is stored.
#' @param primer.df An optional \code{Primers} object. If an evaluated \code{primer.df} is provided,
#' the primer design procedure only optimizes \code{primer.df} and does not perform
#' the initialization and filtering steps. The default is \code{NULL} such that
#' primers are initialized and filtered from scratch.
#' @param updateProgress Shiny progress callback function. The default is \code{NULL}
#' such that no progress is logged.
#' @param primer.length A scalar numeric providing the 
#' target length of the designed primers. The default length 
#' of generated primers is set to \code{18}.
#' @param primer.estimate Whether the number of required primers shall be estimated. By default (\code{FALSE}), the number of required primers is not estimated.
#' @param sample Character vector providing an identifier for the templates.
#' @param primer.lengths Numeric vector of length 2 providing the 
#' minimal and maximal allowed lengths for generated primers.
#' @param allowed.region.definition A character vector providing the definition
#' of region where primers are to be constructed.
#' If \code{allowed.region.definition} is "within", constructed primers lie within the allowed binding region.
#' If \code{allowed.region.definition} is "any", primers overlap with the allowed binding region.
#' The default is "within".
#' 
#' @note Some constraints can only be computed if additional software is installed,
#' please see the documentation of \code{\link{DesignSettings}} for more information.
#' The usage of \code{init.algo = "tree"} requires an installation of
#' the multiple alignment program MAFFT 
#' (http://mafft.cbrc.jp/alignment/software/).
NULL

#######
# Melting temperature computations
######

#' Change in Free Energy.
#'
#' Computes the change in free energy.
#'
#' @param delta.H Change in enthalpy in cal/mol.
#' @param delta.S Change in entropy cal/mol*K.
#' @param temp Temperature in Celsius for which to compute free energy change.
#'
#' @return The change in free energy in kcal/mol.
#' @keywords internal
get.delta.G <- function(delta.H, delta.S, temp = 37) {
    K <- convert.temperature(temp, "K")
    G <- (delta.H - (K * delta.S))/1000
    return(G)  # in kcal/mol
}
#' Conversion from J to cal
#'
#' Converts the input from Joule to calories.
#'
#' @param val.J Numeric Joule value.
#'
#' @return The value correspdoning to \code{val.J} in calories.
#' @keywords internal
joule.to.cal <- function(val.J) {
    return(val.J/4.184)
}
#' Conversion between Celsius and Kelvin
#'
#' Converts the input from Kelvin to Celsius or from Celsius to Kelvin.
#'
#' If \code{temp.scale} is 'K', \code{T_m} is transformed from Celsius
#' to Kelvin. If \code{temp.scale} is 'C', \code{T_m} is transformed from
#' Kelvin to Celsius. The default is to transform from Celsius to Kelvin.

#' @param temp The input temperature.
#' @param temp.scale The desired unit of the output temperature.
#'
#' @return Transforms the input temperature to the specified \code{temp.scale}.
#' @keywords internal
convert.temperature <- function(temp, temp.scale = c("K", "C")) {
    # scale: kelvin -> convert to kelvin scale: celsius -> convert to celsius
    if (length(temp.scale) == 0) {
        stop("Please supply the 'temp.scale' parameter.")
    }
    temp.scale <- match.arg(temp.scale)
    if (temp.scale == "K") {
        return(temp + 273.15)
    } else if (temp.scale == "C") {
        return(temp - 273.15)
    } 
}

#' Baldino Formula
#'
#' Computes the melting temperature using the formulation by Baldino.
#'
#' @param sequences Input sequence strings.
#' @param na_salt_conc Sodium ion concentration.
#' @param mg_salt_conc Magensium ion concentration.
#' @param k_salt_conc Potassium ion concentration.
#' @param tris_salt_conc Tris buffer concentration.
#' @param primer_conc Primer concentration.
#'
#' @return The melting temperature for the input sequences.
#' @references 
#' Rychlik, W. J. S. W., W. J. Spencer, and R. E. Rhoads. "Optimization of the annealing temperature for DNA amplification in vitro." Nucleic acids research 18.21 (1990): 6409-6412.
#' @keywords internal
compute.Tm.baldino <- function(sequences, na_salt_conc, mg_salt_conc, k_salt_conc, 
    tris_salt_conc, primer_conc) {
    l <- nchar(sequences)  # assume longest length of the PCR product
    gc.ratio <- compute.gc.ratio(sequences)
    sodium.eq.concentration <- compute.sodium.equivalent.conc(na_salt_conc, mg_salt_conc, 
                                            k_salt_conc, tris_salt_conc)  
    T_mC <- (0.41 * (gc.ratio * 100)) + 
            (16.6 * log10(sodium.eq.concentration*1000)) - (675/l)
    return(T_mC)
}
#' Non-Thermodynamic Computation of Melting Temperatures.
#'
#' Computes the melting temperature of primers from an empiric formula.
#'
#' @param primer.df A \code{Primers} object.
#' @return A data frame with melting temperature information for the primers.
#' @keywords internal
compute.empiric.melting.temp <- function(primer.df) {
    fw.seqs <- convert.from.iupac(primer.df$Forward)
    rev.seqs <- convert.from.iupac(primer.df$Reverse)
    Tm <- function(x) {
        if (x[[1]] == "") {
            return(NA)
        }
        x <- strsplit(x, split = "")
        A <- sapply(x, function(y) length(which(y == "a")))
        C <- sapply(x, function(y) length(which(y == "c")))
        G <- sapply(x, function(y) length(which(y == "g")))
        T.count <- sapply(x, function(y) length(which(y == "t")))
        # return the smallest reuired melting temperature
        Tm <- 64.9 + 41 * (G + C - 16.4) / (A + T.count + G + C) 
        return(min(Tm))
    }
    Tm.fw <- sapply(fw.seqs, Tm)
    Tm.rev <- sapply(rev.seqs, Tm)
    Tm.K.fw <- convert.temperature(Tm.fw, "K")
    Tm.K.rev <- convert.temperature(Tm.rev, "K")
    # return overall minimal Tm of both directions:
    melting.temp <- sapply(seq_len(nrow(primer.df)), function(x) min(Tm.fw[x], Tm.rev[x], na.rm = TRUE))
    Tm.data <- data.frame("melting_temp" = melting.temp, "Tm_C_fw" = Tm.fw, "Tm_C_rev" = Tm.rev, "Tm_K_fw" = Tm.K.fw, "Tm_K_rev" = Tm.K.rev)
    return(Tm.data)
}

#' Computation of Thermodynamic Melting Temperatures.
#'
#' Use nearest-neighbor thermodynamic computations to find melting temperatures.
#'
#' @param primer.df Primer data frame.
#' @param primer_conc Primer concentration.
#' @param na_salt_conc Sodium ion concentration.
#' @param mg_salt_conc Magensium ion concentration.
#' @param k_salt_conc Potassium ion concentration.
#' @param tris_salt_conc Tris buffer concentration.
#' @param mode.directionality Direction of primers
#' @return Data frame with melting temperature info for the input primers.
#' @keywords internal
compute.melting.temps.thermo <- function(primer.df, primer_conc, na_salt_conc, mg_salt_conc, 
    k_salt_conc, tris_salt_conc, mode.directionality = c("fw", "rev", "both")) {

    used.Tm <- rep(NA, nrow(primer.df))  # Tm to be used in optimization algorithm
    Tm.fw <- call.melt(primer.df$Forward, complement.sequence(primer.df$Forward), 
        primer_conc, na_salt_conc, mg_salt_conc, k_salt_conc, tris_salt_conc)[, 3:6]
    Tm.rev <- call.melt(primer.df$Reverse, complement.sequence(primer.df$Reverse), 
        primer_conc, na_salt_conc, mg_salt_conc, k_salt_conc, tris_salt_conc)[, 3:6]
    cnames <- NULL
    if (mode.directionality == "fw" || mode.directionality == "both" && length(Tm.fw) != 0) {
        cnames <- colnames(Tm.fw)
    } else {
        cnames <- colnames(Tm.rev)
    }
    # set empty data frame to NA
    if (length(Tm.fw) != 0) {
        colnames(Tm.fw) <- paste(cnames, "_fw", sep = "")
    } else if (length(Tm.fw) == 0 && length(Tm.rev) != 0) {
        Tm.fw <- Tm.rev
        Tm.fw[TRUE] <- NA
        colnames(Tm.fw) <- paste(cnames, "_fw", sep = "")
    }
    # set empty data frame to NA
    if (length(Tm.rev) != 0) {
        colnames(Tm.rev) <- paste(cnames, "_rev", sep = "")
    } else if (length(Tm.rev) == 0 && length(Tm.fw) != 0) {
        Tm.rev <- Tm.fw
        Tm.rev[TRUE] <- NA
        colnames(Tm.rev) <- paste(cnames, "_rev", sep = "")
    }
    if (length(Tm.fw) != 0 && length(Tm.rev) != 0) {
        used.Tm <- unlist(lapply(seq_len(nrow(Tm.fw)), function(x) min(c(Tm.fw$Tm_C_fw[x], Tm.rev$Tm_C_rev[x]), 
            na.rm = TRUE)))  # use the min Tm 
    } else if (length(Tm.fw) != 0 && length(Tm.rev) == 0) {
        used.Tm <- Tm.fw$Tm_C_fw
    } else if (length(Tm.rev) != 0 && length(Tm.fw) == 0) {
        used.Tm <- Tm.rev$Tm_C_rev
    } else {
        used.Tm <- NULL
    }
    Tm.data <- cbind(Tm.fw, Tm.rev, melting_temp = used.Tm)
    return(Tm.data)
}
#' Computation of Melting Temperatures.
#'
#' Use nearest-neighbor thermodynamic computations to find melting temperatures.
#'
#' @param primer.df Primer data frame.
#' @param primer_conc Primer concentration.
#' @param na_salt_conc Sodium ion concentration.
#' @param mg_salt_conc Magensium ion concentration.
#' @param k_salt_conc Potassium ion concentration.
#' @param tris_salt_conc Tris buffer concentration.
#' @param mode.directionality Direction of primers
#' @return Data frame with melting temperature info for the input primers.
#' @keywords internal
compute.melting.temps <- function(primer.df, primer_conc, na_salt_conc, mg_salt_conc, 
    k_salt_conc, tris_salt_conc, mode.directionality = c("fw", "rev", "both")) {

    if (length(mode.directionality) == 0) {
        stop("Please supply the 'mode.directionality' argument.")
    }
    mode.directionality <- match.arg(mode.directionality)
    if (check.tool.function()["MELTING"]) {
        # melting is available -> use thermodynamic model
        Tm.data <- compute.melting.temps.thermo(primer.df, primer_conc, na_salt_conc, mg_salt_conc, 
                            k_salt_conc, tris_salt_conc, mode.directionality)
    } else {
        # melting is not available -> use empiric formula without salt correction
        Tm.data <- compute.empiric.melting.temp(primer.df)
    }
    # determine max temperature diff for every primer
    temp.diff <- get.melting.temp.diff(Tm.data$Tm_C_fw, Tm.data$Tm_C_rev)
    result <- cbind(Tm.data,melting_temp_diff = temp.diff)
    return(result)
}
#' Computation of Maximal Melting Temperature Differences.
#'
#' @param Tm.fw The melting temperatures of forward primers.
#' @param Tmr.rev The melting temperatures of reverse primers.
#' @return The worst-case melting temperature difference, for every primer.
#' @keywords internal
get.melting.temp.diff <- function(Tm.fw, Tm.rev) {
    if (length(Tm.fw) != 0) {
        fw.idx <- which(!is.na(Tm.fw))
    } else {
        fw.idx <- NULL
    }
    if (length(Tm.rev) != 0) {
        rev.idx <- which(!is.na(Tm.rev))
    } else {
        rev.idx <- NULL
    }
    max.len <- max(length(Tm.fw), length(Tm.rev))
    temp.diff.fw <- rep(0, max.len)
    temp.diff.rev <- rep(0, max.len)
    if (length(fw.idx) != 0) {
        temp.diff.fw[fw.idx] <- sapply(fw.idx, function(x) max(c(abs(Tm.fw[x] - Tm.fw), abs(Tm.fw[x] - Tm.rev)), na.rm = TRUE))
    }
    if (length(rev.idx) != 0) {
        temp.diff.rev[rev.idx] <- sapply(rev.idx, function(x) max(c(abs(Tm.rev[x] - Tm.rev), abs(Tm.rev[x] - Tm.fw)), na.rm = TRUE))
    }
    temp.diff <- sapply(seq_along(temp.diff.fw), function(x) max(c(temp.diff.fw[x], temp.diff.rev[x])))
    return(temp.diff)
}
#' Thermodynamic melting temperature computations.
#'
#' Computes the melting temperature for the input primers.
#'
#' @param primers List of primer strings.
#' @param complements List with corresponding complements.
#' @param out.file Path to the file where MELTING will write the results.
#' @param primer_conc Primer concentration.
#' @param na_salt_conc Sodium ion concentration.
#' @param mg_salt_conc Magensium ion concentration.
#' @param k_salt_conc Potassium ion concentration.
#' @param tris_salt_conc Tris buffer concentration.
#' @param ID identifiers of the input primers
#' @return Melting temperature data frame. 
#' @keywords internal
call.melt.single <- function(primers, complements, out.file, primer_conc, na_salt_conc, 
    mg_salt_conc, k_salt_conc, tris_salt_conc, ID) {
    # call melt command ID: used to keep track of ambiguous seqs that were generated
    if (length(primers) == 0 || length(complements) == 0) {
        return(NULL)
    }
    melt <- Sys.which("melting-batch")
    if (melt == "") {
        stop("MELTING not available.")
    } else {
		if (grepl("windows", .Platform$OS.type)) {
			# for windows, use the bat file
			jar.file <- Sys.which("melting5.jar")
			melt <- paste("java -cp", jar.file, "melting.BatchMain")
		}
		#message(paste("Using the following melting location: ", melt))
	}
    # melt results from b atch are in joul -> convert to cal use Allawi, SantaLucia
    # 1997 publication nearest neighbor data
    nn <- "all97"  # nearest neighbor model
    am <- "wetdna91"  # from Wetmur1991, used for more than 60 bases primers
    # standard parameters: -P: nucleic acid strand in excess (template or primers) -E
    # <compound>: concentration of compound in molar if [Mg] is 0, don't add the
    # parameter
    salt.cmd <- paste(" -E Na=", format(na_salt_conc, scientific = FALSE), sep = "")
    if (mg_salt_conc != 0) {
        salt.cmd <- paste(salt.cmd, " -E Mg=", format(mg_salt_conc, scientific = FALSE), 
            sep = "")
    }
    if (k_salt_conc != 0) {
        salt.cmd <- paste(salt.cmd, " -E K=", format(k_salt_conc, scientific = FALSE), 
            sep = "")
    }
    if (tris_salt_conc != 0) {
        salt.cmd <- paste(salt.cmd, " -E Tris=", format(tris_salt_conc, scientific = FALSE), 
            sep = "")
    }
    # don't use custom tandem file anymore, not necessary if we don't do mismatch Tm computation.
    #tandem.mm.file <- "AllawiSantaluciaPeyret1997_1998_1999tanmm_mod.xml"
    melt.cmd <- paste(melt, " -am ", am, " -nn ", nn, " -H dnadna -P ", format(primer_conc, 
        scientific = FALSE), salt.cmd, #" -tan :", tandem.mm.file, 
        " ", out.file, sep = "")
    #message(melt.cmd)
    system(melt.cmd, ignore.stdout = TRUE)  # don't igore stdout for debugging
    # read results
    result.file <- paste(out.file, ".results.csv", sep = "")
    raw.data <- readChar(result.file, file.info(result.file)$size)
    # modify the header change a separator from \r to \t
    raw.data <- gsub("\r", "\t", raw.data)
	# remove terminal tabulators:
	raw.data <- gsub("\t\n", "\n", raw.data)
    results <- try(read.delim(text = raw.data, header = TRUE, 
                stringsAsFactors = FALSE), silent = FALSE)
    # raw data is empty sometime. why if there's input -> melting fails i guess?
    if (is(results, "try-error")) {
        warning("MELTING could not compute the Tm:\n", 
            "File: ", result.file, "\n", 
            attr(results, "condition"))
        e <- rep(NA, length(primers))
        return(data.frame(Sequence = e, Complementary = e, DeltaH = e, DeltaS = e, 
            Tm_C = e, Tm_K = e, ID = ID))
    }
    # determine system locale settings:
    decimal.sep <- Sys.localeconv()["mon_decimal_point"]
    if (decimal.sep == ".") {
        decimal.sep <- "\\."
        mill.sep <- ","
    } else {
        mill.sep <- "\\."
    }
    # replace 1000pos commas and convert from J to calories
    results[, 3] <- joule.to.cal(as.numeric(gsub(decimal.sep, ".", gsub(mill.sep, 
        "", results[, 3]))))
    results[, 4] <- joule.to.cal(as.numeric(gsub(decimal.sep, ".", gsub(mill.sep, 
        "", results[, 4]))))
    results[, 5] <- as.numeric(gsub(decimal.sep, ".", gsub(mill.sep, "", results[, 
        5])))
    colnames(results)[5] <- "Tm_C"
    results <- cbind(results, Tm_K = convert.temperature(results$Tm_C, "K"))
    ######### insert missing sequence results, e.g. for '' seqs
    sel <- rep(NA, length(primers))
    for (i in seq_along(primers)) {
        primer <- primers[i]
        comp <- complements[i]
        idx.s <- which(sapply(results$Sequence, function(x) x == primer))
        idx.c <- which(sapply(results$Complementary, function(x) x == comp))
        idx <- intersect(idx.s, idx.c)
        if (length(idx) != 0) {
            sel[i] <- idx[1]
        }
    }
    results <- results[sel, ]
    # insert sequence and complementary for missing entries
    missing.idx <- which(is.na(results$Sequence))
    if (length(missing.idx) != 0) {
        results[missing.idx, "Sequence"] <- primers[missing.idx]
        results[missing.idx, "Complementary"] <- complements[missing.idx]
    }
	# prepare output:
    results <- cbind(results, ID = ID)
    results$Sequence <- tolower(results$Sequence)
    results$Complementary <- tolower(results$Complementary)
    return(results)
}

#' Thermodynamic melting temperature computations.
#'
#' Computes the melting temperature for the input primers.
#'
#' @param primers Character vector of primer strings.
#' @param complements Character vector with complement sequences
#' corresponding to \code{primers}.
#' @param primer_conc Primer concentration.
#' @param na_salt_conc Sodium ion concentration.
#' @param mg_salt_conc Magensium ion concentration.
#' @param k_salt_conc Potassium ion concentration.
#' @param tris_salt_conc Tris buffer concentration.
#' @return Melting temperature data frame. 
#' @keywords internal
#' @references Le Novère N. (2001). MELTING, computing the 
#' melting temperature of nucleic acid duplex. Bioinformatics, 17: 1226-1227. 
#' Dumousseau M., Rodriguez N., Juty N., Le Novère N. (2012) MELTING, 
#' a flexible platform to predict the melting temperatures of nucleic acids. 
#' BMC Bioinformatics, 13: 101.
call.melt <- function(primers, complements, primer_conc, na_salt_conc, mg_salt_conc, 
    k_salt_conc, tris_salt_conc) {
    # complement needs to be specified for mismatch consideration 
    # ./melting -S TGAGGTGCAGCTGGTGGAGTC -H dnadna -P 0.00000005 -E Na=0.05 -v -C ACTCCACGCCGACCACCTCAA 
    # ./melting-batch -H dnadna -P 0.00000005 -E Na=0.05
    # primer_rev_comp.fasta complement from 3' to 5' 1. create file for melt
    if (all(primers == "") || length(primers) == 0 || length(complements) == 0) {
        return(NULL)
    }
    #print("call.melt: primer input")
    #print(primers)
    runID <- digest::digest(c(primers, complements), "md5")
    if (length(primers) != length(complements)) {
        stop("The number of primers did not correspond to the number of complements.")
    }
    if (Sys.which("melting-batch") == "") {
        stop("MELTING executable not available. Please install MELTING to compute melting temperatures.")
    }
    idx <- which(primers != "")
    out.len <- length(primers)
    primers <- primers[idx]
    complements <- complements[idx]
    # disambiguate primers
    unambig.primers <- convert.from.iupac(primers)
    # determine idx of exploded primers
    unambig.primers.c <- convert.from.iupac(complements)  # unambiguated complements
    # need to get all combinations of unambig primers with unambig.primers.c
    ambig.combinations <- lapply(seq_along(unambig.primers), function(x) expand.grid(unambig.primers[[x]], 
        unambig.primers.c[[x]], stringsAsFactors = FALSE))
    ids <- unlist(lapply(seq_along(ambig.combinations), function(x) rep(x, nrow(ambig.combinations[[x]]))))
    out.primers <- unlist(lapply(seq_along(ambig.combinations), function(x) toupper(ambig.combinations[[x]][, 
        1])))
    out.complements <- unlist(lapply(seq_along(ambig.combinations), function(x) toupper(ambig.combinations[[x]][, 
        2])))
    # MELTING can't handle terminal mismatches, remove those
    check.comp <- toupper(complement.sequence(out.complements))
    rm.idx <- which(substring(check.comp, 1, 1) != substring(out.primers, 1, 1) |
        substring(check.comp, nchar(check.comp), nchar(check.comp)) !=
        substring(out.primers, nchar(out.primers), nchar(out.primers)))
    if (length(rm.idx) != 0) {
        out.primers <- out.primers[-rm.idx]
        ids <- ids[-rm.idx]
        out.complements <- out.complements[-rm.idx]
    }
    melt.tab <- data.frame(Identifier = ids, Primer = out.primers, Complement = out.complements)
    # write one file for each parallel call
    seqs.per.file <- ceiling(length(out.primers)/getDoParWorkers())
    no.files <- ceiling(length(out.primers)/seqs.per.file)
    out.files <- rep(NA, no.files)
    total.seqs.covered <- 0
    seq.idx <- vector("list", no.files)  # idx of the primers in out.primers per file
    # write out.files
    for (i in seq_len(no.files)) {
        seqs.added <- min(seqs.per.file, length(out.primers) - total.seqs.covered)
        s <- NULL
        if (i == 1) {
            s <- 1
        } else {
            s <- max(seq.idx[[i - 1]]) + 1
        }
        e <- s + (seqs.added - 1)
        seq.idx[[i]] <- s:e
        total.seqs.covered <- total.seqs.covered + seqs.added
        out.file <- tempfile(paste("melt_seqs_", i, runID, sep = ""), fileext = ".txt")
        out.files[i] <- out.file
        #print("melt Seqs:")
        #print(melt.tab[seq.idx[[i]],])
        write.table(melt.tab[seq.idx[[i]], c("Primer", "Complement")], file = out.file, 
					sep = "\t", quote = FALSE, col.names = FALSE, row.names = FALSE)
    }
    # parallelized call:
    i <- NULL
    results <- foreach(i = seq_along(out.files), .combine = "rbind") %dopar% {
        result <- call.melt.single(out.primers[seq.idx[[i]]], out.complements[seq.idx[[i]]], 
            out.files[i], primer_conc, na_salt_conc, mg_salt_conc, k_salt_conc, tris_salt_conc, 
            ids[seq.idx[[i]]])
    }
    # break results down to individual sequences using the ID column to integrate
    # ambig seq results
    u.IDs <- unique(results$ID)
    results$deltaG <- get.delta.G(results$DeltaH, results$DeltaS)
    # integrate by selecting worst-case deltaG per ambig seq combination
    results <- ddply(results, c("ID"), function(x) arrange(x, substitute(deltaG))[1, ])  # select unique pairs
    # replace Sequence/Complementary stretch with original input (ambiguous)
    results[, "Sequence"] <- primers[results$ID]
    results[, "Complementary"] <- complements[results$ID]
    results <- results[, colnames(results) != "ID"]  # remove ID column
    # create entries for primers without this direction ("")
    out <- results[1,]
    out[,] <- NA
    out <- do.call(rbind, replicate(out.len, out, simplify = FALSE))
    out[idx,] <- results
	# clean up temporary files
	on.exit(file.remove(out.files))
    return(out)
}

#' Sodium-equivalent Concentration
#'
#' Computes the sodium-equivalent concentration for the input ion concentrations.
#'
#' @param na_salt_conc Sodium ion concentration.
#' @param mg_salt_conc Magensium ion concentration.
#' @param k_salt_conc Potassium ion concentration.
#' @param tris_salt_conc Tris buffer concentration.
#' @return The sodium-equivalent concentration of the input ion concentrations.
#' @references
#' Record, M. Thomas. "Effects of Na+ and Mg++ ions on the helix–coil transition of DNA." 
#' Biopolymers 14.10 (1975): 2137-2158.
#'
#' Owczarzy, Richard, et al. "Predicting stability of DNA duplexes in solutions containing magnesium and monovalent cations." 
#' Biochemistry 47.19 (2008): 5336-5353.
#'
#' Peyret, Nicolas. Prediction of nucleic acid hybridization: parameters and algorithms. 
#' Detroit: Wayne State University, 2000.
#' @keywords internal
compute.sodium.equivalent.conc <- function(na_salt_conc, mg_salt_conc, k_salt_conc, 
    tris_salt_conc) {
    # beta: correction factor from Peyret (2000)
    beta <- 3.3
    na_salt_eq <- beta * sqrt(mg_salt_conc) + na_salt_conc + k_salt_conc + tris_salt_conc / 2 # sodium equivalent concentration
    return(na_salt_eq)
}

#' Plotting Functions.
#'
#' @rdname Plots
#' @name Plots
#' @description
#' \describe{
#' \item{\code{plot_cvg_vs_set_size}}{Plots the coverage ratios of the input primer sets 
#' against the size of the sets.}
#' \item{\code{plot_penalty_vs_set_size}}{Plots the penalties of the input primer sets 
#' against the number of primers contained in each set.
#' The penalties are computed using \code{\link{score_primers}}
#' where more information is provided on how to set \code{alpha}.}
#' \item{\code{plot_primer_subsets}}{Visualizes the coverage of optimized primer subsets.}
#' \item{\code{plot_primer}}{Visualizes the binding positions of every primer relative to
#' the target binding region in the corresponding template sequences.}
#' \item{\code{plot_template_cvg}}{Creates a bar plot visualizing the covered templates.}
#' \item{\code{plot_primer_cvg}}{Shows which groups of templates are covered by individual primers.}
#' \item{\code{plot_constraint}}{Shows the distribution of the primer properties.
#' The current constraint settings are indicated with dashed lines.}
#' \item{\code{plot_constraint_fulfillment}}{Visualizes which primers pass the constraints 
#' and which primers break the constraints}
#' \item{\code{plot_cvg_constraints}}{Plots the distribution of the coverage constraint values.}
#' \item{\code{plot_constraint_deviation}}{Plots the deviation of primer properties from the target ranges.}
#' \item{\code{plot_primer_binding_regions}}{Visualizes the number of binding events of the primers
#' with respect to the allowed binding regions in the templates.}
#' \item{\code{plot_conservation}}{Plots the template sequence conservation (range [0,1]) according to
#' the Shannon entropy of the sequences.}
#' }
#'
#' @param primer.df An object of class \code{Primers} containing
#' primers with evaluated primer coverage.
#' @param template.df An object of class \code{Templates} containing the
#' template sequences.
#' @param primer.data List with objects of class \code{Primers} containing
#' the primer sets that are to be compared.
#' @param primer.subsets A list with optimal primer subsets, each of class \code{Primers}. The \code{k}-th list entry should correspond to an object of class \code{Primers}
#' representing the primer subset of size \code{k} whose coverage ratio
#' is the largest among all possible subsets of size \code{k}.
#' @param template.data List with objects of class \code{Templates} containing
#' the templates corresponding to \code{primer.data}.
#' @param primers Either a single \code{Primers} object with evaluated primer coverage
#' or a list containing such \code{Primers} objects.
#' @param templates If \code{primers} is a \code{Primers} object, \code{templates} should be a \code{Templates} object.
#' If \code{primers} is a list, then \code{templates} should be a list of \code{Templates} objects.
#' @param per.mismatch A logical specifying whether the visualization should be stratified
#' according to the allowed number of mismatches. By default,
#' \code{per.mismatch} is set to \code{FALSE} such that the overall coverage
#' is plotted.
#' @param settings An object of class \code{DesignSettings} containing
#' the constraints to be considered.
#' @param active.constraints A character vector containing the identifiers
#' to be considered for plotting. By default, \code{active.constraints} is
#' \code{NULL} such that all computed constraints found in \code{settings} are plotted.
#' @param show.labels Whether the identifiers of the primer sets
#' should be annotated in the plot. The default is \code{TRUE}.
#' @param highlight.set A character vector providing the identifiers
#' of primer sets to highlight. By default, \code{highlight.set} is \code{NULL}
#' such that no highlighting takes place.
#' @param alpha A numeric in the range [0,1] defining the trade-off between
#' the maximal deviation of a constraint (large \code{alpha}) and
#' all constraint deviations (large \code{alpha}).
#' By default, \code{alpha} is set to 0 such that the absolute
#' deviation across all constraints is considered.
#' @param required.cvg The required coverage ratio.
#' The default is 100\%; this value is plotted as a horizontal line.
#' @param identifier Identifiers of primers that are to be considered.
#' If \code{identifier} is set to \code{NULL} (the default), all primers are considered.
#' @param relation Whether binding positions are computed relative to forward ("fw")
#' or reverse ("rev") binding regions. The default is "fw".
#' @param region.names Character vector of length 2 providing the names
#' of the binding and amplification region.
#' @param ... Optional arguments \code{groups} (a character vector of groups to be plotted when \code{primers} is a single primer set), 
#' \code{highlight.set} (the identifier of a primer set to be highlighted when \code{primers} is a list), 
#' \code{ncol} (a numeric indicating the number of facet columns if \code{primers} is a list),
#' \code{deviation.per.primer} (a boolean indicating whether constraint deviations
#' should be plotted per primer rather than per constraint if \code{primers} is a list)
#' @param direction The directionality of primers to be plotted. This can either
#' be "both" to plot primers of both directions (the default), "fw" to plot
#' only forward primers, or "rev" to plot only reverse primers.
#' @param plot.p.vals An optional logical argument indicating whether
#' p-values computed via \code{\link{primer_significance}} should be annotated in the plot. The default is \code{FALSE}.
#' @param group Optional identifiers of template groups for which binding events should
#' be determined. By default, \code{group} is set to \code{NULL} such that
#' all templates are considered.
#' @param entropy.df A data frame with entropies.
#' Each row gives the entropies of a group of related template
#' sequences for all columns of the alignment.
#' @param alignments A list with \code{DNABin} alignment objects
#' corresponding to the groups (rows) in the alignment.
#' @param gap.char The gap char in the alignments. By default,
#' \code{gap.char} is set to "-".
NULL
 

############
# Initialization of Primers
############

#' Primer Length Check.
#'
#' Checks whether it is possible to construct primers of the desired length.
#'
#' @param template.df Template data frame.
#' @param allowed.region.definition Definition of allowed binding regions.
#' @param primer.lengths The desired lengths of the priemrs.
#' @param mode.directionality The primer directionality.
#' @return TRUE, if primers of the desired length can be constructed, 
#' @keywords internal
#' FALSE otherwise.
check.init.primer.length <- function(template.df, 
    allowed.region.definition = c("within", "any"), primer.lengths, 
    mode.directionality = c("fw", "rev", "both")) {

    if (length(allowed.region.definition) == 0) {
        stop("Please supply the 'allowed.region.definition' arg.")
    }
    allowed.region.definition <- match.arg(allowed.region.definition)
    if (length(mode.directionality) == 0) {
        stop("Please supply the 'mode.directionality' arg.")
    }
    mode.directionality <- match.arg(mode.directionality)
    if (length(template.df) == 0) {
        stop("No templates available.")
    }
    if (length(primer.lengths) == 0) {
        stop("No primer length specified.")
    }
    check <- NULL
    if (mode.directionality == "fw") {
        check <- check.init.primer.length.single(template.df$Allowed_fw, allowed.region.definition, 
            min(primer.lengths))
    } else if (mode.directionality == "rev") {
        check <- check.init.primer.length.single(template.df$Allowed_rev, allowed.region.definition, 
            min(primer.lengths))
    } else {
        check.fw <- check.init.primer.length.single(template.df$Allowed_fw, allowed.region.definition, 
            min(primer.lengths))
        check.rev <- check.init.primer.length.single(template.df$Allowed_rev, allowed.region.definition, 
            min(primer.lengths))
        check <- check.fw && check.rev
    }
    return(check)
}
#' Primer Length Check.
#'
#' Checks whether it is possible to construct primers of the desired length.
#'
#' @param allowed String containing the allowed binding sequence.
#' @param allowed.region.definition Definition of allowed binding regions.
#' @param min.len Minimal desired primer lengths.
#' @return TRUE if primers of the desired length can be constructed,
#' FALSE otherwise.
#' @keywords internal
check.init.primer.length.single <- function(allowed, 
    allowed.region.definition = c("within", "any"), min.len) {
    # allowed: allowed binding region min.len: minimal primer length required by user
    if (length(allowed.region.definition) == 0) {
        stop("Please supply the 'allowed.region.definition' arg.")
    }
    allowed.region.definition <- match.arg(allowed.region.definition)
    if (allowed.region.definition == "within" && any(nchar(allowed) < min.len)) {
        warning("Selected region is too short to create primers for all templates.")
        return(FALSE)
    } else {
        return(TRUE)
    }
}
#' Creation of an Initial Primer Set.
#'
#' Creates an initial set of candidate primers for primer design.
#'
#' @param template.df Template data frame.
#' @param primer.lengths Vector containing the permissible primer lengths.
#' @param mode.directionality Direction of primers to be created.
#' @param sample Name of the template sample.
#' @param allowed.region.definition Definition of the allowed binding region.
#' @param init.algo Algorithm for initializing primers.
#' @param max.degen Maximal allowed degeneration of created primers.
#' @param conservation Required conservation of primers.
#' The value of \code{conservation} should be in the range[0,1].
#' @param updateProgress Shiny progress object.
#' @return An initialized data frame of candidate primers.
#' @keywords internal
create.initial.primer.set <- function(template.df, primer.lengths, mode.directionality = c("fw", "rev"), 
    sample, allowed.region.definition = c("within", "any"), 
    init.algo = c("naive", "tree"), max.degen, conservation, updateProgress = NULL) {

    if (length(init.algo) == 0) {
        stop("Please supply the 'init.algo' arg.")
    }
	init.algo <- match.arg(init.algo)
    if (length(mode.directionality) == 0) {
        stop("Please supply the 'mode.directionality' arg.")
    }
	mode.directionality <- match.arg(mode.directionality)
    if (length(allowed.region.definition) == 0) {
        stop("Please supply the 'allowed.region.definition' arg.")
    }
    allowed.region.definition <- match.arg(allowed.region.definition)
    seqs <- template.df$Sequence
    if (mode.directionality == "fw") {
        # Set the positions in the sense sequence
        l.s <- template.df$Allowed_Start_fw
        e.s <- template.df$Allowed_End_fw
    } else if (mode.directionality == "rev") {
        # Start primer design start/end in reverse complement sequence
        l.s <- nchar(template.df$Sequence) - template.df$Allowed_End_rev + 1 # smaller number (e.g. 1)
        e.s <- nchar(template.df$Sequence) - template.df$Allowed_Start_rev + 1 # larger number (e.g. 30)
        seqs <- rev.comp.sequence(seqs)
    }
    # if l.s/e.s is NA -> create primers for the full sequence
    l.s[is.na(l.s)] <- 1
    e.s[is.na(e.s)] <- nchar(seqs)[is.na(e.s)]
    # ensure that required primer lengths can be fulfilled
    if (allowed.region.definition == "within") {
        if (any(max(primer.lengths) > (e.s - l.s + 1))) {
            warning("The specified maximal primer length ('", max(primer.lengths),
                    "') exceeds the length of the shortest target binding region ('", min(e.s - l.s +1), "').")
            # select only feasible primer lengths
            sel <- which(primer.lengths <= min(e.s - l.s + 1))
            primer.lengths <- primer.lengths[sel]
        }
    } else {
        # any intersection with binding range: primer length may not exceed the sequence length
        end <- NULL
        if (mode.directionality == "fw") {
            end <- max(template.df$Allowed_Start_fw)
        } else {
            end <- max(template.df$Allowed_End_rev - nchar(template.df$Sequence)) + 1
        }
        if (any(max(primer.lengths) > (nchar(template.df$Sequence) - end + 1))) {
            warning("Specified maximal primer length ('", max(primer.lengths), "') exceeded the length of at least one sequence ('", min(nchar(template.df$Sequence)), "').")
            sel <- which(primer.lengths <= min(nchar(template.df$Sequence) - end + 1))
            primer.lengths <- primer.lengths[sel]

        }
    }
    # change group IDs to be unique
    group.tags <- uniqtag::uniqtag(unique(template.df$Group), 3)
    seq.groups <- group.tags[match(template.df$Group, unique(template.df$Group))]
    seq.IDs <- uniqtag::uniqtag(template.df$ID, 5)
    if (init.algo == "naive") {
        primers <- create.primers.naive(seqs, seq.IDs, seq.groups, l.s, e.s, primer.lengths, 
            allowed.region.definition, max.degen, sample, mode.directionality, updateProgress)
    } else {
        primers <- create.primers.tree(seqs, seq.IDs, seq.groups, l.s, e.s, primer.lengths, 
            allowed.region.definition, max.degen, conservation, sample, mode.directionality, 
            updateProgress)
    }
    return(primers)
}
#' Primer Identifier Creation.
#' 
#' Creates identifiers for generated primers.
#'
#' @param sample Sample name of the templates.
#' @param seq.IDs Identifiers of the templates.
#' @param seq.identifiers The index of the seq.
#' @param all.starts Primer positions (start).
#' @param all.ends Primer positions (end).
#' @param identifier Direction keyword.
#' @param seq.primers The primer sequences as strings.
#' @return Identifiers for each primer.
#' @keywords internal
get.primer.identifier.string <- function(sample, seq.IDs, seq.identifier, all.starts, all.ends,
                                    identifier, seq.primers) {
    enum <- seq_along(seq.IDs)
    seq.names.short <- paste(paste0(seq.identifier, "-", enum), "|", seq.IDs, "|", 
                        all.starts, ":", all.ends, "|_",
                        identifier, sep = "")
    return(seq.names.short)
}
#' Naive Initialization of Primers.
#'
#' Initialize primers by extracting substrings from all templates.
#'
#' @param seqs The template sequence strings.
#' @param seq.IDs The identifiers of the templates.
#' @param seq.groups The group identifiers of the templates.
#' @param l.s The positions where the allowed region starts for each template.
#' @param e.s The positions where the allowed reigon ends for each template.
#' @param primer.lengths Vector of desired primer lengths.
#' @param allowed.region.definition Definition of the allowed region.
#' @param max.degen Maximum allowed degeneracy of primers.
#' @param sample Template sample identifier.
#' @param updatProgress Shiny progress object.
#' @return Data frame with initialized primer candidates.
#' @keywords internal
create.primers.naive <- function(seqs, seq.IDs, seq.groups, l.s, e.s, primer.lengths, allowed.region.definition, 
    max.degen, sample = "", identifier = "", updateProgress = NULL) {
    if (length(seqs) == 0) {
        return(NULL)
    }
    seq.idx <- NULL
    # use group identifier for primers?
    use.group.identifier <- TRUE
    if (length(unique(seq.groups)) <= 1) {
        use.group.identifier <- FALSE
    }
    all.primer.seqs <- foreach(seq.idx = seq_along(seqs), .combine = c) %dopar% {
        if (is.function(updateProgress)) {
            detail <- identifier
            updateProgress(1/length(seqs), detail, "inc")
        }
        # create possible primer start positions:
        if (allowed.region.definition == "within") {
            primer.starts <- l.s[seq.idx]:(max(1, e.s[seq.idx] - min(primer.lengths)))
        } else {
            # any overlap:
            primer.starts <- (max(1, l.s[seq.idx] - max(primer.lengths))):(e.s[seq.idx])
        }
        # create allowed primer intervals
        all.starts <- unlist(lapply(primer.starts, function(x) rep(x, length(primer.lengths))))
        all.ends <- unlist(lapply(seq_along(primer.starts), function(x) primer.starts[x] + primer.lengths - 1))
        if (allowed.region.definition == "within") {
            # remove primers that extend over the target region
            rm.idx <- which(all.ends > e.s[seq.idx])
        } else {
            # remove primers that would exceed the template length
            rm.idx <- which(all.ends > nchar(seqs[seq.idx]))
        }
        if (length(rm.idx) != 0) {
            all.starts <- all.starts[-rm.idx]
            all.ends <- all.ends[-rm.idx]
        }
        seq.primers <- substring(seqs[seq.idx], all.starts, all.ends)
        # identify primers: 
        if (use.group.identifier) {
            # use group identifiers 
            use.ids <- seq.groups[seq.idx]
        } else {
            # use the first sequence identifier
            use.ids <- seq.IDs[seq.idx]
        }
        names(seq.primers) <- get.primer.identifier.string(sample, abbreviate(use.ids, 5),  
                                    seq.idx, all.starts, all.ends, identifier, seq.primers)
        seq.primers
    }
    all.primer.seqs <- all.primer.seqs[!duplicated(all.primer.seqs)]
    degen <- score_degen(strsplit(all.primer.seqs, split = ""))
    all.primer.seqs <- all.primer.seqs[degen <= max.degen]
    return(all.primer.seqs)
}
#' @rdname PrimerDesign
#' @name PrimerDesign
#' @aliases get_initial_primers
#' @return \code{get_initial_primers} returns a data frame 
#' with candidate primers for optimization.
#' @export 
#' @examples
#' 
#' data(Ippolito)
#' # Naive primer initialization
#' init.primers <- get_initial_primers("InitialPrimers", template.df, 
#'                          c(18,18), "fw", init.algo = "naive")
#' # Tree-based primer initialization (requires MAFFT)
#' \dontrun{
#' init.primers <- get_initial_primers("InitialPrimers", template.df, 
#'                          c(18,18), "fw", init.algo = "tree")
#' }
get_initial_primers <- function(sample, template.df, primer.lengths, 
    mode.directionality = c("fw", "rev"), allowed.region.definition = c("within", "any"), 
    init.algo = c("naive", "tree"), max.degen = 16, conservation = 1.0, updateProgress = NULL) {

    if (length(init.algo) == 0) {
        stop("Please supply the 'init.algo' arg.")
    }
	init.algo <- match.arg(init.algo)
    if (length(mode.directionality) == 0) {
        stop("Please supply the 'mode.directionality' arg.")
    }
	mode.directionality <- match.arg(mode.directionality)
    if (length(allowed.region.definition) == 0) {
        stop("Please supply the 'allowed.region.definition' arg.")
    }
    allowed.region.definition <- match.arg(allowed.region.definition)
    if (length(primer.lengths) != 2) {
        stop("Primer lengths should be an interval")
    }
    if (primer.lengths[2] < primer.lengths[1]) {
        stop("Primer length doesn't specify an interval")
    }
    if (conservation < 0 || conservation > 1) {
        stop("The conservation percentile should be in the range [0,1]")
    }
    primer.lengths <- primer.lengths[1]:primer.lengths[2]
    if (any(primer.lengths <= 0)) {
        stop("All primer lengths have to be strictly positive.")
    }
    if (max.degen <= 0) {
        stop("The maximal primer degeneracy has to be strictly positive.")
    }
    if (is.function(updateProgress)) {
        detail <- "Creating primers"
        updateProgress(1/2, detail, "inc")
    }
    primer.set <- create.initial.primer.set(template.df, primer.lengths, mode.directionality, 
        sample, allowed.region.definition, init.algo, max.degen, conservation, updateProgress)
    if (is.function(updateProgress)) {
        detail <- "Structuring"
        updateProgress(1/2, detail, "inc")
    }
    primer.df <- read_primers.internal(primer.set, names(primer.set), "fw", "rev", 
        merge.ambig = "none", max.degen, sample)
    return(primer.df)
}
#' Creation of Initial Primers
#' 
#' Creates a set of candidate primers.
#'
#' @param template.df Template data frame.
#' @param sample.name Name of the template sample.
#' @param primer.lengths Interval of minimal and maximal desired primer length.
#' @param mode.directionality Direction of primers to be created.
#' @param init.algo Algorithm for initializing primers.
#' @param allowed.region.definition Definition of the allowed binding region.
#' @param max.degen Maximal allowed degeneration of created primers.
#' @param conservation Required conservation of primers.
#' The value of \code{conservation} should be in the range[0,1].
#' @param cur.results.loc Location for writing the primers as csv.
#' @return An initial primer data frame.
#' @keywords internal
initialize.primer.set <- function(template.df, sample.name, primer.lengths, allowed.region.definition, 
    mode.directionality, init.algo, max.degen, conservation, cur.results.loc) {
    out.name <- get.init.file.name(cur.results.loc, sample.name, primer.lengths, 
        mode.directionality, allowed.region.definition, init.algo, max.degen, conservation)


    primer.df <- get_initial_primers(sample.name, template.df, primer.lengths, mode.directionality, 
        allowed.region.definition, init.algo, max.degen, conservation)
    if (length(out.name) != 0) {
        dir.create(dirname(out.name), showWarnings = FALSE, recursive = TRUE)
        write.csv(primer.df, file = out.name, row.names = FALSE)
    }
    return(primer.df)
}

#########
# Primer coverage functions
#########

#' Unique Coverage Indices
#"
#' Computes the indices of templates that are covered uniquely 
#' covered by an individual primer.
#'
#' @param primer.df Primer data frame.
#' @param template.df Template data frame.
#'
#' @return Index of templates uniquely covered per input primer.
#' @keywords internal
compute.unique.covered.idx <- function(primer.df, template.df) {
    # computes the unique coverage of a primer (the number of templates that are not
    # covered by the other primers in the set)
    cvd.idx <- covered.seqs.to.idx(primer.df$Covered_Seqs, template.df)
    unique.idx <- lapply(seq_along(cvd.idx), function(x) setdiff(cvd.idx[[x]], intersect(cvd.idx[[x]], 
        unlist(cvd.idx[-x]))))
    return(unique.idx)
}

#' Conversion of Coverage Strings to Indices.
#'
#' Converts the input coverage strings (comma separated template Identifiers) into indices.
#'
#' @param covered.seqs Strings of covered sequences to be converted.
#' @param template.df Template data frame containing the identifiers of templates
#'
#' @return Indices of covered templates.
#' @keywords internal
covered.seqs.to.idx <- function(covered.seqs, template.df) {
    #print("covered.seqs.to.idx")
    #print(covered.seqs)
    idx <- lapply(strsplit(as.character(covered.seqs), split = ","), 
        function(x) {
            if (all(is.na(x))) {
                NULL
            } else {
                match(as.numeric(x), template.df$Identifier)
            }
    })
    return(idx)
}
#' Conversion of Template Coverage Indices to ID string
#'
#' Converts the input coverage indices to a comma-separated string with the template IDs.
#'
#' @param covered.seqs Indices of covered template sequences.
#' @param template.df Template data frame.
#' @return A string containing the covered template IDs.
#' @keywords internal
covered.seqs.to.ID.string <- function(covered.seqs, template.df) {
    if (length(covered.seqs) == 0) {
        return("")
    }
    ID <- lapply(strsplit(as.character(covered.seqs), split = ","), function(x) paste(template.df[match(as.numeric(x), 
        template.df$Identifier), "ID"], collapse = ","))
    return(ID)
}
#' Conversion of Primer Indices to ID string
#'
#' Converts the input coverage indices to a comma-separated string with the template IDs.
#'
#' @param covered.primers Identifiers of primers covering sequences.
#' @param primer.df Primer data frame.
#' @return String containing the covered template IDs.
#'
#' @return A string containing the IDs of covering primers.
#' @keywords internal
covered.primers.to.ID.string <- function(covered.primers, primer.df) {
    if (length(covered.primers) == 0) {
        return("")
    }
    ID <- lapply(strsplit(as.character(covered.primers), split = ","), function(x) paste(primer.df[match(x, 
        primer.df$Identifier), "ID"], collapse = ","))
    return(ID)
}
#' Covered Templates
#'
#' Get the indices of covered templates.
#'
#' @param Tm.set Primer data frame.
#' @param template.df Template data set.
#'
#' @return Index of templates that are covered by the primers in \code{Tm.set}.
#' @keywords internal
get.covered.templates <- function(Tm.set, template.df) {
    # for each template, return the idx of covering primers Tm.set: primer data frame
    # template.df: template data frame
    cvd.IDs <- strsplit(Tm.set$Covered_Seqs, split = ",")
    cvd.idx <- covered.seqs.to.idx(Tm.set$Covered_Seqs, template.df)
    template.coverage <- lapply(seq_along(template.df$Identifier), function(j) which(sapply(cvd.IDs, 
        function(x) template.df$Identifier[j] %in% x)))
    return(template.coverage)
}
#' 3' Mismatch Check.
#'
#' Check for mismatches at primer 3' ends.
#'
#' @param template.df Template data frame.
#' @param primer.df Primer data frame.
#' @param mode.directionality Primer directionality.
#' @return Returns the distance of mismatches from the 3' terminal end of primers. 
#' @keywords internal
check.3prime.mismatches <- function(template.df, primer.df, 
                           mode.directionality = c("fw", "rev", "both")) {
     if (length(mode.directionality) == 0) {
        stop("Please supply the 'mode.directionality' argument.")
    }
    mode.directionality <- match.arg(mode.directionality)
    fw.idx <- primer.df$Forward != ""
    rev.idx <- primer.df$Reverse != ""
    if (mode.directionality == "fw") {
        mismatches <- get.3prime.mismatch.pos(primer.df$Forward, primer.df$Mismatch_pos_fw)
    } else if (mode.directionality == "rev") {
        mismatches <- get.3prime.mismatch.pos(primer.df$Reverse, primer.df$Mismatch_pos_rev)
    } else {
        fw.mismatches <- get.3prime.mismatch.pos(primer.df$Forward, primer.df$Mismatch_pos_fw)
        rev.mismatches <- get.3prime.mismatch.pos(primer.df$Reverse, primer.df$Mismatch_pos_rev)
        mismatches <- vector("list", length(fw.mismatches))
        for (x in seq_along(fw.mismatches)) {
            if (fw.idx[x] && rev.idx[x]) {
                res <- sapply(seq_along(fw.mismatches[[x]]), function(y) min(c(fw.mismatches[[x]][[y]], 
                  rev.mismatches[[x]][[y]])))
            } else if (fw.idx[x]) {
                res <- fw.mismatches[[x]]
            } else {
                res <- rev.mismatches[[x]]
            }
            mismatches[[x]] <- res
        }
    }
    return(mismatches)
}
#' 3' Hexamer Check.
#'
#' Check whether the 3' hexamer of a primer is fully complementary
#' to the corresponding region in the template.
#'
#' @param template.df Template data frame.
#' @param primer.df Primer data frame.
#' @param mode.directionality Primer directionality.
#' @return Returns \code{TRUE} if the 3' hexamer of a primer is fully complementary
#' to the corresponding template region and \code{FALSE} otherwise.
#' @keywords internal
check.3prime.hexamers <- function(template.df, primer.df, 
                           mode.directionality = c("fw", "rev", "both")) {
    term.pos <- check.3prime.mismatches(template.df, primer.df, 
                           mode.directionality = mode.directionality)
    hexamer.ok <- lapply(term.pos, function(x) x > 6) # TRUE, if hexamer is free of mismatches
    return(hexamer.ok)
}

#' Conversion of Mismatch Postions String to List.
#'
#' @param mismatches A character vector where parenthesis give mismatches for a template binding event.
#' @return A list with the mismatches for every template for every primer.
#' @keywords internal
mismatch.string.to.list <- function(mismatches) {
    pattern <- "(?<=\\()(([0-9]+,*)*(?=\\)))"
    m <- regmatches(mismatches, gregexpr(pattern, mismatches, perl = TRUE))
    mm <- lapply(m, function(x) lapply(strsplit(x, split = ","), function(y) {
        if (length(y) == 0) 
            NA else as.numeric(y)
    }))
    return(mm)
}
#' Identification of 3' Mismatches.
#'
#' Computes the lastmost position of a 3' mismatches of a primer with a template.
#'
#' @param primers Primer sequence strings.
#' @param mismatches Comma-separated strings containing the primer mismatch positions.
#'
#' @return The closest position of a mismatch relative to the 3' end of the primer.
#' Here, 1 indicates the terminal position, 2 the penultimate position, and so on.
#' No mismatch is indicated by an infinite value.
#' @keywords internal
get.3prime.mismatch.pos <- function(primers, mismatches) {
    # mismatches: mismatches string vector from primer.df primers: primer character
    # vector
    mm <- mismatch.string.to.list(mismatches)
    # no mismatch -> position is Inf (needed for filtering)
    res <- lapply(seq_along(mm), function(x) 
        if (length(mm[[x]]) != 0) {
            # select primers with mismatches:
            idx <- which(sapply(mm[[x]], function(y) !is.na(y[1])))
            r <- rep(Inf, length(mm[[x]]))
            r[idx] <- sapply(mm[[x]][idx], function(y) nchar(primers[x]) - max(y) +1 ) # worst-case position (furthest mismatch 3' prime)
            r
        } 
    ) 
    return(res)
}
#' Update Coverage Information.
#'
#' Updates the coverage-related columns in the input primer data frame.
#' Does not modify the entries of template-specific coverage columns
#' such as primer efficiency (comma-separated values).
#'
#' Removes all coverage events of templates whose index is not in \code{sel}.
#'
#' @param filtered.df Primer data frame.
#' @param sel List with indices of covered templates to be retained, one list with template indices to keep per primer.
#' @param template.df Template data frame.
#' @param mode Either \code{on_target} to filter on-target binding events
#' or \code{off_target} to filter off-target binding events. The corresponding
#' \code{sel} argument should be different.
#' @param active.constraints The active coverage constraints.
#' @return A primer data frame with updated coverage information.
#' @keywords internal
update.cvg.data <- function(filtered.df, sel, template.df, mode = c("on_target", "off_target"), active.constraints) {
    # updates all cvg-related columns according to the 'sel' list (covered templates
    # to be removed, e.g. due to primer efficiency filter) sel: list with indices for
    # each covered template of a primer that should be retained. all others are
    # removed (not considered covered anymore!)
    if (length(filtered.df$Forward) == 0 || length(filtered.df$Reverse) == 0) {
        stop("Need some primers to update cvg data.")
    }
    if (length(mode) == 0) {
        stop("Please supply the 'mode' argument.")
    }
    mode <- match.arg(mode)
    updated.df <- filtered.df
    # all cols relating to cvg info were strsplit is necessary need to be updated
    special.cols <- c("Mismatch_pos_fw", "Mismatch_pos_rev")  # () delimiters
    # only update the currently active constraints in order to prevent error when data are outdated for some reason.
    template.constraints <- active.constraints
    #print(template.constraints)
    # off-binding conditions:
    other.cols <- c("Covered_Seqs", "Binding_Position_Start_fw", "Binding_Position_End_fw", 
            "Binding_Position_Start_rev", "Binding_Position_End_rev", "Binding_Region_Allowed", 
            "Nbr_of_mismatches_fw", "Nbr_of_mismatches_rev", "Relative_Forward_Binding_Position_Start_fw", 
            "Relative_Forward_Binding_Position_End_fw", "Relative_Forward_Binding_Position_Start_rev", 
            "Relative_Forward_Binding_Position_End_rev", "Relative_Reverse_Binding_Position_Start_fw", 
            "Relative_Reverse_Binding_Position_End_fw", "Relative_Reverse_Binding_Position_Start_rev", 
            "Relative_Reverse_Binding_Position_End_rev", "Binding_Region_Allowed_fw", 
            "Binding_Region_Allowed_rev")
    if (mode == "off_target") {
        special.cols <- paste0("Off_", special.cols)
        template.constraints <- paste0("Off_", template.constraints)
        other.cols <- paste0("Off_", other.cols)
    }
    cvg.cols <- c(special.cols, template.constraints, other.cols)
    fw.idx <- which(filtered.df$Forward != "")
    rev.idx <- which(filtered.df$Reverse != "")
    for (i in seq_along(cvg.cols)) {
        if (!cvg.cols[i] %in% colnames(filtered.df)) {
            next
        }
        #message(paste('Updating: ', cvg.cols[i]))
        #print(filtered.df[, cvg.cols[i]])
        fw.col <- grepl("fw", cvg.cols[[i]])
        rev.col <- grepl("rev", cvg.cols[[i]])
        special.col <- cvg.cols[[i]] %in% special.cols  # no differentiation necessary (full mismatch bracket is removed)
        if (special.col) {
            s <- stringr::str_extract_all(filtered.df[, cvg.cols[i]], "(\\([^()]*\\))")
        } else {
            s <- strsplit(filtered.df[, cvg.cols[i]], split = ",")
        }
        s.new <- sapply(seq_along(s), function(x) paste(s[[x]][sel[[x]]], collapse = ","))
        if (fw.col) {
            updated.df[fw.idx, cvg.cols[i]] <- s.new[fw.idx]
        } else if (rev.col) {
            updated.df[rev.idx, cvg.cols[i]] <- s.new[rev.idx]
        } else {
            # col is not direction-specific
            updated.df[, cvg.cols[i]] <- s.new
        }
    }
    # recompute summary statistics: Coverage_Ratio, primer_coverage
    if (mode == "on_target") {
        if ("primer_coverage" %in% colnames(updated.df)) {
            updated.df$primer_coverage <- sapply(sel, length)
            updated.df$Coverage_Ratio <- updated.df$primer_coverage/nrow(template.df)
            if ("annealing_DeltaG" %in% colnames(updated.df)) {
                mean.DeltaG <- unlist(lapply(strsplit(updated.df$annealing_DeltaG, split = ","), function(x) mean(as.numeric(x))))
                mean.DeltaG[is.nan(mean.DeltaG)] <- 0
                updated.df$mean_annealing_DeltaG <- mean.DeltaG
            }
        }
        # update mean efficiency
        if ("primer_efficiency" %in% colnames(updated.df) && "primer_efficiency" %in% active.constraints) {
            updated.df$mean_primer_efficiency <- unlist(lapply(strsplit(updated.df$primer_efficiency, split = ","), function(x) mean(as.numeric(x)))) 
            updated.df$mean_primer_efficiency[is.na(updated.df$mean_primer_efficiency)] <- 0
        }
    }
    # update specificity only for off-target update -> true off-target events are known now!
    if (mode == "off_target") {
        # n.b.: specificity is not perfectly accurate since we store only the best binding event in the target region -> the actual specificity could be higher than the computed one.
        off.cvg <- sapply(strsplit(updated.df$Off_Covered_Seqs, split = ","), function(x) length(unique(x))) # unique off-covered seqs to be on the same scale as the on-target coverage, which is also unique
        TP <- sapply(strsplit(updated.df$Binding_Region_Allowed, split = ","), function(x) length(which(as.logical(x)))) # on-target cvg count
        #print("TP")
        #print(TP) 
        #print("off")
        #print(off.cvg)
        updated.df$primer_specificity <- TP / (TP + off.cvg)
        # if there's not binding events -> 0 specificity
        updated.df$primer_specificity[is.na(updated.df$primer_specificity)] <- 0
    }
    return(updated.df)
}
#' Identification of Mismatch Mutations.
#'
#' Identifies primers that induce mutations due to mismatch binding.
#'
#' Checks for one primer and all covered templates whether any templates are bound
#' with mismatches such that a forbidden mutation is induced. A boolean vector indicating 
#' which binding events induce a forbidden mutation is returned.
#'
#' @param primer.seq Primer sequence string.
#' @param pos.start Binding position of primer (start).
#' @param pos.end Binding position of primer (end).
#' @param template.df Template data frame.
#' @param covered.seqs Identifiers of covered templates.
#' @param ORF.data Reading frame information of templates.
#' @param mode.directionality Directionality of primers.
#' @param mutation.types Character vector of the mutation types to be checked for.
#' @return TRUE if the \code{primer.seq} induces a mutation that is
#' forbidden according to the provided \code{mutation.types}.
#' @keywords internal
check.mutations <- function(primer.seq, pos.start, pos.end, template.df, covered.seqs, 
                            ORF.data, mode.directionality = c("fw", "rev"),
                            mutation.types = c("stop_codon", "substitution")) {

    if (length(mode.directionality) == 0) {
        stop("Please supply the 'mode.directionality' argument.")
    }
    if (length(mutation.types) == 0) {
        return(NULL)
    }
    mutation.types <- match.arg(mutation.types, several.ok = TRUE)
    mode.directionality <- match.arg(mode.directionality)
    if (primer.seq == "") {
        return(NULL)
    }
    template.idx <- match(covered.seqs, template.df$Identifier)
    if (length(template.idx) == 0) {
        # nothing to check
        return(NULL)  # no stop codons found
    }
    ORFs <- ORF.data$Frame
    cur.ORF <- ORFs[template.idx]
    my.lex <- template.df[template.idx, ]
    seqs.exon.nt <- my.lex$Sequence
    primer.seqs.nt <- seqs.exon.nt
    # incorporate the changes from the primer seqs into the sequences
    if (mode.directionality == "fw") {
        substr(primer.seqs.nt, pos.start, pos.end) <- primer.seq
    } else {
        substr(primer.seqs.nt, pos.start, pos.end) <- rev.comp.sequence(primer.seq)
    }
    S <- strsplit(seqs.exon.nt, split = "")
    P <- strsplit(primer.seqs.nt, split = "")
    result <- data.frame(matrix(rep(FALSE, length(template.idx) * length(mutation.types)), 
                        nrow = length(template.idx), ncol = length(mutation.types)))
    colnames(result) <- mutation.types
    for (k in seq_along(S)) {
        frame <- cur.ORF[[k]]
        # compare translated amplicon (primer mismatches) and nt seqs
        seq.aa <- paste(seqinr::translate(S[[k]], frame = frame, sens = "F", ambiguous = TRUE), 
                        collapse = "")
        primer.aa <- paste(seqinr::translate(P[[k]], frame = frame, sens = "F", ambiguous = TRUE), 
                                    collapse = "")
        highlight.aa <- highlight.mismatch(tolower(seq.aa), tolower(primer.aa))
        types <- highlight.aa$type
        #print(seq.aa)
        #print(primer.aa)
        #print(unlist(types))
        idx <- which(sapply(types, function(x) mutation.types %in% x)) 
        result[k, idx] <- TRUE # disallowed mutation type found
    }
    return(result)
}

#' Evaluation of Coverage.
#'
#' Re-evaluates the coverage of primers under exclusion of certain templates.
#'
#' This function requires that \code{primers} was already annotated with primer coverage before.
#'
#' @param primers Primer data frame.
#' @param excluded.seqs Identifiers of templates to be excluded.
#' @param template.df Template data frame
#' @return Primer data frame with updated coverage under the exclusion of \code{excluded.seeqs}.
#' @keywords internal
evaluate.diff.primer.cvg <- function(primers, excluded.seqs, template.df) {
    if (!"primer_coverage" %in% colnames(primers)) {
        stop("Differential primer cvg can only be used when cvg has already been computed!")
    }
    if (length(excluded.seqs) == 0) {
        return(primers)  # nothing to adjust
    }
    # need to update the string with covered seqs in the primer data frame
    s <- strsplit(primers$Covered_Seqs, split = ",")
    i <- NULL
    updated.seqs <- foreach(i = seq_along(s), .combine = rbind) %dopar% {
        seqs <- as.numeric(s[[i]])
        m <- match(seqs, excluded.seqs)
        rm.idx <- which(!is.na(m))
        if (length(rm.idx) != 0) {
            seqs <- seqs[-rm.idx]
        }
        # turn vectors into strings again
        cvg.count <- length(seqs)
        seqs <- paste(seqs, collapse = ",", sep = "")
        r <- data.frame(Covered_Seqs = seqs, primer_coverage = cvg.count, Coverage_Ratio = cvg.count/nrow(template.df), 
            stringsAsFactors = FALSE)
    }
    # update the input data frame
    primers[, colnames(updated.seqs)] <- updated.seqs
    # order again by cvg count
    primers <- primers[order(primers$primer_coverage, decreasing = TRUE), ]
    return(primers)
}

#' Selection of Binding Events
#'
#' Selects primer binding events that are within the allowed binding regions.
#'
#' @param bound.fw Indices of covered templates of a single primer.
#' @param bound.to.allowed.region.fw Corresponding allowed binding regions.
#' @param allowed.other.binding.ratio The ratio of other binding events. If
#' this is different from 0, disallowed binding events will also be reported.
#'
#' @return The indices of the allowed binding events.
#' @keywords internal
select.allowed.binding.events <- function(bound.fw, bound.to.allowed.region.fw, allowed.other.binding.ratio) {
    unique.bound.fw <- unique(bound.fw)
    dup.mapping.fw <- lapply(seq_along(unique.bound.fw), function(x) which(bound.fw == 
        unique.bound.fw[x]))
    sel.idx.fw <- vector("list", length(dup.mapping.fw))
    use.disallowed <- allowed.other.binding.ratio != 0
    for (j in seq_along(dup.mapping.fw)) {
        idx <- dup.mapping.fw[[j]]
        if (!use.disallowed) {
            b.fw <- which(bound.to.allowed.region.fw[idx])
        } else {
            b.fw <- seq_along(bound.to.allowed.region.fw[idx])
        }
        if (length(b.fw) != 0) {
            sel.idx.fw[[j]] <- idx[b.fw]
        } 
    }
    return(sel.idx.fw)
}
#' Selection of Individual Binding Events
#'
#' Selects only binding events of interest.
#'
#' @param fw.binding.filtered IRanges with binding events.
#' @param p.idx Index of binding events to keep.
#' @return An IRanges object containing only the selected binding events.
#' @keywords internal
select.binding.events <- function(fw.binding.filtered, p.idx) {
    # select only the target binding events
    my.binding.fw <- fw.binding.filtered[p.idx]
    # adjust metadata
    metadata(my.binding.fw)$primer_idx <- metadata(my.binding.fw)$primer_idx[p.idx]
    metadata(my.binding.fw)$template_idx <- metadata(my.binding.fw)$template_idx[p.idx]
    metadata(my.binding.fw)$allowed_region <- metadata(my.binding.fw)$allowed_region[p.idx]
    return(my.binding.fw)
}
#' Combination of Binding Events.
#'
#' Appends all binding events.
#'
#' @param my.binding.fw Forward binding events of individual primers.
#' @param my.binding.rev Reverse binding events of individual primers.
#' @param fw.m Forward binding events of paired primers.
#' @param rev.m Reverse binding events of paired primers.
#' @return IRanges of all binding events.
#' @keywords internal
combine.binding.events <- function(my.binding.fw, my.binding.rev, fw.m, rev.m) {
    binding <- c(my.binding.fw, my.binding.rev, fw.m, rev.m)
    # integrate metadata for both directions
    metadata(binding)$primer_idx <- c(metadata(my.binding.fw)$primer_idx, metadata(my.binding.rev)$primer_idx, 
        metadata(fw.m)$primer_idx, metadata(rev.m)$primer_idx)
    metadata(binding)$template_idx <- c(metadata(my.binding.fw)$template_idx, 
        metadata(my.binding.rev)$template_idx, metadata(fw.m)$template_idx, metadata(rev.m)$template_idx)
    metadata(binding)$allowed_region <- c(metadata(my.binding.fw)$allowed_region, 
        metadata(my.binding.rev)$allowed_region, metadata(fw.m)$allowed_region, 
        metadata(rev.m)$allowed_region)
    return(binding)
}
#' Merge of Forward/Reverse Binding Information.
#'
#' Determines binding events of individual and pairs of primers.
#'
#' @param primers The primer data frame.
#' @param fw.binding IRanges object with binding events of fw primers.
#' @param rev.binding IRanges object with binding events of rev primers.
#' @param mode.directionality Primer directionality.
#' @param idx.fw Index of fw primers.
#' @param idx.rev Index of rev primers.
#' @return IRanges with correct binding events.
#' @keywords internal
merge.binding.information <- function(primers, fw.binding.filtered, 
    rev.binding.filtered, mode.directionality = c("fw", "rev", "both"), 
    idx.fw, idx.rev) {
    
    if (length(mode.directionality) == 0) {
        stop("Please supply the 'mode.directionality' argument.")
    }
    mode.directionality <- match.arg(mode.directionality)
    if (mode.directionality == "fw") {
        # only forward primers present
        binding <- fw.binding.filtered
        metadata(binding) <- c(metadata(binding), direction = list(rep("fw", length(binding))))
    } else if (mode.directionality == "rev") {
        # only reverse primers present
        binding <- rev.binding.filtered
        metadata(binding) <- c(metadata(binding), direction = list(rep("rev", length(binding))))
    } else {
        # primers of both directions present
        # differentiate between individual primers and primer pairs:
        merge.idx <- intersect(idx.fw, idx.rev) # primer pairs
        keep.idx <- setdiff(seq_len(nrow(primers)), merge.idx) # individual primers
        # get individual binding events:
        my.binding.fw <- select.binding.events(fw.binding.filtered, 
            metadata(fw.binding.filtered)$primer_idx %in% keep.idx)
        my.binding.rev <- select.binding.events(rev.binding.filtered, 
            metadata(rev.binding.filtered)$primer_idx %in% keep.idx)
        # get pair binding events:
        fw.m <- select.binding.events(fw.binding.filtered, 
            metadata(fw.binding.filtered)$primer_idx %in% merge.idx) # fw bindings of pairs
        rev.m <- select.binding.events(rev.binding.filtered, 
            metadata(rev.binding.filtered)$primer_idx %in% merge.idx) # rev bindings of pairs
        # get idx of fw/rev covered templates
        t.fw <- lapply(seq_len(nrow(primers)), function(x) 
                    metadata(fw.m)$template_idx[metadata(fw.m)$primer_idx == x])
        t.rev <- lapply(seq_len(nrow(primers)), function(x) 
                    metadata(rev.m)$template_idx[metadata(rev.m)$primer_idx == x])
        # determine templates covered by both primers of a pair:
        t <- lapply(seq_len(nrow(primers)), function(x) intersect(t.fw[[x]], t.rev[[x]]))
        ####### cave: multiple binding events are retained here ..
        # select fw binding events of pairs:
        rm <- unlist(lapply(seq_along(t), function(x) which(metadata(fw.m)$template_idx %in% 
            t[[x]] & metadata(fw.m)$primer_idx == x)))
        fw.m <- select.binding.events(fw.m, rm) # fw bindings of pairs
        # select rev binding events of pairs:
        rm <- unlist(lapply(seq_along(t), function(x) which(metadata(rev.m)$template_idx %in% 
            t[[x]] & metadata(rev.m)$primer_idx == x)))
        rev.m <- select.binding.events(rev.m, rm) # rev bindings of pairs
        # integrate pair and individual binding data:
        binding <- combine.binding.events(my.binding.fw, my.binding.rev, fw.m, rev.m)
        # add primer directions to metadata:
        direction <- c(rep("fw", length(my.binding.fw)), rep("rev", length(my.binding.rev)), 
            rep("fw", length(fw.m)), rep("rev", length(rev.m)))
        metadata(binding) <- c(metadata(binding), direction = list(direction))
    }
    return(binding)
}
#' Retrieval of Allowed Binding Indices.
#'
#' Retrieves the indices of allowed binding events in \code{binding}
#' for the primer with index \code{x} and type \code{primer.type}.
#'
#' @param binding IRanges binding information.
#' @param primer.type Direction of primer.
#' @param x Index of primer in the primer data frame.
#' @param allowed.other.binding.ratio The ratio of allowed off-target binding events.
#' @return Indices in binding for primer with index code{x} that are allowed.
#' @keywords internal
get.primer.binding.idx <- function(binding, primer.type = c("fw", "rev", "both"), 
                                x, allowed.other.binding.ratio) {
    # determine individual covered sequences
    if (length(primer.type) == 0) {
        stop("Please supply the 'primer.type' argument.")
    }
    primer.type <- match.arg(primer.type)
	fw.idx <- which(metadata(binding)$direction == "fw" & metadata(binding)$primer_idx == x)
	rev.idx <- which(metadata(binding)$direction == "rev" & metadata(binding)$primer_idx == x)
	# in case of duplicate binding events of one primer into the same template, select allowed ones
	fw.idx <- fw.idx[unlist(select.allowed.binding.events(metadata(binding)$template_idx[fw.idx],
												metadata(binding)$allowed_region[fw.idx], 
                                                allowed.other.binding.ratio))]
	rev.idx <- rev.idx[unlist(select.allowed.binding.events(metadata(binding)$template_idx[rev.idx], 
												metadata(binding)$allowed_region[rev.idx], 
                                                allowed.other.binding.ratio))]
	if (primer.type == "both") {
        # require that both primers of pair bind to allowed regions:
		sel.idx.fw <- fw.idx[which(metadata(binding)$template_idx[fw.idx] %in% metadata(binding)$template_idx[rev.idx])]
		sel.idx.rev <- rev.idx[which(metadata(binding)$template_idx[rev.idx] %in% metadata(binding)$template_idx[fw.idx])]
	} else if (primer.type == "fw") {
		sel.idx.fw <- fw.idx
		sel.idx.rev <- NULL
	} else {
		sel.idx.fw <- NULL
		sel.idx.rev <- rev.idx
	}
    result <- list("fw" = sel.idx.fw, "rev" = sel.idx.rev)
	return(result)
}
#' Retrieval of Relative Binding Positions.
#'
#' Retrieves primer binding position relative to allowed regions of either
#' forward or reverse primers, as specified by \code{direction}.
#'
#' @param allowed Positions where binding is allowed in the templates.
#' @param primer.pos Binding position of primer (absolute).
#' @param direction Direction (either fw/rev).
#' @param covered.seqs.idx Indices of covered templates.
#' @return Numeric of relative binding position to allowed region.
#' @keywords internal
get.relative.binding.pos <- function(allowed, primer.pos, direction, covered.seqs.idx) {
    if (direction == "fw") {
        # relative to fw primer binding region
        rel.pos <- sapply(seq_along(primer.pos), function(x) primer.pos[[x]] - 
            allowed[covered.seqs.idx[x]] - 1)
    } else {
        # relative to rev primer binding region
        rel.pos <- sapply(seq_along(primer.pos), function(x) -(primer.pos[[x]] - 
            allowed[covered.seqs.idx[x]] + 1))
    }
    return(rel.pos)
}
#' Selection of Best (smallest number of mismatches) Binding Event per Template Coverage Event.
#'
#' @param binding Binding information.
#' @param fw.mm.info Info about mismatches.
#' @return A list with entries 'fw' and 'rev' giving the best indices of primers.
#' @keywords internal
select_best_binding <- function(binding, fw.mm.info) {
    result <- vector("list", 2)
    unique.t <- unique(metadata(binding)$template_idx)
    sel.idx.fw.rel <- unlist(lapply(seq_along(unique.t), function(z) {
                    my.t.idx <- which(metadata(binding)$template_idx == unique.t[z])
                    nbr.mm <- fw.mm.info$Nbr[my.t.idx]
                    my.t.idx[which.min(nbr.mm)]}))
    return(sel.idx.fw.rel)
}
#' Evaluation of Primer Coverage.
#'
#' Evaluates the coverage of a set of primers.
#'
#' @param template.df Template data frame.
#' @param primers Primer data frame.
#' @param mode.directionality Primer directionality.
#' @param allowed.mismatches The number of allowed mismatches per binding event.
#' @param allowed.other.binding.ratio Ratio of primers that are allowed to bind to non-allowed regions.
#' If \code{allowed.other.binding.ratio} >0 primers are allowed to bind at any location within the templates.
#' However, a warning is given if the ratio of primers binding to non-target regions exceeds the \code{allowed.other.binding.ratio}.
#' @param allowed.region.definition Definition of the target binding sites used for evaluating the coverage.
#' If \code{allowed.region.definition} is \code{within}, primers have to lie within the allowed binding region.
#' If \code{allowed.region.definition} is \code{any}, primers have to overlap with the allowed binding region.
#' The default is that primers have to bind within the target binding region.
#' @param updateProgress Progress callback function for shiny.
#' @return Primer data frame with information on the covered template sequences.
#' @keywords internal
evaluate.basic.cvg <- function(template.df, primers, mode.directionality = c("fw", "rev", "both"), 
                               allowed.mismatches, allowed.other.binding.ratio, 
                               allowed.region.definition = c("within", "any"), updateProgress = NULL) {

    if (length(mode.directionality) == 0) {
        stop("Please supply the 'mode.directionality' argument.")
    }
    if (length(allowed.region.definition) == 0) {
        stop("Please supply the 'allowed.region.definition' argument.")
    }
    mode.directionality <- match.arg(mode.directionality)
    allowed.region.definition <- match.arg(allowed.region.definition)
    if (length(template.df) == 0 || nrow(template.df) == 0 || length(primers) == 0  || nrow(primers) == 0) {
        return(NULL)
    }
    #message("computing cvg for: ", nrow(primers), " primers.")
    # disregard empty primers
    idx.fw <- which(primers$Forward != "")
    idx.rev <- which(primers$Reverse != "")
    seqs <- DNAStringSet(template.df$Sequence)
    seqs.rc <- DNAStringSet(reverseComplement(seqs))
    # should we consider only as 'on-target', the events in the target region or just consider all events?
    doFilter <- ifelse(allowed.other.binding.ratio == 0, TRUE, FALSE)
    ### fw primers
    fw.binding <- evaluate.cvg(seqs, DNAStringSet(primers$Forward), "fw", allowed.mismatches, 
        updateProgress) # binding events of forward primers
    # determine the allowed binding region:
    allowed.fw <- IRanges(start = template.df$Allowed_Start_fw, end = template.df$Allowed_End_fw)
    # determine allowed binding events for forward primers
    fw.binding.data <- annotate.binding.events(fw.binding, allowed.fw, nrow(primers), allowed.region.definition)
    ### rev primers: analogously
    rev.binding <- evaluate.cvg(seqs.rc, DNAStringSet(primers$Reverse), "rev", allowed.mismatches, 
       updateProgress)
    allowed.rev <- IRanges(start = template.df$Allowed_Start_rev, end = template.df$Allowed_End_rev)
    rev.binding.data <- annotate.binding.events(rev.binding, allowed.rev, nrow(primers), allowed.region.definition)
    if (doFilter) {
        # only retain allowed binding events
        fw.binding.filtered <- fw.binding.data$on_target
        rev.binding.filtered <- rev.binding.data$on_target
    } else {
        # consider all binding events
        fw.binding.filtered <- fw.binding.data$all_binding
        rev.binding.filtered <- rev.binding.data$all_binding
    }
    # require paired primers to bind with both directions, merge results:
    binding <- merge.binding.information(primers, fw.binding.filtered, rev.binding.filtered, mode.directionality, idx.fw, idx.rev)
    # off-target binding:
    off.binding <- merge.binding.information(primers, fw.binding.data$off_target, rev.binding.data$off_target, mode.directionality, idx.fw, idx.rev)
    on.df <- compute.basic.details(binding, "on_target", template.df, primers, mode.directionality,
                               allowed.mismatches, allowed.other.binding.ratio, 
                               allowed.region.definition, updateProgress)
    off.df <- compute.basic.details(off.binding, "off_target", template.df, primers, mode.directionality,
                               allowed.mismatches, allowed.other.binding.ratio, 
                               allowed.region.definition, updateProgress)
    colnames(off.df) <- paste0("Off_", colnames(off.df))
    cvg.df <- cbind(on.df, off.df)
    ################
    # E: Specificity
    ################
    off.cvg <- unlist(lapply(covered.seqs.to.idx(cvg.df$Off_Covered_Seqs, template.df), function(x) length(unique(x))))
    TP <- unlist(lapply(strsplit(cvg.df$Binding_Region_Allowed, split = ","), function(x) length(which(as.logical(x)))))
    specificity <- TP / (TP + off.cvg)
    specificity[is.na(specificity)] <- 0
    cvg.df$primer_specificity <- specificity
    return(cvg.df)
}
get.mismatch.info <- function(primers, template.df, binding, covered.seqs.idx, primer.type = c("fw", "rev")) {
    primer.type <- match.arg(primer.type)
    subject <- DNAStringSet(template.df$Sequence[covered.seqs.idx])
    w <- Biostrings::width(subject)  # template lengths (Biostrings to diff with 'width' from IRanges)
    # cumulative index in the concatenation of template strings
    idx <- c(0, cumsum(head(w, length(w) - 1)))
    s <- unlist(subject)
    mm.info <- NULL
    if (primer.type == "fw") {
        f <- Views(s, binding@start + idx, (binding@start + binding@width - 1) + idx)  
        mm.info <- mismatch.info(primers$Forward, f) 
    } else if (primer.type == "rev") {
        # rev
        f <- Views(s, binding@start + idx, 
                    (binding@start + binding@width - 1) + idx)
        mm.info <- mismatch.info(primers$Reverse, reverseComplement(f))
    }
    return(mm.info)
}

#' Computation of Coverage Details
#'
#' Determines binding properties of primers.
#'
#' @param binding An \code{IRanges} object with primer binding information.
#' @param mode Either \code{on_target} for on-target binding or \code{off_target} for off-target binding.
#' @param template.df Template data frame.
#' @param primers Primer data frame.
#' @param mode.directionality Primer directionality.
#' @param allowed.mismatches The number of allowed mismatches per binding event.
#' @param allowed.other.binding.ratio Ratio of primers that are allowed to bind to non-allowed regions.
#' If \code{allowed.other.binding.ratio} >0 primers are allowed to bind at any location within the templates.
#' However, a warning is given if the ratio of primers binding to non-target regions exceeds the \code{allowed.other.binding.ratio}.
#' @param allowed.region.definition Definition of the target binding sites used for evaluating the coverage.
#' If \code{allowed.region.definition} is \code{within}, primers have to lie within the allowed binding region.
#' If \code{allowed.region.definition} is \code{any}, primers have to overlap with the allowed binding region.
#' The default is that primers have to bind within the target binding region.
#' @param updateProgress Progress callback function for shiny.
#' @return Primer data frame with information on the covered template sequences.
#' @keywords internal
compute.basic.details <- function(binding, mode = c("on_target", "off_target"), template.df, 
                               primers, mode.directionality = c("fw", "rev", "both"), 
                               allowed.mismatches, allowed.other.binding.ratio, 
                               allowed.region.definition = c("within", "any"), updateProgress = NULL) {

    mode <- match.arg(mode)
    x <- NULL
    message("Basic Coverage: ", mode)
    ##########
    # Determine indices of coverage in 'binding' for every primer:
    # motivation: reduce memory consumption of 'foreach' call (doesn't require full 'binding' object)
    #############
    fw.bindings <- vector("list", nrow(primers))
    rev.bindings <- vector("list", nrow(primers))
    meta.cols <- names(binding@metadata)
    for (x in seq_len(nrow(primers))) {
        primer.type <- primers$Direction[x]
        primer.indices <- get.primer.binding.idx(binding, primer.type, x,
                                ifelse(mode == "on_target", allowed.other.binding.ratio, 1))
        cur.bindings <- list(fw = binding[primer.indices$fw, ], rev = binding[primer.indices$rev, ])
        # adjust metadata to selection
        for (y in seq_along(cur.bindings)) {
            cur.binding <- cur.bindings[[y]]
            dir <- names(cur.bindings)[y]
            dir.idx <- primer.indices[[dir]] # select only metadata relating to current direction
            for (col in meta.cols) {
                metadata(cur.bindings[[y]])[[col]] <- metadata(cur.binding)[[col]][dir.idx]
            }
        }
        fw.bindings[[x]] <- cur.bindings[["fw"]]
        rev.bindings[[x]] <- cur.bindings[["rev"]]
    }
    # define iteration vars
    # fw.binding/rev.binding are used in the iteration such that the full object is not needed in the loop -> saves memory!
    fw.binding <- NULL
    rev.binding <- NULL
    x <- NULL
    cvg.df <- foreach(fw.binding = fw.bindings, rev.binding = rev.bindings, x = seq_len(nrow(primers)), .combine = "rbind") %dopar% {
		##############
		# NB: DO NOT updateProgress here -> breaks parallel evaluation under Windows in the frontend!
        #if (is.function(updateProgress)) {
         #   detail <- ""
          #  updateProgress(x/nrow(primers), detail, "set")
        #}
		###############
        # blueprint output
        p.result <- data.frame(primer_coverage = integer(1), Coverage_Ratio = numeric(1), 
            Binding_Position_Start_fw = character(1), Binding_Position_End_fw = character(1), 
            Binding_Position_Start_rev = character(1), Binding_Position_End_rev = character(1), 
            Binding_Region_Allowed = character(1), Binding_Region_Allowed_fw = character(1),
            Binding_Region_Allowed_rev = character(1), Nbr_of_mismatches_fw = character(1), 
            Nbr_of_mismatches_rev = character(1), Mismatch_pos_fw = character(1), 
            Mismatch_pos_rev = character(1), Covered_Seqs = character(1), 
            # relative binding positions relative to start/end of primers:
            Relative_Forward_Binding_Position_Start_fw = character(1),
            Relative_Forward_Binding_Position_End_fw = character(1), 
            Relative_Forward_Binding_Position_Start_rev = character(1),
            Relative_Forward_Binding_Position_End_rev = character(1), 
            Relative_Reverse_Binding_Position_Start_fw = character(1),
            Relative_Reverse_Binding_Position_End_fw = character(1),
            Relative_Reverse_Binding_Position_Start_rev = character(1),
            Relative_Reverse_Binding_Position_End_rev = character(1),
            stringsAsFactors = FALSE)
        primer.type <- primers$Direction[x]
        #######
        # A: determine covered templates
        ########
        covered.seqs.idx.fw <- metadata(fw.binding)$template_idx
        covered.seqs.idx.rev <- metadata(rev.binding)$template_idx
        ###############
        # Mismatch Info
        ################
        fw.mm.info <- get.mismatch.info(primers[x,], template.df, fw.binding, covered.seqs.idx.fw, "fw")
        rev.mm.info <- get.mismatch.info(primers[x,], template.df, rev.binding, covered.seqs.idx.rev, "rev")
        # select only the best binding mode in each template target region: select binding w/ smallest nbr of mismatches
        sel.idx.fw <- select_best_binding(fw.binding, fw.mm.info)
        sel.idx.rev <- select_best_binding(rev.binding, rev.mm.info)
        ########
        # update the mismatch info to retain only the selected events
        #########
        # update mismatch info: retain only best mismatch info
        fw.mm.info$Pos <- fw.mm.info$Pos[sel.idx.fw]
        fw.mm.info$Nbr <- fw.mm.info$Nbr[sel.idx.fw]
        rev.mm.info$Pos <- rev.mm.info$Pos[sel.idx.rev]
        rev.mm.info$Nbr <- rev.mm.info$Nbr[sel.idx.rev]
        # update covered seqs: unique for every template
        if (primer.type == "both") {
            # templates should be covered by both directions
			covered.seqs.idx <- intersect(metadata(fw.binding)$template_idx[sel.idx.fw], 
			metadata(rev.binding)$template_idx[sel.idx.rev])
		} else if (primer.type == "fw") {
            covered.seqs.idx <- metadata(fw.binding)$template_idx[sel.idx.fw]
		} else {
            covered.seqs.idx <- metadata(rev.binding)$template_idx[sel.idx.rev]
		}
        covered.seqs <- template.df$Identifier[covered.seqs.idx]
        #####
        # B: determine binding positions in the templates
        ######
        if (length(fw.binding) != 0) {
            # on-target binding: select best binding mode
            primer.exon.start <- fw.binding@start[sel.idx.fw]  # binding pos in the seq
            primer.exon.end <- (fw.binding@start + fw.binding@width - 1)[sel.idx.fw]
       } else {
            primer.exon.start <- NULL
            primer.exon.end <- NULL
       } 
       if (length(rev.binding) != 0) {
            primer.exon.start.rev <- rev.binding@start[sel.idx.rev]
            primer.exon.end.rev <- (rev.binding@start + rev.binding@width - 1)[sel.idx.rev]
        } else {
            primer.exon.start.rev <- NULL
            primer.exon.end.rev <- NULL
        }
        #########
        # C: Relative binding positions
        #########
        # n.b.: not using _initial_here!
        # relative to fw leader:
        rel.primer.start.f <- get.relative.binding.pos(template.df$Allowed_End_fw, primer.exon.start, "fw", covered.seqs.idx)
        rel.primer.end.f <- get.relative.binding.pos(template.df$Allowed_End_fw, primer.exon.end, "fw", covered.seqs.idx)
        rel.primer.start.rev.f <- get.relative.binding.pos(template.df$Allowed_End_fw, primer.exon.start.rev, "fw", covered.seqs.idx)
        rel.primer.end.rev.f <- get.relative.binding.pos(template.df$Allowed_End_fw, primer.exon.end.rev, "fw", covered.seqs.idx)
        # relative to rev leader:
        rel.primer.start.r <- get.relative.binding.pos(template.df$Allowed_Start_rev, primer.exon.start, "rev", covered.seqs.idx)
        rel.primer.end.r <- get.relative.binding.pos(template.df$Allowed_Start_rev, primer.exon.end, "rev", covered.seqs.idx)
        rel.primer.start.rev.r <- get.relative.binding.pos(template.df$Allowed_Start_rev, primer.exon.start.rev, "rev", covered.seqs.idx)
        rel.primer.end.rev.r <- get.relative.binding.pos(template.df$Allowed_Start_rev, primer.exon.end.rev, "rev", covered.seqs.idx)
        ############
        # D: Allowed binding regions check
        ###########
        bound.to.allowed.region.fw <- metadata(fw.binding)$allowed_region[sel.idx.fw]
        bound.to.allowed.region.rev <- metadata(rev.binding)$allowed_region[sel.idx.rev]
        if (primer.type == "both") {
            if (length(bound.to.allowed.region.fw) == 0 || length(bound.to.allowed.region.rev) == 
              0) {
                bound.to.allowed.region <- NULL
            } else {
              bound.to.allowed.region <- bound.to.allowed.region.fw & bound.to.allowed.region.rev
            }
        } else if (primer.type == "fw") {
            bound.to.allowed.region <- bound.to.allowed.region.fw
        } else {
            bound.to.allowed.region <- bound.to.allowed.region.rev
        }
        ############
        # Output results
        #############
        coverage.count <- length(covered.seqs)
        p.result$primer_coverage <- coverage.count
        p.result$Coverage_Ratio <- coverage.count/nrow(template.df)
        # binding position in seqs
        p.result$Binding_Position_Start_fw <- paste(primer.exon.start, collapse = ",")
        p.result$Binding_Position_End_fw <- paste(primer.exon.end, collapse = ",")
        p.result$Binding_Position_Start_rev <- paste(primer.exon.start.rev, collapse = ",")
        p.result$Binding_Position_End_rev <- paste(primer.exon.end.rev, collapse = ",")
        # relative position with regard to fw binding region: fw primer:
        p.result$Relative_Forward_Binding_Position_Start_fw <- paste(rel.primer.start.f, 
            collapse = ",")
        p.result$Relative_Forward_Binding_Position_End_fw <- paste(rel.primer.end.f, 
            collapse = ",")
        # rev primer
        p.result$Relative_Forward_Binding_Position_Start_rev <- paste(rel.primer.start.rev.f, 
            collapse = ",")
        p.result$Relative_Forward_Binding_Position_End_rev <- paste(rel.primer.end.rev.f, 
            collapse = ",")
        # rel to rev binding region:
        p.result$Relative_Reverse_Binding_Position_Start_fw <- paste(rel.primer.start.r, 
            collapse = ",")
        p.result$Relative_Reverse_Binding_Position_End_fw <- paste(rel.primer.end.r, 
            collapse = ",")
        p.result$Relative_Reverse_Binding_Position_Start_rev <- paste(rel.primer.start.rev.r, 
            collapse = ",")
        p.result$Relative_Reverse_Binding_Position_End_rev <- paste(rel.primer.end.rev.r, 
            collapse = ",")
        # allowed binding?
        p.result$Binding_Region_Allowed_fw <- paste(bound.to.allowed.region.fw, 
            collapse = ",")
        p.result$Binding_Region_Allowed_rev <- paste(bound.to.allowed.region.rev, 
            collapse = ",")
        p.result$Binding_Region_Allowed <- paste(bound.to.allowed.region, collapse = ",")
        p.result$Nbr_of_mismatches_fw <- paste(fw.mm.info$Nbr, collapse = ",")
        p.result$Nbr_of_mismatches_rev <- paste(rev.mm.info$Nbr, collapse = ",")
        if (length(fw.mm.info$Pos) != 0) {
            mm.info <- paste(sapply(fw.mm.info$Pos, function(x) paste("(", paste(x, 
              collapse = ","), ")", sep = "")), collapse = ",")
        } else {
            mm.info <- ""
        }
        p.result$Mismatch_pos_fw <- mm.info
        if (length(rev.mm.info$Pos) != 0) {
            mm.info <- paste(sapply(rev.mm.info$Pos, function(x) paste("(", paste(x, 
              collapse = ","), ")", sep = "")), collapse = ",")
        } else {
            mm.info <- ""
        }
        p.result$Mismatch_pos_rev <- mm.info
        p.result$Covered_Seqs <- paste(covered.seqs, collapse = ",")
        # message(paste0(x, "/", nrow(primers))) # iteration status
        p.result
    }
    return(cvg.df)
}
#' Evaluation of Primer Coverage.
#'
#' Evaluates the coverage of a set of primers.
#'
#' @param template.df Template data frame.
#' @param primers Primer data frame.
#' @param mode.directionality Primer directionality.
#' @param settings A \code{DesignSettings} object.
#' @param updateProgress Progress callback function for shiny.
#' @return Primer data frame with information on the covered template sequences.
#' @keywords internal
evaluate.primer.cvg <- function(template.df, primers, mode.directionality = c("fw", "rev", "both"), 
                                settings, updateProgress = NULL) {

    # determine basic coverage: string matching with mismatches
    basic.cvg.df <- evaluate.basic.cvg(template.df, primers, mode.directionality = mode.directionality, 
                                       conOptions(settings)$allowed_mismatches,
                                       conOptions(settings)$allowed_other_binding_ratio,
                                       conOptions(settings)$allowed_region_definition,
                                       updateProgress = updateProgress)
    out.df <- basic.cvg.df
    colnames(out.df) <- paste0("Basic_", colnames(out.df))
    # annotate primers with cvg:
    primers <- update.constraint.values(primers, basic.cvg.df)
    constrained.cvg.df <- evaluate.constrained.cvg(template.df, primers, basic.cvg.df, 
                            mode.directionality, settings, updateProgress = updateProgress)
    out.df <- cbind(out.df, constrained.cvg.df)
    return(out.df)
}
#' Evaluation of Primer Coverage.
#'
#' Evaluates the coverage of a set of primers.
#'
#' @param template.df Template data frame.
#' @param primer.df Primer data frame.
#' @param cvg.df Data frame with basic coverage entries.
#' @param mode.directionality Primer directionality.
#' @param settings A \code{DesignSettings} object.
#' @param updateProgress Progress callback function for shiny.
#' @return Primer data frame with information on the covered template sequences.
#' @keywords internal
evaluate.constrained.cvg <- function(template.df, primer.df, cvg.df, mode.directionality = c("fw", "rev", "both"), 
                                    settings, updateProgress = NULL) {

    if (length(mode.directionality) == 0) {
        stop("Please supply the 'mode.directionality' argument.")
    }
    mode.directionality <- match.arg(mode.directionality)
    if (length(template.df) == 0 || nrow(template.df) == 0 || length(primer.df) == 0  || nrow(primer.df) == 0) {
        return(NULL)
    }
    #t <- proc.time()["elapsed"]
    message("Computing constrained coverage ...")
    #######
    # Selection of binding events
    #######
    cvg.constraints <- cvg_constraints(settings)
    active.constraints <- names(cvg.constraints)
    if (any(grepl("hexamer_coverage", names(conOptions(settings))))) {
        active.constraints <- c(active.constraints, "hexamer_coverage")
    }
    new.df <- check_cvg_constraints(primer.df, template.df, settings, 
                                active.constraints = active.constraints,
                                updateProgress = updateProgress)
    filter.res <- filter.by.constraints(new.df, new.df, cvg.constraints, 
                                      names(cvg.constraints), mode.directionality, template.df)$Filtered
    # only keep the columns relating to coverage:
    keep.cols.basic <- colnames(cvg.df)
    keep.cols.cvg <- colnames(new.df)[unlist(lapply(names(cvg_constraints(settings)), function(x) grep(x, colnames(new.df))))]
    keep.cols <- union(keep.cols.basic, keep.cols.cvg)
    out <- filter.res[, keep.cols[keep.cols %in% colnames(filter.res)]]
    return(out)
}
get_duplex_events <- function(fw.df, annealing.temp, ions, primer.df) {
    if (length(fw.df) == 0 || nrow(fw.df) == 0) {
        return(NULL)
    }
    #print("get_duplex_Events:")
    #print("fw.df:")
    #print(fw.df)
    combi.df <- ddply(fw.df, c("Primer", "Template"), summarize,
                             PrimerIdentifier = paste(substitute(PrimerIdentifier), collapse = ","), TemplateIdentifier = paste(substitute(TemplateIdentifier), collapse = ","))
    p.ids <- unlist(lapply(combi.df$PrimerIdentifier, function(x) strsplit(x, split = ",")[[1]][1]))
    duplex.result.fw <- get.dimer.data(combi.df$Primer, combi.df$Template, annealing.temp[match(p.ids, primer.df$Identifier)], ions, no.structures = TRUE) 
    if (length(duplex.result.fw) != 0) {
        # get lowest DeltaG for all ambiguities
        duplex.result.fw <- ddply(duplex.result.fw, c("Idx1"), function(x) arrange(x, substitute(DeltaG))[1, ])
        # restore original dimensions of dat
        m <- lapply(seq_len(nrow(combi.df)), function(x) {
            primer.id <- strsplit(combi.df$PrimerIdentifier[x], split = ",")[[1]]
            template.id <- strsplit(combi.df$TemplateIdentifier[x], split = ",")[[1]]
            mm <- unlist(lapply(seq_along(primer.id), function(y) which(fw.df$PrimerIdentifier == primer.id[y] & fw.df$TemplateIdentifier == template.id[y])))
        })
        deltaG <- rep(NA, nrow(fw.df))
        for (i in seq_len(nrow(duplex.result.fw))) {
            deltaG[m[[i]]] <- duplex.result.fw$DeltaG[i]
        }
        duplex.result.fw <- cbind(fw.df, DeltaG = deltaG)
    }
    return(duplex.result.fw)
}
#' Determination of the Free Binding Energy.
#'
#' Computest the free energy of annealing between primers and templates. If the \code{mode} is set to "on_target", 
#' the free energies of binding events in the allowed region are computed, while if the \code{mode} is set to
#' "off_target", the free energies of off-target events are computed.
#'
#' @param primer.df A \code{Primers} object.
#' @param template.df A \code{Templates} object.
#' @param annealing.temp The vector of optimal annealing temperatures of the primers.
#' @param settings A \code{DesignSettings} object.
#' @param mode  If the \code{mode} is set to "on_target", 
#' the free energies of binding events in the allowed region are computed, while if the \code{mode} is set to
#' "off_target", the free energies of off-target events are computed.
#' @return A list of lists containing the numeric free energies of the annealing events for every primer.
#' @keywords internal
get.duplex.energies <- function(primer.df, template.df, annealing.temp, settings, mode = c("on_target", "off_target")) {
    ions <- compute.sodium.equivalent.conc(PCR(settings)$Na_concentration, PCR(settings)$Mg_concentration, 
                                           PCR(settings)$K_concentration, PCR(settings)$Tris_concentration)
    if (length(mode) == 0) {
        stop("Please provide the 'mode' argument.")
    }
    #print("duplex energies: temperatures are:")
    #print(annealing.temp)
    mode <- match.arg(mode)
    if (mode == "on_target") {
        # on target binding events
        all.covered.seq.idx <- covered.seqs.to.idx(primer.df$Covered_Seqs, template.df)
        primer.starts.fw <- lapply(strsplit(primer.df$Binding_Position_Start_fw, split = ","), as.numeric)
        primer.ends.fw <- lapply(strsplit(primer.df$Binding_Position_End_fw, split = ","), as.numeric)
        primer.starts.rev <- lapply(strsplit(primer.df$Binding_Position_Start_rev, split = ","), as.numeric)
        primer.ends.rev <- lapply(strsplit(primer.df$Binding_Position_End_rev, split = ","), as.numeric)
    } else {
        # off target binding events
        all.covered.seq.idx <- covered.seqs.to.idx(primer.df$Off_Covered_Seqs, template.df)
        primer.starts.fw <- lapply(strsplit(primer.df$Off_Binding_Position_Start_fw, split = ","), as.numeric)
        primer.ends.fw <- lapply(strsplit(primer.df$Off_Binding_Position_End_fw, split = ","), as.numeric)
        primer.starts.rev <- lapply(strsplit(primer.df$Off_Binding_Position_Start_rev, split = ","), as.numeric)
        primer.ends.rev <- lapply(strsplit(primer.df$Off_Binding_Position_End_rev, split = ","), as.numeric)
    }
    all.covered.templates <- unlist(lapply(all.covered.seq.idx, function(x) template.df$Sequence[x]))
    subject <- DNAStringSet(all.covered.templates)
    w <- Biostrings::width(subject)  # template lengths
    # cumulative index in the concatenation of template strings
    idx <- c(0, cumsum(head(w, length(w) - 1)))
    s <- unlist(subject)
    template.indices <- c(0, cumsum(sapply(all.covered.seq.idx, length)))
    # don't parallelize: dimerization is already parallelized!
    fw.p <- vector("list", nrow(primer.df))
    fw.t <- vector("list", nrow(primer.df))
    fw.ids <- vector("list", nrow(primer.df))
    rev.p <- vector("list", nrow(primer.df))
    rev.t <- vector("list", nrow(primer.df))
    for (i in seq_len(nrow(primer.df))) {
        primer.type <- primer.df$Direction[i]
        if (template.indices[[i]] == template.indices[[i+1]]) {
            # no coverage
            next
        }
        cur.template.indices <- seq(template.indices[[i]] + 1, template.indices[[i+1]])
        cur.idx.adjustment <- idx[cur.template.indices]
        f.fw <- NULL
        f.rev <- NULL
        if (primer.type == "fw") {
            # fw
            f.fw <- Views(s, primer.starts.fw[[i]] + cur.idx.adjustment, primer.ends.fw[[i]] + cur.idx.adjustment)  
        } else if (primer.type == "rev") {
            # rev
            f.rev <- Views(s, primer.starts.rev[[i]] + cur.idx.adjustment, primer.ends.rev[[i]] + cur.idx.adjustment)  
        } else {
            # both
            f.fw <- Views(s, primer.starts.fw[[i]] + cur.idx.adjustment, primer.ends.fw[[i]] + cur.idx.adjustment)  
            f.rev <- Views(s, primer.starts.rev[[i]] + cur.idx.adjustment, primer.ends.rev[[i]] + cur.idx.adjustment)  
        }
        # forward deltaG:
        fw.templates <- rev.comp.sequence(tolower(as.character(f.fw)))
        fw.primers <- rep(primer.df$Forward[i], length(fw.templates))
        # reverse deltaG:
        rev.templates <- tolower(as.character(f.rev))
        rev.primers <- rep(primer.df$Reverse[i], length(rev.templates))
        # write-out:
        fw.p[[i]] <- fw.primers
        fw.t[[i]] <- fw.templates
        rev.p[[i]] <- rev.primers
        rev.t[[i]] <- rev.templates
    }
    fw.ids <- unlist(lapply(seq_along(fw.p), function(x) rep(as.character(primer.df$Identifier[x]), length(fw.p[[x]]))))
    fw.t.ids <- unlist(all.covered.seq.idx[match(unique(fw.ids), primer.df$Identifier)])
    fw.df <- data.frame(PrimerIdentifier = fw.ids, TemplateIdentifier = fw.t.ids, Primer = unlist(fw.p), Template = unlist(fw.t), stringsAsFactors = FALSE)
    # get the free energies for fw primers
    duplex.data.fw <- get_duplex_events(fw.df, annealing.temp, ions, primer.df)
    rev.ids <- unlist(lapply(seq_along(rev.p), function(x) rep(as.character(primer.df$Identifier[x]), length(rev.p[[x]]))))
    rev.t.ids <- unlist(all.covered.seq.idx[match(unique(rev.ids), primer.df$Identifier)])    
    rev.df <- data.frame(PrimerIdentifier = rev.ids, TemplateIdentifier = rev.t.ids, Primer = unlist(rev.p), Template = unlist(rev.t), stringsAsFactors = FALSE)
    # get the free energies for rev primers
    duplex.data.rev <- get_duplex_events(rev.df, annealing.temp, ions, primer.df)
    # unify fw/rev data
    out <- vector("list", nrow(primer.df))
    for (i in seq_len(nrow(primer.df))) {
        # nb: 'nrow' is important, otherwise length of data.frame is defined by its number of columns when the data frame is empty
        if (length(duplex.data.fw) != 0) {
            cur.fw.data <- duplex.data.fw[duplex.data.fw$PrimerIdentifier == primer.df$Identifier[i], ]
        } else {
            cur.fw.data <- data.frame()
        }
        if (length(duplex.data.rev) != 0) {
            cur.rev.data <- duplex.data.rev[duplex.data.rev$PrimerIdentifier == primer.df$Identifier[i], ]
        } else {
            cur.rev.data <- data.frame()
        }
        if (nrow(cur.fw.data) != 0 && nrow(cur.rev.data) != 0) {
            out[[i]] <- sapply(seq_len(nrow(cur.fw.data)), function(x) min(cur.fw.data$DeltaG[x], cur.rev.data$DeltaG))
            #print(i)
            #print(out[[i]])
        } else if (nrow(cur.fw.data) != 0) {
            out[[i]] <- cur.fw.data$DeltaG
        } else if (nrow(cur.rev.data) != 0) {
            out[[i]] <- cur.rev.data$DeltaG
        }
    }
    # control:
    # ctrl.df <- cbind(fw.df, "DeltaG" = unlist(out))
    # print(ctrl.df)
    return(out)
}
#' Identification of Mutations Induced by Mismatch Binding Events.
#'
#' Identifies whether mutations are induced by mismatch binding events.
#'
#' Checks for one primer and all covered templates whether any templates are bound
#' with mismatches such that mismatches are induced. A numeric vector indicating 
#' which binding events induce a forbidden mismatch according to 
#' \code{mutation.types} is returned such that \code{1} indicates forbidden events
#' and \code{0} allowed events.
#'
#' @param primer.df A \code{Primers} object.
#' @param template.df A \code{Template} object.
#' @param mutation.types Character vector of the mutation types to be checked for.
#' @return A list containing data frames where an entry of 1 is present if the \code{primer.seq} induces a mutation that is
#' forbidden according to the provided \code{mutation.types}, otherwise 0.
#' @keywords internal
mismatch.mutation.check <- function(primer.df, template.df, mutation.types = c("stop_codon", "substitution")) {
    mutation.types <- match.arg(mutation.types, several.ok = TRUE) 
    all.covered.seqs <- strsplit(primer.df$Covered_Seqs, split = ",")
    ORF.data <- get.ORFs(template.df)
    primer.starts.fw <- lapply(strsplit(primer.df$Binding_Position_Start_fw, split = ","), as.numeric)
    primer.ends.fw <- lapply(strsplit(primer.df$Binding_Position_End_fw, split = ","), as.numeric)
    primer.starts.rev <- lapply(strsplit(primer.df$Binding_Position_Start_rev, split = ","), as.numeric)
    primer.ends.rev <- lapply(strsplit(primer.df$Binding_Position_End_rev, split = ","), as.numeric)
    x <- NULL
    #for (x in seq_i
    stop.list <- foreach(x = seq_len(nrow(primer.df)), .combine = "c") %dopar% {
        covered.seqs <- all.covered.seqs[[x]]
        stop.check.fw <- check.mutations(primer.df$Forward[x], primer.starts.fw[[x]], primer.ends.fw[[x]],
                                           template.df, covered.seqs, ORF.data, "fw", mutation.types)
        stop.check.rev <- check.mutations(primer.df$Reverse[x], primer.starts.rev[[x]], primer.ends.rev[[x]], 
                                            template.df, covered.seqs, ORF.data, "rev", mutation.types)
        primer.type <- primer.df$Direction[x]
        if (primer.type == "both") {
            # any stop codon in a pair -> exclusion
            if (length(stop.check.fw) != 0 && length(stop.check.rev) != 0) {
                stop.check <- stop.check.fw | stop.check.rev
            } else {
                stop.check <- NULL
           }
        } else if (primer.type == "fw") {
            stop.check <- stop.check.fw
        } else if (primer.type == "rev") {
            stop.check <- stop.check.rev
        }
        # output 1 for stop codon and 0 for no stop codon
        stop.check <- 1 * stop.check # numeric
        list(stop.check)
    }
    return(stop.list)
}
get_feature_matrix <- function(primer.df, template.df, mode = c("on_target", "off_target")) {
    # transform to a long-data frame of primer-coverage events
    mode <- match.arg(mode)
    if (!"primer_coverage" %in% colnames(primer.df)) {
        stop("Primer coverage required for model.")
    }
    if (mode == "on_target") {
        if (!"annealing_DeltaG" %in% colnames(primer.df)) {
            stop("Annealing free energy required for model.")
        }
    } else {
        if (!"off_annealing_DeltaG" %in% colnames(primer.df)) {
            stop("Annealing free energy required for model.")
        }
    }
    full.df <- prepare_mm_plot(primer.df, template.df, mode = mode)
    # consider all identified events (this is constrained here, as we use the 'basic coverage entries' (not filtered yet!)
    full.df <- full.df[full.df$Coverage_Type == "constrained",]
    # select unique binding event for all mismatch contacts:
    # the minimal deltaG is selected to ensure that we consider the
    # disambiguated primer with the least nbr of mismatches
    df <- ddply(full.df, c("Primer", "Template", "Group"), summarize,
                        annealing_DeltaG = min(substitute(annealing_DeltaG)),
                        Number_of_mismatches = min(substitute(Number_of_mismatches)),
                        Position_3terminus = min(substitute(Position_3terminus)),
                        Position_3terminusLocal = max(substitute(Position_3terminusLocal)))
    if (length(df) != 0 && nrow(df) != 0 && (any(is.na(df$annealing_DeltaG) | is.na(df$Position_3terminusLocal)))) {
        warning("Some values of the coverage model feature matrix were NA; this shouldn't happen!")
    }
    return(df)
}
#' Prediction of Primer Coveragee.
#'
#' Predicts primer coverage using a logistic regression model.
#' Converts coverage probabilities to expected false positive rate
#' for a given probability.
#'
#' @param primer.df A \code{Primers} data frame.
#' @param template.df A \code{Templates} data frame.
#' @param settings A \code{DesignSettings} object.
#' @param mode Whether on-target or off-target events shall be considered.
#' @return The predictions for primer coverage
#' @keywords internal
predict_coverage <- function(primer.df, template.df, settings, mode = c("on_target", "off_target"), updateProgress = NULL) {
    ##################
    # sysdata used:
        # CVG_MODEL (the predictive model)
        # FPR_TABLE (probability to FPR conversion for predictions)
    #########
    if (length(mode) == 0) {
        mode <- "on_target"
    } else {
        mode <- match.arg(mode)
    }
    if (!is(primer.df, "Primers")) {
        stop("Please input a primer data frame.")
    }
    if (!is(template.df, "Templates")) {
        stop("Please provide a valid template data frame.")
    }
    mode.directionality <- get.analysis.mode(primer.df)
    # compute annealingDeltaG:
    cvg_constraints(settings) <- list("annealing_DeltaG" = c("max" = Inf))
    #print(paste0("mode: ", mode))
    if (mode == "on_target") {
        p.df <- compute.constraints(primer.df, mode.directionality, 
                    template.df, settings, 
                    active.constraints = "annealing_DeltaG")
    } else {
        p.df <- compute.constraints(primer.df, mode.directionality, 
                    template.df, settings, 
                    active.constraints = "off_annealing_DeltaG")
    }
    primer.df <- update.constraint.values(primer.df, p.df)
    pred.matrix <- get_feature_matrix(primer.df, template.df, mode = mode)
    if (length(pred.matrix) != 0 && nrow(pred.matrix) != 0) {
        pred <- try(predict(CVG_MODEL, newdata = pred.matrix, type = "response"))
    } else {
        # if 'newdata' is NULL, the training data would be predicted ..
        pred <- NULL
    }
    if (is(pred, "try-error")) {
        stop("Could not predict coverage using the logistic model. Maybe some columns are missing in the data frame? Provided columns were: ", colnames(pred.matrix), ". Dimension of matrix:", dim(pred.matrix))
    }
    # check whether predictions are available for all primer coverage events
    if (length(pred) != 0 && any(is.na(pred))) {
        idx <- which(is.na(pred))
        warning("No prediction available for: ", paste(pred.matrix$Primer[idx], collapse = ","))
    }
    #print(pred)
    # transform coverage probabilities to FPR
    idx <- unlist(lapply(pred, function(x) which.min(abs(x - FPR_TABLE$Cutoff))))
    fpr <- FPR_TABLE$FPR[idx]
    # check function
    #summary <- data.frame("Coverage_Probability" = round(pred, 3), "FPR" = round(fpr, 3))
    #print(summary)
    # need to output values as a list, one for each primer; one entry for every coverage event
    if (mode == "on_target") {
        idx <- covered.seqs.to.idx(primer.df$Covered_Seqs, template.df)
    } else {
        idx <- covered.seqs.to.idx(primer.df$Off_Covered_Seqs, template.df)
    }
    pred.idx <- match(pred.matrix$Template, template.df$ID) # idx of templates in the template data frame 
    # annotate each primer with its FPR vector
    if (length(pred.matrix) != 0) {
        out <- lapply(seq_len(nrow(primer.df)), function(x) {
            p.idx <- which(as.character(pred.matrix$Primer) == as.character(primer.df$ID[x]))
            t.idx.pred <- pred.idx[p.idx] # template idx
            t.idx.p <- idx[[x]]
            m <- match(t.idx.pred, t.idx.p)
            if (any(is.na(m))) {
                stop("Fatal error in predict_coverage: could not find coverage event! Maybe the template IDs were non-unique?!")
            }
            fpr[p.idx][m]
        })
    } else {
        out <- vector("list", nrow(primer.df))
    }
    return(out)
}
#' Evaluation of Coverage.
#'
#' Evaluates primer coverage.
#' 
#' @param template.seqs Template sequences as a \code{DNAStringSet}.
#' @param primers Primer sequences as a \code{DNAStringSet}.
#' @param mode.directionality Directionality of primres
#' @param allowed.mismatches Allowed number of mismatches between a primer and a template.
#' @param updateProgress Progress function for shiny
#' @return IRanges object with primer coverage information.
#' @keywords internal
evaluate.cvg <- function(template.seqs, primers, 
                    mode.directionality = c("fw", "rev"), 
                    allowed.mismatches, updateProgress = NULL) {

    # todo: implement progress function?
    if (length(mode.directionality) == 0) {
        stop("Please provide the 'mode.directionality' arg.")
    }
    mode.directionality <- match.arg(mode.directionality)
    if (length(template.seqs) == 0 || length(primers) == 0) {
        return(NULL)
    }
    if (!is(primers, "DNAStringSet")) {
        stop("primers must be a DNAStringSet.")
    }
    if (!is(primers, "DNAStringSet")) {
        stop("seqs must be a DNAStringSet.")
    }
    # consider ambiguous templates up to a cutoff of degeneration
    seqs <- my.disambiguate(template.seqs)
    # determine mapping from disambiguated seqs to input seqs
    l.p <- unlist(lapply(primers, length))
    l.s <- unlist(lapply(seqs, length))
    l.p <- rep(seq_along(primers), l.p)
    l.s <- rep(seq_along(seqs), l.s)
    seqs <- unlist(seqs)
    seqDB <- unlist(seqs)  # one continuous string of all template seqs
    # determine mapping of pos in seqDB to template id
    w <- Biostrings::width(seqs)  # template lengths
    idx <- c(0, cumsum(w))
    w <- c(0, w)
    template.mapping <- IRanges(start = idx[seq_len(length(idx) - 1)] + 1, width = w[2:length(w)])
    names(template.mapping) <- l.s
    f <- IRanges()  # forward start/end positions in the templates
    fp <- integer()  # fp: forward primer indices, one index for every template hit
    non.empty.idx <- which(Biostrings::width(primers) != 0)
    i <- GRanges() # need to define 'i'; used 'GRanges' outside the foreach loop to prevent building error on macOS
    f <- foreach(i = seq_along(non.empty.idx), .combine = c) %dopar% {

        idx <- non.empty.idx[i]
        # matchPattern returns all matches below the max mismatch cutoff! need to select the best match later on
        temp <- matchPattern(primers[[idx]], seqDB, max.mismatch = allowed.mismatches, 
                                        with.indels = FALSE, fixed = FALSE) # fixed FALSE to allow ambigs
        temp <- as(temp, "IRanges")
        # add additional info
        fp <- rep(idx, length(temp))
        # store the primer idx as metadata using GRanges
        if (length(temp) != 0) {
            temp <- GRanges(seqnames = mode.directionality, ranges = temp, primer_idx = fp)  # changed from RangedData (deprecated)
        } else {
            temp <- GRanges()
        }
    }
    if (length(f) == 0) {
        return(IRanges())
    }
    fp <- as.numeric(as.character(f$primer_idx))  # primer indices
    f <- Views(seqDB, start(f), end(f))  # extract the sub-sequence of the primer matches
    # try to map from 'f' Iranges to template.mapping to find which template seqs
    # were bound
    overlaps <- findOverlaps(f, template.mapping)  # f is query, template.mapping is the subject
    fw.table <- f[as.matrix(overlaps)[, 1]] # as.matrix from IRanges ..
    fw.idx <- fp[as.matrix(overlaps)[, 1]]
    fw.template.table <- template.mapping[as.matrix(overlaps)[, 2]]  # names: identifier of seq
    # extract regions of binding: subtract fw.table from fw.template.table
    fw.tab <- as(fw.table, "IRanges")
    fw.starts <- fw.tab@start - fw.template.table@start + 1
    fw.ends <- fw.starts + fw.tab@width - 1
    template.ends <- fw.template.table@width
    r <- fw.starts > 0 & fw.ends <= template.ends
    df <- data.frame(Primer = fw.idx, Seq = fw.template.table@NAMES)
    if (mode.directionality == "rev") {
        binding.regions <- IRanges(start = fw.template.table@width[r] - fw.starts[r] + 
                                    1 - fw.tab@width[r] + 1, width = fw.tab@width[r])
    } else {
        binding.regions <- IRanges(start = fw.starts[r], width = fw.tab@width[r])
    }
    metadata(binding.regions) <- list(primer_idx = fw.idx[r], template_idx = as.numeric(fw.template.table@NAMES[r]))
    return(binding.regions)
}
#' Information about Mismatches.
#' 
#' Computes information about mismatch binding events.
#'
#' @param primer Primer character vector.
#' @param seqs Template binding sequences of primers as a \code{XStringsView} object.
#' @return List with positions and number of mismatches of the \code{primer} in the 
#' \code{seqs}. The list contains the field \code{mm.pos} containing a list
#' with the positions of the mismatches and the field \code{Nbr} containing
#' a numeric vector with the number of mismatches per template binding event.
#' @keywords internal
mismatch.info <- function(primer, seqs) {
    if (length(seqs) == 0) {
        return(NULL)
    }
    p <- my.disambiguate(DNAStringSet(primer))[[1]]
    s <- my.disambiguate(DNAStringSet(seqs))
    # select template with minimal mismatches
    As <- lapply(p, function(p.seq) {
        modes <- lapply(s, function(s.seq) compareStrings(rep(tolower(as.character(p.seq)), length(s.seq)), tolower(as.character(s.seq))))
        mm.pos <- lapply(modes, function(x) lapply(strsplit(x, split = ""), function(y) which(y == "?")))
        mm.count <- lapply(mm.pos, function(x) sapply(x, length))
        idx <- unlist(lapply(mm.count, function(x) which.min(x)))
        res <- lapply(seq_along(idx), function(x) mm.pos[[x]][[idx[x]]])
    })
    # select primer with minimal mismathches
    mm.pos <- lapply(seq_along(seqs), function(x) {
        nbr <- sapply(As, function(p.seqs) length(p.seqs[[x]]))
        sel <- which.min(nbr)
        As[[sel]][[x]]
    })
    mm.count <- sapply(mm.pos, length) # number of mismatches
    result <- list(Pos = mm.pos, Nbr = mm.count)
    return(result)
}
#' Annotation of Primer Binding Events.
#'
#' Annotates whether primer binding events are in the allowed binding region or not. 
#'
#' @param fw.binding IRanges with coverage information.
#' @param allowed.range IRanges of the allowed binding ranges in the templates.
#' @param nbr.primers Number of primers to consider.
#' @param allowed.region.definition Definition of the allowed binding region
#' @return IRanges with annotations of (preliminary) specificity and allowed binding. 
#' The field \code{all_binding} contains all binding regions, \code{on_target} contains
#' all events in the target region, and \code{off_target} contains all off-target
#' binding events.
#' @keywords internal
annotate.binding.events <- function(fw.binding, allowed.range, nbr.primers,
                                 allowed.region.definition = c("within", "any")) {
    if (length(allowed.region.definition) == 0) {
        stop("Please provide the 'allowed.region.definition' argument.")
    }
    allowed.region.definition <- match.arg(allowed.region.definition)
    out <- list("on_target" = IRanges(), "all_binding" = IRanges(), "off_target" = IRanges())
    if (length(fw.binding) == 0) {
        return(out)  # nothing to filter/annotate
    }
    t.idx <- metadata(fw.binding)$template_idx
    p.idx <- metadata(fw.binding)$primer_idx
    t.u <- unique(t.idx)
    # determine allowed binding events
    p.u <- unique(p.idx)
    # determine the allowed range corresponding to each binding event in 'fw.binding' -> replicate allowed range
    cur.bindings <- vector("list", nbr.primers)
    template.indices <- vector("list", nbr.primers)
    for (i in seq_along(p.u)) {
        idx <- which(p.idx == p.u[i])
        template.indices[[i]] <- t.idx[idx]  # current set of templates to be checked for primer binding
        cur.bindings[[i]] <- fw.binding[idx]
    }
    # unify all bindings
    all.bindings <- do.call(c, cur.bindings)
    all.allowed <- allowed.range[unlist(template.indices)]
    # either checks for 'any' overlaps or 'within' allowed region overlaps of primers with template regions
    allowed <- overlapsAny(all.bindings, all.allowed, type = allowed.region.definition)
    metadata(fw.binding) <- c(metadata(fw.binding), allowed_region = list(allowed))
    # select on-target binding events
    allowed.binding <- fw.binding[allowed]
    # update existing metadata
    for (col in names(allowed.binding@metadata)) {
        metadata(allowed.binding)[[col]] <- metadata(fw.binding)[[col]][allowed]
    }
    # select off-target binding events: 
    disallowed.binding <- fw.binding[!allowed]
    # update existing metadata
    for (col in names(disallowed.binding@metadata)) {
        metadata(disallowed.binding)[[col]] <- metadata(fw.binding)[[col]][!allowed]
    }
    out <- list("on_target" = allowed.binding, "all_binding" = fw.binding, "off_target" = disallowed.binding)
    return(out)
}

####################
# Create data for comparison of primers
#############
remove.all.duplicates.from.fasta <- function(primers, primers.fixed) {
    for (i in seq_along(primers)) {
        sample <- names(primers)[i]
        message(sample)
        primer.loc <- primers[i]
        print(primer.loc)
        primer.files <- list.files(path = primer.loc, pattern = ".fasta", full.names = TRUE)
        out.loc <- primers.fixed[sample]
        dir.create(out.loc, showWarnings=FALSE)
        for (j in seq_along(primer.files)) {
            primer.location <- primer.files[j]
            cur.primers <- read.fasta(primer.location)
            seqs <- sapply(cur.primers, function(x) paste(as.character(x), collapse = ""))
            # check for duplicates
            dup.idx <- which(duplicated(seqs))
            # remove > character at the start from annotation
            ids <- sapply(cur.primers, function(x) substring(attr(x, "Annot"), 2))
            if (length(dup.idx) != 0) {
                warning(paste("Duplicates found in: ", primer.location, sep = ""))
                seqs <- seqs[-dup.idx]
                ids <- ids[-dup.idx]
            }
            # remove reverse primers
            rev.idx <- which(grepl("_rev", ids))
            if (length(rev.idx) != 0) {
                message(paste("Removing reverse primers: ", paste(ids[rev.idx], collapse = ","), sep = ""))
                seqs <- seqs[-rev.idx]
                ids <- ids[-rev.idx]
            }
            out.file <- file.path(out.loc, basename(primer.location))
            message(out.file)
            write.fasta(as.list(seqs), names = as.list(ids), file.out = out.file)
        }
    }
}
create.primer.set.evaluation.data <- function(data, settings, comp.results.loc, 
                                              primer.locations, rm.partials, max.degen, fw.id, rev.id) {

    primer.folders <- list.files(path = primer.locations, full.names = TRUE)
    constraint.settings <- constraints(settings)
    na_salt_conc <- PCR(settings)$Na_concentration
    mg_salt_conc <- PCR(settings)$Mg_concentration
    k_salt_conc <- PCR(settings)$K_concentration
    tris_salt_conc <- PCR(settings)$Tris_concentration
    primer_conc <- PCR(settings)$primer_concentration
    for (i in seq_along(data)) {
        sample <- names(data)[i]
        message(sample)
        lex.df <- data[[i]]
        cur.results.loc <- file.path(comp.results.loc, sample)
        dir.create(cur.results.loc, showWarnings=FALSE)
        write.csv(lex.df, file = file.path(comp.results.loc, paste(sample, "_templates.csv", sep = "")))
        # list all primer sets in folder
        primer.files <- list.files(path = primer.folders[i], full.names = TRUE)
        for (j in seq_along(primer.files)) {
            primer.location <- primer.files[j]
            bname <- basename(primer.location)
            run.info <- unlist(strsplit(bname, split = "_"))
            if (grepl("IPS", bname)) {
                run <- run.info[2]
            } else {
                run <- run.info[1]
            }
            if (any(grepl("1st", run.info))) {
                run <- paste(run, "_1st", sep = "")
            } 
            if (any(grepl("2nd", run.info))) {
                run <- paste(run, "_2nd", sep = "")
            }
            run <- sub(".fasta", "", run)
            message(paste("Loading: ", primer.location, sep = ""))
            message(paste("Primer set: ", run, sep = ""))
            set.name <- sub(".fasta", "", basename(primer.location))
            primer.df  <- read_primers(primer.location, fw.id, rev.id, merge.ambig = "none", max.degen)
            mode.directionality <- get.analysis.mode(primer.df)
            # need primer cvg to compute annealing  temperature estimate
            active.constraints <- names(constraint.settings)
            constraint.df <- check_constraints(primer.df, lex.df, settings, 
                                active.constraints)
            # add Run info to constraint.df
            fname <- file.path(cur.results.loc, paste(sample, "_", set.name, ".csv", sep = ""))
            message(paste("Writing: ", fname, sep = ""))
            write.csv(constraint.df, file = fname, row.names =FALSE)
        }
    }
}

get_templates <- function() {
    # initial retrieval templates from IMGT
    script.loc <- system.file("inst", "shiny", "shiny_server", 
                              "extra_IMGT_template_set_extractor.py",
                    package = "openPrimeRui")
    if (script.loc == "") {
        stop("Please load the 'openPrimeRui' package first.")
    }
    # define what we want to extract:
    species <- "Homo sapiens"
    loci <- c("IGH", "IGK", "IGL")
    funcs <- c("any", "functional")
    for (i in seq_along(loci)) {
        locus <- loci[i]
        for (j in seq_along(funcs)) {
            func <- funcs[i]
            # retrieve & store leader+variable in extdata folder
            openPrimeRui::retrieve.IMGT.templates(species, locus, func, TRUE)
        }
    }
    message("templates reside in extdata target dir now :-)")
}
get_primers <- function() {
    # initial retrieval of primers from IMGT
    # n.b.: primers are not written to extdata here because they were processed by hand
    script.loc <- system.file("data-raw", "IMGT_primer_set_extractor.py",
                    package = "openPrimeR")
    cmd <- paste("Python", script.loc)
    res <- system(cmd) # stored in data/IMGT_data/IMGT_primers
    # combine literature and IMGT primer sets
    data.dir <- system.file("data-raw", "IMGT_data", package = "openPrimeR")
    IMGT.loc <- file.path(data.dir, "IMGT_primers")
    # n.b.: for literature primers: apply mac2unix first to change formatting from mac
    lit.loc <- file.path(data.dir, "literature_primers")
    primer.loc <- file.path(data.dir, "primers_raw") # primers_raw/ -> (manual) primers_mod/ -> primers/
    # copy primers from both locations
    IMGT.dirs <- list.files(IMGT.loc, full.names = TRUE)
    out.dirs <- file.path(primer.loc, basename(IMGT.dirs))
    names(out.dirs) <- basename(IMGT.dirs)
    lit.dirs <- list.files(lit.loc, full.names = TRUE)
    for (i in seq_along(out.dirs)) {
        out.dir <- out.dirs[i]
        dir.create(out.dir, recursive = TRUE)
        res <- dir.copy(IMGT.dirs[i], out.dir, overwrite = TRUE)
        res <- dir.copy(lit.dirs[i], out.dir, overwrite = TRUE)
    }
    warning(paste("extdata was not updated with primers, need to manually modify adaptors.",
                "If you don't want this, use the existing processed primers."))
    # adaptor removal was done by hand -> manually create primers/ folder entries
}
generate_evaluation_primers <- function(settings) {
    # evaluates the comparison primer sets according  to the provided settings
    # load templates:
    template.loc <- system.file("extdata", "IMGT_data", "templates",
                                package = "openPrimeR")
    hdr.structure <- c("ACCESSION", "GROUP", "SPECIES", "FUNCTION")

    template.locs <- file.path(template.loc, c("Homo_sapiens_IGH_functional_exon_exp.fasta",
                                               "Homo_sapiens_IGK_functional_exon.fasta",
                                               "Homo_sapiens_IGL_functional_exon.fasta"))
    leader.locs <- file.path(template.loc, c("Homo_sapiens_IGH_functional_leader_exp.fasta",
                                               "Homo_sapiens_IGK_functional_leader.fasta",
                                               "Homo_sapiens_IGL_functional_leader.fasta"))
    functional.data <- lapply(seq_along(template.locs), function(x) {
                            template.file <- template.locs[x]
                            leader.file <- leader.locs[x]
                            template.df <- read_templates(template.file, hdr.structure, "|", "GROUP", 
                                                          rm.keywords = "partial")
                            template.df <- assign_binding_regions(template.df, leader.file, NULL)
                            template.df <- adjust_binding_regions(template.df, 
                                           c(-max(template.df$Allowed_End_fw), 0), NULL) # allow binding in the first exon base as well
    })
    names(functional.data) <- c("IGH", "IGK", "IGL") # change names for shiny app comparison data
    # store template data in comparison folder:
    for (i in seq_along(functional.data)) {
        fname <- file.path(system.file("extdata", "IMGT_data", "comparison", "templates", package = "openPrimeR"),
                    paste0(names(functional.data)[i], "_templates.csv"))
        write_templates(functional.data[[i]], fname, "CSV")
    }
    # primers to load for comparison:
    primer.locations <- system.file("extdata", "IMGT_data", "primers",
                                    package = "openPrimeR")
    # output of set evaluations:
    comp.primer.out <- system.file("extdata", "IMGT_data", "comparison",
                                    "primer_sets", package = "openPrimeR")
    create.primer.set.evaluation.data(functional.data, settings,
                                      comp.primer.out, primer.locations, 
                                      rm.partials = TRUE, max.degen = 16,
                                      fw.id = "_fw", rev.id = "_rev")
}
generate_Comparison <- function(settings) {
    # store comparison data for IGH: primers and templates  
    comp.files.primer <- list.files(path = system.file("extdata", "IMGT_data", 
                            "comparison", "primer_sets", "IGH", package = "openPrimeR"),
                            pattern = "*\\.csv", full.names = TRUE)
    comp.files.seqs <- rep(system.file("extdata", "IMGT_data", "comparison", 
                        "templates", "IGH_templates.csv", package = "openPrimeR"),
                        length(comp.files.primer))
    primer.data <- read_primers(comp.files.primer)
    template.data <- read_templates(comp.files.seqs)
    # update template cvg:
    template.data <- lapply(seq_along(template.data), function(x) {
                            update_template_cvg(template.data[[x]], primer.data[[x]])
    })
    out.loc <- system.file("data", "Comparison.rda", package = "openPrimeR")
    save(primer.data, template.data, settings, file = out.loc, compress = "xz")
}
###########
# SOURCES
#############
devtools::load_all()
# load settings
settings.filename <- system.file("extdata", "settings", "B_Taq_PCR_evaluate.xml", package = "openPrimeR")
settings <- read_settings(settings.filename)
refresh_src <- FALSE # should source primers/templates be retrieved once again?
if (refresh_src) {
    # WARNING: should never be run as primers were selected by hand after retrieval from the data sources
    get_templates()
    get_primers()
}
generate_evaluation_primers(settings) # evaluate primers
# n.b. need to copy templates from primer to template folder manually atm
generate_Comparison(settings) # create rda file

#' @rdname Settings
#' @name Settings
#' @aliases DesignSettings-class
#' @details
#' Note that for the \code{DesignSettings} class, the fields \code{Input_Constraints}, \code{Input_Constraint_Boundaries}, and \code{Coverage_Constraints} should 
#' contain entries with at most two components using the fields \code{min} and/or \code{max}.
#' The \code{Input_Constraint_Boundaries} should always be at least as general as the
#' specified \code{Input_Constraints}.
#'
#' @exportClass DesignSettings
#' @return The \code{DesignSettings} constructor defines a \code{DesignSettings} object.
#' 
#' @family settings functions
#' @keywords Settings
#' @seealso \code{\link{read_settings}} for reading settings from XML files,
#' \code{\link{write_settings}} for storing settings as XML files,
#' \code{\link{constraints}} for accessing constraints,
#' \code{\link{constraintLimits}} for accessing constraint boundaries,
#' \code{\link{cvg_constraints}} for accessing coverage constraints,
#' \code{\link{conOptions}} for accessing constraint options,
#' \code{\link{PCR}} for accessing the PCR conditions.
#' @examples
#' 
#' # Load a settings object
#' filename <- system.file("extdata", "settings", 
#'                  "C_Taq_PCR_high_stringency.xml", package = "openPrimeR")
#' settings <- read_settings(filename)
#' # Modify the constraints
#' constraints(settings)$gc_clamp["min"] <- 0
#' # Modify the constraint limits for designing primers
#' constraintLimits(settings)$gc_clamp["max"] <- 6
#' # Modify the coverage constraints
#' cvg_constraints(settings)$primer_efficiency["min"] <- 0.001
#' # Modify the PCR conditions
#' PCR(settings)$Na_concentration <- 0.0001
#' # Modify the constraint options
#' conOptions(settings)$allowed_mismatches <- 0
DesignSettings <- setClass("DesignSettings",
	slots = c( # constructor arguments:
			Input_Constraints = "ConstraintSettings",
			Input_Constraint_Boundaries = "ConstraintSettings",
            Coverage_Constraints = "CoverageConstraints",
			PCR_conditions = "PCR_Conditions",
			constraint_settings = "ConstraintOptions"
			),
	validity=function(object)
	{
		return(check_settings_validity(object))
	}
)
#' Check of limit correctness.
#'
#' Checks whether a constraint limit is more general than the setting.
#'
#' @param setting A single constraint setting.
#' @param limit A single constraint limit.
#' @return A vector containing \code{TRUE} if the limit is more general
#' than the constraint setting and \code{FALSE} otherwise.
#' @keywords internal
check_limit_value <- function(setting, limit) {
    if (length(limit) == 0) { # no limit specified
        # nothing to check
        return(TRUE)
    }
    if (length(names(setting)) == 0 || !all(names(setting)  %in% c("min", "max"))) {
        stop("Please provide min/max as fields for constraints.")
    }
    if (length(names(limit)) == 0 || !all(names(limit) %in% c("min", "max"))) {
        stop("Please provide min/max as fields for constraint limits.")
    }
    ret <- NA
    if (length(setting) == 2) {
        ret <- c(limit[1] <= setting[1], limit[2] >= setting[2])
    } else if (length(setting) == 1 && names(setting) == "max") {
        ret <- limit["max"] >= setting["max"]
    } else if (length(setting) == 1 && names(setting) == "min") {
        ret <- limit["min"] <= setting["min"]
    } else {
        message("Unkown constraint length:")
        message(setting)
        message(limit)
        ret <- NA
    }
    na.idx <- which(is.na(ret))
    ret[na.idx] <- FALSE # missing value in one of the settings
    return(ret)
}
#' Correction of Constraint Boundaries.
#'
#' Fixes the constraint boundaries if they are more narrow than the current settings.
#'
#' @param constraints A list with constraint settings.
#' @param constraint.limits A list with constraint limits.
#' @param fix.limit Whether the constraint limits should be
#' adjusted. If \code{FALSE}, the constraint settings are adjusted.
#' @return The corrected constraint limits.
#' @keywords internal
fix_constraint_boundaries <- function(constraints, constraint.limits, fix.limit = TRUE) {
    iter.range <- NULL
    if (fix.limit) {
        iter.val <- constraints # the constraints for iteration
        other.val <- constraint.limits # the constraints to be adjusted
    } else {
        iter.val <- constraint.limits
        other.val <- constraints
    }
    for (i in seq_along(iter.val)) {
        con <- names(iter.val)[i]
        if (!con %in% names(other.val)) {
            next
        }
        other.setting <- other.val[[con]]
        setting <- iter.val[[con]]
        #print(con)
        #print(setting)
        #print(other.setting)
        if (fix.limit) {
            test <- check_limit_value(setting, other.setting)
        } else {
            test <- check_limit_value(other.setting, setting)
        }
        if (!all(test)) {
            #warning(paste("Had to fix constraint limit of constraint:",
                #con), ". Constraint limits should be more general than the setting.")
            # limit needs to be adjusted to the current setting
            other.setting[names(setting[!test])] <- setting[!test]
            other.val[[con]] <- other.setting
        } 
    }
    # sync the constraints with the constraintlimits: addition/deletion of constraints
    # remove all constraints that are present in 'other.val' but not in 'iter.val'
    rm.cons <- setdiff(names(other.val), names(iter.val))
    if (length(rm.cons) != 0) {
        other.val <- other.val[-which(names(other.val) %in% rm.cons)]
    }
    # add constraints that are present in 'iter.val' but not in 'other.val'
    add.cons <- setdiff(names(iter.val), names(other.val))
    # set missing entries to have the identical constraint and boundary since we don't know better:
    other.val[add.cons] <- iter.val[add.cons]
    # ensure that constraints are ordered correctly 
    # (the order agrees with the order of the options, since we ordered in constraints()/constraintLimits())
    con.order <- names(iter.val)
    other.val <- other.val[con.order[con.order %in% names(other.val)]]
    return(other.val)
}

setMethod("show", "DesignSettings", function(object) {
    con.tab <- create.constraint.table(constraints(object), 
                                       constraintLimits(object))
    # raw constraint name output: users should know the identifiers that are correct from the output.
    inactive.cons <- setdiff(names(object@Input_Constraints@status), names(constraints(object)))
    con.result <- list("Active" = con.tab, "Inactive" = inactive.cons)
    cvg.con.tab <- create.constraint.table(cvg_constraints(object), NULL)
    inactive.cons <- setdiff(names(object@Coverage_Constraints@status), names(cvg_constraints(object)))
    cvg.con.result <- list("Active" = cvg.con.tab, "Inactive" = inactive.cons)
    PCR.tab <- create.PCR.table(PCR(object))
    inactive.options <- setdiff(names(object@PCR_conditions@status), names(PCR(object)))
    PCR.result <- list("Active" = PCR.tab, "Inactive" = inactive.options)
    opt.tab <- create.options.table(conOptions(object))
    inactive.options <- setdiff(names(object@constraint_settings@status), names(conOptions(object)))
    opt.result <- list("Active" = opt.tab, "Inactive" = inactive.options)
    out <- list("Constraints" = con.result, "Coverage Constraints" = cvg.con.result,
                "PCR conditions" = PCR.result, "Options" = opt.result)
})

setMethod("initialize", "DesignSettings",
    function(.Object, constraint.settings, constraint.limits, cvg.constraints,
             PCR.conditions, constraint.options) {
        # just call the default constructor
        #print("INITIALIZATION:")
        # .Object is the prototype from the class for which slots should be set
        if (is(constraint.settings, "ConstraintSettings")) {
            # ok
        } else if (is.null(constraint.settings)) {
            constraint.settings <- ConstraintSettings()   
        } else if (is(constraint.settings, "list")) {
            obj <- ConstraintSettings()
            constraints(obj) <- constraint.settings
            constraint.settings <- obj
        } else {
            stop("Please provide 'constraint.settings' as a 'ConstraintSettings' object or a list that can be converted to a 'ConstraintSettings' object.")
        }
        .Object@Input_Constraints <- constraint.settings
        if (is(constraint.limits, "ConstraintSettings")) {
            # ok
         } else if (is.null(constraint.limits)) {
            constraint.limits <- ConstraintSettings()   
         } else if (is(constraint.limits, "list")) {
            obj <- ConstraintSettings()
            constraints(obj) <- constraint.limits
            constraint.limits <- obj
         } else {
            stop("Please provide 'constraint.limits' as a 'ConstraintSettings' object or a 'list'.")
        }
        .Object@Input_Constraint_Boundaries <- constraint.limits
        if (is(cvg.constraints, "CoverageConstraints")) {
            # ok
        } else if (is.null(cvg.constraints)) {
            cvg.constraints <- CoverageConstraints() # no cvg constraints 
        } else if (is(cvg.constraints, "list")) {  
            obj <- CoverageConstraints()
            constraints(obj) <- cvg.constraints
            cvg.constraints <- obj
        } else {
            stop("Please provide 'cvg.constraints' as a 'CoverageConstraints' object or a 'list'.")
        }
        .Object@Coverage_Constraints <- cvg.constraints
        if (is(PCR.conditions, "PCR_Conditions")) {
            # ok
        } else if (is.null(PCR.conditions)) {
            stop("'PCR_conditions' are mandatory, but were not provided.")
        } else if (is(PCR.conditions, "list")) {
            obj <- PCR_Conditions()
            constraints(obj) <- PCR.conditions
            PCR.conditions <- obj
        } else {
            stop("Please provide 'PCR.conditions' as a 'PCR_Conditions' object or a 'list'.")
        }
        .Object@PCR_conditions <- PCR.conditions
        if (is(constraint.options, "ConstraintOptions")) {
            # ok
        } else if (is.null(constraint.options)) {
            stop("'ConstraintOptions' are mandatory, but were not provided.")
        } else if (is(constraint.options, "list")) {
            obj <- ConstraintOptions()
            constraints(obj) <- constraint.options
            constraint.options <- obj
        } else {
            stop("Please provide 'constraint.options' as a 'ConstraintOptions' object.")
        }
        .Object@constraint_settings <- constraint.options
        if (is(.Object, "try-error")) {
            stop("Could not initialize DesignSettings: did you supply all args?\n", attr(.Object, "condition"))
        }
        # specify the order of constraint evaluation for the design procedure 
        # users can modify the order by changing the options() setting
        con.order <- getOption("openPrimeR.constraint_order")
        # reduce possible constraints according to available tools
        sel.constraints <- con_select(names(constraints(.Object@Input_Constraints)))
        rm.cons <- setdiff(names(constraints(.Object@Input_Constraints)), sel.constraints)
        if (length(rm.cons) != 0) {
            message("Due to missing external tools the following constraints were ignored:",
                    paste(rm.cons, collapse = ","))
        }
        sel.o <- con.order[con.order %in% sel.constraints]
        constraints(.Object@Input_Constraints) <- constraints(.Object@Input_Constraints)[sel.o]
        sel.constraints <- con_select(names(constraints(.Object@Input_Constraint_Boundaries)))
        sel.o <- con.order[con.order %in% sel.constraints]
        constraint.boundaries <- fix_constraint_boundaries(constraints(.Object@Input_Constraints), constraints(.Object@Input_Constraint_Boundaries)[sel.o])
        constraints(.Object@Input_Constraint_Boundaries) <- constraint.boundaries
        cvg.sel <- select.constraints(names(constraints(.Object@Coverage_Constraints))) # select.constraints here since we're not storing cvg constraints in option
        rm.cvg <- setdiff(names(constraints(.Object@Coverage_Constraints)), cvg.sel)
        if (length(rm.cvg) != 0) {
            message("Due to missing external tools the following coverage constraints were ignored:",
                    paste(rm.cvg, collapse = ","))
        }
        constraints(.Object@Coverage_Constraints) <- constraints(.Object@Coverage_Constraints)[cvg.sel]
        validObject(.Object)
        .Object
    }
) 

#' Check Setting Names.
#'
#' Checks whether the specified settings hvae the correct names.
#'
#' @param known.options Allowed setting names.
#' @param input.options Input setting names
#'
#' @return Mapping of \code{input.options} to \code{known.options} or \code{NULL} if invalid.
#' @keywords internal
check_names <- function(known.options, input.options) {
    # checks whether the names of 
    errors <- NULL
    m <- match(input.options, known.options)
    if (any(is.na(m))) {
        # unknown option
        errors <- c(errors, paste("Unknown/missing field in object:", paste(input.options[is.na(m)], collapse = ",")))
    }
    if (length(errors) != 0) {
        return(errors)
    } else {
        return(m)
    }
}

#' Check Setting Validity.
#'
#' Checks whether the input settings are valid or not.
#'
#' @param known.options Vector with names and classes of allowed options.
#' @param options Active options to be checked.
#' @param mandatory.options Fields that have to be present.
#' @return \code{TRUE} if the setting is valid, \code{FALSE} otherwise.
#' @keywords internal
check_setting <- function(known.options, options, mandatory.options = NULL) {
    # checks whether a setting in DesignSettings is valid or not.
    if (length(options) == 0) {
        return(TRUE)
    }
    # check for correct field names:
    errors <- NULL
    if (!is.null(known.options)) {
        m <- check_names(names(known.options), names(options))
        if (is.character(m)) { # error
            return(m)
        }
        c <- sapply(seq_along(options), function(x) !(class(options[[x]]) %in% known.options[m][[x]]) & !is.null(options[[x]]))
        if (any(c)) {
            idx <- which(c)
            msgs <- sapply(idx, function(x) paste("Supplied class incorrect: ",  names(options[which(c)]), 
                    ".Class was: ", sapply(options, class)[c], ", but should have been: ", 
                    paste(known.options[m][[which(c)]], collapse = " or "), ".", sep = ""))
            return(msgs)
        }
    }
    # check for mandatory field names:
    if (!is.null(mandatory.options)) {
        m <- check_names(names(options), names(mandatory.options))
        if (is.character(m)) {
            # error
            return(m)
        }
        if (is(options, "data.frame")) {
            c <- sapply(seq_along(mandatory.options), function(x) !(class(asS3(options)[[m[x]]]) %in% mandatory.options[[x]]))
        } else {
            c <- sapply(seq_along(mandatory.options), function(x) !(class(options[m][[x]]) %in% mandatory.options[[x]]))
        }
        if (any(c)) {
            idx <- which(c)
            msgs <- sapply(idx, function(x) paste("Supplied class was incorrect for '",  names(mandatory.options)[x],
                    "'. Class was '", sapply(options, class)[m[x]], "', but should have been '", 
                    paste(mandatory.options[[x]], collapse = " or "), "'.", sep = ""))
            return(msgs)
        }
    }
    return(TRUE)
}

#' Check Constraint Intervals
#'
#' Checks the validity of constraint intervals.
#'
#' @param constraints A list with constraint settings.
#' @return \code{TRUE}, if all constraints specificy valid intervals, 
#' \code{FALSE} otherwise.
#' @keywords internal
check_interval <- function(constraints) {
    check <- rep(FALSE, length(constraints))
    for (i in seq_along(constraints)) {
        con <- constraints[[i]]
        #print("check_interval:")
        #print(names(constraints)[i])
        #print(con)
        if (!all(is.numeric(con) | is.integer(con))) {
            return("Please supply only integer/numeric values.")
        }
        if (length(names(con)) == 0) {
            check[i] <- FALSE
        } else if (length(con) == 1) {
            # only min or max condition
            if (names(con) %in% c("min", "max")) {
                check[i] <- TRUE
            }
        } else if (length(con) == 2) {
            # min & max conditions
            if (any(names(con) == "min") &&
                any(names(con) == "max") && 
                con["min"] <= con["max"]) {
                check[i] <- TRUE
            } else {
                # min/max not specified or min larger than max
                check[i] <- FALSE
            }
        } else {
            # other lengths are not allowed
            check[i] <- FALSE
        }
    }
    if (all(check)) {
        return(TRUE)
    } else {
        details <- names(constraints)[!check]
        msg <- (paste("Constraint interval (min/max) was not specified properly ",
                    "for the following constraints:", 
                    paste(details, collapse = ","), 
                    ".\nDid you supply a named vector containing the components 'min' and/or 'max' such that 'min' <= 'max' if both entries were indeed supplied?", sep = ""))
        return(msg)
    }
}
#' Validity Check for Limits.
#'
#' Checks whether the constraint limits are at least as
#' general as the constraint settings. This ensures that
#' the relaxation works in the proper direction.
#'
#' @param constraint.settings A list with the constraint settings.
#' @param constraint.limits A list with the constraint relaxation limits.
#' @return \code{TRUE} if the limits are at least as wide as the constraints, 
#' \code{FALSE} otherwise.
#' @keywords internal
check_limits <- function(constraint.settings, constraint.limits) {
    check <- sapply(seq_along(constraint.settings), function(x) {
        id <- names(constraint.settings)[x]
        setting <- constraint.settings[[id]]
        limit <- constraint.limits[[id]]
        return(all(check_limit_value(setting, limit)))
        }
    )
    idx <- which(!check)
    if (length(idx) != 0) {
        msg <- paste("Whoops - A Limit was tighter than the constraint setting:",
                    paste(names(constraint.limits)[idx], collapse = ","))
        warning(msg)
        return(FALSE)
    } else {
        return(TRUE)
    }
}
#' Validity Check for DesignSettings.
#'
#' Validates whether a DesignSettings object has the correct structure.
#'
#' @param object A DesignSettings object to be checked for validity.
#' @return \code{TRUE} if \code{object} is valid, FALSE otherwise.
#' @keywords internal
check_settings_validity <- function(object) {
    # ensure that constraint limits are at least as wide as the constraint settings
    # check_limits shouldn't produce an error, just warn :-)
    check.limits <- check_limits(constraints(object@Input_Constraints), constraints(object@Input_Constraint_Boundaries))
	return(TRUE)
}
##########
# GETTERS
############

#' @rdname Settings
#' @name Settings
#' @aliases constraints
#' @exportMethod constraints
#' @return \code{constraints} gets a list with the active constraint settings.
#' @keywords Settings
setGeneric("constraints", function(x) standardGeneric("constraints"))

#' @rdname Settings
#' @name Settings
#' @aliases constraints,DesignSettings-method
setMethod("constraints", "DesignSettings", function(x) {
    sel <- names(constraints(x@Input_Constraints))
    #print("Constraint getter:")
    #print(class(x@Input_Constraints))
    constraints(x@Input_Constraints)[con_select(sel)]
})

#' @rdname Settings
#' @name Settings
#' @aliases constraints,AbstractConstraintSettings-method
setMethod("constraints", c("AbstractConstraintSettings"), 
    function(x) {
        return(x@settings)    
    }
)

#' @rdname Settings
#' @name Settings
#' @aliases cvg_constraints
#' @exportMethod cvg_constraints
#' @return \code{cvg_constraints} returns the list of active coverage constraints.
setGeneric("cvg_constraints", function(x) standardGeneric("cvg_constraints"))

#' @rdname Settings
#' @name Settings
#' @aliases cvg_constraints,DesignSettings-method
setMethod("cvg_constraints", "DesignSettings", function(x) {
    sel <- names(constraints(x@Coverage_Constraints))
    # select only the possible constraints from the settings (tool dependencies):
    constraints(x@Coverage_Constraints)[select.constraints(sel)]
})

#' Getter for Filtering Constraints.
#'
#' Gets the constraints on the physicochemical properties
#' that are used for the filtering procedure when designing primers
#' using the \code{Input_Constraints} slot of the provided
#' \code{DesignSettings} object \code{x}.
#' @return Gets the list of filtering constraints.
#' @name filters
#' @keywords internal
#' @rdname filters-methods
setGeneric("filters", function(x) standardGeneric("filters"))

#' @rdname filters-methods
#' @aliases filters,DesignSettings-method
#' @param x A \code{DesignSettings} object.
setMethod("filters", "DesignSettings", function(x) {
    # return all filters except for those only relevant to the optimization procedure:
    opti.names <- c("melting_temp_diff", "cross_dimerization")
    sel <- setdiff(names(constraints(x@Input_Constraints)), opti.names)
    constraints(x@Input_Constraints)[con_select(sel)]
})

#' Getter for Filtering Constraint Limits.
#'
#' Gets the limits on the constraints
#' that are used for the filtering procedure when designing primers
#' using the \code{Input_Constraint_Boundaries} slot of the provided
#' \code{DesignSettings} object \code{x}.
#' @return Gets the list of filtering limits.
#' @name filterLimits
#' @rdname filterLimits-methods
#' @keywords internal
setGeneric("filterLimits", function(x) standardGeneric("filterLimits"))

#' @rdname filterLimits-methods
#' @aliases filterLimits,DesignSettings-method
#' @param x A \code{DesignSettings} object.
setMethod("filterLimits", "DesignSettings", function(x) {
    opti.names <- c("melting_temp_diff", "cross_dimerization")
    sel <- setdiff(names(constraints(x@Input_Constraint_Boundaries)), opti.names)
    constraints(x@Input_Constraint_Boundaries)[con_select(sel)]
})

#' Getter for Optimization Constraints.
#'
#' Gets the constraints on the physicochemical properties
#' that are applied just before the optimization procedure
#' using the \code{Input_Constraints} slot of the provided
#' \code{DesignSettings} object \code{x}.
#' @return Gets the list of optimization constraints.
#' @name opti
#' @rdname opti-methods
#' @keywords internal
setGeneric("opti", function(x) standardGeneric("opti"))

#' @rdname opti-methods
#' @aliases opti,DesignSettings-method
#' @param x A \code{DesignSettings} object.
setMethod("opti", "DesignSettings", function(x) {
    # define the constraints for filtering in the optimization procedure
    opti.names <- c("melting_temp_diff", "cross_dimerization")
    sel.constraints <- con_select(names(constraints(x@Input_Constraints)))[names(constraints(x@Input_Constraints)) %in% opti.names]
    opti.cons <- constraints(x@Input_Constraints)[sel.constraints]
    opti.cons
})

#' Getter for Optimization Constraint Limits.
#'
#' Gets the limits for the constraints
#' that are applied just before the optimization procedure
#' using the \code{Input_Constraint_Boundaries} slot of the provided
#' \code{DesignSettings} object \code{x}.
#' @return Gets the optimization constraint limits.
#' @name optiLimits
#' @rdname optiLimits-methods
#' @keywords internal
setGeneric("optiLimits", function(x) standardGeneric("optiLimits"))

#' @rdname optiLimits-methods
#' @aliases optiLimits,DesignSettings-method
#' @param x A \code{DesignSettings} object.
setMethod("optiLimits", "DesignSettings", function(x) {
    opti.names <- c("melting_temp_diff", "cross_dimerization")
    sel <- names(constraints(x@Input_Constraint_Boundaries))[names(constraints(x@Input_Constraint_Boundaries)) %in% opti.names]
    opti.cons <- constraints(x@Input_Constraint_Boundaries)[sel]
    opti.cons
})

#' @rdname Settings
#' @name Settings
#' @aliases PCR
#' @return \code{PCR} gets the list of PCR conditions defined in the
#' provided \code{DesignSettings} object.
#' @exportMethod PCR
setGeneric("PCR", function(x) standardGeneric("PCR"))

#' @rdname Settings
#' @name Settings
#' @aliases PCR,DesignSettings-method
setMethod("PCR", "DesignSettings", function(x) {
    constraints(x@PCR_conditions)
})

#' @rdname Settings
#' @name Settings
#' @aliases conOptions
#' @exportMethod conOptions
#' @return \code{conOptions} returns a list with constraint options.
setGeneric("conOptions", function(x) standardGeneric("conOptions"))

#' @rdname Settings
#' @name Settings
#' @aliases conOptions,DesignSettings-method
setMethod("conOptions", "DesignSettings", 
    function(x) {
        constraints(x@constraint_settings)
    }
)

#' @rdname Settings
#' @name Settings
#' @aliases constraintLimits
#' @return \code{constraintLimits} gets the list of constraint limits.
#' @exportMethod constraintLimits
setGeneric("constraintLimits", function(x) standardGeneric("constraintLimits"))

#' @rdname Settings
#' @name Settings
#' @aliases constraintLimits,DesignSettings-method
setMethod("constraintLimits", "DesignSettings", 
	function(x) {
        sel <- con_select(names(constraints(x@Input_Constraint_Boundaries)))
		constraints(x@Input_Constraint_Boundaries)[sel]
	}
)

###############
# SETTERS
###############

#' @rdname Settings
#' @name Settings
#' @aliases constraints<-
#' @details
#' For an overview of permissible constraints,
#' please consider the \code{\link{ConstraintSettings}} documentation.
#'
#' @exportMethod constraints<-
#' @return \code{constraints<-} sets the list of constraints in a \code{DesignSettings} object.
#' @examples
#' 
#' # Load some settings
#' data(Ippolito)
#' # View the active constraints
#' constraints(settings)
#' # Require a minimal GC clamp extent of 0
#' constraints(settings)$gc_clamp["min"] <- 0
#' # View available constraints
#' settings
setGeneric("constraints<-", function(x, value) standardGeneric("constraints<-"))

#' @rdname Settings
#' @name Settings
#' @aliases constraints<-,DesignSettings,list-method
setReplaceMethod("constraints", c("DesignSettings", "list"), 
    # NB: setReplaceMethod: second argument MUST be named 'value'
	function(x, value) {
        # modify constraint limits if necessary
        # a) ensure that constraints are in the right order
        con.order <- getOption("openPrimeR.constraint_order")
        sel.constraints <- con.order[con.order %in% names(value)]
        if (length(sel.constraints) != length(value)) {
           warning("The following constraint identifiers are not valid and were ignored: ",
                    paste0(names(value)[!names(value) %in% con.order], sep = ","), ".\nThe following constraints are valid: ", paste0(con.order, collapse = ","))
        }
        value <- value[con.order[con.order %in% names(value)]]
        # b) ensure that limit and constraint entries are concordant
        #print("con value: ")
        #print(value)
        fixed.limits <- fix_constraint_boundaries(value, constraintLimits(x))
        constraints(x@Input_Constraint_Boundaries) <- fixed.limits
		constraints(x@Input_Constraints) <- value
		validObject(x)
		x
	}
)

#' @rdname Settings
#' @name Settings
#' @aliases constraints<-,AbstractConstraintSettings,list-method
setReplaceMethod("constraints", c("AbstractConstraintSettings", "list"),
    function(x, value) {
        m <- match(names(value), names(x@status))
        if (any(is.na(m))) {
            warning("The following constraints were ignored due to invalidity: ",
                    paste0(names(value)[is.na(m)], sep = ","),
                    ".\nOnly the follwing options are valid: ",
                    paste0(names(x@status), collapse = ","))
        }
        value <- value[!is.na(m)]
        x@settings <- value
        # activate the input constraints, deactivate all others
        x@status[names(value)] <- TRUE
        deactivate.idx <- setdiff(seq_along(x@status), m[!is.na(m)])
        x@status[deactivate.idx] <- FALSE
        # check for correct list structure with the validity function:
        validObject(x)
        return(x)
    }
)



#' @rdname Settings
#' @name Settings
#' @aliases cvg_constraints<-
#' @exportMethod cvg_constraints<-
#' @return \code{cvg_constraints<-} sets the list of coverage constraints in the provided \code{DesignSettings} object.
#' @examples
#' 
#' # Load some settings
#' data(Ippolito)
#' # View all active coverage constraints
#' cvg_constraints(settings)
#' # Increase the maximal false positive rate to increase the sensitiviity of coverage predictions
#' cvg_constraints(settings)$coverage_model <- c("max" = 0.1) 
#' # View available coverage constraints:
#' settings
setGeneric("cvg_constraints<-", function(x, value) standardGeneric("cvg_constraints<-"))

#' @rdname Settings
#' @name Settings
#' @aliases cvg_constraints<-,DesignSettings-method
setReplaceMethod("cvg_constraints", "DesignSettings", 
	function(x, value) {
		constraints(x@Coverage_Constraints) <- value
		validObject(x)
		x
	}
)
#' @rdname Settings
#' @name Settings
#' @aliases constraintLimits<-
#' @exportMethod constraintLimits<-
#' @return \code{constraintLimits<-} sets the list of constraint limits in 
#' the provided \code{DesignSettings} object.
#' @examples
#'
#' # Load some settings
#' data(Ippolito)
#' # View the active constraint limits
#' constraintLimits(settings)
#' # Extend the GC relaxation limit
#' constraintLimits(settings)$gc_clamp <- c("min" = 0, "max" = 6)
#' # View available constraints
#' settings
setGeneric("constraintLimits<-", function(x, value) standardGeneric("constraintLimits<-"))

#' @rdname Settings
#' @name Settings
#' @aliases constraintLimits<-,DesignSettings-method
setReplaceMethod("constraintLimits", "DesignSettings", 
	function(x, value) {
        # modify the settings if necessary
        # a) ensure that constraints are in the right order
        con.order <- getOption("openPrimeR.constraint_order")
        value <- value[con.order[con.order %in% names(value)]]
        # b) ensure that constraint limits and settings are still compatible
        fixed.settings <- fix_constraint_boundaries(constraints(x), value, fix.limit = FALSE)
		constraints(x@Input_Constraint_Boundaries) <- value
        constraints(x@Input_Constraints) <- fixed.settings
		validObject(x)
		x
	}
)

#' @rdname Settings
#' @name Settings
#' @aliases PCR<-
#' @return \code{PCR<-} sets the constraint options in the provided
#' \code{DesignSettings} object.
#' @exportMethod PCR<-
#' @examples
#' 
#' # Load some settings
#' data(Ippolito)
#' # View the active PCR conditions
#' PCR(settings)
#' # Evaluate primers with a fixed annealing temperature
#' PCR(settings)$annealing_temperature <- 50 # celsius
#' # View available PCR conditions
#' settings
setGeneric("PCR<-", function(x, value) standardGeneric("PCR<-"))

#' @rdname Settings
#' @name Settings
#' @aliases PCR<-,DesignSettings-method
setReplaceMethod("PCR", "DesignSettings", 
	function(x, value) {
		constraints(x@PCR_conditions) <- value
		validObject(x)
		x
	}
)

#' @rdname Settings
#' @name Settings
#' @aliases conOptions<-
#' @exportMethod conOptions<-
#' @return \code{conOptions<-} sets the specified list of constraint options in the provided \code{DesignSettings} object.
#' @examples
#'
#' # Load some settings
#' data(Ippolito)
#' # View the active constraint options
#' conOptions(settings)
#' # Prevent mismatch binding events
#' conOptions(settings)$allowed_mismatches <- 0
#' # View available constraint options
#' settings
setGeneric("conOptions<-", function(x, value) standardGeneric("conOptions<-"))

#' @rdname Settings
#' @name Settings
#' @aliases conOptions<-,DesignSettings-method
setReplaceMethod("conOptions", "DesignSettings", 
	function(x, value) {
		constraints(x@constraint_settings) <- value
		validObject(x)
		x
	}
)
#################################

#' Parse XML Constraint Data.
#'
#' Parses the constraint settings contained in an XML object.
#'
#' @param xml_data XML object from a parsed XML file.
#' @return List with constraint settings.
#' @keywords internal
parse.constraints <- function(xml_data) {
    result <- vector("list", length(xml_data))
    for (i in seq_along(xml_data)) {
        # for every set of constraints (filtering/optimization etc.)
        data <- xml_data[[i]]
        names(result)[i] <- data$name
        con.idx <- which(names(data) == "constraint")  # the entries we are looking for 
        values <- vector("list", length(con.idx))  # one entry for every constraint (melting temp, gc content etc.)
        for (j in seq_along(con.idx)) {
            idx <- con.idx[j]
            cur.con <- data[[idx]]
            names(values)[j] <- cur.con$name
            vals <- data[[idx]]$values
            # check whether constraint is min/max or value
            is.min.max.con <- any(grepl("=", vals))
            if (is.min.max.con) {
                v <- sapply(unlist(strsplit(vals, split = ",")), function(x) strsplit(x, 
                  split = "="))
                con.type <- sapply(v, function(x) x[1])
                settings <- sapply(v, function(x) as.numeric(x[2]))  # min/max something else setting?
                names(settings) <- con.type
            } else {
                settings <- suppressWarnings(as.numeric(vals))
                if (is.na(settings)) {
                  # vals was string (active/inactive setting)
                  settings <- vals  # don't transform to numeric
                }
            }
            values[[j]] <- settings
        }
        result[[i]] <- values
    }
    return(result)
}

#' @rdname Input
#' @name Input
#' @aliases read_settings
#' @details
#' When loading a settings file with \code{read_settings}, 
#' if \code{filename} is not provided,
#' a default XMl settings file is loaded. Please review the 
#' function's examples to learn more about the default settings. If you want
#' to load custom settings, you can store a modified \code{DesignSettings}
#' object as an XML file using \code{\link{write_settings}}.
#'
#' @return \code{read_settings} returns an object of class \code{DesignSettings}.
#' @export
#' @examples
#' 
#' # Select available settings
#' available.settings <- list.files(
#'      system.file("extdata", "settings", package = "openPrimeR"), 
#'      pattern = "*.xml", full.names = TRUE)
#' # Select one of the settings and load them
#' filename <- available.settings[1]
#' settings <- read_settings(filename)
read_settings <- function(filename = list.files(
                             system.file("extdata", "settings", package = "openPrimeR"), 
                             pattern = "*.xml", full.names = TRUE),
                          frontend = FALSE) {
    # read xml constraint data from filename to constraint.settings objects
    if (length(filename) == 0) {
        stop("Please provide a non-empty filename.")
    }
    filename <- filename[1]
    if (!file.exists(filename)) {
        stop("Could not find settings file: '", filename, "'")
    }
    message("Reading settings file: ", basename(filename))
    # Validate XML prior to reading the settings
    schema.file <- system.file("extdata", "settings", 
                            "settings_schema.xsd", package = "openPrimeR")
    con.setting <- try({
        xsd <- XML::xmlTreeParse(schema.file,
                                isSchema = TRUE, useInternal = TRUE)
        doc <- XML::xmlInternalTreeParse(filename)
        xml.check <- XML::xmlSchemaValidate(xsd, doc)
        if (xml.check$status != 0) {
            msg <- xml.check$errors
            stop(paste("The settings XML file ", filename, 
                 " did not adhere to the required schema. The error was: ",
                 msg, sep = ""))
        }
        data <- XML::xmlParse(filename)
        xml_data <- XML::xmlToList(data)
        con.setting <- parse.constraints(xml_data)
        con.setting
    })
    if (is(con.setting, "try-error")) {
        msg <- "Error while parsing settings XML file. Please check your input."
        my.error("XML_Parsing_Error", msg)
    }
    # convert active/inactive to boolean
    # for each slot, convert all 'active'/'inactive' annotations to TRUE/FALSE
    for (i in seq_along(con.setting)) {
        c.setting <- lapply(con.setting[[i]], function(x)  {
            if (length(x) == 1 && x[[1]] %in% c("active", "inactive")) {
                return(ifelse(x[[1]] == "active", TRUE , ifelse(x[[1]] == "inactive", FALSE, x)))
            } else {
                return(x)
            }
        })
        con.setting[[i]] <- c.setting
    }
    if (!frontend) {
        # for backend: convert the settings to non-prefixed notation
        con.setting$PCR_conditions <- convert.PCR.units(con.setting$PCR_conditions, to.mol = TRUE)
    }
	# construct DesignSettings object
    settings <- DesignSettings(con.setting$Input_Constraints,
               con.setting$Input_Constraint_Boundaries, 
               con.setting$Coverage_Constraints, 
               con.setting$PCR_conditions, con.setting$constraint_settings)
    return(settings)
}
#' Conversion of PCR Units
#'
#' Converts frontend PCR concentration units to the units used for the backend.
#'
#' @param pcr.settings List with several PCR settings (concentrations).
#' @param to.mol If \code{TRUE}, convert to the molar concentration.
#' If \code{FALSE} convert to the unit representation in the XML.
#' @return List with concentrations for usage in the backend.
#' @keywords internal
convert.PCR.units <- function(pcr.settings, to.mol = TRUE) {
    # Ion concentrations are in mM in XML
    ions <- c("Na_concentration", "Mg_concentration", "K_concentration",
            "Tris_concentration")
    if (!all(ions %in% names(pcr.settings))) {
        stop("Concentration names have changed in xml.")
    }
    for (i in seq_along(ions)){
        if (to.mol) {
            # convert from mM to molar
            pcr.settings[[ions[i]]] <- pcr.settings[[ions[i]]] * 1e-3
        } else {
            # convert from molar to mM
            pcr.settings[[ions[i]]] <- pcr.settings[[ions[i]]] / 1e-3
        }
    }
    # Primer/template concentrations are in nM in XML
    other <- c("primer_concentration", "template_concentration")
    if (!all(other %in% names(pcr.settings))) {
        stop("Concentration names have changed in xml.")
    }
    for (i in seq_along(other)){
        if (to.mol) {
            pcr.settings[[other[i]]] <- pcr.settings[[other[i]]] * 1e-9
        } else {
            pcr.settings[[other[i]]] <- pcr.settings[[other[i]]] / 1e-9
        }
    }
    return(pcr.settings)
}

#' Gather all Coverage Constraints.
#'
#' Constructor for coverage constraint settings.
#'
#' @param allowed.stop.codons Whether mismatch binding events inducing
#' stop codons in the amino acid sequence are allowed.
#' @param allowed.efficiency Min/max for primer efficiency.
#' @param disallowed.mismatch.pos The positions from the 3' terminal
#' end of primers where mismatches shall be prevented.
#' @param allowed.anneal.deltaG Maximal allowed free energy of template-primer annealing.
#' @param allowed.substitutions Whether mismatch binding events inducing substitutions
#' in the amino acid sequence are allowed.
#' @return List with all coverage constraint settings.
#' @keywords internal
get.cvg.constraint.settings <- function(allowed.stop.codons, allowed.efficiency, 
                                        disallowed.mismatch.pos, allowed.anneal.deltaG,
                                        allowed.substitutions, allowed.coverage.model) {

    if (allowed.stop.codons) {
        # stop codons allowed
        #allowed.stop.range <- c("min" = 0, "max" = 1)
        allowed.stop.range <- NULL # don't consider constraint
    } else {
        # no stop codons allowed
        allowed.stop.range <- c("min" = 0, "max" = 0) 
    }
    if (allowed.substitutions) {
        # substitutions allowed
        #allowed.sub.range <- c("min" = 0, "max" = 1)
        allowed.sub.range <- NULL # don't consider constraint
    } else {
        # no substitutions allowed
        allowed.sub.range <- c("min" = 0, "max" = 0) 
    }
    settings <- list()
    if (length(allowed.stop.codons) != 0 && length(allowed.stop.range) != 0) {
        settings$stop_codon <- allowed.stop.range
    }
    if (length(allowed.efficiency) != 0) {
        settings$primer_efficiency <- allowed.efficiency
    }
    if (length(allowed.anneal.deltaG) != 0) {
        settings$annealing_DeltaG <- allowed.anneal.deltaG
    }
    if (length(disallowed.mismatch.pos) != 0) {
        settings$terminal_mismatch_pos <- c(disallowed.mismatch.pos + 1)
    }
    if (length(allowed.sub.range) != 0 && length(allowed.sub.range) != 0) {
        settings$substitution <- allowed.sub.range
    }
    if (length(allowed.coverage.model) != 0) {
        settings$coverage_model <- allowed.coverage.model
    }
    return(settings)
} 

#' Gather all Other Constraints (for Shiny frontend).
#'
#' Constructor for other constraint settings (non-PCR, non-filtering, non-optimization).
#'
#' @param allowed_mismatches Allowed mismatches for primers binding events.
#' @param allowed_other_binding_ratio Ratio of primers allowed to bind to non-target regions.
#' @param allowed_region_definition The definition of the allowed region.
#' @return List with all other constraint settings.
#' @keywords internal
get.other.constraint.settings <- function(allowed_mismatches, 
    allowed_other_binding_ratio, allowed_region_definition) {
  
    settings <- list(allowed_mismatches = allowed_mismatches, 
        allowed_other_binding_ratio = allowed_other_binding_ratio, 
        allowed_region_definition = allowed_region_definition)
    #if (hexamer_coverage == "active") { 
        ## don't add hexamer coverage as constraint if inactive anyway
        #settings <- settings["hexamer_coverage"] <- "active"
    #}
    return(settings)
} 
#' Constraint XML Format.
#'
#' Format constraint settings for XML output.
#'
#' @param constraints List with constraint settings.
#' @param set.name Identifier for the constraint settings.
#' @return XML string containing the constraint settings.
#' @keywords internal
constraints.xml.format <- function(constraints, set.name) {
    # create a consistent format for the constraints for xml output
    main <- "constraint"
    out <- rep("", length(constraints))
    for (i in seq_along(constraints)) {
        # message(constraints[[i]])
        if (length(names(constraints[[i]])) != 0) {
            val.string <- paste(names(constraints[[i]]), "=", constraints[[i]], collapse = ",", 
                sep = "")
        } else {
            # no min/max -> just the value itself
            val.string <- paste(constraints[[i]], collapse = ",", sep = "")
        }
        # warning: saveXML/listToXML creates many textConnections and doesn't close them!!
        # realiziation: I don't even need the XML::saveXML call after using listToXml
        a <- xmlToChar(listToXml(names(constraints)[[i]], "name"))
        b <- xmlToChar(listToXml(val.string, "values"))
        out[i] <- paste(a, b, sep = "\n")
    }
    # force closing of all connections: this could be very bad if any other connection is open for a reason...
    if (length(out) != 0) {
        result <- paste("<constraint>", out, "</constraint>", collapse = "\n", sep = "\n")
    } else {
        result <- ""
    }
    result <- paste("<constraint_set>\n", "<name>", set.name, "</name>\n", result, 
        "</constraint_set>", sep = "")
    return(result)
}

#' @rdname Output
#' @name Output
#' @aliases write_settings
#' @return \code{write_settings} returns the status from closing the connection to the output file.
#' @export
#' @examples
#' 
#' # Store settings to disk
#' xml <- system.file("extdata", "settings", 
#'        "C_Taq_PCR_high_stringency.xml", package = "openPrimeR")
#' settings <- read_settings(xml)
#' out.file <- tempfile("my_settings", fileext = ".xml")
#' write_settings(settings, out.file)
write_settings <- function(settings, fname) {
    XML <- create.constraint.XML(constraints(settings), constraintLimits(settings), 
                                 cvg_constraints(settings), PCR(settings), conOptions(settings))
    f <- file(fname)
    write(XML, f)
    return(close(f))
}
#' XML Output of Constraints
#'
#' Creates an XML summarizing all settings.
#'
#' @param filtering.constraints List with constraint settings for filtering.
#' @param c.f.lim Relaxation limits for the filtering constraints.
#' @param cvg.constraints List with constraints for coverage computations.
#' @param PCR.settings Settings for the PCR.
#' @param constraint.settings Other settings of constraints (e.g. coverage).
#' @return String in XML format containing all constraint settings.
#' @keywords internal
create.constraint.XML <- function(filtering.constraints, c.f.lim, 
                                  cvg.constraints, PCR.settings, constraint.settings) {
    comment <- "<!--These settings were automatically generated by openPrimeR. Please do not edit!-->"
    root.node <- "<design_settings>"
    end.node <- "</design_settings>"
    prefix <- character(0)
    ###############
    # Constraints
    ##############
    if (length(filtering.constraints) != 0) {
        m <- match(names(filtering.constraints), names(c.f.lim))
        c.f.lim <- c.f.lim[m[!is.na(m)]]
    }
    if (length(filtering.constraints) != 0) {
        c.f.in <- constraints.xml.format(filtering.constraints, "Input_Constraints")
    } else {
        c.f.in <- ""
    }
    ######################
    # Constraints limits
    ######################
    if (length(c.f.lim) != 0) {
        c.f.lim <- constraints.xml.format(c.f.lim, "Input_Constraint_Boundaries")
    } else {
        c.f.lim <- ""
    }
    ######################
    # Coverage constraints
    ######################
    if (length(cvg.constraints) != 0) {
        c.cvg <- constraints.xml.format(cvg.constraints, "Coverage_Constraints")
    } else {
        c.cvg <- ""
    }
    #############
    # PCR settings
    ##############
    # convert the raw, molar concentrations to XML units (nM and such)
    PCR.settings <- convert.PCR.units(PCR.settings, to.mol = FALSE)
    # turn TRUE/FALSE to active/inactive
    PCR.settings <- lapply(PCR.settings, function(x) ifelse(is.logical(x[[1]]), 
                                  ifelse(x[[1]] == TRUE, "active", ifelse(x[[1]] == FALSE, "inactive", x[[1]])), 
                                  x[[1]]))
    PCR.settings <- constraints.xml.format(PCR.settings, "PCR_conditions")
    #############################
    # Constraint options
    #######################
    # need to replace TRUE/FALSE with active/inactive
    constraint.settings <- lapply(constraint.settings, function(x) ifelse(is.logical(x[[1]]), 
                                  ifelse(x[[1]] == TRUE, "active", ifelse(x[[1]] == FALSE, "inactive", x[[1]])), 
                                  x[[1]]))
    constraint.settings <- constraints.xml.format(constraint.settings, "constraint_settings")  # constraint options
    sep <- "\n"
    out <- paste(comment, root.node, c.f.in, c.f.lim, c.cvg, PCR.settings, 
                 constraint.settings, end.node, sep = sep)
    return(out)
}

#' Conversion of XML to Character.
#'
#' Converts an XML object to a character string.
#'
#' @param xml An xml object to be converted to character.
#' @return A character vector.
#' @keywords internal
xmlToChar <- function(xml) {
    file <- textConnection(NULL, "w")
    sink(file)
    on.exit({sink();close(file)})
    # we need this print statement here:
    print(xml)
    paste(textConnectionValue(file), collapse = "\n")
}
#' Convert List to XML.
#'
#' Can convert list or other object to an xml object using xmlNode.
#'
#' @title List to XML
#' @param item 
#' @param tag xml tag
#' @return xmlNode
#' @author David LeBauer, Carl Davidson, Rob Kooper
#' @keywords internal
listToXml <- function(item, tag) {
    # just textnode, or empty node with attributes
    if (typeof(item) != "list") {
        if (length(item) > 1) {
            xml <- XML::xmlNode(tag)
            for (name in names(item)) {
                XML::xmlAttrs(xml)[[name]] <- item[[name]]
            }
            return(xml)
        } else {
            return(XML::xmlNode(tag, item))
        }
    }
    
    # create the node
    if (identical(names(item), c("text", ".attrs"))) {
        # special case a node with text and attributes
        xml <- XML::xmlNode(tag, item[["text"]])
    } else {
        # node with child nodes
        xml <- XML::xmlNode(tag)
        for (i in seq_along(item)) {
            if (names(item)[i] != ".attrs") {
                xml <- XML::append.xmlNode(xml, listToXml(item[[i]], names(item)[i]))
            }
        }
    }
    
    # add attributes to node
    attrs <- item[[".attrs"]]
    for (name in names(attrs)) {
        XML::xmlAttrs(xml)[[name]] <- attrs[[name]]
    }
    return(xml)
}
#' Gather all PCR settings.
#'
#' Gathers all PCR settings (e.g. for XML output).
#'
#' @param annealing_temp Annealing temperature in Celsius.
#' @param Na_concentration Sodium ion concentration.
#' @param Mg_concentration Magensium ion concentration.
#' @param K_concentration Potassium ion concentration.
#' @param Tris_concentration Tris buffer concentration.
#' @param primer_concentration Primer concentration.
#' @param template_concentration Template concentration.
#' @return List with all PCR settings.
#' @keywords internal
get.PCR.settings <- function(use_taq_polymerase, annealing_temp, Na_concentration, Mg_concentration, 
    K_concentration, Tris_concentration, primer_concentration, template_concentration, nbr.cycles) {
        settings <- list(use_taq_polymerase = use_taq_polymerase, Na_concentration = Na_concentration, 
        Mg_concentration = Mg_concentration, K_concentration = K_concentration, Tris_concentration = Tris_concentration, 
        primer_concentration = primer_concentration, template_concentration = template_concentration, cycles = nbr.cycles)
    if (length(annealing_temp) != 0) {
        settings <- c("annealing_temp" = annealing_temp, settings)
    }
    return(settings)
}

#' Conversion of Constraints List to Data Frame.
#'
#' Converts the input constraints to a data frame representation.
#'
#' @param limit.constraints A list with constraints.
#' @param out.names The desired column names.
#' @param format.type The type of formatting to be performed on the table
#' @return A data frame giving an overview of the constraints.
#' @keywords internal
constraints.to.df <- function(limit.constraints, out.names, format.type = c("backend", "shiny", "report")) {
    if (length(out.names) != 1) {
        stop("out.names must have length 1")
    }
    format.type <- match.arg(format.type)
    cnames <- c("Constraint", "min", "max")
    df <- NULL
    for (i in seq_along(limit.constraints)) {
        c <- limit.constraints[[i]]
        con.name <- names(limit.constraints)[i]
        criteria <- names(c)
        idx <- sapply(criteria, function(x) which(x == cnames))
        idx.length <- unlist(lapply(idx, length))
        sel <- which(idx.length != 0)  # add sel to df
        n.sel <- which(idx.length == 0)
        nrows <- max(idx.length) + length(n.sel)
        d <- data.frame(matrix(rep(NA, nrows * length(cnames)), nrow = nrows))
        colnames(d) <- cnames
        rnames <- rep(NA, nrow(d))
        if (length(sel) != 0) {
            rnames[1] <- con.name
            dir <- criteria[sel]
            d[1, dir] <- c[sel]
        }
        if (length(n.sel) != 0) {
            origin.name <- criteria[n.sel]
            dir <- ifelse(grepl("min", origin.name), "min", "max")
            name <- sapply(seq_along(origin.name), function(x) gsub(paste(".", dir[x], 
                sep = ""), "", paste(con.name, ":", origin.name[x], sep = "")))
            rnames[n.sel] <- name
            d.s <- ifelse(length(sel) != 0, 2, 1)
            d.e <- d.s + length(n.sel) - 1
            for (j in seq_along(dir)) {
                d[(d.s:d.e)[j], dir[j]] <- c[n.sel][j]
            }
        }
        rownames(d) <- rnames
        df <- rbind(df, d)
    }
    if (length(df) != 0) {
        df$Constraint <- rownames(df)
        if (format.type == "shiny") {
            # annotate units with html units
            df$Constraint <- unlist(constraints_to_unit(df$Constraint, TRUE, "HTML"))
        } else if (format.type == "report") {
            df$Constraint <- unlist(constraints_to_unit(df$Constraint, TRUE, "report"))
        } else {
            # don't annotate units
            #df$Constraint <- unlist(constraints_to_unit(df$Constraint, FALSE)) # keep the input fields
        }
        df[,2:3] <- apply(df[,2:3], 2, function(x) round(x,3))
        # introduce interval notation:
        if (format.type %in% c("shiny", "report")) {
            # change NA to infinity symbols
            if (format.type == "shiny") {
                s1 <- "-&infin;"
                s2 <- "&infin;"
            } else {
                s1 <- "$-\\infty$"
                s2 <- "$\\infty$"
            }
            df[,2][is.na(df[,2])] <- s1
            df[,3][is.na(df[,3])] <- s2
        }
        df[,2] <- paste0("[", df[,2], ", ", df[,3], "]")
        df <- df[, -3] # remove the third column
        # modify column name of 2nd column
        colnames(df)[2] <- out.names
    }
    rownames(df) <- NULL
    return(df)
}

#' Renaming of Constraint Options.
#'
#' Renames the input list with constraint options.
#'
#' @param constraint.options A list with constraint options.
#' @return A list with renamed constraint options.
#' @keywords internal
rename.constraint.options <- function(constraint.options) {
    new.names <- Hmisc::capitalize(gsub("_", " ", names(constraint.options)))
    names(constraint.options) <- new.names
    return(constraint.options)
}
#' Output a Constraint Overview Table
#'
#' Outputs a table showing the values of constraints.
#'
#' @param constraints List with constraint settings.
#' @param constraint.limits List with constraint limits.
#' @param constraints.used.fw Constraints used for forward primer design.
#' @param constraints.used.rev Constraints used for reverse primer design.
#' @param format.type The type of formatting to be performed on entries.
#' @return Data frame with summary of constraints.
#' @keywords internal
create.constraint.table <- function(constraints, constraint.limits = NULL, constraints.used.fw = NULL, 
                                    constraints.used.rev = NULL, format.type = c("backend", "shiny", "report")) {
    if (length(constraints) == 0) {
        return(NULL)
    }
    format.type <- match.arg(format.type)
    input.constraints <- constraints
    limit.constraints <- constraint.limits
    output.constraints.fw <- constraints.used.fw
    output.constraints.rev <- constraints.used.rev
    if (length(input.constraints) != 0) {
        input.constraints <- constraints.to.df(input.constraints, "Target range", format.type)
    }
    if (length(output.constraints.fw) != 0) {
        out.names <- "Used range (fw)"
        # reorder
        m <- match(names(constraints), names(output.constraints.fw))
        output.constraints.fw <- constraints.to.df(output.constraints.fw[m[!is.na(m)]], out.names, format.type)
    }
    if (length(output.constraints.rev) != 0) {
        out.names <- "Used range (rev)"
        # reorder
        m <- match(names(constraints), names(output.constraints.rev))
        output.constraints.rev <- constraints.to.df(output.constraints.rev[m[!is.na(m)]], out.names, format.type)
    }
    if (length(input.constraints) != 0 && length(limit.constraints) != 0) {
        # only show relevant limits/reorder
        m <- match(names(constraints), names(constraint.limits))
        limit.constraints <- limit.constraints[m[!is.na(m)]]
        out.names <- "Limit range"
        limit.constraints <- constraints.to.df(limit.constraints, out.names, format.type)
    }
    result <- input.constraints
    if (length(limit.constraints) != 0) {
        result <- cbind(result, limit.constraints[,which(colnames(limit.constraints) != "Constraint"), drop = FALSE])
    }
    if (length(output.constraints.fw) != 0) {
        result <- cbind(result, output.constraints.fw[, which(colnames(output.constraints.fw) != "Constraint"), drop = FALSE])
    }
    if (length(output.constraints.rev) != 0) {
        result <- cbind(result, output.constraints.rev[, which(colnames(output.constraints.rev) != "Constraint"), drop = FALSE])
    }
    if (length(result) != 0) {
        # reorder: put Constraint name first
        o <- which(colnames(result) == "Constraint")
        o <- c(o, which(colnames(result) != "Constraint"))
        result <- result[,o]
    }
    rownames(result) <- NULL
    return(result)
}

#' Creation of a Table for Other Constraint Settings.
#'
#' @param other.settings List with other constraint settings.
#' @param format.type How the table shall be formatted.
#' @return A data frame.
#' @keywords internal
create.other.table <- function(other.settings, col.names, format.type) {
    option.names <- names(other.settings)
    if (format.type == "backend") {
        # nothing to format
    } else if (format.type == "shiny") {
        option.names <- constraints_to_unit(option.names, TRUE, "HTML")
    } else { # report format
        option.names <- constraints_to_unit(option.names, TRUE, format.type)
    }
    option.names <- unname(unlist(option.names))
    df <- data.frame(option.names, unlist(as.character(other.settings)))
    if (length(col.names) != ncol(df)) {
        stop("Dimensions of col.names and df do not agree.")
    }
    colnames(df) <- col.names
    return(df)
}
#' Creation of a Table for Constraint Options.
#'
#' @param other.settings List with constraint options
#' @param format.type How the table shall be formatted.
#' @return A data frame.
#' @keywords internal
create.options.table <- function(other.settings, 
                        format.type = c("backend", "shiny", "report")) {
    format.type <- match.arg(format.type)
    df <- create.other.table(other.settings, c("Option", "Setting"), format.type)
    rownames(df) <- NULL
    return(df)
}
#' Creation of a Table for PCR Conditions
#'
#' @param other.settings List with PCR settings.
#' @param format.tyep How the table shall be formatted.
#' @return A data frame.
#' @keywords internal
create.PCR.table <- function(other.settings, 
                    format.type = c("backend", "shiny", "report")) {
    format.type <- match.arg(format.type)
    # hide Tris setting if it is 0
    if ("Tris_concentration" %in% names(other.settings)) {
        if (other.settings$Tris_concentration == 0) {
            other.settings <- other.settings[which(names(other.settings) != "Tris_concentration")]
        }
    }
    df <- create.other.table(other.settings, c("Condition", "Setting"), format.type)

    rownames(df) <- NULL
    return(df)
}
#' Constraint list comparison
#'
#' Determines whether two list with constraints are identical.
#'
#' @param A First constraint list.
#' @param B Second constraint list.
#'
#' @return TRUE if the constraints are identical, FALSE else.
#' @keywords internal
compare.constraints <- function(A, B) {
    if (length(A) != length(B)) {
        return(FALSE)
    }
    m <- match(names(A), names(B))
    if (any(is.na(m))) {
        # names don't match
        return(FALSE)
    }
    # need to check further because the names of all lists agree
    for (i in seq_along(A)) {
        if (any(A[[i]] != B[[m[i]]])) {
            return(FALSE)
        }
    }
    return(TRUE)
}
#' Retrieval of Tool Information.
#'
#' Constructs a data frame containing information about the tools.
#'
#' @return A data frame with information about the required tools.
#' @keywords internal
get.static.tool.info <- function() {
    tools <- c("MELTING", "ViennaRNA", "OligoArrayAux", "MAFFT",
              "Selenium", "Pandoc", "PhantomJS")
    blank <- rep(NA, length(tools))
    names(blank) <- tools
    purposes <- blank
    purposes["MELTING"] <- "Melting temperatures"
    purposes["ViennaRNA"] <- "Secondary structures"
    purposes["OligoArrayAux"] <- "Primer efficiencies and dimerization"
    purposes["MAFFT"] <- "Multiple sequence alignments"
    purposes["Selenium"] <- "IMGT queries"
    purposes["Pandoc"] <- "PDF reports"
    purposes["PhantomJS"] <- "IMGT queries"
    tool.ex <- blank
    tool.ex["MELTING"] <- "melting-batch"
    tool.ex["ViennaRNA"] <- "RNAfold"
    tool.ex["OligoArrayAux"] <- "hybrid-min"
    tool.ex["MAFFT"] <- "mafft"
    tool.ex["Selenium"] <- "Python"
    tool.ex["Pandoc"] <- "Pandoc"
    tool.ex["PhantomJS"] <- "phantomjs"
    locations <- sapply(tool.ex, Sys.which)
    tool.URLs <- blank
    tool.URLs["MELTING"] <- "http://www.ebi.ac.uk/biomodels/tools/melting/"
    tool.URLs["ViennaRNA"] <- "http://www.tbi.univie.ac.at/RNA/"
    tool.URLs["OligoArrayAux"] <- "http://unafold.rna.albany.edu/OligoArrayAux.php"
    tool.URLs["MAFFT"] <- "http://mafft.cbrc.jp/alignment/software/"
    tool.URLs["Selenium"] <- "http://selenium-python.readthedocs.io/"
    tool.URLs["Pandoc"] <- "http://pandoc.org"
    tool.URLs["PhantomJS"] <- "http://phantomjs.org/"
    result <- data.frame("Tool" = names(purposes), "Purpose" = purposes, "Executable" = tool.ex, "URL" = tool.URLs)
    return(result)
}
#' Creation of an Overview of Third-Party Tools.
#'
#' Creates a table of required third-party tools and their installation status.
#'
#' @param AVAILBLE.TOOLS A vector whose names give the required tools and whose
#' entries give their installation status as logicals.
#' @param If \code{for.shiny} is \code{TRUE}, provide the URLs for the tool using HTML.
#' @return A data frame with information on third-part tools.
#' @keywords internal
build.tool.overview <- function(AVAILABLE.TOOLS, for.shiny = FALSE) {
    status <- ifelse(AVAILABLE.TOOLS, "Available", "Unavailable")
    tool.info <- get.static.tool.info()
    m <- match(tool.info$Tool, names(status))
    status <- status[m]
    tool.df <- data.frame(Tool = tool.info$Tool, Status = status, Purpose = tool.info$Purpose, 
                          Executable = tool.info$Executable, URL = tool.info$URL, stringsAsFactors = FALSE)
    if (for.shiny) {
        # add URLs to names
        tool.names <- paste0("<a href='", tool.info$URL, "' target='_blank'>", tool.info$URL, "</a>")
        tool.df$Tool <- tool.names
        # remove URL column
        tool.df <- tool.df[,colnames(tool.df) != "URL"]
    }
    return(tool.df)
}
#' Selection of Constraints.
#'
#' Selects constraints that can be computed according to installed third-party software.
#' This function is only used for initializing the 'constraint_order' option.
#'
#' @param active.constraints A vector whose names give the constraints to be checked.
#' @return A vector of useable constraint identifiers.
#' @keywords internal
select.constraints <- function(active.constraints) {
    tool.info <- build.tool.overview(check.tool.function(), for.shiny = FALSE)
    # check for each tool:
    melting.available <- tool.info[tool.info$Tool == "MELTING", "Status"] == "Available"
    vienna.available <- tool.info[tool.info$Tool == "ViennaRNA", "Status"] == "Available"
    oligo.available <- tool.info[tool.info$Tool == "OligoArrayAux", "Status"] == "Available"
    rm.constraints <- NULL
    #melting.constraints <- c("melting_temp_range", "melting_temp_diff")
    melting.constraints <- NULL # melting temp is now computed by empiric formula if melting is not present
    # this ensures we can also compute an annealing temperature (required for some constraints ..)
    vienna.constraints <- c("secondary_structure")
    oligo.constraints <- c("primer_efficiency", "self_dimerization", "cross_dimerization", "annealing_DeltaG", "coverage_model")
    if (!melting.available) {
        #message("Melting not available")
        rm.constraints <- c(rm.constraints, melting.constraints)
    }
    if (!vienna.available) {
        #message("ViennaRNA not available")
        rm.constraints <- c(rm.constraints, vienna.constraints)
    }
    if (!oligo.available) {
        #message("OligoArrayAux not available")
        rm.constraints <- c(rm.constraints, oligo.constraints)
    }
    ignore.constraints <- intersect(active.constraints, rm.constraints)
    new.constraints <- setdiff(active.constraints, ignore.constraints)
    return(new.constraints)
}

#' Quick Selection of Constraints.
#'
#' Select constraints that can be used according to third-party tools quickly.
#'
#' @param active.constraints Identifiers of constraints.
#' @return The identifiers of constraints that can be computed.
#' @keywords internal
con_select <- function(active.constraints) {
    out <- active.constraints[active.constraints %in% getOption("openPrimeR.constraint_order")]
    return(out)
}

tool.info <- get.static.tool.info()
#' Format Constraint Names.
#'
#' Formats constraint names for frontend output.
#'
#' @param constraints The character vector of constraints to transform.
#' @return A character vector with formatted constraint names.
#' @keywords internal
format.constraints <- function(constraints) {
    mapping <- list("primer_coverage" = "Coverage", 
        "primer_specificity" = "Specificity",
        "primer_length" = "Length",
        "gc_clamp" = "GC clamp",
        "gc_ratio" = "GC ratio",
        "no_runs" = "Runs",
        "no_repeats" = "Repeats", 
        "self_dimerization" = "Self dimers",
        "cross_dimerization" = "Cross dimers",
        "melting_temp_range" = "Tm range",
        "secondary_structure" = "Structures",
        "melting_temp_diff" = "Tm deviation",
        # coverage constraints
        "primer_efficiency" = "Efficiency",
        "annealing_DeltaG" = "Annealing",
        "stop_codon" = "Stop codons",
        "substitution" = "Substitutions",
        "terminal_mismatch_pos" = "3' Mismatch Position",
        "coverage_model" = "Coverage Model FPR",
        # PCR settings
        "use_taq_polymerase" = "Taq polymerase",
        "Na_concentration" = "[Na]",
        "Mg_concentration" = "[Mg]",
        "K_concentration" = "[K]",
        "Tris_concentration" = "[Tris]",
        "primer_concentration" = "[Primer]",
        "template_concentration" = "[Template]",
        "cycles" = "PCR cycles",
        "annealing_temp" = "Annealing temperature",
        # Other options
        "allowed_mismatches" = "Allowed mismatches",
        "allowed_other_binding_ratio" = "Allowed off-target binding ratio",
        "allowed_region_definition" = "Binding region definition")
    m <- match(constraints, names(mapping))
    idx <- which(is.na(m))
    out <- mapping[m]
    if (length(idx) != 0) {
        msg <- paste("Could not format the following constraints:",
                paste(constraints[idx], collapse = ","))
        warning(msg)
        # keep original name if not found
        out[idx] <- constraints[idx]
    }
    return(out)
}
#' Mapping of Constraints to Units.
#'
#' Maps constraints to units for plotting.
#'
#' @param constraint The names of the constraints to convert to their plot identifiers (units).
#' @param use.unit Whether constraint names should be annotated with their units.
#' @param use.HTML Whether constraint units should be annotated with HTML units.
#' @return A list of constraint names.
#' @keywords internal
constraints_to_unit <- function(constraint, use.unit = TRUE, 
                    format.type = c("backend", "HTML", "report")) {
    format.type <- match.arg(format.type)
    if (length(constraint) == 0) {
        # nothing to modify
        return(constraint)
    }
    mod.con <- format.constraints(constraint)
    unit.mapping.R <- list(
        melting_temp_range = expression(paste("T"[m], " [", degree * 
                                        C, "]", sep = "")),
        melting_temp_diff = expression(paste("T"[m], " deviation [", degree * 
                                        C, "]", sep = "")),
        self_dimerization = expression(paste("Self dimer ", Delta, " G [kcal/mol]", sep = "")),
        cross_dimerization = expression(paste("Cross dimer ", Delta, " G [kcal/mol]", sep = "")),
        secondary_structure = expression(paste("Structure ", Delta, " G [kcal/mol]", sep = "")),
        annealing_DeltaG = expression(paste("Annealing ", Delta, " G [kcal/mol]", sep = ""))
        )
    unit.mapping.report <- list(
        melting_temp_range = "T\\textsubscript{m} range [\\textdegree C]",
        melting_temp_diff = "T\\textsubscript{m}  deviation [\\textdegree C]",
        self_dimerization = "Self dimer $\\Delta$G[$\\frac{\\text{kcal}}{\\text{mol}}$]",
        cross_dimerization = "Cross dimer $\\Delta$G[$\\frac{\\text{kcal}}{\\text{mol}}$]",
        secondary_structure = "Structure $\\Delta$G[$\\frac{\\text{kcal}}{\\text{mol}}$]",
        annealing_DeltaG = "Annealing $\\Delta$G[$\\frac{\\text{kcal}}{\\text{mol}}$]",
        # PCR conditions
        Na_concentration = "[Na\\textsuperscript{+}] [M]",
        Mg_concentration = "[Mg\\textsuperscript{2+}] [M]",
        K_concentration = "[K\\textsuperscript{+}] [M]",
        Tris_concentration = "[Tris buffer] [M]",
        primer_concentration = "[Primer] [M]",
        template_concentration = "[Template] [M]",
        annealing_temp = "Annealing temperature [\\textdegree C]"
        )
    unit.mapping.html <- list(
        melting_temp_range = "T<sub>m</sub> range [&#x2103;]",
        melting_temp_diff = "T<sub>m</sub> deviation [&#x2103;]",
        self_dimerization = "Self dimer &Delta;G[<sup>kcal</sup>&frasl;<sub>mol</sub>]",
        cross_dimerization = "Cross dimer &Delta;G[<sup>kcal</sup>&frasl;<sub>mol</sub>]",
        secondary_structure = "Structure &Delta;G[<sup>kcal</sup>&frasl;<sub>mol</sub>]",
        annealing_DeltaG = "Annealing &Delta;G[<sup>kcal</sup>&frasl;<sub>mol</sub>]",
        # PCR conditions
        Na_concentration = "[Na<sup>+</sup>] [M]",
        Mg_concentration = "[Mg<sup>2+</sup> [M]",
        K_concentration = "[K<sup>+</sup>] [M]",
        Tris_concentration = "[Tris buffer] [M]",
        primer_concentration = "[Primer] [M]",
        template_concentration = "[Template] [M]",
        annealing_temp = "Annealing temperature [&#x2103;]"
        )
    if (use.unit) {
        if (format.type == "HTML") {
            mapping <- unit.mapping.html
        } else if (format.type == "report") {
            mapping <- unit.mapping.report
        } else {
            mapping <- unit.mapping.R
        }
        m <- match(constraint, names(mapping))
        idx <- which(!is.na(m))
        if (length(idx) != 0) {
            mod.con[idx] <- lapply(idx, function(x) mapping[[m[x]]])
        }
    }
    return(mod.con)
}

#' @rdname Settings
#' @name Settings
#' @aliases parallel_setup
#' @return \code{parallel_setup} returns \code{NULL}.
#' @export
#' @examples
#' # Use two cores for parallel processing:
#' parallel_setup(2)
parallel_setup <- function(cores = NULL) {
    doParallel.available <- requireNamespace("doParallel", quietly = TRUE)
    if (doParallel.available) { # no parallel support for windows at the moment (unserialize errors if we don't do clusterexport/clustercall)
        if (length(cores) == 0) {
            avail.cores <- parallel::detectCores()
            # use half the available cores at most
            cores <- max(1, floor(avail.cores / 2))
        } else {
            if (!is(cores, "numeric")) {
                stop("Please supply a numeric for 'cores'.")
            }
        }
        cores <- min(floor(cores), parallel::detectCores()) # use at most all of the available cores
        doParallel::registerDoParallel(cores = cores)
        # also set mc.cores for 'mclapply'
        options(mc.cores = cores)
        message("The number of cores for was set to '", cores, "' by 'parallel_setup()'.")
    } else {
        warning("Please install 'doParallel' to use multiple cores.")
    }
}

context("Input/Output")

test_that("Settings IO Consistency", {
    # load settings
    filename <- system.file("extdata", "settings", 
                  "C_Taq_PCR_high_stringency.xml", package = "openPrimeR")
    settings <- read_settings(filename)
	settings.file <- tempfile("test_settings", fileext = ".xml")
    write_settings(settings, settings.file)
    # verify that test_settings agree with settings:
    test.settings <- read_settings(settings.file)
    for (i in seq_along(settings)) {
        slot.name <- slotNames(settings)[i]
        setting.slots <- slot(settings, slot.name)
        test.slots <- slot(test.settings, slot.name)
        expect_that(length(setting.slots), equals(length(test.slots)))
        expect_that(names(setting.slots), equals(names(test.slots)))
        expect_that(setting.slots, equals(test.slots))
    }
}) 
 

context("Settings") 

test_that("Settings validity", {
    # check whether settings validity is detected correctly
    data(Ippolito)
    # check that settings can be changed without errors
    constraints(settings)[["gc_clamp"]] <- c("min" = 0, "max" = 5)
    constraints(settings) <- constraints(settings)[1:4]
    constraintLimits(settings) <- constraintLimits(settings)[1:4]
    PCR(settings)$Na_concentration <- 0.00001
    conOptions(settings)$allowed_mismatches <- 2
    # check required fields -> should produce errors
    expect_error({PCR(settings) <- PCR(settings)[-which(names(PCR(settings)) == "Na_concentration")]})
    expect_error({conOptions(settings) <- conOptions(settings)[-which(names(conOptions(settings)) == "allowed_mismatches")]})
    expect_error({constraints(settings)$gc_clamp <- c("min" = 64, "max" = 30)})
    expect_error({constraints(settings)$gc_clamp <- c("max" = 30, "min" = 40)})
    test <- NULL
})

test_that("Settings concordance", {
    # check whether settings adjust when constraints/constraintLimit changes
    data(Ippolito)
    # widen the constraints beyond the limits:
    constraints(settings)$gc_clamp[2] <- 20
    new.setting <- constraints(settings)$gc_clamp[2]
    expect_that(constraintLimits(settings)$gc_clamp[2], equals(new.setting))
    # narrow down the limits beyond the constraint
    new.setting <- c("min" = 0, "max" = 0)
    constraintLimits(settings)$no_runs <- new.setting
    expect_that(constraints(settings)$no_runs, equals(new.setting))
    # check whether removal keeps constraints/limits concordant
    constraints(settings) <- constraints(settings)[names(constraints(settings)) %in% c("gc_clamp", "no_runs")]
    expect_that(names(constraints(settings)), equals(names(constraintLimits(settings))))
    # the same for addition of constraints
    constraintLimits(settings)$primer_length <- c("min" = 18, "max" = 30)
    expect_that(constraints(settings)$primer_length, equals(constraintLimits(settings)$primer_length))
})


test_that("Settings IO", {
    # check whether settings are stored/loaded correctly
    settings.xml <- system.file("extdata", "settings", 
                    "C_Taq_PCR_high_stringency.xml", package = "openPrimeR")
    settings <- read_settings(settings.xml)
    constraints(settings)$melting_temp_diff <- c("max" = 3)
    out.xml <- tempfile("my_settings", fileext = ".xml")
    write_settings(settings, out.xml)
    loaded.settings <- try(read_settings(out.xml), silent = TRUE)
    expect_that(class(loaded.settings), equals(class(settings)))
    expect_that(length(slotNames(loaded.settings)), equals(length(slotNames(settings))))
    for (i in seq_along(slotNames(loaded.settings))) {
        loaded.name <- slotNames(loaded.settings)[i]
        old.name <- slotNames(settings)[i]
        # ensure that slot names are identical
        expect_that(loaded.name, equals(old.name))
        loaded <- constraints(slot(loaded.settings, loaded.name))
        old <- constraints(slot(settings, loaded.name))
        # check that list entries are identical
        for (j in seq_along(loaded)) {
            con.name <- names(loaded)[j]
            loaded.val <- loaded[[j]]
            old.val <- old[[con.name]]
            expect_that(loaded.val, equals(old.val))
        }
    }
})

######
# function for global primer design. TODO: improve & export for use in package. (need a pairwise function for this still) -> generalized primer design function should accept a function as the argument ...
######
design_primers_global_single <- function(template.df, settings, mode.directionality = c("fw", "rev"), 
                                         sample.name = NULL, target.temps = NULL, init.algo = c("naive", "tree"),
                                         max.degen = 16, conservation = 1, cur.results.loc = NULL, prefilter = FALSE,
                                         max.penalties = c(0, 0.25, 0.5, 0.75, 1, 1.5, 2, 3, 5, Inf)) {
    # load example primers:
    if (length(sample.name) == 0) {
        sample.name <- "TEST"
    }
    cur.results.loc <- NULL
    primer.df <- initialize.primer.set(template.df, sample.name, constraints(settings)$primer_length,  # N = 10,707
                conOptions(settings)$allowed_region_definition, mode.directionality, init.algo, max.degen, conservation, cur.results.loc)
    message("Number of primers: ", nrow(primer.df))
    # don't sub-sample the primer set
    #my.sel <- sample(seq_len(nrow(primer.df)), 50)
    #primer.df <- primer.df[my.sel, ]
    # 0) Pre-filter? Necessary to reduce possible number of primers sometimes (> 10, 000 makes problems)
    if (prefilter) {
        filter.settings <- settings
        print(class(filter.settings))
        filter.constraints <- list("melting_temp_range" = c("min" = 50, "max" = Inf))
        print(filter.constraints)
        #filter.constraints <- list("gc_ratio" = c("min" = 0.3, "max" = 0.7), # ~10%
                                #"melting_temp_range" = c("min" = 50, "max" = 67), # ~3%
                                #"no_runs" = c("min" = 0, "max" = 5), # ~0%
                                #"no_repeats" = c("min" = 0, "max" = 5)) # ~0%
        constraints(filter.settings) <- filter.constraints
        primer.df <- filter_primers(primer.df, template.df, filter.settings)
        message("Number of primers after filtering: ", nrow(primer.df)) # N = 9,930 > ok? -> deal with cross dimer computation? .. could use greedy? or filter more
    }
    # A) Determine all constraints for all primers
    # select constraints that need to be computed:
    exclude.constraints.check <- c("primer_length", "gc_ratio", 
                             "no_runs", "no_repeats",
                             "melting_temp_diff",
                             "cross_dimerization")
    compute.constraints <- setdiff(names(constraints(settings)), exclude.constraints.check)
    primer.df <- check_constraints(primer.df, template.df, settings, 
                                   compute.constraints)
    message("Done with evaluating the primers!")
    # B) Score all primers
    # exclude 'exclude.constraints' from scoring
    exclude.constraints <- c("primer_length", "gc_ratio", 
                             "no_runs", "no_repeats",
                             "primer_coverage", "melting_temp_diff",
                             "cross_dimerization")
    score.constraints <- setdiff(names(constraints(settings)), exclude.constraints)
    penalty.df <- score_primers(primer.df, settings, active.constraints = score.constraints)
    penalties <- penalty.df$Penalty[match(primer.df$ID, penalty.df$ID)]
    primer.df$Penalty <- penalties
    #if (length(cur.results.loc != 0)) {
        #png(file.path(cur.results.loc, paste0("design_penalties_", sample.name, ".png")))
        #hist(penalties)
        #dev.off()
        #save(primer.df, file = file.path(cur.results.loc, paste0("eval_primers_", sample.name, ".Rdata")))
    #}
    #load(file.path("data", "global_primers.Rdata"))
    # C) Optimize (NB: consider melting_temp_diff, cross_dimerization)!
    # build dimerization matrix
    Tm.brackets <- create.Tm.brackets(primer.df, template.df, settings, target.temps)
    primer_conc <- PCR(settings)$primer_concentration
    na_salt_conc <- PCR(settings)$Na_concentration
    mg_salt_conc <- PCR(settings)$Mg_concentration
    k_salt_conc <- PCR(settings)$K_concentration
    tris_salt_conc <- PCR(settings)$Tris_concentration
    annealing.temp <- min(Tm.brackets$df$annealing_temp) # conservative dimerization calls
    message("Computing cross dimers ...")
    G.df <- compute.all.cross.dimers(primer.df, primer_conc, na_salt_conc, 
                            mg_salt_conc, k_salt_conc, tris_salt_conc, annealing.temp,
                            no.structures = TRUE)
    if (length(cur.results.loc) != 0) {
        save(G.df, file = file.path("data", paste0("dimer_energies_", sample.name, ".Rdata")))
    }
    #load(file.path("data", "global_dimer_energy.Rdata"))
    G.matrix <- create.G.matrix(primer.df, G.df) 
    message("Cross dimers computed!")
    cvg.matrix <- get.coverage.matrix(primer.df, template.df) # TODO: this could be a bit faster ..
    max.temp.diff <- constraints(settings)$melting_temp_diff["max"]
    dimer.cutoff <- constraints(settings)$cross_dimerization["min"]
    penalty.sets <- vector("list", length(max.penalties))
    message("Optimizing ...")
    for (i in seq_along(max.penalties)) {
        print(i)
        # explore several maximal penalties:
        max.penalty <- max.penalties[i]
        if (is.infinite(max.penalty)) {
            cur.max.temp.diff <- 50
        } else {
            cur.max.temp.diff <- max.temp.diff + (max.penalty * max.temp.diff)
        }
        cur.settings <- settings
        constraints(cur.settings)$melting_temp_diff <- c("max" = unname(cur.max.temp.diff))
        cur.temp.brackets <- create.Tm.brackets(primer.df, template.df, cur.settings, target.temps)
        cur.dimer.cutoff <- dimer.cutoff + (max.penalty * dimer.cutoff)
        D <- compute.dimer.matrix(G.matrix, cur.dimer.cutoff)
        temperature.sets <- vector("list", length(nrow(cur.temp.brackets))) 
        # define a temperature range
        for (j in seq_along(cur.temp.brackets$df$annealing_temp)) {
            # select only primers with penalties <= max.penalty and within melting temperature range
            target.Ta <- cur.temp.brackets$df$annealing_temp[j]
            run.id <- paste0("pen=", round(max.penalty, 2), "_Ta=", round(target.Ta,0))
            min.Tm <- cur.temp.brackets$df$min_Tm[j]
            max.Tm <- cur.temp.brackets$df$max_Tm[j]
            sel <- which(primer.df$Penalty <= max.penalty & primer.df$melting_temp <= max.Tm & primer.df$melting_temp >= min.Tm)
            if (length(sel) == 0) {
                # No primers available fulfilling the conditions
                next
            }
            cur.primers <- primer.df[sel,] 
            cur.D <- D[sel,sel, drop = FALSE]
            cur.cvg.matrix <- cvg.matrix[, sel, drop = FALSE]
            ILP <- ILPConstrained(cur.D, cur.cvg.matrix)
            #write.lp(ILP, "test_ILP.txt")
            return.val <- solve(ILP)
            if (return.val != 0) {
                # Optimization didn't work
                warning(return.val)
            }
            vars <- lpSolveAPI::get.variables(ILP)
            primer.vars <- vars[seq_len(nrow(cur.primers))]
            sel.idx <- which(primer.vars == 1)
            opt.primers <- cur.primers[sel.idx, ]
            # set Run identifier
            if (length(opt.primers) != 0 && nrow(opt.primers) != 0) {
                opt.primers$Run <- run.id
            }
            print(as.numeric(get_cvg_ratio(opt.primers, template.df)))
            temperature.sets[[j]] <- opt.primers
        }
        penalty.sets[[i]] <- temperature.sets
    }
    message("Storing optimization results")
    primer.data <- unlist(penalty.sets, recursive = FALSE)
    # select non-empty sets only:
    sel <- which(unlist(lapply(primer.data, function(x) length(x) != 0 && nrow(x) != 0)))
    primer.data <- primer.data[sel]
    set.names <- unlist(lapply(primer.data, function(x) x$Run[1]))
    names(primer.data) <- set.names
    # enrich sets with missing constraint info:
    for (i in seq_along(primer.data)) {
        print(i)
        primer.data[[i]] <- check_constraints(primer.data[[i]], template.df, settings, exclude.constraints.check)
    }
    return(primer.data)
}
penalize_primer <- function(deviations, alpha = 0.5) {
    # alpha: tuning parameter for penalizing maximum deviation
    if (!is.numeric(alpha) || alpha <0 || alpha > 1) {
        stop("Please provide a numeric in the interval [0,1] for 'alpha'.")
    }
    abs.dev <- abs(deviations)
    max.idx <- which.max(abs.dev)
    # minimal penalty is 0. 
    score <- (alpha * abs.dev[max.idx]) + (1 - alpha) * sum(abs.dev)
    return(score)
}
#' @rdname Scoring
#' @name Scoring
#' @details
#' \code{score_primers} determines the penalty of a primer in the following way.
#' Let \code{d} be a vector indicating the absolute deviations from 
#' individual constraints and let \code{p} be the scalar penalty that
#' is assigned to a primer. We define
#' \deqn{p = \alpha \cdot \max_i d_i + \sum_i (1 - \alpha) \cdot d_i}
#' such that for large values of \code{alpha} the maximal deviation 
#' dominates giving rise to a local penalty (reflecting the largest
#' absolute deviation) and for small \code{alpha} the total deviation
#' dominates giving rise to a global penalty 
#' (reflecting the sum of constraint deviations). 
#' When \code{alpha} is 1 only the most extreme absolute deviation is
#' considered and when \code{alpha} is 0 the sum of all absolute 
#' deviations is computed.
#'
#' @return \code{score_primers} returns a data frame containing
#' scores for individual primers.
#' @export
#' @examples
#'
#' # Score the primers
#' data(Ippolito)
#' primer.scores <- score_primers(primer.df, settings)
score_primers <- function(primer.df, settings, active.constraints = names(constraints(settings)), alpha = 0.5) {
    if (!is(primer.df, "Primers")) {
        stop("Please input a 'Primers' object as 'primer.df'.")
    }
    my_penalty <- function(deviation) {
        # local function where 'alpha' doesn't need to be supplied for ddply
        penalize_primer(deviation, alpha = alpha)
    }
    constraint.settings <- constraints(settings)[active.constraints]
    deviations <- get_constraint_deviation_data(primer.df, constraint.settings)
    penalty.df <- ddply(deviations, c("ID"), 
                              here(summarize), Penalty = my_penalty(substitute(Deviation)), Deviation = sum(abs(substitute(Deviation))))
    return(penalty.df)
}

#' @rdname Data
#' @name Data
#'
#' @usage data(Tiller)
#'
#' @aliases Tiller tiller.template.df tiller.primer.df tiller.settings
#'
#' @references Tiller, Thomas, et al. "Efficient generation of monoclonal antibodies from single human B cells by single cell RT-PCR and expression vector cloning." Journal of immunological methods 329.1 (2008): 112-124.
#' @examples
#'
#' # Load the data from Tiller et al.
#' data(Tiller)
#' tiller.primer.df
#' tiller.template.df
#' constraints(tiller.settings)
NULL

#' Primer Evaluation.
#' 
#' @rdname PrimerEval
#' @name PrimerEval
#'
#' @description
#' \describe{
#' \item{\code{check_constraints}}{Determines whether a set of primers
#' fulfills the constraints on the properties of the primers.}
#' \item{\code{check_restriction_sites}}{Checks a set of primers 
#' for the presence of restriction sites. 
#' To reduce the number of possible restriction sites,
#' only unambiguous restriction sites are taken into account and 
#' only common (typically used) restriction sites are checked if a common
#' restriction site can be found in a sequence.}
#' \item{\code{filter_primers}}{Filters a primer set according to 
#' the specified constraints such that all primers
#' that do not fulfill the constraints are removed from the primer set.}
#' \item{\code{primer_significance}}{Uses Fisher's exact test to determine 
#' the significance of a primer set according to its ratio of fulfilled 
#' constraints.}
#' \item{\code{subset_primer_set}}{Determines subsets of the input primer set
#' that are optimal with regard to the number of covered template sequences.}
#' }
#'
#' @param primer.df A \code{Primers} object containing the primers
#' whose properties are to be checked.
#' @param template.df A \code{Templates} object containing the 
#' template sequences corresponding to \code{primer.df}.
#' @param settings A \code{DesignSettings} object containing the 
#' constraints that are to be considered.
#' @param active.constraints A subset of the constraint identifiers 
#' provided by \code{settings} that are to be checked
#' for fulfillment. By default \code{active.constraints} is \code{NULL} such that
#' all constraints found in \code{settings} are evaluated. Otherwise,
#' only the constraints specified via \code{active.constraints} 
#' that are available in \code{settings} are considered.
#' @param to.compute.constraints Constraints that are to be computed.
#' By default, \code{to.compute.constraints} is set to \code{NULL} such that
#' all \code{active.constraints} are computed. If \code{to.compute.constraints}
#' is a subset of \code{active.constraints}, all constraints specified
#' via \code{active.constraints} are evaluated for fulfillment,
#' but only the constraints in \code{to.compute.constraints} are newly calculated.
#' @param for.shiny Whether the output of the function shall be
#' formatted as HTML. The default setting is \code{FALSE}.
#' @param updateProgress Progress callback function for shiny. The defaut is
#' \code{NULL} meaning that no progress is monitored via the Shiny interface.
#' @return \code{check_constraints} returns a \code{Primers} object 
#' that is augmented with columns providing the results for the evaluated 
#' constraints.
#' The \code{constraints_passed} column indicates whether all 
#' \code{active.constraints} were fulfilled.
#' The \code{EVAL_*} columns indicate the fulfillment of primer-specific constraints.
#' The \code{T_EVAL_*} columns indicate the fulfillment of template-specific
#' (e.g. coverage-based) constraints.
#' For the coverage computations, columns prefixed by \code{Basic_},
#' indicate the results from string matching, while all other results
#' (e.g. \code{primer_coverage}) indicate the expected coverage
#' after applying the coverage constraints specified in \code{settings}.
#' Columns prefixed by \code{Off_} indicate off-target binding results.
#' @param adapter.action The action to be performed when adapter sequences
#' are found. Either "warn" to issue a warning about adapter sequences or
#' "rm" to remove identified adapter sequences. Currently, only
#' the default setting ("warn") is supported.
#' @param selected Names of restriction sites that are to be checked.
#' By default \code{selected} is \code{NULL} in which case all REBASE 
#' restriction sites are taken into account.
#' @param only.confident.calls Whether only confident calls
#' of restriction sites are returned.
#' All restriction site call is considered \emph{confident} if the restriction site
#' is located in a region that does not match the template sequences.
#' Note that this classification requires that the provided primers
#' are somehow complementary to the provided templates.
#' In contrast, non-confident restriction site calls are 
#' based solely on the primer sequences and do not take the templates
#' into account, resulting in more false positive calls of restriction sites.
#' @param set.name An identifier for the input primers. If \code{NULL},
#' the run identifier is used.
#' @param k The spacing between generated primer subset sizes. By default,
#' \code{k} is set to 1 such that all primer subsets are constructed.
#' @param groups The identifiers of template groups according to which 
#' coverage should be determined. By default, \code{groups} is set to 
#' \code{NULL} such that all all covered templates are considered.
#' @param identifier An identifier for storing the primer set. By default,
#' \code{identifier} is set to  \code{NULL}.
#' @param cur.results.loc Directory for storing the results. By default,
#' \code{cur.results.loc} is set to \code{NULL}, which means that
#' no results are stored.
#' 
#' @note Please note that some constraint computations 
#' may require the installation of additional programs; for more information
#' please view the documentation of \code{\link{DesignSettings}}.
NULL

#' @rdname Plots
#' @name Plots
#' @aliases plot_cvg_vs_set_size
#' @return \code{plot_cvg_vs_set_size} returns a plot of coverage vs set size.
#' @export
#' @examples
#' 
#' # Plot coverage vs primer set size
#' data(Comparison)
#' p <- plot_cvg_vs_set_size(primer.data, template.data)
plot_cvg_vs_set_size <- function(primer.data, template.data, show.labels = TRUE, highlight.set = NULL) {
    if (length(primer.data) == 0 || length(template.data) == 0) {
        return(NULL)
    }
    # check type of primer and template data
    template.classes <- sapply(template.data, function(x) class(x))
    primer.classes <- sapply(primer.data, function(x) class(x))
    if (any(template.classes != "Templates") || any(primer.classes != "Primers")) {
        stop("Check types of primers/templates.")
    }
    cvg <- unlist(lapply(seq_along(primer.data), function(x) get_cvg_ratio(primer.data[[x]], 
                                                      template.data[[x]])))
    set.size <- unlist(lapply(primer.data, function(primer.df) nrow(primer.df)))
    set.names <- get.run.names(primer.data)
    plot.df <- data.frame("Run" = set.names, "Coverage" = cvg, "Set_Size" = set.size)
    title <- "Coverage vs set size"
     if (length(highlight.set) != 0) {
        # check whether highlight set is specified correctly
        m <- match(highlight.set, plot.df$Run)
        na.idx <- which(is.na(m))
        if (length(na.idx) != 0) {
            msg <- paste("Highlight set not available in data:",
                paste(highlight.set[na.idx], collapse = ","))
            warning(msg)
            highlight.set <- highlight.set[!is.na(m)]
        }
    }
    pal <- getOption("openPrimeR.plot_colors")["Run"] # the RColorBrewer palette to use
    colors <- colorRampPalette(brewer.pal(8, pal))(length(unique(plot.df[, "Run"])))
    # determine rate of constraint fulfillment:
    fulfilled.rate <- unlist(lapply(primer.data, function(x) {
        fulfilled.counts <- create_fulfilled_counts(x)
        if (nrow(fulfilled.counts) != 0) {
            fulfilled.counts <- fulfilled.counts[!grepl("Run", colnames(fulfilled.counts))]
            ok.names <- which(!grepl("_failure", colnames(fulfilled.counts)))
            rate <- sum(fulfilled.counts[ok.names]) / (nrow(x) * length(ok.names))
        } else {
            # empty primer set
            rate <- 0
        }
    }))
    #print(fulfilled.rate)
    plot.df$Constraint_Fulfillment_Rate <- fulfilled.rate
    p <- ggplot(plot.df, aes(x = .data[["Set_Size"]], y = .data[["Coverage"]], size = .data[["Constraint_Fulfillment_Rate"]]), key = substitute(Run)) + 
    # trans = scales::probability_trans("exp")
        scale_radius(limits = c(0,1), range = c(1,10),
            labels = scales::percent, 
            name = "Constraint Fulfillment") + 
        theme(axis.text.x = element_text(angle = 90, hjust = 1, vjust = 0.5)) +
        scale_y_continuous(labels = scales::percent, limits = c(0, 1)) +
        xlab("Number of primers") +
        ggtitle(title) +
        # only show integer values
        scale_x_continuous(breaks = function(x) unique(as.integer(pretty(x)))) +
        scale_colour_manual(values = colors)

    # decide whether to add some jitter
    overplotted <- sapply(seq_len(nrow(plot.df)), function(x) any(plot.df$Set_Size[x] - plot.df$Set_Size[-x] == 0 & abs(plot.df$Coverage[x] - plot.df$Coverage[-x]) < 0.2))
    if (any(overplotted)) {
        # there's overplotting of points -> add some jitter
        p <- p + geom_point(aes(fill = .data[["Run"]], color = NULL), alpha = 0.65, pch=21, position = position_jitter(width = 0.35, height = 0.0))
    } else {
        p <- p + geom_point(aes(fill = .data[["Run"]], color = NULL), alpha = 0.65, pch=21)
    }
    if (show.labels) {
        p <- p + geom_text(aes(label = .data[["Run"]],
                  hjust = 0, vjust = 0), check_overlap = TRUE, size = 3) 
    }
    if (length(highlight.set) != 0) {
        p <- p + geom_point(data = plot.df[plot.df$Run %in% highlight.set,], colour="black", shape=1, size=5, stroke = 3, alpha = 0.6)  
   } 
    return(p)
}

#' @rdname Plots
#' @name Plots
#' @aliases plot_penalty_vs_set_size
#' @return \code{plot_penalty_vs_set_size} returns a plot of constraint penalties
#' vs primer set sizes.
#' @export
#' @examples
#' 
#' # Plot penalties vs number of primers
#' data(Comparison)
#' p <- plot_penalty_vs_set_size(primer.data, settings)
plot_penalty_vs_set_size <- function(primer.data, settings, 
        active.constraints = names(constraints(settings)), alpha = 0) {
    if (length(primer.data) == 0) {
        return(NULL)
    }
    # check types
    primer.classes <- sapply(primer.data, function(x) class(x))
    if (any(primer.classes != "Primers")) {
        stop("Ensure that 'primer.data' is a list of 'Primers' objects.")
    }
    set.size <- unlist(lapply(primer.data, function(primer.df) rep(nrow(primer.df), nrow(primer.df))))
    penalty.data <- lapply(primer.data, function(primer.df) score_primers(primer.df, settings, active.constraints = active.constraints, alpha = alpha))
    penalties <- unlist(lapply(penalty.data, function(x) x$Penalty))
    run.names <- get.run.names(primer.data)
    run.names <- unlist(lapply(seq_along(run.names), function(x) rep(run.names[x], nrow(primer.data[[x]]))))
    plot.df <- data.frame("Run" = run.names, "Penalty" = penalties, "Set_Size" = set.size)
    title <- "Penalty vs set size"
    pal <- getOption("openPrimeR.plot_colors")["Run"] # the RColorBrewer palette to use
    colors <- colorRampPalette(brewer.pal(8, pal))(length(unique(plot.df[, "Run"])))
    # determine rate of constraint fulfillment:
    p <- ggplot(plot.df, aes(x = .data[["Set_Size"]], y = .data[["Penalty"]], fill = .data[["Run"]], group = .data[["Run"]])) + 
        geom_boxplot() + 
        theme(axis.text.x = element_text(angle = 90, hjust = 1, vjust = 0.5)) +
        xlab("Number of primers") +
        ggtitle(title) +
        # only show integer values
        scale_x_continuous(breaks = function(x) unique(as.integer(pretty(x)))) +
        scale_fill_manual(values = colors)
   return(p)
}

#' @export
cbind.fill <- function(...){
    nm <- list(...) 
    nm <- lapply(nm, as.matrix)
    n <- max(sapply(nm, nrow)) 
    do.call(cbind, lapply(nm, function (x) 
        rbind(x, matrix(, n-nrow(x), ncol(x))))) 
}

#' Primer Binding Region Data
#'
#' Collects all data concerning primer binding regions.
#'
#' @param primer.df Primer data frame.
#' @param template.df Template data frame.
#' @param direction Primer direction
#' @param group The groups for which binding data shall be retrieved.
#' @param relation Binding region data relative to forward/reverse binding region?
#' @return Data frame with primer binding data.
#' @keywords internal
primer.binding.regions.data <- function(primer.df, template.df, 
    direction = c("both", "fw", "rev"), group = NULL, relation = c("fw", "rev")) {
    if (length(relation) == 0) {
        stop("Please provide the 'relation' arg.")
    }
    relation <- match.arg(relation)
    if (length(direction) == 0) {
        stop("Please provide the 'direction' arg.")
    }
    direction <- match.arg(direction)
    if (length(primer.df) == 0 || nrow(primer.df) == 0) {
        return(NULL)
    }
    if (!"primer_coverage" %in% colnames(primer.df)) {
        stop(primer.df$Run[1], ": Please compute the primer coverage first.")
    }
    if (!is.null(group) && !"all" %in% group) { # select template subset
        idx <- which(template.df$Group %in% group)
        template.df <- template.df[idx,]
    }
    use.location.cols <- NULL
    if (relation == "fw") {
        use.location.cols <- c("Relative_Forward_Binding_Position_Start_fw", "Relative_Forward_Binding_Position_End_fw", 
            "Relative_Forward_Binding_Position_Start_rev", "Relative_Forward_Binding_Position_End_rev")
    } else {
        use.location.cols <- c("Relative_Reverse_Binding_Position_Start_fw", "Relative_Reverse_Binding_Position_End_fw", 
            "Relative_Reverse_Binding_Position_Start_rev", "Relative_Reverse_Binding_Position_End_rev")
    }
    # columns for 'start of binding':
    # 3 prime of primer: start
    # beginning of the target region'
    if (direction == "fw") {
        location.cols <- use.location.cols[c(1,2)]
    } else if (direction == "rev") {
        location.cols <- use.location.cols[c(3,4)]
    } else {
        location.cols <- use.location.cols
    }
    # select primer data belonging to the right group
    idx <- covered.seqs.to.idx(primer.df$Covered_Seqs, template.df)
    dfs <- do.call(rbind, lapply(seq_along(idx), function(x) if (length(idx[[x]]) == 0) 
        NULL else 
            data.frame(ID = primer.df$ID[x], Group = template.df[idx[[x]], "Group"], 
        StringIdx = seq_along(idx[[x]]))))  # id of primer + gene group
    if (any(!location.cols %in% colnames(primer.df))) {
        warning("Primer set doesn't have primers of indicated direction for plotting.")
        return(NULL)
    }
    # remove NA group entries
    if (length(dfs) != 0) {
        dfs <- dfs[!is.na(dfs$Group), ]
    }
    locations <- apply(asS3(primer.df)[, location.cols, drop = FALSE], 2, function(x) strsplit(as.character(x), 
        split = ","))
    all.IDs <- primer.df$ID
    if (length(group) != 0 && !"all" %in% group) {
        all.IDs <- dfs$ID
        m <- match(all.IDs, primer.df$ID)
        ## identify binding positions for the used location cols (fw and rev binding)
        locations <- lapply(seq_along(locations), function(y) lapply(seq_along(m),
                        function(x) {
                            pos <- locations[[y]][[m[x]]]
                            if (length(pos) != 0) {
                                pos[dfs$StringIdx[x]]
                            } else {
                                NULL
                            }
                        }
                    ))
    }
    my.locations <- lapply(locations, function(x) as.numeric(unlist(x)))
    my.IDs <- NULL
    cvg.per.primer <- lapply(locations, function(y) unlist(lapply(y, length)))
    IDs <- unlist(lapply(seq_along(cvg.per.primer), 
        function(x) unlist(lapply(seq_along(cvg.per.primer[[x]]), function(y) rep(all.IDs[y], cvg.per.primer[[x]][y])))))

    if (length(location.cols) == 2) {
        # one direction
        my.IDs <- all.IDs[unlist(lapply(seq_along(locations[[1]]), function(x)
                         rep(x, length(unlist(locations[[1]][x])))))]
        directions <- unlist(lapply(seq_along(cvg.per.primer[[1]]),    
                        function(y) rep(primer.df$Direction[y], cvg.per.primer[[1]][y])))
        location.df <- data.frame(ID = my.IDs, Direction = directions, do.call(cbind, my.locations))
    } else {
        # both directions
        my.IDs.fw <- all.IDs[unlist(lapply(seq_along(locations[[1]]), function(x)
                         rep(x, length(unlist(locations[[1]][x])))))]
         my.IDs.rev <- all.IDs[unlist(lapply(seq_along(locations[[3]]), function(x)
                         rep(x, length(unlist(locations[[3]][x])))))]
        my.IDs <- c(my.IDs.fw, my.IDs.rev)
        directions.fw <- unlist(lapply(seq_along(cvg.per.primer[[1]]),    
                        function(y) rep(primer.df$Direction[y], cvg.per.primer[[1]][y])))
        directions.rev <- unlist(lapply(seq_along(cvg.per.primer[[3]]),    
                        function(y) rep(primer.df$Direction[y], cvg.per.primer[[3]][y])))
        location.df.fw <- data.frame(ID = my.IDs.fw, Direction = directions.fw, cbind(my.locations[[1]], my.locations[[2]]))
        location.df.rev <- data.frame(ID = my.IDs.rev, Direction = directions.rev, cbind(my.locations[[3]], my.locations[[4]]))
        location.df <- rbind(location.df.fw, location.df.rev)
    }
    colnames(location.df) <- c("ID", "Direction", "Start", "End")
    if (relation == "rev") {
        old.cnames <- colnames(location.df)
        colnames(location.df)[old.cnames == "Start"] <- "End"
        colnames(location.df)[old.cnames == "End"] <- "Start"
    }
    if (nrow(location.df) == 0) {
        return(NULL)
    }
    return(location.df)
}

#' Data Preparation for Mismatch Plot.
#' @param primer.df A \code{Primers} object.
#' @param template.df A \code{Templates} object.
#' @param mode Whether to compute for on-target or off-target events.
#' @return A data frame with binding information for every primer.
#' @keywords internal
prepare_mm_plot <- function(primer.df, template.df, 
                    mode = c("on_target", "off_target")) {
    if (length(mode) == 0) {
        mode <- "on_target"
    } else {
        mode <- match.arg(mode)
    }
    all.mm.cols <- c("Mismatch_pos_fw", "Mismatch_pos_rev")
    if (mode == "off_target") {
        all.mm.cols <- paste0("Off_", all.mm.cols)
    }
    all.cvg.cols <- c("Covered_Seqs")
    if (mode == "off_target") {
        all.cvg.cols <- paste0("Off_", all.cvg.cols)
    }
    cvg.criteria <- c("off_annealing_DeltaG", "annealing_DeltaG", "off_primer_efficiency", "primer_efficiency", "coverage_model", "off_coverage_model")
    # get constrained coverage data:
    cvg.defs <- c("constrained", "basic")
    dfs <- vector("list", length(cvg.defs))
    for (i in seq_along(cvg.defs)) {
        # define current columns depending on coverage definition
        cvg.def <- cvg.defs[i]
        if (cvg.def == "basic") { 
            cvg.cols <- paste0("Basic_", all.cvg.cols)
            mm.cols <- paste0("Basic_", all.mm.cols)
        } else {
            cvg.cols <- all.cvg.cols
            mm.cols <- all.mm.cols
        }
        if (any(!mm.cols %in% colnames(primer.df))) {
            # no data for selected coverage definition
            next
        }
        abs.mm.pos.fw <- mismatch.string.to.list(primer.df[, mm.cols[1]])
        abs.mm.pos.rev <- mismatch.string.to.list(primer.df[, mm.cols[2]])
        # determine mismatch positions for forward and reverse primers
        fw.idx <- which(primer.df$Direction == "fw")
        rev.idx <- which(primer.df$Direction == "rev")
        both.idx <- which(primer.df$Direction == "both")
        # for primer pairs: consider for every coverage event the fw/rev primer with the higher nbr of mismatches
        # new idea: for 'both' primers: add both entries
        abs.mm.pos <- rep(NA, nrow(primer.df))
        abs.mm.pos[fw.idx] <- abs.mm.pos.fw[fw.idx]
        abs.mm.pos[rev.idx] <- abs.mm.pos.rev[rev.idx]
        abs.mm.pos[both.idx] <- lapply(seq_along(both.idx), function(x) c(abs.mm.pos.fw[both.idx[x]], abs.mm.pos.rev[both.idx[x]]))
        pos.3prime <- rep(NA, length(abs.mm.pos))
        pos.3prime[fw.idx] <- lapply(seq_along(fw.idx), function(x) lapply(abs.mm.pos[[fw.idx[x]]], function(y) nchar(primer.df$Forward[x]) - y + 1))
        pos.3prime[rev.idx] <- lapply(seq_along(rev.idx), function(x) lapply(abs.mm.pos[[rev.idx[x]]], function(y) nchar(primer.df$Reverse[x]) - y + 1))
        pos.3prime[both.idx] <- lapply(both.idx, function(x) unlist(lapply(seq_along(abs.mm.pos[[x]]), function(y) lapply(seq_along(abs.mm.pos[[x]][[y]]),
            function(z) {
                if (y == 1) { # fw primers
                    nchar(primer.df$Forward[x]) - abs.mm.pos[[x]][[y]][[z]] + 1
                } else { # rev primers
                    nchar(primer.df$Reverse[x]) - abs.mm.pos[[x]][[y]][[z]] + 1
                }
            }
         )), recursive = FALSE))
        pos.worst <- lapply(seq_along(abs.mm.pos), function(x) lapply(pos.3prime[[x]], function(y) rep(min(y), length(y))))
        primer.ids <- lapply(seq_along(pos.3prime), function(x) rep(as.character(primer.df$ID[x]), length(unlist(pos.3prime[[x]]))))
        directions <- lapply(seq_len(nrow(primer.df)), function(x) {
            dir <- primer.df$Direction[x]
            rep(dir, primer.df$primer_coverage[x])

            if (dir == "both") {
                c(rep("fw",  length(unlist(abs.mm.pos[[x]][[1]]))), 
                  rep("rev", length(unlist(abs.mm.pos[[x]][[2]]))))
            } else {
                rep(dir, length(unlist(abs.mm.pos[[x]])))
            }
         })
        cvd.idx <- covered.seqs.to.idx(primer.df[, cvg.cols], template.df)
        if (length(unlist(cvd.idx)) == 0) {
            #warning("Nothing covered")
            next
        }
        template.ids <- lapply(seq_len(nrow(primer.df)), function(x) {
            dir <- primer.df$Direction[x]
            if (dir == "both") {
                # fw and rev
                id1 <- unlist(lapply(seq_along(abs.mm.pos[[x]][[1]]), function(y) rep(template.df$ID[cvd.idx[[x]]][y], length(abs.mm.pos[[x]][[1]][[y]]))))
                id2 <- unlist(lapply(seq_along(abs.mm.pos[[x]][[2]]), function(y) rep(template.df$ID[cvd.idx[[x]]][y], length(abs.mm.pos[[x]][[2]][[y]]))))
                ids <- c(id1, id2)
            } else {
                ids <- unlist(lapply(seq_along(abs.mm.pos[[x]]), function(y) rep(template.df$ID[cvd.idx[[x]]][y], length(abs.mm.pos[[x]][[y]]))))
            }
            return(ids)
         })
        # count number of mismatches per primer-template pair
        nbr.mismatches <- lapply(seq_along(pos.3prime), function(x) lapply(pos.3prime[[x]], function(y) rep(length(which(!is.na(y))), length(y))))
        # annotate with group of templates
        m <- match(unlist(template.ids), template.df$ID)
        template.groups <- template.df$Group[m]
		#print(primer.ids) # character on windows
        df <- data.frame(Primer = as.character(unlist(primer.ids)), 
                        Direction = unlist(directions),
                        Template = unlist(template.ids),
                        Group = template.groups,
                        Position_3prime = unlist(pos.3prime),
                        Position_3terminus = unlist(pos.worst),
                        Number_of_mismatches = unlist(nbr.mismatches),
                        Coverage_Type = as.factor(rep(cvg.def, length(template.groups))))
		#print(df$Template)
        # if coverage criteria are available, also include these
        for (z in seq_along(cvg.criteria)) {
            crit <- cvg.criteria[z]
            if (crit %in% colnames(primer.df) && any(!is.na(primer.df[, crit]))) {
                val <- lapply(strsplit(primer.df[, crit], split = ","), as.numeric)
                val <- lapply(seq_along(pos.3prime), function(x)  {
                    dir <- primer.df$Direction[x]
                    if (dir == "both") {
                        val1 <- unlist(lapply(seq_along(abs.mm.pos[[x]][[1]]), function(y) rep(val[[x]][y], length(abs.mm.pos[[x]][[1]][[y]]))), recursive = FALSE)
                        val2 <- unlist(lapply(seq_along(abs.mm.pos[[x]][[2]]), function(y) rep(val[[x]][y], length(abs.mm.pos[[x]][[2]][[y]]))), recursive = FALSE)
                        c(val1, val2)
                    } else {
                        lapply(seq_along(pos.3prime[[x]]), function(y) rep(val[[x]][y], length(pos.3prime[[x]][[y]])))
                    }
                })
                df[, crit] <- unlist(val)
            }
        }
        ####
        # add coverage model position:
        #######
        df$Position_3terminusLocal <- df$Position_3terminus
        df$Position_3terminusLocal[is.na(df$Position_3terminusLocal) | df$Position_3terminusLocal >= 7]  <- 7
        # ensure that position is increasing monotonically: 0 is no mismatch, 6 is mismatch at last 3' hexamer position
        df$Position_3terminusLocal <- abs(df$Position_3terminusLocal - 7)
        # if we have a subsetted template data frame, some primer matches might not be identified -> remove those that shouldn't be considered.
        df <- df[!is.na(df$Template ),]
        dfs[[i]] <- df
    }
    df <- do.call(rbind, dfs)
    if (length(df) != 0) {
        # ensure that the order of the factors doesn't change
        df$Primer <- factor(df$Primer, levels(primer.df$ID))
    }
    # rename off_columns for later use for cvg model
    if (length(df) != 0 && mode == "off_target") {
        # note: this may 'overwrite' the regular columns ..
        colnames(df) <- gsub("^off_", "", colnames(df))
    }
    return(df)
}
prepare_template_cvg_data <- function(df, mode.directionality) {
    # df: coverage for one coverage type
    # select individual binding mode per primer-template pair instead of the representation per individual mismatch
    #########
    ddf <-  ddply(df, c("Primer", "Direction", "Template", "Group"), summarize,
                            Position = unique(substitute(Position_3terminus)), 
                            Number_of_mismatches = unique(substitute(Number_of_mismatches)))

    ##########################
    # take the minium number of mismatches binding mode per direction 
    dff <- ddply(ddf, c("Template", "Direction"), function(x) arrange(x, 
                                        substitute(Number_of_mismatches))[1, ])
    # if mode.directionality is "both" -> retain only templates where we have coverage from a fw AND a rev primer
    if (mode.directionality == "both") {
        # for individual primers, require coverage by fw & rev primers:
        dir.count <- ddply(dff, "Template", summarize, DirectionCount = length(unique(substitute(Direction))))
        # remove all template coverage events where we don't have two primers covering the template
        rm.template.id <- setdiff(dir.count$Template[dir.count$DirectionCount <= 1], dff$Template[dff$Direction == "both"])
        m <- which(dff$Template %in% rm.template.id)
        if (length(m) != 0) {
            dff <- dff[-m,]
        }
    }
    # take the maximum number of mismatches per template over all directions
    # result is: max(min(fw), min(rev))
    dff <- ddply(dff, c("Template"), function(x) arrange(x, 
                                        -substitute(Number_of_mismatches))[1, ])
    return(dff)
}
#' Retrieval of Template Coverage Data.
#'
#' Determines the coverage of the templates for individual
#' allowed mismatch settings and coverage definitions.
#'
#' @param primer.df A \code{Primers} object.
#' @param template.df A \code{Templates} object.
#' @return Computes a data frame providing the coverage of the 
#' templates for the basic as well as expected (constrained) coverage.
#' @keywords internal
get_template_cvg_data <- function(primer.df, template.df) {
    mode.directionality <- get.analysis.mode(primer.df)
    full.df <- prepare_mm_plot(primer.df, template.df)
    # select only unique primer-template pairs with a certain number of mismatches
    unique.cvg.types <- unique(full.df$Coverage_Type)
    plot.data <- vector("list", length(unique.cvg.types))
    for (z in seq_along(unique.cvg.types)) {
        df <- full.df[full.df$Coverage_Type == unique.cvg.types[z],]    
        # select representative binding events for every template
        ddf <- prepare_template_cvg_data(df, mode.directionality)
         # if there's no mismatches, plot these primers with mismatch position @ position 0
        ddf[is.na(ddf$Position), "Position"] <- 0
        ddf$Status <- unique.cvg.types[z]
        plot.data[[z]] <- ddf
    }
    plot.df <- do.call(rbind, plot.data)
    return(plot.df)
}

#' @rdname AnalysisStats
#' @name AnalysisStats
#' @aliases get_cvg_stats_primer
#' @details
#' For \code{get_cvg_stats_primer}, the cells corresponding
#' to columns with numeric identifiers 
#' indicate the percentage of coverage events occurring
#' with a certain number of mismatches. For example
#' column \code{3} provides the number of coverage events
#' where there are exactly three mismatches between primers and templates.
#' The column \code{Group_Coverage} provides a listing
#' of the percentage of covered templates per group.
#'
#' @return \code{get_cvg_stats_primer} returns a list with the following entries. \code{cvg_per_nbr_mismatches} contains a data frame listing 
#' the number of binding events broken down according to the number
#' of expected mismatches between primers and templates.
#' \code{cvg_per_group} contains a data frame listing the the coverage
#' of individual primers per group of templates.
#'
#' @export
#' @examples
#' 
#' data(Ippolito)
#' # Determine coverage stats per primer
#' primer.cvg.stats <- get_cvg_stats_primer(primer.df, template.df)
get_cvg_stats_primer <- function(primer.df, template.df,
                                cvg.definition = c("constrained", "basic")) {
    if (!is(primer.df, "Primers")) {
        stop("'primer.df' should be a 'Primers' object.")
    }
    if (!is(template.df, "Templates")) {
        stop("'template.df' should be a 'Templates' object.")
    }
    cvg.definition <- match.arg(cvg.definition)
    full.df <- prepare_mm_plot(primer.df, template.df)
    # select only constrained cvg events
    full.df <- full.df[full.df$Coverage_Type == cvg.definition, ]
    df <- ddply(full.df, c("Primer", "Template", "Group"), summarize,
                            Position = unique(substitute(Position_3terminus)), 
                            Number_of_mismatches = unique(substitute(Number_of_mismatches)))
    if (length(df) == 0 || nrow(df) == 0) {
        # no stats available
        return(NULL)
    }
    mm.range <- seq(0, max(df$Number_of_mismatches))
    # create a matrix containing the coverage of every primer for every mismatch setting
    mm.stats <- matrix(rep(0, length(mm.range) * nrow(primer.df)), nrow = nrow(primer.df), ncol = length(mm.range))
    for (i in seq_along(mm.range)) {
        mm <- mm.range[i]
        sub.df <- df[df$Number_of_mismatches == mm,]
        count.df <- ddply(sub.df, c("Primer"), summarize,
                            Coverage = length(unique(substitute(Template))))
        m <- match(count.df$Primer, primer.df$ID)
        mm.stats[m,i] <- count.df$Coverage
    }
    ####
    # annotate coverage counts per mismatch number with percentages
    ####
    mm.stats.x <- mm.stats / rowSums(mm.stats)
    mm.stats.x[is.na(mm.stats.x)] <- 0
    for (i in seq_len(nrow(mm.stats))) {
        mm.stats[i, ] <- paste0(mm.stats[i,], " (", round(mm.stats.x[i,], 3) * 100, "%)")
    }
    mm.stats <- data.frame(mm.stats)
    colnames(mm.stats) <- mm.range
    mm.stats <- cbind("Primer" = primer.df$ID, mm.stats)
    ########
    # determine number of templates covered from each group
    ########
    count.df <- ddply(df, c("Primer", "Group"), summarize,
                            Coverage = length(unique(substitute(Template))),
                            CoverageRatio = length(unique(substitute(Template))))
    count.df$CoverageRatio <- count.df$CoverageRatio / unlist(lapply(count.df$Group, function(x) length(which(template.df$Group == x))))
    # create table entry: value + percentage
    count.df$TabEntry <- paste0(count.df$Coverage, " (", 
                            paste0(round(count.df$CoverageRatio * 100, 1), "%"), ")")
    # add percentages
    # order by factorlevels to ensure group order when using 'reshape'
    #print(count.df)
    cols <- c("Primer", "Group", "TabEntry")
    count.df <- count.df[order(count.df$Group), cols]
    # to wide format
    group.df <- reshape(count.df, idvar = "Primer", timevar = "Group", direction = "wide")
    colnames(group.df) <- gsub("TabEntry.", "", colnames(group.df), fixed = TRUE)
    # replace NA with 0
    group.df[is.na(group.df)] <- "0 (0%)"
    out <- list("cvg_per_nbr_mismatches" = mm.stats, "cvg_per_group" = group.df)
    return(out)
}
#' @rdname Plots
#' @name Plots
#' @aliases plot_primer_subsets
#' @details
#' The \code{primer.subsets} argument for \code{plot_primer_subsets} can be computed using
#' \code{\link{subset_primer_set}}. 
#' The line plot indicates the ratio of covered templates when considering
#' all primers in a primer set of a given size. The bar plots indicate
#' the coverage ratios of individual primers in a set. The target coverage
#' ratio is indicated by a horizontal line. Bars exceeding the target ratio
#' possibly indicate the existence of redundant coverage events.
#' @return \code{plot_primer_subsets} plots the coverages of the primer subsets provided via \code{primer.subsets}.
#' @export
#' @examples
#' 
#' # Plot the coverage of optimal primer subsets
#' data(Ippolito)
#' primer.subsets <- subset_primer_set(primer.df, template.df, k = 3)
#' p <- plot_primer_subsets(primer.subsets, template.df)
plot_primer_subsets <- function(primer.subsets, template.df, required.cvg = 1) {
                            
    if (length(primer.subsets) == 0) {
        return(NULL)
    }
    if (!all(sapply(primer.subsets, function(x) is(x, "Primers")))) {
        stop("All subsets should represent primer sets.")
    }
    if (!is(template.df, "Templates")) {
        stop("Please provide a valid template data frame.")
    }
    # create plot.df: need cvg and size of set
    N <- sapply(primer.subsets, nrow)
    cvg <- sapply(seq_along(primer.subsets), function(x) get_cvg_ratio(primer.subsets[[x]], 
        template.df))
    # retrieve the IDs of the primer(s) that are different in the next subset size
    IDs <- lapply(primer.subsets, function(x) x$ID)
    # collect all set members?
    collect.all <- TRUE
    if (!collect.all) {
        # only consider primers that are different from previous ones
    IDs <- lapply(seq_along(IDs), function(x) {
        if (x == 1) {
            out <- IDs[[x]]
        } else {
            out <- setdiff(IDs[[x]], IDs[[x-1]])
        }
        })
    }
    primer.df <- tail(primer.subsets, n = 1)[[1]] # the complete primer set
    individual.cvg <- lapply(IDs, function(x) sapply(x, function(y) get_cvg_ratio(primer.df[match(y, primer.df$ID),], template.df))) # cvg of added primers for this subset size
    #IDs <- unlist(lapply(IDs, function(x) paste(x, collapse = "\n")))
    subset.df <- data.frame(Set_Size = N, Coverage = cvg, Type = "Overall")
    single.df <- data.frame(Set_Size = unlist(lapply(seq_along(N), function(x) rep(N[x], length(individual.cvg[[x]])))),
                            Single_Coverage = unlist(individual.cvg),
                            ID = unlist(IDs), Type = "Individual")
    set.labels <- seq_along(primer.subsets)
    max.cvg <- sum(single.df[single.df$Set_Size == tail(N, n = 1), "Single_Coverage"])
    # order primers by time of addition to set
    levels <- rev(unique(unlist(lapply(primer.subsets, function(x) x$ID)))) # first primers at the bottom
    single.df$ID <- factor(single.df$ID, levels = levels, labels = abbreviate(levels, getOption("openPrimeR.plot_abbrev")))
    pal <- getOption("openPrimeR.plot_colors")["Primer"] # the RColorBrewer palette to use
    colors <- colorRampPalette(brewer.pal(8, pal))(length(unique(single.df[, "ID"])))
    indi.color <- ifelse(length(colors) >= 1, colors[1], NA)
    p <- ggplot() + 
    # add bars for individual cvg
        geom_bar(data = single.df, stat = 'identity', 
                aes(x = .data[["Set_Size"]], y = .data[["Single_Coverage"]], 
                           fill = .data[["ID"]], colour = .data[["Type"]]), 
                alpha = 0.80) +
        scale_color_manual(values=c("Individual"= NA, "Overall" = "grey30"),
            guide = ggplot2::guide_legend(override.aes = list(linetype = c(0, 1), 
                                fill = c(indi.color, NA), 
                                colour = c(NA, "grey20")),
                                title = "Coverage type")) +
        scale_fill_manual(values = colors) +
        geom_point(data = subset.df, size = 1.5, aes(x = .data[["Set_Size"]], y = .data[["Coverage"]], colour = .data[["Type"]]))  +
        geom_line(data = subset.df, size = 1, aes(x = .data[["Set_Size"]], 
                  y = .data[["Coverage"]], colour = .data[["Type"]])) +
        xlab("Subset size") + ylab("Coverage of templates") + 
        ggtitle("Primer subset coverage") + 
        scale_y_continuous(labels = scales::percent, 
                            limits = c(0, max(max.cvg, 1))) + 
        scale_x_continuous(breaks = seq_along(primer.subsets),
                           labels = set.labels) +
        geom_hline(yintercept = required.cvg, linetype = 2, colour = "red") + # indicate the desired cvg
        theme(axis.text.x = element_text(angle = 60, hjust = 1))
    if (length(unique(single.df$ID)) > 30) {
        # don't show individual primer legend if nbr of primers exceeds 30
        p <- p + guides(fill = "none")
    }
    return(p)
}

#' Data for Primer Plot.
#'
#' Constructs a data frame containing information
#' about primer binding events.
#'
#' @param primer.df An object of class \code{Primers} containing
#' primers with evaluated primer coverage.
#' @param template.df An object of class \code{Templates} with template sequences
#' corresponding to \code{primer.df}.
#' @param identifier Identifiers of primers that are to be considered.
#' If \code{identifier} is set to \code{NULL} (the default), all primers are considered.
#' @param relation Compute binding positions relative to forward (\code{fw}) or reverse (\code{rev}) binding regions.
#' The default is "fw".
#' @return Data frame with primer binding data.
#' @keywords internal
get_plot_primer_data <- function(primer.df, template.df, identifier = NULL, relation = c("fw", "rev")) {

    if (!is(primer.df, "Primers")) {
        stop("Please input a valid primer data frame.")
    }
    if (!is(template.df, "Templates")) {
        stop("Please provide a valid template data frame.")
    }
    if (!"primer_coverage" %in% colnames(primer.df)) {
        stop(primer.df$Run[1], ": Please compute the primer coverage first.")
    }
    if (is.null(identifier)) { # plot all primers
        identifier <- primer.df$Identifier
    }
    if (length(relation) == 0) {
        stop("Please provide the 'relation' arg.")
    }
    relation <- match.arg(relation)
    m <- match(identifier, primer.df$Identifier)  # determine primers to be plotted
    if (any(is.na(m))) {
        stop("Could not find specified identifiers.")
    }
    if (length(m) == 0 || is.na(identifier) || identifier == "") {
        # nothing to plot
        return(NULL)
    }
    primer.info <- primer.df[m, ]  # info about selected primers
    s.idx <- lapply(covered.seqs.to.idx(primer.info$Covered_Seqs, template.df), function(x) if (length(x) != 0) {which(!is.na(x))} else {NULL})  # idx of of covered seqs per primer
    sel.idx <- which(!is.na(unlist(covered.seqs.to.idx(primer.info$Covered_Seqs, 
        template.df))))  # unlist idx of all primers covering something
    # need to find the idx of fw covering and rev covering seqs
    fw.idx <- which(primer.info$Forward != "")
    rev.idx <- which(primer.info$Reverse != "")
    if (any(primer.info$Forward != "")) {
        sel.idx.fw <- which(!is.na(unlist(covered.seqs.to.idx(primer.info$Covered_Seqs[fw.idx], 
            template.df))))
        s.idx.fw <- sapply(covered.seqs.to.idx(primer.info$Covered_Seqs[fw.idx], 
            template.df), function(x) if (length(x) != 0) {which(!is.na(x))} else {NULL})
    } else {
        sel.idx.fw <- NULL
        s.idx.fw <- NULL
    }
    if (any(primer.info$Reverse != "")) {
        sel.idx.rev <- which(!is.na(unlist(covered.seqs.to.idx(primer.info$Covered_Seqs[rev.idx], 
            template.df))))
        s.idx.rev <- sapply(covered.seqs.to.idx(primer.info$Covered_Seqs[rev.idx], 
            template.df), function(x) if (length(x) != 0) {which(!is.na(x))} else {NULL})
    } else {
        sel.idx.rev <- NULL
        s.idx.rev <- NULL
    }
    # retrieve coverage info:
    covered.seqs <- lapply(seq_along(primer.info$Covered_Seqs), function(x) as.numeric(unlist(strsplit(primer.info$Covered_Seqs[x], 
        split = ",")[s.idx[[x]]])))
    # determine relative primer starts and ends depending on whether we want to
    # analyze regarding fw allowed region or rev allowed region
    if (relation == "fw") {
        primer.start <- lapply(seq_len(nrow(primer.info)), function(x) as.numeric(unlist(strsplit(primer.info$Relative_Forward_Binding_Position_Start_fw[x], 
            split = ","))[s.idx[[x]]]))
        primer.end <- lapply(seq_len(nrow(primer.info)), function(x) as.numeric(unlist(strsplit(primer.info$Relative_Forward_Binding_Position_End_fw[x], 
            split = ","))[s.idx[[x]]]))
        primer.start.rev <- lapply(seq_len(nrow(primer.info)), function(x) as.numeric(unlist(strsplit(primer.info$Relative_Forward_Binding_Position_End_rev[x], 
            split = ","))[s.idx[[x]]]))
        primer.end.rev <- lapply(seq_len(nrow(primer.info)), function(x) as.numeric(unlist(strsplit(primer.info$Relative_Forward_Binding_Position_Start_rev[x], 
            split = ","))[s.idx[[x]]]))
    } else {
        primer.start <- lapply(seq_len(nrow(primer.info)), function(x) as.numeric(unlist(strsplit(primer.info$Relative_Reverse_Binding_Position_Start_fw[x], 
            split = ","))[s.idx[[x]]]))
        primer.end <- lapply(seq_len(nrow(primer.info)), function(x) as.numeric(unlist(strsplit(primer.info$Relative_Reverse_Binding_Position_End_fw[x], 
            split = ","))[s.idx[[x]]]))
        #print("primer posis:")
        #print(primer.start)
        #print(primer.end)
        primer.start.rev <- lapply(seq_len(nrow(primer.info)), function(x) as.numeric(unlist(strsplit(primer.info$Relative_Reverse_Binding_Position_End_rev[x], 
            split = ","))[s.idx[[x]]]))
        primer.end.rev <- lapply(seq_len(nrow(primer.info)), function(x) as.numeric(unlist(strsplit(primer.info$Relative_Reverse_Binding_Position_Start_rev[x], 
            split = ","))[s.idx[[x]]]))
        
    }
    covered.seq.identifiers <- as.numeric(unlist(strsplit(primer.info$Covered_Seqs, 
        split = ",")))[sel.idx]
    # determine nbr covered for given template.df
    nbr.covered <- sapply(seq_along(seq_len(nrow(primer.info))), function(x) length(strsplit(primer.info$Covered_Seqs[x], 
        split = ",")[[1]][s.idx[[x]]]))  # nbr of covered seqs per primer
    m <- match(covered.seq.identifiers, template.df$Identifier)
    not.na.mapping <- which(!is.na(m))
    m <- m[not.na.mapping]
    covered.seqs <- template.df[m, ]
    not.m <- setdiff(seq_len(nrow(template.df)), m)
    uncovered.seqs <- template.df[not.m, ]
    repeat.idx <- unlist(sapply(seq_along(nbr.covered), function(x) rep(x, each = nbr.covered[x])))  # assignment of index in primer.info to all covered seqs from each primers
    primer.info.m <- primer.info[repeat.idx, ]
    total.covered <- sum(nbr.covered)
    total.y <- total.covered + nrow(template.df)  # dimension of plot: nbr of primers + nbr of seqs
    plot.df <- data.frame(Type = rep(NA, total.y), y = rep(NA, total.y), 
                          Map = rep(NA, total.y))
    y.idx <- 1
    mode.directionality <- get.analysis.mode(primer.info)
    for (i in seq_len(nrow(template.df))) {
        id <- template.df$Identifier[i]
        # determine nbr of occurrences of this seq in covered.seq.identifiers vector
        count.idx <- which(covered.seq.identifiers == id)  # how often was this sequence covered?
        p.idx <- repeat.idx[count.idx] # index in primer info
        covered <- FALSE
        if (mode.directionality == "both") {
            # require at least 1 fw and 1 rev primer
            fw.count <- length(which(primer.info[p.idx,]$Forward != ""))
            rev.count <- length(which(primer.info[p.idx,]$Reverse != ""))
            count <- min(c(fw.count, rev.count))
        } else {
            # require any primer to cover a sequence
            count <- length(count.idx)
        }
        if (count != 0) { # 
            covered <- TRUE
        }
        seq.idx <- y.idx
        primer.idx <- (seq.idx + 1):(seq.idx + 1 + length(count.idx) - 1)
        plot.df$Type[seq.idx] <- ifelse(covered, "Covered Sequence", "Not-covered Sequence")
        plot.df$y[seq.idx] <- seq.idx
        plot.df$Map[seq.idx] <- i
        if (length(count.idx) != 0) {
            # determine fw and rev primers
            plot.df$Type[primer.idx] <- "Primer"
            plot.df$y[primer.idx] <- primer.idx
            plot.df$Map[primer.idx] <- p.idx  # gives the index of primer in primer.info
        } 
        y.idx <- y.idx + 1 + length(count.idx)
    }
    # enrich with info
    primer.data.idx <- plot.df$Map[which(plot.df$Type == "Primer")]
    pos.idx <- get.unlist.idx(primer.start, primer.data.idx)
    covered.seq.identifiers <- sapply(seq_along(primer.info$Covered_Seqs), function(x) as.numeric(strsplit(primer.info$Covered_Seqs[x], 
        split = ",")[[1]][s.idx[[x]]]))
    primer.cov.idx <- sapply(covered.seq.identifiers, function(x) match(x, template.df$Identifier))
    primer.gene.groups <- sapply(primer.cov.idx, function(x) template.df[x, "Group"])
    fw.primer.name <- "fw primer"
    rev.primer.name <- "rev primer"
    primer.names <- c(fw.primer.name, rev.primer.name)
    primer.data <- data.frame(Identifier = as.character(primer.info$Identifier[primer.data.idx]), 
                    ID = as.character(primer.info$ID[primer.data.idx]),
                    stringsAsFactors = FALSE) 
    primer.data <- cbind(primer.data, x_start = unlist(primer.start)[pos.idx][not.na.mapping], 
        x_end = unlist(primer.end)[pos.idx][not.na.mapping], x_start_rev = unlist(primer.start.rev)[pos.idx][not.na.mapping], 
        x_end_rev = unlist(primer.end.rev)[pos.idx][not.na.mapping], Group = unlist(primer.gene.groups)[pos.idx][not.na.mapping])
    seq.data.idx <- plot.df$Map[which(plot.df$Type != "Primer")]
    #print("primer data:")
    #print(primer.data)
    # n.b.: not using initial here anymore -> plot relates to the current binding region
    seq.data <- data.frame(Identifier = template.df$Identifier[seq.data.idx], 
        ID = template.df$ID[seq.data.idx], 
        x_start = -template.df$Allowed_End_fw[seq.data.idx], 
        x_end = nchar(template.df$Sequence)[seq.data.idx] - 
                template.df$Allowed_End_fw[seq.data.idx] - 1, 
        x_start_rev = NA,  # only relevant for primers
        x_end_rev = NA, 
        Group = template.df$Group[seq.data.idx],
        stringsAsFactors = FALSE)
    if (relation == "fw") {
        seq.data$x_start <- -template.df$Allowed_End_fw[seq.data.idx]
        seq.data$x_end <- nchar(template.df$Sequence)[seq.data.idx] - 
                            template.df$Allowed_End_fw[seq.data.idx] - 1 
    } else {
        seq.data$x_start <- template.df$Allowed_Start_rev[seq.data.idx] - nchar(template.df$Sequence)[seq.data.idx]
        seq.data$x_end <- nchar(template.df$Sequence[seq.data.idx]) - seq.data$x_start - 1
    }
    add.df <- rbind(primer.data, seq.data)[FALSE, ]  # primer data + seq data should make up the plot.df
    add.df[which(plot.df$Type == "Primer"), ] <- primer.data
    add.df[which(plot.df$Type != "Primer"), ] <- seq.data
    rownames(add.df) <- NULL
    rownames(plot.df) <- NULL
    plot.data <- cbind(plot.df, add.df)
    plot.data$Type <- factor(plot.data$Type, levels = c("Primer", "Covered Sequence", 
        "Not-covered Sequence"))
    ########## determine limits of plot
    #seq.idx <- which(plot.data$Type != "Primer")
    #s.min <- min(plot.data[seq.idx, "x_start"], na.rm = TRUE)
    #seq.max <- max(plot.data[seq.idx, "x_end"], na.rm = TRUE)
    #primer.idx <- which(plot.data$Type == "Primer")
    #x.lim <- c(s.min, seq.max)  # max was p.max before
    #print("before plot lim adjustment")
    #print(plot.data)
    #plot.data$x_start <- sapply(plot.data$x_start, function(x) max(x, x.lim[1]))
    #plot.data$x_end <- sapply(plot.data$x_end, function(x) min(x, x.lim[2]))
    #plot.data$x_start_rev <- sapply(plot.data$x_start_rev, function(x) max(x, x.lim[1]))
    #plot.data$x_end_rev <- sapply(plot.data$x_end_rev, function(x) min(x, x.lim[2]))
    ########## 
    cur.plot.df <- plot.data
    cur.plot.df$ID <- as.character(cur.plot.df$ID)
    # plot only the templates
    rel.cols <- c("ID", "y", "Type", "x_start", "x_end", "x_start_rev", "x_end_rev")
    m <- melt(cur.plot.df[, rel.cols], id = c("y", "Type", "ID"))
    # remove NA entries
    na.idx <- which(is.na(m[, "value"]))
    if (length(na.idx) != 0) {
        m <- m[-na.idx, ]
    }
    # update types
    types <- ifelse(m$variable == "x_start_rev" | m$variable == "x_end_rev", "Reverse Primer", 
        ifelse(m$Type == "Primer", "Forward Primer", as.character(m$Type)))
    m$Type <- types
    m$variable <- ifelse(grepl("start", m$variable), "x_start", "x_end")
    d <- dcast(m, ID + y + Type ~ variable)
    return(d)
}
#' @rdname Plots
#' @name Plots
#' @aliases plot_primer
#' @return \code{plot_primer} plots the primer binding sites in the templates.
#' @export
#' @examples
#' 
#' # Plot of individual primer binding positions
#' data(Ippolito)
#' p <- plot_primer(primer.df[1,], template.df[1:30,])
plot_primer <- function(primer.df, template.df, identifier = NULL, 
                        relation = c("fw", "rev"), 
                        region.names = c("Binding region", "Amplification region")) {

    if (length(region.names) != 2) {
        stop("Need 2 region names.")
    }
    relation <- match.arg(relation)
    d <- get_plot_primer_data(primer.df, template.df, identifier = identifier, relation = relation)
    # region annotation:
    y.val <- 0.05 * nrow(d)
    y.ext <- min(10, y.val)
    ymin <- min(-y.ext, -3)
    ymax <- min(-1, -0.01 * nrow(d))
    region.df <- create_region_boxes(list(primer.df), list(template.df), relation, region.names, ymin, ymax, max(d$x_end, na.rm = TRUE))

    if (length(d) == 0 || nrow(d) == 0) {
        # nothing to plot
        return(NULL)
    }
    col <- c("Forward Primer" = brewer.pal(8, "Accent")[5], 
            "Reverse Primer" = "#0B4D46", 
            "Covered Sequence" = "black", 
            "Not-covered Sequence" = "grey70")
    m <- match(unique(d$Type), names(col))
    col <- col[m[which(!is.na(m))]]
    label.idx <- which(d$Type %in% c("Covered Sequence", "Not-covered Sequence"))
    label.idx.primer <- which(d$Type %in% c("Forward Primer", "Reverse Primer") & 
        !duplicated(d$y))
    label.idx <- c(label.idx, label.idx.primer)
    label.order <- order(d$y[label.idx])
    labels <- d$ID[label.idx][label.order]
    template.data <- d[d$Type %in% c("Covered Sequence", "Not-covered Sequence"), ]
    primer.data <- d[d$Type %in% c("Forward Primer", "Reverse Primer"), ]
    if (length(primer.data) == 0 || nrow(primer.data) == 0) {
        # no primers to plot
        return(NULL)
    }
    break.step.size <- 10
    lim.min <- min(d$x_start, na.rm = TRUE)
    lim.max <- max(d$x_end, na.rm = TRUE)
    ticks <- c(seq(lim.min, lim.max, break.step.size))
    # determine coverage count
    N.total <- length(which(!d$Type %in% c("Forward Primer", "Reverse Primer")))
    N.covered <- length(which(d$Type == "Covered Sequence"))
    N.p <- N.covered/N.total
    stat.text <- paste(N.covered, " of ", N.total, " (", round(N.p, 3) * 100, "%) seqs covered", 
        sep = "")
    title <- stat.text
    segment.size <- 1.5
    arrow.size <- 2
    # arrow angle needs to be a function of the number of coverage events -> the more events, the larger the angle: from 5 to 30
    # number of primers
    min.arrow.angle <- 30 # was 5 before, not necessary anymore
    max.arrow.angle <- 30
    arrow.angle <- max(min(0.5 * (nrow(primer.data) + nrow(template.data)), max.arrow.angle), min.arrow.angle)
    # arrow length should be scaled according to the x-extent of the plot
    x.extent <- max(template.data$x_end) - min(template.data$x_start)
    p.extent <- max(primer.data$x_end) - min(primer.data$x_start)
    arrow.length <- 80 * (p.extent / x.extent^2)
    x.ticks <- c(seq(floor(lim.min/10) * 10, floor(lim.max/10) * 10, break.step.size))
    x.labels <- x.ticks
    x.labels[x.labels > 0] <- paste0("+", x.labels[x.labels > 0])
    r.colors <- c("#e5f4ff", "#ffefe5")
    ggplot() + 
        # need show.legend = FALSE for the segment, otherwise arrow shows
        geom_segment(data = primer.data, show.legend = FALSE,
            lineend = "butt", size = arrow.size,  
            aes(x = .data[["x_start"]], xend = .data[["x_end"]], y = .data[["y"]], 
                       yend = .data[["y"]], colour = .data[["Type"]]),
            arrow = arrow(angle = arrow.angle, ends = "last", type = "closed")) + 
             geom_segment(data = template.data, 
                     aes(x = .data[["x_start"]], xend = .data[["x_end"]], 
                                y = .data[["y"]], yend = .data[["y"]], colour = .data[["Type"]]), 
                    lineend = "square", size = segment.size) + 

        # x-axis rectangles to annotate binding/amplification region:
        geom_rect(data = region.df, 
            mapping = aes(xmin=.data[["xmin"]], xmax=.data[["xmax"]], 
                        ymin=.data[["ymin"]], ymax=.data[["ymax"]]), 
            fill = r.colors, alpha = 0.5,
            colour = "#3d3835", 
            size = segment.size) +
        # text for region annotation
        geom_text(data=region.df, 
                  aes_string(x = "xmin+(xmax-xmin)/2", 
                             y = "ymin+(ymax-ymin)/2", 
                             label = "Region"), 
                  size = 4) +
        ggtitle(title) + 
        xlab("Relative primer position") + 
        ylab("Sequence") + 
        scale_x_continuous(limits = c(lim.min, lim.max), 
                            breaks = x.ticks,
                            labels = x.labels) + 
        theme(axis.text.x = element_text(angle = 90, hjust = 1, vjust = 0.5)) +
        geom_vline(data = region.df, aes(xintercept = .data[["RelStartPosition"]]), colour = "red") + # start of binding region
        geom_vline(xintercept = -1, colour = "red") + 
        #scale_y_continuous(limits = c(1, max(d$y, na.rm = TRUE)), 
        scale_y_continuous(
                breaks = seq_along(labels), labels = abbreviate(labels, getOption("openPrimeR.plot_abbrev"))) + 
        scale_colour_manual(values = col) +
        theme(legend.title = element_text(face = "bold"),
              legend.position = "top")
        # arrow arg doesn't work for linetype ...
        #guides(colour = guide_legend(override.aes = 
                #list(size = 1, linetype = "solid", arrow = arrow(length = unit(1, "cm")))))
       
}

#' Plot of Excluded Primers
#' 
#' Plots histogram of excluded primers.
#'
#' @param excluded.df Data frame with excluded primers.
#' @param filtered.stats Data frame with statistics of the filtering procedure.
#' @param template.df Template data frame.
#' @return A plot of excluded primers.
#' @keywords internal
plot.excluded.hist <- function(excluded.df, filtered.stats, template.df) {
    # histogram of excluded sequences
    if (length(excluded.df) == 0 || nrow(excluded.df) == 0) {
        # nothing to plot
        return(NULL)
    }
    if (!"Coverage_Ratio" %in% excluded.df || all(is.na(excluded.df$Coverage_Ratio))) {
        # no cvg to plot
        return(NULL)
    }
    ylab <- "Coverage"
    xlab <- "Constraint"
    tit <- "Excluded primer coverage"
    plot.df <- excluded.df
    plot.df$Filter_Reason <- factor(plot.df$Exclusion_Reason, levels = unique(filtered.stats$Constraint))
    ggplot(plot.df, aes(x = .data[["Filter_Reason"]], y = .data[["Coverage_Ratio"]], colour = .data[["Coverage_Ratio"]])) + 
        ylab(ylab) + xlab(xlab) + ggtitle(tit) + 
        theme(axis.text.x = element_text(angle = 90, hjust = 1)) + 
        scale_colour_gradient(name = "Coverage") + 
        geom_boxplot(outlier.colour = NA) + 
        geom_point(position = position_jitter(width = 0.5, height = 0)) +
        scale_y_continuous(labels = scales::percent) +
        facet_wrap(~Direction)
}

#' @rdname Plots
#' @name Plots
#' @aliases plot_template_cvg
#' @return \code{plot_template_cvg} creates a plot showing the number of covered template sequences.
#' @family templates
#' @export
#' @include primers.R templates.R
#' @examples
#' 
#' # Visualize the template coverage of a single primer set
#' data(Ippolito)
#' p.cvg <- plot_template_cvg(primer.df, template.df)
#' # Stratify by allowed mismatches:
#' p.mm.cvg <- plot_template_cvg(primer.df, template.df, per.mismatch = TRUE)
#' # Compare the coverage of multiple primer sets
#' data(Comparison)
#' p.cmp.cvg <- plot_template_cvg(primer.data[1:2], template.data[1:2])
#' # Stratify by allowed mismatches:
#' p.cmp.cvg.mm <- plot_template_cvg(primer.data[1:2], template.data[1:2], 
#'                                  per.mismatch = TRUE)
setGeneric("plot_template_cvg", 
    function(primers, templates, per.mismatch = FALSE, ...) {
        standardGeneric("plot_template_cvg")
})

#' Bar Plot of Template Coverage.
#'
#' Creates a bar plot showing the coverage for every group of template sequences.
#'
#' @param primers A \code{Primers} object with evaluated primer coverage.
#' @param templates A \code{Templates} object containing the template sequences.
#' @param per.mismatch Whether to stratify by mismatches.
#' @param groups Identifiers of template groups for which plot should be created. By default, \code{groups} is set to \code{NULL} such that all 
#' templates are considered.
#' according to the number of mismatches between primer-template pairs.
#' @return A plot showing the number of covered template sequences.
#' @keywords internal
setMethod("plot_template_cvg", 
    signature(primers = "Primers", templates = "Templates"),
    function(primers, templates, per.mismatch, groups = NULL) {
    if (length(primers) == 0 || length(templates) == 0) {
        return(NULL)
    } 
    if (!is(primers, "Primers")) {
        stop("Please input a valid primer data frame.")
    }
    if (!is(templates, "Templates")) {
        stop("Please input a valid template data frame.")
    }
    if (per.mismatch) {
        # stratify by mismatches
        p <- plot_template_cvg_mismatches(primers, templates, groups)
    } else {
        p <- plot_template_cvg_unstratified(primers, templates, groups)
    }
    return(p)
})

#' Bar Plot of Template Coverage.
#'
#' Creates a bar plot showing the coverage for every group of template sequences.
#'
#' @param primers A \code{Primers} object with evaluated primer coverage.
#' @param templates A \code{Templates} object containing the template sequences.
#' @param groups Identifiers of template groups for which plot should be created. By default, \code{groups} is set to \code{NULL} such that all 
#' templates are considered.
#' according to the number of mismatches between primer-template pairs.
#' @return A plot showing the number of covered template sequences.
#' @keywords internal
plot_template_cvg_unstratified <- function(primers, templates, groups = NULL) {
    # get statistics on expected coverage
    stats <- get_cvg_stats(primers, templates)
    if (length(stats) == 0) {
        warning("No coverage statistics available for the input primers.")
        return(NULL)
    }
    cvg.con <- paste0(round(stats[stats$Group == "Total", "Coverage_Ratio"] * 100, 0), "%")
    # compute coverage stats according to text identity:
    stats.txt <- get_cvg_stats(primers, templates, allowed.mismatches = 0, cvg.definition = "basic")
    cvg.txt <- paste0(round(stats.txt[stats.txt$Group == "Total", "Coverage_Ratio"] * 100, 0), "%")
    # need to melt
    # change names of columns
    vars <- c("Group", "primer_coverage", "N")
    p.df.con <- melt(stats[, vars], c("Group"), variable.name = "Status", value.name = "Count")
    levels(p.df.con$Status) <- c("Expected Coverage (E)", "Available Templates")
    vars <- c("Group", "primer_coverage")
    p.df.txt <- melt(stats.txt[, vars], c("Group"), variable.name = "Status", value.name = "Count")
    p.df.txt$Status <- "Identity Coverage (I)"
    # integrate both coverage results
    plot.df <- rbind(p.df.txt, p.df.con)
    plot.df$Status <- factor(plot.df$Status, levels = unique(plot.df$Status))
    # remove total cvg column
    plot.df <- plot.df[plot.df$Group != "Total",]
    # select groups
    if (length(groups) != 0 && !"all" %in% groups) {
        plot.df <- plot.df[plot.df$Group %in% groups, ]
    }
    colnames(plot.df)[colnames(plot.df) == "N"] <- "Available Templates"
    xlab <- "Group"
    ylab <- "Number of templates"
    title <- paste0("Covered templates (I = ", cvg.txt, ", E = ", cvg.con, ")")
    cols.1 <- brewer.pal(8, "Accent")[c(5, 8)]
    cols.2 <- brewer.pal(8, "Blues")[3]
    colors <- c(cols.2, cols.1)
    names(colors) <- levels(plot.df$Status)
    p <- ggplot(plot.df) + geom_bar(aes(x = .data[["Group"]], y = .data[["Count"]], fill = .data[["Status"]]), stat = "identity", 
        position = "dodge") + xlab(xlab) + ggtitle(title) + 
        ylab(ylab) + 
        theme(axis.text.x = element_text(
                angle = 90, 
                hjust = 1, vjust = 0.5)) +
        scale_fill_manual(values = colors)
    return(p)
}
#' Preparation of Data for Plotting Mismatch Template Coverage.
#'
#' Creates a data frame for plotting a bar plot for the covered templates per allowed mismatches.
#'
#' @param primer.df A \code{Primers} object.
#' @param template.df A \code{Templates} object.
#' @param allowed.mismatches An optional numeric specifying the number of mismatches to be considered at most for plotting.
#' If not provided, the maximal number of mismatches found for the input primer 
#' set is used.
#' @return A data frame for creating a plot.
#' @keywords internal
prepare_template_cvg_mm_data <- function(primer.df, template.df, allowed.mismatches = NULL) {
    plot.df <- get_template_cvg_data(primer.df, template.df)
    #########################
    # get data for available templates for every mismatch category:
    t.df <- data.frame("Template" = unique(template.df$ID), 
                       "Group" = NA, "Position" = NA, "Number_of_mismatches" = NA)
    m <- match(t.df$Template, template.df$ID)
    t.df$Group <- template.df$Group[m]
    if (length(allowed.mismatches) == 0) {
        allowed.mismatches <- max(plot.df$Number_of_mismatches)
    }
    mm.settings <- seq(0, allowed.mismatches)
    mm.settings <- mm.settings[order(mm.settings)]
    # identify available number of template sequences
    available.data <- vector("list", length(mm.settings))
    for (i in seq_along(mm.settings)) {
        df.a <- t.df
        df.a$Number_of_mismatches <- mm.settings[i]
        available.data[[i]] <- df.a
    }
    available.df <- do.call(rbind, available.data)
    available.df$Status <- as.factor("Available")
    p.df <- rbind(plot.df[, !colnames(plot.df) %in% c("Primer", "Direction")], available.df)
    cvg.groups <- c("Coverage", "Basic Coverage")
    all.groups <- c(cvg.groups, "Available Templates")
    levels(p.df$Status) <- all.groups
    # do not consider the exact nbr of mismatches, but the cumulative number (cvg with >= x mismatches rather than cvg with == x mismatches)
    cum.data <- p.df[p.df$Status == "Available Templates", ] # start from available data
    cum.data$Maximal_mismatches <- cum.data$Number_of_mismatches
    for (i in seq_along(mm.settings)) {
        mm <- mm.settings[i]
        for (j in seq_along(cvg.groups)) {
            cvg.status <- cvg.groups[j]
            event.idx <- which(p.df$Status == cvg.status & p.df$Number_of_mismatches <= mm)
            if (length(event.idx) != 0) {
                cur.df <- p.df[event.idx,]
                cur.df$Maximal_mismatches <- mm
                cum.data <- rbind(cum.data, cur.df) 
            }
        }
    }
    o <- order(unique(as.character(cum.data$Group)))
    cum.data$Group <- factor(cum.data$Group, levels = unique(as.character(cum.data$Group))[o])
    # need to plot using identity as stat in order to ensure that bars have all the same width
    plot.df <- ddply(cum.data, c("Group", "Maximal_mismatches", "Status"), summarize, Count = length(substitute(Group)))
    additional.df <- expand.grid(Maximal_mismatches = mm.settings, Group = unique(template.df$Group), Status = unique(plot.df$Status), Count = 0)
    plot.df <- merge(plot.df, additional.df, all = TRUE)
    # for the added, duplicate events, select the 'real events'
    plot.df <- ddply(plot.df, c("Group", "Maximal_mismatches", "Status"), summarize,
                           Count = max(substitute(Count)))
    total.percentages <- TRUE
    if (!total.percentages) {
        idx <- which(plot.df$Status == "Available Templates")
        idx <- idx[!duplicated(plot.df$Group[idx])]
        available.per.group <- plot.df$Count[idx]
        names(available.per.group) <- plot.df$Group[idx]
        m <- match(plot.df$Group, names(available.per.group))
        plot.df$Coverage_Ratio <- plot.df$Count / available.per.group[m]
    } else {
        plot.df$Coverage_Ratio <- plot.df$Count / nrow(template.df)
    }
    return(plot.df)
}
#' Bar Plot of Template Coverage for Mismatches.
#'
#' Creates a bar plot showing the coverage for every group of template sequences.
#'
#' @param primer.df A \code{Primers} object.
#' @param template.df A \code{Templates} object.
#' @param groups Identifiers of template groups for which plot should be created. By default, \code{groups} is set to \code{NULL} such that all 
#' templates are considered.
#' @param nfacets The number of facets columns to plot. By default,
#' \code{nfacets} is set to 2.
#' @return A plot showing the number of covered template sequences.
#' @keywords internal
plot_template_cvg_mismatches <- function(primer.df, template.df, groups = NULL,
                                        nfacets = 2) {
    if (!is.null(groups) && !"all" %in% groups) {
        # select a template subset
        idx <- which(template.df$Group %in% groups)
        template.df <- template.df[idx,]
    }
    plot.df <- prepare_template_cvg_mm_data(primer.df, template.df)
    if (nrow(plot.df) == 0) {
        return(NULL)
    }
    # compute cvg ratio per mismatch setting to show in facet labels:
    cvg.per.mm <- ddply(plot.df, c("Maximal_mismatches", "Status"), here(summarize),
                           Coverage_Ratio = sum(substitute(Count)) / nrow(template.df)) 
    cvg.per.mm <- cvg.per.mm[cvg.per.mm$Status == "Coverage",]
    cvg.info <- paste0("(", round(cvg.per.mm$Coverage_Ratio * 100, 0), "% coverage)")
    plot.df$Maximal_mismatches <- paste0(plot.df$Maximal_mismatches, " ", cvg.info[match(plot.df$Maximal_mismatches, cvg.per.mm$Maximal_mismatches)])
    xlab <- "Group"
    ylab <- "Number of templates"
    title <- "Covered templates: mismatches"
    cols.1 <- brewer.pal(8, "Accent")[c(5, 8)]
    cols.2 <- brewer.pal(8, "Blues")[3]
    colors <- c(cols.1[1], cols.2, cols.1[2])
    names(colors) <- levels(plot.df$Status)
    p <- ggplot(plot.df) + 
        geom_bar(aes(x = .data[["Group"]], y = .data[["Count"]], fill = "Status"),
                 position = "dodge", stat = "identity") + 
        xlab(xlab) + ggtitle(title) + 
        ylab(ylab) + 
        theme(axis.text.x = element_text(
                angle = 90, 
                hjust = 1, vjust = 0.5)) +
        scale_fill_manual(values = colors) + 
        facet_wrap(~Maximal_mismatches, ncol = nfacets,
        labeller = label_bquote("Mismatches"<=.(substitute(Maximal_mismatches))))
    return(p)
}
#' Templates Coverage for Multiple Primer Sets.
#'
#' Plots the coverage of multiple primer sets.
#'
#' @param primers List with primer data frames.
#' @param templates List with template data frames.
#' @param colors Color for every primer set.
#' @param highlight.set Primer sets to be highlighted.
#' @return A plot for comparing primer coverage.
#' @keywords internal
setMethod("plot_template_cvg", 
    signature(primers = "list", templates = "list"),
    function(primers, templates, per.mismatch, highlight.set = NULL) {
    if (length(primers) == 0 || length(templates) == 0) {
        return(NULL)
    }
    if (per.mismatch) {
        p <- plot_template_cvg_comparison_mismatch(primers, templates, highlight.set)
    } else {
        p <- plot_template_cvg_comparison_unstratified(primers, templates, highlight.set)
    }
    return(p)
})
#' Getter for Run Names.
#'
#' Retrieves the run names of the input data.
#'
#' @param primer.data A list with \code{Primers} or \code{Templates}.
#' @return A vector with identifiers for every set.
#' @keywords internal
get.run.names <- function(primer.data) {
    # if primer.data is a named list, use the names of the list
    if (length(primer.data) == 0) {
        return(NULL)
    }
    if (length(names(primer.data) != 0)) {
        run.names <- names(primer.data)
    } else {
        # primer.data is not a named list -> use the 'Run' identifier instead
        #run.names <- unname(unlist(lapply(primer.data, function(x) unique(x$Run))))
        run.names <- lapply(primer.data, function(x) unique(x$Run))
        empty.idx <- which(unlist(lapply(primer.data, function(x) nrow(x) == 0)))
        run.names[empty.idx] <- "Unknown"
        run.names <- unname(unlist(run.names))
    }
    return(run.names)
}
#' Templates Coverage for Multiple Primer Sets.
#'
#' Plots the coverage of multiple primer sets.
#'
#' @param primers List with primer data frames.
#' @param templates List with template data frames.
#' @param highlight.set Primer sets to be highlighted.
#' @return A plot for comparing primer coverage.
#' @keywords internal
plot_template_cvg_comparison_unstratified <- function(primers, templates, highlight.set = NULL) {

    if (length(primers) == 0 || length(templates) == 0) {
        return(NULL)
    }
    run.names <- get.run.names(primers)
    cvg <- lapply(seq_along(primers), function(x) {
        stats <-  openPrimeR::get_cvg_stats(primers[[x]], 
                    templates[[x]],
                    total.percentages = TRUE)
        if (length(stats) != 0) {
            stats <- cbind("Run" = run.names[x], stats)
        }
    })
    plot.df <- do.call(rbind.fill, cvg)
    if (length(plot.df) == 0 || nrow(plot.df) == 0) {
        return(NULL)
    }
    # remove the total group
    plot.df <- plot.df[plot.df$Group != "Total", ]
    plot.df$Group <- factor(plot.df$Group, levels = unique(plot.df$Group)[order(unique(plot.df$Group))])
    title <- "Coverage of the template sequences"
    plot.df$Run <- abbreviate(plot.df$Run, getOption("openPrimeR.plot_abbrev"))
    plot.df$Run <- factor(plot.df$Run, levels = unique(plot.df$Run)[order(as.character(unique(plot.df$Run)))])
    if (length(highlight.set) != 0) {
        # check whether highlight set is specified correctly
        m <- match(highlight.set, plot.df$Run)
        na.idx <- which(is.na(m))
        if (length(na.idx) != 0) {
            msg <- paste("Highlight set not available in data:",
                paste(highlight.set[na.idx], collapse = ","))
            warning(msg)
            highlight.set <- highlight.set[!is.na(m)]
        }
    }
    pal <- getOption("openPrimeR.plot_colors")["Group"] # the RColorBrewer palette to use
    colors <- colorRampPalette(brewer.pal(8, pal))(length(unique(plot.df$Group)))
    # add label for top of the bars (the first group) 
    labels <- rep(NA, nrow(plot.df))
    for (i in seq_along(levels(plot.df$Run))) {
        run <- levels(plot.df$Run)[i]
        idx <- which(plot.df$Run == run)
        cur.data <- plot.df[idx,]
        res <- sum(cur.data$Coverage_Ratio) # the coverage to show for this bar
        out.idx <- which(plot.df$Run == run & plot.df$Group == unique(cur.data$Group)[1]) # row index to show the coverage
        labels[out.idx] <- res
    }
    plot.df$Label <- ifelse(is.na(labels), "", paste0(round(labels * 100, 1), "%"))

    p <- ggplot(plot.df, aes(x = .data[["Run"]], y = .data[["Coverage_Ratio"]], fill = .data[["Group"]])) + 
        geom_bar(stat = "identity") + 
        theme(axis.text.x = element_text(angle = 60, hjust = 1)) +
        scale_y_continuous(labels = scales::percent, limits = c(0, 1)) +
        ggtitle(title) +
        ylab("Coverage") +
        scale_fill_manual(values = colors) + 
        # add coverage above bars:
        geom_text(aes(label = plot.df$Label),
        #geom_text(aes(label = substitute(Label)),
                    size = 5, 
                    position = position_stack(vjust = 0.5),
                    check_overlap = FALSE)
    if (length(highlight.set) != 0) {
        # highlight selected sets
        sel <- levels(plot.df$Run) %in% highlight.set
        p <- p + theme(
                axis.text.x = element_text(
                    face=ifelse(sel, "bold","plain"),
                    colour = ifelse(sel, 
                            "grey20", "grey30")))
    }
    return(p)
}
#' Templates Coverage for Multiple Primer Sets.
#'
#' Plots the coverage of multiple primer sets.
#'
#' @param primers List with primer data frames.
#' @param templates List with template data frames.
#' @param highlight.set Primer sets to be highlighted.
#' @return A plot for comparing primer coverage.
#' @keywords internal
plot_template_cvg_comparison_mismatch <- function(primers, templates, 
            highlight.set = NULL) {

    if (length(primers) == 0 || length(templates) == 0) {
        return(NULL)
    }
    run.names <- get.run.names(primers)
    max.mm <- max(unlist(lapply(primers, function(x) as.numeric(strsplit(c(x$Nbr_of_mismatches_fw, x$Nbr_of_mismatches_rev), split = ",")[[1]]))))
    cvg <- parallel::mclapply(seq_along(primers), function(x) {
        # supply allowed mismatch arg to show the same number of mismatches for all sets in the plot's facets
        plot.df <- prepare_template_cvg_mm_data(primers[[x]], templates[[x]],
                    allowed.mismatches = max.mm)
        if (length(plot.df) != 0) {
            # only select the expected coverage events
            plot.df <- plot.df[plot.df$Status == "Coverage",]
            # annotate with Run
            plot.df <- cbind("Run" = run.names[x], plot.df)
        }
    })
    plot.df <- do.call(rbind.fill, cvg)
    if (length(plot.df) == 0 || nrow(plot.df) == 0) {
        return(NULL)
    }
    title <- "Coverage of the template sequences"
    plot.df$Run <- abbreviate(plot.df$Run, getOption("openPrimeR.plot_abbrev"))
    plot.df$Run <- factor(plot.df$Run, levels = unique(plot.df$Run)[order(as.character(unique(plot.df$Run)))])
    if (length(highlight.set) != 0) {
        # check whether highlight set is specified correctly
        m <- match(highlight.set, plot.df$Run)
        na.idx <- which(is.na(m))
        if (length(na.idx) != 0) {
            msg <- paste("Highlight set not available in data:",
                paste(highlight.set[na.idx], collapse = ","))
            warning(msg)
            highlight.set <- highlight.set[!is.na(m)]
        }
    }
    pal <- getOption("openPrimeR.plot_colors")["Group"] # the RColorBrewer palette to use
    colors <- colorRampPalette(brewer.pal(8, pal))(length(unique(plot.df$Group)))
    p <- ggplot(plot.df, aes(x = .data[["Run"]], y = .data[["Coverage_Ratio"]], fill = .data[["Group"]], group = .data[["Group"]])) + 
        geom_bar(stat = "identity") + 
        facet_wrap(~Maximal_mismatches,
            labeller = label_bquote("Mismatches"<=.(substitute(Maximal_mismatches)))
        ) + 
        theme(axis.text.x = element_text(angle = 90, hjust = 1)) +
        scale_y_continuous(labels = scales::percent, limits = c(0, 1)) +
        ggtitle(title) +
        ylab("Coverage") +
        scale_fill_manual(values = colors)
    if (length(highlight.set) != 0) {
        # highlight selected sets
        sel <- levels(plot.df$Run) %in% highlight.set
        p <- p + theme(
                axis.text.x = element_text(
                    face=ifelse(sel, "bold","plain"),
                    colour = ifelse(sel, "grey20", "grey30")))
    }
    return(p)
}

#' @rdname AnalysisStats
#' @name AnalysisStats
#' @aliases get_cvg_stats
#' @return \code{get_cvg_stats} returns a data frame whose entries provide 
#' the coverage of templates per group of templates.
#' @export
#' @include primers.R templates.R
#' @examples
#' 
#' # Coverage statistics for a single primer set
#' data(Ippolito)
#' cvg.stats <- get_cvg_stats(primer.df, template.df)
#' # Coverage statistics for multiple primer sets
#' data(Comparison)
#' cvg.stats.comp <- get_cvg_stats(primer.data[1:2], template.data[1:2])
setGeneric("get_cvg_stats", 
    function(primers, templates, for.viewing = FALSE, total.percentages = FALSE, 
            allowed.mismatches = Inf, cvg.definition = c("constrained", "basic")) {
        standardGeneric("get_cvg_stats")
})
#' Coverage Statistics of a Primer Set.
#'
#' Retrieve statistics on the templates that are covered by a primer set.
#'
#' @param primer.df An object of class \code{Primers} containing
#' primers with evaluated coverage.
#' @param template.df An object of class \code{Templates} containing
#' templates with evaluated coverage.
#' @param for.viewing Whether the table should be formatted
#' for viewing rather than processing.
#' @param total.percentages Whether group coverage percentages
#' should relate to all template sequences or just those templates
#' belonging to a specific group.
#' @param allowed.mismatches The maximal allowed number of mismatches.
#' By default, the number of mismatches is not restricted. 
#' @param cvg.definition If \code{cvg.definition} is set to
#' "constrained", the statistics for the expected
#' coverage (after applying the coverage constraints) are retrieved.
#' If \code{cvg.definition} is set to "basic", the coverage is determined 
#' solely by string matching (i.e. without applying the coverage constraints).
#' By default, \code{cvg.definition} is set to "constrained".
#' @return Data frame with coverage statistics.
#' @keywords internal
setMethod("get_cvg_stats", signature(primers = "Primers"), 
    function(primers, templates, for.viewing = FALSE, total.percentages = FALSE, 
             allowed.mismatches = Inf, cvg.definition = c("constrained", "basic")) {

    # compute some stats on the coverage table structure: group | nbr templates |
    # [nbr covered | covered percent |]
    if (!is(primers, "Primers")) {
        stop("Please input a valid primer data frame.")
    }
    if (!is(templates, "Templates")) {
        stop("Please input a valid template data frame.")
    }
    if (!"Covered_Seqs" %in% colnames(primers) || !"primer_coverage" %in% colnames(primers)) {
        warning("get_cvg_stats: Primer/template coverage not available.")
        return(NULL)  # not possible
    }
    cvg.definition <- match.arg(cvg.definition)
    # update primer & template coverage according to the input conditions:
    primers <- update_primer_cvg(primers, templates, allowed.mismatches = allowed.mismatches, cvg.definition = cvg.definition)
    templates <- update_template_cvg(templates, primers)
    if (total.percentages) {
        # normalize coverage by total number of templates
        df <- ddply(templates, "Group", here(summarize), N = nrow(templates),
            N_primer = length(unique(strsplit(as.character(substitute(Covered_By_Primers)), split = ",")[[1]])), 
            primer_coverage = length(which(substitute(primer_coverage) > 0)), 
            Coverage_Ratio = sum(substitute(primer_coverage))/nrow(templates),
            N_primer_fw = length(unique(strsplit(as.character(substitute(Covered_By_Primers_fw)), split = ",")[[1]])), 
            primer_coverage_fw = length(which(substitute(primer_coverage_fw) > 0)), 
            Coverage_Ratio_fw = sum(substitute(primer_coverage_fw))/nrow(templates),
            N_primer_rev = length(unique(strsplit(as.character(substitute(Covered_By_Primers_rev)), split = ",")[[1]])), 
            primer_coverage_rev = length(which(substitute(primer_coverage_rev) > 0)), 
            Coverage_Ratio_rev = sum(substitute(primer_coverage_rev))/nrow(templates))
    } else {
        # normalize coverage by number of templates per group
        df <- ddply(templates, "Group", here(summarize), N = length(substitute(Group)), N_primer = length(unique(strsplit(as.character(substitute(Covered_By_Primers)), 
            split = ",")[[1]])), primer_coverage = length(which(substitute(primer_coverage) > 0)), 
            Coverage_Ratio = sum(substitute(primer_coverage))/length(substitute(Group)), N_primer_fw = length(unique(strsplit(as.character(substitute(Covered_By_Primers_fw)), 
                split = ",")[[1]])), primer_coverage_fw = length(which(substitute(primer_coverage_fw) > 
                0)), Coverage_Ratio_fw = sum(substitute(primer_coverage_fw))/length(substitute(Group)), 
            N_primer_rev = length(unique(strsplit(as.character(substitute(Covered_By_Primers_rev)), 
                split = ",")[[1]])), primer_coverage_rev = length(which(substitute(primer_coverage_rev) > 
                0)), Coverage_Ratio_rev = sum(substitute(primer_coverage_rev))/length(substitute(Group)))
    }
    # order by groups
    o <- order(df$Group)
    df <- df[o, ]
    # summary of all groups in 'Total' row
    df <- rbind(data.frame(Group = "Total", N = nrow(templates), N_primer = nrow(primers), 
        primer_coverage = sum(df$primer_coverage), Coverage_Ratio = sum(df$primer_coverage)/nrow(templates), 
        N_primer_fw = length(which(primers$Forward != "")), primer_coverage_fw = sum(df$primer_coverage_fw), 
        Coverage_Ratio_fw = sum(df$primer_coverage_fw)/nrow(templates), N_primer_rev = length(which(primers$Reverse != 
            "")), primer_coverage_rev = sum(df$primer_coverage_rev), Coverage_Ratio_rev = sum(df$primer_coverage_rev)/nrow(templates), 
        stringsAsFactors = FALSE), data.frame(df, stringsAsFactors = FALSE))
    df$Coverage <- paste(df$primer_coverage, " of ", df$N, " (", round(df$Coverage_Ratio * 
        100, 2), "%)", sep = "")
    df$Coverage_fw <- paste(df$primer_coverage_fw, " of ", df$N, " (", round(df$Coverage_Ratio_fw * 
        100, 2), "%)", sep = "")
    df$Coverage_rev <- paste(df$primer_coverage_rev, " of ", df$N, " (", round(df$Coverage_Ratio_rev * 
        100, 2), "%)", sep = "")
    out.df <- df
    if (for.viewing) {
        cols.both <- c("Group", "Coverage", 
                "Coverage_fw", "Coverage_rev")
        out.names <-  c("Group", "Coverage", "Coverage (fw)", "Coverage (rev)")
        mode <- "both"
        if (sum(out.df$N_primer_rev) == 0 || sum(out.df$N_primer_fw) == 0) {
            mode <- "single"
        }
        if (mode == "single") {
            out.df <- out.df[, cols.both[seq_len(2)]]
            colnames(out.df) <- out.names[seq_len(2)]
        } else {
            out.df <- out.df[, cols.both]
            colnames(out.df) <- out.names
        }
    }
    return(out.df)
})
#' Coverage Statistics for Multiple Primer Sets.
#'
#' Retrieve statistics on covered templates for multiple primer sets.
#'
#' @param primers A list with objects of class \code{Primers} containing
#' primers with evaluated coverage.
#' @param templates A list with objects of class \code{Templates} containing
#' templates with evaluated coverage.
#' @param for.viewing Whether the table should be formatted
#' for viewing rather than processing.
#' @param total.percentages Whether group coverage percentages
#' should relate to all template sequences or just those templates
#' belonging to a specific group.
#' @param allowed.mismatches The maximal allowed number of mismatches.
#' By default, the number of mismatches is not restricted. 
#' @param cvg.definition If \code{cvg.definition} is set to
#' "constrained", the statistics for the expected
#' coverage (after applying the coverage constraints) are retrieved.
#' If \code{cvg.definition} is set to "basic", the coverage is determined 
#' solely by string matching (i.e. without applying the coverage constraints).
#' By default, \code{cvg.definition} is set to "constrained".
#' @return Data frame with coverage statistics.
#' @keywords internal
setMethod("get_cvg_stats", signature(primers = "list"),
    function(primers, templates, for.viewing = FALSE, total.percentages = FALSE,
             allowed.mismatches = Inf, cvg.definition = c("constrained", "basic")) {


    if (length(primers) == 0 || length(templates) == 0) {
        return(NULL)
    }
    template.classes <- sapply(templates, function(x) class(x))
    primer.classes <- sapply(primers, function(x) class(x))
    if (any(template.classes != "Templates") || any(primer.classes != "Primers")) {
        stop("Check types of primers/templates.")
    }
    runs <- get.run.names(primers)
    stats <- lapply(seq_along(primers), function(x)  {
                df <- get_cvg_stats(primers[[x]], templates[[x]], FALSE, total.percentages) 
                groups <- as.character(df$Group)
                if (for.viewing) {
                    # only select the coverage entry
                    cvg <- df[, which(colnames(df) == "Coverage_Ratio") ]
                    # format cvg
                    cvg <- paste0(round(cvg * 100, 1), "%")
                    df <- data.frame(Run = runs[x], Coverage = t(cvg), stringsAsFactors = FALSE)
                } else {
                    df <- data.frame(cbind(Run = runs[x], t(df)), stringsAsFactors = FALSE)
                }
                colnames(df) <- c("Run", groups)
                return(df)
            })
    stat.df <- do.call(rbind.fill, stats)
    if (ncol(stat.df) >= 3) {
        # don't change Run & Total columns
        idx <- seq(3, ncol(stat.df))
        group.order <- colnames(stat.df)[idx]
        group.order <- group.order[order(group.order)]
        colnames(stat.df)[idx] <- group.order
    }
    # order rows by 'Run':
    stat.df <- stat.df[order(stat.df$Run), ]
    return(stat.df)
})
#' @rdname Plots
#' @name Plots
#' @aliases plot_primer_cvg
#' @return \code{plot_primer_cvg} creates a plot showing the coverage of individual primers.
#' @export
#' @include primers.R templates.R
#' @examples
#' 
#' # Plot expected coverage per primer
#' data(Ippolito)
#' p.cvg <- plot_primer_cvg(primer.df, template.df)
#' # Plot coverage stratified by allowed mismatches:
#' p.cvg.mm <- plot_primer_cvg(primer.df, template.df, per.mismatch = TRUE)
#' # Plot coverage of multiple primer sets
#' data(Comparison)
#' p.cvg.cmp <- plot_primer_cvg(primer.data[1:3], template.data[1:3])
setGeneric("plot_primer_cvg", 
    function(primers, templates, per.mismatch = FALSE, ...) {
        standardGeneric("plot_primer_cvg")
})
#' Plot Individual Primer Coverage.
#'
#' Shows which templates are covered by individual primers.
#'
#' @param p.df Primer data frame.
#' @param template.df Template data frame.
#' @param per.mismatch Whether to stratify by allowed mismatches.
#' @param excluded.seqs Identifiers of templates that should not be considered.
#' @param per.mismatch Whether the coverage should
#' be broken down for individual settings of allowed mismatches.
#' @return A bar plot showing the coverage of individual primers.
#' @keywords internal
setMethod("plot_primer_cvg", 
    signature(primers = "Primers", templates = "Templates"),
    function(primers, templates, per.mismatch = FALSE, groups = NULL) {
    
    if (length(primers) == 0 || nrow(primers) == 0 || length(templates) == 0) {
        return(NULL)
    }
    if (!is(primers, "Primers")) {
        stop("Please input a valid primer data frame.")
    }
    if (!is(templates, "Templates")) {
        stop("Please input a valid template data frame.")
    }
    if (per.mismatch) {
        p <- plot_primer_cvg_mismatches(primers, templates)
    } else {
        p <- plot_primer_cvg_unstratified(primers, templates, groups = groups)
    }
    return(p)
})
#' Plot Multiple Primer Coverages.
#'
#' Plots the coverage of individual primers for multiple sets.
#'
#' @param primers List with \code{Primers} objects.
#' @param templates List with \code{Templates} objects.
#' @param per.mismatch Whether the coverage should
#' be broken down for individual settings of allowed mismatches.
#' @return A bar plot showing the coverage of individual primers.
#' @keywords internal
setMethod("plot_primer_cvg", 
    signature(primers = "list", templates = "list"),
    function(primers, templates, per.mismatch = FALSE) {

    if (length(primers) == 0 || length(templates) == 0) {
        return(NULL)
    }
    if (per.mismatch) {
        warning("Not supported yet.")
        return(NULL)
    } else {
        # alias for existing exported function
        p <- plot_constraint(primers, NULL, "primer_coverage")
    }
    return(p)
})
#' Plot Individual Primer Coverage.
#'
#' Plots the coverage of individual primers.
#'
#' @param p.df Primer data frame.
#' @param template.df Template data frame.
#' @param groups Optional identifiers of template groups to be considered.
#' If not provided, all template groups are considered.
#' @return A bar plot showing the coverage of individual primers.
#' @keywords internal
plot_primer_cvg_unstratified <- function(p.df, template.df, groups = NULL) {
    # Select only the selected groups of templates:
     if (!is.null(groups) && !"all" %in% groups) { # select subset
        idx <- which(template.df$Group %in% groups)
        # set excluded seqs for updating the coverage
        excluded.seqs <- setdiff(template.df$Identifier[seq_len(nrow(template.df))], template.df$Identifier[idx])
        # select relevant templates
        template.df <- template.df[idx,]
        # re-evalaute coverage with the new templates
        p.df <- evaluate.diff.primer.cvg(p.df, excluded.seqs, template.df)
    }
    title <- "Template coverage per primer"
    xlab <- "Primer Identifier"
    title <- "Template coverage per primer and group"
    # annotate p.df with strat column
    m <- covered.seqs.to.idx(p.df$Covered_Seqs, template.df)
    plot.df <- NULL
    strat <- "Group"
    # TODO: maybe modify the code inline with prepare_mm_plot?
    for (i in seq_along(m)) {
        missing.groups <- unique(template.df$Group)
        new.data <- NULL
        if (length(m[[i]]) != 0) {
            hit.groups <- unique(template.df[m[[i]], strat])
            new.data <- data.frame(ID = rep(p.df[i, "ID"], length(hit.groups)))
            new.data[, strat] <- hit.groups
            cvd.idx <- m[[i]]
            cvg.idx <- lapply(seq_along(hit.groups), function(x) {
              s <- hit.groups[x]
              sel <- which(template.df[cvd.idx, strat] == s)
              cvd.idx[sel]
            })
            cvg.counts <- sapply(cvg.idx, length)
            cvd.seqs <- sapply(cvg.idx, function(x) paste(template.df$Identifier[x], 
              collapse = ","))
            available.seqs.per.group <- sapply(hit.groups, function(x) length(which(template.df$Group == x)))
            new.data$primer_coverage <- cvg.counts/available.seqs.per.group # percentage of group templates covered
            new.data$Covered_Seqs <- cvd.seqs
            missing.groups <- setdiff(unique(template.df$Group), new.data$Group)
        }
        # also add all combinations for missing groups -> ensures bar width stays constant throughout the plot observations
        missing.df <- data.frame(ID = rep(p.df$ID[i], length(missing.groups)), Group = missing.groups, 
                                primer_coverage = rep(0, length(missing.groups)), Covered_Seqs = rep("", length(missing.groups)))
        new.data <- rbind(new.data, missing.df)
        plot.df <- rbind(plot.df, new.data)
    }
    ###
    ylab <- "Covered Templates"
    pal <- getOption("openPrimeR.plot_colors")[strat] # the RColorBrewer palette to use
    group.colors <- colorRampPalette(brewer.pal(8, pal))(length(unique(plot.df[, strat])))
    plot.df$ID <- factor(plot.df$ID)
    unique.cvd.idx <- compute.unique.covered.idx(plot.df, template.df)
    unique.cvg <- sapply(unique.cvd.idx, length)
    unique.cvg.ratio <- unique.cvg/nrow(template.df)
    p.df.unique <- plot.df
    p.df.unique$Coverage_Type <- "Unique Coverage"
    plot.df$Coverage_Type <- "Overall Coverage"
    p.df.unique$primer_coverage <- unique.cvg.ratio
    plot.df <- rbind(plot.df, p.df.unique)
    levels(plot.df$ID) <- abbreviate(levels(plot.df$ID), getOption("openPrimeR.plot_abbrev"))
    plot.df <- plot.df[plot.df$Coverage_Type == "Overall Coverage",] # only plot overall cvg
    bar.width <- 0.75
    dist.buffer <- (1- 0.75) / 2
    # add shading for every primer region in the plot to differentiate them better
    base.pos <- seq_along(unique(plot.df$ID))
    starts <- base.pos - 0.5 + dist.buffer
    ends <- base.pos + 0.5 - dist.buffer
    rects <- data.frame(xstart = starts, xend = ends,
                        ymin = 0, ymax = 1)
    rects$col <- factor(rep(c(0,1), nrow(rects))[seq_len(nrow(rects))])
    # only retain rectangles alternately
    rects <- rects[rects$col == 0, ]
    p <- ggplot() +
        geom_bar(data = plot.df, stat = "identity", position = "dodge", 
                    aes(x = .data[["ID"]], y = .data[["primer_coverage"]],
                    fill = .data[[strat]]), width = bar.width, show.legend = TRUE) + 
        geom_rect(data = rects, aes(xmin = .data[["xstart"]], xmax = .data[["xend"]], ymin = .data[["ymin"]], ymax = .data[["ymax"]]), 
                  fill = "grey60", alpha = 0.15, inherit.aes = FALSE, show.legend = FALSE) +
        xlab(xlab) + ggtitle(title) + ylab(ylab) + 
        theme(axis.text.x = element_text(angle = 60, 
                                hjust = 1)) +
        scale_fill_manual(values = group.colors) + 
        scale_y_continuous(limits = c(0, 1), labels = scales::percent)
    return(p)
}

#' Data for Mismatch Primer Coverage Plot.
#'
#' Ensures that there's an entry for every possible mismatch setting.
#'
#' @param primer.df A \code{Primers} object.
#' @param template.df A \code{Templates} object.
#' @return A data frame for plotting mismatch primer coverage.
#' @keywords internal
get_primer_cvg_mm_plot_df <- function(primer.df, template.df) {
    full.df <- prepare_mm_plot(primer.df, template.df)
    full.df <- full.df[full.df$Coverage_Type == "constrained",]
    # go from individual mismatches to individual coverage events
    df <- ddply(full.df, c("Primer", "Template", "Group"), summarize,
                            Position = unique(substitute(Position_3terminus)), 
                            Number_of_mismatches = unique(substitute(Number_of_mismatches)))
    # for every primer and group, ensure that there's an entry for every number of mismatches:
    count.df <- ddply(df, c("Primer", "Group", "Number_of_mismatches"), summarize,
                            Coverage = length(unique(substitute(Template))))
    if (nrow(df) == 0) {
        # just plot a single facet to show nothing is covered
        max.mm <- 0 
    } else {
        max.mm <- max(df$Number_of_mismatches)
    }
    result.data <- vector("list", max.mm + 1)
    for (i in seq(0, max.mm)) {
        additional.df <- data.frame(Primer = unlist(lapply(seq_len(nrow(primer.df)), function(x) rep(primer.df$ID[x], length(unique(template.df$Group))))),
                                    Group = unlist(lapply(seq_len(nrow(primer.df)), function(x) unique(template.df$Group))),
                                    Number_of_mismatches = i, Coverage = 0)
        count.df <- merge(count.df, additional.df, all = TRUE)
    }
    # ensure that for duplicate entries (0 coverage entries), the one with the maximal coverage is chosen:
    count.df <- ddply(count.df, c("Primer", "Group", "Number_of_mismatches"), summarize,
                            Coverage = max(substitute(Coverage)))
    # make the counts cumulative
    count.df$Cumulative_Coverage <- ave(count.df$Coverage, count.df$Primer, count.df$Group, FUN = cumsum)
    # determine ratio of covered seqs per group
    group.counts <- table(template.df$Group) # match to count.df$Group for division
    m <- match(count.df$Group, names(group.counts))
    count.df$Coverage_Ratio <- count.df$Cumulative_Coverage / as.vector(group.counts[m])
    o <- order(unique(as.character(count.df$Group)))
    count.df$Group <- factor(count.df$Group, levels = unique(as.character(count.df$Group))[o])
    colnames(count.df)[colnames(count.df) == "Number_of_mismatches"] <- "Maximal_mismatches" 
    count.df$Primer <- factor(abbreviate(count.df$Primer, getOption("openPrimeR.plot_abbrev")), levels = abbreviate(levels(count.df$Primer), getOption("openPrimeR.plot_abbrev"))) # shorter identifiers
    return(count.df)
}
#' Plot of Individual Primer Coverage and Mismatches.
#'
#' Plots the coverage of individual primers for different mismatch settings.
#'
#' @param primer.df A \code{Primers} object.
#' @param template.df A \code{Templates} object.
#' @param groups Optional identifiers of template groups to be considered.
#' If not provided, all template groups are considered.
#' @param nfacets A numeric providing the number of facet columns to use. 
#' By default, \code{nfacets} is set to \code{NULL} such that a
#' suitable number of columns is chosen automatically.
#' @return A bar plot showing the coverage of individual primers for different mismatch settings.
#' @keywords internal
plot_primer_cvg_mismatches <- function(primer.df, template.df, groups = NULL,
                                       nfacets = NULL) {
    # select template group subset
    if (!is.null(groups) && !"all" %in% groups) { # select subset
        idx <- which(template.df$Group %in% groups)
        template.df <- template.df[idx,]
    }
    # retrieve cvg stats of each primer:
    count.df <- get_primer_cvg_mm_plot_df(primer.df, template.df)
    pal <- getOption("openPrimeR.plot_colors")["Group"] # the RColorBrewer palette to use
    group.colors <- colorRampPalette(brewer.pal(8, pal))(length(unique(count.df[, "Group"])))
    p <- ggplot(count.df) + 
        geom_bar(width = 0.75, position = "stack", stat = "identity",
            aes(.data[["Primer"]], 
                .data[["Cumulative_Coverage"]],
                fill = .data[["Group"]])) + 
        xlab("Primer") + ggtitle("Mismatch primer coverage") + 
        ylab("Number of covered templates") + 
        theme(axis.text.x = element_text(
                angle = 90,  # angle @ 90 to prevent overplotting when facetting
                hjust = 1)) +
        scale_fill_manual(values = group.colors) + 
        facet_wrap(~Maximal_mismatches, ncol = nfacets,
            labeller = label_bquote("Mismatches"<=.(substitute(Maximal_mismatches))))
    return(p)
}

#' Plot Filtering Runtimes
#'
#' Plots the runtimes of individual evaluation steps in the filtering procedure.
#'
#' @param filered.stats Stats from filtering.
#' @return A plot of the runtime for each filtering step.
#' @keywords internal
plot.filtering.runtime <- function(filtered.stats) {
    plot.stats <- filtered.stats[which(filtered.stats$Constraint != ""), ]  # make sure that we don't have duplicates when we form levels if filtering stopped early
    plot.stats$Constraint <- factor(plot.stats$Constraint, levels = rev(unique(plot.stats$Constraint)))
    constraint.names <- unlist(constraints_to_unit(levels(plot.stats$Constraint), FALSE))
    if (nrow(filtered.stats) == 0) {
        return(NULL)
    }
    ggplot(plot.stats, aes(y = .data[["Time"]], x = .data[["Constraint"]], fill = .data[["Time"]])) + geom_bar(stat = "identity") + 
        coord_flip() + xlab("Applied constraint") + ylab("Required computation time [s]") + 
        ggtitle("Runtime for each filtering step") + 
        theme(axis.text.x = element_text(hjust = 1)) +
        scale_fill_gradient() + 
        facet_wrap(~Direction) +
        scale_x_discrete(labels = constraint.names)
}
#' Plot of Filtering Stats for Coverage.
#'
#' Plots the remaining coverage after each filtering step.
#'
#' @param stats Statistics of the filtering procedure.
#' @param stats.relax Statistic of the relaxation procedure.
#' @return A plot showing the possible coverage after each filtering step.
#' @keywords internal
plot.filtering.stats.cvg <- function(stats, stats.relax = NULL) {
    plot.stats <- stats_plot_data(stats, stats.relax)
    if (length(plot.stats) == 0 || nrow(plot.stats) == 0) {
        return(NULL)
    }
    if (all(is.na(plot.stats$Current_Coverage))) { # no coverage available
        return(NULL)
    }
    yl <- "Primer coverage after filtering"
    ttl <- "Filtering coverage"
    filter.color <- brewer.pal(8, "Accent")[5]
    relax.color <- brewer.pal(8, "Blues")[4]
    color <- c("Filtering" = filter.color, "Relaxation" = relax.color)

    constraint.names <- unlist(constraints_to_unit(levels(plot.stats$Constraint), FALSE))
    ggplot(plot.stats, aes(y = .data[["Current_Coverage"]], x = .data[["Constraint"]], fill = .data[["Type"]])) + 
        geom_bar(stat = "identity") + coord_flip() + xlab("Applied constraint") + 
        ylab(yl) + ggtitle(ttl) + 
        scale_y_continuous(labels = scales::percent, limits = c(0, 1)) +
        theme(axis.text.x = element_text(hjust = 1)) +
        scale_fill_manual(values = color) + 
        facet_wrap(~Direction) + 
        scale_x_discrete(labels = constraint.names)
}
#' Combination of Filtering Stats.
#'
#' Summarizes filtering/relaxation statistics for plotting.
#'
#' @param stats Statistics of the filtering procedure.
#' @param stats.relax Statistic of the relaxation procedure.
#' @return A data frame combinin filtering/relaxation stats.
#' @keywords internal
stats_plot_data <- function(stats, stats.relax) {
    if (length(stats) == 0 && nrow(stats) == 0) {
        return(NULL)
    }
    plot.stats <- stats
    plot.stats$Constraint <- factor(plot.stats$Constraint, levels = rev(unique(plot.stats$Constraint)))
    ratios <- ddply(plot.stats, "Direction", summarize, 
                Ratio = substitute(Remaining) / (substitute(Remaining)[1] + substitute(Excluded)[1]))$Ratio
    plot.stats$Ratio <- ratios
    plot.stats$Type <- "Filtering"
    if (length(stats.relax) != 0 && nrow(stats.relax) != 0) {
        # augment with relaxation info
        plot.relax <- stats.relax
        plot.relax$Type <- "Relaxation"
        # don't take first element for reverse .. split here
        fw.idx <- plot.stats$Direction == "fw"
        fw.count <- plot.stats$Remaining[fw.idx][1] + plot.stats$Excluded[fw.idx][1]
        rev.idx <- plot.stats$Direction == "rev"
        rev.count <- plot.stats$Remaining[rev.idx][1] + plot.stats$Excluded[rev.idx][1]
        fw.ratio <- plot.relax$Remaining[plot.relax$Direction == "fw"] / fw.count
        rev.ratio <- plot.relax$Remaining[plot.relax$Direction == "rev"] / fw.count
        plot.relax$Ratio <- c(fw.ratio, rev.ratio)
        # adjust actual coverage using the relaxation info
        plot.relax$Current_Coverage[is.na(plot.relax$Current_Coverage)] <- 0
        cum.sums <- ddply(plot.relax, "Direction", summarize, 
                Cumulative_Coverage = cumsum(substitute(Current_Coverage)))
        plot.relax$Current_Coverage <- cum.sums$Cumulative_Coverage
        idx <- which(!is.na(plot.relax$Current_Coverage))
        if (length(idx) != 0) {
            plot.stats$Current_Coverage[idx] <- sapply(seq_along(idx), function(x) max(0,plot.stats$Current_Coverage[idx[x]] - plot.relax$Current_Coverage[idx[x]]))
        }
        plot.stats <- rbind(plot.stats, plot.relax)
        plot.stats$Type <- factor(plot.stats$Type, levels = c("Relaxation", "Filtering"))
    } 
    return(plot.stats)
}
#' Plot of Overall Filtering Stats.
#' 
#' Plots the number of primers remaining after each filtering step.
#'
#' @param stats Statistics on the filtering procedure
#' @param stats.relax Statistic on the filtering procedure after relaxation.
#' @return A plot for the number of primers after filtering. 
#' @keywords internal
plot.filtering.stats <- function(stats, stats.relax = NULL) {
    plot.stats <- stats_plot_data(stats, stats.relax)
    if (length(plot.stats) == 0 || nrow(plot.stats) == 0) {
        return(NULL)
    }
    constraint.names <- unlist(constraints_to_unit(levels(plot.stats$Constraint), FALSE))
    filter.color <- brewer.pal(8, "Accent")[5]
    relax.color <- brewer.pal(8, "Blues")[4]
    color <- c("Filtering" = filter.color, "Relaxation" = relax.color)
    ggplot(plot.stats, aes(y = .data[["Ratio"]], x = .data[["Constraint"]], fill = .data[["Type"]])) + 
        geom_bar(stat = "identity") + coord_flip() + xlab("Filtering constraint") + 
        ylab("Remaining primers") + 
        ggtitle("Filtering overview") + 
        scale_y_continuous(labels = scales::percent, limits = c(0, 1)) +
        scale_fill_manual(values = color) + 
        theme(axis.text.x = element_text(hjust = 1)) +
        facet_wrap(~Direction) + 
        scale_x_discrete(labels = constraint.names)
}

###############
# Greedy optimization of primer sets
#################

#' Reorder Primers 
#'
#' Reorders a primer set according to the IDs of primers.
#'
#' @param filtered.primers Primer data frame.
#' @param primer.ID.order new ordering of IDs in the data frame.
#' @return Reordered primer data frame.
#' @keywords internal
reorder.primer.table <- function(filtered.primers, primer.ID.order) {
    # reorders the input primers table according to the given ID order
    # filtered.primers: input primer df primer.ID.order: new order of ID column in
    # primer.df
    if (length(primer.ID.order) == 0) {
        return(filtered.primers)
    }
    filtered.primers$ID <- factor(filtered.primers$ID, levels = primer.ID.order)
    o <- order(filtered.primers$ID)
    filtered.primers <- filtered.primers[o, ]
    return(filtered.primers)
}
#' Greedy Choice
#'
#' Selects the currently best primer for Greedy primer selection.
#'
#' @param result Data frame of current optimized primer data set that is to be augmented.
#' @param primers Data frame of candidate primers for addition to \code{result}.
#' @param deltaG.cutoff Free energy cutoff for cross-dimerization.
#' @param target.temp Target annealing temperature in Celsius.
#' @param primer_conc Primer concentration.
#' @param na_salt_conc Sodium ion concentration.
#' @param mg_salt_conc Magensium ion concentration.
#' @param k_salt_conc Potassium ion concentration.
#' @param tris_salt_conc Tris ion concentration.
#' @return The index of a suitable primer according to Greedy selection.
#' @keywords internal
select.best.primer.idx <- function(result, primers, deltaG.cutoff, target.temp,
    primer_conc, na_salt_conc, mg_salt_conc, k_salt_conc, tris_salt_conc)  {

    if (length(primers) == 0 || nrow(primers) == 0) {
        return(NULL)
    }
    candidates <- NULL  # idx of all selection candidates
    bad.primer.df <- data.frame(Identifier = numeric(0), Index = numeric(0), melting_temp = numeric(0), 
        deltaG_dimer = numeric(0))
    for (idx in seq_len(nrow(primers))) {
        # check whether constraints are fulfilled by selection candidates best primers
        # already found, check if cur primer cvg is the same as the best primer cvg of
        # the current primer is not as high as the selected ones -> don't consider for
        # inclusion
        if (length(candidates) != 0) {
            if (primers[idx, "primer_coverage"] < primers[candidates[1], "primer_coverage"]) {
                break
            }
        }
        if (!is.na(deltaG.cutoff) && length(result) != 0 && nrow(result) != 0) {
            # check for cross-dimerization between current primers and candidate primer
            tmp.result <- suppressWarnings(my_rbind(result, primers[idx, ]))
            comp <- compute.all.cross.dimers(tmp.result, primer_conc, na_salt_conc, 
                    mg_salt_conc, k_salt_conc, tris_salt_conc, 
                    target.temp, check.idx = nrow(tmp.result), 
                    no.structures = TRUE)
            if (any(comp$DeltaG < deltaG.cutoff)) {
                # any deltaG exceeds the allowed cutoff -> don't select this primer
                min.deltaG <- min(comp$DeltaG)
                #message(paste("bad primer id: ", primers[idx, "Identifier"], ", dimerization deltaG: ", 
                  #min.deltaG))
                bad.primer <- data.frame(Identifier = primers[idx, "Identifier"], 
                  Index = idx, melting_temp = NA, deltaG_dimer = min.deltaG)
                bad.primer.df <- rbind(bad.primer.df, bad.primer)
            } else {
                candidates <- c(candidates, idx)
            }
        } else {
            # if there's no results yet, the first primer is automatically a candidate ..
            candidates <- c(candidates, idx)
        }
    }
    if (length(candidates) != 0) {
        # select the best primer from the candidates (all have the same cvg) by
        # resolution through primer efficiency (if available) select primer by highest
        # primer efficiency
        if ("primer_efficiency" %in% colnames(primers)) {
            o <- order(primers[candidates, "mean_primer_efficiency"], decreasing = TRUE)
            idx <- candidates[o][1]
        } else {
            idx <- candidates[1]
        }
    }
    res <- list(sel_idx = idx, filtered_info = bad.primer.df)
    return(res)
}
#' Filter by Melting Temperature Difference
#'
#' Filters primers by melting temperature differences.
#'
#' @param target.temp Target melting temperature in Celsius.
#' @param selected.primers Current candidate primer data frame.
#' @param max.Tm.delta Maximum allowed difference of primer melting temperatures to target temperature.
#' @return Filtered primer data frame.
#' @keywords internal
filter_primers.by.Tm.delta <- function(target.temp, selected.primers, max.Tm.delta) {
    # target.temp: target melting temperature of primers in the optimization set
    # selected.primers: current candidates for selection max.Tm.delta: maximal
    # allowed difference of primer melting temperature to target temperature
    r <- selected.primers
    sel.idx <- NULL
    filtered.df <- data.frame(Identifier = numeric(0), Index = numeric(0), melting_temp = numeric(0), 
        deltaG_dimer = numeric(0))  # excluded primers
    if (!is.na(max.Tm.delta)) {
        Tm.diff <- abs(selected.primers$melting_temp - target.temp)
        sel.idx <- which(Tm.diff > max.Tm.delta)
        if (length(sel.idx) != 0) {
            f.df <- data.frame(Identifier = selected.primers$Identifier[sel.idx], 
                Index = sel.idx, melting_temp = Tm.diff[sel.idx], deltaG_dimer = NA)
            filtered.df <- rbind(filtered.df, f.df)
            r <- selected.primers[-sel.idx, ]
        }
    }
    out <- list(data = r, filtered = Primers(filtered.df))  # data: selected primers, filtered: excluded primers. melting temp entry of filtered: refers to DeltaTm!
    return(out)
}
#' Check of Primer and Template Correspondence.
#'
#' Checks whether the primers relate to the correct templates.
#'
#' @param primer.df An object of class \code{Primers}.
#' @param template.df An object of class \code{Templates}.
#' @return \code{TRUE} if the primers and templates seem to correspond,
#' \code{FALSE} otherwise.
#' @keywords internal
check_correspondence <- function(primer.df, template.df) {
    # check whether primers and templates correspond to one another
    cvd.seqs <- unlist(sapply(strsplit(primer.df$Covered_Seqs, split = ","), function(x) as.numeric(x)))
    if (length(cvd.seqs) == 0) {
        # no covered seqs -> always corresponds
        return(TRUE)
    }
    cvd.seqs <- cvd.seqs[!is.na(cvd.seqs)] # exclude primers for which cvg is not available.
    m <- match(cvd.seqs, template.df$Identifier)
    if (any(is.na(m))) {
        return(FALSE)
        msg <- paste("Could not match all coverage events to the input template data frame.",
            "Please check whether the primers and templates correspond.")
        stop(msg)
    } else {
        return(TRUE)
    }
}
#' Greedy Optimization
#'
#' Greedy approach for solving the primer set coverage problem.
#'
#' @param primers Primer data frame to be optimized.
#' @param template.df Template data frame.
#' @param mode.directionality Primer direction.
#' @param cur.opti.constraints List with optimization constraint settings.
#' @param target.temp Target annealing temperature of the optimized primer set in Celsius.
#' @param allowed.mismatches The number of mismatches primers are allowed to have with the templates.
#' @param opti.limits List with optimization limits.
#' @param primer_conc Primer concentration.
#' @param template_conc Template concentration.
#' @param na_salt_conc Sodium ion concentration.
#' @param mg_salt_conc Magensium ion concentration.
#' @param k_salt_conc Potassium ion concentration.
#' @param tris_salt_conc Tris ion concentration.
#' @param updateProgress Shiny progress callback function.
#' @return List with optimization data.
#' @keywords internal
optimize.primer.cvg <- function(primers, template.df, mode.directionality, 
    cur.opti.constraints, target.temp, allowed.mismatches, opti.limits, primer_conc, 
    na_salt_conc, mg_salt_conc, k_salt_conc, tris_salt_conc, updateProgress = NULL) {

    #print("optimizing primer cvg")
    deltaG.cutoff <- NA
    if ("cross_dimerization" %in% names(cur.opti.constraints)) {
        deltaG.cutoff <- cur.opti.constraints[["cross_dimerization"]]["min"]
    }
    l.df <- template.df
    max.cvg <- get_cvg_ratio(primers, template.df)
    cur.cvg <- 0
    it <- 0
    result <- NULL # not initialized here due to Primers class.
    all.bad.primers <- data.frame(Identifier = numeric(0), Index = numeric(0), 
                        melting_temp = numeric(0), deltaG_dimer = numeric(0))
    # order primers by cvg for greedy selection
    selected.primers <- primers[order(primers$primer_coverage, decreasing = TRUE), ]  
    while ((cur.cvg != max.cvg) && nrow(selected.primers) >= 1) {
        it <- it + 1
        #message(it)
        best.primer.data <- select.best.primer.idx(result, selected.primers, 
            deltaG.cutoff, target.temp, primer_conc, 
            na_salt_conc, mg_salt_conc, k_salt_conc, tris_salt_conc)
        if (length(best.primer.data) == 0 || is.na(best.primer.data$sel_idx)) {
            # no primer could be selected..
            selected.primers <- primers[FALSE, ]
            break
        } else {
            best.primer.idx <- best.primer.data$sel_idx
            bad.primer.df <- best.primer.data$filtered_info  # e.g. cross-dimerizing primers
            all.bad.primers <- rbind(all.bad.primers, bad.primer.df)
            best.covered <- as.numeric(unlist(strsplit(selected.primers[best.primer.idx, "Covered_Seqs"], split = ",")))
            if (length(best.covered) == 0) {
                # cannot improve anymore: best (highest cvg primer) didn't cover anything
                break
            }
            # selected.primers gives only the coverage count for sequences that haven't been covered yet :-)
            cumulative.coverage <- sum(result$Marginal_Coverage_Gain) + selected.primers[best.primer.idx, "primer_coverage"]  
            # compute current coverage
            cur.cvg <- cumulative.coverage/nrow(template.df)
            # use primer entry from original table to have the 'true' covered seqs in the
            # result, not only the marginal gain (primer_coverage is not the marginal gain!!)
            primer.entry <- primers[match(selected.primers[best.primer.idx, "Identifier"], primers$Identifier), ]
            entry <- data.frame(Iteration = it, 
                                Marginal_Coverage_Gain = selected.primers[best.primer.idx, "primer_coverage"], 
                                Cumulative_Coverage = cumulative.coverage, Cumulative_Coverage_Ratio = cur.cvg, 
                                primer.entry, stringsAsFactors = FALSE)
            result <- rbind(result, entry)
            rm.idx <- c(bad.primer.df$Index, best.primer.idx)
            selected.primers <- selected.primers[-rm.idx, ]  # remove the selected primer and selection candidates that were filtered this round
            # re-compute coverage by removing templates that are covered by newly added primer (best.covered)
            selected.primers <- evaluate.diff.primer.cvg(selected.primers, best.covered, template.df)  
        }
    }
    # reset cvg to consistent values from before (not marginal gain anymore):
    if (length(result) != 0) {
        m <- match(result$Identifier, primers$Identifier)
        result[, colnames(primers)] <- primers[m, ]
        # remove redundant primers
        cvg.matrix <- get.coverage.matrix(primers, template.df)
        primer.idx <- remove.redundant.cols(m, cvg.matrix)  # updated primer.idx with unnecessary primers removed
        result <- result[match(primers$Identifier[primer.idx], result$Identifier), , drop = FALSE]
    } else {
        result <- primers # could not optimize anything
    }
    out <- list(data = Primers(result), bad_primers = all.bad.primers)
    return(out)
}
#' Greedy Optimization.
#'
#' Greedy approach for solving the primer set coverage problem.
#'
#' @param primers Primer data frame to be optimized.
#' @param settings A \code{DesignSettings} object.
#' @param template.df Template data frame.
#' @param mode.directionality Primer direction.
#' @param required.cvg Target coverage ratio.
#' @param allowed.mismatches The number of mismatches primers are allowed to have with the templates.
#' @param primer_conc Primer concentration.
#' @param na_salt_conc Sodium ion concentration.
#' @param mg_salt_conc Magensium ion concentration.
#' @param k_salt_conc Potassium ion concentration.
#' @param tris_salt_conc Tris ion concentration.
#' @param template_conc Template data frame.
#' @param allowed.other.binding.ratio Ratio of primers allowed to bind to non-target regions.
#' @param allowed.stop.codons  Consider mismatch binding events that induce stop codons.
#' @param allowed.region.definition Definition of the target binding sites used for evaluating the coverage.
#' If \code{allowed.region.definition} is \code{within}, primers have to lie within the allowed binding region.
#' If \code{allowed.region.definition} is \code{any}, primers have to overlap with the allowed binding region.
#' The default is that primers have to bind within the target binding region.
#' @param disallowed.mismatch.pos The number of positions from the primer 3' end where mismatches should not be allowed.
#' All primers binding templates with mismatches within \code{disallowed.mismatch.pos} from the 3' end are disregarded.
#' @param target.temps Target melting temperatures for optimized sets in Celsius.
#' @param fw.primers List with already optimized primer data frames corresponding to \code{target.temps}.
#' @param updateProgress Shiny progress callback function.
#' @return List with optimization data.
#' @keywords internal
select.primers.by.cvg <- function(primers, settings, template.df, mode.directionality = c("fw", "rev"), 
    required.cvg = 1, allowed.mismatches, primer_conc, na_salt_conc, mg_salt_conc, 
    k_salt_conc, tris_salt_conc, template_conc, allowed.other.binding.ratio,
    allowed.stop.codons, allowed.region.definition = c("within", "any"), 
    disallowed.mismatch.pos, target.temps = NULL, fw.primers = NULL, updateProgress = NULL) {

    if (length(mode.directionality) == 0) {
        stop("Please supply the 'mode.directionality' arg.")
    }
    mode.directionality <- match.arg(mode.directionality)
    if (length(allowed.region.definition) == 0) {
        stop("Please supply the 'allowed.region.definition' arg.")
    }
    allowed.region.definition <- match.arg(allowed.region.definition)
    opti.constraints <- opti(settings)
    opti.limits <- optiLimits(settings)
    if (length(primers) == 0) {
        return(NULL)
    } else if (nrow(primers) == 0) {
        # nothing to optimize
        return(list(opti = primers, all_results = list(primers), all_used_constraints = list(settings), 
            used_constraints = settings))
    }
    # add primer cvg if not present yet: needed for optimization
    if (!"primer_coverage" %in% colnames(primers)) {
        stop("Optimization requires pre-computed primer coverage")
    }
    # determine computation constraints:
    Tm.brackets <- create.Tm.brackets(primers, template.df, settings, target.temps)
    primers <- Tm.brackets$primers
    Tm.groups <- Tm.brackets$group_idx
    Tm.group.df <- Tm.brackets$df
    opti.results <- vector("list", nrow(Tm.group.df))
    # 1. precompute constraints for all temperatures
    diagnostic.location <- NULL
    Tm.data <- compute.Tm.sets(primers, template.df, Tm.brackets, settings, mode.directionality, 
        primer_conc, template_conc, na_salt_conc, mg_salt_conc, k_salt_conc, tris_salt_conc, 
        allowed.mismatches, allowed.other.binding.ratio, 
        allowed.stop.codons, allowed.region.definition, disallowed.mismatch.pos, 
        TRUE, required.cvg,
        fw.primers, diagnostic.location, updateProgress)
    Tm.sets <- Tm.data$sets
    Tm.settings <- Tm.data$settings
    #############
    # 2. optimize all temperature primer sets
    Tm.group.idx <- NULL
    max.nbr.relaxations <- 3
    opti.results.data <- vector("list", length(Tm.sets)) # for debug
    for (Tm.group.idx in seq_along(Tm.sets)) { # for debug only
    #opti.results.data <- foreach(Tm.group.idx = seq_along(Tm.sets), .combine = c) %dopar% {
        Tm.set <- Tm.sets[[Tm.group.idx]]
        cur.settings <- Tm.settings[[Tm.group.idx]]
        if (length(Tm.set) == 0 || nrow(Tm.set) == 0) {
            # no Tm set -> nothing to optimize. this shouldn't really happen except if there's no data.
            Tm.out <- list(list(data = Tm.set, used_constraints = opti.constraints))
        } else {
            #message(paste("Nbr of primers: ", dim(Tm.set)[1], sep = ""))
            target.temp <- Tm.group.df$target_Tm[Tm.group.idx]
            annealing.temp <- Tm.group.df$annealing_temp[Tm.group.idx]
            if (is.function(updateProgress)) {
                detail <- paste("Tm group: ", Tm.group.idx, sep = "")
                updateProgress(1/length(opti.results), detail, "inc")
            }
            #message(paste("Optimizing for Tm-group: ", Tm.group.idx, "/", nrow(Tm.group.df), 
            #    sep = ""))
            target.cvg <- min(get_cvg_ratio(Tm.set, template.df), required.cvg)
            result.data <- optimize.primer.cvg(Tm.set, template.df, mode.directionality, 
                opti(cur.settings), annealing.temp, allowed.mismatches, 
                opti.limits, primer_conc, na_salt_conc, mg_salt_conc, k_salt_conc, 
                tris_salt_conc, updateProgress)
            result <- result.data$data
            result.cvg <- get_cvg_ratio(result, template.df)
            # relaxation
            while (result.cvg < target.cvg && "cross_dimerization" %in% names(opti(cur.settings))) {
                # don't lose any coverage here in the optimization ...
                # adjust DeltaG
                message("Greedy deltaG relaxation. Current coverage: ", result.cvg, 
                            "Target coverage: ", target.cvg, 
                            ". Old deltaG: ", constraints(cur.settings)$cross_dimerization["min"])
                constraints(cur.settings)$cross_dimerization <- relax.opti.constraints(opti(cur.settings), optiLimits(Tm.settings[[Tm.group.idx]]), 
                                                                                      opti(Tm.settings[[Tm.group.idx]]))$cross_dimerization
                constraintLimits(cur.settings)$cross_dimerization <- set.new.limits(optiLimits(cur.settings), optiLimits(Tm.settings[[Tm.group.idx]]), 
                                                                                      opti(Tm.settings[[Tm.group.idx]]))$cross_dimerization
                result.data <- optimize.primer.cvg(Tm.set, template.df, mode.directionality, 
                    opti(cur.settings), annealing.temp, allowed.mismatches, 
                    opti.limits, primer_conc, na_salt_conc, mg_salt_conc, k_salt_conc, 
                    tris_salt_conc,  updateProgress)
               result <- result.data$data
               result.cvg <- get_cvg_ratio(result, template.df)
               message("-> New coverage: ", result.cvg, 
                            ". New deltaG: ", constraints(cur.settings)$cross_dimerization["min"])
            }
            Tm.out <- list(list(data = result.data$data, used_constraints = cur.settings))
            opti.results.data[[Tm.group.idx]] <- Tm.out[[1]] # for debug only
        }
    }
    opti.results <- lapply(opti.results.data, function(x) x$data)  # extract optimized sets
    names(opti.results) <- Tm.group.df$target_Tm # uses the target Tm's before relaxation
    # remove empty results
    opti.results <- opti.results[lengths(opti.results) != 0]

    best.idx <- select.best.opti.result(opti.results, template.df)
    if (length(best.idx) == 0) {
        # could not determine a solution, arbitrarily return the first opti.result
        best.idx <- 1
    }
    opti.result <- opti.results[[best.idx]]
    all.used.constraints <- lapply(opti.results.data, function(x) x$used_constraints)  # all used settings
    result <- list(opti = opti.result, 
        used_constraints = opti.results.data[[best.idx]]$used_constraints, 
        all_used_constraints = all.used.constraints, 
        all_results = opti.results)
    #message("optimization w/ greedy finished")
    return(result)
}
#' Selection of Best Greedy Result
#'
#' Selects best greedy primer data set.
#'
#' @param opti.results List with primer data frames for different target melting temperatures.
#' @param template.df Template data frame.
#' @return The index of the best primer data set.
#' @keywords internal
select.best.opti.result <- function(opti.results, template.df) {
    # could have a single function for ILP and greedy result selection?
    cvg.per.group <- sapply(opti.results, function(x) if (length(x) != 0 && nrow(x) != 
        0) {
        max(x$Cumulative_Coverage)
    } else {
        0
    })
    if (all(cvg.per.group == 0)) {
        message("Could not find a solution with primers covering any of the templates.")
        if (length(opti.results) > 0) {
            return(1)
        } else {
            return(NULL)
        }
    }
    cvg.ratios <- sapply(cvg.per.group, function(x) x/nrow(template.df))
    #message("Cumulative coverage per group: ")
    #message(cvg.per.group)
    #message("Coverage ratios: ")
    #message(cvg.ratios)
    # TIE BREAKING: at the moment: just use primer_efficiency: later: score primer
    # set using all available parameters
    eps <- 0.01  # if the cvg difference is just 1%, choose the set with better efficiency
    m.cvg <- max(cvg.per.group, na.rm = TRUE)
    max.idx <- which(cvg.per.group == m.cvg)[1]
    sel.idx <- which((cvg.ratios[max.idx] - cvg.ratios) < eps)
    idx <- NULL
    if ("primer_efficiency" %in% colnames(opti.results[[1]])) {
        eff <- unlist(lapply(opti.results[sel.idx], function(x) mean(x$mean_primer_efficiency, 
            na.rm = TRUE)))
        # message(opti.results[[1]]$primer_efficiency)
        #message(paste("Primer set mean efficiencies:", eff, sep = ""))
        idx <- sel.idx[which.max(eff)]
    } else {
        # no primer efficiency available -> just select the max
        idx <- max.idx[1]
    }
    #message(paste("Selected primer set: ", idx, sep = ""))
    return(idx)
}

############
# Primer optimization algorithms
###########

#' File Name for Initialized Primers.
#'
#' Constructs a filename for initialized primers.
#'
#' @param cur.results.loc Directory where the file should be stored.
#' @param GROUP Sample name of templates.
#' @param primer.lengths Interval of desired primer lengths.
#' @param mode.directionality Directionality of the primers
#' @param allowed.region.definition Definition of the allowed region.
#' @param init.algo Initialization algorithm identifier.
#' @param max.degen Maximum degeneracy of primers.
#' @param conservation Required ratio of primer conservation.
#' @return A filename for the initialized primers.
#' @keywords internal
get.init.file.name <- function(cur.results.loc, GROUP, primer.lengths, mode.directionality, 
    allowed.region.definition, init.algo, max.degen, conservation) {
    # get the filename for the initial primer set
    if (length(cur.results.loc) == 0) {
        return(NULL)
    }
    algo.info <- ""
    if (init.algo == "tree") {
        algo.info <- paste(algo.info, "degen=", max.degen, "_con=", conservation, 
            sep = "")
    }
    path <- file.path(cur.results.loc, paste(GROUP, "_initial_primers_", init.algo, 
        "_", primer.lengths[1], "_", primer.lengths[2], "_", algo.info, mode.directionality, 
        ".csv", sep = ""))
    return(path)
}

#' Write Out Optimization Data
#'
#' Writes out all data relating to the optimization of primers.
#'
#' @param opti.results.loc Folder where optimization data reside.
#' @param optimal.primers.data List with optimization results.
#' @param mode.directionality Direction of primers.
#' @param settings Settings used in the optimization procedure.
#' List containing fw, rev settings.
#' @param sample.name Name of template sample.
#' @param template.df Template data frame.
#' @param max.degen Maximal degeneracy of primers.
#' @return Write-out of primer information to \code{opti.results.loc}.
#' @keywords internal
write.out.primer.info <- function(opti.results.loc, optimal.primers.data, 
    mode.directionality, settings, sample.name, template.df, max.degen) {
    if (length(opti.results.loc) != 0) {
        optimal.primers <- optimal.primers.data$opti
        all.opti.results <- optimal.primers.data$all_results
        for (i in seq_along(all.opti.results)) {
            write.csv(all.opti.results[[i]], file = file.path(opti.results.loc, paste(sample.name, 
                "_optimized_set_target_temp_", names(all.opti.results)[i], ".csv", 
                sep = "")), row.names = FALSE)
        }
        template.df <- update_template_cvg(template.df, optimal.primers)
        write.csv(template.df, file = file.path(opti.results.loc, paste(sample.name, "_optimized_templates.csv", 
            sep = "")), row.names = FALSE)
        # save constraints:
        for (i in seq_along(settings)) {
            setting <- settings[[i]]
            write_settings(setting, file.path(opti.results.loc, paste0(sample.name, 
                           "_constraint_settings_", names(settings)[i], ".xml")))
        }
        write.csv(optimal.primers, file = file.path(opti.results.loc, paste(sample.name, 
            "_FINAL_optimized_primers.csv", sep = "")), row.names = FALSE)
        merged.result <- merge.ambig.primers(optimal.primers, mode.directionality, 
            max.degen)
        write_primers(merged.result, file.path(opti.results.loc, 
                        paste(sample.name, "_FINAL_optimized_primers.fasta", sep = "")))
    }
}
#' @rdname PrimerDesign
#' @name PrimerDesign
#' @details
#' The primer design algorithm used by \code{design_primers} consists
#' of three steps: primer initialization, filtering, and optimization.
#' The method for initializing a set of candidate primers is determined
#' via \code{init.algo}. If \code{init.algo} is set to \emph{naive}, primers
#' are created by extracting substrings from all input template sequences.
#' If \code{init.algo} is set to \emph{tree}, degenerate primers are created by
#' merging similar subsequences by forming their consensus sequence up to
#' a degeneracy of at most \code{max.degen}. The tree-based initialization
#' is recommended for related sequences.
#'
#' The candidate primer set is filtered according to the constraints
#' specified in the \code{settings} object. In some cases, it is necessary
#' to relax the constraints in order to reach the desired \code{required.cvg}.
#' In these cases, primers that fail the input constraints may be selected. 
#' If you would like to skip the initialization and filtering stages,
#' you can provide an evaluated \code{Primers} object via \code{primer.df}.
#'
#' Optimizing a primer set entails finding the smallest subset of primers
#' maximizing the coverage, which is done by solving the set cover problem.
#' If melting temperature differences are a constraint,
#' the optimization procedure automatically samples ranges of melting
#' temperatures to find optimal sets for all possible temperatures.
#' You can select the used optimization algorithm via \code{optia.algo}, where
#' you can set "Greedy" for a greedy algorithm or "ILP for 
#' an integer linear program formulation (ILP).
#' While the worst-case runtime of the
#' greedy algorithm is shorter than the worst-case runtime of the ILP, 
#' the greedy solution may yield larger primer sets than the ILP solution.
#'
#' @return \code{design_primers} returns a list with the following fields:
#' \describe{
#' \item{\code{opti}:}{A \code{Primers} object providing the designed primer set.}
#' \item{\code{used_constraints}:}{A list with \code{DesignSettings} objects
#' for each primer direction providing the (possibly relaxed) constraints used
#' for designing the optimal primers.}
#' \item{\code{all_results}:}{A list containing objects of class \code{Primers}.
#' Each list entry corresponds to an optimal primer set for a given
#' melting temperature.}
#' \item{\code{all_used_constraints}:}{A list containing \code{DesignSettings} object for each optimized set in \code{all_results}.}
#' \item{\code{filtered}:}{A list with data providing information on the results
#' of the filtering procedure.}
#' } 
#' @export
#' @examples
#' 
#' # Define PCR settings and primer criteria
#' data(Ippolito)
#' # design only with minimal set of constraints
#' constraints(settings)$primer_length <- c("min" = 18, "max" = 18)
#' constraints(settings) <- constraints(settings)[c("primer_length", "primer_coverage")]
#' # Design only forward primers using a greedy algorithm
#' optimal.primers.greedy <- design_primers(template.df[1:2,], "both", settings, init.algo = "naive")
#' # Usage of the tree-based initialization strategy (requires MAFFT)
#' \dontrun{
#' out.dir <- tempdir()
#' optimal.primers.tree <- design_primers(template.df[1:2,], "both", settings,
#'                          init.algo = "tree", opti.algo = "ILP",
#'                          max.degen = 16,
#'                          cur.results.loc = out.dir)
#' }
design_primers <- function(template.df, mode.directionality = c("both", "fw", "rev"),
     settings, init.algo =  c("naive", "tree"), opti.algo = c("Greedy", "ILP"),
     required.cvg = 1.0, timeout = Inf, max.degen = 16, conservation = 1.0,
     sample.name = NULL, cur.results.loc = NULL, 
     primer.df = NULL, updateProgress = NULL) {
  
    # never allow any binding outside the target region for designing
    init.algo <- match.arg(init.algo)
    if (init.algo == "tree" && !check.tool.function()["MAFFT"]) {
        stop("If you would like to use the tree-based primer initialization strategy, please install MAFFT (http://mafft.cbrc.jp/alignment/software/) first.")
    }

    conOptions(settings)$allowed_other_binding_ratio <- 0
    mode.directionality <- match.arg(mode.directionality)
    opti.algo <- match.arg(opti.algo)
    if (length(settings) == 0 || !is(settings, "DesignSettings")) {
        stop("Please provide a DesignSettings object.")
    }
    if (!is(template.df, "Templates")) {
        stop("Please supply a valid template data frame.")
    }
    if (required.cvg < 0 || required.cvg > 1) {
        stop("Required coverage should be in [0,1].")
    }
    if (max.degen < 0) {
        stop("The maximal degeneracy should be positive.")
    }
    if (conservation < 0 || conservation > 1) {
        stop("The top-conservation percentile should be in [0,1].")
    }
    # check whether all required constraints are provided
    required.constraints <- c("primer_length", "primer_coverage")
    if (!all(required.constraints %in% names(constraints(settings)))) {
        stop("Please provide a settings object containing at least primer_length and primer_coverage as constraints.")
    }
    if (is.null(sample.name)) {
        if (nrow(template.df) >= 1)  {
            sample.name <- template.df$Run[1]
        }
    }
    if (sample.name == "") {
        # sample name should be different from "" for input from csv
        sample.name <- "openPrimeR"
    }
    allowed.region.definition <- conOptions(settings)$allowed_region_definition
    # check whether binding regions are long enough to design primers
    length.check <- check.init.primer.length(template.df, allowed.region.definition,
                    filters(settings)$primer_length, mode.directionality)
    if (!length.check) {
        # allowed region is too short
        my.warning("AllowedRegionTooShort", "The allowed binding region was too short. Please adjust the binding region such that primers can be created for all templates.")
        return(NULL)
    }
    if (length(primer.df) != 0) {
        # skip the init/filtering steps
        if (!is(primer.df, "Primers")) {
            stop("primer.df is not a Primers objectc.")
        } 
        if (!"primer_coverage" %in% colnames(primer.df)) {
            stop("primer.df does not have coverage entries.")
        }
        fw.primers <- primer.df[primer.df$Forward != "", ]
        rev.primers <- primer.df[primer.df$Reverse != "", ]
        if (mode.directionality == "fw") {
            single.primers <- fw.primers
        } else if (mode.directionality == "rev") {
            single.primers <- rev.primers
        } else {
            if (nrow(fw.primers) == 0 || nrow(rev.primers) == 0) {
                warning("'mode.directionality' is 'both', but only primers of one directionality were provided.") 
            }
            single.primers <- NULL
        }
    } else {
        fw.primers <- NULL
        rev.primers <- NULL
        single.primers <- NULL
    }
    target.temps <- NULL
    if (mode.directionality == "fw" || mode.directionality == "rev") {
        optimal.primers.data <- design_primers.single(template.df, sample.name,
            mode.directionality, settings, timeout, opti.algo, 
            allowed.region.definition, init.algo, max.degen, 
            conservation, target.temps, required.cvg, 
            cur.results.loc = cur.results.loc, 
            primer.df = single.primers, updateProgress = updateProgress)
    } else {
        # both directions are to be optimized
        message("#####\n# (FW) Primer design for forward primers\n#####")
        optimal.primer.data.fw <- design_primers.single(template.df, sample.name, "fw", 
            settings, timeout, opti.algo, allowed.region.definition, init.algo, max.degen, 
            conservation, target.temps, required.cvg, 
            cur.results.loc = cur.results.loc,
            primer.df = fw.primers, updateProgress = updateProgress)
        if (is.function(updateProgress)) { # reset progress to 0 for rev opti
            updateProgress(0, "", "set")
        }
        # exclude primer temperature-sets with 0 cvg
        non.zero.cvg <- unlist(lapply(optimal.primer.data.fw$all_results, function(x) sum(x$primer_coverage) != 0))
        optimal.primer.data.fw$all_results <- optimal.primer.data.fw$all_results[non.zero.cvg]
        optimal.primer.data.fw$all_used_constraints <- optimal.primer.data.fw$all_used_constraints[non.zero.cvg]
        if (length(optimal.primer.data.fw$all_results) == 0) {
            warning("Could not construct a primer set for both directions with any coverage given the current settings.")
            return(NULL)
        }
        # optimize reverse primers for the same target temperatures as the fw primers
        if ("melting_temp_diff" %in% names(opti(settings))) {
            target.temps <- unlist(sapply(seq_along(optimal.primer.data.fw$all_results), function(x) mean(optimal.primer.data.fw$all_results[[x]][,"melting_temp"], na.rm= TRUE)))
            names(optimal.primer.data.fw$all_results) <- target.temps
            if ("melting_temp_range" %in% names(filters(settings))) {
                # adjust the melting_temp range here to save time later
                delta <- unname(opti(settings)$melting_temp_diff["max"] / 2)
                new.Tm.range <- c("min" = min(target.temps) - delta, "max" = max(target.temps) + delta)
                new.Tm.limit <- c(new.Tm.range["min"] - delta, new.Tm.range["max"] + delta)
                constraints(settings)$melting_temp_range <- new.Tm.range
                constraintLimits(settings)$melting_temp_range <- new.Tm.limit
            }
        } else {
            target.temps <- rep(NA, length(optimal.primer.data.fw$all_results))
        }
        # adjust the templates for rever design to include only templates that can be covered by both direction primers
        opti.fw <- optimal.primer.data.fw$opti
        cvg.idx <- unique(unlist(covered.seqs.to.idx(opti.fw$Covered_Seqs, template.df)))
        required.nbr <- required.cvg * nrow(template.df) # required nbr of covered templates
        rev.required.cvg <- min(length(cvg.idx) / required.nbr, 1) # update required cvg to the new value considering only the already-covered templates
        rev.template.df <- template.df[cvg.idx[order(cvg.idx)],] # update template data frame to retain only the already covered templates
        primers.fw <- optimal.primer.data.fw$all_results  # arg for cross dimerization considerations between fw/rev primers
		message("#####\n# (REV) Primer design for reverse primers\n#####")
        optimal.primer.data.rev <- design_primers.single(rev.template.df, sample.name, "rev", 
            settings, timeout, opti.algo, allowed.region.definition, init.algo,
            max.degen, conservation, target.temps, rev.required.cvg, 
            cur.results.loc = cur.results.loc, primer.df = rev.primers, updateProgress = updateProgress)
		message("#####\n# (BOTH) Aggregating results\n#####")
        opti.fw <- optimal.primer.data.fw$all_results # base consideration of templates on the best set from the 'fw' run
        opti.rev <- optimal.primer.data.rev$all_results

        if ("melting_temp_diff" %in% names(opti(settings))) {
            # ensure that fw/rev have the same Tm entries.
            # if melting temp diff is not active, the names will be NA, so we don't care.
            Tms <- intersect(names(opti.fw), names(opti.rev))
            opti.fw <- opti.fw[Tms]
            opti.rev <- opti.rev[Tms]
        }
        if ("melting_temp_diff" %in% names(constraints(settings))) {
            # match fw and rev primers for melting temps if melting temp diff is active
            # matching of sets should be improved ...?
            allowed.diff.fw <- max(sapply(optimal.primer.data.fw$all_used_constraints, function(x) constraints(x)$melting_temp_diff))
            allowed.diff.rev <- max(sapply(optimal.primer.data.fw$all_used_constraints, function(x) constraints(x)$melting_temp_diff))
            allowed.diff <- max(c(allowed.diff.fw, allowed.diff.rev))
            fw.tm <- as.numeric(names(opti.fw))
            rev.tm <- as.numeric(names(opti.rev))
            # create a matrix indicating whether the i-th forward set is compatible with the j-th reverse set
            compatible <- do.call(rbind, lapply(fw.tm, function(x) abs(x - rev.tm) < allowed.diff))
            compatible.df <- data.frame(which(compatible, arr.ind = TRUE))
            colnames(compatible.df) <- c("fw_idx", "rev_idx")
        } else {
            # set all pairs of fw/rev sets as compatible
            compatible.df <- data.frame(fw_idx = seq_along(opti.fw),
                                        rev_idx = seq_along(opti.rev))
        }
        fw.rev.data <- evaluate.fw.rev.combinations(opti.fw, opti.rev, compatible.df, template.df)
        sel.set <- select.best.primer.set(fw.rev.data$stats)
        if (length(sel.set) == 0) {
            warning("Could not select an optimal fw-rev combination of primers.")
            optimal.primers <- NULL
            used.settings.fw <- NULL
            used.settings.rev <- NULL
        } else {
            optimal.primers <- fw.rev.data$sets[[sel.set]]
            # get used settings for the output:
            sel.idx.fw <- fw.rev.data$stats$fw_index[sel.set]
            sel.idx.rev <- fw.rev.data$stats$rev_index[sel.set]
            used.settings.fw <- optimal.primer.data.fw$all_used_constraints[[sel.idx.fw]]
            used.settings.rev <- optimal.primer.data.rev$all_used_constraints[[sel.idx.rev]]
        }
        used.settings <- list("fw" = used.settings.fw, "rev" = used.settings.rev)
        # update melting temp diff and cross dimerization to account for fw/rev primers
        for (i in seq_along(fw.rev.data$sets)) {
            cur.set <- fw.rev.data$sets[[i]]
            cur.set <- update.opti.results(cur.set, settings, template.df)
            fw.rev.data$sets[[i]] <- cur.set
        }
        cur.set <- optimal.primers
        cur.set <- update.opti.results(cur.set, settings, template.df)
        optimal.primers <- cur.set
        # construct result: settings
        # unify filtering stats from fw and rev primers
        filter.data <- list()
        for (i in seq_along(optimal.primer.data.fw$filtered)) {
            field <- names(optimal.primer.data.fw$filtered)[i]
            if (field != "used_settings") {
                # constraints are dealt with in another way ..
                filter.data[[field]] <- my_rbind(optimal.primer.data.fw$filtered[[i]], optimal.primer.data.rev$filtered[[i]])
            }
        }
        # gather filtered primers: combine fw and rev filtered primers
        optimal.primers.data <- list(opti = optimal.primers, all_results = fw.rev.data$sets, 
            used_constraints = used.settings,
            filtered = filter.data)
        if (length(cur.results.loc) != 0 && length(sel.set) != 0) {
            # store data
            message("Storing results")
            cur.results.loc <- file.path(cur.results.loc, "both")
            dir.create(cur.results.loc, showWarnings = FALSE)
            PCR.settings <- PCR(settings)
            template.df <- update_template_cvg(template.df, optimal.primers)
            write.out.primer.info(cur.results.loc, optimal.primers.data, 
                mode.directionality, used.settings,
                sample.name, template.df, max.degen)
            plot.all.cvg.info(sample.name, cur.results.loc, optimal.primers, template.df, 
                mode.directionality, "optimized", PCR.settings$primer_concentration,
                PCR.settings$Na_concentration, PCR.settings$Mg_concentration, 
                PCR.settings$K_concentration, PCR.settings$Tris_concentration, 
                settings, required.cvg, used.settings)  # plot opti coverage info
        }
    }
    opti.data <- optimal.primers.data$all_results
    cvg <- unlist(lapply(opti.data, function(x) get_cvg_ratio(x, template.df, mode.directionality = mode.directionality, as.char = TRUE)))
    message("Coverage of each designed set: ", paste(cvg, collapse = ", "))
    return(optimal.primers.data)
}
#' Selection of Best Primer Set.
#' 
#' Selects the best primer set according to coverage and melting temperature differences among primers in the set.
#'
#' @param stats Statistics of the primer sets to be evaluated.
#' @return The index of the best primer set.
#' @keywords internal
select.best.primer.set <- function(stats) {
    if (length(stats) == 0) {
        return(NULL)
    }
    # stats: properties of each considered primer set
    idx <- which(stats$Coverage == max(stats$Coverage))  # highest coverage sets
    if (any(!is.na(stats$TempDiff[idx]))) {
        idx <- idx[which.min(stats$TempDiff[idx])]  # highest cvg set with lowest temperature difference between primers
    }
    return(idx)
}
#' Evaluation of Set Combinations
#'
#' Evaluates the combinations of forward and reverse primer sets.
#'
#' @param opti.fw List with forward optimal primer sets.
#' @param opti.rev List with reverse optimal primer sets.
#' @param compatible.df Data frame containing the indicices
#' of temperature-compatible forward and reverse primers sets.
#' @param opti.rev.indices Indices for accessing \code{opti.rev}.
#' @param template.df Template data frame for which primers were designed.
#' @return List with information on the combinations of forward and reverse primers
#' as well as the combined data frames themselves.
#' @keywords internal
evaluate.fw.rev.combinations <- function(opti.fw, opti.rev, compatible.df, template.df) {
    stat.df <- vector("list", nrow(compatible.df)) # stats of compatible sets
    # select compatible combinations of sets via the indices
    primer.sets <- vector("list", length(opti.fw))
    # set fw temperatures as names as temperatures of fw/rev sets should correspond:
    names(primer.sets) <- names(opti.fw)
    for (i in seq_len(nrow(compatible.df))) {
        fw.idx <- compatible.df$fw_idx[i]
        rev.idx <- compatible.df$rev_idx[i]
        combi.df <- my_rbind(opti.fw[[fw.idx]], opti.rev[[rev.idx]]) 
        primer.sets[[i]] <- combi.df
        cvg <- get_cvg_ratio(combi.df, template.df)
        temp.diff <- NA
        if ("melting_temp" %in% colnames(combi.df)) {
            temp.diff <- max(combi.df$melting_temp) - min(combi.df$melting_temp)
        }
        stats <- data.frame(Coverage = cvg, 
            TempDiff = temp.diff, fw_index = fw.idx,
            rev_index = rev.idx)
        stat.df[[i]] <- stats
    }
    stat.df <- do.call(rbind, stat.df)
    out <- list(stats = stat.df, sets = primer.sets)
    return(out)
}

#' Augmentation of Optimized Primer Data.
#'
#' Adds melting_temp_diff and cross_dimerization info to optimized sets.
#'
#' @param primer.df A primer data frame.
#' @param settings A \code{DesignSettings} object.
#' @return An updated primer data frame.
#' @keywords internal
update.opti.results <- function(primer.df, settings, template.df) {
    if (length(primer.df) == 0 || nrow(primer.df) == 0) {
        return(primer.df)
    }
    mode.directionality <- get.analysis.mode(primer.df)
    annealing.temp <- compute_annealing_temp(primer.df, mode.directionality, 
                                template.df, PCR(settings)$Na_concentration,
                                PCR(settings)$Mg_concentration, 
                                PCR(settings)$K_concentration, 
                                PCR(settings)$Tris_concentration, 
                                PCR(settings)$primer_concentration)
    if ("melting_temp_diff" %in% colnames(primer.df)) {
        temp.diff <- get.melting.temp.diff(primer.df$Tm_C_fw, primer.df$Tm_C_rev)
        primer.df[, "melting_temp_diff"] <- temp.diff
        primer.df[, "EVAL_melting_temp_diff"] <- temp.diff <= opti(settings)$melting_temp_diff["max"]
    }
    cross.dimer.df <- NULL
    if ("cross_dimerization" %in% names(constraints(settings))) {
        # here we have no_structs not set, since we want to annotate the designed data correctly
        cross.dimer.df <- compute.all.cross.dimers.frontend(primer.df, PCR(settings)$primer_concentration, 
                    PCR(settings)$Na_concentration, PCR(settings)$Mg_concentration, 
                    PCR(settings)$K_concentration, PCR(settings)$Tris_concentration, 
                    annealing.temp)
    }
    if (length(cross.dimer.df) != 0) {
        # no entries in constraint.values when there's nothing to report ...
        primer.df <- update.constraint.values(primer.df, cross.dimer.df)
        primer.df$EVAL_cross_dimerization <- (primer.df$Cross_Dimer_DeltaG >= constraints(settings)$cross_dimerization["min"])
    }
    return(primer.df)
}
#' Design Primers for a Single Direction
#'
#' Designs primers for a single direction.
#'
#'
#' @param template.df Template data frame with sequences for which primers shall be designed.
#' @param sample.name Identifier for the templates.
#' @param mode.directionality Template strands for which primers shall be designed.
#' Primers can be designed either only for forward strands, only for reverse strands, or
#' both strand directions.
#' @param settings A \code{DesignSettings} object specifying the 
#' criteria for designing primers.
#' @param timeout Timeout in seconds for the optimization with ILPs.
#' @param opti.algo The algorithm to be used for solving the primer set covering problem.
#' @param allowed.region.definition Definition of the target binding sites used for evaluating the coverage.
#' If \code{allowed.region.definition} is "within", primers have to lie within the allowed binding region.
#' If \code{allowed.region.definition} is "any", primers have to overlap with the allowed binding region.
#' The default is that primers have to bind within the target binding region.
#' @param init.algo The algorithm to be used for initializing primers.
#' If \code{init.algo} is \code{naive}, then primers are constructed from substrings of the input template sequences.
#' If \code{init.algo} is \code{tree}, phylogenetic trees are used to form degenerate primers whose degeneracy is
#' bounded by \code{max.degen}. 
#' @param max.degen Maximal degeneracy of merged primers.
#' @param conservation When using the tree-based primer initialization, consider only the \code{conservation}
#' percentile of regions with the highest conservation.
#' @param target.temps Target melting temperatures for optimized primer sets in Celsius.
#' Only required when optimizing primers for both strand directions and one optimization was already performed.
#' @param required.cvg The target ratio of covered template sequences. 
#' If the target ratio cannot be reached, the constraint settings are relaxed up to the relaxation limits.
#' @param fw.primers List with optimized primer data frames corresponding to \code{target.temps}. 
#' Only required for optimizing both strand directions and only 
#' in the second optimization run in order to check for cross dimerization.
#' @param cur.results.loc Directory for storing results of the primer design procedure.
#' @param primer.df A data frame of evaluated primer candidates that can be optimized directly.
#' @param updateProgress Shiny progress callback function.
#' @return A list containing the results of the primer design procedure:
#' \describe{
#' \item{\code{opti}:}{A \code{Primers} object representing the set of optimized primers.}
#' \item{\code{all_results}:}{A list containing the optimal results for each sampled melting temperature range
#'                            in terms of a \code{Primers} object 
#'                            in case that the \code{melting_temp_diff} constraint was active. Otherwise,
#'                            \code{all_results} only has a single entry representing a primer set relating to an undefined melting temperature.}
#' \item{\code{used_constraints}:}{A \code{DesignSettings} object with the (adjusted) analysis settings.}
#' \item{\code{filtered}:}{A \code{Primers} object containing the primer candidates that passed the filtering procedure
#'                         and which gave rise to the final optimal set.}
#' }
#' @keywords internal
design_primers.single <- function(template.df, sample.name, 
    mode.directionality = c("fw", "rev"), 
    settings, timeout, opti.algo, allowed.region.definition, init.algo, 
    max.degen, conservation, target.temps, required.cvg, 
    fw.primers = NULL, 
    cur.results.loc = NULL, primer.df = NULL, updateProgress = NULL) {
   
    if (length(mode.directionality) == 0) {
        stop("Please supply the 'mode.directionality' argument.")
    }
    mode.directionality <- match.arg(mode.directionality)
    ####################
    # 'Unlist' settings
    #######################
    constraint.settings <- filters(settings)
    constraint.limits <- filterLimits(settings)
    PCR.settings <- PCR(settings)
    primer_conc <- PCR.settings$primer_concentration
    template_conc <- PCR.settings$template_concentration
    na_salt_conc <- PCR.settings$Na_concentration
    mg_salt_conc <- PCR.settings$Mg_concentration
    k_salt_conc <- PCR.settings$K_concentration
    tris_salt_conc <- PCR.settings$Tris_concentration
    other.settings <- conOptions(settings)
    allowed.mismatches <- other.settings$allowed_mismatches
    allowed.stop.codons <- other.settings$allowed_stop_codons
    allowed.other.binding.ratio <- other.settings$allowed_other_binding_ratio
    disallowed.mismatch.pos <- other.settings$disallowed_mismatch_pos
    ######################
    primer.lengths <- constraint.settings$primer_length
    opti.results.loc <- NULL
    filtering.results.loc <- NULL
    if (length(cur.results.loc) != 0) {
        # create some diagnostic folders
        cur.results.loc <- file.path(cur.results.loc, mode.directionality)
        filtering.results.loc <- file.path(cur.results.loc, "filtering")
        opti.results.loc <- file.path(cur.results.loc, "optimization", opti.algo)
        dir.create(cur.results.loc, showWarnings = FALSE, recursive = TRUE)
        dir.create(filtering.results.loc, showWarnings = FALSE, recursive = TRUE)
        dir.create(opti.results.loc, showWarnings = FALSE, recursive = TRUE)
    }
    # load primers/initialize primers
    used.filtering.constraints <- NULL
    annealing.temp <- NULL  # no pre-determined annealing temperature for filtering should be required
    filtered.result <- NULL
    used.settings <- settings
    if (length(primer.df) == 0) {
        # initialize and filter primer set
		message("# Phase 1: Initialization of primers")
        if (is.function(updateProgress)) {
            detail <- paste("Initialization_", mode.directionality, sep = "")
            updateProgress(1/3, detail, "inc")
        }
        primer.df <- initialize.primer.set(template.df, sample.name, primer.lengths, allowed.region.definition, 
            mode.directionality, init.algo, max.degen, conservation, cur.results.loc)
        # filter primers
		message("# Phase 2: Selection of primers")
        if (is.function(updateProgress)) {
            detail <- paste("Filtering_", mode.directionality, sep = "")
            updateProgress(1/3, detail, "inc")
        }
        filtered.result <- filter.primer.set.opti(primer.df, sample.name, 
            template.df, settings, mode.directionality, required.cvg, 
            filtering.results.loc, target.temps)
        filtered.df <- filtered.result$data
        used.settings <- filtered.result$used_settings # update the setings object
    } else {
        filtered.df <- primer.df
    }
    if (is.function(updateProgress)) {
        detail <- paste("Optimizing_", mode.directionality, sep = "")
        updateProgress(1/3, detail, "inc")
    }
    # ensure that melting temperature is available if melting_temp_diff constraint is active:
    if ("melting_temp_diff" %in% names(opti(used.settings)) && !"melting_temp" %in% colnames(filtered.df) && 
        Sys.which("melting-batch") != "") {
        # compute melting temperatures if they don't exist
        melting.temps <- compute.melting.temps(filtered.df, primer_conc, na_salt_conc, 
            mg_salt_conc, k_salt_conc, tris_salt_conc, mode.directionality)
        filtered.df <- cbind(filtered.df, melting.temps)
    }
	message("# Phase 3: Optimization of primers")
    if (opti.algo == "ILP") {
        # ILP OPTIMIZATION optimize with ILPs
        optimal.primers.data <- optimize.ILP(filtered.df, template.df, used.settings,
            primer_conc, template_conc, na_salt_conc, mg_salt_conc, 
            k_salt_conc, tris_salt_conc, allowed.mismatches, 
            allowed.other.binding.ratio, allowed.stop.codons, allowed.region.definition, 
            disallowed.mismatch.pos, target.temps, required.cvg, fw.primers = fw.primers, 
            diagnostic.location = opti.results.loc, timeout = timeout)
    } else {
        # GREEDY OPTIMIZATION
        optimal.primers.data <- select.primers.by.cvg(filtered.df, used.settings, 
            template.df, mode.directionality, required.cvg, constraint.settings$primer_coverage["primer_mismatches.max"], 
            primer_conc, na_salt_conc, mg_salt_conc, k_salt_conc, tris_salt_conc, 
            template_conc, allowed.other.binding.ratio, 
            allowed.stop.codons, allowed.region.definition, 
            disallowed.mismatch.pos, target.temps, fw.primers)
    }
    opti.set <- optimal.primers.data$opti
    cvg.s <- paste0(round(get_cvg_ratio(opti.set, template.df) * 100, 2), "%")
    message("\to Selected ", nrow(opti.set), " primers (", cvg.s, " coverage).")
    # update melting_temp_diff & cross-dimer entries in all optimized sets
    for (i in seq_along(optimal.primers.data$all_results)) {
        cur.set <- optimal.primers.data$all_results[[i]]
        cur.set <- update.opti.results(cur.set, used.settings, template.df)
        optimal.primers.data$all_results[[i]] <- cur.set
    }
    opti.set <- update.opti.results(opti.set, used.settings, template.df)
    optimal.primers.data$opti <- opti.set
    # add filtering results to output
    optimal.primers.data$filtered <- filtered.result
    ###########
    # create a list of settings for the output
    used.settings <- list(optimal.primers.data$used_constraints)
    names(used.settings) <- mode.directionality
    optimal.primers.data$used_constraints <- used.settings
    if (length(opti.results.loc) != 0) {
        message("Storing results")
        write.out.primer.info(opti.results.loc, optimal.primers.data, mode.directionality, 
        used.settings, sample.name, template.df, max.degen)
        visualize.all.results(sample.name, filtering.results.loc, opti.results.loc, primer_conc, 
        na_salt_conc, mg_salt_conc, k_salt_conc, tris_salt_conc, 
        settings, mode.directionality, used.settings, required.cvg)
    }
    return(optimal.primers.data)
}

#' Visualization of Design Results.
#'
#' Visualizes all results from designing primers.
#'
#' @param sample Identifier of the design run.
#' @param filtering.results.loc Location of filtering results.
#' @param opti.results.loc Location of optimization results.
#' @param primer_conc Primer concentration.
#' @param template_conc Template concentration.
#' @param na_salt_conc Sodium ion concentration.
#' @param mg_salt_conc Magensium ion concentration.
#' @param k_salt_conc Potassium ion concentration.
#' @param tris_salt_conc Tris ion concentration.
#' @param settings The \code{DesignSettings} object.
#' @param mode.directionality Strand direction for which primers were designed.
#' @param used.settings A list with the used settings for optimization (fields "fw" and "rev").
#' @param required.cvg The required coverage.
#' @return Writes visualizations to files in.
#' @keywords internal
visualize.all.results <- function(sample, filtering.results.loc, opti.results.loc, 
    primer_conc, na_salt_conc, mg_salt_conc, k_salt_conc, tris_salt_conc, 
    settings, mode.directionality, used.settings, required.cvg) {
    if (length(filtering.results.loc) == 0 || length(opti.results.loc) == 0) {
        return(NULL)  # not suppoed to plot anything
    }
    #### Plot filtering stats
    excluded.df <- read_primers(file.path(filtering.results.loc, paste(sample, "_excluded_primers.csv", 
        sep = "")))
    template.df <- read_templates(file.path(filtering.results.loc, paste(sample, "_filtered_templates.csv", 
        sep = "")))
    filtered.df <- read_primers(file.path(filtering.results.loc, paste(sample, 
        "_filtered_primers.csv", sep = "")))
    filtered.stats <- read.csv(file.path(filtering.results.loc, paste(sample, "_filtered_primers_stats.csv", 
        sep = "")), stringsAsFactors = FALSE)
    relax.stats <- file.path(filtering.results.loc, paste(sample, "_filtered_primers_stats_relax.csv", 
        sep = ""))
    if (file.size(relax.stats) > 3) {
        stats.relax <- read.csv(file.path(filtering.results.loc, paste(sample, "_filtered_primers_stats_relax.csv", 
            sep = "")), stringsAsFactors = FALSE)
    } else {
        stats.relax <- NULL
    }
    visualize.filtering.results(sample, filtering.results.loc, mode.directionality, 
        excluded.df, template.df, filtered.df, filtered.stats, stats.relax, primer_conc, 
        na_salt_conc, mg_salt_conc, k_salt_conc, tris_salt_conc, 
        settings)
    # optimization plots: differentiate between different optimization methods
    optimal.primers <- try(read_primers(file.path(opti.results.loc, paste(sample, 
        "_FINAL_optimized_primers.csv", sep = ""))))
    if (is(optimal.primers, "try-error")) {
        return()  # no available results here
    }
    optimal.primers[is.na(optimal.primers)] <- ""
    template.df <- update_template_cvg(template.df, optimal.primers)  # update lex cvg 
    plot.all.cvg.info(sample, opti.results.loc, optimal.primers, template.df, mode.directionality, 
        "optimized", primer_conc, na_salt_conc, mg_salt_conc, k_salt_conc, tris_salt_conc, 
        settings, required.cvg, used.settings) 
}
#' Visualization of Filtering Results.
#'
#' Visualizes the filtering results.
#'
#' @param sample Primer design run identifier.
#' @param results.loc Location where the filtering results are stored.
#' @param mode.directionality Design direction.
#' @param excluded.df Data frame with excluded primers.
#' @param template.df Template data frame.
#' @param filtered.df Primer data frame containing the primers that passed the constraints.
#' @param stats.relax Filtering statistics after relaxation.
#' @param primer_conc Primer concentration.
#' @param template_conc Template concentration.
#' @param na_salt_conc Sodium ion concentration.
#' @param mg_salt_conc Magensium ion concentration.
#' @param k_salt_conc Potassium ion concentration.
#' @param tris_salt_conc Tris ion concentration.
#' @param settings A \code{DesignSettings} object.
#' @param required.cvg The required coverage.
#' @return Write-out of filtering results.
#' @keywords internal
visualize.filtering.results <- function(sample, results.loc, mode.directionality, 
    excluded.df, template.df, filtered.df, filtered.stats, stats.relax, primer_conc, na_salt_conc, 
    mg_salt_conc, k_salt_conc, tris_salt_conc, settings, required.cvg) {
    plot.all.filtering.stats(results.loc, sample, excluded.df, filtered.stats, stats.relax, 
        template.df)
    # create coverage plots/stats
    plot.all.cvg.info(sample, results.loc, filtered.df, template.df, mode.directionality, 
        "filtering", primer_conc, na_salt_conc, mg_salt_conc, k_salt_conc, tris_salt_conc, 
        settings, required.cvg)
    cvg.stats <- get_cvg_stats(filtered.df, template.df)
    write.csv(cvg.stats, file.path(results.loc, paste(sample, "_filtering_cvg_stats.csv", 
        sep = "")), row.names = FALSE)
}

#' Plot Filtering Stats.
#'
#' Plots filtering statistics.
#'
#' @param results.loc Location where the filtering results are stored.
#' @param sample Primer design run identifier.
#' @param excluded.df Data frame with excluded primers.
#' @param filtered.stats Filtering statistics data frame.
#' @param stats.relax Filtering statistics after relaxation.
#' @param template.df Template data frame.
#' @return Write-out of filtering results.
#' @keywords internal
plot.all.filtering.stats <- function(results.loc, sample, excluded.df, filtered.stats, 
    stats.relax, template.df) {
    # Plot all filtering stats
    p <- plot.excluded.hist(excluded.df, filtered.stats, template.df)
    my_ggsave(file.path(results.loc, paste(sample, "_excluded_hist.png", sep = "")), p, 
        units = "cm", width = 40, height = 20)
    p <- plot.filtering.stats(filtered.stats, stats.relax)
    my_ggsave(file.path(results.loc, paste(sample, "_filtering_stats.png", sep = "")), p,
        units = "cm", width = 30, height = 15)
    p <- plot.filtering.stats.cvg(filtered.stats, stats.relax)
    my_ggsave(file.path(results.loc, paste(sample, "_filtering_stats_cvg.png", sep = "")), p,
        units = "cm", width = 30, height = 15)
    p <- plot.filtering.runtime(filtered.stats)
    my_ggsave(file.path(results.loc, paste(sample, "_filtering_runtime.png", sep = "")), p)
} 
#' Plots Coverage Information
#'
#' Visualizes all coverage-related data.
#'
#' 
#' @param sample Primer design run identifier.
#' @param results.loc Location where the filtering results are stored.
#' @param primers Primer data frame.
#' @param template.df Template data frame.
#' @param mode.directionality Design direction.
#' @param identifier Identifies whether filtering or optimization info should be displayed.
#' @param primer_conc Primer concentration.
#' @param na_salt_conc Sodium ion concentration.
#' @param mg_salt_conc Magensium ion concentration.
#' @param k_salt_conc Potassium ion concentration.
#' @param tris_salt_conc Tris ion concentration.
#' @param settings The \code{DesignSettings} object.
#' @param required.cvg The required coverage.
#' @param used.settings A list containing \code{DesignSettings} objects for the 'fw' and 'rev' optimization.
#' @return Writes plots to files.
#' @keywords internal
plot.all.cvg.info <- function(sample, results.loc, primers, template.df, mode.directionality, 
    identifier = c("filtering", "optimized"), primer_conc, na_salt_conc, mg_salt_conc, k_salt_conc, tris_salt_conc, 
    settings, required.cvg, used.settings = NULL) {

    if (length(identifier) == 0) {
        stop("Please supply the 'identifier' arg.")
    }
    identifier <- match.arg(identifier)
    if (length(primers) == 0 || length(template.df) == 0 || length(results.loc) == 0) {
        return(NULL)
    }
    constraint.settings <- constraints(settings)
    opti.constraints <- opti(settings)
    out.loc <- file.path(results.loc, "plots")
    dir.create(out.loc, showWarnings = FALSE)
    # primer group coverage overview table
    cvg.stats <- get_cvg_stats(primers, template.df)
    write.csv(cvg.stats, file = file.path(out.loc, paste(sample, "_", identifier, 
        "_cvg_stats.csv", sep = "")), row.names = FALSE)
    plot_template_cvg(primers, template.df)
    my_ggsave(file.path(out.loc, paste(sample, "_", identifier, "_template_cvg.png", sep = "")), 
        units = "cm", width = 15, height = 15)
    plot_template_cvg(primers, template.df, per.mismatch = TRUE)
    my_ggsave(file.path(out.loc, paste(sample, "_", identifier, "_template_cvg_mismatch.png", sep = "")), 
        units = "cm", width = 15, height = 15)
    # primer binding regions
    if (mode.directionality == "both") {
        relations <- c("fw", "rev")  # forward primer info
    } else {
        relations <- mode.directionality
    }
    for (i in seq_along(relations)) {
        relation <- relations[i]
        plot_primer_binding_regions(primers, template.df, relation, group = "all")
        my_ggsave(file.path(out.loc, paste(sample, "_", identifier, "_primer_binding_regions_", relation, ".png", 
            sep = "")), units = "cm", width = 30, height = 15)
    }
    if (identifier != "filtering") {
        # don't plot for the filtered primers: too many ..
        # create pdf report
        file <- file.path(out.loc, get_report_fname("report", sample))
        create_report(primers, template.df, file, settings, 
                    sample.name = sample, used.settings = used.settings, 
                    required.cvg = required.cvg)
        plot_primer_cvg(primers, template.df)
        my_ggsave(file.path(out.loc, paste(sample, "_", identifier, "_primer_cvg.png", 
            sep = "")), units = "cm", width = 25, height = 15)
        for (i in seq_along(relations)) {
            relation <- relations[i]
            p <- plot_primer(primers, template.df, primers$Identifier, relation)
            my_ggsave(file.path(out.loc, paste(sample, "_", identifier, "_", relation, "_primer_plot.png", 
                sep = "")), plot = p, height = 90, units = "cm", width = 30, dpi = 300)
        }
        # self dimerization
        self.dimer.cutoff <- ifelse("self_dimerization" %in% names(constraint.settings), 
            constraint.settings$self_dimerization, NA)
        annealing.temp <- compute_annealing_temp(primers, mode.directionality, 
                                template.df, PCR(settings)$Na_concentration,
                                PCR(settings)$Mg_concentration, 
                                PCR(settings)$K_concentration, 
                                PCR(settings)$Tris_concentration, 
                                PCR(settings)$primer_concentration)
        if (!is.na(self.dimer.cutoff)) {

            # only plot if data were supposed to be computed
            self.dimer.data <- compute.all.self.dimers(primers, primer_conc, 
                na_salt_conc, mg_salt_conc, k_salt_conc, tris_salt_conc, annealing.temp)
            se <- self.dimer.data
            #se <- view.self.dimer.table(self.dimer.data)
            write.csv(se, file = file.path(out.loc, paste(sample, "_", identifier, "_self_dimer_details.csv", 
                sep = "")), row.names = FALSE)
            write.csv(self.dimer.data, file = file.path(out.loc, paste(sample, "_", identifier, 
                "_self_dimer_details_raw.csv", sep = "")), row.names = FALSE)
            p <- plot.dimer.dist(se, self.dimer.cutoff)
            my_ggsave(file.path(out.loc, paste(sample, "_", identifier, "_self_dimer_DeltaG.png", 
                sep = "")), plot = p)
            self.dim.table <- dimerization.table(se, self.dimer.cutoff, "Self-Dimerization")
            write.csv(self.dim.table, file = file.path(out.loc, paste(sample, "_", identifier, 
                "_self_dimer_counts.csv", sep = "")), row.names = FALSE)
            # cross dimerization
            cross.dimer.cutoff <- ifelse("cross_dimerization" %in% names(opti.constraints),
                opti.constraints$cross_dimerization, -8)
            all.cross.dimer.data <- compute.all.cross.dimers.unfiltered(primers, primer_conc, 
                na_salt_conc, mg_salt_conc, k_salt_conc, tris_salt_conc, annealing.temp)
            cross.dimer.data <- compute.all.cross.dimers(primers, primer_conc, 
                na_salt_conc, mg_salt_conc, k_salt_conc, tris_salt_conc, annealing.temp,
                all.cross.dimer.data)
            if (length(cross.dimer.data) != 0) {
                cross.dimer.data <- na.omit(cross.dimer.data)
            }
            se <- cross.dimer.data
            #se <- view.cross.dimer.table(cross.dimer.data)
            write.csv(se, file = file.path(out.loc, paste(sample, "_", identifier, "_cross_dimer_details.csv", 
                sep = "")), row.names = FALSE)
            write.csv(cross.dimer.data, file = file.path(out.loc, paste(sample, "_", 
                identifier, "_cross_dimer_details_raw.csv", sep = "")), row.names = FALSE)
            p <- plot.dimer.dist(se, cross.dimer.cutoff)
            my_ggsave(file.path(out.loc, paste(sample, "_", identifier, "_cross_dimer_DeltaG.png", 
                sep = "")), plot = p)
            cross.dim.table <- dimerization.table(se, cross.dimer.cutoff, "Cross-Dimerization")
            write.csv(cross.dim.table, file = file.path(out.loc, paste(sample, "_", identifier, 
                "_cross_dimer_counts.csv", sep = "")), row.names = FALSE)
        }
        # mismatches
        mm.table.fw <- compute.mismatch.table(primers, template.df, "fw")
        write.csv(mm.table.fw, file = file.path(out.loc, paste(sample, "_", identifier, 
            "_mismatch_priming.csv", sep = "")), row.names = FALSE)
        # constraint evaluation: which constraints are fulfilled?
        p <- plot_constraint_fulfillment(primers, settings)
        my_ggsave(file.path(out.loc, paste(sample, "_", identifier, "_constraints_evaluation.png", 
            sep = "")), plot = p)
        # output optimal set subsets
        groups <- NULL  # consider cvg of all groups
        compute.all.primer.subsets.ILP(primers, template.df, 1, groups, out.loc, required.cvg)
    }
}
#' Computation of Primer Subsets
#'
#' Computes all optimal primer subsets and stores their plots.
#'
#' @param primer.df Primer data frame.
#' @param template.df Template data frame.
#' @param k Subset size-increment.
#' @param groups Identifiers of template groups in order to limit coverage to
#' certain groups of template sequences.
#' @param cur.results.loc Location for storing the results.
#' @param required.cvg The required coverage ratio.
#' @return Write-out of results.
#' @keywords internal
compute.all.primer.subsets.ILP <- function(primer.df, template.df, k, groups, cur.results.loc, required.cvg = 1) {
    if (length(primer.df) == 0 || nrow(primer.df) == 0) {
        return(NULL)
    }
    identifier <- "subsets"
    if (length(groups) != 0) {
        identifier <- paste(identifier, "_groups_", paste(groups, collapse = "_", 
            sep = ""), sep = "")
    }
    top.primers.overall <- subset_primer_set(primer.df, template.df, k, groups, 
                                            identifier, cur.results.loc)
    p <- plot_primer_subsets(top.primers.overall, template.df, 
                        required.cvg = required.cvg)
    my_ggsave(file.path(cur.results.loc, paste(identifier, "_coverage_ramp.png", sep = "")), p,
        units = "cm", width = 15, height = 15)
}
#' Creation of Melting Temperature Groups
#'
#' Creates a data frame identifying target melting temperatures of individual primer sets.
#'
#' @param primers An object of class \code{Primers} for which to create groups based on melting temperatures.
#' @param template.df An object of class \code{Templates} corresponding to the primers.
#' @param settings A \code{DesignSettings} objects.
#' @param target.temps Pre-defined target melting temperatures to use instead
#' of automatically determining groups from the \code{primers}.
#' @return Data frame with target melting temperatures for individual primer sets.
#' @keywords internal
create.Tm.brackets <- function(primers, template.df, settings, target.temps = NULL) {
    group.assignment <- rep(1, nrow(primers)) # all in one group by default
    group.df <- NULL
    group.df <- data.frame(GroupID = 1, min_Tm = NA, max_Tm = NA, 
                            target_Tm = NA, annealing_temp = NA)
    opti.constraints <- opti(settings)
    max.Tm.delta <- opti.constraints[["melting_temp_diff"]]["max"]
    # ensure that melting temperatures are available if melting_temp_diff is to be constrained
    if ("melting_temp_diff" %in% names(opti.constraints) && !"melting_temp" %in% colnames(primers) && nrow(primers) != 0) {
        # melting temp is not available although melting_temp_diff should be respected -> compute melting temperatures here
        primers <- check_constraints(primers, template.df, settings, active.constraints = "melting_temp_diff")
    } else if ("melting_temp" %in% colnames(primers) && "melting_temp_diff" %in% names(opti.constraints)) {
        # we have a melting temp, but we might still have missing entries ...
        na.idx <- which(is.na(primers$melting_temp))
        if (length(na.idx) != 0) {
            primers[na.idx,] <- check_constraints(primers[na.idx,], template.df, settings, active.constraints = "melting_temp_diff")
        }
    }
    if (length(target.temps) != 0 && "melting_temp_diff" %in% names(opti.constraints)) {
        # init group.df with target temps
        group.df <- data.frame(GroupID = seq_along(target.temps), min_Tm = target.temps - (max.Tm.delta/2), 
                                max_Tm = target.temps + (max.Tm.delta/2), target_Tm = target.temps, annealing_temp = NA)
    } else if ("melting_temp_diff" %in% names(opti.constraints) && nrow(primers) != 0) {
        # no target temperatures specified -> determine grouping automatically:
        # ensure that groups overlap by selecting steps half the size of the allowed temperature difference
        groups <- c(seq(min(primers$melting_temp, na.rm = TRUE), max(primers$melting_temp, na.rm = TRUE), ifelse(max.Tm.delta !=0, max.Tm.delta / 2, 1)), max(primers$melting_temp, na.rm = TRUE))
        if (length(groups) == 2) {
            # ensure that 'group.df' can be constructed
            groups <- c(groups, max(primers$melting_temp, na.rm = TRUE))
        }
        group.df <- data.frame(GroupID = seq_len(length(groups) - 2), min_Tm = groups[seq_len(length(groups) - 
            2)], max_Tm = groups[3:length(groups)])
        group.df$target_Tm <- group.df$min_Tm + (group.df$max_Tm - group.df$min_Tm)/2
    }
    # assign annealing temperatures
    if (nrow(primers) != 0 && "melting_temp_diff" %in% names(opti.constraints)) {
        group.df$annealing_temp <- sapply(group.df$min_Tm, function(x) annealing.temp.rule.of.thumb(x))
        # assign primers to groups if melting temperature is available
        group.assignment <- sapply(primers$melting_temp, function(x) group.df$GroupID[which(x >= 
            group.df$min_Tm & x < group.df$max_Tm)])
        group.assignment <- sapply(group.assignment, function(x) ifelse(length(x) == 
            0, NA, x))
        mean.temp <- mean(primers$melting_temp)  # mean melting temperature of the data set serves as the starting position of temperature optimization
        target.temp.order <- order(abs(group.df$target_Tm - mean.temp))
        group.df <- group.df[target.temp.order, ]
    } else if (nrow(primers) != 0 && !"melting_temp_diff"  %in% names(opti.constraints)) {
        # all primers are in one group -> compute annealing temp for all
        # necessary for computations that depend on the Ta
        mode.directionality <- get.analysis.mode(primers)
        annealing.temp <- compute_annealing_temp(primers, mode.directionality, 
                                template.df, PCR(settings)$Na_concentration,
                                PCR(settings)$Mg_concentration, 
                                PCR(settings)$K_concentration, 
                                PCR(settings)$Tris_concentration, 
                                PCR(settings)$primer_concentration)
        group.df[, "annealing_temp"] <- min(annealing.temp)
        # the melting temp range doesn't need to be computed in this case
        # no adjustment of groups necessary
    }
    return(list(group_idx = group.assignment, df = group.df, primers = primers))
}
#' Cross-Dimerization Filtering
#'
#' Removes cross-dimerizing primers from the input data.
#'
#' @param primers.rev The primer data set to be filtered for cross-dimers.
#' @param primers.fw The primer data set to be checked against cross-dimerization.
#' @param opti.constraints List with optimization constraint settings.
#' @param annealing.temp The PCR annealing temperature.
#' @param primer_conc Primer concentration.
#' @param na_salt_conc Sodium ion concentration.
#' @param mg_salt_conc Magensium ion concentration.
#' @param k_salt_conc Potassium ion concentration.
#' @param tris_salt_conc Tris ion concentration.
#' @return \code{primers.rev} with removed cross-dimerizing primers.
#' @keywords internal
#filter.cross.comp <- function(primers.rev, primers.fw, opti.constraints, annealing.temp, 
    #primer_conc, na_salt_conc, mg_salt_conc, k_salt_conc, tris_salt_conc) {
    ## determines whether primers in first set (rev) dimerize with primers in second
    ## set (fw) and removes primers from first set. this function is used during
    ## optimization, when one direction has already been optimized (primers.fw) and
    ## the other one hasn't been yet.  no adjustment necessary ...
    #if (!"cross_dimerization" %in% names(opti.constraints) || length(primers.fw) == 
        #0 || length(primers.rev) == 0) {
        #return(primers.rev)
    #}
    #primers.1 <- primers.rev$Reverse  # rows: selection
    #primers.2 <- primers.fw$Forward  # cols
    #G.df <- get.cross.dimers(primers.1, primers.2, primer_conc, na_salt_conc, 
        #mg_salt_conc, k_salt_conc, tris_salt_conc, annealing.temp, mode = "asymmetric")
    #deltaG.cutoff <- opti.constraints$cross_dimerization["min"]
    ## problem with create.G.matrix: dimension is defined by primer.df ... always
    # square!
    #G.matrix <- create.G.matrix(primers.rev, G.df, primers.fw)  # min deltaG values of cross-dimerization conformations for every pair of primers
    #D <- compute.dimer.matrix(G.matrix, deltaG.cutoff)  # 1 for dimers, 0 else
    ## filter:
    #rm.idx <- which(apply(D, 1, function(x) any(x == 1)))
    #if (length(rm.idx) != 0) {
        #primers.rev <- primers.rev[-rm.idx, ]
    #}
    #return(primers.rev)
#}
#' Creation of Melting Temperature Sets.
#'
#' Stratifies primers according to their melting temperatures and checks temperature-dependent constraints.
#'
#' Cross-dimerization between forward and reverse primers is checked here, in case this is a second 
#' optimization run and a list \code{primers.fw} with primer data frames is provided.
#'
#' @param primer.df Primer data frame.
#' @param template.df Template data frame.
#' @param Tm.brackets Data frame with target primer melting temperatures.
#' @param settings A \code{DesignSettings} object.
#' @param mode.directionality Identifier of strand for which primers shall be designed.
#' @param primer_conc Primer concentration.
#' @param template_conc Template concentration.
#' @param na_salt_conc Sodium ion concentration.
#' @param mg_salt_conc Magensium ion concentration.
#' @param k_salt_conc Potassium ion concentration.
#' @param tris_salt_conc Tris ion concentration.
#' @param allowed.mismatches The number of mismatches primers are allowed to have with the templates.
#' @param allowed.other.binding.ratio Ratio of primers allowed to bind to non-target regions.
#' @param allowed.stop.codons  Consider mismatch binding events that induce stop codons.
#' @param allowed.region.definition Definition of the allowed region.
#' @param disallowed.mismatch.pos The number of positions from the primer 3' end where mismatches should not be allowed.
#' All primers binding templates with mismatches within \code{disallowed.mismatch.pos} from the 3' end are disregarded.
#' @param opti.mode Compute optimization constraints and relax delta Tm if necessary.
#' @param required.cvg Target coverage ratio.
#' @param primers.fw Already designed primer sets for the target temperatures given in \code{Tm.brackets}.
#' Used to determine cross-dimerization.
#' @param diagnostic.location Directory for storing results.
#' @param updateProgress Shiny progress callback function.
#' @return Primer data frames for every target temperature.
#' @keywords internal
compute.Tm.sets <- function(primer.df, template.df, Tm.brackets, settings, mode.directionality = c("fw", "rev"), 
    primer_conc, template_conc, na_salt_conc, mg_salt_conc, k_salt_conc, tris_salt_conc, 
    allowed.mismatches, allowed.other.binding.ratio, 
    allowed.stop.codons, allowed.region.definition, disallowed.mismatch.pos, 
    opti.mode = FALSE, required.cvg = NULL, 
    primers.fw = NULL, diagnostic.location = NULL, updateProgress = NULL) {
  
    # determine constraints that need to be computed: consider only primer efficiency
    # and secondary struct from opti constraints
    if (length(mode.directionality) == 0) {
        stop("Please supply the 'mode.directionality' arg.")
    }
    mode.directionality <- match.arg(mode.directionality) 
    # load initial settings:
    opti.constraints <- opti(settings)
    opti.limits <- optiLimits(settings)
    comp.constraints <- NULL  # active constraints for filtering during optimization (annealing-temp dependent)
    if (opti.mode) {
        # compute constraint values for selected constraints
        if ("secondary_structure" %in% names(opti.constraints)) {
            # should be computed in filters now ...
            comp.constraints <- c(comp.constraints, "secondary_structure")
        }
    }
    target.temps <- Tm.brackets$df$target_Tm
    Tm.sets <- vector("list", length(target.temps))
    out.settings <- vector("list", length(target.temps))
    if (opti.mode) {
        message("Determining melting temperature sets:")
    } else {
        message("Checking melting temperature set coverage:")
    }
    for (i in seq_along(target.temps)) {
        # don't parallelize here, it's more important to parallelize constraint computations
        cur.opti.constraints <- opti.constraints
        cur.opti.limits <- opti.limits
        #cat(paste('Opti constraint filtering\n\to Iteration: ', i, '/',
            #length(target.temps), '\n\to Target temperature: ', target.temps[i], 
            #\n\to Constraints: ', paste(comp.constraints, collapse = ','), '\n', sep= ''))
        target.temp <- target.temps[i]
        # compute absolute melting temperature differences
        annealing.temp <- Tm.brackets$df$annealing_temp[i]

        # select Tm set:
        if (!is.na(target.temp) & "melting_temp" %in% colnames(primer.df)) {
            # target is not specified or melting temperature is not available
            cur.sel <- which(abs(primer.df$melting_temp - target.temp) <= (cur.opti.constraints$melting_temp_diff["max"]/2))  # was based on opti limits before, but this was removed so that ILP does not need temperature constraint
        } else {
            cur.sel <- seq_len(nrow(primer.df))  # select all
        }
        # available primers for current temp
        Tm.set <- primer.df[cur.sel, ]
        message("Melting temperature set @ ", target.temps[i], " with ",
                 nrow(Tm.set), " primers (", get_cvg_ratio(Tm.set, template.df, as.char = TRUE), " coverage)")
        if (length(Tm.set) == 0 || nrow(Tm.set) == 0) {
            # set to empty data frame!
            Tm.sets[[i]] <- Tm.set
            out.settings[[i]] <- settings
            next
        }
        if (opti.mode) {
            ###########
            # evaluate and relax constraints
            ############
            # try not to lose any coverage:
            # since we may have just selected 1 set with 'required.cvg' we cannot
            # try to maintain the existing cvg -> at least one set provided by the relaxation should achieve the specified target cvg from the design function
            target.cvg <- min(get_cvg_ratio(primer.df, template.df), required.cvg) # try to attain the required cvg if possible, or less, if we cannot anymore.
            message("\to Computing optimization constraints for target coverage ", paste0(round(target.cvg * 100, 2), "%"))
            # 1. Melting temp difference relaxation:
            while (get_cvg_ratio(Tm.set, template.df) < target.cvg && "melting_temp_diff" %in% names(opti.constraints) && "melting_temp" %in% colnames(primer.df)) {
                # relax Delta Tm
                cur.opti.constraints$melting_temp_diff <- relax.opti.constraints(cur.opti.constraints, opti.limits, opti.constraints)$melting_temp_diff
                cur.opti.limits$melting_temp_diff <- set.new.limits(cur.opti.limits, opti.limits, opti.constraints)$melting_temp_diff

                # create new Tm set according to relaxed DeltaG setting:
                cur.sel <- which(abs(primer.df$melting_temp - target.temp) <= (cur.opti.constraints$melting_temp_diff["max"]/2))
                Tm.set <- primer.df[cur.sel,]
            }
            #message("After melting temp deviation filter: ", get_cvg_ratio(Tm.set, template.df))
            # 2. Computation of cross-terms between fw/rev primers
            if (mode.directionality == "rev" && length(primers.fw) != 0 && 
                    "cross_dimerization" %in% names(cur.opti.constraints)) {
                # consider cross-terms of fw and rev primers when optimizing both fw and rev primers
                ions <- compute.sodium.equivalent.conc(na_salt_conc, mg_salt_conc, 
                                           k_salt_conc, tris_salt_conc)
                G.df <- get.cross.dimers(Tm.set$Reverse, primers.fw[[i]]$Forward, 
                                ions, annealing.temp, 
                                no.structures = TRUE, mode = "asymmetric")
                G.matrix <- create.G.matrix(Tm.set, G.df, primers.fw[[i]])  # min deltaG values of cross-dimerization conformations for every pair of primers
                D <- compute.dimer.matrix(G.matrix, cur.opti.constraints$cross_dimerization)  # 1 for dimers, 0 else
                # filr new Tm.set -> check cvg
                rm.idx <- which(apply(D, 1, function(x) any(x == 1)))
                new.Tm.set <- Tm.set
                if (length(rm.idx) != 0) {
                    new.Tm.set <- new.Tm.set[-rm.idx, ]
                }
                # relax cross term conditions:
                while (get_cvg_ratio(new.Tm.set, template.df) < target.cvg) { # relax and re-evaluate
                    cur.opti.constraints$cross_dimerization <- relax.opti.constraints(cur.opti.constraints, opti.limits, opti.constraints)$cross_dimerization
                    cur.opti.limits$cross_dimerization <- set.new.limits(cur.opti.limits, opti.limits, opti.constraints)$cross_dimerization
                    D <- compute.dimer.matrix(G.matrix, cur.opti.constraints$cross_dimerization)
                    rm.idx <- which(apply(D, 1, function(x) any(x == 1)))
                    new.Tm.set <- Tm.set
                    if (length(rm.idx) != 0) {
                        new.Tm.set <- new.Tm.set[-rm.idx, ]
                    }
                }
                Tm.set <- new.Tm.set # set to filtered df
            }
            #message("After cross dimerization filter: ", get_cvg_ratio(Tm.set, template.df))
            # 3. Computation of other Constraints, e.g. primer efficiency
            # there's no relaxation here. the filtered data is only used if we can still reach the target coverage.
            constraint.df <- compute.constraints(Tm.set, mode.directionality, template.df, settings, updateProgress)
            Tm.set.data <- filter.by.constraints(Tm.set, constraint.df, cur.opti.constraints[comp.constraints], 
                comp.constraints, mode.directionality, template.df)
            new.Tm.set <- Tm.set.data$Filtered
            if (get_cvg_ratio(new.Tm.set, template.df) >= target.cvg ) {
                # only use efficiency-filtered data, if we're at the target cvg, otherwise use the non-filtered data
                Tm.set <- new.Tm.set
            }
            #message("After cross dimerization filter: ", get_cvg_ratio(Tm.set, template.df))
            # write out filtered data set to diagnostic location output info
            if (length(diagnostic.location) != 0) {
                write.csv(Tm.set, file = file.path(diagnostic.location, paste("filtered_opti_data_target_temp_", 
                    target.temp, ".csv", sep = "")), row.names = FALSE)  # write every iteration in case we have to stop the analysis due to time constraints
                write.csv(Tm.set.data$Excluded, file = file.path(diagnostic.location, 
                    paste("filtered_opti_data_excluded_target_temp_", target.temp, ".csv", 
                      sep = "")), row.names = FALSE)  # write every iteration in case we have to stop the analysis due to time constraints
            }
        }
        # output relaxed settings and sets
        cur.settings <- settings
        constraintLimits(cur.settings)[names(cur.opti.limits)] <- cur.opti.limits
        constraints(cur.settings)[names(cur.opti.constraints)] <- cur.opti.constraints
        out.settings[[i]] <- cur.settings
        Tm.sets[[i]] <- Tm.set # store melting temperature-target set
    }
    result <- list("sets" = Tm.sets, "settings" = out.settings)
    return(result)
}
#' Relaxation of Optimization Constraints
#'
#' Relax optimization constraints.
#'
#' @param cur.opti.constraints List with optimization constraint settings.
#' @param initial.opti.limits Initial optimization limits.
#' @param initial.opti.constraints Initial optimization constraints.
#' @return Relaxed optimization constraints.
#' @keywords internal
relax.opti.constraints <- function(cur.opti.constraints, initial.opti.limits, initial.opti.constraints) {
    if (length(cur.opti.constraints) == 0) {
        # nothing to relax ..
        return(NULL)
    }
    cur.opti.constraints <- set.new.limits(cur.opti.constraints, initial.opti.limits, initial.opti.constraints)
    return(cur.opti.constraints)
}

# load templates
fasta.file <- system.file("extdata", "IMGT_data", "templates", 
              "Homo_sapiens_IGH_functional_exon.fasta", package = "openPrimeR")
hdr.structure <- c("ACCESSION", "GROUP", "SPECIES", "FUNCTION")
template.df <- read_templates(fasta.file, hdr.structure, "|", "GROUP") 

# load primers
primer.location <- system.file("extdata", "IMGT_data", "primers", "IGHV", 
                               "Ippolito2012.fasta", package = "openPrimeR")
primer.df <- read_primers(primer.location, "_fw", "_rev")

# load settings
settings.xml <- system.file("extdata", "settings", 
                "C_Taq_PCR_high_stringency.xml", package = "openPrimeR")
settings <- read_settings(settings.xml)

#######
# Functions for checking templates for introduced stop codons due to mismatches
#######

#' Highlight mismatches
#' 
#' Collects information on the mutations present in the input and highlights the
#' mutations in the sequence.
#' 
#' @param seq character vector of the original sequence
#' @param mm.seq  character vector of the mutated sequence 
#' 
#' @return A list highlighting the mutations and additional information 
#' (mutation type, number, etc.)
#' @keywords internal
highlight.mismatch <- function(seq, mm.seq) {
    # search for introduced mutations / stop codons
    #print(seq)
    #print(mm.seq)
    mutation.types <- vector("list", length(seq))
    nbr.mismatches <- rep(NA, length(seq))
    pos <- vector("list", length(seq))  # mismatch posis
    for (i in seq_along(seq)) {
        s <- unlist(strsplit(seq[i], split = ""))
        mm <- unlist(strsplit(mm.seq[i], split = ""))
        if (length(s) != length(mm)) {
            warning("Sequence lengths do not agree. Mutation analysis might be incorrect.")
        }
        idx <- which(s != mm)
        ####################
        # dont search for identical stop codons anymore, leads to problems
        # if the input seq already contains stop codons (no coverage possible)
        #stop.idx <- intersect(which(s == "*"), which(mm == "*"))
        #idx <- c(idx, stop.idx)
        ##############################
        pos[[i]] <- idx
        if (length(idx) > 0) {
            s[idx] <- toupper(s[idx])
            mm[idx] <- toupper(mm[idx])
        }
        seq[i] <- paste(s, collapse = "")
        mm.seq[i] <- paste(mm, collapse = "")
        # determine mutation type
        if (length(idx) == 0) {
            # no AA mutation found -> silent genomic mutation
            mutation.types[[i]] <- "silent"
        } else {
            #print("muts:")
            #print(mm[idx])
            stop.idx <- which(mm[idx] == "*")
            substitution.idx <- setdiff(idx, stop.idx)
            if (length(stop.idx) != 0) {
                # stop codon in AA seq
                mutation.types[[i]] <- "stop_codon"
            } 
            if (length(substitution.idx) != 0) {
                # DNA substitution 
                # else in order to call stop codons as stop codons
                # and not as substitutions (subs are not as bad!)
                mutation.types[[i]] <- c(mutation.types[[i]], "substitution")
            }
        }
        nbr.mismatches[i] <- length(idx)
    }
    result <- list(seq = seq, mm = mm.seq, type = mutation.types, nbr = nbr.mismatches, pos = pos)
    #print(result)
    return(result)
}
#' Identification of ORFs.
#' 
#' Given a template data frame, identify the exon reading frames in the sequences.
#' 
#' @param template.df template data frame.
#' 
#' @return Returns a data frame containing the shift of the ORF (either 0,1, or 2) 
#' for every sequence, as well as a comment in case of problems.
#' @keywords internal
get.ORFs <- function(template.df) {
    nSeqs <- seq_len(nrow(template.df))
    S <- strsplit(template.df$Sequence, split = "")
    ORFs <- rep(NA, length(nSeqs))
    comments <- rep("", length(ORFs))
    frames <- rep(FALSE, length(nSeqs))
    # consider all possible reading frames:
    frames <- 0:2
    k <- NULL
    reading.frames <- foreach (k = seq_along(frames), .combine = "c") %dopar% {
        f <- frames[k] # current reading frame
        seq.aa <- unlist(lapply(S, function(x) paste(seqinr::translate(x, frame = f, sens = "F", 
                                    NAstring = "X", ambiguous = TRUE), collapse = "")))
        sel <- !grepl("\\*", seq.aa) # true if there's no stop codon in the translated seq
        list(data.frame("No_Stop_Codon" = sel, "Translation" = seq.aa, stringsAsFactors = FALSE))
    }
    frame.df <- do.call(cbind, lapply(reading.frames, function(x) x$No_Stop_Codon)) # one column for every ORF
    non.stop.counts <- apply(frame.df, 1, function(x) length(which(x)))
    # if non.stop.count is 1 -> only possible frame found
    # if non.stop.count is 0 -> non-functional seq -> determine right framee
    # if non.stop.count is greater than 1 -> multiple possible frames -> determine right frame
    translations <- rep(NA, nrow(template.df))
    idx <- which(non.stop.counts == 1)
    ORFs[idx] <- apply(frame.df[idx, , drop = FALSE], 1, which) - 1
    # determine the correct ORF for sequences with stop codons / multiple non-stop codon translations
    na.idx <- which(non.stop.counts > 1)
    na.idx.na <- which(non.stop.counts == 0) 
    if (length(na.idx) != 0) {
        comments[na.idx] <- "Multiple frames without stop codons found. Translation frame was chosen based on similarity."
    }
    if (length(na.idx.na) != 0) {
        comments[na.idx.na] <- "All template translation frames contained stop codons. Check your input sequence. Translation frame was chosen based on similarity."
        warning("Templates with IDs: ", paste(template.df$ID[na.idx.na], collapse = ","), ". ", paste0(unique(comments[na.idx.na], collapse = ",")))
    }
    missing.idx <- c(na.idx, na.idx.na) # idx of missing ORF seqs
    # use the existing translations for determining the ORFs for the missing seqs
    seq.aa <- sapply(seq_along(idx), function(x) reading.frames[[ORFs[idx[x]] + 1]]$Translation[idx[x]])
    translations[idx] <- seq.aa
    if (length(seq.aa) != 0) {
        for (k in seq_along(missing.idx)) {
            cur.idx <- missing.idx[k]
            cur.translations <- sapply(reading.frames, function(x) x$Translation[cur.idx])
            # determine best fit to any other correctly translated sequence
            dists <- unlist(lapply(cur.translations, function(test.seq) min(as.vector(stringdist::stringdistmatrix(test.seq, seq.aa, method = "lv")))))
            #print(dists) # for stop codon seq, it may not be clear which frame to choose ..
            min.idx <- which.min(dists)
            ORFs[cur.idx] <- min.idx - 1
            # store best translations:
            if (k %in% na.idx) {
                # multiple frames without stop codon found
                translations[cur.idx] <- cur.translations[min.idx]
            } else {
                # no translation without stop found: give all translations
                translations[cur.idx] <- paste0(cur.translations, collapse = "|")
            }
        }
    }
    # there should be no remaining ORFs with NA or -1, except if seq.aa is empty
    idx <- which(is.na(ORFs))
    if (length(idx) != 0) {
        ORFs[idx] <- 0  # just set first ORF arbitrarily
        warning(paste0(template.df$ID[idx], ". Could not determine the ORFs for all sequences."))
        comments[idx] <- paste0(comments[idx], "ORF unknown: no available seqs for comparison.")
    }
    result <- data.frame(ID = template.df$ID, Frame = ORFs, Comment = comments, Sequence = template.df$Sequence, Translation = translations)
    return(result)
}

#' Adjust ORF position
#' 
#' Adjusts the reading frame according to the position at which we consider a
#' subsequence. 
#' 
#' @param ORFs the reading frames (0,1,2).
#' @param seq.start the position where a sequence is extracted .
#' 
#' @return The adjusted reading frames for the given start positions.
#' @keywords internal
adjust.ORF.start <- function(ORFs, seq.start) {
    frame.offset <- (seq.start - 1 - ORFs)%%3
    new.frame <- (3 - frame.offset)%%3
    return(new.frame)
}

#' Split a sequence 
#'
#' Splits a sequence at a specified positions
#'
#' @param target The target string.
#' @param index The position for the split.
#'
#' @return List with splitted strings
#' @keywords internal
split_str_by_index <- function(target, index) {
    index <- sort(index)
    substr(rep(target, length(index) + 1), start = c(1, index), stop = c(index - 
        1, nchar(target)))
}

#' Interleave strings

#' Combines the input vectors in an interleaved fashion. 
#' @param v1 Input string.
#' @param v2 Input string.
#'
#' @return The interleaved combination of \code{v1} and \code{v2}.
#' @keywords internal
interleave <- function(v1, v2) {
    ord1 <- 2 * (seq_along(v1)) - 1
    ord2 <- 2 * (seq_along(v2))
    result <- c(v1, v2)[order(c(ord1, ord2))]
    return(result)
}
#' String Insertion.
#'
#' Inserts a string into another string at the speficied position.
#'
#' @param target The string to be modified.
#' @param insert The string to be inserted.
#' @param index The position where the insertion should take place.
#'
#' @return A string where \code{insert} is inserted into \code{target} at position \code{index}.
#' @keywords internal
insert_str <- function(target, insert, index) {
      insert <- insert[order(index)]
    index <- sort(index)
    return(paste(interleave(split_str_by_index(target, index), insert), collapse = ""))
}
#' Format mismatches
#' 
#' Formats a sequence for highlighting mismatches in an alignment.
#'
#' @param seq The input sequence.
#' @param pos  The mismatch positions to be formatted.
#' @param format.type Vector of giving the style (bold/italics) for each pos.
#'
#' @return The input sequence with highlighted mismatch positions.
#' @keywords internal
format.seq.ali <- function(seq, pos, format.type) {
       s <- toupper(seq)
    if (length(pos) == 0) {
        # no mismatches
        return(s)
    }
    if (length(format.type) != length(pos)) {
        stop("style length should be the same as format pos length")
    }
    style.start <- rep(NA, length(format.type))
    style.end <- rep(NA, length(format.type))
    idx <- format.type == "bold"
    style.start[idx] <- "<b>"
    style.end[idx] <- "</b>"
    idx <- format.type == "italics"
    style.start[idx] <- "<i>"
    style.end[idx] <- "</i>"
    ins <- unlist(lapply(seq_along(style.start), function(x) c(style.start[x], style.end[x])))
    positions <- unlist(lapply(pos, function(x) c(x, x + 1)))
    s <- insert_str(s, ins, positions)
    return(s)
}
#' Mismatch overview table
#'
#' Computes a table summarizing all of the mismatches caused by the primers in the
#' input data frame.
#' @param primer.df Primer data frame.
#' @param template.df Template data.
#' @param mode.directionality Direction of primers.
#'
#' @return: A data frame summarizing all mismatches of the input primers with the input templates.
#' @keywords internal
compute.mismatch.table <- function(primer.df, template.df, mode.directionality = c("fw", "rev")) {

    if (length(primer.df) == 0 || nrow(primer.df) == 0) {
        return(NULL)
    }
    mismatch.table <- NULL
    seq.idx <- covered.seqs.to.idx(primer.df$Covered_Seqs, template.df)  # seq idx isn't correct for j= 2
    ORF.data <- get.ORFs(template.df)
    ORFs <- ORF.data$Frame
    ali.style <- "<div style = 'font-family: Courier,courier;'>"
    for (j in seq_len(nrow(primer.df))) {
        # message(j)
        primer <- primer.df[j, ]
        primer.seq <- NA
        if (mode.directionality == "fw") {
            primer.seq <- primer$Forward
        } else {
            primer.seq <- primer$Reverse
        }
        if (primer.seq == "" || primer$primer_coverage == 0) {
            # nothing to check :-)
            next
        }
        idx <- seq.idx[[j]]
        ids <- template.df$ID[idx]
        cur.ORF <- ORFs[idx]
        comments <- ORF.data$Comment[idx]
        # retrieve primer-specific ORFs for ORFs that are NA (stop codon might be in
        # another region than the primer-binding region)
        if (mode.directionality == "fw") {
            start.pos <- as.numeric(unlist(strsplit(primer$Binding_Position_Start_fw, 
                split = ",")))
            end.pos <- as.numeric(unlist(strsplit(primer$Binding_Position_End_fw, 
                split = ",")))
            extract.pos <- rep(1, length(start.pos))  # extract string from first position
        } else {
            start.pos <- as.numeric(unlist(strsplit(primer$Binding_Position_Start_rev, 
                split = ",")))
            end.pos <- as.numeric(unlist(strsplit(primer$Binding_Position_End_rev, 
                split = ",")))
            extract.pos <- start.pos  # extract string from start.pos of binding
        }
        seqs.exon.nt <- template.df$Sequence[idx]
        primer.seqs.nt <- seqs.exon.nt
        # incorporate the changes from the primer seqs into the sequences
        is.indel <- ifelse(nchar(substr(primer.seqs.nt, start.pos, end.pos)) != nchar(primer.seq), 
            TRUE, FALSE)  # with indels: this is different than 
        if (mode.directionality == "fw") {
            substr(primer.seqs.nt, start.pos, end.pos) <- primer.seq
        } else {
            substr(primer.seqs.nt, start.pos, end.pos) <- rev.comp.sequence(primer.seq)
        }
        sel <- which(primer.seqs.nt != seqs.exon.nt)  # mismatches were introduced
        if (length(sel) == 0) {
            # nothing to analyze
            next
        }
        primer.seqs.nt <- primer.seqs.nt[sel]
        seqs.exon.nt <- seqs.exon.nt[sel]
        cur.ORF <- cur.ORF[sel]
        ids <- ids[sel]
        primer.id <- primer$ID
        comments <- comments[sel]
        # see if translation has an asterisk
        S <- strsplit(gsub("-", "", seqs.exon.nt), split = "")
        P <- strsplit(gsub("-", "", primer.seqs.nt), split = "")
        seq.aa <- lapply(seq_along(cur.ORF), function(x) seqinr::translate(S[[x]], frame = cur.ORF[x], 
            sens = "F", NAstring = "X", ambiguous = TRUE))
        seq.aa.string <- sapply(seq.aa, function(x) paste(x, collapse = ""))
        idx <- which(sapply(seq.aa, function(x) "*" %in% x))
        cur.ORF <- as.list(cur.ORF)
        if (length(idx) != 0) {
            comments[idx] <- paste(comments[idx], "Best primer binding region translation (", 
                paste(seq.aa.string[idx], sep = ""), ") had a stop codon. Displaying all three translation frames.", 
                sep = "")
            for (x in idx) {
                cur.ORF[[x]] <- 0:2
            }
        }
        highlight.nt <- highlight.mismatch(tolower(seqs.exon.nt), tolower(primer.seqs.nt))
        for (k in seq_along(S)) {
            primer.pos <- regexpr(primer.seq, tolower(highlight.nt$mm[k]))  # pos of primer binding
            primer.pos <- as.numeric(primer.pos):(as.numeric(primer.pos) + 
                          attr(primer.pos, "match.length") - 1)
            style.info <- rep("bold", length(highlight.nt$pos[[k]]))
            form.pos <- highlight.nt$pos[[k]]
            # form.pos <- c(form.pos, primer.pos) style.info <- c(style.info, rep('italics',
            # length(primer.pos)))
            mm.format <- format.seq.ali(highlight.nt$mm[k], form.pos, style.info)
            NT.alignment <- paste(ali.style, format.seq.ali(highlight.nt$seq[k], 
                highlight.nt$pos[[k]], rep("bold", length(highlight.nt$pos[[k]]))), 
                "<br>", mm.format, "</div>", sep = "")
            frames <- cur.ORF[[k]]
            seq.aa <- unlist(lapply(seq_along(frames), function(x) paste(seqinr::translate(S[[k]], 
                frame = frames[x], sens = "F", NAstring = "X", ambiguous = TRUE), 
                collapse = "")))
            primer.aa <- unlist(lapply(seq_along(frames), function(x) paste(seqinr::translate(P[[k]], 
                frame = frames[x], sens = "F", NAstring = "X", ambiguous = TRUE), 
                collapse = "")))
            highlight.aa <- highlight.mismatch(tolower(seq.aa), tolower(primer.aa))
            if (length(highlight.aa$seq) == 1) {
                AA.alignment <- paste(ali.style, format.seq.ali(highlight.aa$seq[1], 
                                highlight.aa$pos[[1]], rep("bold", length(highlight.aa$pos[[1]]))), 
                                "<br>", format.seq.ali(highlight.aa$mm[[1]], highlight.aa$pos[[1]], 
                                rep("bold", length(highlight.aa$pos[[1]]))), "</div>", sep = "")
            } else {
                AA.alignment <- NA
            }
            types <- highlight.aa$type
            out.types <- sapply(types, function(x) paste(x, collapse = ";")) 
            output <- data.frame(Primer = primer.id, Primer_Identifier = as.character(primer$Identifier), 
                                Template = ids[k], Primer_Seq = primer$Forward, 
                                Mutation_Type = out.types,
                                NT_nbr_mm = highlight.nt$nbr[k], AA_nbr_mm = paste(highlight.aa$nbr, 
                                collapse = ";", sep = ""), Seq_NT = highlight.nt$seq[k], 
                                Primer_NT = highlight.nt$mm[k], 
                                Seq_AA = paste(highlight.aa$seq, collapse = ";", sep = ""), 
                                Primer_AA = paste(highlight.aa$mm, collapse = ";", sep = ""), 
                                Comment = comments[k], Alignment_NT = NT.alignment, 
                                Alignment_AA = AA.alignment, stringsAsFactors = FALSE)
            mismatch.table <- rbind(mismatch.table, output)
        }
    }
    return(mismatch.table)
}

#' Template Functionalities.
#'
#' @rdname TemplatesFunctions
#' @name TemplatesFunctions
#'
#' @description
#' \describe{
#' \item{\code{adjust_binding_regions}}{Adjusts the existing annotation 
#' of binding regions by specifying
#' a new binding interval relative to the existing binding region.}
#' \item{\code{assign_binding_regions}}{Assigns the primer target binding 
#' regions to a set of template sequences.}
#' \item{\code{update_template_cvg}}{Annotates the template coverage.}
#' \item{\code{select_regions_by_conservation}}{Computes Shannon entropy
#' for the defined binding regions and determines the most conserved regions.}
#' }
#'
#' @param region.fw Interval of new binding regions relative to the forward binding region defined in \code{template.df}.
#' @param region.rev Interval of new binding regions relative to the reverse binding region defined in \code{template.df}
#' @param template.df An object of class \code{Templates}.
#' @param fw Binding regions for forward primers. Either a numeric interval indicating a uniform
#' binding range relative to the template 5' end or a path to a FASTA file providing
#' binding sequences for every template. If \code{fw} is missing, only
#' \code{rev} is considered.
#' @param rev Binding regions for reverse primers. Either a numeric interval indicating a uniform
#' binding range relative to the template 3' end or the path to a FASTA file providing
#' binding sequences for every template. If \code{rev} is missing,
#' only \code{fw} is considered.
#' @param optimize.region If \code{TRUE}, the binding regions
#' specified via \code{fw} and \code{rev} are 
#' adjusted such that binding regions that may form secondary structures are 
#' avoided. This feature requires ViennaRNA (see notes). If \code{FALSE}
#' (the default), the input binding regions are not modified.
#' @param primer.length A numeric scalar providing the probe length that is used for
#' adjusting the primer binding regions when \code{optimize.region} is \code{TRUE}.
#' @param gap.char The character in the input file representing gaps. 
#' @param primer.df An object of class \code{Primers} containing
#' primers with annotated coverage that are to be used to update 
#' the template coverage in \code{template.df}.
#' @param mode.directionality The directionality of primers/templates.
#' @param win.len The extent of the desired primer binding region.
#' This should be smaller than the \code{allowed.region}. The default is 30.
#' @param by.group Shall the determination of binding regions be stratified
#' according to the groups defined in \code{template.df}. By default,
#' this is set to \code{TRUE}.
NULL

#' @rdname AnalysisStats
#' @name AnalysisStats
#' @return \code{get_comparison_table} returns a data frame summarizing
#' the properties of the provided primer data sets.
#' @export
#' @examples
#' 
#' # Summarize the properties of multiple primer sets
#' data(Comparison)
#' tab <- get_comparison_table(template.data[1:3], primer.data[1:3], "IGH")
get_comparison_table <- function(templates, primers, sample.name = NULL) {
    if (is(templates, "Templates") && is(primers, "Primers")) {
        templates <- list(templates)
        primers <- list(primers)
    } 
    if (!is(templates, "list") || !is(primers, "list")) {
        stop("'templates' and 'primers' should be lists.")
    }
    if (length(templates) != 0) {
        if (length(sample.name) == 0) {
            sample.name <- sapply(templates, function(x) unique(x$Run))
        } else if (length(sample.name) == 1) {
            sample.name <- rep(sample.name, length(templates))
        } else if (length(sample.name) != length(templates)) {
            stop("Incorrect number of sample names provided.")
        }
        if (length(templates) != length(primers)) {
            stop("Please ensure that you provide the same number of templates as primers.")
        }
    }
    template.classes <- sapply(templates, function(x) class(x))
    if (any(template.classes != "Templates")) {
        stop("Input templates should be a list with objects ",
             "of class 'Templates'.")

    }
    primer.classes <- sapply(primers, function(x) class(x))
    if (any(primer.classes != "Primers")) {
        stop("Input primers should be a list with objects ",
             "of class 'Primers'.")
    }
    all.results <- vector("list", length(primers))
    for (j in seq_along(primers)) {
        primer.df <- primers[[j]]
        lex.seq <- templates[[j]]
        set.name <- unique(primer.df$Run)
        mode.directionality <- get.analysis.mode(primer.df)
        set.stats <- primer.set.parameter.stats(primer.df, mode.directionality, lex.seq)
        if (length(set.stats) != 0) {
            cur.result <- data.frame(Primers = set.name, Templates = sample.name[j], stringsAsFactors = FALSE)
            set.stats <- cbind(cur.result, set.stats, row.names = NULL, stringsAsFactors = FALSE)
        }
        all.results[[j]] <- set.stats
    }
    sample.result <- do.call(rbind.fill, all.results)
    # sort sample result by cvg
    if ("Coverage" %in% colnames(sample.result)) {
        hits <- regexpr("[0-9\\.]+%", sample.result$Coverage)
        cvg <- regmatches(sample.result$Coverage, hits)
        cvg <- substr(cvg, 1, nchar(cvg) - 1)
        o <- order(as.numeric(cvg), decreasing = TRUE)
    }
    if (length(sample.result) != 0) {
        # nice labels for columns without underscores:
        colnames(sample.result) <- gsub("_", " ", colnames(sample.result))
    }
    return(sample.result)
}
#' Preparation of Comparison Plot for Evaluation.
#'
#' @param primer.data List with objects of class \code{Primers}. Each 
#' list entry corresponds to a single primer set.
#' @param constraint.settings List with settings for each constraint.
#' If \code{NULL} (the default), use the available evaluation results
#' in \code{primer.data}.
#' @param plot.p.vals Whether p-values from Fisher's exact test
#' should be annotated for every primer set.
#' @return Plot indicating the ratio of primers
#' fulfilling the constraints specified in \code{constraint.settings}
#' for each primer set in \code{primer.data}. 
#' @keywords internal
prepare.constraint.plot <- function(primer.data, 
        constraint.settings, plot.p.vals = FALSE) {
    if (length(primer.data) == 0) {
        return(NULL)
    }
    if (length(constraint.settings) == 0) {
        return(NULL)
    }
    # check type of primers
    primer.classes <- sapply(primer.data, function(x) class(x))
    if (any(primer.classes != "Primers")) {
        stop("Input primers should be lists with objects ",
             "of class 'Primers' ...")
    }
    # determine available columns of the primer sets
    eval.cols <- get.eval.cols(primer.data, constraint.settings)
    if (length(eval.cols) == 0) {
        # nothing to plot
        return(NULL)
    }
    if (length(constraint.settings) != 0) {
        # re-evaluate primer.data with the available input constraints
        eval.names <- gsub("^EVAL_", "", eval.cols)
        primer.data <- eval.comparison.primers(primer.data, constraint.settings[names(constraint.settings) %in% eval.names])
    } 
    # determine frequency of fulfilled constraints
    new.df <- NULL
    p.vals <- rep(NA, length(primer.data))
    for (i in seq_along(primer.data)) {
        primer.df <- asS3(primer.data[[i]])
        if (nrow(primer.df) != 0) {
            if (plot.p.vals) {
                # compute p-vals
                p.data <- primer_significance(primer.df, active.constraints = names(constraint.settings))
                if (length(p.data) != 0) {
                    p.val <- as.numeric(p.data)
                } else {
                    p.val <- NA
                }
                p.vals[i] <- p.val
            } else {
                p.val <- NA
            }
            eval.df <- primer.df[, eval.cols, drop = FALSE]
            if (nrow(eval.df) != 0) {
                df <- data.frame(t(apply(eval.df, 2, function(x) length(which(x))/nrow(eval.df))))
                df <- cbind(Run = unique(primer.df$Run), P_value = p.val, df)
                m.df <- melt(df, c("Run", "P_value"), variable.name = "Constraint", value.name = "Ratio")
            } else {
                m.df <- NULL
            }
            new.df <- rbind(new.df, m.df)
        }
    }
    if (length(new.df) != 0 && length(p.vals) != 0) {
        # adjust/assign p-vals for multiple hypothesis testing
        m <- match(new.df$P_value, p.vals) 
        p.vals <- p.adjust(p.vals, method = "bonf")
        new.df$P_value <- p.vals[m]
    }
    # re-name constraints
    if (length(new.df) != 0) {
        new.df$Constraint <- gsub("EVAL_", "", new.df$Constraint)
        levels <- unique(new.df$Constraint)[order(as.character(unique(new.df$Constraint)))]
        new.df$Constraint <- factor(new.df$Constraint, levels = levels)
        # order Runs by name
        new.df$Run <- factor(new.df$Run, levels = unique(new.df$Run)[order(as.character(unique(new.df$Run)))])
        if (plot.p.vals) {
            p.rep <- format(new.df$P_value, scientific = TRUE, digits = 2)
            new.df$Run <- paste(new.df$Run, " (p-val: ", p.rep, ")", sep = "")
        }
    }
    return(new.df)
}
#' Plot Primer Mismatches
#'
#' Plots primer mismatches for every set.
#'
#' @param primer.data List with primer data frames.
#' @param template.data List with template data frames.
#' @param allowed.mismatches Allowed mismatches.
#' @param highlight.set Primer sets to highlight in the plot.
#' @return Plot of mismatches for comparison.
#' @keywords internal
plot_primer.comparison.mismatches <- function(primer.data, template.data, allowed.mismatches,
        highlight.set = NULL) {
    con.cols <- list("Nbr_of_mismatches_fw", "Nbr_of_mismatches_rev")
    names(con.cols) <- unlist(con.cols)
    con.identifier <- "Number of mismatches"
    new.data <- primer.data
    for (i in seq_along(primer.data)) {
        primer.df <- primer.data[[i]]
        if (any(!"Nbr_of_mismatches_fw" %in% colnames(primer.df))) {
            return(NULL)
        }
        mm.fw <- strsplit(as.character(primer.df$Nbr_of_mismatches_fw), split = ",")
        mm.rev <- strsplit(as.character(primer.df$Nbr_of_mismatches_rev), split = ",")
        r.fw <- do.call(rbind, lapply(seq_along(mm.fw), function(x) primer.df[rep(x, 
            length(mm.fw[[x]])), ]))
        if (length(r.fw) != 0) {
            r.fw$Nbr_of_mismatches_fw <- as.numeric(unlist(mm.fw))
            r.fw$Nbr_of_mismatches_rev <- rep(NA, nrow(r.fw))
        }
        r.rev <- do.call(rbind, lapply(seq_along(mm.rev), function(x) primer.df[rep(x, 
            length(mm.rev[[x]])), ]))
        if (length(r.rev) != 0) {
            r.rev$Nbr_of_mismatches_rev <- as.numeric(unlist(mm.rev))
            r.rev$Nbr_of_mismatches_fw <- rep(NA, nrow(r.rev))
        }
        df <- rbind(r.fw, r.rev)
        if (length(df) != 0 && nrow(df) != 0) {
            new.data[[i]] <- df
        } else {
            # data frame was empty
            new.data[[i]] <- primer.df
        }
    }
    boundaries <- list("Nbr_of_mismatches_fw" = allowed.mismatches,
                        "Nbr_of_mismatches_rev" = allowed.mismatches)
    p <- plot_primer.comparison.box(new.data, con.identifier, con.cols, 
        boundaries, highlight.set = highlight.set)
    return(p)
}

#' Boxplot for Primer Comparison
#'
#' Constructs a box plot showing constraint values for each primer set.
#'
#' @param primer.data List with primer data frames.
#' @param con.identifier Identifier of constraint to be plotted.
#' @param con.cols Column names with the constraint values in the primer data frames.
#' @param boundaries List with constraint settings.
#' @param y.limits Limits for the extent of the y-axis.
#' @param show.points If \code{TRUE} (the default), individual data points
#' are visualized in the boxplot, otherwise they are not shown.
#' @param highlight.set The identifier of a primer set to highlight in the plot.
#' @param nfacets A numeric providing the number of facet columns to show.
#' By default \code{nfacets} is \code{NULL} such that the number of 
#' facet columns is chosen automatically.
#' @return A boxplot for primer comparison.
#' @keywords internal
plot_primer.comparison.box <- function(primer.data,
    con.identifier, con.cols, boundaries, y.limits = NULL, 
    show.points = TRUE, highlight.set = NULL, nfacets = NULL) {

    if (length(primer.data) == 0) {
        return(NULL)
    }
    if (length(con.cols) == 0) {
        # no constraints to be plotted according to 'con.cols'
        return(NULL)
    }
    #print("plot_primer.box: nbr facet cols is:")
    #print(nfacets)
    plot.df <- rbind.primer.data(primer.data)
    if (length(plot.df) == 0) {
        return(NULL)
    }
    if (any(unlist(lapply(con.cols, function(x) !x %in% colnames(plot.df))))) {
        return(NULL)
    }
    plot.df <- melt(plot.df, id.vars = c("ID", "Run", "Forward", "Reverse"), 
        measure.vars = unlist(con.cols),
        variable.name = "Constraint", 
        value.name = "Value")
    # check whether there's any variable available for plotting ...
    if (length(highlight.set) != 0) {
        # check whether highlight set is specified correctly
        m <- match(highlight.set, plot.df$Run)
        na.idx <- which(is.na(m))
        if (length(na.idx) != 0) {
            msg <- paste("Highlight set not available in data:",
                paste(highlight.set[na.idx], collapse = ","))
            warning(msg)
            highlight.set <- highlight.set[!is.na(m)]
        }
    }
    # Select only fw values for fw primers etc.
    fw.idx <- which(plot.df$Forward != "")
    rev.idx <- which(plot.df$Reverse != "")
    plot.df$Direction <- ifelse(grepl("_fw", plot.df$Constraint), "Forward", ifelse(grepl("_rev", plot.df$Constraint), "Reverse", "Overall"))
    plot.df.fw <- plot.df[plot.df$Direction == "Forward" & plot.df$ID %in% plot.df$ID[fw.idx],]
    plot.df.rev <- plot.df[plot.df$Direction == "Reverse" & plot.df$ID %in% plot.df$ID[rev.idx],]
    plot.df <- rbind(plot.df.fw, plot.df.rev, plot.df[plot.df$Direction == "Overall",])
    ########
    con.names <- unlist(lapply(plot.df$Constraint, function(x) names(con.cols)[sapply(seq_along(con.cols), function(y) any(x %in% con.cols[[y]]))]))
    con.levels <- unique(con.names)[order(unique(con.names))]
    constraint.names <- unlist(constraints_to_unit(con.levels, TRUE))
    plot.df$Constraint <- factor(con.names, levels = con.levels,
                                labels = sapply(constraint.names, deparse))
    #######
    #plot.df$Constraint <- gsub("_fw", "", plot.df$Constraint)
    #plot.df$Constraint <- gsub("_rev", "", plot.df$Constraint)
    title <- "Constraints"
    boundary.data <- NULL # store plot boundaries
    cons <- as.character(levels(plot.df$Constraint))
    #mod.names <- unlist(lapply(con.cols, function(x) gsub("_fw", "", x[1])))
    for (i in seq_along(cons)) {
        con <- cons[i] 
        con.name <- con.levels[i]
        bound <- boundaries[[con.name]]
        if (length(bound) == 0) {
            warning(paste("Constraint boundary for ", con.name, " is not available.", sep = ""))
            next
        }
        cur.bound <- data.frame(Constraint = rep(con, length(bound)),
                                Z = bound)
        boundary.data <- rbind(boundary.data, cur.bound)
    }
    plot.df$Run <- abbreviate(plot.df$Run, getOption("openPrimeR.plot_abbrev"))
    plot.df$Run <- factor(plot.df$Run, levels = unique(plot.df$Run)[order(as.character(unique(plot.df$Run)))])
    p <- ggplot(plot.df, aes(x = .data[["Run"]], y = .data[["Value"]], colour = .data[["Run"]])) + 
        theme(axis.text.x = element_text(angle = 90, hjust = 1)) +
        ggtitle(title) + ylab("Value") +
        geom_boxplot(outlier.shape = NA) +
        guides(colour = "none") # remove legend: is redundant here
    if (length(con.identifier) > 1) {
        # multiple constraints are plotted
        p <- p + facet_wrap(as.formula("~Constraint"), 
                scales = "free_y", ncol = nfacets, 
                labeller = ggplot2::label_parsed)
    } else {
        # a single constraint is plotted
        p <- p + ylab(con.identifier[[1]])
    }
    if (show.points) {
        p <- p + geom_point(alpha = 0.25, position = position_jitter(width = 0.25, 
            height = 0))
    }
    # set colors:
    pal <- getOption("openPrimeR.plot_colors")["Run"] # the RColorBrewer palette to use
    colors <- colorRampPalette(brewer.pal(8, pal))(length(unique(plot.df$Run)))
    p <- p + scale_colour_manual(values = colors)
    if (length(boundary.data) != 0) {
        p <- p + geom_hline(data = boundary.data, aes(yintercept = .data[["Z"]]),
                    linewidth = 0.25, 
                    colour = "red", linetype = 2)
    }
    if (length(y.limits) != 0) {
        p <- p + coord_cartesian(ylim = y.limits)  # don't lose bars outside of limits
    }
    annotation.constraints <- c("cross_dimerization", "self_dimerization", "secondary_structure")
    if (length(con.cols) == 1 && names(con.cols)[1] %in% annotation.constraints) {
        # requires library grid
        if (names(con.cols)[1] %in% c("cross_dimerization", "self_dimerization")) {
            text.high <- "Dimerization\nunlikely"
            text.low <- "Dimerization\nlikely"
        } else if (names(con.cols)[1] == "secondary_structure") {
            text.high <- "Secondary structure\nunlikely"
            text.low <- "Secondary structure\nlikely"
        }
        label.col <- "grey20"
        text_high <- grid::textGrob(text.high, gp=grid::gpar(fontsize=13, fontface="bold", col = label.col ))
        text_low <- grid::textGrob(text.low, gp=grid::gpar(fontsize=13, fontface="bold", col = label.col))
        x.pos <- -1 # position on the x axis of annotation
        high.pos <- max(plot.df$Value, na.rm = TRUE) + 0.1 * min(plot.df$Value, na.rm=TRUE)
        low.pos <- min(plot.df$Value, na.rm = TRUE) - 0.1 * min(plot.df$Value, na.rm= TRUE)
        #######
        p <- p + theme(plot.margin = unit(c(1,1,1,6), "lines")) +
        annotation_custom(text_high, ymin = high.pos, ymax = high.pos, xmin=x.pos,xmax=x.pos) + 
        annotation_custom(text_low,ymin = low.pos, ymax = low.pos,xmin=x.pos,xmax=x.pos) +
        scale_x_discrete(expand = c(0.1,0.0)) +
        geom_segment(aes(x = x.pos, y = high.pos + 0.25 * min(plot.df$Value, na.rm = TRUE), xend = x.pos, 
                    yend = low.pos - 0.25 * min(plot.df$Value, na.rm = TRUE)), color = label.col, size = 1.25,
                    arrow = arrow(length = unit(0.03, "npc")))
    }
    if (length(highlight.set) != 0) {
        # highlight selected sets
        sel <- levels(plot.df$Run) %in% highlight.set
        p <- p + theme(
                axis.text.x = element_text(
                    face=ifelse(sel, "bold","plain"),
                    colour = ifelse(sel, 
                            "grey20", "grey30")))
                    #size = ifelse(sel, 
                            #x.labels.size + 1, x.labels.size)))
    }
    return(p)
}
#' Comparison Coverage Stats.
#'
#' Computes coverage stats for primer comparison.
#'
#' @param primer.data List with primer data frames.
#' @param template.data List with template data frames.
#' @return Coverage statistics for comparing primers.
#' @keywords internal
comparison.cvg <- function(primer.data, template.data) {
    info <- comparison.stats.raw(primer.data, template.data)
    if (length(info) != 0) {
        cvg <- paste(round(info$Coverage * 100, 2), "%", sep = "")
        cvg.string <- paste(info$Nbr_Primers, " primers cover ", 
                    info$Nbr_Covered, 
                    " of ", info$Nbr_Templates, " (", cvg, 
                    ") templates", sep = "")
        result <- data.frame(Run = info$Run, 
                    Coverage = round(info$Coverage, 3), 
                    Info = cvg.string, stringsAsFactors = FALSE)
        return(result)
    } else {
        return(NULL)
    }
}
#' Computation of Raw Stats for Primer Comparison
#' 
#' Computes raw stats for primer comparison.
#'
#' @param primer.data List with primer data frames.
#' @param template.data List with template data frames.
#' @return Raw statistics for primer comparison.
#' @keywords internal
comparison.stats.raw <- function(primer.data, template.data) {
    if (length(primer.data) == 0 || length(template.data) == 0 ||
        is.null(template.data[[1]])) {
        return(NULL)
    }
    run.names <- get.run.names(primer.data)
    N <- unlist(lapply(primer.data, function(x) nrow(x)))
    N.templates <- unlist(lapply(template.data, function(x) nrow(x)))
    cvg.info <-  cvg <- lapply(seq_along(primer.data), function(x) get_cvg_ratio(primer.data[[x]], 
        template.data[[x]]))
    nbr.covered <- unlist(lapply(cvg.info, function(x) attr(x, "no_covered")))
    cvg <- unlist(lapply(cvg.info, function(x) as.numeric(x)))
    o <- order(cvg, decreasing = TRUE)
    if (length(cvg) == length(run.names)) {
        # if we have enough data (primers available for each set)
        info <- data.frame(Run = run.names, Coverage = cvg, Nbr_Primers = N, Nbr_Templates = N.templates, 
            Nbr_Covered = nbr.covered, stringsAsFactors = FALSE)
        info <- info[o, ]
        return(info)
    } else {
        return(NULL)
    }
}
build.comparison.table <- function(primers, templates) {
    primers <- set.run.names(primers) # ensure that primer run names are unique
    primer.runs <- get.run.names(primers)
    template.runs <- get.run.names(templates)
    m <- which.max(c(length(primer.runs), length(template.runs)))
    if (m == 1) { # more primer sets than templates
        template.runs <- c(template.runs, rep(NA, length(primer.runs) - length(template.runs)))
    } else { # more templates than primer sets
        primer.runs <- c(primer.runs, rep(NA, length(template.runs) - length(primer.runs)))
    }
    out <- data.frame(PrimerSet = primer.runs, 
                      Templates = template.runs) 
    # numeric index is important for selection!
    rownames(out) <- seq_len(nrow(out))
    if (length(templates) != length(primers)) {
        templates <- replicate(length(primers), templates[1])
    }
    cvg.data <- comparison.cvg(primers, templates)
    my.ids <- seq_along(primers)
    if (length(cvg.data) != 0) {
        # match out to cvg.data -> sorted
        m <- match(as.character(out$PrimerSet), as.character(cvg.data$Run)) # assumes that IDs are unique!!!
        add.df <- cvg.data[m,]
        out <- cbind(out, Coverage = add.df$Coverage, Info = add.df$Info)
        # sort by highest coverage
        o <- order(out$Coverage, decreasing = TRUE)
        out <- out[o, ]
        out$Coverage <- paste0(round(out$Coverage * 100, 2), "%")
    }
    return(out)
}

context("Primer Comparison") 

test_that("primer_comparison", {
    # check that the filtering procedure excludes the correct primers
    data(Comparison)
    primer.data <- primer.data[1:3]
    template.data <- template.data[1:3]
    constraint.settings <- constraints(settings)
    constraints <- names(constraint.settings)
    # create basic plots etc.
    p <- plot_constraint_deviation(primer.data, settings)
    p <- plot_primer_binding_regions(primer.data, template.data)
    p <- plot_constraint(primer.data, settings, constraints)
    p <- plot_cvg_constraints(primer.data, settings)
    p <- plot_constraint_fulfillment(primer.data, settings)
    p <- plot_template_cvg(primer.data, template.data)
    tab <- get_comparison_table(template.data, primer.data)
    ######
    # test what happens if a template set is lacking required objects
    #######
    # missing coverage
    primer.data[[1]] <- primer.data[[1]][, !colnames(primer.data[[1]]) %in% c("Covered_Seqs", "primer_coverage")]
    p <- expect_warning(plot_template_cvg(primer.data, template.data)) # ok :-) 
    p <- expect_error(plot_primer_binding_regions(primer.data, template.data), ".*primer coverage.*")
    tab <- expect_warning(get_comparison_table(template.data, primer.data)) # ok 
    # missing constraint
    primer.data[[1]] <- primer.data[[1]][, !colnames(primer.data[[1]]) %in% c("melting_temp")]
    p <- expect_warning(plot_constraint_deviation(primer.data, settings)) # ok
    p <- expect_warning(plot_constraint(primer.data, settings, "melting_temp_range")) # ok
    p <- expect_warning(plot_constraint_fulfillment(primer.data, settings))
})

#' Primer Analysis Statistics.
#'
#' @rdname AnalysisStats
#' @name AnalysisStats
#'
#' @description
#' \describe{
#' \item{\code{get_cvg_ratio}}{Determines the ratio of template sequences 
#' that are covered by the evaluated input primers. The ratio
#' is in the interval [0,1] where 0 indicates 0\% coverage (no templates
#' covered) and 1 indicates 100\% coverage (all templates covered).}
#' \item{\code{get_cvg_stats}}{Retrieve statistics on covered templates, 
#' either for a single or multiple primer sets.}
#' \item{\code{get_cvg_stats_primer}}{Creates a table summarizing the 
#' coverage events of individual primers.}
#' \item{\code{get_comparison_table}}{Creates an overview of the
#' properties of multiple primer sets by providing
#' the inter-quartile range of primer properties in bracket notation.}
#' }
#' 
#' @param primer.df A \code{Primers} object containing the primers.
#' @param template.df A \code{Templates} object containing
#' the template sequences corresponding to \code{primer.df}. 
#' @param allowed.mismatches The number of allowed mismatches
#' for determining the coverage of the templates. By default,
#' all annotated coverage events are considered.
#' @param mode.directionality If \code{mode.directionality} is provided,
#' the coverage of templates is computed for a specific direction of primers.
#' Either "fw" (forward coverage only), "rev" (reverse coverage only), or "both" for both directions. By default, \code{mode.directionality} is \code{NULL}
#' such that the directionality of the primers is determined automatically.
#' @param as.char Whether the coverage ratio should
#' be outputted as a percentage-formatted character vector. By default,
#' \code{as.char} is set to \code{FALSE} such that a numeric is returned.
#' @param primers To retrieve statistics for a single primer set, 
#' please provide an object of class \code{Primers} containing a set of
#' evaluated primers.
#' To retrieve statistics for multiple primer sets, pelase provide
#' a list with evaluated \code{Primers} objects.
#' @param templates If \code{primers} is an object of class \code{Primers},
#' please provide an object of class \code{Templates} containing the
#' template sequences targeted by \code{primers}. If \code{primers} is a list,
#' \code{templates} should be a list of \code{Template} objects.
#' @param for.viewing Whether the table should be formatted
#' to be human-readable. By default, \code{for.viewing} is \code{FALSE}.
#' @param total.percentages Whether group coverage percentages
#' should be computed in relation to the total number of template sequences
#' or in relation to the number of templates belonging to a specific group.
#' By default, \code{total.percentages} is \code{FALSE} suc that the
#' percentages are group-specific.
#' @param cvg.definition If \code{cvg.definition} is set to
#' "constrained", the statistics for the expected
#' coverage (after applying the coverage constraints) are retrieved.
#' If \code{cvg.definition} is set to "basic", the coverage is determined 
#' solely by string matching (i.e. without applying the coverage constraints).
#' By default, \code{cvg.definition} is set to "constrained".
#' @param sample.name Either a single identifier or a character vector 
#' of identifiers for every
#' \code{Templates} object in \code{templates}.
#' By default, \code{sample.name} is \code{NULL} such that the
#' \code{Run} annotations in the provided \code{Templates} objects are used.
NULL

########
# Constraint: cross dimerization
########

#' Self dimerization
#' 
#' Computes possible self-dimers.
#'
#' @param primers.1 Input primers
#' @param primers.2 (Copy/reverse) of the input primers 
#' @param ions Sodium-equivalent ionic concentration.
#' @param annealing.temp The annealing temperature.
#' @param no.structures Whether the dimerization structure shall be computed.
#' @return Possible self-dimer conformations.
#' @keywords internal
get.self.dimers <- function(primers.1, primers.2, ions, annealing.temp,
                            no.structures = FALSE) {
    # pairs (fw - fw, fw-rev, rev-rev)
    if (length(primers.1) != length(primers.2)) {
        stop("Both primer vectors should have the same length.")
    }
    idx <- which(primers.1 != "" & primers.2 != "")
    dimer.data <- get.dimer.data(primers.1[idx], primers.2[idx],
                                  annealing.temp, ions, no.structures = no.structures)
    # return the correct indices:
    dimer.data$Idx1 <- idx[dimer.data$Idx1]
    dimer.data$Idx2 <- idx[dimer.data$Idx2]
    return(dimer.data)
}

#' Cross dimers
#'
#' Computes all possible primer cross-dimers.
#' 
#' @param primers.1 Input primers.
#' @param primers.2 Input primers.
#' @param ions Sodium-equivalent ionic concentration.
#' @param annealing.temp The PCR annealing temperature.
#' @param check.idx indices of primers for checking cross-dimerization
#' @param no.structures Whether to compute structures of dimers.
#' @param mode 'symmetric', if \code{primers.1} and \code{primers.2}
#' carry the same information (i.e. fw-fw, rev-rev, fw-rev), 'asymetric' else.
#'
#' @return Data frame with potential cross dimers.
#' @keywords internal
get.cross.dimers <- function(primers.1, primers.2, ions, 
                            annealing.temp, check.idx = NULL, 
                            no.structures = FALSE,
                            mode = c("symmetric", "asymmetric")) {
    mode <- match.arg(mode) 
    if (length(primers.1) == 0 || length(primers.2) == 0) {
        return(NULL)
    }
    if (length(annealing.temp) == 1) {
        annealing.temp <- rep(annealing.temp, length(primers.1))
    }
    #message("Creating dimer combinations ...")
    if (mode == "symmetric") {
        # fewer options for combinations: can exclude self-dimers, as well as already
        # considered combinations (e.g. if we have 1-2, we don't need to 2-1 if 2 in the
        # right set and 2 in the left set correspond to each other).
        combis <- do.call("rbind", parallel::mclapply(seq_along(primers.1), function(x) expand.grid(x, 
            seq_along(primers.2)[-(seq_len(x))])))
    } else {
        # all combinations
        combis <- do.call("rbind", parallel::mclapply(seq_along(primers.1), function(x) expand.grid(x, 
            seq_along(primers.2))))
    }
    if (length(check.idx) != 0) {
        combi.idx <- which(apply(combis, 1, function(x) any(x %in% check.idx)))
        combis <- combis[combi.idx, ]
    }
    idx <- which(primers.1[combis[, 1]] != "" & primers.2[combis[, 2]] != "")  # only consider primer pairings representing actual sequences
    if (length(idx) == 0) {
        return(NULL)
    }
    combis <- combis[idx, ]
    # for each possible cross-dimer use the minimal annealing temp of the pairs
    use.Ta <- unlist(lapply(seq_len(nrow(combis)), function(x)
                    min(annealing.temp[combis[x,1]], annealing.temp[combis[x,2]])))
    result <- get.dimer.data(primers.1[combis[, 1]], primers.2[combis[, 2]], use.Ta, ions, no.structures)
    result <- result[, c("DeltaG", "Structure", "Idx1", "Idx2")]
    # modify indices from internal index to used index
    result$Idx1 <- combis[, 1][result$Idx1]
    result$Idx2 <- combis[, 2][result$Idx2]
    return(result)
}

#' Create free energy matrix
#'
#' Creates a matrix giving the deltaG values of all primers.
#'
#' @param primer.df Primer data frame.
#' @param G.df Free energy data for the primers.
#' @param primer.df.2 Optional second primer data frame
#'
#' @return Matrix with the smallest free dimerization energy for every primer pair.
#' @keywords internal
create.G.matrix <- function(primer.df, G.df, primer.df.2 = NULL) {
    # asymmetric: when primer.df.2 represents some other primers that aren't matched to primer.df (during optimization)
    mode <- "asymmetric"  
    if (length(primer.df.2) == 0) {
        primer.df.2 <- primer.df
        mode <- "symmetric"
    }
    if (length(G.df) == 0) {
        # no possible dimerizations detected
        out <- matrix(rep(0, nrow(primer.df) * nrow(primer.df.2)), nrow = nrow(primer.df), 
            ncol = nrow(primer.df.2))
        return(out)
    }
    G <- matrix(rep(0, nrow(primer.df) * nrow(primer.df.2)), nrow = nrow(primer.df), 
        ncol = nrow(primer.df.2))
    # write entries
    G[cbind(G.df$Idx1, G.df$Idx2)] <- G.df$DeltaG
    if (mode == "symmetric") {
        G[cbind(G.df$Idx2, G.df$Idx1)] <- G.df$DeltaG
    }
    return(G)
}

#' Dimerization matrix
#'
#' Computes a matrix indicating all dimerizing primers according to a DeltaG cutoff.
#'
#' @param G Matrix with free energies of all considered primer pairs.
#' @param deltaG.cutoff Primers with free energies below the cutoff are considered dimerizing.
#'
#' @return Binary matrix with dimerization events according to the \code{deltaG.cutoff}.
#' Contains '1' if primers (i,j) dimerize and '0' else.
#' @keywords internal
compute.dimer.matrix <- function(G, deltaG.cutoff = -7) {
    D <- G
    D[G <= deltaG.cutoff] <- 1  # primers i and j dimerize
    D[G > deltaG.cutoff] <- 0  # primer i and j don't dimerize
    # print some infos
    nbr.pairs <- dim(G)[1] * dim(G)[1]
    nbr.dimers <- length(which(D == 1))
    ratio.dimers <- nbr.dimers/(dim(G)[1] * dim(G)[1])
    if (nrow(D) == 0 || all(D == 0)) {
        message("No dimerizing primers detected :-)")
    } else {
        dimerizing.primers <- length(unique(which(D == 1, arr.ind = TRUE)[, 1]))
        cat(paste("Dimerization info: \n\to DeltaG cutoff: ", deltaG.cutoff, "\n\to Number of pairs: ", 
            nbr.pairs, "\n\to Number of dimers: ", nbr.dimers, " (", round(ratio.dimers * 
                100, 2), "%)\n\to Number of dimerizing primers: ", dimerizing.primers, 
            "\n", sep = ""))
    }
    return(D)
}
#' Self dimers
#'
#' Computes all possible self dimers for the primers in the input data frame.
#'
#' @param primer.df Input primer data frame.
#' @param primer_conc Primer concentration.
#' @param na_salt_conc Sodium ion concentration.
#' @param mg_salt_conc Magensium ion concentration.
#' @param k_salt_conc Potassium ion concentration.
#' @param tris_salt_conc Tris buffer concentration.
#' @param annealing.temp The PCR annealing temperature.
#' @param for.shiny Whether the output is to be formatted for HTML.
#' @param no.structures Whether dimerization structures shall be outputted.
#' @return Data frame with thermodynamic information on all self dimers.
#' @keywords internal
compute.all.self.dimers <- function(primer.df, primer_conc, na_salt_conc, 
                     mg_salt_conc, k_salt_conc, tris_salt_conc, annealing.temp, for.shiny = FALSE, no.structures = FALSE) {
    # computes data frame with deltaG values of possible primer self-dimers considers
    # all directions: a primer can bind fw-fw, fw-rev, rev-rev
    #message("Computing self dimers @ ", annealing.temp)
    ions <- compute.sodium.equivalent.conc(na_salt_conc, mg_salt_conc, 
                                           k_salt_conc, tris_salt_conc)

    fw.fw <- get.self.dimers(primer.df$Forward, primer.df$Forward, 
                             ions, annealing.temp, no.structures = no.structures)  # fw-fw primer dimers
    if (length(fw.fw) != 0) {
        fw.fw$Direction <- "fw-fw"
    }
    rev.rev <- get.self.dimers(primer.df$Reverse, primer.df$Reverse, 
                               ions, annealing.temp, no.structures = no.structures)  # rev-rev primer dimers
    if (length(rev.rev) != 0) {
        rev.rev$Direction <- "rev-rev"
    }
    fw.rev <- get.self.dimers(primer.df$Forward, primer.df$Reverse, 
                              ions, annealing.temp, no.structures = no.structures)  # fw-rev primer dimers
    if (length(fw.rev) != 0) {
        fw.rev$Direction <- "fw-rev"
    }
    results <- rbind(fw.fw, rev.rev, fw.rev)
    if (length(results) != 0) {
        # remove NAs and add ID
        results <- na.omit(results)
        # results$Idx1 is missing :o
        results$ID <- primer.df$ID[results$Idx1]
    }
    if (for.shiny) {
        results <- view.dimer.df(results, "Self")
    }
    return(results)
}

#' Selection of cross dimer index
#'
#' Select the index with the smallest DeltaG value.
#'
#' @param deltaG Data frame with thermodynamic info.
#' @param primers The corresponding primers.
#'
#' @return The indices with smallest DeltaG for every primer.
#' @keywords internal
select.min.cross.idx <- function(deltaG, primers) {
    # select smallest deltaG idx from data frame of interaction terms, in the order
    # of the primers (NA if nothing is available)
    indices <- seq_along(primers)
    sel.idx <- unlist(lapply(indices, function(x) {
        idx <- which(deltaG$Idx1 == x | deltaG$Idx2 == x)
        if (length(idx) == 0) {
            return(NA)
        }
        min.idx <- which.min(deltaG$DeltaG[idx])
        ret <- idx[min.idx]
    }))
    return(sel.idx)
}
#' Cross dimerization
#'
#' Compute DeltaG data frame for possible primer cross-dimers.
#'
#' @param primer.df Input primers data frame.
#' @param primer_conc Primer concentration.
#' @param na_salt_conc Sodium ion concentration.
#' @param mg_salt_conc Magensium ion concentration.
#' @param k_salt_conc Potassium ion concentration.
#' @param tris_salt_conc Tris buffer concentration.
#' @param annealing.temp The PCR annealing temperature.
#' @param check.idx Indices of primers for checking cross-dimerization.
#' @param for.shiny Whether to format for HTML output.
#' @param no.structures Whether dimer structures shall not be determined.
#' If \code{TRUE}, structures are not computed resulting in faster runtimes.
#' @return All cross dimers.
#' @keywords internal
# p.df <- do.call(my_rbind, replicate(30, primer.df, simplify = FALSE))
# X <- compute.all.cross.dimers.unfiltered(p.df, primer_conc, na_salt_conc, mg_salt_conc, k_salt_conc, tris_salt_conc, annealing.temp, check.idx = NULL, for.shiny = FALSE, no.structures = FALSE)
compute.all.cross.dimers.unfiltered <- function(primer.df, primer_conc, na_salt_conc, 
    mg_salt_conc, k_salt_conc, tris_salt_conc, annealing.temp, check.idx = NULL, for.shiny = FALSE, no.structures = FALSE) {
    # computes all combinations of cross dimers: fw-fw, rev-rev, fw-rev
    # check.idx: only compute cross-dimers for a subset of primers
     ions <- compute.sodium.equivalent.conc(na_salt_conc, mg_salt_conc, 
                                           k_salt_conc, tris_salt_conc)
    # message("Computing cross dimers @ ", annealing.temp)
    # fw-fw dimers
    fw.fw <- get.cross.dimers(primer.df$Forward, primer.df$Forward, 
                              ions, annealing.temp, check.idx, no.structures)
    if (length(fw.fw) != 0) {
        fw.fw$Direction <- "fw-fw"
    }
    # rev-rev dimers
    rev.rev <- get.cross.dimers(primer.df$Reverse, primer.df$Reverse, 
                                ions, annealing.temp, check.idx, no.structures)
    if (length(rev.rev) != 0) {
        rev.rev$Direction <- "rev-rev"
    }
    # fw-rev dimers
    fw.rev <- get.cross.dimers(primer.df$Forward, primer.df$Reverse, 
                               ions, annealing.temp, check.idx, no.structures)
    if (length(fw.rev) != 0) {
        fw.rev$Direction <- "fw-rev"
    }
    results <- rbind(fw.fw, rev.rev, fw.rev)
    if (length(results) != 0) {
        # remove NAs and add ID
        results <- na.omit(results)
        results$Primer_1 <- as.character(primer.df$ID[results$Idx1])
        results$Primer_2 <- as.character(primer.df$ID[results$Idx2])
    }
    if (for.shiny) {
        results <- view.dimer.df(results, "Cross")
    }
    return(results)
}
#' Cross dimerization
#'
#' Compute worst-case DeltaG data frame with all possible primer cross-dimers.
#'
#' @param primer.df Input primers data frame.
#' @param primer_conc Primer concentration.
#' @param na_salt_conc Sodium ion concentration.
#' @param mg_salt_conc Magensium ion concentration.
#' @param k_salt_conc Potassium ion concentration.
#' @param tris_salt_conc Tris buffer concentration.
#' @param annealing.temp The PCR annealing temperature.
#' @param results (optional) Cross dimer data frame (unfiltered)
#' @param check.idx Indices of primers for checking cross-dimerization.
#' @param for.shiny Whether the table is inteded for HTML display.
#' @param no.structures Whether dimerization structures shall not be outputted.
#' @return Worst-case cross dimers.
#' @keywords internal
compute.all.cross.dimers <- function(primer.df, primer_conc, na_salt_conc, mg_salt_conc, 
    k_salt_conc, tris_salt_conc, annealing.temp, results = NULL, check.idx = NULL, for.shiny = FALSE, no.structures = FALSE) {
    if (length(results) == 0) {
        results <- compute.all.cross.dimers.unfiltered(primer.df, primer_conc, 
                        na_salt_conc, mg_salt_conc, k_salt_conc, tris_salt_conc, 
                        annealing.temp, check.idx, no.structures = no.structures)
    }
    if (length(results) == 0) {
        # no cross dimers found
        return(NULL)
    } 
    # filter out duplicate conformations for the same primer pair
    # replaced plyr call with dplyr call for speed for large matrices ..
    #result <- ddply(results, c("Idx1", "Idx2"), function(x) arrange(x, substitute(DeltaG))[1, ])
    ######
    # slicing info:
    # https://github.com/tidyverse/dplyr/issues/3034
    #########
    result <- as.data.frame(results %>% 
                            dplyr::group_by(.dots = c("Idx1", "Idx2")) %>% 
                            dplyr::slice(which.min(!! quote(DeltaG)))) # !! is inline operator to 'unquote' an expression
    if (for.shiny) {
        result <- view.dimer.df(result, "Cross")
    }
    return(result)
}

#' Formats a Dimerization Structure for HTML.
#' @param structures A character vector of dimerization structures.
#' @return HTML-formatted character vectors.
#' @keywords internal
html.format.structure <- function(structures) {
    # check if structures are already formatted
    if (length(structures) == 0) {
        return(structures)
    }
    if (grepl("<div", structures[[1]])) {
        # already formatted!
        return(structures)
    }
    # format structures for non-NA indices only
    idx <- which(!is.na(structures))
    pre <- "<div class = 'verbatim'>"
    post <- "</div>"
    s <- strsplit(structures[idx], "\n")
    s <- unlist(lapply(s, function(x) {
        if (length(x) != 0) {
            paste0("<b>5'-</b>", x[[1]], "<b>-3'</b>", "<br>", 
                paste0("   ", x[[2]]), 
                "<br>", "<b>3'-</b>", x[[3]], "<b>-5'</b>")
        } else {
            ""
        }
    }))
    out <- paste0(pre, s, post)
    result <- rep(NA, length(structures))
    result[idx] <- out
    return(out)
}
#' Formatted dimerization data.
#'
#' Format a dimerization data frame for frontend output.
#'
#' @param dimers Dimerization data frame.
#' @param type Type of dimerization.
#'
#' @return A data frame whose columns are formatted in a user-readable way.
#' @keywords internal
view.dimer.df <- function(dimers, type = c("Self", "Cross")) {
    if (length(type) == 0) {
        stop("Please supply a 'type' argument.")
    }
    type <- match.arg(type)
    out <- dimers
    if (type == "Self") {
        struct.col <- "Self_Dimer_Structure"
    } else {
        struct.col <- "Cross_Dimer_Structure"
    }
    if (!struct.col %in% colnames(out)) {
        # name wasn't changed yet ..
        struct.col <- "Structure"
    }
    structures <- out[, struct.col]
    out[, struct.col] <- html.format.structure(structures)
    rm.cols <- c("Ambig_Index1", "Ambig_Index2") # Idx1, Idx2 retained for shiny
    m <- match(rm.cols, colnames(out))
    m <- m[!is.na(m)]
    if (length(m) != 0) {
        out <- out[,-m]
    }
    #out <- out[, "Primer2" != colnames(out)]  # remove primer2 col
    ## re-write Idx1 to contain the idx of the 'other' primer
    #if (type == "Cross") {
    #    idx <- rep(NA, nrow(primer.df))
    #    for (i in seq_along(primer.df$Forward)) {
    #        if (is.na(out$Idx1[i])) {
    #            next
    #        }
    #        if (out$Idx1[i] == i) {
    #            idx[i] <- out$Idx2[i]
    #        } else {
    #            idx[i] <- out$Idx1[i]
    #            
    #        }
    #    }
    #    out$Idx1 <- primer.df$ID[idx]
    #    out <- out[, "Idx2" != colnames(out)]
    #}
    return(out)
}

#' Self Dimerization.
#'
#' Computes all self dimers in a user-formatted way.
#'
#' @param primer.df Input primer data frame.
#' @param primer_conc Primer concentration.
#' @param na_salt_conc Sodium ion concentration.
#' @param mg_salt_conc Magensium ion concentration.
#' @param k_salt_conc Potassium ion concentration.
#' @param tris_salt_conc Tris buffer concentration.
#' @param annealing.temp The PCR annealing temperature.
#' @param for.shiny Whether to format the table for HTML output.
#' @param no.structures Whether dimerization structures shall be outputted.
#' @return A formatted data frame with self-dimerization infos
#' @keywords internal
compute.all.self.dimers.frontend <- function(primer.df, primer_conc, na_salt_conc, 
                                    mg_salt_conc, k_salt_conc, tris_salt_conc, 
                                    annealing.temp, for.shiny = FALSE,
                                    no.structures = FALSE) {
    if (Sys.which("hybrid-min") == "") {
        stop("Cannot compute self dimers without an installation of OligoArrayAux for computing the thermoydnamic properties.")
    }
    self.dimers <- compute.all.self.dimers(primer.df, primer_conc, na_salt_conc, 
                                           mg_salt_conc, k_salt_conc, tris_salt_conc, annealing.temp, for.shiny = for.shiny, no.structures = no.structures)
    # select lowest energy binding mode of a self-dimer:
    self.dimers <- ddply(self.dimers, c("Idx1"), function(x) arrange(x, substitute(DeltaG))[1, ]) # i should retain the self primer index, right?
    self.dimers <- self.dimers[, c("DeltaG", "Structure", "Idx1", "Idx2")]
    if (length(self.dimers) != 0) {
        colnames(self.dimers) <- paste0("Self_Dimer_", colnames(self.dimers))
    }
    if (for.shiny) {
        self.dimers <- view.dimer.df(self.dimers, "Self")
    }
    return(self.dimers)
}
#' Cross Dimerization.
#'
#' Computes all cross dimers in a user-formatted way.
#'
#' @param primer.df Input primer data frame.
#' @param primer_conc Primer concentration.
#' @param na_salt_conc Sodium ion concentration.
#' @param mg_salt_conc Magensium ion concentration.
#' @param k_salt_conc Potassium ion concentration.
#' @param tris_salt_conc Tris buffer concentration.
#' @param annealing.temp The PCR annealing temperature.
#' @param for.shiny Whether to format the table for HTML output.
#' @param no.structures Whether to compute structures of dimers.
#' @return A formatted data frame with cross-dimerization infos
#' @keywords internal
compute.all.cross.dimers.frontend <- function(primer.df, primer_conc, na_salt_conc, 
    mg_salt_conc, k_salt_conc, tris_salt_conc, annealing.temp, for.shiny = FALSE, no.structures = FALSE) {
    # computes all combinations of cross dimers: fw-fw, rev-rev, fw-rev
     if (Sys.which("hybrid-min") == "") {
        stop("Cannot compute cross dimers without an installation of OligoArrayAux for computing the thermoydnamic properties.")
    }
    results <- compute.all.cross.dimers.unfiltered(primer.df, primer_conc, na_salt_conc, 
        mg_salt_conc, k_salt_conc, tris_salt_conc, annealing.temp, no.structures = no.structures)
    sel.cols <-  c("DeltaG", "Structure", "Idx1", "Idx2")
    result.cols <- paste0("Cross_Dimer_", sel.cols)
    if (length(results) == 0) {
        # no cross dimers found
        count <- nrow(primer.df)
        out <- data.frame(Cross_Dimer_DeltaG = rep(0, count),
                   Cross_Dimer_Structure = rep("", count),
                   Cross_Dimer_Idx1 = rep(NA, count),
                   Cross_Dimer_Idx2 = rep(NA, count),
                   stringsAsFactors =  FALSE)
        return(out)
    }
    sel.idx <- select.min.cross.idx(results, primer.df$Forward)
    result <- results[sel.idx, sel.cols]
    colnames(result) <- result.cols
    rownames(result) <- NULL
    if (for.shiny) {
        result <- view.dimer.df(result, "Cross")
    }
    return(result)
}
#' Parser for OligoArrayAux Dimerization Data.
#'
#' Parses the free energies and structures of OligoArrayAux.
#'
#' @param deltaG.file A path to a file with OligoArrayAux energies.
#' @param struct.file A path to a file with OligoArrayAux structures.
#' @return A data frame with structures and free energies.
#' @keywords internal
parse.oligo.results <- function(deltaG.file, struct.file) {
    time <- Sys.time()
    results <- try(read.delim(deltaG.file, header = TRUE, 
                stringsAsFactors = FALSE), silent = TRUE)
    if (is(results, "try-error")) {
        warning("Oligo error")
    }
    deltaG <- results[,2]
    #message("DeltaG time: ", Sys.time() - time)
    #time <- Sys.time()
    K.eq <- results[,3] # eq constant; 0 if prefiltered -> no structure!
    prefilter.idx <- which(K.eq == 0) 
    non.prefilter.idx <- which(K.eq != 0)
    if (length(prefilter.idx) == 0) {
        deltaG[prefilter.idx] <- 0 # assume it's no dimer
    }
    if (file.exists(struct.file)) {
        results <- try(readLines(struct.file), silent = TRUE)
        if (is(results, "try-error")) {
            stop("OligoArrayAux: output not produced.")
        }
        last.split <- 1
        split.idx <- which(results == "")
        sel.idx <- setdiff(seq_along(results), split.idx)
        structs <- results[sel.idx]
        structs <- align.structures(structs)
        #message("Struct time: ", Sys.time() - time)
        all.structs <- rep("", length(deltaG))
        if (length(non.prefilter.idx) != 0) {
            all.structs[non.prefilter.idx] <- structs
        }
    } else {
        # non-structure mode
        all.structs <- rep("", length(deltaG))
    }
    out <- data.frame("DeltaG" = deltaG, "Structure" = all.structs, stringsAsFactors = FALSE)
    return(out)
}
#' Identification of Sequence Matches.
#'
#' Identifies matches between two strings
#' provided by OligoArrayAux.
#'
#' @param s1 The aligned nucleotide sequence character vector.
#' @param s2 The aligned, matching substring of \code{s1}.
#' @return A match vector (\code{M} for matches, \code{X} for mismatches).
#' @keywords internal
get.matches <- function(s1, s2) {
    if (any(nchar(s1) != nchar(s2))) { 
        stop("Character vectors of same length required.")
    }
    s1 <- strsplit(s1, "")
    s2 <- strsplit(s2, "")
    match.vector <- lapply(seq_along(s1), function(x) {
        ifelse(s1[[x]] == " ", "X", "M")
    })
    return(match.vector)
}
#' Combination of OligoArrayAux Structure Sequences.
#'
#' Combines the input strings.
#' @param s1 A character vector to be combined with \code{s2}.
#' @param s2 A character vector to be included into \code{s1}.
#' @return A character vector.
#' @keywords internal
combine.strings <- function(s1, s2) {
    s1 <- strsplit(s1, split = "")[[1]]
    s2 <- strsplit(s2, split = "")[[1]]
    idx <- which(s2 != " ")
    out <- s1
    out[idx] <- s2[idx]
    out <- paste0(out, collapse = "")
    return(out)
}
#' Formatting of Dimerization Structures.
#'
#' Formats the given dimerization structures nicely.
#'
#' @param s1 strutcs A character vector, where a block of 4 elements
#' contains: sequence 1 (with removed overlaps), part of sequence 1
#' overlappign with sequence 2, part of sequence 2 
#' overapping with sequence 1, and sequence 2 (with removed overlaps).
#' @return A list of two elements givin the conformation of the first
#' and the second sequence, respectively.
#' @keywords internal
align.structures <- function(structs) {
    # returns a list with two elements: conformation of seq1, conformation of seq2
    # every 
    it <- seq(1, length(structs), 4)
    s1 <- unlist(lapply(seq_along(it), function(x) combine.strings(structs[it[x]], structs[it[x]+1])))
    s2 <- unlist(lapply(seq_along(it), function(x) combine.strings(structs[it[x]+3], structs[it[x]+2])))
    struct.matrix <- do.call(rbind, lapply(seq_along(it), function(x) c(structs[it[x]+1], structs[it[x]+2])))
    match.vector <- get.matches(struct.matrix[,1],struct.matrix[,2])
    match.rep <- unlist(lapply(match.vector, function(x) {
        x[x == "M"] <- "|"
        x[x == "X"] <- " "
        paste(x, collapse = "")
    }))
    out <- paste(s1, match.rep, s2, sep = "\n")
    return(out)
}
expand.grid.unique <- function(x, y, include.equals = TRUE)
{
    x <- unique(x)

    y <- unique(y)

    g <- function(i, x, y)
    {
        z <- setdiff(y, x[seq_len(i-include.equals)])
        if(length(z)) cbind(x[i], z, deparse.level=0)
    }
    if (length(x) <= length(y)) {
        do.call(rbind, lapply(seq_along(x), function(i) g(i, x, y)))
    } else {
        res <- do.call(rbind, lapply(seq_along(y), function(i) g(i, y, x)))
        res[, c(1,2)] <- res[, c(2,1)]
    }
}
#' Preparation of Input for Dimerization.
#' @param s1 Nucleotide character vectors (5' to 3')
#' @param s2 Nucleotide character vectors  (5' to 3')
#' @return A list with two fields containing character vectors.
#' @keywords internal
prepare.dimer.seqs <- function(s1, s2) {
    if (length(s1) == 0 || length(s2) == 0) {
        return(NULL)
    }
    # disambiguate sequences
    s1 <- my.disambiguate(DNAStringSet(s1))
    s2 <- my.disambiguate(DNAStringSet(s2))
    # assign indices
    s1.counts <- IRanges::width(s1@partitioning)
    s2.counts <- IRanges::width(s2@partitioning)
    # indices to indicate the original sequence 
    idx.s1 <- parallel::mclapply(seq_along(s1), function(x) rep(x, s1.counts[x]))
    idx.s2 <- parallel::mclapply(seq_along(s2), function(x) rep(x, s2.counts[x]))
    # index to indicate the position in the list of disambiguated sequences:
    idx.a1 <-  parallel::mclapply(seq_along(s1), function(x) seq_len(s1.counts[x]))
    idx.a2 <-  parallel::mclapply(seq_along(s2), function(x) seq_len(s2.counts[x]))
    # get all combinations
    ambig.options <- parallel::mclapply(seq_along(s1), function(x) {
                    combis <- expand.grid.unique(idx.a1[[x]], idx.a2[[x]])})
    combis.ambig <- do.call(rbind, ambig.options) # index for disambiguated seqs
    colnames(combis.ambig) <- c("Ambig_Index1", "Ambig_Index2")
    combis.ori <- parallel::mclapply(seq_along(s1), function(x) {
                            rep(x, nrow(ambig.options[[x]]))
                })# index for original sequences
    combis.ori <- data.frame(Idx1 = unlist(combis.ori), Idx2 = unlist(combis.ori))
    combis <- cbind(combis.ori, combis.ambig) # TODO: there was a bug here when cbinding for a large set (> 20 mio entries, different number of elements ...), check what happens for global design
    s1 <- CharacterList(s1)
    s2 <- CharacterList(s2)
    # suppress warning that number of columns don't match
    # warning doesn't matter since we select by the correct Ambig_Index
    s1.mat <- suppressWarnings(do.call(rbind, as.list(s1)))
    s2.mat <- suppressWarnings(do.call(rbind, as.list(s2)))
    # this extraction part still takes away some time since we need to iterate over the full list of combinations to select the elements we want
    S1 <- tolower(s1.mat[as.matrix(combis[, c("Idx1", "Ambig_Index1")])])
    S2 <- tolower(s2.mat[as.matrix(combis[, c("Idx2", "Ambig_Index2")])])
    out <- list("Seqs1" = S1, "Seqs2" = S2, "combis" = combis)
    return(out)
}
#' Simple Batchification
#' @param tasks The tasks to assign to individual batches.
#' @return A list of lists containing indices corresponding to \code{tasks}, each list gives a batch.
#' @keywords internal
batchify.simple <- function(tasks) {
    tasks.per.job <- ceiling(length(tasks)/getDoParWorkers())
    no.jobs <- ceiling(length(tasks)/tasks.per.job)
    total.tasks <- 0
    batches <- vector("list", no.jobs)  # idx of the primers in out.primers per file
    for (i in seq_len(no.jobs)) {
        tasks.added <- min(tasks.per.job, length(tasks) - total.tasks)
        s <- NULL
        if (i == 1) {
            s <- 1
        } else {
            s <- max(batches[[i - 1]]) + 1
        }
        e <- s + (tasks.added - 1)
        batches[[i]] <- s:e
        total.tasks <- total.tasks + tasks.added
    }
    return(batches)
}
#' Batchification by Temperature.
#' @param tasks The tasks to assign to individual batches.
#' @param annealing.temp The annealing temperatures according to which batches are to be created.
#' @return A list of lists containing indices corresponding to \code{tasks}, each list gives a batch.
#' @keywords internal
batchify.temp <- function(tasks, annealing.temp) {
    # range of sampled annealing temperatures for batchification
    if (length(annealing.temp) != length(tasks)) {
        stop("Need matching length of annealing temperatures and tasks for batchification! Number of provided annealing temperatures: ", length(annealing.temp), "; Number of tasks: ", length(tasks))
    }
    temps <- seq(min(annealing.temp), max(annealing.temp), 2)
    # assign every task to a temp
    temp.idx <- sapply(annealing.temp, function(x) which.min(abs(x - temps)))
    job.temps <- unique(temps[temp.idx])
    no.jobs <- length(job.temps)
    batches <- vector("list", no.jobs)  # idx of the primers in out.primers per file
    names(batches) <- job.temps
    for (i in seq_len(no.jobs)) {
        cur.temp.idx <- which(job.temps[i] == temps)
        cur.temp <- job.temps[i]
        batches[[as.character(cur.temp)]] <- tasks[temp.idx == cur.temp.idx]
    }
    return(batches)
}
#' Creates multiple Batches for Parallelization.
#' @param tasks An integer vector with indices representing individual computations.
#' @param annealing.temps Temperatures according to which to batchify.
#' @return A list of lists containing indices corresponding to \code{tasks}, each list gives a batch.
#' @keywords internal
batchify <- function(tasks, annealing.temps = NULL) {
    if (length(tasks) == 0) {
        # nothing to batchify
        return(NULL)
    }
    if (length(annealing.temps) == 0) {
       batches <- batchify.simple(tasks)
    } else {
        batches <- batchify.temp(tasks, annealing.temps)
    } 
    return(batches)
}
#' Retrieval of dimerization energies.
#'
#' Uses OligoArrayAux to compute dimerization candidates.
#'
#' @param s1 Nucleotide character vectors (5' to 3')
#' @param s2 Nucleotide character vectors  (5' to 3')
#' @param annealing.temp The PCR annealing temperature in Celsius.
#' @param ions The sodium-equivalent ions used in the PCR.
#' @param no.structures Whether to compute structures of dimers.
#' @return A data frame containing free energies in the field
#' \code{DeltaG} and the dimerization structure in \code{Structure}.
#' @keywords internal
get.dimer.data <- function(s1, s2, annealing.temp, ions, no.structures) {
    seq.data <- prepare.dimer.seqs(s1, s2)
    # update annealing temps to match 'seq.data'
    annealing.temp <- annealing.temp[match(seq.data$combis$Idx1, seq_along(annealing.temp))]
    s1 <- seq.data$Seqs1
    s2 <- seq.data$Seqs2
    combis <- seq.data$combis
    if (length(s1) == 0 || length(s2) == 0) {
        # nothing to be computed ..
        return(NULL)
    }
    if (length(s1) != length(s2)) {
        stop("Cannot compute dimerization: s1 and s2 have different lengths.")
    }
    # parallelize -> split into batches
    if (length(annealing.temp) == 1)  {
        # split up seqs independent of annealing temp into batches for parallelization
        batches <- batchify(seq_along(s1))
        annealing.temp <- rep(annealing.temp, length(batches))
    } else {
        # split up seqs dependent on annealing temp
        batches <- batchify(seq_along(s1), annealing.temp)
		# adjust to batch temperatures:
		annealing.temp <- as.numeric(names(batches))
    }
    i <- NULL
    #print(annealing.temp)
    #print("Computing free energies ...")
    results <- foreach(i = seq_along(batches), .combine = rbind) %dopar% {
        batch <- batches[[i]]
        f1 <- tempfile(pattern = "oligo_dimers_1_", fileext = ".txt")
        f2 <- tempfile(pattern = "oligo_dimers_2_", fileext = ".txt")
        out.prefix <- tempfile(pattern = "oligo_dimers_out_", fileext = "")
        # select sequences
        my.s1 <- s1[batch]
        my.s2 <- s2[batch]
        # write sequences to files
        seqinr::write.fasta(as.list(my.s1), as.list(names(my.s1)), f1)
        seqinr::write.fasta(as.list(my.s2), as.list(names(my.s2)), f2)
        if (length(annealing.temp[i]) == 0 || is.na(annealing.temp[i])) {
            warning("Dimerization: Annealing temperature was NA. Using setting of 50.")
            annealing.temp[i] <- 50
        }
        # filter doesn't speed up the computations so much, just a little bit
        # so let's use the default
        filter <- 2
        # hybrid-min -n DNA -t 50 -T 50 -N 0.1778085 -q tccttcctcatcttcctg caggaggaggaagaacca
        call <- paste0("hybrid-min -n DNA -t ",
                    annealing.temp[i],
                    " -T ",
                    annealing.temp[i],
                    " -N ",
                    ions,
                    " -o ",
                    out.prefix,
                    " --prefilter=",
                    filter, " ")
        #print(call)
        if (no.structures) {
            call <- paste(call, "-E")
        }
        call <- paste(call, f1, f2)
        #message(call)
        system(call, ignore.stdout = TRUE)
        # retrieve results
        deltaG.file <- paste0(out.prefix, ".dG")
        struct.file <- paste0(out.prefix, ".asc")
        results <- parse.oligo.results(deltaG.file, struct.file)
        unlink(c(f1, f2)) 
        results
    }
    # annotate with indices of combinations and re-order from batch order
    results <- cbind(results[match(seq_along(s1), unlist(batches)), ], combis)
    return(results)
}

#' AbstractClass for Constraint Settings.
#'
#' The \code{ConstraintSettings} class encapsulates the constraints
#' on the physicochemical properties of primers.
#'
#' @slot status Named boolean vector indicating 
#' which of the possible constraints are active (\code{TRUE})
#' and which are not (\code{FALSE}).
#' @slot settings Named list containing the 
#' settings of the active constraints
#' @return An \code{AbstractConstraintSettings} object.
#' @keywords internal
AbstractConstraintSettings <- setClass("AbstractConstraintSettings",
	slots = c(status = "vector", settings = "list"),
)

#' @rdname Settings
#' @name Settings
#' @aliases ConstraintSettings-class
#'
#' @section primer_coverage:
#' Computing the primer coverage involves identifying
#' which templates are expected to be amplified (covered) by which primers.
#' The \code{primer_coverage} constraint
#' determines the minimal and maximal number of coverage events per primer
#' that are required. The computation of primer coverage is governed
#' by the coverage constraints postulated via \code{\link{CoverageConstraints}}
#' and the constraint options defined via \code{\link{ConstraintOptions}}. 
#'
#' @section primer_specificity:
#' Primer specificity is automatically determined during the primer coverage
#' computations but the constraint is only checked when the \code{primer_specificity}
#' field is available. The specificity of a primer is defined as its ratio of
#' on-target vs total coverage events (including off-target coverage). Low-specificity
#' primers should be excluded as they may not amplify the target region effectively.
#' 
#' @section primer_length:
#' The length of a primer is defined by its number of bases. Typical
#' primers have lengths between 18 and 22. Longer primers may guarantee higher
#' specificities.
#'
#' @section gc_clamp:
#' The GC clamp refers to the presence of GCs at the 3' end of a primer.
#' For the \code{gc_clamp} constraint, we consider the number of 3' terminal GCs.
#' For example, the primer \emph{actgaaatttcaccg} has a GC clamp of length 3.
#' The presence of a GC clamp is supposed to aid the stability of the
#' polymerase complex. At the same time, long GC clamps should be avoided.
#'
#' @section no_runs:
#' Homopolymer runs (e.g. the primer \emph{aaaaa} has a run of 5 A's)
#' may lead to secondary structure formation and unspecific binding
#' and should therefore be avoided.
#'
#' @section no_repeats:
#' Dinucleotide repeats (e.g. the primer \emph{tatata} has 3 TA repeats)
#' should be avoided for the same reason a long homopolymer runs.
#'
#' @section self_dimerization:
#' Self dimerization refers to a primer that binds to itself rather than
#' to one of the templates. Primers exhibiting self dimers should
#' be avoided as they may prevent the primer from amplifying the templates.
#' Therefore primers with small free energies of dimerization should 
#' be avoided.
#'
#' @section melting_temp_range:
#' The melting temperature is the temperature at which 50% of the primers
#' are in duplex with templates and 50% are still present as random coils.
#' Hence, primers exhibiting high melting temperatures have high affinities
#' to the templates, while primers with small melting temperatures
#' have small affinities. The melting temperatures of the primers
#' determine the annealing temperature of the PCR, which is why
#' the melting temperatures of the primers should not deviate too much (see
#' \code{melting_temp_diff}).
#'
#' @section melting_temp_diff:
#' The differences between the melting temperatures of primers in a
#' set of primers should not deviate too much as the annealing temperaturte
#' of a PCR should be based on the smallest melting temperature
#' of a primer in the set. If there are other primers in the set 
#' exhibiting considerably higher melting temperatures,
#' these primers may bind inspecifically due to the low annealing temperature.
#'
#' @section cross_dimerization:
#' When two different primers bind to each each other rather than
#' to the templates, this is called cross dimerization.
#' Cross dimerization should be prevent at all costs because
#' such primers cannot effectively amplify their target templates.
#' Cross dimerizing primers can be excluding primers
#' exhibiting small free energies of cross dimerization.
#'
#' @section secondary_structure:
#' When a primer exhibits secondary structure, this may prevent it from
#' binding to the templates. To prevent this, 
#' primers with low free energies of secondary structure formation
#' can be excluded.
#'
#' @return The \code{ConstraintSettings} constructor defines a new
#' \code{ConstraintSettings} object.
#' @note
#' The following external programs are required for constraint computations:
#' \describe{
#' \item{MELTING (http://www.ebi.ac.uk/biomodels/tools/melting/):}{Thermodynamic computations (optional) for determining melting temperatures for the constraints \code{melting_temp_diff} and \code{melting_temp_range}}
#' \item{OligoArrayAux (http://unafold.rna.albany.edu/OligoArrayAux.php):}{Thermodynamic computations used for computing \code{self_dimerization} and \code{cross_dimerization}.
#' Also required for computing \code{primer_coverage} when a constraint based on the free energy of annealing is active.}
#' \item{ViennaRNA (http://www.tbi.univie.ac.at/RNA/):}{Secondary structure predictions used for the constraint \code{secondary_structure}}
#' }
#' @export
#' @examples
#'
#' # Initializing a new 'ConstraintSettings' object:
#' constraint.settings <- new("ConstraintSettings")
#' # Retrieving the constraint settings from a 'DesignSettings' object:
#' data(Ippolito) # loads a 'DesignSettings' object into 'settings'
#' constraints(settings)
#' # Modifying the constraint settings:
#' constraints(settings)$no_runs["max"] <- 10
#' constraints(settings) <- constraints(settings)[names(constraints(settings)) != "gc_clamp"]
ConstraintSettings <- setClass("ConstraintSettings", contains="AbstractConstraintSettings",
	slots = c( # constructor arguments:
			status = "vector",
            settings = "list"
			),
    # define initial values using prototype
    prototype = prototype(status = {
            possible.constraints <- c("primer_coverage", "primer_length", 
                "primer_specificity", "gc_clamp", "gc_ratio","no_runs",
                       "no_repeats", "self_dimerization", "melting_temp_diff",
                       "melting_temp_range", "cross_dimerization", 
                       "secondary_structure")
            cons <- rep(FALSE, length(possible.constraints))
            names(cons) <- possible.constraints
            cons
        },
        settings = list() # no active constraints initially
   ),
    validity=function(object)
	{
		return(check_constraint_settings_validity(object))
	}
)
setMethod("initialize", "ConstraintSettings",
    function(.Object, ...) {
        .Object <- callNextMethod(...)  # passed to parent constructor
        .Object
})
setMethod("show", "ConstraintSettings", function(object) {
    print(constraints.to.df(object@settings, out.names = "Target range"))
    invisible()
})
check.abstract.constraint <- function(settings) {
    # checks for errors when defining abstract constraints
    errors <- NULL
    classes <- sapply(settings, class)
    if (!all(classes %in% c("numeric", "integer"))) {
        msg <- "Please provide only numeric constraints."
        errors <- c(errors, msg)
    }
    # check for interval property of the constraints:
    interval.check <- check_interval(settings)
    #print("checked interval:")
    #print(interval.check)
    if (!is.logical(interval.check)) {
        errors <- c(errors, interval.check)
    }
    return(errors)
}

#' Check the Validity of the Constraint Settings.
#'
#' Checks whether the status and the active constraints match.
#' Determines whether the constraints are allowed/known.
#'
#' @param object An \code{AbstractConstraintSettings} object.
#' @return \code{TRUE} if the constraint settings are valid, \code{FALSE} otherwise.
#' @keywords internal
check_constraint_settings_validity <- function(object) {
    # store errors here:
    errors <- character(0)
    # create reference objects for checking possible fields of the settings
    if (is(object, "ConstraintSettings")) {
        ref.object <- ConstraintSettings() # containing all possible constraints
    } else if (is(object, "CoverageConstraints")) {
        ref.object <- CoverageConstraints()
    } else if (is(object, "ConstraintOptions")) {
        ref.object <- ConstraintOptions()
    } else if (is(object, "PCR_Conditions")) {
        ref.object <- PCR_Conditions()
    } else {
        msg <- paste0("Unknown class: ", class(object))
        errors <- c(errors, msg)
    }
    # possible constraints should always stay the same:
    if (length(names(object@status)) == 0 || 
    any(names(ref.object@status) != names(object@status))) {
        m <- match(names(object@status), names(ref.object@status))
        idx <- which(is.na(m))
        error <- paste0("'status' was modified from the default.",
                    "Unknown constraint: ",
                    paste0(names(object@status)[idx], collapse = ","))
        errors <- c(errors, error)
    } 
    # if a constraint is active, it should have a setting:
    active.cons <- names(object@status)[which(object@status)]
    if (any(!active.cons %in% names(object@settings))) {
        error <- "All active constraints should have a setting."
        errors <- c(errors, error)
    }
    # if a constraint has a setting, it should be active
    setting.cons <- names(object@settings)
    if (length(setting.cons) != 0 && any(!setting.cons %in% names(object@status))) {
        error <- "Constraint with setting wasnt' active."    
        errors <- c(errors, error)
    }
    # check the types of the constraint settings:
    if (is(object, "ConstraintSettings") || is(object, "CoverageConstraints")) {
        #print("cvg interval check")
        #print(object@settings)
        # check interval properties for the constraints
        #############
        settings <- NULL
        if (is(object, "ConstraintSettings")) {
            cur.settings <- object@settings
        } else if (is(object, "CoverageConstraints")) {
            cur.settings <- object@settings
        }
        cur.errors <- check.abstract.constraint(cur.settings)
        errors <- c(errors, cur.errors)
    } else if (is(object, "ConstraintOptions") || is(object, "PCR_Conditions")) {
        # check mandatory options for the other settings
        m <- match(names(ref.object@settings), names(object@settings))
        if (any(is.na(m))) {
            idx <- which(is.na(m))
            msg <- paste0("There were missing mandatory settings: ", 
                 paste0(names(constraints(ref.object))[idx], collapse = ","))
            errors <- c(errors, msg)
        }
        # check values of options
        if (is(object, "ConstraintOptions")) {
            # character options:
            region.ok <- match.arg(constraints(object)$allowed_region_definition, c("within", "any"))
            hexamer.ok <- match.arg(constraints(object)$hexamer_coverage, c("active", "inactive"))
            # other options should be numeric: 
            sel <- which(!names(constraints(object)) %in% c("allowed_region_definition", "hexamer_coverage"))
            classes <- sapply(constraints(object)[sel], class)
            if (!all(classes %in% c("numeric", "integer"))) {
                msg <- paste0("Incorrect option class: should have been numeric.")
                errors <- c(errors, msg)
            }
        }
        # check values of PCR
        if (is(object, "PCR_Conditions")) {
            if (!is(constraints(object)$use_taq_polymerase, "logical")) {
                msg <- "'use_taq_polymerase' should be logical."
                errors <- c(errors, msg)
            }
            # other PCR conditions should be numeric:
            sel <- which(!names(constraints(object)) %in% c("use_taq_polymerase"))
            classes <- sapply(constraints(object)[sel], class)
            if (!all(classes %in% c("numeric", "integer"))) {
                msg <- paste0("Incorrect class for PCR setting: should have been numeric.")
                errors <- c(errors, msg)
            }
            # warn about Tris if != 0 due to MELTING bug
            if ("Tris_concentration" %in% names(constraints(object))
                && constraints(object)$Tris_concentration != 0) {
                    warning("Tris_concentration should remain at 0 since melting temperature estimate is unstable otherwise.")
            }
        }
    } 
    return (length(errors) == 0)
}

#' @rdname Settings
#' @name Settings
#' @aliases CoverageConstraints-class
#' @return The \code{CoverageConstraints} constructor initializes a new
#' \code{CoverageConstraints} object.
#' @note
#' The following external programs are required for computing the coverage constraints:
#' \describe{
#' \item{OligoArrayAux (http://unafold.rna.albany.edu/OligoArrayAux.php):}{Thermodynamic computations used for 
#' computing the coverage constraints \code{annealing_DeltaG}, \code{primer_efficiency}, and \code{coverage_model}}
#' }
#' @export
#' @examples
#' 
#' # Initialize a new 'CoverageConstraints' object:
#' cvg.constraints <- new("CoverageConstraints")
#' # Retrieving the coverage constraints from a 'DesignSettings' object:
#' data(Ippolito) # loads a 'DesignSettings' object into 'settings'
#' cvg_constraints(settings)
#' # Modifying the coverage constraints
#' cvg_constraints(settings)$primer_efficiency["min"] <- 0.001
CoverageConstraints <- setClass("CoverageConstraints", contains="AbstractConstraintSettings",
    prototype = prototype(status = {
            possible.constraints <- c("coverage_model", "stop_codon", "substitution", 
                "terminal_mismatch_pos", "primer_efficiency", "annealing_DeltaG", 
                "coverage_model")
            cons <- rep(FALSE, length(possible.constraints))
            names(cons) <- possible.constraints
            cons
        },
        settings = list() # no active constraints initially
   ),
    validity=function(object)
	{
		return(check_constraint_settings_validity(object))
	}
)
setMethod("show", "CoverageConstraints", function(object) {
    print(constraints.to.df(object@settings, out.names = "Setting"))
    invisible()
})
setMethod("initialize", "CoverageConstraints",
    function(.Object, ...) {
        .Object <- callNextMethod(...)  # passed to parent constructor
        .Object
})

#' @rdname Settings
#' @name Settings
#' @aliases ConstraintOptions-class
#' @return The \code{ConstraintOptions} constructor returns
#' a new \code{ConstraintOptions} object.
#' @export
#' @examples
#' 
#' # Initialize a new 'ConstraintOptions' object:
#' constraint.options <- new("ConstraintOptions")
#' # Retrieve the constraint options from a 'DesignSettings' object:
#' data(Ippolito) # loads a 'DesignSettings' object into 'settings'
#' conOptions(settings)
#' # Prevent off-target binding:
#' conOptions(settings)$allowed_other_binding_ratio <- 0
ConstraintOptions <- setClass("ConstraintOptions", contains="AbstractConstraintSettings",
    prototype = prototype(status = {
        # set mandatory options to TRUE
        option.status <- c("allowed_mismatches" = TRUE, "allowed_other_binding_ratio" = TRUE,
                          "allowed_region_definition" = TRUE, "hexamer_coverage" = FALSE)
        option.status
    }, 
    # ensure that 'settings' reflects the option status (mandatory should be present in default):
    settings = list("allowed_mismatches" = 7, "allowed_other_binding_ratio" = 1, 
                    "allowed_region_definition" = "within")
   ),
   validity=function(object) {
		return(check_constraint_settings_validity(object))
	}
)
setMethod("show", "ConstraintOptions", function(object) {
    print(create.options.table(object@settings))
    invisible()
})
setMethod("initialize", "ConstraintOptions",
    function(.Object, ...) {
        .Object <- callNextMethod(...)  # passed to parent constructor
        .Object
})

#' @rdname Settings
#' @name Settings
#' @aliases PCR_Conditions-class
#' @return The \code{PCR_Conditions} constructor defines a new \code{PCR_Conditions} object.
#' @export
#' @examples
#' 
#' # Initialize a new 'PCR_Conditions' object:
#' PCR.conditions <- new("PCR_Conditions")
#' # Retrieving the PCR conditions from a 'DesignSettings' object:
#' data(Ippolito) # loads a 'DesignSettings' object into 'settings'
#' PCR(settings)
#' # Modifying the PCR conditions:
#' PCR(settings)$use_taq_polymerase <- FALSE
PCR_Conditions <- setClass("PCR_Conditions", contains="AbstractConstraintSettings",
    prototype = prototype(status = {
        # set mandatory options to TRUE
        option.status <- c("use_taq_polymerase" = TRUE, "annealing_temp" = FALSE,
                        "Na_concentration" = TRUE, "Mg_concentration" = TRUE,
                        "K_concentration" = TRUE, "Tris_concentration" = FALSE,
                        "primer_concentration" = TRUE, "template_concentration" = TRUE,
                        "cycles" = FALSE)
        option.status
    }, 
    # ensure that 'settings' reflects the option status (mandatory should occur in settings):
    settings = list("use_taq_polymerase" = TRUE, "Na_concentration" = 0,
                    "Mg_concentration" = 0.0015, "K_concentration" = 0.05, 
                    "Tris_concentration" = 0, "primer_concentration" = 2e-07,
                    "template_concentration" = 1.28e-11)
   ),
   validity=function(object) {
		return(check_constraint_settings_validity(object))
	}
)
setMethod("show", "PCR_Conditions", function(object) {
    print(create.PCR.table(object@settings))
    invisible()
})

setMethod("initialize", "PCR_Conditions",
    function(.Object, ...) {
        .Object <- callNextMethod(...)  # passed to parent constructor
        .Object
})

#' @rdname Data
#' @name Data
#'
#' @usage data(Ippolito)
#'
#' @aliases template.df primer.df settings
#'
#' @references Ippolito GC, Hoi KH, Reddy ST, Carroll SM, Ge X, Rogosch T, 
#' Zemlin M, Shultz LD, Ellington AD, VanDenBerg CL, Georgiou G. 2012. 
#' Antibody Repertoires in Humanized NOD-scid-IL2R gamma null Mice and Human B
#' Cells Reveals Human-Like Diversification and Tolerance Checkpoints 
#' in the Mouse. PLoS One 7:e35497.
#' @examples
#' 
#' # Load the data from Ippolito et al.
#' data(Ippolito)
#' primer.df
#' template.df
#' constraints(settings)
NULL

#' Output Functionalities.
#'
#' @rdname Output
#' @name Output
#' @description
#' \describe{
#' \item{\code{write_primers}}{Writes a set of primers to disk, either as a FASTA or CSV file.}
#' \item{\code{write_settings}}{ Stores primer analysis settings to a file in XML format.}
#' \item{\code{write_templates}}{Stores a set of templates as a FASTA or CSV file.}
#' \item{\code{create_report}}{Creates a PDF report for analyzed primer sets.}
#' \item{\code{create_coverage_xls}}{Creation of an XLS spreadsheet 
#' providing an overview of the covered
#' template sequences for each primer. Each cell in the spreadsheet
#' indicates a coverage event between a primer and template using
#' color codes. Identified coverage events are indicated by green, while
#' primer-template pairs without coverage are indicated by red.
#' In case that a primer binding condition (see \code{\link{CoverageConstraints}})
#' was active when computing the coverage, the numeric value of the
#' coverage condition is annotated for each cell.}
#' }
#' 
#' @param primer.df An object of class \code{Primers}.
#' @param fname The path to the output file.
#' @param ftype A character vector giving the type of the file.
#' This can either be "FASTA" or "CSV" (default: "FASTA").
#' @param template.df An object of class \code{Templates}.
#' @param settings A \code{DesignSettings} object to be stored to disk.
#' @param primers To create a report for a single primer set, please provide
#' an evaluated \code{Primers} object.
#' For creating a report comparing multiple primer sets, please provide
#' a list of \code{Primers} objects.
#' @param templates If \code{primers} is a \code{Primers} object, \code{templates} should be a \code{Templates} object.
#' If \code{primers} is a list of \code{Primers} objects, \code{templates}
#' should be a list of \code{Templates} objects of the same length as \code{primers}.
#' @param sample.name An identifier for your analysis. By default ( 
#' \code{NULL}), the sample identifier is selected from the
#' \code{Run} column of the input templates.
#' @param used.settings A named list (with fields \code{fw} and \code{rev}) containing the relaxed settings
#' for designing forward/reverse primers. By default (\code{NULL}), 
#' the relaxed settings are not shown in the report.
#' @param ... \code{required.cvg} (optional, default: 1), the desired coverage ratio if \code{primers} is a single primer set.
NULL
 

#' Input Functionalities.
#'
#' @rdname Input
#' @name Input
#' @description
#' \describe{
#' \item{\code{read_primers}}{Reads one or multiple input files with 
#' primer sequences. The input can either be in FASTA or in CSV format.}
#' \item{\code{read_templates}}{Read one or multiple files with template sequences in FASTA or CSV format.}
#' \item{\code{read_settings}}{Loads primer analysis settings from an XML file.}
#' \item{\code{Templates}}{The \code{Templates} class encapsulates a data frame
#' containing the sequencs of the templates, their binding regions,
#' as well as additional information (e.g. template coverage).}
#' \item{\code{Primers}}{The \code{Primers} class encapsulates a data frame
#' representing a set of primers. Objects of this class
#' store all properties associated with a set of primers,
#' for example the results from evaluating the properties
#' of a primer set or from determining its coverage.}
#' }
#' @param fname Character vector providing either a single or multiple
#' paths to FASTA or CSV files.
#' @param fw.id For FASTA input, the identifier for forward primers in the FASTA headers.
#' @param rev.id For FASTA input, the identifier for reverse primers in the FASTA headers.
#' @param merge.ambig Indicates whether similar primers should be merged
#' ("merge") using IUPAC ambiguity codes or whether primers should 
#' be disambiguated ("unmerge").
#' By default \code{merge.ambig} is set to "none", leaving primers as they are.
#' @param max.degen A scalar numeric providing the maximum
#' allowed degeneracy for merging primers if
#' \code{merge.ambig} is set to "merge".
#' Degeneracy is defined by the number of disambiguated sequences 
#' that are represented by a degenerate primer.
#' @param template.df An object of class \code{Templates}.
#' If \code{template.df} is provided for \code{read_primers} then the primers are checked
#' for restriction sites upon input; otherwhise they are not checked.
#' @param adapter.action The action to be performed when \code{template.df} is
#' provided for identifying adapter sequences.
#' Either "warn" to issue warning about adapter sequences or
#' "rm" to remove identified adapter sequences. The default is "warn".
#' @param sample.name An identifier for the input primers. 
#' @param updateProgress A Shiny progress callback function. This is
#' \code{NULL} by default such that no progress is tracked.
#' @param hdr.structure A character vector describing the information contained in the FASTA headers. In case that the headers of \code{fasta.file} contain
#' template group information, please include the keyword "GROUP" in
#' \code{hdr.structure}. If the numer of elements provided via \code{hdr.structure}
#' is shorter than the actual header structure, the missing fields are ignored.
#' @param delim Delimiter for the information in the FASTA headers.
#' @param id.column Field in the header to be used as the identifier of individual
#' template sequences.
#' @param rm.keywords A vector of keywords that are used to remove templates whose headers contain any of the keywords.
#' @param remove.duplicates Whether duplicate sequence shall be removed.
#' @param fw.region The positional interval from the template 5' end specifying the  
#' binding sites for forward primers. The default \code{fw.region} is set to
#' the first 30 bases of the templates.
#' @param rev.region The positional interval from the template 3' end specifying
#' the binding sites for reverse primers. The default \code{rev.region}
#' is set to the last 30 bases of the templates.
#' @param gap.char The character in the input file representing gaps. 
#' Gaps are automatically removed upon input and the default character is "-".
#' @param run An identifier for the set of template sequences. By default,
#' \code{run} is \code{NULL} and its value is set via \code{template.file}.
#' @param filename Path to a valid XML file containing the 
#' primer analysis settings. By default, \code{filename} is set
#' to all settings that are shipped with openPrimeR and the lexicographically
#' first file is loaded.
#' @param frontend Indicates whether settings shall be loaded for the Shiny frontend. 
#' In this case no unit conversions for the PCR settings are performed.
#' The default setting is \code{FALSE} such that the correct units are used.
#' @param ... A data frame fulfilling the structural requirements
#' for initializing a \code{Templates} or \code{Primers} object.
NULL

#######
# Use alignments+trees to initalize primers
#######

#' Multiple Sequence Alignment
#'
#' Computes a multiple sequence alignment using MAFFT.
#'
#' @param seqs The sequences to be aligned.
#' @param names The identifiers of the sequences.
#'
#' @return An alignment object as created from seqinr's read.alignment method.
#' @keywords internal
#' @references 
#' Katoh, Misawa, Kuma, Miyata 2002 (Nucleic Acids Res. 30:3059-3066) 
#' MAFFT: a novel method for rapid multiple sequence alignment based
#' on fast Fourier transform. 
align.seqs <- function(seqs, names) {
    runID <- digest::digest(seqs, "md5")
    file.loc <- tempfile(paste0("in_ali_templates_", runID), fileext = ".fasta")
    seqinr::write.fasta(as.list(seqs), file.out = file.loc, names = as.list(names))
    out.loc <- tempfile(paste0("out_ali_templates_", runID), fileext = ".fasta")
    call <- paste("mafft", " --auto --quiet ", file.loc, " > ", out.loc, sep = "")
    if (Sys.which("mafft") == "") {
        stop("Cannot align sequences since MAFFT is not installed on your system.")
    }
    system(call)  #, ignore.stderr = TRUE) # cannot be called with suppressed output ..
    seq.ali <- try(seqinr::read.alignment(out.loc, format = "fasta"), silent = FALSE)
    if (is(seq.ali, "try-error")) {
        # alignment not possible
        warning("MAFFT wasn't able to align the templates. Returning.")
        return(NULL)
    }
	# clean up temporary files
	on.exit({
		file.remove(c(file.loc, out.loc))
	})
    return(seq.ali)
}
#' Identication of Short Primers.
#'
#' Identify initial primers that are too short.
#'
#' @param primer.candidates Primer alignment.
#' @param min.len Minimal primer length.
#' @return The index of proposed primers that are shorter than \code{min.len}.
#' @keywords internal
prefilter.primer.candidates <- function(primer.candidates, min.len) {
    n.gaps <- apply(as.character(primer.candidates), 1, function(x) length(which(x == 
        "-")))
    len <- ncol(primer.candidates) - n.gaps  # ungapped length
    rm.idx <- which(len < min.len)
    return(rm.idx)
}
#' Filtering of Primer Candidates
#'
#' Filters primer candidates according to length and duplications.
#"
#' @param primer.candidates Alignment of candidate primers.
#' @param min.len Minimal required length of primers.
#' @return Filtered alignment of candidate primers.
#' @keywords internal
filter.primer.candidates <- function(primer.candidates, min.len) {
    # remove entries in the primer candidate matrix whose ungapped sequences are
    # shorter than min.len
    
    # idea: count gaps
    rm.idx <- prefilter.primer.candidates(primer.candidates, min.len)
    if (length(rm.idx) != 0) {
        # message(paste('Removed ',length(rm.idx), ' candidates that were shorter than ',
        # min.len, ' bases.', sep = ''))
        primer.candidates <- primer.candidates[-rm.idx, ]
    }
    # remove duplicated candidates: we lose the implicit cvg info from the tree with
    # this, but we are faster.
    if (nrow(primer.candidates) != 0) {
        al <- ape::as.alignment(primer.candidates)
        rm.idx <- which(duplicated(al$seq))
        if (length(rm.idx) != 0) {
            al$nb <- al$nb - length(rm.idx)
            al$seq <- al$seq[-rm.idx]
            al$nam <- al$nam[-rm.idx]
        }
        primer.candidates <- ape::as.DNAbin(al)
    }
    return(primer.candidates)
}

#' @rdname Scoring
#' @name Scoring
#' @aliases score_degen
#' @details
#' \code{score_degen} computes the degeneration of an ambiguous sequence
#' by considering the number of unambiguous sequences that
#' are represented by the the ambiguous sequence.
#' Let a sequence \code{S} of length \code{n} be represented by a collection of sets such that
#' \deqn{S = {s_1, s_2, \ldots, s_n}}
#' where \eqn{s_i} indicates the set of unambiguous bases found
#' at position \eqn{i} of the primer. Then the degeneracy \eqn{D} of a primer
#' can be defined as 
#' \deqn{D = \prod_i{|s_i|}}
#' where \eqn{|s_i|} provides the number of disambiguated bases at position \eqn{i}.
#'
#' @return \code{score_degen} finds the number of unambiguous sequences
#' that are represented by \code{seq}.
#' @export
#' @examples
#' # Compute degeneration for sequences with differing number of ambiguous bases
#' seq <- strsplit(c("ctggaattacggtacc", "taggaaccggrtaagc", "rtaaasrygtar"), split = "")
#' degen <- score_degen(seq)
score_degen <- function(seq, gap.char = "-") {
    if (length(seq) == 0) {
        return(NA)
    }
    if (!is(seq, "list")) {
        stop("'seq' should be a list of split characters.")
    }
    gapped.codemap <- IUPAC_CODE_MAP
	if (length(gap.char) != 0 && nchar(gap.char) == 1) {
		gapped.codemap[gap.char] <- gap.char
	}
    degen <- unlist(lapply(seq, function(x) prod(nchar(gapped.codemap[toupper(x)]))))
    return(degen)
}
#' Tree Ancestry
#'
#' Checks whether \code{ancestor.node} is an ancestor to the nodes 
#' specified in \code{test.node}.
#'
#' @param tree The phylogenetic tree to be tested.
#' @param ancestor.node A node to be checked for being an ancestor to \code{test.node}.
#' @param test.node Possible descendants of \code{ancestor.node}.
#' @return TRUE, if \code{ancestor.node} is an ancestor to any node in \code{test.node}.
#' @keywords internal
ancestor_of <- function(tree, ancestor.node, test.node) {
    e <- tree$edge
    return(any(test.node %in% e[e[, 1] == ancestor.node, 2]))
}
#' Computation of Consensus.
#'
#' Computes the consensus of the sequences in the input alignment.
#'
#' @param ali An alignment object.
#' @return A consensus sequence without gap characters.
#' @keywords internal
get.consensus.seq <- function(ali) {
    c.seq <- seqinr::consensus(ali, method = "IUPAC")
    na.idx <- which(is.na(c.seq))
    if (length(na.idx) != 0) {
        c.seq.m <- seqinr::consensus(ali[, na.idx, drop = FALSE], method = "majority")
        # a) replace gaps in the consensus with gaps if they are the majority
        gap.idx <- which(c.seq.m == "-")
        c.seq[na.idx][gap.idx] <- c.seq.m[gap.idx]
        if (length(gap.idx) != 0) {
            na.idx <- na.idx[-gap.idx]
        }
        # b) form consensus for remaining positions where non-gap characters are the
        # majority, excluding gaps from consensus
        if (length(na.idx) != 0) {
            non.gap.idx <- apply(ali[, na.idx, drop = FALSE], 2, function(y) y != 
                "-")
            c.seq.m <- sapply(seq_len(ncol(non.gap.idx)), function(x) seqinr::consensus(ali[non.gap.idx[, 
                x], na.idx[x], drop = FALSE], method = "IUPAC"))
            c.seq[na.idx] <- c.seq.m
        }
    }
    # remove gaps from consensus
    c.seq <- c.seq[c.seq != "-"]
    return(c.seq)
}
#' Determine Tree Consensus Sequences
#'
#' Creates all possible consensus sequences from a phylogenetic tree.
#'
#' Ambiguous sequences are only generated with a degeneracy of at most \code{max.degen}. 
#' The tree is iterated from leaves to the top, i.e., starting from least degeneracy to most degeneracy. 
#' Merges only take place when the degeneracy of the resulting sequence would
#' be at most \code{max.degen}. Gaps are removed from the alignments.
#'
#' @param tree The phylogenetic tree.
#' @param max.degen The maximal degeneration of consensus primers.
#' @param primer.candidates Alignment of primers.
#' @return Data frame with consensus primers extracted from the tree.
#' @keywords internal
get.tree.seqs <- function(tree, max.degen, primer.candidates) {
    if (length(tree) == 0) {
        # too few seqs to build tree
        return(NULL)
    }
    N.tips <- length(tree$tip.label)
    selected.seqs <- NULL
    node.ignore.list <- NULL
    degeneracy.count <- NULL
    primer.IDs <- NULL
    merge.idx <- NULL
    for (i in seq_len(tree$Nnode)) {
        # message(i)
        node <- N.tips + tree$Nnode - i + 1
        stree <- ape::extract.clade(tree, node)  # get tree from current node
        ######### 
        # plot(stree) # visualize subtree 
        # nodelabels()
        # tiplabels() 
        ##############
        # ignore nodes that are
        # ancestors of trees that already exceeded the degeneracy count
        if (ancestor_of(tree, node, node.ignore.list)) {
            node.ignore.list <- c(node.ignore.list, node)
            next
        }
        seq.ids <- stree$tip.label  # seqs of this tree
        m <- match(seq.ids, rownames(primer.candidates))
        if (any(is.na(m))) {
            stop(paste("Integrity error: tree did not contain expected ID: ", seq.ids[is.na(m)], 
                " (hashing problem).", sep = ""))
        }
        ali <- primer.candidates[m, ]
        c.seq <- get.consensus.seq(as.character(ali))
        degen <- score_degen(list(c.seq))
        if (degen <= max.degen) {
            selected.seqs <- c(selected.seqs, paste(c.seq, collapse = ""))
            degeneracy.count <- c(degeneracy.count, degen)
            id <- paste(seq.ids, collapse = ",")
            primer.IDs <- c(primer.IDs, id)
            merge.idx <- c(merge.idx, list(m[order(m)]))
        } else {
            # message(paste('ignored: degeneracy was ', degen, sep = ''))
            node.ignore.list <- c(node.ignore.list, node)
        }
    }
    if (length(selected.seqs) != 0) {
        node.result <- data.frame(ID = paste(primer.IDs, sep = ""), Sequence = selected.seqs, 
            Degeneracy = degeneracy.count, Merge_Idx = sapply(merge.idx, function(x) paste(x, 
                collapse = ",", sep = "")), stringsAsFactors = FALSE)
        # remove duplicates
        node.result <- node.result[!duplicated(node.result$Sequence), ]
    } else {
        node.result <- NULL
    }
    return(node.result)
}
#' Ranges for Initial Primers.
#'
#' Creates a data frame indicating primer starts and ends.
#'
#' @param end.position End positions of primers.
#' @param p.lens Desired primer lengths.
#' @param start.posiion The start positions of primers.
#' @param step.size A numeric giving the steps with which start
#' positions are cycled. Should be 1 for primer design (evaluate all positions)
#' and higher values can be used for windowing.
#' @param groups Character vector with group annotation.
#' @return Data frame with ranges for initial primers.
#' @keywords internal
create.primer.ranges <- function(end.position, p.lens, start.position,
                                 step.size = 1, groups = NULL) {
    primer.ranges <- vector("list", length(p.lens))
    if (length(groups) == 0) {
        groups <- rep("default", length(end.position))
    }
    for (i in seq_along(p.lens)) {
        # ensure that we only select regions that aren't larger than the allowed binding region:
        p.len <- min(p.lens[i], end.position[i] - start.position[i])
        starts <- seq(start.position[i], end.position[i] - p.len + 1, step.size)
        ends <- starts + p.len - 1
        idx <- starts >= 1 & ends <= end.position[i]
        primer.ranges[[i]] <- data.frame(Start = starts[idx], End = ends[idx],
                                         Group = rep(groups[i], length(starts[idx])))
    }
    primer.range <- do.call(rbind, primer.ranges)
    primer.range <- ddply(primer.range, "Group", 
                        summarize,
                        Start = unique(substitute(Start)),
                        End = unique(substitute(End)))
    return(primer.range)
}
#' Conservation Scores
#'
#' Scores the conservation of alignment regions.
#'
#' @param primer.range A data frame with starts/ends of primers.
#' @param ali.entropy Entropies corresponding to the alignment
#' @return Entropies indicating conservation (similarity) of regions.
#' @keywords internal
score.conservation <- function(primer.range, ali.entropy) {
    # mean entropy per region
    scores <- unlist(lapply(seq_len(nrow(primer.range)), function(x) mean(ali.entropy[primer.range$Start[x]:primer.range$End[x]])))
    return(scores)
}
#' Selection by Conservation
#'
#' Selects primer regions for initialization of primers according
#' to their conservation scores.
#'
#' The conservation scores are computed using the entropies computed from
#' the alignment of the template sequence regions of interests.
#'
#' @param primer.range Data frame with primer starts/stops.
#' @param ali.entropy Entropy values for the alignment.
#' @param conservation Desired ratio of primer conservation.
#' Only regions with a conservation of at least \code{conservation}
#' are considered for the initialization of primers.
#' @param bin \code{DNABin} alignment of templates.
#' @param gap.char The character for alignment gaps.
#' @return Updated primer regions according to the desired conservation.
#' @keywords internal
select.primer.region.by.conservation <- function(primer.range, ali.entropy, 
                                                conservation, bin, gap.char = "-") {
    # conservation: select only the top percent (ratio) of alignments 100% ->
    # consider all sub-alignments
    if (conservation == 1) {
        # conservation is 1 -> return top 100% of conserved ranges -> all candidate
        # ranges
        return(primer.range)
    }
    # otherwise: limit to conserved regions yielding mostly primers with required
    # length
    primer.candidates <- lapply(seq_len(nrow(primer.range)), function(x) bin[, primer.range[x, 
        "Start"]:primer.range[x, "End"]])  # sub-alignment
    # don't consider regions that are basically only gaps
    # determine the gap positions in every possible binding region
    gap.idx <- detect.gap.columns(primer.candidates, gap.cutoff = 0.95, gap.char = gap.char) # NB: could improve the speed of gap detection
    # select only regions without major gap columns
    allowed.gaps <- 0
    sel.idx <- NULL
    selected <- FALSE
    # TODO: check region detection for 'rev'
    while (!selected) { 
        #print(gap.idx)
        sel.idx <- which(unlist(lapply(gap.idx, function(x) length(x) <= allowed.gaps)))
        if (length(sel.idx) != 0) {
            # region found
            selected <- TRUE
        } else {
            # allow more gaps in selected regions
            allowed.gaps <- allowed.gaps + 1
        }
    }
    # turn selected index into idx for the whole alignment
    selected.range <- primer.range[sel.idx,]
    #warning("Need to check that the right regions are selected here!")
    c.score <- score.conservation(selected.range, ali.entropy)
    q <- quantile(c.score, seq(0, 1, 0.01))
    # entropy cutoff for the specified conservation level:
    sel.score <- q[paste(round(conservation, 2) * 100, "%", sep = "")]
    #print(sel.score)
    # index of regions below the entropy cutoff
    idx <- which(c.score <= sel.score)
    new.primer.range <- primer.range[sel.idx[idx], ]
    # add scores:
    new.primer.range$Entropy <- c.score[idx]
    return(new.primer.range)
}
#' Hierarchical Clustering.
#'
#' Performs hierarchical clustering on aligned primer sequences.
#'
#' The clustering is performed to identify similar groups of primer candidates
#' that can be merged to form degenerate primers.
#'
#' @param primer.candidates Alignment of primer candidates.
#' @return Phylogeny of the input \code{primer.candidates}.
#' @keywords internal
hclust.tree <- function(primer.candidates) {
    if (nrow(primer.candidates) < 2) {
        # cannot build a tree for a small number of seqs
        return(NULL)
    }
    # dist: use Hamming distance (nbr of required substitutions)
    s <- as.character(primer.candidates)
    seqs <- unlist(lapply(seq_len(nrow(s)), function(x) paste(s[x,], collapse = "")))
    names(seqs) <- rownames(primer.candidates)
    # don't parallelize here -> nthread =1, otherwise foreach loop hangs 
    dist <- stringdist::stringdistmatrix(seqs, method = "hamming",
                                        useNames = "names", nthread = 1) 
    #dist <- cluster::daisy(data.frame(as.character(primer.candidates)), metric = "gower")
    clusters <- hclust(dist)
    tree <- ape::as.phylo(clusters)
    return(tree)
}
#' Tree-based Initialization of Primers.
#'
#' Creates a set of candidate primers using a tree-based algorithm.
#'
#' First, primers are aligned and their sequence similarity is determined to
#' compute a phylogenetic tree using hierarchical clustering.
#' Next, the tree is traversed from leaves to top in order to identify
#' groups of primers that can be merged (consensus) without exceeding the maximum
#' degeneracy of primers. 
#'
#' @param seqs Template sequences.
#' @param seq.IDs Identifiers of template sequences.
#' @param seq.groups Group identifiers of template sequences.
#' @param start For each template the start of the interval where primers 
#' should be created.
#' @param end For each template the end of the interval where primers
#' should be created.
#' @param primer.lengths Vector of desired primer lengths.
#' @param allowed.region.definition Definition of allowed regions.
#' @param max.degen Maximal degeneracy of primers.
#' @param conservation Required conservation of template regions considered
#' for the generation of primers. Conservation identifies the top conserved 
#' percentile of possible primers. 
#' @param sample Sample name for the analysis.
#' @param identifier Identifier (e.g. for directionality).
#' @param updateProgress Shiny progress object.
#' @return A vector with initialized primers.
#' @keywords internal
create.primers.tree <- function(seqs, seq.IDs, seq.groups, start, end, primer.lengths, allowed.region.definition, 
    max.degen, conservation, sample = "", identifier = "", updateProgress = NULL) {
    # use template group as primer identifier or individual primers?
    use.group.identifier <- TRUE
    if (length(unique(seq.groups)) <= 1) {
        use.group.identifier <- FALSE
    } 
    ##### select region of interest
    primer.lengths <- unique(primer.lengths) # ensure that we don't create duplicate primers
    input.start <- start
    input.end <- end
    if (allowed.region.definition == "any") {
        # we're selecting the minimal extension (min primer length) here to ensure that
        # no shorter primers are generated that do not overlap with the target region.
        start <- sapply(start, function(x) max(x - min(primer.lengths) + 1, 1))
        end <- sapply(seq_along(end), function(x) min(end[x] + min(primer.lengths) - 
            1, nchar(seqs)[x]))
    }
    input.seqs <- seqs
    seqs <- substring(seqs, start, end)
    ######### align templates
	message("i) Aligning sequences")
    lex.ali <- align.seqs(seqs, seq.IDs)
    if (is.null(lex.ali)) {
        # alignment wasn't possible
        return(NULL)
    }
	# sanitize the sequences: (carriage returns cause problems in windows)
	lex.ali$seq <- gsub("\r", "", lex.ali$seq)
    bin <- ape::as.DNAbin(lex.ali)
    ali.entropy <- shannon.entropy(as.character(bin))  # determine conservation according to Shannon entropy
    ####### define primer extents
    primer.range <- create.primer.ranges(rep(ncol(bin), length(primer.lengths)), primer.lengths,
                                         rep(1, length(primer.lengths)))
    #print("primer range:")
    #print(primer.range)
    primer.range <- select.primer.region.by.conservation(primer.range, ali.entropy, 
                        conservation, bin)
    ####### create primers from subalignments for every possible primer placement
    i <- NULL
	message("ii) Hierarchical clustering and tree construction")
    min.len <- min(primer.lengths)
    tree.data <- vector("list", nrow(primer.range))
    for (i in seq_len(nrow(primer.range))) { 
        # n.b. removed foreach loop due to memory issues -> no big impact on runtime here
    #tree.seqs <- foreach(i = seq_len(nrow(primer.range)), .combine = "rbind") %dopar% {
        if (is.function(updateProgress)) {
            detail <- identifier
            updateProgress(1/nrow(primer.range), detail, "inc")
        }
        range <- primer.range[i, "Start"]:primer.range[i, "End"]
        primer.candidates <- bin[, range]  # sub-alignment
        t.seqs <- NULL
        # n.b.: disadvantage -> for gappy regions, no primers may be generated here since primers would be too short!
        primer.candidates <- filter.primer.candidates(primer.candidates, min.len)
        ####### construct primers from tree
        tree <- hclust.tree(primer.candidates)
        t.seqs <- get.tree.seqs(tree, max.degen, primer.candidates)
        if (length(t.seqs) != 0 && nrow(t.seqs) != 0) {
            t.seqs <- t.seqs[nchar(t.seqs$Sequence) >= min.len, ]  # select only seqs of target length (can be shorter after merge & gap removal)
            node.ids <- strsplit(t.seqs$ID, split = ",")
            if (use.group.identifier) {
                # use group identifiers 
                m <- lapply(node.ids, function(x) match(x, seq.IDs))
                seq.ids <- sapply(m, function(x) paste0(unique(seq.groups[x]), collapse = ","))
            } else {
                # use the first sequence identifier
                seq.ids <- unlist(lapply(node.ids, function(x) x[1]))
            }
            primer.identifiers <- get.primer.identifier.string(sample, 
                    paste("D_", seq.ids, "_", t.seqs$Degeneracy, sep = ""), 
                    i, primer.range[i, "Start"], primer.range[i, "End"], 
                    identifier, t.seqs$Sequence)
            t.seqs$ID <- primer.identifiers
        }
        tree.data[[i]] <- t.seqs
    }
    tree.seqs <- do.call(rbind, tree.data)
    if (length(tree.seqs) == 0) {
        warning("No degenerate primers were created (gappy alignments, degeneration cutoff?)!")
    }
	message("iii) Augmentation with naive primers")
    # A) create naive primers in the target regions using the alignment to respect conservation:
    ali.matrix <- as.character(bin)
    rownames(ali.matrix) <- NULL
    #print("primer range:")
    #print(primer.range)
    naive.primers <- unlist(lapply(seq_len(nrow(primer.range)), function(x) {
                    range <- primer.range[x, "Start"]:primer.range[x, "End"]
                    if (use.group.identifier) {
                        # use group identifiers 
                        use.ids <- seq.groups
                    } else {
                        # use the first sequence identifier
                        use.ids <- seq.IDs
                    }
                    id <- get.primer.identifier.string(sample, paste0("N_", abbreviate(use.ids, 5), "_", seq_along(use.ids)), 
                                 x, min(range), max(range), identifier, 
                                 paste(rep("X", max(range) - min(range) + 1), collapse = ""))
                    out <- apply(ali.matrix, 1, function(s)  {
                        s <- s[range] # select substring
                        non.gap.idx <- which(s != "-") # remove gaps
                        paste(s[non.gap.idx], collapse = "")
                    })
                    # filter for length
                    len.ok <- which(nchar(out) >= min(primer.lengths))
                    names(out) <- id
                    out <- out[len.ok]
                }))
    # B) create primers for templates for which no primers have been generated yet (due to gappy alignments)
    if (length(naive.primers) != 0) { 
        naive.IDs <- sapply(strsplit(names(naive.primers), split = "\\|"), function(x) x[[1]])
    } else {
        naive.IDs <- NULL
        warning("Naive primer generation from alignment failed. Creating primers from sequences instead.")
    }
    missing.idx <- match(setdiff(seq.IDs, naive.IDs), seq.IDs)
    if (length(missing.idx) != 0) {
        # create primers that are still missing
        missing.primers <- create.primers.naive(input.seqs[missing.idx], seq.groups[missing.idx], 
            seq.IDs[missing.idx], input.start[missing.idx], input.end[missing.idx], primer.lengths, 
            allowed.region.definition, max.degen, sample, identifier, updateProgress)
        naive.primers <- c(naive.primers, missing.primers)
    }
    tree.primers <- tree.seqs$Sequence
    names(tree.primers) <- tree.seqs$ID
    all.seqs <- c(naive.primers, tree.primers)
    # remove duplicates
    all.seqs <- all.seqs[!duplicated(all.seqs)]
    return(all.seqs)
}
#' Shannon Entropy
#' 
#' Computation of Shannon entropy for an alignment.
#'
#' @param ali An alignment of primer sequences.
#' @return The Shannon entropy for the alignment.
#' @keywords internal
shannon.entropy <- function(ali) {
    #gapped.codemap <- IUPAC_CODE_MAP
    #gapped.codemap["-"] <- "-"
    # don't consider ambiguous positions to keep the score tight.
    symbols <- c("a", "c", "g", "t")
    all.counts <- apply(ali, 2, function(x) table(factor(x, levels = symbols)))
    freqs <- apply(all.counts, 2, function(x) x/nrow(ali))
    entropy <- apply(freqs, 2, function(x) -sum(x[x!=0] * log(x[x != 0], base = length(symbols))))
    return(entropy)
}

################
# Constraint: GC clamp
#################

#' GC clamp
#'
#' Determines the number of consecutive G/Cs at the 3' end. 
#'
#' @param y Pirmer sequence from 5' to 3'.
#'
#' @return The length of the GC clamp.
#' @keywords internal
evaluate.GC.clamp <- function(y) {
    y.iupac <- convert.from.iupac(y)
    n <- 5  # consider the last 5 bases
    i <- NULL
    GC.counts <- foreach(i = seq_along(y.iupac), .combine = c) %dopar% {
        x <- y.iupac[[i]]
        K <- rep(n, length(x))
        K[which(nchar(x) < K)] <- nchar(x)[which(nchar(x) < K)]  # nbr of bases to look at
        tails <- sapply(seq_along(x), function(y) substr(x[y], nchar(x[y]) - K[y] + 
            1, nchar(x[y])))
        # count CG hits
        G.hits <- gregexpr("g", tails)
        C.hits <- gregexpr("c", tails)
        hits <- lapply(seq_along(G.hits), function(x) {
            if (G.hits[[x]][1] == -1 && C.hits[[x]][1] == -1) {
                return(NA)
            } else if (G.hits[[x]][1] == -1) {
                return(C.hits[[x]])
            } else if (C.hits[[x]][1] == -1) {
                return(G.hits[[x]])
            } else {
                h <- c(G.hits[[x]], C.hits[[x]])
                h <- h[order(h)]
            }
        })
        # for disambiguated seqs: take the shortest GC clamp found in any seq
        # reason: if we want to have a gc clamp, we dont want to have an ambig
        # seq that represents a seq that actually doesn't have a gc clamp
        gc.counts <- unlist(lapply(seq_along(hits), 
                   function(x) consecutive.GC.count(hits[[x]], nchar(tails)[x])))
        gc.count <- 0
        if (any(gc.counts > 3)) {
            # more than 3 GCs are undesirable
            gc.count <- max(gc.counts)    
        } else {
            # few GCs (e.g. 0) are undesirable
            gc.count <- min(gc.counts)
        }
    }
    return(GC.counts)
}
#' Consecutive GCs
#'
#' Determines the maximum number of consecutive G/Cs
#'
#' @param y Positions where G/C occurs.
#' Positions are numbered from 1 to 5 where 5 is the end of the primer.
#' @param len Is the number of bases from the primer end considered.
#'
#' @return The maximal number of consecutive G/Cs.
#' @keywords internal
consecutive.GC.count <- function(y, len) {
    if (!any(y %in% len)) {
        # no gc clamp because no G/C at the termianl primer pos
        return(0)  
    }
    diffs <- rle(diff(y))
    # select the last consecutive element
    sel <- length(diffs$values)
    if (sel == 0) {
        return(1)
    }
    if (diffs$values[sel] != 1) {
        return(1)
    }
    count <- diffs$lengths[sel] + 1  # x diffs correspond to x+1 numbers
    return(count)
}

##########################
# Template functionalities
#########################

#' Validates a Template Object.
#'
#' Checks whether a Templates object is valid or not.
#'
#' @param object An input data frame to be checked for being a template data frame.
#' @return \code{TRUE}, if the object is valid, \code{FALSE} otherwise.
#' @keywords internal
validate_templates <-  function(object) {
    # specify minimal set of columns that should be present in a template data frame:
    errors <- NULL
    required.fields <- list("ID" = "character", "Header" = "character", 
                  "Identifier" = c("integer", "numeric"), "Sequence_Length" = c("integer", "numeric"),
                  "Group" = "character", "Allowed_Start_fw" = c("integer", "numeric"),
                  "Allowed_End_fw" = c("integer", "numeric"), "Allowed_Start_rev" = c("integer", "numeric"),
                  "Allowed_End_rev" = c("integer", "numeric"), "Allowed_fw" = "character", "Allowed_rev" = "character",
                  "Allowed_Start_fw_initial" = c("integer", "numeric"), "Allowed_End_fw_initial" = c("integer", "numeric"), 
                  "Allowed_Start_rev_initial" = c("integer", "numeric"),
                  "Allowed_End_rev_initial" = c("integer", "numeric"), "Sequence" = "character",
                  "Run" = "character")
    if (!is(object, "data.frame")) {
        errors <- c(errors, "Input was no data frame.")
        return(FALSE)
    }
    possible.fields <- NULL # don't check for additional fields here
    check.fields <- check_setting(possible.fields, object, required.fields)
    if (is.character(check.fields)) {
        errors <- c(errors, check.fields)
    }
    if (length(errors) != 0) {
        return(errors) 
    } else {
        # ensure that 'Run' is unique
        check.run <- length(unique(object$Run)) <= 1
        if (check.run) {
            return(TRUE)
        } else {
            msg <- "The 'Run' column may only contain a single unique value."
            return(msg)
        }
    }
}

#' @rdname Input
#' @name Input
#' @aliases Templates-class
#' @details
#' In the following you can find a description
#' of the most important columns that can be found
#' in an object of class \code{Templates}. Note that
#' angle brackets in the column names 
#' indicate the existence of multiple possibilities.
#' \describe{
#' \item{\code{ID}}{The identifiers of the templates.}
#' \item{\code{Identifier}}{The internal identifiers of the templates.}
#' \item{\code{Group}}{The identifiers of the groups that the templates belong to.}
#' \item{\code{Allowed_Start_<fw|rev>}}{The start of the interval in the templates
#' where binding is allowed for forward and reverse primers, respectively.}
#' \item{\code{Allowed_End_<fw|rev>}}{The end of the interval in the templates
#' where binding is allowed for forward and reverse primers, respectively.}
#' \item{\code{Allowed_<fw|rev>}}{The template sequence where binding
#' is allowed for forward and reverse primers, respectively.}
#' \item{\code{Run}}{An identifier for the set of template sequences.}
#' \item{\code{Covered_By_Primers}}{The identifiers of primers covering the templates,
#' when the template coverage has been annotated.}
#' \item{\code{primer_coverage}}{The number of primers covering the templates,
#' when the template coverage has been annotated.}
#' }
#' @return The \code{Templates} constructor returns 
#' a \code{Templates} object, an instance of a data frame.
#' @exportClass Templates
#' @export
#' @examples
#' 
#' # Load a set of templates:
#' fasta.file <- system.file("extdata", "IMGT_data", "templates", 
#'      "Homo_sapiens_IGH_functional_exon.fasta", package = "openPrimeR")
#' hdr.structure <- c("ACCESSION", "GROUP", "SPECIES", "FUNCTION")
#' template.df <- read_templates(fasta.file, hdr.structure, "|", "GROUP")
setClass("Templates",
   contains = c("data.frame"), validity = validate_templates)


setMethod("initialize", "Templates",
    function(.Object, ...) {
        # just call data frame constructor
        .Object <- callNextMethod(.Object, ...)
    }
) 
setMethod("show", "Templates", function(object) {
    # overwrite the 'print' function
    my_show_df(asS3(object))
})
#' @rdname Input
#' @name Input
#' @aliases Templates-class
#' @export
Templates <- function(...) new("Templates", ...) 

#' cbind for Template class.
#'
#' Ensures that the cbind result has the appropriate class.
#'
#' @param ... Parameters for cbind function.
#' @return Column binded Templates data frame.
#' @keywords internal
#' @export
#' @examples
#' data(Ippolito)
#' template.df <- cbind(template.df, seq_len(nrow(template.df)))
cbind.Templates <- function(...) {
    df <- cbind.data.frame(...)
    df <- Templates(df)
    return(df)
}
#' rbind for Template class.
#'
#' Ensures that the rbind result has the appropriate class.
#'
#' @param ... Parameters for rbind function.
#' @return Row-binded Templates data frame.
#' @keywords internal
#' @export
#' @examples
#' data(Ippolito)
#' template.df <- rbind(template.df, template.df)
rbind.Templates <- function(...) {
    df <- rbind.data.frame(...)
    df <- Templates(df)
    return(df)
}

#' S4 cbind for Templates.
#'
#' S4 cbind function for Templates data frame.
#'
#' @export
#' @return Cbinded template data frame.
#' @rdname Templates-method
#' @param x The Templates data frame.
#' @param y Another data frame.
#' @keywords internal
#' @examples
#' data(Ippolito)
#' template.df <- cbind2(template.df, seq_len(nrow(template.df)))
setMethod("cbind2", "Templates", 
    # need this in case of S3 dispatch fails.
    function(x, y, ...) {
        df <- cbind.data.frame(x, y, ...)
        df <- Templates(df)
        return(df)
    }
)
#' S4 rbind for Templates.
#'
#' S4 rbind function for templates.
#'
#' @export
#' @rdname Templates-method
#' @return Rbinded template data frame.
#' @keywords internal
#' @examples
#' data(Ippolito)
#' template.df <- rbind2(template.df, template.df)
setMethod("rbind2", "Templates", 
    # need this in case of S3 dispatch fails.
    function(x, y, ...) {
        df <- rbind.data.frame(x, y, ...)
        df <- Templates(df)
        return(df)
    }
)

#' Slicing Operator for Templates.
#'
#' Slicing of Templates data frame object.
#'
#' @param x The Template data frame.
#' @param i The row index.
#' @param j The column index.
#' @param ... Other arguments to the slice operator.
#' @param drop Simplify data frame?
#' @exportMethod [
#' @rdname Templates-method
#' @keywords internal
#' @aliases [,Templates-method [,Templates,ANY-method
#' @return Subsetted template data frame.
#' @examples
#' data(Ippolito)
#' template.df <- template.df[1:2,]
setMethod("[", c("Templates", "ANY", "ANY"),
    function(x, i, j, ..., drop = TRUE) {
        if (missing(drop)) {
            df <- asS3(x)[i, j, ...]
        } else {
            df <- asS3(x)[i, j, ..., drop = drop]
        }
        if (is(df, "data.frame")) {
            # only set my class if we haven't simplified.
            #options("show.error.messages" = FALSE)
            t.df <- suppressWarnings(try(Templates(df), silent = TRUE))
            #options("show.error.messages" = TRUE)
            if (is(t.df, "try-error")) {
                # we removed some crucial columns -> No Templates object anymore.
                t.df <- asS3(df)
            }
            df <- t.df
        }
        return(df)
    }
)
#' Dollar Operator for Templates.
#'
#' Set a column in a Templates data frame.
#'
#' @exportMethod $<-
#' @rdname Templates-method
#' @param name The name of the column.
#' @param value The values of the column.
#' @return Templates with replaced column.
#' @keywords internal
#' @examples
#' data(Ippolito)
#' template.df$ID[1] <- "newID"
setMethod("$<-", "Templates", 
    function(x, name, value) {
        df <- asS3(x)
        eval(parse(text = paste("df$", name, " <- value", sep = "")))
        df <- Templates(df)
        return(df)
    }
)
#' Removal of Partial Sequences.
#'
#' Removes partial template sequences.
#'
#' @param template.df Template data frame.
#' @param Header keywords indiating templates that should be excluded.
#' @return Template data frame with partial sequences removed.
#' @keywords internal
remove.seqs.by.keyword <- function(template.df, keyword = "partial") {
    if (length(keyword) == 0) {
        # nothing to remove
        return(template.df)
    }
    rm.idx <- unique(unlist(lapply(keyword, function(x) grep(x, template.df$Header))))
    if (length(rm.idx) != 0) {
        template.df <- template.df[-rm.idx, ]
    }
    return(template.df)
}
#' Parse FASTA headers
#'
#' Parses the headers of a FASTA file.
#'
#' @param hdr The headers (> entries) of a FASTA File.
#' @param delim The delimiter used to separate distinct fields in the headers.
#' For example, \code{|} for headers such as > E.coli | GeneX | ...
#' @param hdr.str Names of the fields appearing in the header.
#' @param id.column Field in the header to be used used as an identifier for the sequences.
#' @return Data frame with structured information from the headers.
#' @keywords internal
parse.header <- function(hdr, delim, hdr.str, id.column) {
    result <- data.frame(Header = hdr, stringsAsFactors = FALSE)
    data.df <- NULL
    if (length(hdr.str) != 0 && delim != "") {
        s <- strsplit(gsub(">", "", hdr), split = delim, fixed = TRUE)
        data.df <- data.frame(matrix(rep(NA, length(hdr.str) * length(hdr)), ncol = length(hdr.str)), 
            stringsAsFactors = FALSE)
        colnames(data.df) <- hdr.str
        for (i in seq_along(s)) {
            h <- s[[i]]
            if (length(hdr.str) > length(h)) {
                my.error("TemplateHeaderStructure", "Header is too short for given header structure!")
            }
            elements <- h[seq_along(hdr.str)]
            data.df[i, ] <- elements
        }
    } else if (length(hdr.str) != 0 && delim == "") {
        # no delimiter or header structure given -> use the complete header info for the first specified field
        s <- gsub(">", "", hdr)
        data.df <- data.frame(Info = s, stringsAsFactors = FALSE)
        colnames(data.df) <- hdr.str[1]
    } 
    if (length(data.df) != 0) {
        result <- cbind(result, data.df)
    }
    if (!id.column %in% colnames(result)) {
        my.warning("TemplateIDColNotFound", paste("ID column: ", id.column, " was not found in the header of the template. Using first available header variable as ID.", 
            sep = ""))
        result <- cbind(ID = result[, 1], result, stringsAsFactors = FALSE)
    } else {
        result <- cbind(ID = result[, id.column], result, stringsAsFactors = FALSE)
    }
    # create an identifier to match to leader sequences: should be unique and be
    # present in leader file
    if (!"Group" %in% colnames(result)) {
        result$Group <- "default"
    }
    return(result)
}

#' @rdname Input
#' @name Input
#' @aliases read_templates
#' @details
#' When loading a FASTA file with \code{read_templates},
#' the input arguments \code{hdr.structure}, \code{delim}, \code{id.column}, 
#' \code{rm.keywords}, \code{remove.duplicates}, \code{fw.region}, 
#' \code{rev.region}, \code{gap.character}, and \code{run} are utilized.
#' Most importantly, \code{hdr.structure} and \code{delim} should
#' match the FASTA header structure.
#' To learn more about setting the primer binding regions, consider the
#' \code{\link{assign_binding_regions}} function.
#' In contrast, when a CSV file is loaded with \code{read_templates},
#' the data are loaded without performing any modifications because the CSV file
#' should represent an object of class \code{\link{Templates}}, which 
#' can be stored using the \code{\link{write_templates}} function.
#'
#' @return \code{read_templates} returns a single object of class
#' \code{Templates} if a single filename was provided or a list
#' of such objects if multiple file names were provided.
#' @export
#' @examples
#' # Load templates from a FASTA file
#' fasta.file <- system.file("extdata", "IMGT_data", "templates", 
#'           "Homo_sapiens_IGH_functional_exon.fasta", package = "openPrimeR")
#' hdr.structure <- c("ACCESSION", "GROUP", "SPECIES", "FUNCTION")
#' template.df.fasta <- read_templates(fasta.file, hdr.structure, "|", "GROUP")
#' # Load mutliple FASTA files
#' fasta.files <- c(fasta.file, fasta.file)
#' template.df.fastas <- read_templates(fasta.files, hdr.structure, "|", "GROUP")
#' # Load templates from a previously stored CSV file
#' csv.file <- system.file("extdata", "IMGT_data", "comparison", 
#'                "templates", "IGH_templates.csv", package = "openPrimeR")
#' template.df.csv <- read_templates(csv.file)
#' # Load multiple CSV files:
#' csv.files <- c(csv.file, csv.file)
#' template.df.csvs <- read_templates(csv.files)
#' # Load a mixture of FASTA/CSV files:
#' mixed.files <- c(csv.file, fasta.file)
#' template.data <- read_templates(mixed.files)
read_templates <- function(fname, hdr.structure = NULL, delim = NULL, id.column = NULL, 
                           rm.keywords = NULL, remove.duplicates = FALSE, fw.region = c(1,30), 
                           rev.region = c(1,30), gap.char = "-", run = NULL) {

    if (length(fname) > 1) {
        template.df <- read_templates_multiple(fname, 
            hdr.structure = hdr.structure, delim = delim, 
            id.column = id.column, rm.keywords = rm.keywords,
            remove.duplicates = remove.duplicates, 
            fw.region = fw.region, rev.region = rev.region, 
            gap.character = gap.char, run = run)
    } else {
        template.df <- read_templates_single(fname, 
            hdr.structure = hdr.structure, delim = delim, 
            id.column = id.column, rm.keywords = rm.keywords,
            remove.duplicates = remove.duplicates, 
            fw.region = fw.region, rev.region = rev.region, 
            gap.character = gap.char, run = run)

    }
    return(template.df)
}
#' Read Template CSV File
#'
#' Reads an input template CSV file.
#'
#' @param fname The filename of the input template CSV file.
#'
#' @return A template data frame.
#' @keywords internal
read_templates_csv <- function(fname) {
    if (!file.exists(fname)) {
        stop("Templates CSV file at '", fname, "' could not be found.")
    }
    templates <- try(read.csv(fname, stringsAsFactors = FALSE, row.names = NULL), silent = TRUE)
    if (is(templates, "try-error")) {
        my.error("TemplateFormatIncorrect", 
            paste0("Could not read the input file: ",
            " '", fname, "' as a CSV. Please check your input!"))
    }
    char.cols <- c("Allowed_fw", "Allowed_rev", 
                    "Covered_By_Primers", "Covered_By_Primers_fw",
                    "Covered_By_Primers_rev")
    num.cols <- c("Allowed_Start_fw","Allowed_Start_rev", 
                    "Allowed_End_fw", "Allowed_End_rev",
                    "Allowed_Start_fw_initial", "Allowed_Start_rev_initial",
                    "Allowed_End_fw_initial", "Allowed_End_rev_initial",
                    "primer_coverage", "primer_coverage_fw", 
                    "primer_coverage_rev")
    c.cols <- char.cols[char.cols %in% colnames(templates)]
    templates[, c.cols] <- apply(templates[, c.cols], 2, as.character)
    n.cols <- num.cols[num.cols %in% colnames(templates)]
    templates[, n.cols] <- apply(templates[,n.cols], 2, as.numeric)
    # set NA to empty string for character columns since "" is converted to NA by write.csv()
    templates[, c.cols][is.na(templates[, c.cols])] <- c("") 
    templates <- try(Templates(templates))
    if (is(templates, "try-error")) {
	error_message <- attr(templates, "condition")$message
        my.error("TemplateFormatIncorrect", 
            paste0("Could not construct a 'Templates' object from the",
                " CSV input file '", fname, "'. Please check that",
                " the structure of the input file is correct. Error:, ", error_message), silent = TRUE)
    }
    return(templates)
}

#' Input of a Single Template File.
#' 
#' Read template sequences from a FASTA or CSV file.
#'
#' When supplying a FASTA file with template sequences,
#' the input arguments \code{hdr.structure}, \code{delim}, \code{id.column}, 
#' \code{rm.keywords}, \code{remove.duplicates}, \code{fw.region}, 
#' \code{rev.region}, \code{gap.character}, and \code{run} are utilized.
#' Most importantly, \code{hdr.structure} and \code{delim} should
#' match the FASTA header structure.
#' When supplying a CSV file with template sequences,
#' the data are loaded without any modification and the CSV file
#' should represent an object of class \code{\link{Templates}}, which 
#' can be stored using the \code{\link{write_templates}} function.
#'
#' @param template.file Path to a FASTA or CSV file containing the template sequences.
#' @param hdr.structure A character vector describing the information contained in the FASTA headers. In case that the headers of \code{fasta.file} contain
#' template group information, please include the keyword "GROUP" in
#' \code{hdr.structure}.
#' @param delim Delimiter for the information in the FASTA headers.
#' @param id.column Field in the header to be used as the identifier.
#' @param rm.keywords A vector of keywords that are used to remove templates whose headers contain any of the keywords.
#' @param remove.duplicates Whether duplicate sequence shall be removed.
#' @param fw.region The positional interval from the template 5' end specifying the 
#' binding sites for forward primers.
#' @param rev.region The positional interval from the template 3' end specifying
#' the binding sites for reverse primers.
#' @param gap.character The character in the input file representing gaps.
#' Gaps are automatically removed upon input.
#' @param run An identifier for the template sequences.
#' @return An object of class \code{Templates}.
#' @keywords internal
read_templates_single <- function(template.file, hdr.structure = NULL, delim = NULL, id.column = NULL, 
                           rm.keywords = NULL, remove.duplicates = FALSE, fw.region = c(1,30), 
                           rev.region = c(1,30), gap.character = "-", run = NULL) {
    template.df <- try(read_templates_fasta(template.file, hdr.structure, 
                       delim, id.column, rm.keywords, remove.duplicates, 
                       fw.region, rev.region, gap.character, run), silent = TRUE)
    if (is(template.df, "try-error")) {
	message("read_templates_csv since fasta failed")
        template.df <- try(read_templates_csv(template.file), silent = TRUE)
        if (is(template.df, "try-error")) {
	    error_message <- attr(template.df, "condition")$message
            my.error("TemplateFormatIncorrect", 
                paste0("Unsupported template input file type or error reading data for file: '", template.file, "'. Error: ", error_message))
        }    
    }
    # check template quality: degeneration
    degen <- score_degen(strsplit(template.df$Sequence, split = ""), gap.char = gap.character)
    degen.idx <- which(degen > 2^10)
    if (length(degen.idx) > 0) {
        msg <- paste("At least one template sequence was highly degenerate.",
                    "Please check the quality of the input template sequences.",
                    "The following sequences will not be disambiguated:",
                    paste(template.df$ID[degen.idx], collapse = ","))
        my.warning("TemplateDegeneration", msg)
    }
    # ensure that IDs of templates are unique
    dup.idx <- which(duplicated(template.df$ID))
    if (length(dup.idx) != 0) {
        # make IDs unique
        non.unique.IDs <- unique(template.df$ID[dup.idx])
        for (ID in non.unique.IDs) {
            idx <- which(template.df$ID == ID)
            new.IDs <- paste0(ID,  "_", seq_along(idx))
            template.df$ID[idx] <- new.IDs
        }
    }
    return(template.df)
}
####
#' Input of Multiple Templates.
#'
#' Reads multiple template CSV/FASTA files.
#'
#' @param filenames Names of FASTA/CSV files containing template data.
#' @param hdr.structure A character vector describing the information contained in the FASTA headers. In case that the headers of \code{fasta.file} contain
#' template group information, please include the keyword "GROUP" in
#' \code{hdr.structure}.
#' @param delim Delimiter for the information in the FASTA headers.
#' @param id.column Field in the header to be used as the identifier.
#' @param rm.keywords A vector of keywords that are used to remove templates whose headers contain any of the keywords.
#' @param remove.duplicates Whether duplicate sequence shall be removed.
#' @param fw.region The positional interval from the template 5' end specifying the 
#' binding sites for forward primers.
#' @param rev.region The positional interval from the template 3' end specifying
#' the binding sites for reverse primers.
#' @param gap.character The character in the input file representing gaps.
#' Gaps are automatically removed upon input.
#' @param run An identifier for the template sequences.
#' @return A list containing objects of class \code{Templates}.
#' @keywords internal
read_templates_multiple <- function(filenames,  hdr.structure = NULL, delim = NULL, id.column = NULL, 
                           rm.keywords = NULL, remove.duplicates = FALSE, fw.region = c(1,30), 
                           rev.region = c(1,30), gap.character = "-", run = NULL) {
    data <- lapply(filenames, function(x) read_templates_single(x,
            hdr.structure = hdr.structure, delim = delim, 
            id.column = id.column, rm.keywords = rm.keywords,
            remove.duplicates = remove.duplicates, 
            fw.region = fw.region, rev.region = rev.region, 
            gap.character = gap.character, run = run))
    data <- set.run.names(data, filenames)
    return(data)
}

#' Input of Template Sequences.
#' 
#' Read template sequences from a FASTA file.
#'
#' @param fasta.file Path to a FASTA file containing the template sequences.
#' @param hdr.structure Names describing the information contained in the FASTA headers. In case that the headers of \code{fasta.file} contain
#' template group information, please include the keyword "GROUP" in
#' \code{hdr.structure}.
#' @param delim Delimiter for the information in the FASTA headers.
#' @param id.column Field in the header to be used as the identifier.
#' @param rm.keywords A vector of keywords that are used to remove templates whose headers contain any of the keywords.
#' @param remove.duplicates Whether duplicate sequence shall be removed.
#' @param fw.region The positional interval from the template 5' end specifying the 
#' binding sites for forward primers.
#' @param rev.region The positional interval from the template 3' end specifying
#' the binding sites for reverse primers.
#' @param gap.character The character in the input file representing gaps.
#' Gaps are automatically removed upon input.
#' @param run An identifier for the template sequences.
#' @return An object of class \code{Templates}.
#' @keywords internal
#' @examples
#' fasta.file <- system.file("extdata", "IMGT_data", "templates", 
#'               "Homo_sapiens_IGH_functional_exon.fasta", package = "openPrimeR")
#' hdr.structure <- c("ACCESSION", "GROUP", "SPECIES", "FUNCTION")
#' template.df <- read_templates(fasta.file, hdr.structure, "|", "GROUP")
read_templates_fasta <- function(fasta.file, hdr.structure = NULL, delim = NULL, id.column = NULL, 
                           rm.keywords = NULL, remove.duplicates = FALSE, fw.region = c(1,30), 
                           rev.region = c(1,30), gap.character = "-", run = NULL) {
    allowed.nts <- c(tolower(names(IUPAC_CODE_MAP)), gap.character)
    seqs <- my.read.fasta(fasta.file, allowed.nts)
    if (is.null(seqs)) {
        return(NULL)
    }
    if (is.null(run)) {
        # use filename w/o extension as identifier
        run <- sub("^([^.]*).*", "\\1", basename(fasta.file))
    }
    if (length(id.column) != 0) { 
        if (length(id.column) != 1) {
            stop("Please specify only a single id.column.")
        }
        id.col.ok <- any(grepl(id.column[1], hdr.structure))
        if (!id.col.ok) {
            my.error("ID_Column_Not_Found", paste0("The specified",
                "ID column  was not part of the header structure."))
        }
    } else {
        # use the first hdr.structure field as the ID if not id.col is specified
        id.column <- hdr.structure[1]
    }
    if (length(hdr.structure) != 0 && length(delim) == 0) {
        warning("Header structure provided, but no delimiter. Cannot read meta information!")
    }
    if (length(hdr.structure) == 0 && length(delim) != 0) {
        warning("Header delimiter supplied, but no header structure. Cannot read meta information!")
    }
    if (length(delim) == 0) {
        delim <- ""
    }
    if (length(hdr.structure) == 0) {
        # no hdr.structure supplied
        id.column <- "HEADER"  # use the full header as the ID column
    }
    # retrieve seqs/replace gap char
    s <- sapply(seqs, function(x) paste(x, collapse = ""))
    hdr <- sapply(seqs, function(x) attr(x, "Annot"))
    # determine IDs and groups parse hdr structure
    hdr.str <- paste(substr(hdr.structure, 1, 1), 
                substr(tolower(hdr.structure), 2, nchar(hdr.structure)), 
                        sep = "")
    id.column <- paste(substr(id.column, 1, 1), substr(tolower(id.column), 2, nchar(id.column)), 
        sep = "")  # rename according to renaming in parse.hdr.structure
    ids <- parse.header(hdr, delim, hdr.str, id.column)  # data frame from parsed header
    if (length(ids) == 0) {
        return(NULL)
    }
    d <- data.frame(ids, Identifier = seq_along(s), InputSequence = s, Sequence_Length = nchar(s), 
        stringsAsFactors = FALSE)
    rownames(d) <- NULL
    # annotate gene groups: try to parse IMGT header structure
    gene.groups <- try(parse.IMGT.gene.groups(d$Group))
    if (is(gene.groups[1], "try-error")) {
            gene.groups <- NULL
    }
    if (!is.null(gene.groups)) {
        # IMGT header structure -> get some more info
        groups <- gene.groups$Gene_Group 
        if (!all(is.na(groups))) {
            d$Group <- groups
        }
    }
    # exclude partial seqs
    if (length(rm.keywords) != 0) {
        d <- remove.seqs.by.keyword(d, rm.keywords)
    }
    # exclude duplicates
    if (remove.duplicates) {
        idx <- which(duplicated(d$InputSequence))
        if (length(idx) != 0) {
            d <- d[-idx,]
        }
    }
    # add default binding regions:
    d$Sequence <- ungap_sequence(d$InputSequence, gap.character)
    default.leaders <- create.uniform.leaders(fw.region, rev.region, d, gap.character)
    d <- cbind(d, default.leaders)
    d$Run <- run
    # re-order:
    last.cols <- c("Sequence", "InputSequence", "Run")
    cols <- setdiff(colnames(d), last.cols)
    cols <- c(cols, last.cols)
    d <- d[, cols]
    template.df <- Templates(d)
    return(template.df)
}
#' Read Sequences.
#'
#' Reads an input FASTA file.
#'
#' @param fasta.file The path to a FASTA file.
#' @param The character indicating gaps in sequences.
#' @return  A data frame with sequences.
#' @keywords internal
read.sequences <- function(fasta.file, gap.character) {
    if (length(fasta.file) == 0) {
        return(NULL)
    }
    # supress warning when there's an extra-newline tryCatch(
    allowed.nts <- c(tolower(names(IUPAC_CODE_MAP)), gap.character)
    seqs <- my.read.fasta(fasta.file, allowed.nts)
    s <- sapply(seqs, function(x) paste(x, collapse = ""))
    hdr <- sapply(seqs, function(x) attr(x, "Annot"))
    d <- data.frame(Identifier = seq_along(s), Header = hdr, Sequence = s, Sequence_Length = nchar(s), 
        stringsAsFactors = FALSE)
    rownames(d) <- NULL
    return(d)
}
#' @rdname TemplatesFunctions
#' @name TemplatesFunctions
#' @aliases update_template_cvg
#' @return \code{update_template_cvg} returns an object of class 
#' \code{Templates} with updated coverage columns.
#' @export
#' @examples
#'
#' # Annotate the coverage of the templates
#' data(Ippolito)
#' template.df <- update_template_cvg(template.df, primer.df)
update_template_cvg <- function(template.df, primer.df, 
                                mode.directionality = NULL) {

    if (length(template.df) == 0 || length(primer.df) == 0) {
        # nothing to update ...
        return(NULL)
    }
    if (!is(primer.df, "Primers")) {
        stop("Please supply a valid primer data frame.")
    }
    if (!is(template.df, "Templates")) {
        stop("Please provide a valid template data frame.")
    }
    # check whether primers and templates correspond to one another
    if (!check_correspondence(primer.df, template.df)) {
        msg <- paste("Could not match all coverage events to the input template data frame.",
            "Please check whether the primers and templates correspond.")
        stop(msg)
    }
    if (is.null(mode.directionality)) {
        mode.directionality <- get.analysis.mode(primer.df)
    } else {
        mode.directionality <- match.arg(mode.directionality, c("fw", "rev", "both"))
    }
    cvg.matrix <- get.coverage.matrix(primer.df, template.df)  # template x primers: 1/0
    if (length(cvg.matrix) != 0) {
        # require any primer to cover a template
        # for direction "both": covered seqs are fw AND rev primers 
        idx <- lapply(seq_len(nrow(cvg.matrix)), function(x) 
            which(cvg.matrix[x,] != 0))  # idx of covering primers per template
    } else {
        idx <- NULL
    }
    # add columns:
    cols <- c("Covered_By_Primers")
    extra.cols <- c(paste(cols, "_fw", sep = ""), paste(cols, "_rev", sep = ""))
    all.cols <- c(cols, extra.cols)
    for (i in seq_along(all.cols)) {
        template.df[, all.cols[i]] <- rep("", nrow(template.df))
    }
    cols <- c("primer_coverage")
    extra.cols <- c(paste(cols, "_fw", sep = ""), paste(cols, "_rev", sep = ""))
    all.cols <- c(cols, extra.cols)
    for (i in seq_along(all.cols)) {
        template.df[, all.cols[i]] <- rep(0, nrow(template.df))
    }
    if (length(idx) != 0) {
        fw.primer.idx <- which(primer.df$Forward != "")
        rev.primer.idx <- which(primer.df$Reverse != "")
        if (mode.directionality == "both") { # require fw & rev binding events
            # coverage by both a foward and a reverse primer?
            both.idx <- sapply(idx, function(x) any(x %in% fw.primer.idx) & any(x %in% rev.primer.idx))
            primer.IDs <- lapply(seq_along(idx), function(x) if(both.idx[x]) primer.df$Identifier[idx[[x]]] else numeric(0))
        } else {
            primer.IDs <- lapply(idx, function(x) primer.df$Identifier[x])
            
        }
        template.df$Covered_By_Primers <- sapply(primer.IDs, function(x) paste(x, collapse = ","))
        # n.b.: for primers with two directions, template cvg is total nbr of fw/rev primers!
        template.df$primer_coverage <- sapply(primer.IDs, length) 
        # individual direction coverage:
        new.entries <- data.frame(
                Covered_By_Primers_fw = "", 
                primer_coverage_fw = 0, 
                Covered_By_Primers_rev = "", 
                primer_coverage_rev = 0, stringsAsFactors = FALSE)
        col.idx <- which(!colnames(new.entries) %in% colnames(template.df))
        if (length(col.idx) != 0) {
            # add missing fields 
            template.df <- cbind(template.df, new.entries[,col.idx])
        }
        if (mode.directionality == "fw" || mode.directionality == "both") {
            idx.fw <- lapply(idx, function(x) intersect(x, fw.primer.idx))
            primer.IDs.fw <- lapply(idx.fw, function(x) primer.df$Identifier[x])
            template.df$Covered_By_Primers_fw <- sapply(primer.IDs.fw, function(x) paste(x, 
                collapse = ","))
            template.df$primer_coverage_fw <- sapply(primer.IDs.fw, length)
        }
        if (mode.directionality == "rev" || mode.directionality == "both") {
            idx.rev <- lapply(idx, function(x) intersect(x, rev.primer.idx))
            primer.IDs.rev <- lapply(idx.rev, function(x) primer.df$Identifier[x])
            template.df$Covered_By_Primers_rev <- sapply(primer.IDs.rev, function(x) paste(x, 
                collapse = ","))
            template.df$primer_coverage_rev <- sapply(primer.IDs.rev, length)
        }
    } 
    return(template.df)
}

#' @rdname TemplatesFunctions
#' @name TemplatesFunctions
#' @aliases adjust_binding_regions
#' @details
#' When modifying binding regions with \code{adjust_binding_regions}, new
#' binding intervals can be specified via \code{fw} and \code{rev}
#' for forward and reverse primers, respectively. The new regions should
#' be provided relative to the existing definition of binding regions 
#' in \code{template.df}.
#' For specifying the new binding regions, position \code{0} refers to 
#' the first position after the end of the existing
#' binding region. Hence, negative positions relate to regions within the
#' existing binding region, while non-negative values relate to positions
#' outside the defined binding region.
#'
#' @return \code{adjust_binding_regions} returns a \code{Templates} object with
#' updated binding regions.
#' @export
#' @examples
#' data(Ippolito)
#' # Extend the binding region by one position
#' relative.interval <- c(-max(template.df$Allowed_End_fw), 0)
#' template.df.adj <- adjust_binding_regions(template.df, relative.interval)
#' # compare old and new annotations:
#' head(cbind(template.df$Allowed_Start_fw, template.df$Allowed_End_fw))
#' head(cbind(template.df.adj$Allowed_Start_fw, template.df.adj$Allowed_End_fw))
adjust_binding_regions <- function(template.df, region.fw, region.rev) {
    if (!is(template.df, "Templates")) {
        stop("'template.df' should be a 'Templates' object.")
    }
    if (missing(region.fw)) {
        region.fw <- NULL
    }
    if (missing(region.rev)) {
        region.rev <- NULL
    }
    cols <- c("Allowed_Start_fw", "Allowed_End_fw", "Allowed_Start_rev", "Allowed_End_rev")
    if (length(region.fw) != 0) {
        if (length(region.fw) != 2) {
            stop("'region.fw' wasn't an interval.")
        }
        min <- region.fw[1]
        max <- region.fw[2]
        template.df <- update.individual.binding.region(min, max, template.df, "fw")
    }
    if (length(region.rev) != 0) {
        if (length(region.rev) != 2) {
            stop("'region.rev' wasn't an interval.")
        }
        min <- region.rev[1]
        max <- region.rev[2]
        template.df <- update.individual.binding.region(min, max, template.df, "rev")
    }
    return(template.df)
}
#' Adjustment of Existing Binding Regions for one Direction.
#'
#' Adjusts the existing annotation of binding regions by specifying
#' an interval relative to the existing binding region. 
#'
#' Position \code{0} indicates the first position after the existing
#' binding region. Hence, negative positions adjust the binding region towards
#' the existing binding regions and non-negative positionis extend
#' the existing binding region definition aways from the existing target region.
#'
#' @param min Position where binding should start.
#' @param max End position of binding.
#' @param Template data frame.
#' @param mode.directionality Directionality of primers.
#' @return Template data frame with updated binding regions.
#' @keywords internal
update.individual.binding.region <- function(min, max, template.df, mode.directionality) {
    # min/max: new setting for binding position relative to target region
    allowed.start.initial <- paste("Allowed_Start_", mode.directionality, "_initial", 
        sep = "")
    allowed.start <- paste("Allowed_Start_", mode.directionality, sep = "")
    allowed.end.initial <- paste("Allowed_End_", mode.directionality, "_initial", 
        sep = "")
    allowed.end <- paste("Allowed_End_", mode.directionality, sep = "")
    allowed.region <- paste("Allowed_", mode.directionality, sep = "")
    if (mode.directionality == "fw") {
        target.pos <- template.df[, allowed.end.initial] + 1  # absolute target position
    } else {
        target.pos <- template.df[, allowed.start.initial] - 1
    }
    if (length(min) != 0 && length(max) != 0) {
        # start
        if (mode.directionality == "fw") {
            new.start <- sapply(target.pos + min, function(x) max(x, 1))
            sel <- new.start >= nchar(template.df$Sequence)
            new.start[sel] <- nchar(template.df$Sequence)[sel]
            template.df[, allowed.start] <- new.start
        } else {
            new.end <- sapply(seq_len(nrow(template.df)), function(x) min(target.pos[x] - 
                min, nchar(template.df$Sequence[x])))
            new.end[new.end <= 0] <- 1
            template.df[, allowed.end] <- new.end
        }
        # end
        if (mode.directionality == "fw") {
            new.end <- sapply(seq_len(nrow(template.df)), function(x) min(target.pos[x] + 
                max, nchar(template.df$Sequence[x])))
            new.end[new.end <= 0] <- 1
            template.df[, allowed.end] <- new.end
        } else {
            new.start <- sapply(target.pos - max, function(x) max(x, 1))
            template.df[, allowed.start] <- new.start
        }
        # update allowed region
        template.df[, allowed.region] <- substring(template.df$Sequence, template.df[, 
            allowed.start], template.df[, allowed.end])
    }
    return(template.df)
}
#' @rdname TemplatesFunctions
#' @name TemplatesFunctions
#' @aliases assign_binding_regions
#' @details
#' Binding regions are defined using \code{assign_binding_regions}, where
#' the arguments \code{fw} and \code{rev} provide data describing
#' the binding regions of the forward and reverse primers, respectively.
#' To specify binding regions for each template
#' individually, \code{fw} and \code{rev} should provide the paths to FASTA files. The headers of these
#' FASTA file should match the headers of the loaded \code{template.df} 
#' and the sequences in the files specified by \code{fw} and \code{rev} should indicate
#' the target binding regions. 
#'
#' To specify uniform binding regions,
#' \code{fw} and \code{rev} should be numeric intervals indicating the allowed
#' binding range for primers in the templates. Setting the forward interval
#' to (1,30) indicates that the first 30 bases should be used for forward primers
#' and specifying the reverse interval to (1,30) indicates that the last
#' 30 bases should be used for reverse primer binding.
#'
#' If \code{optimize.region} is \code{TRUE}, the input binding region is
#' adjusted such that regions forming secondary structures are avoided.
#'
#' @return \code{assign_binding_regions} returns an object of class \code{Templates} with newly assigned binding regions.
#' @export
#' @note \code{assign_binding_regions} requires the program ViennaRNA
#' (https://www.tbi.univie.ac.at/RNA/)
#' for adjusting the binding regions when \code{optimize.region} 
#' is set to \code{TRUE}.
#' @examples
#' data(Ippolito)
#' # Assignment of individual binding regions
#' l.fasta.file <- system.file("extdata", "IMGT_data", "templates", 
#'      "Homo_sapiens_IGH_functional_leader.fasta", package = "openPrimeR")
#' template.df.individual <- assign_binding_regions(template.df, l.fasta.file, NULL)
#' # Assign the first/last 30 bases as forward/reverse binding regions
#' template.df.uniform <- assign_binding_regions(template.df, c(1,30), c(1,30))
#' # Optimization of binding regions (requires ViennaRNA)
#' \dontrun{template.df.opti <- assign_binding_regions(template.df, c(1,30), c(1,30),
#'                      optimize.region = TRUE, primer.length = 20)}
assign_binding_regions <- function(template.df, fw = NULL, rev = NULL, optimize.region = FALSE, 
                                   primer.length = 20, gap.char = "-") {
    if (is.null(fw) && !is.null(rev)) {
        fw <- eval(parse(text = paste(class(rev), "(0)", sep = "")))
    } else if (is.null(rev) && !is.null(fw)) {
        rev <- eval(parse(text = paste(class(fw), "(0)", sep = "")))
    } else if (is.null(fw) && is.null(rev)) {
        warning("'fw' and 'rev' were both NULL. No binding regions were annotated.")
        return(template.df)
    }
    if (!is(template.df, "Templates")) {
        stop("Please supply a valid template data frame.")
    }
    if (optimize.region && length(primer.length) != 1) {
        stop("Please specify a numeric, scalar primer length.")
    }
    UseMethod("assign_binding_regions", fw)
}

#' Numeric Assignment of Binding Regions.
#'
#' Numeric S3 generic case for assigning binding regions.
#'
#' @param template.df Template data frame.
#' @param fw Binding region data forward primers.
#' @param rev Binding region data for reverse primers.
#' @param optimize.region Should the primer binding region be optimized using secondary structure prediction?
#' @param primer.length Probe length for optimizing template secondary structure.
#' @return The template data frame with assigned binding regions.
#' @export
#' @keywords internal
assign_binding_regions.numeric <- function(template.df, fw, rev, optimize.region = FALSE, 
                                           primer.length = 20, gap.char = "-") {

    if (missing(rev)) {
        rev <- NULL
    }
    if (length(fw) > 2 || length(rev) > 2 ||
        length(fw) == 1 || length(rev) == 1) {
        stop("fw or rev were no intervals.")
    }
    template.df$Sequence <- ungap_sequence(template.df$InputSequence, gap.char)
    l.seq <- create.uniform.leaders(fw, rev, template.df, gap.char)
    template.df <- add.uniform.leaders.to.seqs(template.df, l.seq)
    # optimization of regions:
    if (optimize.region) {
        if (length(fw) != 0 && length(rev) != 0) {
            mode.directionality <- "both"
        } else if (length(fw) != 0) {
            mode.directionality <- "fw"
        } else {
            mode.directionality <- "rev"
        }
        opti.regions <- optimize.template.binding.regions.dir(template.df, NULL, primer.length, mode.directionality)
        opti.intervals <- opti.regions$Intervals
        template.df <- update.binding.regions(template.df, opti.intervals)
    }
    return(template.df)
}
#' Character Assignment of Binding Regions
#'
#' Generic method for assigning the binding region using individual binding regions.
#'
#' @param template.df Template data frame.
#' @param fw FASTA file specifying the forward binding regions.
#' @param rev FASTA file specifying the reverse binding regions.
#' @param optimize.region Should the primer binding region be optimized using secondary structure prediction?
#' @param primer.length Probe length for optimizing template secondary structure.
#' @param gap.char The gap character for aligned sequences.
#' @return Template data frame with assigned binding regions.
#' @export
#' @keywords internal
assign_binding_regions.character <- function(template.df, fw, rev, optimize.region = FALSE, 
                                    primer.length = 20, gap.char = "-") {

    fw.df <- read.leaders(fw, "fw", gap.character = gap.char)
    rev.df <- read.leaders(rev, "rev", gap.character = gap.char)
    uni.leaders <- unify.leaders(fw.df, rev.df, template.df, gap.char)
    template.df <- get.leader.exon.regions(template.df, uni.leaders)
    # optimization of regions:
    if (optimize.region) {
        if (length(fw) != 0 && length(rev) != 0) {
            mode.directionality <- "both"
        } else if (length(fw) != 0) {
            mode.directionality <- "fw"
        } else {
            mode.directionality <- "rev"
        }
        opti.regions <- optimize.template.binding.regions.dir(template.df, NULL, primer.length, mode.directionality)
        opti.intervals <- opti.regions$Intervals
        template.df <- update.binding.regions(template.df, opti.intervals)
    }
    return(template.df)
}
#' Update of Binding Regions.
#'
#' Updates the binding regions in the templates by providing
#' new intervals for forward and reverse binding regions.
#'
#' @param template.df An object of class \code{Templates}.
#' @param opti.regions List with new binding intervals.
#' The list can contain the components \code{fw} and \code{rev} providing
#' numeric vectors of length 2 providing the start and end
#' of the binding regions in the templates, for forward
#' and reverse binding regions, respectively.
#' @return A \code{Templates} object with updated binding regions.
#' @keywords internal
update.binding.regions <- function(template.df, opti.regions) {
    # updates binding regions
    if ("fw" %in% names(opti.regions)) {
        if (length(opti.regions[["fw"]]) != 0) {
            starts <- sapply(opti.regions[["fw"]], function(x) x[1])
            sel <- !is.na(starts)  # only modify those where we could find a possible region ...
            template.df$Allowed_Start_fw[sel] <- starts[sel]
            template.df$Allowed_End_fw[sel] <- sapply(opti.regions[["fw"]], function(x) x[2])[sel]
            template.df$Allowed_fw[sel] <- substring(template.df$Sequence[sel], template.df$Allowed_Start_fw[sel], 
                template.df$Allowed_End_fw[sel])
        }
    }
    if ("rev" %in% names(opti.regions)) {
        if (length(opti.regions[["rev"]]) != 0) {
            starts <- sapply(opti.regions[["rev"]], function(x) x[1])
            sel <- !is.na(starts)
            template.df$Allowed_Start_rev[sel] <- starts[sel]
            template.df$Allowed_End_rev[sel] <- sapply(opti.regions[["rev"]], function(x) x[2])[sel]
            template.df$Allowed_rev[sel] <- substring(template.df$Sequence[sel], template.df$Allowed_Start_rev[sel], 
                template.df$Allowed_End_rev[sel])
        }
    }
    return(template.df)
}
#' Ungapping of Sequences.
#'
#' Removes gaps from the input sequences.
#'
#' @param seqs The input character vector with sequences
#' @param gap.char The character used to represent gaps.
#' @return \code{seqs} with gaps removed.
#' @keywords internal
ungap_sequence <- function(seqs, gap.char = "-") {
    seqs <- gsub(gap.char, "", seqs)
    return(seqs)
}
binding.interval.ungapped <- function(template.df, start, end, gap.char) {
    # adjust the binding positions with gaps to those without gaps 
    # gap count from start of seq until the end of the allowed region
    gap.count.before <- sapply(strsplit(substring(template.df$InputSequence, 1, start), split = ""),
                    function(x) length(which(x == gap.char)))
    gap.count.within <- sapply(strsplit(substring(template.df$InputSequence, start, end), split = ""),
                    function(x) length(which(x == gap.char)))
    start <- start - gap.count.before
    end <- end - gap.count.before - gap.count.within
    return(list(start, end))
}
#' Retrieval of Binding Regions
#'
#' Retrives information about individual binding regions.
#'
#' @param template.df Template data frame.
#' @param direction Identify forward and reverse.
#' @param start Start positions.
#' @param end End positions.
#' @param gap.char The character for gaps in alignments.
#' @param init.from.leader Whether the binding regions are initialized
#'        from leader sequences.
#' @return Data frame with information on allowed binding regions.
#' @keywords internal
retrieve.leader.region <- function(template.df, direction = c("fw", "rev"), start, end, gap.char,
                                   init.from.leader) {
    if (length(start) == 0 || length(end) == 0) {
        return(NULL)
    }
    if (length(direction) == 0) {
        stop("Please provide the 'direction' argument.")
    }
    direction <- match.arg(direction)
    if (length(start) == 1 && length(end) == 1) {
        # start and end refer to intervals (uniform range)
        start <- rep(start, nrow(template.df))
        end <- rep(end, nrow(template.df))
    } # otherwise: regions are already ok in the input.
    if (direction == "rev" && !init.from.leader) {
        # only for uniform binding regions: need to change the interpretation of rev positions
        ori.start <- start
        start <- nchar(template.df$InputSequence) - end + 1
        end <- nchar(template.df$InputSequence) - ori.start + 1
    }
    # store start/end in possibly gappy input
    ali.start <- start
    ali.end <- end
    # gap count from start of seq until the end of the allowed region
    # adjust binding region in the ungapped sequence according to the number of gaps in between
    new.interval <- binding.interval.ungapped(template.df, start, end, gap.char)
    # work with start/end in ungapped seqs
    start <- new.interval[[1]]
    end <- new.interval[[2]]
    # use the ungapped sequences now to work on the correct positions 
    leader <- substring(template.df$Sequence, start, end)
    # deal with incorrect positions
    idx <- which(start < 0 & end < 0)  # interval is not in range -> no leader available
    if (length(idx) != 0) {
        start[idx] <- NA
        end[idx] <- NA
    }
    idx <- which(end < 0)  # only end is too small
    if (length(idx) != 0) {
        end[idx] <- 0  # no leader available
    }
    idx <- which(start < 0)  # only start is too early
    if (length(idx) != 0) {
        start[idx] <- 1
    }
    idx <- which(unlist(lapply(seq_along(end), function(x) end[x] > (nchar(template.df$Sequence[x]) + 
        1) && start[x] > (nchar(template.df$Sequence[x]) + 1))))
    if (length(idx) != 0) {
        # not in the defined range
        end[idx] <- NA
        start[idx] <- NA
    }
    idx <- which(unlist(lapply(seq_along(start), function(x) start[x] > (1 + nchar(template.df$Sequence[x])))))  #only start is larger than the range
    if (length(idx) != 0) {
        start[idx] <- 0
        end[idx] <- 0
    }
    idx <- which(unlist(lapply(seq_along(end), function(x) end[x] > (1 + nchar(template.df$Sequence[x])))))  #only end is larger than the range
    if (length(idx) != 0) {
        end[idx] <- nchar(template.df$Sequence)[idx]
    }
    df <- build_leader_df(direction, leader, start, end, ali.start, ali.end)
    return(df)
}
#' Building of Leader Data Frame.
#'
#' Constructs the leader data frame.
#'
#' @param direction The primer direction for which we are annotating binding regions.
#' @param leader The binding region sequence.
#' @param start The start positions of the binding region.
#' @param end The end positions of the binding region.
#' @param ali.start The start positions of the binding region in the aligned input.
#' @param ali.end The end positions of the binding region in the aligned input.
#' @return A data frame with binding region information.
#' @keywords internal
build_leader_df <- function(direction = c("fw", "rev"), leader, start, end,
                            ali.start, ali.end) {
    if (length(direction) == 0) {
        stop("Please provide the 'direction' argument.")
    }
    direction <- match.arg(direction)
    # initial: stored such that binding range can be modified later in relation to initial configuration
    df <- data.frame(Allowed_fw = character(length(leader)),  # binding region string
                    Allowed_rev = character(length(leader)),
                    # binding region intervals:
                    Allowed_Start_fw = numeric(length(start)), 
                    Allowed_Start_rev = numeric(length(start)),
                    Allowed_End_fw = numeric(length(end)), 
                    Allowed_End_rev = numeric(length(end)),
                    # the binding region in gapped sequence positions from alignemnts
                    Allowed_Start_fw_ali = numeric(length(start)), 
                    Allowed_Start_rev_ali = numeric(length(start)),
                    Allowed_End_fw_ali = numeric(length(end)), 
                    Allowed_End_rev_ali = numeric(length(end)),
                    # initial: the initially specified binding region (without gaps)
                    Allowed_Start_fw_initial = numeric(length(start)), 
                    Allowed_Start_rev_initial = numeric(length(start)), 
                    Allowed_End_fw_initial = numeric(length(end)), 
                    Allowed_End_rev_initial = numeric(length(end)),
                    # initial positions in aligned input:
                    Allowed_Start_fw_initial_ali = numeric(length(start)), 
                    Allowed_Start_rev_initial_ali = numeric(length(start)), 
                    Allowed_End_fw_initial_ali = numeric(length(end)), 
                    Allowed_End_rev_initial_ali = numeric(length(end)),
                    stringsAsFactors = FALSE)
    if (direction == "rev") {
        df$Allowed_rev <- leader
        df$Allowed_Start_rev <- start
        df$Allowed_End_rev <- end
        df$Allowed_Start_rev_ali <- ali.start
        df$Allowed_End_rev_ali <- ali.end
        df$Allowed_Start_rev_initial <- df$Allowed_Start_rev
        df$Allowed_End_rev_initial <- df$Allowed_End_rev
        df$Allowed_Start_rev_initial_ali <- df$Allowed_Start_rev_ali
        df$Allowed_End_rev_initial_ali <- df$Allowed_End_rev_ali
    } else {
        df$Allowed_fw <- leader
        df$Allowed_Start_fw <- start
        df$Allowed_End_fw <- end
        df$Allowed_Start_fw_ali <- ali.start
        df$Allowed_End_fw_ali <- ali.end
        df$Allowed_Start_fw_initial <- df$Allowed_Start_fw
        df$Allowed_End_fw_initial <- df$Allowed_End_fw
        df$Allowed_Start_fw_initial_ali <- df$Allowed_Start_fw_ali
        df$Allowed_End_fw_initial_ali <- df$Allowed_End_fw_ali
    }
    # we return only the modified binding direction such that 
    # previous settings in template.df are not overwritten:
    sel.cols <- grepl(direction, colnames(df))
    df <- df[,sel.cols]
    return(df)
}
#' Uniform Binding Ranges.
#'
#' Creates uniform binding regions for all templates.
#'
#' @param fw.interval Binding region for forward templates.
#' @param rev.interval Binding region for reverse templates.
#' @param template.df Template data frame.
#' @param gap.char The character for gaps in alignments.
#' @return Data frame with binding region information.
#' @keywords internal
create.uniform.leaders <- function(fw.interval, rev.interval, template.df, gap.char) {
    if (length(fw.interval) == 0 && length(rev.interval) == 0) {
        warning("No fw/rev intervals were specified for specifying a uniform binding range.")
    }
    if (length(fw.interval) == 2) {
        if (fw.interval[1] > fw.interval[2]) {
            stop("Forward binding region was no proper interval.")
        }
    } else if (length(fw.interval) > 0) {
        stop("Forward binding region was no proper interval.")
    }
    if (length(rev.interval) == 2) {
        if (rev.interval[1] > rev.interval[2]) {
            stop("Reverse binding region was no proper interval.")
        }
    } else if (length(rev.interval) > 1) {
        stop("Reverse binding region was no proper interval.")
    }
    fw.leaders <- retrieve.leader.region(template.df, "fw", fw.interval[1], fw.interval[2], gap.char, FALSE)
    rev.leaders <- retrieve.leader.region(template.df, "rev", rev.interval[1], rev.interval[2], gap.char, FALSE)
    if (length(fw.leaders) == 0) {
        result <- rev.leaders
    } else if (length(rev.leaders) == 0) {
        result <- fw.leaders
    } else {
        result <- cbind(fw.leaders[,grep("_fw", colnames(fw.leaders))], 
                        rev.leaders[, grep("_rev", colnames(rev.leaders))])
    }
    col.order <- c("Allowed_Start_fw", "Allowed_End_fw", "Allowed_Start_rev", "Allowed_End_rev", 
        "Allowed_fw", "Allowed_rev")
    all.cols <- c(col.order[col.order %in% colnames(result)], setdiff(colnames(result), 
        col.order))
    result <- result[, all.cols]
    return(result)
}
#' Add Uniform Binding Regions.
#'
#' Augments a template data frame with uniform binding regions.
#'
#' @param lex.seq Template data frame
#' @param leaders Data frame with uniform binding regions.
#' @return Template data frame with updated binding regions.
#' @keywords internal
add.uniform.leaders.to.seqs <- function(lex.seq, leaders) {
    result <- lex.seq
    if ("TemplateIdentifier" %in% colnames(leaders)) {
        # need to map from leader annotation to templates
        m <- match(leaders$TemplateIdentifier, lex.seq$Identifier)
    } else {
        # leader annotation is not a subset of the templates
        m <- seq_len(nrow(leaders))
    }
    for (i in colnames(leaders)) {
        result[m, i] <- leaders[, i]
    }
    return(result)
}
#' Read Individual Binding Regions
#'
#' Reads individual binding regions into a data frame.
#'
#' @param fasta.file Path to a FASTA file with binding regions.
#' @param direction String identifying whether the FASTA file
#' contains information pertaining to the binding region of forward
#' or reverse primers.
#' @param rm.keywords A vector of keywords that are used to remove templates whose headers contain any of the keywords.
#' @param gap.character The character for indicating gaps in sequences.
#' @return A data frame with individual binding regions.
#' @keywords internal
read.leaders <- function(fasta.file, direction = c("fw", "rev"), rm.keywords = NULL, gap.character) {
    if (length(fasta.file) == 0) {
        return(NULL)
    }
    d <- read.sequences(fasta.file, gap.character)
    if (direction == "fw") {
        colnames(d)[which(colnames(d) == "Sequence")] <- "Allowed_fw"
    } else {
        colnames(d)[which(colnames(d) == "Sequence")] <- "Allowed_rev"
    }
    if (length(rm.keywords) != 0) {
        d <- remove.seqs.by.keyword(d, rm.keywords)
    }
    return(d)
}
#' Parser for IMGT Groups.
#'
#' Parses IMGT group information contained in FASTA headers.
#'
#' @param IDs Group information strings to be parsed.
#' @return Data frame with structured group information.
#' @keywords internal
parse.IMGT.gene.groups <- function(IDs) {
    # regexp: (gene)-(subgroup)-(localization in the locus)-(extracted_labels)
    exp <- "([^-]+)-(.+)(\\*[^\\|]+)"
    # some genes are ORs (orphons) -> still the same gene group?  orphons appear
    # outside the main genetic loci but are still assigned to groups ..  clan
    # annotations: clan1: ighv1, ighv5, ighv7
    idx <- stringr::str_match(IDs, exp)
    if (any(apply(idx, 1, is.na))) {
        # at least one entry not in IMGT format
        return(NULL)
    }
    # remove OR annotation from GeneGroup
    colnames(idx) <- c("Full", "Gene", "Subgroup", "Allele")
    s <- strsplit(idx[, "Gene"], split = "/")
    g <- sapply(s, function(x) x[1])
    idx[, "Gene"] <- g
    # remove localization from Subgroup
    s <- strsplit(idx[, "Subgroup"],  split = "-")
    g <- sapply(s, function(x) x[1])
    idx[, "Subgroup"] <- g
    identifier <- paste(idx[, "Gene"], idx[, "Subgroup"], sep = "-")
    result <- data.frame(Group_Identifier = identifier, Gene_Group = idx[, "Gene"], 
        Sub_Group = idx[, "Subgroup"], Allele = idx[, "Allele"], 
        stringsAsFactors = FALSE)
    # message(result)
    return(result)
}
#' Unification of Leaders
#'
#' Unifies individual binding regions for forward and reverse primers.
#'
#' @param l.seq.fw Data frame with binding information for forward primers.
#' @param l.seq.rev Data frame with binding information for reverse primers.
#' @param lex.seq Template data frame.
#' @param gap.char The character for indicating alignment gaps.
#' @return Template data frame with annotated binding regions.
#' @keywords internal
unify.leaders <- function(l.seq.fw, l.seq.rev, lex.seq, gap.char) {
    fw.df <- NULL
    rev.df <- NULL
    if (length(l.seq.fw) == 0 && length(l.seq.rev) == 0) {
        return(lex.seq)
    }
    template.df <- lex.seq
    if (length(l.seq.fw) != 0) {
        fw.df <- get.leader.exon.regions.single(l.seq.fw, lex.seq, "fw", gap.char) 
        template.df <- add.uniform.leaders.to.seqs(template.df, fw.df)
    }
    if (length(l.seq.rev) != 0) {
        rev.df <- get.leader.exon.regions.single(l.seq.rev, lex.seq, "rev", gap.char)
        template.df <- add.uniform.leaders.to.seqs(template.df, rev.df)
    }
    template.df$Allowed_Start_fw_initial <- template.df$Allowed_Start_fw
    template.df$Allowed_End_fw_initial <- template.df$Allowed_End_fw
    template.df$Allowed_End_rev_initial <- template.df$Allowed_End_rev
    template.df$Allowed_Start_rev_initial <- template.df$Allowed_Start_rev
    template.df$Sequence <- ungap_sequence(template.df$InputSequence, gap.char) # use the ungapped seq from now on only
    return(template.df)
}
#' Assign Binding Regions
#' 
#' Augments a template data frame with individual binding regions.
#' 
#' @param lex.seqs Data frame with template sequences.
#' @param uni.leaders Data frame with individual allowed binding regions.
#' @return Template data frame with annotated binding regions.
#' @keywords internal
get.leader.exon.regions <- function(lex.seqs, uni.leaders) {
    if (length(uni.leaders) == 0) {
        return(NULL)
    }
    # add missing columns
    if (!"Allowed_Start_fw" %in% colnames(uni.leaders)) {
        uni.leaders$Allowed_Start_fw <- lex.seqs$Allowed_Start_fw
        uni.leaders$Allowed_End_fw <- lex.seqs$Allowed_End_fw
        uni.leaders$Allowed_fw <- ""
        uni.leaders$Allowed_Start_fw_initial <- uni.leaders$Allowed_Start_fw
        uni.leaders$Allowed_End_fw_initial <- uni.leaders$Allowed_End_fw
        
    }
    if (!"Allowed_Start_rev" %in% colnames(uni.leaders)) {
        uni.leaders$Allowed_Start_rev <- lex.seqs$Allowed_Start_rev
        uni.leaders$Allowed_End_rev <- lex.seqs$Allowed_End_rev
        uni.leaders$Allowed_rev <- ""
        uni.leaders$Allowed_Start_rev_initial <- uni.leaders$Allowed_Start_rev
        uni.leaders$Allowed_End_rev_initial <- uni.leaders$Allowed_End_rev
    }
    # add entries from lex.seqs, where uni.leaders didn't have anything
    fw.idx <- which(uni.leaders$Allowed_fw == "")
    rev.idx <- which(uni.leaders$Allowed_rev == "")
    if (length(fw.idx) != 0) {
        uni.leaders$Allowed_Start_fw[fw.idx] <- lex.seqs$Allowed_Start_fw[fw.idx]
        uni.leaders$Allowed_End_fw[fw.idx] <- lex.seqs$Allowed_End_fw[fw.idx]
        uni.leaders$Allowed_Start_fw_initial[fw.idx] <- lex.seqs$Allowed_Start_fw[fw.idx]
        uni.leaders$Allowed_End_fw_initial[fw.idx] <- lex.seqs$Allowed_End_fw[fw.idx]
    }
    if (length(rev.idx) != 0) {
        uni.leaders$Allowed_Start_rev[rev.idx] <- lex.seqs$Allowed_Start_rev[rev.idx]
        uni.leaders$Allowed_End_rev[rev.idx] <- lex.seqs$Allowed_End_rev[rev.idx]
        uni.leaders$Allowed_Start_rev_initial[rev.idx] <- lex.seqs$Allowed_Start_rev[rev.idx]
        uni.leaders$Allowed_End_rev_initial[rev.idx] <- lex.seqs$Allowed_End_rev[rev.idx]
        
    }
    # remove duplicate cols
    dup.idx <- which(duplicated(colnames(uni.leaders)))
    if (length(dup.idx) != 0) {
        uni.leaders <- uni.leaders[, -dup.idx]
    }
    return(uni.leaders)
}
#' Individual Binding Annotation
#'
#' Annotate individual binding regions.
#'
#' @param l.seq Data frame with individual binding regions.
#' @param lex.seq Template data frame.
#' @param direction The primer direction for which the binding info is valid.
#' @param gap.char The character for gaps in alignments.
#' @return Template data  frame with annotated binding regions.
#' @keywords internal
get.leader.exon.regions.single <- function(l.seq, lex.seq, 
                     direction = c("fw", "rev"), gap.char) {
    if (length(direction) == 0) {
        stop("Please provide the 'direction' argument.")
    }
    direction <- match.arg(direction)
    # match leader sequences to lex seqs
    # try to match accurately:
    idx <- lapply(lex.seq$Header, function(x) match(x, l.seq$Header))  # hit from seqs to leaders
    any.missing <- any(unlist(lapply(idx, function(x) is.na(x))))
    if (any.missing) {
        # could not match template header to leader headers, try approximate matching
        # NB: substring here is a necessary hack for mapping IMGT leaders to seqs
        idx <- lapply(lex.seq$Header, function(x) grep(substring(x, 1, 30), l.seq$Header, fixed = TRUE))
    }
    # ensure that each template has a unique leader
    hit.len <- sapply(idx, length)
    nbr.of.hits <- sum(hit.len)
    if (any(hit.len == 0 | hit.len > 1)) {
        if (any(hit.len == 0)) {
            error.msg <- "Some templates did not have a corresponding binding region. The binding regions of templates for which no allowed region was specified were not changed."
            my.idx <- which(hit.len == 0)
            error.msg <- paste(error.msg, "The IDs of the sequences are:\n", paste(lex.seq$ID[my.idx], 
                collapse = "\n", sep = ""))
            my.warning("MissingLeaders", error.msg)
        } else {
            error.msg <- "Some templates had more than one corresponding binding region. Only the first specified binding region was considered in these cases."
            my.idx <- which(hit.len > 1)
            error.msg <- paste(error.msg, "The IDs of the sequences are:\n", paste(lex.seq$ID[my.idx], 
                collapse = "\n", sep = ""))
            my.warning("RedundantLeaders", error.msg)
        }
    } else if (nbr.of.hits != nrow(l.seq)) {
        # not all leaders were matched to a template sequence
        used.leaders <- unique(unlist(idx))
        not.used.leaders <- setdiff(seq_len(nrow(l.seq)), used.leaders)
        leader.info <- paste("Leaders with the following headers were not used: ", 
            paste(l.seq$Header[not.used.leaders], collapse = ", "), sep = "")
        my.warning("Not_all_leaders_matched", "Some allowed regions did not match any template header and were ignored.")
    }
    # select indices that are available if multiple leader regions are given for one
    # sample, take the 1st one arbtirarily
    sel.idx.zero <- which(sapply(idx, length) == 0)
    idx <- sapply(seq_along(idx), function(x) if (x %in% sel.idx.zero) {
        NA
    } else {
        idx[[x]][1]
    })
    sel.idx.lex <- which(!is.na(idx))  # ids of the templates where a leader seq was found
    idx <- unlist(idx)
    if (length(sel.idx.lex) == 0) {
        my.error("Leaders_no_matches", "No matches between templates and allowed binding regions. No binding regions were annotated")
    }
    sel.idx <- idx[sel.idx.lex]  # ids of the leader.seqs with a template
    ###### 
    template.df <- lex.seq[sel.idx.lex, ]
    idx.add <- which(!(colnames(l.seq) %in% colnames(template.df)))
    leader.col <- paste("Allowed", "_", direction, sep = "")
    template.df[, leader.col] <- l.seq$Allowed[sel.idx]  # replace the leader
    # rename the other columns that are already present and add
    cols <- setdiff(colnames(l.seq), leader.col)
    idx.replace <- which(cols %in% colnames(template.df))
    if (length(idx.replace) != 0) {
        template.df[, cols[idx.replace]] <- l.seq[sel.idx, cols[idx.replace]]
    }
    # add the new columns also ..
    if (length(idx.add) != 0) {
        template.df[, colnames(l.seq)[idx.add]] <- l.seq[sel.idx, idx.add]
    }
    
    # find leader position in sequences with exact matching if we have a forward
    # restriction, select the first match we find, otherwise select the last match we
    # find
    leader.pos <- lapply(seq_len(nrow(template.df)), function(x) gregexpr(template.df[x, leader.col], 
        template.df$InputSequence[x]))
    leader.idx <- NULL
    if (direction == "fw") {
        # always take the first match we find
        leader.idx <- rep(1, length(leader.pos))
    } else {
        # always take the last match we find
        leader.idx <- unlist(lapply(seq_along(leader.pos), function(x) length(unlist(leader.pos[[x]]))))
    }
    leader.s <- sapply(seq_along(leader.pos), function(x) as.numeric(unlist(leader.pos[[x]])[leader.idx[x]]))
    leader.e <- sapply(seq_along(leader.pos), function(x) leader.s[x] + attr(leader.pos[[x]][[1]], 
        "match.length")[leader.idx[x]] - 1)
    # verify that all leader.s and leader.e are defined
    leader.s.ok <- leader.s > 0
    leader.e.ok <- leader.e > 0
    if (any(!leader.s.ok) || any(!leader.e.ok)) {
        warn.msg <- "Some of the input allowed binding sites could not be found in the templates. The affected sequences are:\n"
        idx1 <- which(!leader.s.ok)
        idx2 <- which(!leader.e.ok)
        idx <- union(idx1, idx2)
        warn.msg <- paste(warn.msg, paste(template.df$ID[idx], collapse = "\n", sep = ""))
        my.warning("LeadersNotFound", warn.msg)
    }
    final.df <- retrieve.leader.region(template.df, direction, leader.s, leader.e, gap.char, TRUE) # retrieve leader region: overwrites values of other directions
    final.df <- cbind(final.df, "TemplateIdentifier" = template.df$Identifier)
    if (FALSE) {
        if (any(leader.s.ok)) {
        # set allowed region start/end
        allowed.col.e <- paste("Allowed_End", "_", direction, sep = "")
        allowed.col.s <- paste("Allowed_Start", "_", direction, sep = "")
        template.df[leader.s.ok, allowed.col.s] <- leader.s[leader.s.ok]
        template.df[leader.s.ok, allowed.col.e] <- leader.e[leader.s.ok]
        }
    }
    return(final.df)
}

#' @rdname Output
#' @name Output
#' @aliases write_templates
#' @return \code{write_templates} stores templates to \code{fname}.
#' @export
#' @examples
#'
#' data(Ippolito)
#' # Store templates as FASTA
#' fname.fasta <- tempfile("my_templates", fileext = ".fasta")
#' write_templates(template.df, fname.fasta)
#' # Store templates as CSV
#' fname.csv <- tempfile("my_templates", fileext = ".csv")
#' write_templates(template.df, fname.csv, "CSV")
write_templates <- function(template.df, fname, ftype = c("FASTA", "CSV")) {
    ftype <- match.arg(ftype)
    if (ftype == "FASTA") {
        seqs <- template.df$Sequence
        labels <- template.df$ID
        seqinr::write.fasta(as.list(seqs), as.list(labels), fname)
    } else if (ftype == "CSV") {
        write.csv(template.df, file = fname, row.names = FALSE)
    } else {
        warning("Unknown filetype: ", ftype)
    }
}

#' @rdname Scoring
#' @name Scoring
#' @aliases score_conservation
#' @return A list containing \code{Entropies} and \code{Alignments}.
#' \code{Entropies} is a data frame with conservation scores. 
#' Each column indicates a position in the alignment of template sequences
#' and each row gives the entropies of the sequences 
#' belonging to a specific group of template sequences.
#' \code{Alignments} is a list of \code{DNABin} objects, where each
#' object gives the alignment corresponding to one group of template sequences.
#' @export
#' @note \code{score_conservation} requires the MAFFT software 
#' for multiple alignments (http://mafft.cbrc.jp/alignment/software/).
#' @examples
#' \dontrun{
#' data(Ippolito)
#' entropy.data <- score_conservation(template.df, gap.char = "-", win.len = 18, by.group = TRUE)
#' }
score_conservation <- function(template.df, gap.char = "-", 
                               win.len = 30, by.group = TRUE) {
    
    # only align sequences if not aligned already
    if (!check.tool.function()["MAFFT"]) {
        stop("MAFFT is required for scoring the conservation (http://mafft.cbrc.jp/alignment/software/).")
    }
    if (!any(grepl(gap.char, template.df$InputSequence))) {
        # no gap char found -> sequences are not aligned yet
        message("Aligning sequences ...")
        ali <- align.seqs(template.df$InputSequence, template.df$ID)
    } else {
        message("Assuming that the sequences are pre-aligned.")
        ali <- seqinr::as.alignment(nb = length(template.df$InputSequence),
                                    nam = template.df$ID,
                                    seq = template.df$InputSequence)
    }
    if (is.null(ali)) {
        # alignment wasn't possible
        return(NULL)
    }
    # sanitize the sequences: (carriage returns cause problems in windows)
	ali$seq <- gsub("\r", "", ali$seq)
    bin <- ape::as.DNAbin(ali)
    # create groups in the alignment according to user-anntotation
    # may automatic grouping in the future (clustering?)
    if (by.group) {
        groups <- unique(template.df$Group)
        bins <- lapply(seq_along(groups), function(x) bin[which(template.df$Group == groups[x]),])
    } else {
        groups <- "default"
        bins <- list(bin)
    }
    # determine conservation according to Shannon entropy for each alignment group
    ali.entropy <- lapply(bins, function(x) shannon.entropy(as.character(x)))
    # create a data frame representation
    entropy.df <- do.call(rbind, ali.entropy)
    rownames(entropy.df) <- groups
    colnames(entropy.df) <- seq_len(ncol(entropy.df))
    names(bins) <- groups
    out <- list("Alignments" = bins, "Entropies" = entropy.df)
    return(out)
}
#' Identification of Gappy Columns in Alignments.
#' @param bins A list of \code{DNABin} alignments.
#' @param gap.cutoff The required percentage of gaps
#' for consideration as a gap column.
#' @param gap.char The gap character in the alignments.
#' @return A list with indices giving the gap columns
#' for every alignment in \code{bins}.
#' @keywords internal
detect.gap.columns <- function(bins, gap.cutoff = 0.95, gap.char = "-") {
    gap.cutoff <- 0.95 # if there is a column 95% gaps, ignore the whole region
    gap.freq <- lapply(bins, function(x) {
        mat <- as.character(x)
        count <- unlist(lapply(seq_len(ncol(mat)), function(y) 
            length(which(mat[,y] == gap.char))))
        freq <- count / nrow(mat)
    })
    # exclude regions with gap percentage above cutoff -> useless to construct primers here 
    mask.idx <- lapply(gap.freq, function(x) which(x >= gap.cutoff))
    return(mask.idx)
}
#' @rdname Plots
#' @name Plots
#' @aliases plot_conservation
#' @return \code{plot_conseration} returns a plot showing the degree of sequence conservation in the templates.
#' @export
#' @note Computing the conservation scores for using \code{plot_conservation} requires
#' the MAFFT software for multiple alignments (http://mafft.cbrc.jp/alignment/software/).
#' @examples
#'
#' data(Ippolito)
#' # Select binding regions for every group of templates and plot:
#' template.df <- select_regions_by_conservation(template.df, win.len = 30)
#' if (length(template.df) != 0) {
#'      p1 <- plot_conservation(attr(template.df, "entropies"), 
#'                              attr(template.df, "alignments"), template.df)
#' }
#' # Select binding regions for all templates and plot:
#' data(Ippolito)
#' template.df <- select_regions_by_conservation(template.df, by.group = FALSE)
#' if (length(template.df) != 0) {
#'      p2 <- plot_conservation(attr(template.df, "entropies"), 
#'                              attr(template.df, "alignments"), template.df)
#' } 
plot_conservation <- function(entropy.df, alignments, template.df, gap.char = "-") {
    # set gappy columns to 0 conservation for all groups
    gap.idx <- detect.gap.columns(alignments, gap.cutoff = 0.95, gap.char = gap.char)
    m <- match(rownames(entropy.df), names(gap.idx))
    for (i in seq_len(nrow(entropy.df))) { # for each alignment group
        entropy.df[i, gap.idx[[m[i]]]] <- 1 # max entropy
    }
    plot.df <- melt(entropy.df, value.name = "Entropy")
    # since we have [0,1] interval entropies, we can simply convert to conservation:
    colnames(plot.df) <- c("Group", "Position", "Entropy")
    plot.df$Conservation <- 1 - plot.df$Entropy
    # get region data for every group
    get.region.df <- function(allowed.start.fw.old, allowed.end.fw.old,
                              allowed.start.fw.new, allowed.end.fw.new,
                              allowed.start.rev.old, allowed.end.rev.old,
                              allowed.start.rev.new, allowed.end.rev.new) {
        # select unique interval here, but how?
        fw.ori.range <- unique(cbind(allowed.start.fw.old,
                                allowed.end.fw.old))
        fw.ori.df <- data.frame("Start" = fw.ori.range[,1],
                                "End" = fw.ori.range[,2],
                                "Direction" = "fw",
                                "Type" = "Old")
        fw.new.range <- unique(cbind(allowed.start.fw.new,
                                allowed.end.fw.new))
        fw.new.df <- data.frame("Start" = fw.new.range[,1],
                                "End" = fw.new.range[,2],
                                "Direction" = "fw",
                                "Type" = "New")
        rev.ori.range <- unique(cbind(allowed.start.rev.old,
                                allowed.end.rev.old))
        rev.ori.df <- data.frame("Start" = rev.ori.range[,1],
                                "End" = rev.ori.range[,2],
                                "Direction" = "rev",
                                "Type" = "Old")
        rev.new.range <- unique(cbind(allowed.start.rev.new,
                                allowed.end.rev.new))
        rev.new.df <- data.frame("Start" = rev.new.range[,1],
                                "End" = rev.new.range[,2],
                                "Direction" = "rev",
                                "Type" = "New")
        range.df <- rbind(fw.ori.df, fw.new.df, 
                          rev.ori.df, rev.new.df)
        return(range.df)
    }
    groups <- names(alignments)
    allowed.df <- vector("list", length(groups))
    for (i in seq_along(alignments)) {
        if (groups[i] == "default") { 
            # capture all templates
            sub.df <- template.df
        } else {
            # select individual groups
            sub.df <- template.df[template.df$Group == groups[i],]
        }
        region.df <- get.region.df(sub.df$Allowed_Start_fw_initial_ali,
                               sub.df$Allowed_End_fw_initial_ali,
                               sub.df$Allowed_Start_fw_ali,
                               sub.df$Allowed_End_fw_ali,
                               sub.df$Allowed_Start_rev_initial_ali,
                               sub.df$Allowed_End_rev_initial_ali,
                               sub.df$Allowed_Start_rev_ali,
                               sub.df$Allowed_End_rev_ali)
        region.df$Group <- groups[i]
        allowed.df[[i]] <- region.df
    }
    allowed.df <- do.call(rbind, allowed.df)
    group.colors <- brewer.pal(8, "Set1")
    group.colors <- colorRampPalette(group.colors)(length(groups))
    group.colors <- c(group.colors, "grey40")
    # only plot the new allowed regions
    allowed.new <- allowed.df[allowed.df$Type == "New",]
    allowed.old <- allowed.df[allowed.df$Type == "Old",]
    # get max bar height for any group for a position
    #max.conservation <- ddply(plot.df, c("Position"), summarize, "TotalConservation" = sum(substitute(Conservation)))
    #box.extent <- max(max.conservation$TotalConservation) * 0.1
    box.extent <- max(plot.df$Conservation) * 0.1
    box.ymax <- -0.05 * max(plot.df$Conservation)
    box.ymin <- box.ymax - box.extent
    # select only unique old regions
    allowed.old <- allowed.old[row.names(unique(allowed.old[, c("Start", "End")])), ]
    p <- ggplot() + 
        geom_bar(data = plot.df, aes(x = .data[["Position"]], 
                 y = .data[["Conservation"]],
                 fill = .data[["Conservation"]]),
                 stat = "identity",
                 position = "stack") +
        xlab("Position") +
        ylab("Conservation") +
        scale_y_continuous(labels = scales::percent) +
        facet_wrap(~Group, ncol = 2) 
    # add rectangles for binding region annotation
    p <- p + geom_rect(data = allowed.new, 
                      aes(xmin = .data[["Start"]], 
                                 xmax = .data[["End"]], 
                                 ymax = box.ymax,
                                 ymin = box.ymin),
                                 #color = "Type",
                                 alpha = 1) +
            # don't plot old regions: may not be updated to aligned binding regions
            #geom_rect(data = allowed.old,
                #aes_string(xmin = "Start", 
                                 #xmax = "End", 
                                 #fill = "Type",
                                 #ymax = box.ymax,
                                 #ymin = box.ymin),
                                 #alpha = 0.2) +
      #scale_fill_manual(values = group.colors) +
      scale_fill_gradient()
      ggtitle("Sequence conservation")
    return(p)
}
#' Updates Binding Region in the Alignment by conservation.
#' @param template.df A \code{Templates} object.
#' @param bins A list with \code{DNAbin} alignments, one for each group of template sequences.
#' @param entropy.df A data frame with entropy information.
#' @param gap.char The gap character for alignments.
#' @param win.len The desired length of the new binding region.
#' @param direction The direction for which the binding range shall be adjusted.
#' @return A \code{Templates} object with modified binding regions.
#' @keywords internal
update.binding.ranges.by.conservation <- function(template.df, 
                                    bins,
                                    entropy.df,
                                    gap.char = "-", 
                                    win.len = 30, direction = c("fw", "rev")) {
    if (length(direction) == 0) {
        stop("Please provide the 'direction' argument.")
    }
    direction <- match.arg(direction)
    if (direction == "fw") {
        allowed.region <- cbind(template.df$Allowed_Start_fw_initial_ali, 
                                template.df$Allowed_End_fw_initial_ali)
    } else {
        allowed.region <- cbind(template.df$Allowed_Start_rev_initial_ali, 
                                template.df$Allowed_End_rev_initial_ali)

    }
    step.size <- 1 # shift each window by 5 positions for speedup of computations
    primer.range <- create.primer.ranges(allowed.region[,2], rep(win.len, nrow(allowed.region)), 
                                        start.position = allowed.region[,1], step.size = step.size,
                                        groups = template.df$Group) 
    primer.ranges <- parallel::mclapply(seq_along(bins), 
        function(x) select.primer.region.by.conservation(primer.range[primer.range$Group == names(bins)[x], ], 
            entropy.df[x,], 0.1, bins[[x]], gap.char) # select top 10% within the allowed region
    )
    # annotate ranges with groups
    for (i in seq_along(primer.ranges)) {
        primer.ranges[[i]] <- cbind(Group = names(bins)[i], primer.ranges[[i]]) # there was an error here ..
    }
    ranges <- do.call(rbind, primer.ranges)
    # select range with smallest entropy for every group
    selected.ranges <- ddply(ranges, "Group", function(x) arrange(x, substitute(Entropy))[1,])
    # assign new binding regions for every group of template sequences individually
    new.templates <- template.df
    for (i in seq_along(selected.ranges$Group)) {
        group <- selected.ranges$Group[i]
        ali <- as.character(bins[[i]])
        ali.seqs <- unlist(lapply(seq_len(nrow(ali)),function(x) paste0(ali[x,], collapse = "")))
        # identify templates belonging to this group
        idx <- which(template.df$Group == group)
        sub.df <- template.df[idx,]
        # set the new aligned sequence
        sub.df$InputSequence <- ali.seqs
        region <- c(selected.ranges[i, "Start"], selected.ranges[i, "End"])
        if (direction == "fw")  {
            region.fw <- region
            region.rev <- NULL
        } else {
            region.fw <- NULL
            # convert from absolute posi to relative posi from the end of the seq
            region.rev <- rev(ncol(ali) - region)
        }
        sub.df <- assign_binding_regions(sub.df, region.fw, region.rev)
        # store old (initial binding regions)
        if (direction == "fw") {
            sub.df$Allowed_Start_fw_initial_ali <- template.df[idx, "Allowed_Start_fw_initial_ali"]
            sub.df$Allowed_End_fw_initial_ali <- template.df[idx, "Allowed_End_fw_initial_ali"]
        } else {
            sub.df$Allowed_Start_rev_initial_ali <- template.df[idx, "Allowed_Start_rev_initial_ali"]
            sub.df$Allowed_End_rev_initial_ali <- template.df[idx, "Allowed_End_rev_initial_ali"]
        }
        # overwrite template entry
        new.templates[idx,] <- sub.df
    }
    return(new.templates)
}

#' @rdname TemplatesFunctions
#' @name TemplatesFunctions
#' @aliases select_regions_by_conservation
#' @return \code{select_regions_by_conservation} returns a
#' \code{Templates} object with adjusted binding regions.
#' The attribute \code{entropies} gives a data frame with positional entropies
#' and the attribute \code{alignments} gives the alignments of the templates.
#' @export
#' @note \code{select_regions_by_conservation} requires the MAFFT software
#' for multiple alignments 
#' (http://mafft.cbrc.jp/alignment/software/).
#' @examples
#' data(Ippolito)
#' new.template.df <- select_regions_by_conservation(template.df)
select_regions_by_conservation <- function(template.df, 
                                    gap.char = "-", 
                                    win.len = 30, by.group = TRUE,
                                    mode.directionality = c("both", "fw", "rev")) {
    if (!is(template.df, "Templates")) {
        stop("Please input an object of class 'Templates'.")
    }
    if (!check.tool.function()["MAFFT"]) {
        warning("MAFFT is required for selecting conserved regions (http://mafft.cbrc.jp/alignment/software/).")
        return(NULL)
    }
    direction <- match.arg(mode.directionality)
    # note: time could be saved by not computing the entropy of the full sequence, but only the entropy of the possible binding regions?
    # note: individual binding regions are not really respected as selection is via groups and not individual sequences.
    entropy.data <- score_conservation(template.df, gap.char, win.len, by.group)
    if (!by.group) {
        # ignore group annotation for updating binding regions
        template.df$Group <- "default"
    }
    bins <- entropy.data$Alignments # list with DNABin objects
    entropy.df <- entropy.data$Entropies # list with data frames giving the entropies
    if (direction == "fw" || direction == "rev") {
        template.df <- update.binding.ranges.by.conservation(template.df, bins, entropy.df, gap.char, win.len, direction)
    } else if (direction == "both") {
        t.fw <- update.binding.ranges.by.conservation(template.df, bins, entropy.df, gap.char, win.len, "fw")
        template.df <- update.binding.ranges.by.conservation(t.fw, bins, entropy.df, gap.char, win.len, "rev")
    }
    # add additional data (for plotting) as attributes
    attr(template.df, "entropies") <- entropy.df
    attr(template.df, "alignments") <- bins
    return(template.df)
}

context("Templates") 

test_that("Template CSV Input", {
    # tests whether csv input/output of templates works
    data(Ippolito)
    # test whether correct template files are loaded
    template.files <- system.file("extdata", "IMGT_data", "comparison", "templates", 
                              "IGH_templates.csv", package = "openPrimeR")
    template.data <- read_templates(template.files)
    # test whether we can store personal templates
    template.csv <- tempfile("my_templates", fileext = ".csv")
    write.csv(template.df, file = template.csv, row.names = FALSE)
    # reset rownames for test, would be different due to subsetting otherwise
    rownames(template.df) <- NULL
    template.data <- read_templates(template.csv)
    expect_that(template.data, equals(template.df))
    # removed some tests checking for errors when reading since these make problems with testthat environment ..
})
 

#' Creation of Fulfilled/Failed Constraint Counts.
#'
#' Creates counts of fullfilled/failed constraints.
#'
#' @param primer.df An evaluated \code{Primers} object.
#' @param eval.cols Evaluation columns in \code{primer.df} to consider.
#' By default (\code{NULL}) all evaluation columns are considered.
#' @return A data frame with the number of fulfilled/failed constraints
#' for \code{primer.df}.
#' @keywords internal
create_fulfilled_counts <- function(primer.df, eval.cols = NULL) {
    # compute a data frame of passed/failed constraints
    primer.df <- asS3(primer.df)
    if (length(eval.cols) != 0) {
        m <- match(eval.cols, colnames(primer.df))
        na.idx <- which(is.na(m))
        if (length(na.idx) != 0) {
            msg <- paste("EVAL columns unknown: ", 
                paste(eval.cols, collapse = ","), sep = "")
            warning(msg) 
            m <- m[-na.idx]
        }
        primer.df <- cbind("Run" = primer.df$Run, primer.df[,m, drop = FALSE])
    }
    constraint.idx <- grep("^EVAL_", colnames(primer.df))
    if (length(constraint.idx) == 0) {
        return(NULL)
    }
    constraint.names <- colnames(primer.df)[constraint.idx]
    cols <- c("Run", constraint.names)
    count.df <- ddply(primer.df[,cols], "Run", 
                    catcolwise(sum))
    # enrich with failures:
    failure.df <- ddply(primer.df[,cols], "Run", 
                    catcolwise(function(x) length(x) - sum(x)))
    failure.names <- paste(constraint.names,"_failure", sep = "")
    colnames(failure.df)[colnames(failure.df) %in% constraint.names] <- 
        failure.names
    count.df <- cbind(count.df, failure.df)
    return(count.df)
}

#' @rdname PrimerEval
#' @name PrimerEval
#' @aliases primer_significance
#' @details
#' \code{primer_significance} computes the significance by comparing
#' the total count of fulfilled and failed constraints
#' with the corresponding counts of primer sets from the literature.
#' Significant p-values indicate primer sets whose rate of constraint 
#' fulfillment is higher compared to the reference sets.
#'
#' @return \code{primer_significance} returns a numeric providing
#' the p-value of the primer set according to Fisher's exact test.
#' The returned value has the following attributes: 
#' \describe{
#' \item{\code{test}}{The results of the significance test}
#' \item{\code{tab}}{The confusion matrix for Fisher's exact test}
#' \item{\code{constraints}}{The names of the considered constraints}
#' }
#' @export
#' @examples
#' 
#' # Determine the significance of a primer set
#' data(Ippolito)
#' p.data <- primer_significance(primer.df, "Ippolito")
#' attr(p.data,"tab") # the confusion matrix
#' attr(p.data, "test") # results from Fisher's test
#' attr(p.data, "constraints") # considered constraints for the test
primer_significance <- function(primer.df, set.name = NULL, active.constraints = NULL) {
    if (length(primer.df) == 0 || nrow(primer.df) == 0) {
        stop("Cannot compute significance for empty sets.")
    }
    if (is.null(set.name)) {
        set.name <- primer.df$Run[1]
    } 
    if (length(set.name) > 1) {
        stop("Set name should have length 1.")
    }
    # evaluate counts of primer.df
    eval.cols <- NULL
    if (length(active.constraints) != 0) {
        eval.cols <- paste0("EVAL_", active.constraints)
        # select only available constraints
        eval.cols <- eval.cols[eval.cols %in% colnames(primer.df)]
    }
    my.data <- create_fulfilled_counts(primer.df, eval.cols)
    if (is.null(my.data)) { # no evaluated constraints
        warning("Cannot compute a p-value since no constraints have been previously evaluated for the input primers.")
        return(NULL)
    }
    # always exclude specificity and coverage as these are problem-specific
    excl.cons <- c("EVAL_primer_specificity", "EVAL_primer_coverage")
    excl.cons <- c(excl.cons, paste(excl.cons, "_failure", sep = ""))
    constraint.names <- colnames(my.data)[grep("^EVAL_", colnames(my.data))]
    constraint.names <- setdiff(constraint.names, excl.cons)
    if (length(constraint.names) == 0) {
        warning("No computed constraint of the provided primer set is suitable for determining set significance.")
        # no constraints to be tested
        return(NULL)
    }
    ok.names <- constraint.names[!grepl("_failure", constraint.names)]
    failure.names <- constraint.names[grepl("_failure", constraint.names)]
    # now, summarize across all constraints for a fisher's exact test
    #   -> Passed constraint count | Failed constraint count
    my.ok <- unlist(apply(my.data[, ok.names, drop = FALSE], 1, function(x) sum(x, na.rm = TRUE)))
    my.fail <- unlist(apply(my.data[, failure.names, drop = FALSE], 1, function(x) sum(x, na.rm = TRUE)))
    # reference sysdata is available within the package automatically: REF.pass.counts, REF.fail.counts
    m.ok <- match(constraint.names, names(REF.pass.counts))
    m.fail <- match(constraint.names, names(REF.fail.counts))
    ref.ok <- sum(unlist(REF.pass.counts[, m.ok[!is.na(m.ok)]]))
    ref.fail <- sum(unlist(REF.fail.counts[, m.fail[!is.na(m.fail)]]))
    # test primers:
    tab <- matrix(c(my.ok, my.fail, ref.ok, ref.fail), nrow = 2, 
                  ncol = 2,byrow = TRUE)
    rownames(tab) <- c(set.name, "Reference")
    colnames(tab) <- c("Fulfilled", "Failed")
    # alternative: greater -> check only if odds ratio increases using the tested primers
    # test primers should fulfill more constraints and have less failures
    test.result <- fisher.test(tab, alternative = "greater")
    out.constraint.names <- constraint.names[!grepl("_failure", constraint.names)]
    out.constraint.names <- gsub("EVAL_", "", out.constraint.names)
    result <- test.result$p.value
    attr(result, "test") <- test.result
    attr(result, "table") <- tab
    attr(result, "constraints") <- out.constraint.names
    return(result)
}

 

context("Primers") 

test_that("Primer CSV Input", {
    # tests whether csv input/output of primers works
    data(Ippolito)
    primer.csv <- tempfile("my_primers", fileext = ".xml")
    rownames(primer.df) <- NULL
    write.csv(primer.df, file = primer.csv, row.names = FALSE)
    csv.primers <- read_primers(primer.csv)
    expect_that(csv.primers, equals(primer.df))
    # check for error behavior on loading incorrect primers
    test.df <- data.frame(A = 1:3, B = 6:8)
    write.csv(test.df, file = primer.csv, row.names = FALSE)
    expect_error(read_primers(primer.csv))
})

test_that("Primer Set Significance", {
    # check that significance computations provide reasonable results
    data(Ippolito)
    primer.location <- system.file("extdata", "IMGT_data", "primers", "IGHV", 
                        "Ippolito2012.fasta", package = "openPrimeR")
    primers <- read_primers(primer.location, "_fw", "_rev")
    expect_warning(primer_significance(primers)) # nothing evaluated -> warning
    expect_warning(primer_significance(primer.df, active.constraints = "primer_coverage"))
    p <- primer_significance(primer.df) # should not be significant
    expect_gt(as.numeric(p), 5e-2)
})

test_that("Primer fw+rev", {
    # check that primer sets with pairs of primers are analyzed correctly
    data(Ippolito)
    primer.location <- system.file("extdata", "IMGT_data", "primers", "Test", 
                        "test_both.fasta", package = "openPrimeR")
    primers <- read_primers(primer.location, "_fw", "_rev")
    constraint.df <- check_constraints(primers[1:2,], template.df, settings)
    temp.file <- tempfile(fileext = ".pdf")
    # check whether a report can be generated
    create_report(constraint.df, template.df, temp.file, settings)
})

# Store Ippolito's primers for IGH, the templates, and the settings.
fasta.file <- system.file("extdata", "IMGT_data", "templates", 
            "Homo_sapiens_IGH_functional_exon.fasta", 
            package = "openPrimeR")
hdr.structure <- c("ACCESSION", "GROUP", "SPECIES", "FUNCTION")
# load templates
template.df <- openPrimeR::read_templates(fasta.file, hdr.structure, "|", "GROUP", rm.keywords = "partial")
# assign leader:
leader.file <- system.file("extdata", "IMGT_data", "templates",
            "Homo_sapiens_IGH_functional_leader.fasta", package = "openPrimeR")
template.df <- assign_binding_regions(template.df, leader.file, NULL)
primer.location <- system.file("extdata", "IMGT_data", "primers", "IGHV", 
                        "Ippolito2012.fasta", package = "openPrimeR")
primer.df <- openPrimeR::read_primers(primer.location, "_fw", "_rev")
# load settings for analysis
filename <- system.file("extdata", "settings", 
            "C_Taq_PCR_high_stringency.xml", package = "openPrimeR")
settings <- openPrimeR::read_settings(filename)
# evaluate constraints
# allow also other binding events:
conOptions(settings)$allowed_other_binding_ratio <- 1.0
conOptions(settings)$allowed_mismatches <- 5
primer.df <- openPrimeR::check_constraints(primer.df, template.df, settings)
template.df <- openPrimeR::update_template_cvg(template.df, primer.df)
out.loc <- system.file("data", "Ippolito.rda", package = "openPrimeR")
# compression level: 'xz' ensures small size
save(primer.df, template.df, settings, file = out.loc, compress = "xz")

#' The openPrimeR Tutorial. 
#'
#' Starts a Shiny app containing the openPrimeR tutorial, which
#' was built using the learnr package. The application
#' starts locally and should open a new tab in your default
#' browser. If no browser is opened, please consider the console
#' output to identify the local port on which the server is running.
#'
#' @param dev A logical indicating whether to start the development version of the tutorial (default: \code{FALSE}).
#' @note
#' The Shiny app can be started only if you fulfill all 
#' of the suggested package dependencies for the Shiny framework,
#' so please ensure that you've installed openPrimeR including
#' all suggested dependencies.
#'
#' @export
#' @return Opens the openPrimeR tutorial in a web browser.
#' @examples
#' \dontrun{
#' # Open the tutorial
#' if (interactive()) {
#' runTutorial()
#' }
#' }
runTutorial <- function(dev = FALSE) {
    # n.b.: need to manually call to create html files to be included in pkg
    if (dev) {
        # development
        rmarkdown::run(system.file("inst", "tutorials", "introduction", "introduction.Rmd", package = "openPrimeR"))
    } else {
        # release
        learnr::run_tutorial("introduction", package = "openPrimeR")
    }
}

#' Data Sets.
#'
#' @rdname Data
#' @name Data
#' @docType data
#' @keywords datasets
#' @description
#' \describe{
#' \item{\code{Ippolito}}{IGHV primer data from Ippolito et al.}
#' \item{\code{Tiller}}{IGHV primer data from Tiller et al.}
#' \item{\code{Comparison}}{Evaluated primer sets targeting the 
#' functional human IGH immunoglobulin
#' genes. The sets were generated using the default evaluation settings
#' of openPrimeR. The primer sets were gathered from
#' IMGT and the literature.}
#' \item{\code{RefCoverage}}{Experimental results of multiplex PCR.}
#' }
#'  
#' @format For the \code{RefCoverage} data set, the \code{feature.matrix} data frame contains the properties
#' of the primer set from Tiller et al. as well as a primer set
#' that was designed by openPrimeR. The column \code{Experimental_Coverage}
#' indicates the experimentally determined coverage, while the other
#' columns relate to properties of the primers that were computed
#' with openPrimeR. 
#' The \code{ref.data} list contains the raw experimental coverage
#' of individual primers from the primer sets from Tiller and openPrimeR, which
#' both target templates from the IGH locus.
#' The rows of the data frames indicate primers and the columns indicate
#' IGH templates for which experimental coverage was determined.
#' The cell entries are hex codes. Each hex code represents a color 
#' indicating a certain experimental coverage status. 
#' Hex codes representing red shades indicate no or little amplification, 
#' while hex codes for green shades indicate high yields.
#'
#' For the \code{Ippolito} data set, 
#' \code{primer.df} provides a \code{Primers} object containing
#' the evaluated set of primers from Tiller et al.
#' \code{template.df} provides
#' a \code{Templates} object containing functional, human IGHV
#' templates for, and \code{settings} provides a 
#' \code{DesignSettings} object providing the used analysis settings.
#'
#' For the \code{Comparison} data set, \code{primer.data} and \code{template.data} are
#' lists of \code{Primers} and \code{Templates} objects, respectively.
#' 
#' For the \code{Tiller} data set, \code{tiller.primer.df} 
#' provides a \code{Primers} object, \code{tiller.template.df} provides 
#' the corresponding \code{Templates} object, and \code{tiller.settings} 
#' provides the \code{DesignSettings} object that was used
#' for evaluating \code{tiller.primer.df}.
NULL

################
# Generate procedures for creating internal Rdata here.
# o Computes the reference distribution for Fisher's exact test on the number of
# fulfilled constraints
#################


#' Creation of Reference Data for Significance Tests
#'
#' Creates reference data for significance test on the ratio of
#' fulfilled constraints.
#'
#' @param set.name Primer set to exclude from the data.
#' @param top.k Select only the top k coverage sets for every locus. 
#' If \code{top.k} is \code{NULL}, all primer sets 
#' are selected to form the reference distribution.
#' @param settings Settings for evaluating constraints that are not part of the original data.
#' @keywords internal
sig_eval_ref_data <- function(set.name = "openPrimeR2017", top.k = NULL, settings) {
    # load comparison data as reference frequencies
    # other options: kolmogorov smirnov (compare 2 distributions)
    # F-test to compare two variances (mine should be more narrow -> close to the target)
    comp.loc <- system.file("extdata", "IMGT_data", "comparison", "primer_sets",
                    package = "openPrimeR")
    comp.folders <- list.dirs(comp.loc, full.names = TRUE, recursive = FALSE)
    comp.loc.t <- list.files(
                    system.file("extdata", "IMGT_data", "comparison", 
                                "templates",package = "openPrimeR"),
                    full.names = TRUE)
    count.data <- NULL
    template.data <- read_templates(comp.loc.t)
    for (i in seq_along(comp.folders)) { # for each Ig group (IGHV etc)
        primer.set.location <- comp.folders[i]
        primer.files <- list.files(path = primer.set.location, pattern = ".csv", full.names = TRUE)
        primer.data <- read_primers(primer.files)
        # add melting_temp_diff column here
        # TODO: use settings here?
        primer.data <- lapply(primer.data, function(x) cbind(x, "EVAL_melting_temp_diff" = x$melting_temp_diff < constraints(settings)$melting_temp_diff))
        template.df <- template.data[[i]]
        locus <- template.df$Run[1]
        cvg <- sapply(primer.data, function(x) get_cvg_ratio(x, template.df))
        # select reference set for this locus:
        if (!is.null(top.k)) {
            # selet top-k coverage sets
            o <- order(cvg, decreasing = TRUE)
            ref.sets <- primer.data[o[1:top.k]]
            if (set.name %in% names(ref.sets)) {
                # need another reference to have k sets (will remove `set.name` later
                ref.sets <- primer.data[o[1:(top.k+1)]]
            } 
        } else {
            # select all cvg sets
            ref.sets <- primer.data
        }

        # count number of fulfilled constraints of all other sets
        # construct count data
        ref.df <- asS3(do.call(my_rbind, ref.sets))
        count.df <- create_fulfilled_counts(ref.df)
        count.df <- cbind(Locus = locus, count.df)
        count.data <- my_rbind(count.data, count.df)
    }       
    # exclude my set from ref.sets
    ref.data <- count.data[count.data$Run != set.name,]
    # summarize across all primer sets
    ref.data <- plyr::numcolwise(function(x) sum(x, na.rm = TRUE))(ref.data)
    return(ref.data)
}
generate.country.dists <- function() {
    # compute distances between countries for determining the fastest mirror for custom install scripts of the package
    library(cshapes)
    library(countrycode)
    dmat <- distmatrix(as.Date("2002-1-1"), type = "capdist", useGW = FALSE)  # correlates of war
    my.names <- tolower(countrycode(rownames(dmat), "cown", "iso2c"))  # turn countries to iso with 2 character encoding as used by R
    country.dists <- dmat
    colnames(country.dists) <- my.names
    rownames(country.dists) <- my.names
    return(country.dists)
}
#' Explore the properties of different coverage distribtuions.
#' @param template.df List with templates.
#' @param primer.data List with primers.
#' @param k Length of primers.
#' @return Does not return anything but performs exploration.
#' @keywords internal
explore.dists <- function(template.data, k = 18) {
    # library(logspline)
    for (i in seq_along(template.data)) {
        template.df <- template.data[[i]]
        sample <- names(template.data)[i]
        print(sample)
        cvg.df <- estimate.cvg(template.df, k = 18, "both", sample = sample)
        x <- c(cvg.df$fw$Coverage_Ratio, cvg.df$rev$Coverage_Ratio)
        png(paste0(sample, "_fit_discovery.png"))
        descdist(x, discrete = FALSE)
        dev.off()
        fit.beta <- fitdist(x, "beta") # beta has high error
        print(fit.beta)
        fit.gamma <- fitdist(x, "gamma")
        print(fit.gamma)
        png(paste0(sample, "_gamma_fit.png"))
        plot(fit.gamma)
        dev.off()
        png(paste0(sample, "_beta_fit.png"))
        plot(fit.beta)
        dev.off()
    }
}
plot.ref.dists <- function() {
    library(distrEx)
    ref.dist <- list("plain(Easily~Feasible):~beta(1,10)" = Beta(1, 10),
                     "plain(Feasible):~beta(0.8,20)" = Beta(0.8, 20), 
                    "plain(Hardly~Feasible):~beta(0.6,40)" = Beta(0.60, 40), 
                    "plain(Unfeasible):~beta(0.3,200)" = Beta(0.3, 200))
    results <- vector("list", length(ref.dist))
    distribution_rep <- list()
    for (i in seq_along(ref.dist)) {
        dist <- ref.dist[[i]]
        p1 <- dist@param@shape1
        p2 <- dist@param@shape2
        s <- rbeta(10000, shape1 = p1, shape2 = p2)
        res <- data.frame("Distribution" = factor(names(ref.dist)[i]),
                          "x" = s)
        results[[i]] <- res
        distribution_rep[[names(ref.dist)[i]]] <- expression(beta(p1, p2))
        
    }
    df <- do.call(rbind, results)
    library(ggplot2)
    ggplot(df) + geom_histogram(aes(x = x)) + facet_wrap(~Distribution,
        labeller = label_parsed) + 
      #Distribution = distribution_rep)) + 
        labs(x = "Coverage Ratio", y = "Number of Samples") +
        scale_x_continuous(labels=scales::percent)
    ggsave("beta_distribution.pdf")

}
#' Creation of Reference Coverage Ratio Distributions.
#'
#' Creates reference coverage ratio distributions in order to classify
#' primer design problems into three categories: easy, medium, and hard.
#'
#' @return A list with three reference beta distributions of coverage ratios.
#' @keywords internal
create.ref.dists <- function() {
    #library(fitdistrplus)
    library(distrEx)
    # reference dists were determined empirically
    # using the commented code from explore.dists() to determine good fits
    ###################
    #data(Ippolito)
    #template.data <- list("HCV" = read_templates("data/HCV/target_seqs.fasta"), 
                          #"IGH" = template.df)
    #explore.dists(template.data, k = 18)
    #x.easy <- hist(rbeta(10000, shape1 = 0.9, shape2 = 15))
    #x.medium <- hist(rbeta(10000, shape1 = 0.60, shape2 = 40))
    # we made the hard distribution a bit like a normal distribution
    # since in this case the parameter estimation may not give the 
    # actual distribution of the data. this is only the case for hard problems
    # where the space of possible coverage ratios is tight.
    #x.hard.before <- hist(rbeta(10000, shape1 = 0.3, shape2 = 200))
    #x.hard <- hist(rbeta(10000, shape1 = 2, shape2 = 1000))
    #x.HCV <- rbeta(10000, shape1 = 9.96, shape2 = 1606)
    #x.IGH <- rbeta(10000, shape1 = 0.798, shape2 = 23.825)
    #hist(x.easy)
    #hist(x.medium)
    #hist(x.hard)
    #hist(x.HCV)
    #hist(x.IGH)
    #KL.matrix <- cbind(Easy = ref.dist$easy, Medium = ref.dist$medium, Hard = ref.dist$hard, TestVeryEasy = Beta(0.8, 10), TestEasy = Beta(0.7, 30), TestMedium = Beta(1, 100), TestHard = Beta(0.3, 1000))
    #combis <- expand.grid(seq_len(ncol(KL.matrix)), seq_len(ncol(KL.matrix)))
    #dist <- matrix(rep(NA, nrow(combis)), nrow = ncol(KL.matrix))
    #colnames(dist) <- colnames(KL.matrix)
    #rownames(dist) <- colnames(KL.matrix)
    #for (i in 1:nrow(combis)) {
        #x <- combis[i,1]
        #y <- combis[i,2]
        #dist[x,y] <-  TotalVarDist(KL.matrix[1,x][[1]], unlist(KL.matrix[1,y][[1]])) 
    #}
    # store three reference distributions
    ref.dist <- list("very_easy" = Beta(1, 10),
                     "easy" = Beta(0.8, 20), 
                    "medium" = Beta(0.60, 40), 
                    "hard" = Beta(0.3, 200),
                    "very_hard" = Beta(10, 1000))
    return(ref.dist)
}
get_model_formula <- function() {
    # formula used for creating coverage model
    return(Experimental_Coverage ~ annealing_DeltaG + Position_3terminusLocal + annealing_DeltaG * Position_3terminusLocal)
}
get.train.idx <- function(feature.matrix) {
    # sample from training set to keep the data more balanced
    deltaG.cuts <- -c(0, 5, 9, 13, Inf) 
    strat <- cut(feature.matrix$annealing_DeltaG, deltaG.cuts)
    strat.types <- unique(strat)
    strat.counts <- sapply(strat.types, function(x) length(which(x == strat)))
    set.seed(1)
    n.samples <- min(strat.counts)
    # controlling for equal distribution of "Run": from 300 (equal distribution of DeltaG) to 168 samples
    train.idx <- vector("list", length(strat.types))
    set.seed(1)
    for (i in seq_along(strat.types)) {
        s <- strat.types[i]
        idx <- which(strat == s)
        cur.runs <- feature.matrix$Run[idx]
        # select from each run randomly:
        run.dist <- table(cur.runs)
        run.sample.size <- min(run.dist, n.samples/2)
        run.ids <- names(run.dist)
        sel <- unlist(lapply(run.ids, function(x) sample(idx[which(feature.matrix$Run[idx] == x)], run.sample.size)))
        #sel <- sapply(sample(idx, n.samples)
        train.idx[[i]] <- sel
    }
    train.idx <- unlist(train.idx)
    return(train.idx)
}
create.cvg.model <- function(feature.matrix) {
    # create model for predicting coverage
    #train.idx <- get.train.idx(feature.matrix) # remove bias from training data
    cvg.model <- glm(get_model_formula(), family = "binomial", data = feature.matrix) #[train.idx,])
    return(cvg.model)
}
create.FPR.table <- function(feature.matrix) {
    # perform 10 repeats of 5-fold cross-validation using logistic regression models
    library(caret)
    library(e1071)
    #train.idx <- get.train.idx(feature.matrix)
    ctrl <- trainControl(method = "repeatedcv", number = 5, repeats = 10, savePredictions = TRUE, classProbs = TRUE)
    model <- train(get_model_formula(), data = feature.matrix, method = "glm", 
                family="binomial", trControl = ctrl, tuneLength = 5) #, subset = train.idx)
    # retrieve all fold-predictions (coverage probabilities) across all repetitions
    cutoffs <- seq(0, 1, 0.0001)
    n.pos <- length(which(model$pred$obs == "Amplified"))
    n.neg <- length(which(model$pred$obs == "Unamplified"))
    FPR.data <- lapply(cutoffs, function(x) {
        #print(x)
        cut.pred <- factor(ifelse(model$pred$Amplified > x, "Amplified", "Unamplified"), levels = c("Unamplified", "Amplified"))
        tpr <- length(which(model$pred$obs == "Amplified" & cut.pred == "Amplified")) / n.pos
        fpr <- length(which(model$pred$obs == "Unamplified" & cut.pred == "Amplified")) / n.neg
        data.frame(Cutoff = x, FPR = fpr, TPR = tpr)
    })
    FPR.df <- do.call(rbind, FPR.data)
    # verify:
    print(FPR.df[500,])
    return(FPR.df)
}
#########
# generate p-value ref data
#######
# use default settings for evaluating melting temp diff:
filename <- system.file("extdata", "settings", 
                  "C_Taq_PCR_high_stringency.xml", package = "openPrimeR")
settings <- read_settings(filename)
ref.data <- sig_eval_ref_data(settings = settings)
REF.pass.counts <- ref.data[, !grepl("failure", colnames(ref.data))]
REF.fail.counts <- ref.data[, grepl("failure", colnames(ref.data))]
colnames(REF.fail.counts) <- gsub("_failure", "", colnames(REF.fail.counts))
###
# generate common restriction site Rdata
####
# seqRFLP data from REBASE:
# pkg: seqRFLP (>= 1.0.1)
utils::data("enzdata", package = "seqRFLP")
# common restriction enzymes retrieved from:
# https://www.addgene.org/mol-bio-reference/restriction-enzymes/
# annotate common sites in enzdata
common.file <- system.file("data-raw", "common_restriction_sites.txt", package = "openPrimeR")
common.sites <- as.character(read.table(common.file)[,1])
m <- match(common.sites, enzdata$nam)
enzdata$common <- FALSE
enzdata$common[m[!is.na(m)]] <- TRUE
# re-format search strings
# apostrophe -> indicates the sense cutting position, 
    # e.g. G_ACGT'C indicates a cut into ...GACGT and C...
# underscore -> indicates the antisense cutting position, 
    # e.g. G_ACGT'C indicates a cut into ...G and ACGTC....
enzdata$rep <- toupper(
            gsub("_", "", 
                gsub("'", "", enzdata$site, fixed=TRUE), 
                fixed = TRUE))
####
# design problem classification -> coverage distribution references
####
cvg.ref.dists <- create.ref.dists()
#######
# build supervised model for predicting primer coverage:
#######
data(RefCoverage) # load ref.data and feature.matrix
CVG_MODEL <- create.cvg.model(feature.matrix)
FPR_TABLE <- create.FPR.table(feature.matrix)
########
# create country dists for install helper function for the shiny app
########
###############
# not part of sysdata:
#country.dists <- generate.country.dists() # selection of CRAN mirrors using closest distance to countries
# store country dists to extdata as shiny app isn't exactly in the package (cannot access sysdata)
#save(country.dists, file = file.path(system.file("extdata", package = "openPrimeR"), "country_dists.Rdata"))
###################################
############
# store sysdata
##########
devtools::use_data(REF.pass.counts,  # p-value: pass counts for constraint fulfillment
                   REF.fail.counts,  # p-value: fail counts for constraint fulfillment
                   cvg.ref.dists, # reference coverage distributions (problem difficulty estimation)
                   enzdata, # enzyme restriction sites
                   CVG_MODEL, # logistic model for predicting coveragee
                   FPR_TABLE, # conversion table from coverage-probabililties to FPR
                   internal = TRUE, 
                   overwrite = TRUE) #, pkg = "openPrimeR")

#' Scoring Functions.
#'
#' @rdname Scoring
#' @name Scoring
#' 
#' @description
#' \describe{
#' \item{\code{score_degen}}{Determines the degeneration score of a sequence.}
#' \item{\code{score_conservation}}{Determines the sequence conservation
#' scores of a set of templates using Shannon entropy.}
#' \item{\code{score_primers}}{Computes scores for a set of primers
#' based on the deviations of the primers from the constraints.}
#' }
#' 
#' @param seq A list of vectors containing individual characters of a nucleotide sequence.
#' @param gap.char The gap character in the sequences.
#' The default is "-".
#' @param template.df A \code{Templates} object providing the set of templates.
#' @param win.len The size of a window for evaluating conservation.
#' The default window size is set to 30.
#' @param by.group Whether the determination of binding regions 
#' should be stratified according to the groups defined in \code{template.df}.
#' The default is \code{TRUE}.
#' @param primer.df A \code{Primers} object containing the primers.
#' @param settings A \code{DesignSettings} object containing the
#' analysis settings.
#' @param active.constraints A character vector of constraint identifiers
#' that are considered for scoring the primers.
#' @param alpha A numeric that is used to determine the trade-off
#' between the impact of the maximal observed deviation and the total
#' deviation. At its default \code{alpha} is set to 0.5 such that
#' the maximal deviation and the total deviation have an equal weight
#' when computing the penalties.
NULL

###########
# Number of runs
###########

#' Number of Runs
#'
#' Computes the longest run of a single character in the input sequence.
#' 
#' @param x Primer character sequences.
#'
#' @return The longest repeat of a single character in \code{x}.
#' @keywords internal
nbr.of.runs <- function(x) {
    if (length(x) == 0 || !is(x, "character")) {
        return(NULL)
    }
    # disambiguate primers
    x.iupac <- convert.from.iupac(x)
    result <- rep(0, length(x))
    idx <- which(x != "")
    s <- lapply(x.iupac[idx], function(y) strsplit(y, split = ""))
    i <- NULL
    runs <- foreach(i = seq_along(idx), .combine = c) %dopar% {
        r <- max(sapply(s[[i]], function(y) max(rle(y)$lengths)))
    }
    if (length(idx) != 0) {
        result[idx] <- runs
    }
    return(result)
}

#' @rdname Data
#' @name Data
#' @usage data(Comparison)
#'
#' @aliases primer.data template.data
#'
#' @references IMGT®, the international ImMunoGeneTics information system® http://www.imgt.org (founder and director: Marie-Paule Lefranc, Montpellier, France).
#' @examples
#'
#' # Load the comparison data
#' data(Comparison)
#' # Explore the first entry of the primer and template data:
#' primer.data[[1]]
#' template.data[[1]]
#' # Summarize the primer properties:
#' get_comparison_table(template.data, primer.data)
NULL

#' Coverage Matrix
#'
#' Constructs a coverage matrix where rows indicate templates and columns indicate primers.
#'
#' Entry (i,j) in the matrix is equal to 1 if primer j covers template i and otherwise 0.
#'
#' @param primer.df Primer data frame.
#' @param template.df Template data frame.
#' @param constraints A character vector of coverage constraints
#' to be used as entries for the coverage matrix instead of the 0/1 encoding.
#' At its default setting (\code{NULL}), the 0/1 encoding is used.
#' @return The binary coverage matrix.
#' @keywords internal
get.coverage.matrix <- function(primer.df, template.df, constraints = NULL) {
    # create an mxn matrix where templates are rows and primers are columns a_ij is 1
    # if template i is covered by primer j, 0 else
    if (nrow(primer.df) == 0) {
        return(NULL)
    }
    template.coverage <- get.covered.templates(primer.df, template.df)
    if (any(!constraints  %in% c("primer_efficiency", "annealing_DeltaG"))) {
        stop("Incorrect coverage constraint: ", constraints)
    }
    if (length(constraints) > 1) {
        warning("Only one constraint possible at a time. Selecting the first one.")
        constraints <- constraints[1]
        print(constraints)
    }
    # transform to matrix
    if (length(constraints) == 0) {
        A <- matrix(rep(0, nrow(template.df) * nrow(primer.df)), nrow = nrow(template.df), ncol = nrow(primer.df))
    } else {
        # insert NA to identify templates that weren't covered
        A <- matrix(rep(NA, nrow(template.df) * nrow(primer.df)), nrow = nrow(template.df), ncol = nrow(primer.df))
        cvd <- covered.seqs.to.idx(primer.df$Covered_Seqs, template.df)
        con.vals <- lapply(strsplit(primer.df[, constraints], split = ","), function(vals) as.numeric(vals))
    }
    for (i in seq_along(template.coverage)) {
        # for every template
        primer.idx <- template.coverage[[i]]
        if (length(constraints) == 0) {
            A[i, primer.idx] <- 1
        } else {
            cur.cvd <- cvd[primer.idx]
            cur.idx <- sapply(cur.cvd, function(x) which(i == x))
            if (length(cur.idx) > 0) {
                # retrieve value @ cur.idx
                cur.values <- con.vals[primer.idx]
                sel.values <- sapply(seq_along(cur.values), function(x) round(as.numeric(cur.values[[x]][cur.idx[x]]), 3))
                A[i, primer.idx] <- sel.values
            }
        }
    }
    rownames(A) <- template.df$ID
    colnames(A) <- primer.df$ID
    return(A)
}
#' Addition of Dimerization Constraints
#'
#' Updates ILP formulation with dimerization constraints.
#'
#' @param lprec An ILP instance.
#' @param D.idx Data frame giving the indices of dimerizing primer pairs.
#' @param indices Row indices for setting the dimerization constraints in \code{lprec}.
#' @return \code{lprec} with added dimerization constraints.
#' @keywords internal
add.dimerization.constraints <- function(lprec, D.idx, indices) {
    # D.idx: pairs of primers that dimerize indices: rows in which to insert the
    # dimerization constraints in the lprec
    if (nrow(D.idx) > 0) {
        #pb <- txtProgressBar(min = 0, max = nrow(D.idx), style = 3)
        for (j in seq_len(nrow(D.idx))) {
            # pairs of dimerizing primers
            row <- D.idx[j, 1]
            col <- D.idx[j, 2]
            set.row(lprec, indices[j], rep(1, 2), indices = c(row, col))  # don't include pairs of dimerizing primers
            #setTxtProgressBar(pb, j)
        }
    }
    if (length(indices) != 0) {
        set.constr.type(lprec, rep("<=", length(indices)), indices)
        set.rhs(lprec, rep(1, length(indices)), indices)
        # don't set rownames -> takes quite long rownames <- paste('Dimer', indices, sep
        # = '') rownames(lprec)[indices] <- rownames
    }
    return(lprec)
}
#' Addition of Coverage Constraints.
#'
#' Adds coverage constraints to ILP instance.
#'
#' @param lprec An ILP instance.
#' @param covered.templates Indices of covered template sequences.
#' @param template.coverage List containing the indices of covering primers for each template.
#' @return \code{lprec} with coverage constraints.
#' @keywords internal
add.coverage.constraints <- function(lprec, covered.templates, template.coverage) {
    # add coverage constraints to ILP 'lprec'
    nbr.coverage.constraints <- length(covered.templates)
    indices <- seq_len(nbr.coverage.constraints)
    cur.nbr.constraints <- 0
    if (nbr.coverage.constraints == 0) {
        indices <- NULL
    }
    #cat(paste("Adding ", length(covered.templates), " coverage constraints. \n", 
        #sep = ""))
    #if (length(covered.templates) > 0) {
        #pb <- txtProgressBar(min = 0, max = length(covered.templates), style = 3)
    #}
    for (j in seq_along(covered.templates)) {
        # select the idx of those primers that are covering the current template
        idx <- template.coverage[[covered.templates[j]]]
        set.row(lprec, j, rep(1, length(idx)), indices = idx)
        #setTxtProgressBar(pb, j)
    }
    if (length(indices) != 0) {
        set.constr.type(lprec, rep(">=", length(indices)), indices)
        set.rhs(lprec, rep(1, length(indices)), indices)
        # rownames <- paste('Cvg', indices, sep = '') rownames(lprec)[indices] <-
        # rownames
    }
    return(lprec)
}

#' Template Coverage.
#'
#' Determines the indices of covering primers for every template.
#'
#' @param cvg.matrix Binary matrix of covering events.
#' @return A list with covering primers for every template.
#' @keywords internal
J.cvg <- function(cvg.matrix) {
    if (length(cvg.matrix) == 0) {
        return(NULL)
    }
    idx <- lapply(seq_len(nrow(cvg.matrix)), function(x) which(cvg.matrix[x, ] > 
        0))
    return(idx)
}
#' Primer Coverage.
#'
#' Determines the indices of covered templates for every primer.
#'
#' @param cvg.matrix Binary matrix of covering events.
#' @return A list with covered templates for every primer.
#' @keywords internal
I.cvg <- function(cvg.matrix) {
    if (length(cvg.matrix) == 0) {
        return(NULL)
    }
    idx <- lapply(seq_len(ncol(cvg.matrix)), function(x) which(cvg.matrix[, x] > 
        0))
    return(idx)
}
#' Construct Coverage ILP.
#'
#' Constructs an ILP modeling the primer set cover problem.
#'
#' @param D Binary dimerization matrix.
#' @param cvg.matrix Binary coverage matrix.
#' @param time.limit Time limit for ILP optimization in seconds.
#' @param presolve.active Whether the ILP presolver should be used.
#' This is set to \code{FALSE} by default, since presolving may lead
#' to inferior solutions. However, for large problems presolving
#' might be useful.
#' @return An instance of the set cover ILP.
#' @keywords internal
ILPConstrained <- function(D, cvg.matrix, time.limit = NULL, presolve.active = FALSE) {
    # melting temperature and binding region conditions are relaxed. here, binding
    # region is modeled with another aux var.
    
    # D: dimerization with entry d_ij = 1 if primers i and j dimerize, 0 else
    # time.limit: stop model computation when timeout is surpassed cvg.matrix:
    # templates x primers. 1 if template is covered
    X <- ncol(cvg.matrix)  # nbr vars
    template.coverage <- J.cvg(cvg.matrix)
    covered.templates <- which(apply(cvg.matrix, 1, sum) != 0)  # number of templates we can cover
    nbr.coverage.constraints <- length(covered.templates)
    # remove symmetric entries in D -> halve the number of dimerization constraints
    D[lower.tri(D)] <- 0  # lower triangle doesn't contain additional info
    D.idx <- which(D == 1, arr.ind = TRUE)  # all pairs of dimerizing primers
    nbr.dimer.constraints <- nrow(D.idx)
    total.nbr.constraints <- nbr.coverage.constraints + nbr.dimer.constraints
    ####### create MILP with binary variables for each primer and real auxiliary variables
    ####### for the relaxed conditions
    lprec <- make.lp(total.nbr.constraints, X)  # 0 constraints and X vars
    name.lp(lprec, "openPrimeR_set_cover_ILP")
    for (col in seq_len(X)) {
        # define primer vars as integer (bounds of 0,1)
        set.type(lprec, col, "binary")
    }
    coefs <- c(rep(1, X))  # costs: uniform
    set.objfn(lprec, coefs)
    # set ILP options
    if (presolve.active) {
        # timeout does not seem to work for loading ILP data, but only for simplex
        # optimization
        lp.info <- lp.control(lprec, sense = "min", timeout = ifelse(is.finite(time.limit), 
            time.limit, 0), verbose = "neutral", epslevel = "tight", presolve = c("lindep", 
            "rows", "probefix", "probereduce", "rowdominate", "impliedfree", "reducegcd", 
            "bounds", "duals", "sensduals", "mergerows", "cols", "colfixdual"))
    } else {
        # timeout does not seem to work for loading ILP data, but only for simplex
        # optimization
        lp.info <- lp.control(lprec, sense = "min", timeout = ifelse(is.finite(time.limit), 
            time.limit, 0), verbose = "neutral", epslevel = "tight")
    }
    rownames <- NULL
    ####### coverage constraint: every template that can be covered should have at least 1
    ####### covering primer
    cur.nbr.constraints <- 0
    lprec <- add.coverage.constraints(lprec, covered.templates, template.coverage)
    cur.nbr.constraints <- cur.nbr.constraints + length(covered.templates)
    # set new indices for lprec access of dimer constraint:
    if (nbr.dimer.constraints != 0) {
        indices <- (cur.nbr.constraints + 1):(cur.nbr.constraints + nbr.dimer.constraints)
    } else {
        indices <- NULL
    }
    ############### dimerization constraint
    lprec <- add.dimerization.constraints(lprec, D.idx, indices)
    cur.nbr.constraints <- cur.nbr.constraints + length(nbr.dimer.constraints)
    # write.lp(lprec, 'mynewLP.lp', 'lp') #write it to a file in LP format
    return(lprec)
}

#' Retrieval of ILP Decisions
#'
#' Retrieves ILP decision variables. 
#'
#' The original dimension of the ILP is required to determine the correct decisions
#' when presolve has been active and dimensions of the ILP might have changed.
#'
#' @param ILP A solved ILP instance.
#' @param original.dim Dimension of \code{ILP} before using presolve.
#' @return The ILP decision variables.
#' @keywords internal
get.ILP.vars <- function(ILP, original.dim = NULL) {
    if (length(original.dim) == 0) {
        return(get.variables(ILP))  # does not contain variables that were removed by presolve
    } else {
        # if presolve removed some variables, we need to know the original ILP
        # dimensionality, to find the 'original' decision variable solutions
        v <- get.primal.solution(ILP, orig = TRUE)  # contains also the variables that were removed by presolve
        vars <- v[(original.dim[1] + 1):length(v)]
        return(vars)
    }
}
#' Selection of Best ILP
#'
#' Selects the best solution from multiple solved ILP instances.
#'
#' @param ILP.df Data frame with ILP result properties.
#' @return The index of the best solution.
#' @keywords internal
select.best.ILP <- function(ILP.df) {
    # select best ILP according to the smallest objective function value after
    # selecting only those ILPs maximizing the coverage ILP.df: overview of solved LP
    # data
    if (length(ILP.df) == 0 || nrow(ILP.df) == 0) {
        warning("Cannot select best ILP as ILP.df was empty.")
        message(ILP.df)
        return(NULL)
    }
    if (all(is.na(ILP.df$Coverage))) {
        # arbitrarily select a set
        sel.idx <- 1
    } else {
        max.cvg <- max(ILP.df$Coverage, na.rm = TRUE)
        sel.idx <- which(ILP.df$Coverage == max.cvg)
        if (any(!is.na(ILP.df$MaxTempDiffToTargetTemp))) {
            sel.idx <- sel.idx[which.min(ILP.df$MaxTempDiffToTargetTemp[sel.idx])]
        } else {
            sel.idx <- sel.idx[1]
        }
    }
    return(sel.idx)
}
#' Construction of ILP Results.
#'
#' Constructs a data frame summarizing the properties of an ILP solution.
#'
#' @param ILP A solved ILP instance.
#' @param vars The ILP decision variables.
#' @param primer.df The primer data frame correspdong to the \code{ILP}.
#' @param template.df The template data frame.
#' @param i Index for the ILP.
#' @param target.temp Target melting temperature in Celsius.
#' @param time Runtime of the ILP.
#' @param deltaG_Cutoff Free energy cutoff used for the dimerization constraint.
#' @param deltaG_Limit The free energy boundary for dimerization.
#' @return Data frame summarizing the ILP solution.
#' @keywords internal
build.ILP.df <- function(ILP, vars, primer.df, template.df, i, target.temp, time = NA, 
    deltaG_Cutoff = NA, deltaG_Limit = NA) {
    # Builds a data frame summarizing a solved ILP
    
    # i: iteration time: time measurement of run need non-null entries for df...
    if (is.null(deltaG_Cutoff)) {
        deltaG_Cutoff <- NA
    }
    if (is.null(deltaG_Limit)) {
        deltaG_Limit <- NA 
    }
    data <- data.frame(Identifier = NA, Iteration = i, Set_Size = NA, Penalty = NA, 
        Coverage = NA, Iterations = NA, Nodes = NA, Time = time, MaxTempDiffToTargetTemp = NA, 
        Objective = NA, deltaG_Cutoff = deltaG_Cutoff, Variables = NA, 
        stringsAsFactors = FALSE)
    if (length(ILP) == 0 || nrow(primer.df) == 0) {
        # no solution found :-(
        return(data)
    }
    # vars <- get.ILP.vars(ILP, original.dim) # this should be used for external
    # representation
    primer.vars <- vars[seq_len(nrow(primer.df))]
    solution.idx <- which(primer.vars > 0.5)
    solution.size <- length(solution.idx)
    cur.primers <- primer.df[solution.idx, ]  # solution primers
    cur.cvg <- get_cvg_ratio(cur.primers, template.df)
    iter <- get.total.iter(ILP)
    val <- get.objective(ILP)
    nodes <- get.total.nodes(ILP)
    # data-associated properties
    max.temp.diff <- NA
    if ("melting_temp" %in% colnames(cur.primers)) {
        max.temp.diff <- max(abs(cur.primers$melting_temp - target.temp))
    }
    id <- paste(template.df$Group[1], "_", target.temp, sep = "")
    data <- data.frame(Identifier = id, Target_Temperature = target.temp, Iteration = i, 
        Set_Size = solution.size, Coverage = cur.cvg, Iterations = iter, Nodes = nodes, 
        Time = time, MaxTempDiffToTargetTemp = max.temp.diff,
        Objective = val, DeltaG_Cutoff = deltaG_Cutoff, DeltaG_Limit = deltaG_Limit, Variables = paste(vars, collapse = ","), 
        stringsAsFactors = FALSE)
    return(data)
}
#' Solve an ILP
#'
#' Constructs and solves an ILP and outputs a list with the reuslts.
#'
#' @param cur.D Binary dimerization matrix.
#' @param cur.G Free energy matrix for cross-dimerization.
#' @param cur.settings Current \code{DesignSettings} object.
#' @param deltaG.cutoff Cutoff for dimerization free energy.
#' @param deltaG.limit Relaxation limit for free energy cutoff.
#' @param cur.cvg.matrix Binary coverage matrix.
#' @param time.limit Time limit for solving the ILP in seconds.
#' @param required.cvg The target coverage of the designed primer set.
#' @param primer.df A \code{Primers} object.
#' @param template.df A \code{Templates} object.
#' @return List with ILP solution data.
#' @keywords internal
solve.ILP <- function(cur.D, cur.G, cur.settings, cur.cvg.matrix, 
                      time.limit, required.cvg, primer.df, template.df) {
    if (length(cur.cvg.matrix) == 0) {
        # no data to optimize
        return(NULL)
    }
    target.cvg <- min(get_cvg_ratio(primer.df, template.df), required.cvg)
    message("Solving ILP; Target cvg: ", target.cvg)
    deltaG.cutoff <- opti(cur.settings)$cross_dimerization
    deltaG.limit <- optiLimits(cur.settings)$cross_dimerization
    ILP <- ILPConstrained(cur.D, cur.cvg.matrix, time.limit)
    original.dim <- dim(ILP)
    return.val <- solve(ILP)  # presolve can also (even when only 'rows' is active, remove variables from the model)
    cvg.ratio <- 0
    if (return.val == 1) {
        # suboptimal solution (due to timeout)
        info.string <- paste("lpsolve: Stopped some of the iterations due to reaching the timeout.", 
            sep = "")
        warning(info.string)
    } else if (return.val != 0) {
        info.string <- paste("lpsolve could not find a solution. The return value was: ", 
            return.val, ". Please check the lpsolve documentation for more details.", 
            sep = "")
        if (return.val != 2) {
            stop(info.string)
        }
    } else if (return.val == 0) {
        vars <- get.ILP.vars(ILP, original.dim)
        primer.idx <- which(vars[seq_len(nrow(primer.df))] > 0.5)
        # get cvg ratio
        cur.cvg <- get_cvg_ratio(primer.df[primer.idx,], template.df)
    }
    initial.deltaG.cutoff <- deltaG.cutoff
    initial.deltaG.limit <- deltaG.limit
    relax.counter <- 0
    # keep on iterating until we're solveable OR we've reached the target cvg
    # OR cvg cannot be improved by changing dimerization cutoff
    while (return.val == 2 || cur.cvg < target.cvg && !all(cur.D == 0)) { 
        relax.counter <- relax.counter + 1
        # relax dimerization constraint
        deltaG.cutoff <- relax.opti.constraints(list("cross_dimerization" = deltaG.cutoff), list("cross_dimerization" = initial.deltaG.limit), 
                                                list("cross_dimerization" = initial.deltaG.cutoff))$cross_dimerization
        deltaG.limit <- relax.opti.constraints(list("cross_dimerization" = deltaG.limit), list("cross_dimerization" = initial.deltaG.limit), 
                                                list("cross_dimerization" = initial.deltaG.cutoff))$cross_dimerization
        message(paste("Setting deltaG.cutoff to ", deltaG.cutoff, sep = ""))
        cur.D <- compute.dimer.matrix(cur.G, deltaG.cutoff["min"])  # 1 for dimers, 0 else
        # update ILP with new dimerization values
        ILP <- ILPConstrained(cur.D, cur.cvg.matrix, time.limit)
        return.val <- solve(ILP)  # presolve can also (even when only 'rows' is active, remove variables from the model)
        original.dim <- dim(ILP)  # keep original.dim updated 
        if (return.val == 0) {
            vars <- get.ILP.vars(ILP, original.dim) 
            primer.idx <- which(vars[seq_len(nrow(primer.df))] > 0.5)
            cur.cvg <- get_cvg_ratio(primer.df[primer.idx,], template.df)
            message("Current cvg is: ", cur.cvg, ". Required cvg is: ", target.cvg)
        }
        #write.lp(ILP, file.path(cur.results.loc, paste0("LP_", relax.counter, ".lp")), 'lp') #write it to a file in LP format
    }
    result <- list(ILP = ILP, DeltaG_Cutoff = deltaG.cutoff["min"], DeltaG_Limit = deltaG.limit["min"], original_dim = original.dim)
    return(result)
}
#' Identification of Redudant Primers.
#'
#' Identifies primers that are redundant.
#'
#' Redundant primers do not reduce the coverage when removed.
#'
#' @param cvg.matrix Binary matrix of coverage events.
#' @return TRUE for redundant primers, FALSE otherwise.
#' @keywords internal
get.redundant.cols <- function(cvg.matrix) {
    template.coverage <- I.cvg(cvg.matrix)
    total.cvg <- length(unique(unlist(template.coverage)))
    is.redundant.col <- unlist(lapply(seq_along(template.coverage), function(x) length(unique(unlist(template.coverage[-x]))) == 
        total.cvg))
    return(is.redundant.col)
}
#' Removal of Redundant Primers.
#'
#' Removes redundant primers from an optimal solution. 
#'
#' An optimal solution can contain primers with redundant coverage when using presolve or greedy optimization.
#'
#' @param S Indices of primers that are selected in an optimal solution.
#' @param cvg.matrix Binary matrix of coverage events.
#' @return Updated indices of selected primers \code{S} such that indices
#' representing primers with redundant coverage are removed.
#' @keywords internal
remove.redundant.cols <- function(S, cvg.matrix) {
    redundant.cols <- get.redundant.cols(cvg.matrix[, S, drop = FALSE])  # boolean: redundant primers in solution
    
    R <- S[redundant.cols]  # index of redundant columns
    non.R <- setdiff(S, R)  # index of non-redundant columns
    if (length(R) == 0) {
        # no redundant columns -> no need to solve another SCP instance
        return(S)
    }
    M <- setdiff(unique(unlist(I.cvg(cvg.matrix[, R, drop = FALSE]))), unique(unlist(I.cvg(cvg.matrix[, 
        non.R, drop = FALSE]))))  # template indices covered by redundant cols only
    if (length(M) == 0) {
        # redundant primers don't afford any additional coverage (this shouldn't happen)
        S <- S[-R]  # remove all redundant primers        
    } else {
        cur.D <- matrix(rep(0, length(R) * length(R)), nrow = length(R))  # matrix without dimerization events
        # solve set cover for redundant primers and corresponding templates
        ILP <- ILPConstrained(cur.D, cvg.matrix[M, R, drop = FALSE], time.limit = Inf, 
                             presolve.active = FALSE) 
        return.val <- solve(ILP)
        vars <- get.variables(ILP)
        S <- S[-R[vars == 0]]
    }
    return(S)
}
#' Solver for ILP Set Cover
#'
#' Solves the primer set cover problem using an ILP formulation.
#'
#' @param primer.df Primer data frame to be optimized.
#' @param template.df Template data frame with sequences.
#' @param settings A \code{DesignSettings} object.
#' @param primer_conc Primer concentration.
#' @param template_conc Template concentration.
#' @param na_salt_conc Sodium ion concentration.
#' @param mg_salt_conc Magensium ion concentration.
#' @param k_salt_conc Potassium ion concentration.
#' @param tris_salt_conc Tris ion concentration.
#' @param allowed.mismatches The number of mismatches primers are allowed to have with the templates.
#' @param allowed.other.binding.ratio Ratio of primers allowed to bind to non-target regions.
#' @param allowed.stop.codons  Consider mismatch binding events that induce stop codons.
#' @param allowed.region.definition Definition of the target binding sites used for evaluating the coverage.
#' If \code{allowed.region.definition} is \code{within}, primers have to lie within the allowed binding region.
#' If \code{allowed.region.definition} is \code{any}, primers have to overlap with the allowed binding region.
#' The default is that primers have to bind within the target binding region.
#' @param disallowed.mismatch.pos The number of positions from the primer 3' end where mismatches should not be allowed.
#' All primers binding templates with mismatches within \code{disallowed.mismatch.pos} from the 3' end are disregarded.
#' @param target.temps Target melting temperatures for primer sets in Celsius.
#' @param required.cvg Target coverage ratio of the templates by the primers.
#' @param fw.primers List with optimized primer data frames corresponding to \code{target.temps}. 
#' Only required for optimizing both strand directions and only 
#' in the second optimization run in order to check for cross dimerization.
#' @param diagnostic.location Directory for storing results.
#' @param timeout Timeout in seconds for the optimization with ILPs.
#' @param updateProgress Shiny progress callback function.
#' @return List with optimization results.
#' @keywords internal
optimize.ILP <- function(primer.df, template.df, settings, primer_conc, 
    template_conc, na_salt_conc, mg_salt_conc, k_salt_conc, tris_salt_conc, 
    allowed.mismatches, allowed.other.binding.ratio, allowed.stop.codons, 
    allowed.region.definition, disallowed.mismatch.pos, target.temps, 
    required.cvg, fw.primers = NULL, diagnostic.location = NULL, 
    timeout = Inf, updateProgress = NULL) {
    # todo: add updateprogress arg in server.R
    
    # sanity check nothing to optimize
    if (length(primer.df) == 0) {
        return(NULL)
    }
    #if (!"melting_temp" %in% colnames(primer.df)) {
        #stop("Melting temp has to be computed before optimizing")
    #}
    opti.constraints <- opti(settings)
    opti.limits <- optiLimits(settings)
    cvg.matrix <- get.coverage.matrix(primer.df, template.df)
    mode.directionality <- get.analysis.mode(primer.df)
    Tm.brackets <- create.Tm.brackets(primer.df, template.df, settings, target.temps)
    primer.df <- Tm.brackets$primers
    target.temps <- Tm.brackets$df$target_Tm
    #### COMPUTE CONSTRAINTS cross-dimerization
    if ("cross_dimerization" %in% names(opti.constraints)) {
        deltaG.cutoff <- opti.constraints$cross_dimerization["min"]
        deltaG.limit <- opti.limits$cross_dimerization["min"]
        if (length(deltaG.cutoff) == 0 || length(deltaG.limit) == 0) {
            stop("DeltaG constraint/limit was active but values were not specified.")
        }
        message("Computing cross dimers. This may take a while ...")
        # nb: G matrix is computed for the smallest annealing temperature only
        # computing for every Ta would be more accurate but would
        # require more computation time
        G.df <- compute.all.cross.dimers(primer.df, primer_conc, na_salt_conc, 
                            mg_salt_conc, k_salt_conc, tris_salt_conc, 
                            min(Tm.brackets$df$annealing_temp),
                            no.structures = TRUE)
        G.matrix <- create.G.matrix(primer.df, G.df)  # min deltaG values of cross-dimerization conformations for every pair of primers
        # load G matrix for debugging: G.matrix <-
        # read.csv(file.path(diagnostic.location, 'G_matrix.csv'))
        if (length(diagnostic.location) != 0) {
            write.csv(G.matrix, file.path(diagnostic.location, "G_matrix.csv"), row.names = FALSE)
        }
        D <- compute.dimer.matrix(G.matrix, deltaG.cutoff)  # 1 for dimers, 0 else
    } else {
        # set dimerization D-matrix to zeros -> all primers are compatible
        D <- matrix(rep(0, nrow(primer.df) * nrow(primer.df)), nrow = nrow(primer.df), 
            ncol = nrow(primer.df))
        G.matrix <- matrix(rep(0, nrow(primer.df) * nrow(primer.df)), nrow = nrow(primer.df), 
            ncol = nrow(primer.df))
        deltaG.cutoff <- NULL
        deltaG.limit <- NULL
    }
    # determine annealing-temperature dependent constraints and filter according to them
    Tm.data <- compute.Tm.sets(primer.df, template.df, Tm.brackets, settings, 
                mode.directionality, primer_conc, template_conc, na_salt_conc, mg_salt_conc, 
                k_salt_conc, tris_salt_conc, allowed.mismatches, allowed.other.binding.ratio, 
                allowed.stop.codons, allowed.region.definition, disallowed.mismatch.pos, 
                TRUE, required.cvg, fw.primers, diagnostic.location, updateProgress)  
    Tm.sets <- Tm.data$sets
    Tm.settings <- Tm.data$settings # relaxed settings
    ######### 
    i <- NULL

    ILP.df <- foreach(i = seq_along(target.temps), .combine = my_rbind) %dopar% {
            target.temp <- target.temps[i]
            Tm.set <- Tm.sets[[i]]
            cur.settings <- Tm.settings[[i]]
            if (nrow(Tm.set) == 0) {
                # create empty entry
                data <- build.ILP.df(NULL, NULL, Tm.set, template.df, i, target.temp)
            }
            # Tm.set <- read_primers(file.path(opti.results.loc,
            # paste('filtered_opti_data_target_temp_', target.temp, '.csv', sep = '')))
            #cat(paste("Optimization for target temp ", target.temp, " commences ...\n", 
                #sep = ""))
            sel.idx <- match(Tm.set$Identifier, primer.df$Identifier)  # index for accessing global properties of primer.df
            cur.G <- G.matrix[sel.idx, sel.idx, drop = FALSE]
            cur.D <- D[sel.idx, sel.idx, drop = FALSE]
            cur.cvg.matrix <- cvg.matrix[, sel.idx, drop = FALSE]
            time <- Sys.time()  # measure optimization time
            #### test sample.size <- 200 cur.G <- cur.G[1:sample.size,1:sample.size] Tm.set <-
            #### Tm.set[1:sample.size,]
            solution.data <- solve.ILP(cur.D, cur.G, cur.settings, 
                                       cur.cvg.matrix, timeout, required.cvg, Tm.set, template.df)
            if (length(solution.data) == 0 || is(solution.data, "try-error")) {
                # optimization was not possible for some reason (most likely due to dimerization
                # constraints)
                data <- build.ILP.df(NULL, NULL, Tm.set, template.df, i, target.temp)
            } else {
                vars <- get.ILP.vars(solution.data$ILP, solution.data$original_dim)
                primer.idx <- which(vars[seq_len(nrow(Tm.set))] > 0.5)
                primer.idx <- remove.redundant.cols(primer.idx, cur.cvg.matrix)
                vars[!seq_len(nrow(Tm.set)) %in% primer.idx] <- 0  # remove redundant cols from vars
                solution <- solution.data$ILP
                relaxed.deltaG.cutoff <- solution.data$DeltaG_Cutoff
                relaxed.deltaG.limit <- solution.data$DeltaG_Limit
                # for parallelization, need to extract the relevant infos from the ILP solutions,
                # because objects are destroyed when threads end (solutions are only ptrs to
                # lprec objects)
                time <- as.numeric(difftime(Sys.time(), time, units = "secs"))
                data <- build.ILP.df(solution, vars, Tm.set, template.df, i, target.temp, 
                  time = time, deltaG_Cutoff = relaxed.deltaG.cutoff, deltaG_Limit = relaxed.deltaG.limit)
            }
            data
    }
    solution.idx <- select.best.ILP(ILP.df)
    if (length(solution.idx) == 0) {
        warning("No optimal primers found. No primer set could be optimized with ILPs. Check your constraint settings (dimerization!).")
        return(NULL)
    }
    all.optimal.sets <- get.sets.from.decisions(ILP.df, Tm.sets)
    if (length(diagnostic.location) != 0) {
        # output info
        optimal <- rep(FALSE, nrow(ILP.df))
        optimal[solution.idx] <- TRUE
        lambda.idx <- which(colnames(ILP.df) == "Objective")
        ILP.df <- cbind(ILP.df[, seq_len(lambda.idx)], Optimal = optimal, ILP.df[, ((lambda.idx + 
            1):ncol(ILP.df))], stringsAsFactors = FALSE)
        write.csv(ILP.df, file = file.path(diagnostic.location, "ILP_summary.csv"), 
            row.names = FALSE)
    }
    optimal.primers <- all.optimal.sets[[solution.idx]]
    relaxed.deltaG_Cutoff <- ILP.df$deltaG_Cutoff[solution.idx]
    # output the used opti constraints for each Tm set
    if ("cross_dimerization" %in% names(opti.constraints)) {
        # deltaG is the only relaxed opti constraint for ILPs
        for (i in seq_along(Tm.settings)) {
            if (!is.na(ILP.df$Identifier[i])) {
                constraintLimits(Tm.settings[[i]])$cross_dimerization["min"] <- ILP.df$DeltaG_Limit[i]
                constraints(Tm.settings[[i]])$cross_dimerization["min"] <- ILP.df$DeltaG_Cutoff[i]
            }
        }
    }
    result <- list(opti = optimal.primers, all_results = all.optimal.sets, 
        all_used_constraints = Tm.settings, 
        used_constraints = Tm.settings[[solution.idx]]) 
    return(result)
}
#' Optimal Sets from Decision Variables
#'
#' Determines primer sets from decision variables from ILP.
#'
#' @param ILP.df Data frame with ILP optimization results.
#' @param Tm.sets List with primer data frames for every target melting temperature.
#' @return A list with optimal primer data sets for every target temperature.
#' @keywords internal
get.sets.from.decisions <- function(ILP.df, Tm.sets) {
    results <- vector("list", length(Tm.sets))
    for (i in seq_along(Tm.sets)) {
        Tm.set <- Tm.sets[[i]]  # solution filtered data (with efficiencies!)
        primer.idx <- which(as.numeric(strsplit(as.character(ILP.df$Variables[i]), 
            split = ",")[[1]])[seq_len(nrow(Tm.set))] > 0.5)
        optimal.primers <- Tm.set[primer.idx, ]
        results[[i]] <- optimal.primers
    }
    names(results) <- ILP.df$Target_Temperature
    return(results)
}
#' Subset ILP Constructor
#'
#' Constructs an ILP for selecting optimal primer subsets.
#'
#' Here, "optimal" refers to a subset of a certain size that maximizes the coverage.
#'
#' @param primer.df Primer data frame to be subsetted.
#' @param template.df Template data frame.
#' @param k Required number of primers to be selected.
#' @return An ILP for choosing the primer subset of size \code{k} with the largest coverage.
#' @keywords internal
subset.ILP <- function(primer.df, template.df, k) {
    if (length(primer.df) == 0 || nrow(primer.df) == 0)  {
        return(NULL)
    }
    if (length(template.df) == 0 || nrow(template.df) == 0) {
        return(NULL)
    }
    if (k <= 0) {
        stop("Subset size 'k' should be positive.")
    }
    # template.coverage: for each template, the indices of covering primers
    template.coverage <- get.covered.templates(primer.df, template.df)
    covered.templates <- which(sapply(template.coverage, length) != 0)  # nbr of coverage constraints
    nbr.coverage.constraints <- length(covered.templates)
    xp <- nrow(primer.df)  # primer decision variables
    xt <- nbr.coverage.constraints  # template decision variables: only those templates that are covered are considered
    # we have two variables (x_i: primer selected, y_i: template covered)
    X <- xp + xt  # nbr of decision variables
    # no coverage constraint, instead constraint on the set size
    total.nbr.constraints <- 1 + nbr.coverage.constraints  # constraint for set size + template coverage constraint
    lprec <- make.lp(total.nbr.constraints, X)
    name.lp(lprec, "openPrimeR_SubsetSelection_ILP")
    for (col in seq_len(X)) {
        # define primer vars as integer (bounds of 0,1)
        set.type(lprec, col, "binary")
    }
    coefs <- c(rep(0, xp), rep(1, xt))  # maximize the sum of templates that are covered
    set.objfn(lprec, coefs)
    # set ILP options
    lp.info <- lp.control(lprec, sense = "max", verbose = "neutral", 
        epslevel = "tight")
    rownames <- NULL
    # add constraints subset size constraint
    set.row(lprec, 1, c(rep(1, xp), rep(0, xt)))
    set.constr.type(lprec, "=", 1) # set should be of specified size
    set.rhs(lprec, k, 1)
    rownames(lprec)[1] <- "SetSize_Constraint"
    ####### coverage constraint: 
    for (j in seq_along(covered.templates)) {
        # select the idx of those primers that are covering the current template
        template.lp.idx <- j + xp
        idx <- template.coverage[[covered.templates[j]]]  # index of primers covering the j-th coverable template
        set.row(lprec, j + 1, c(-1, rep(1, length(idx))), indices = c(template.lp.idx, 
            idx))  # j+1 because we already have the coverage constraint in lprec row 1
        #setTxtProgressBar(pb, j)
    }
    #cat("\n")
    if (length(covered.templates) != 0) {
        set.constr.type(lprec, rep(">=", length(covered.templates)), 2:(nbr.coverage.constraints + 
            1))
        set.rhs(lprec, rep(0, length(covered.templates)), 2:(nbr.coverage.constraints + 
            1))
        rownames(lprec)[2:(nbr.coverage.constraints + 1)] <- "Coverage_Constraint"
    }
    cnames.p <- paste("Primer_", seq_len(nrow(primer.df)), sep = "")
    if (length(covered.templates) != 0) {
        cnames.t <- paste("Template_", covered.templates, sep = "")
    } else {
        cnames.t <- NULL
    }
    colnames(lprec) <- c(cnames.p, cnames.t)
    return(lprec)
}
#' Determination of Primer Coverage for Groups.
#'
#' Modifies a primer data frame to retain only coverage events
#' relating to the selected groups of templates.
#'
#' @param primer.df Primer data frame.
#' @param template.df Template data frame.
#' @param groups Template groups for which coverage should be determined.
#' @return \code{primer.df} with coverage considered only for \code{groups}.
#' @keywords internal
primer.coverage.for.groups <- function(primer.df, template.df, groups) {
    if (!is.null(groups)) {
        # filter template.df for groups
        template.df <- template.df[template.df$Group %in% groups, ]
    }
    idx <- covered.seqs.to.idx(primer.df$Covered_Seqs, template.df)
    cvg.counts <- sapply(idx, function(x) ifelse(length(x) == 0, 0, length(x[!is.na(x)]))) # na check
    names(cvg.counts) <- seq_len(nrow(primer.df))
    primer.df$primer_coverage <- cvg.counts
    # modify covered_seqs of primers
    template.IDs <- unlist(lapply(idx, function(x) {
        if (is.null(x)) {
            ""
        } else {
            paste(template.df$Identifier[x[!is.na(x)]], 
                    collapse = ",")
        }
   }))
    primer.df$Covered_Seqs <- template.IDs
    primer.df$Coverage_Ratio <- cvg.counts/nrow(template.df)
    ## remove all fields that weren't updated
    # if other fields are required -> re-evaluate the primers
    keep.cols <- c("Forward", "Reverse", "ID", "Identifier", 
                    "primer_length_fw", "primer_length_rev",
                    "Run", "primer_coverage", "Coverage_Ratio", 
                    "Covered_Seqs", "Direction")
    primer.df <- primer.df[, colnames(primer.df) %in% keep.cols]
    return(primer.df)
}

#' @rdname PrimerEval
#' @name PrimerEval
#' @details
#' \code{subset_primer_set} determines optimal subsets of the input primer set
#' by solving an integer-linear program.
#' Since the quality of the primers (in terms of properties) is not taken into
#' account when creating the subsets, this method should only be used
#' for primer sets that are already of high quality. 
#' 
#' @return \code{subset_primer_set} returns a list with optimal primer subsets,
#' each of class \code{Primers}.
#' @export
#' @examples
#' 
#' # Determine optimal primer subsets
#' data(Ippolito)
#' primer.subsets <- subset_primer_set(primer.df, template.df, k = 3)
subset_primer_set <- function(primer.df, template.df, k = 1, groups = NULL, identifier = NULL, cur.results.loc = NULL) {
    if (k <= 0) {
        stop("k has to be positive ...")
    }
    if (length(primer.df) == 0 || nrow(primer.df) == 0)  {
        return(NULL)
    }
    if (length(template.df) == 0 || nrow(template.df) == 0) {
        return(NULL)
    }
    if (k > nrow(primer.df)) {
        stop("k shouldn't exceed size of primer set")
    }
    if (!is(primer.df, "Primers")) {
        stop("Please supply a valid primer data frame.")
    }
    if (!"primer_coverage" %in% colnames(primer.df)) {
        return(NULL)
    }
    if (!is(template.df, "Templates")) {
        stop("Please provide a valid template data frame.")    
    }
    # need either individual primers or only primer pairs for subsets to make sense
    if ((!all(primer.df$Forward == "") && any(primer.df$Reverse != "")) || (!all(primer.df$Reverse == 
        "") && any(primer.df$Forward != "")) || (any(primer.df$Forward != "") && 
        any(primer.df$Reverse != "") && !all(primer.df$Forward != "" & primer.df$Reverse != 
        ""))) {
            primer.df <- pair_primers(primer.df, template.df)
    }
     # set the same analysis mode for all subsets:
    dirs <- unique(primer.df$Direction)
    if (length(dirs) >= 2) {
        mode.directionality <- "both"
    } else {
        mode.directionality <- dirs
    }
    p.df <- primer.coverage.for.groups(primer.df, template.df, groups)
    out.loc <- file.path(cur.results.loc, identifier)
    if (length(out.loc) != 0) {
        dir.create(out.loc, showWarnings = FALSE)
    }
    K <- seq_len(100) * k
    K <- K[K <= nrow(primer.df)]
    # for (i in seq_along(K)) {
    k <- NULL
    top.primers <- foreach(k = seq_along(K), .combine = c) %dopar% {
        ILP <- subset.ILP(p.df, template.df, K[k])
        return.val <- solve(ILP)
        if (return.val != 0) {
            warning(return.val)
        }
        if (length(out.loc) != 0) {
            write.lp(ILP, file.path(out.loc, paste("ILP_", K[k], sep = "")), "lp")  #write it to a file in LP format
        }
        vars <- get.variables(ILP)
        primer.vars <- vars[seq_len(nrow(p.df))]
        sel.idx <- which(primer.vars == 1)
        cur.top.primers <- p.df[sel.idx, ]
        fname <- file.path(out.loc, paste("subset_k=", K[k], ".csv", sep = ""))
        if (length(fname) != 0) {
            write.csv(cur.top.primers, fname, row.names = FALSE)
        }
        list(cur.top.primers)
    }
    return(top.primers)
}

################
# Input/Output
################

#' Coverage Info Text
#'
#' Creates a string with information on the coverage
#'
#' @param stats Data frame with coverage statistics.
#' @param selected.group Retrieve information for a subgroup of templates only.
#' @param ident An identifier for the coverage.
#' @return A string with information on the coverage.
#' @keywords internal
create.cvg.text <- function(stats, selected.group = NULL, ident = NULL) {
    if (length(selected.group) == 0 || "all" %in% selected.group) {
        selected.group <- "Total"
    } 
    if (selected.group[1] == "Total") {
        out.selected.group <- "" # for output: change id
    } else {
        out.selected.group <- selected.group
    }
    if (length(ident) == 0) { # identifier for coverage type
        ident <- "Coverage"
    }
    idx <- which(stats$Group %in% selected.group)
    if (length(idx) != 0) {
        stats <- stats[idx, ]
        # differentiate fw & rev cvg
        text <- paste0(ident, ": ", stats$N_primer, " primers (", stats$N_primer_fw, " fw/", stats$N_primer_rev, 
            " rev) cover ", stats$Coverage, " ", out.selected.group, " template sequences.")
    } else {
        text <- "Templates not covered."
    }
    return(text)
}
#' Check for Report Dependencies.
#'
#' Checks whether the dependencies for rmarkdown::render() are fulfilled.
#'
#' @return A logical vector giving the dependency availability status.
#' @keywords internal
check_report_deps <- function() {
	pandoc.available <- Sys.which("pandoc") != ""
	latex.available <- Sys.which("pdflatex") != ""
	out <- c("Pandoc" = pandoc.available, "LateX" = latex.available)
	return(out)
}
#' Sanitiziation of Filename.
#'
#' Ensures that a filename is valid for the file system.
#'
#' @param path The path to the file to be sanitized, without file extension.
#' @param suffix The suffix (e.g. ".png") of a file.
#' @param sub_char The character used to replacing disallowed chars.
#' @return The sanitized filename
#' @keywords internal
sanitize_path <- function(path, suffix = '', sub.char = "_") {
  if (grepl('[^~:_./\\[:alnum:]-]', path)) {
    warning('Replaced special characters in path "', path, '" -> "',
            path <- gsub('[^~:_./\\[:alnum:]-]', sub.char, path), '"')
  }
  # replace . with sub.char except ../ and ./
  s = strsplit(path, '[/\\\\]')[[1L]]
  i = (s != '.') & (s != '..') & grepl('\\.', s)
  if (any(i)) {
    s[i] = gsub('\\.', sub.char, s[i])
    path = paste(s, collapse = '/')
    warning('Dots in path replaced with", sub.char, "("', path, '")')
  }
  out <- paste0(path, suffix)
  return(out)
}
#' Creation of a Filename for Reports.
#'
#' Creates the filename for reports.
#'
#' @param report.name The identifier for the report type.
#' @param sample.name The identifier of the sample that was analyzed.
#' @return A character vector.
#' @keywords internal
get_report_fname <- function(report.name, sample.name) {
    date <- format(Sys.time(), "%Y-%m-%d")
    out <- paste(report.name, "_", sample.name, "_", date, sep = "")
    out <- sanitize_path(out, ".pdf")
    return(out)
}
#' Renders an rmarkdown file using Pandoc.
#'
#' Creates a PDF report using rmarkdown and Pandoc by passing the specified
#' \code{params} to the markdown file given by \code{report_template} and 
#' storing the PDF in \code{out.file}.
#'
#' @param params A list with parameters for the R markdown parser.
#' @param report_template A character vector giving the basename of the
#' Rmarkdown template to use for report creation.
#' @param out.file The filename of the report PDF to be created.
#' @return Creates a PDF in \code{out.file} if successful.
#' @keywords internal
render_report <- function(params, report_template, out.file) {
    deps <- check_report_deps()
	missing.deps <- names(deps)[which(!deps)]
	if (length(missing.deps) != 0) {
		msg <- paste("Cannot create PDF report. Dependencies are missing:\n",
					paste(missing.deps, collapse = ","))
		warning(msg)
		return()
	}
    report.template.dir <- system.file("extdata", "report", 
                            package = "openPrimeR")
    # copy report file to temporary directory to make sure we can write
    generation.dir <- tempdir() # dir where report is generated
    dir.copy(report.template.dir, generation.dir, overwrite = TRUE)
    tempReport <- file.path(generation.dir, report_template)
    if (rmarkdown::pandoc_available()) {
        dir.create(dirname(out.file), showWarnings = FALSE, recursive = TRUE)
        out.file <- file.path(normalizePath(dirname(out.file)), basename(out.file))
        res <- try(rmarkdown::render(tempReport, output_file = out.file, 
            params = params, envir = new.env(parent = globalenv()),
            quiet = TRUE))
        if (is(res, "try-error")) {
            warning("No PDF file was produced, please check the console output and whether all Latex dependencies are installed (Texlive available?).")
        }
    } else {
        msg <- paste0("Pandoc for rmarkdown is not available on your system.", 
                "Please install it first to generate a report.")
        warning(msg)
        return()
    }
}

#' Read FASTA File.
#'
#' Reads the input FASTA file.
#'
#' @param fasta.file The path to a FASTA file.
#' @param NTs The allowed set of nucleotides.
#' @return List with vectors of chars.
#' @keywords internal
my.read.fasta <- function(fasta.file, NTs) {
    # NTs: supported NTs
    if (!file.exists(fasta.file)) {
        stop(paste("The input fasta.file did not exist at: '",
            fasta.file, "'", sep = ""))
    }
    seqs <- suppressWarnings(seqinr::read.fasta(fasta.file, forceDNAtolower = TRUE))
    # replace empty chars
    for (i in seq_along(seqs)) {
        if (" " %in% seqs[[i]]) {
            new.seq <- seqs[[i]][seqs[[i]] != " "]
            seqs[[i]] <- seqinr::as.SeqFastadna(new.seq, attr(seqs[[i]], "name"), attr(seqs[[i]], 
                "Annot"))
        }
    }
    # check whether the seqs are in the NT alphabet
    sanity.check <- unlist(lapply(seqs, function(x) all(x %in% NTs)))
    idx <- which(!sanity.check)
    # message(idx)
    if (length(idx) != 0) {
        msg <- "Some sequences contained non-supported characters. Supported are the following characters: "
        sup <- paste(NTs, collapse = ", ")
        loc <- paste(sapply(idx, function(x) attr(seqs[[x]], "name")), collapse = ", ")
        pre <- "The sequences with the following headers contained non-supported characters: "
        msg <- paste(msg, sup, ". ", pre, loc, sep = "")
        my.error("FastaAlphabetError", msg)
    }
    return(seqs)
}
#' Wrapper for the ggplot2::ggsave function.
#'
#' Saves a plot using ggplot2's ggsave function.
#'
#' @param filename The filename to store the plot.
#' @param plot The ggplot object.
#' @param ... Further arguments to the ggplot2 ggsave function.
#' @return Stores \code{p} in \code{fname}.
#' @keywords internal
my_ggsave <- function(filename, plot = ggplot2::last_plot(), ...) {
   if (length(plot) != 0) {
        check <- try(suppressMessages(ggsave(filename = filename, plot = plot, ...)))
        if (is(check, "try-error")) {
            # downgrade error to warning
            warning(attr(check, "condition") )
        }
   }
   return()
}
add_cvg_to_workbook <- function(cvg.matrix, wb, start.row, start.col, is.first.entry = FALSE) {
    # wb workbook
    # start.col The column to start writing into (counting from 1)
    # position for row/column style separators:
    # c <- seq(start.col, ncol(cvg.matrix) - 1)
    # r <- seq(start.row, start.row + nrow(cvg.matrix) - 1)
    if (is.first.entry) {
        #r <- r + 1
        openxlsx::writeData(wb, 1, cvg.matrix, 
              startCol = start.col, startRow = start.row,
              rowNames = FALSE, colNames = TRUE, borders = "all",
              borderColour = "#000000")

    } else {
        openxlsx::writeData(wb, 1, cvg.matrix, 
              startCol = start.col, startRow = start.row,
              rowNames = FALSE, colNames = FALSE, borders = "all",
              borderColour = "#000000")

    }
    # styles:
    #bodyStyle <- createStyle(border="Bottom", 
                            #borderColour = "#000000",
                            #borderStyle = "thick")
    #r <- tail(r,1)
    # doesnt work:
    #addStyle(wb, sheet = 1, bodyStyle, rows = r, cols = c, gridExpand = TRUE)
    #setColWidths(wb, 1, cols=1, widths = 21) ## set column width for row names column
    return(wb)
}

#' @rdname Output
#' @name Output
#' @aliases create_coverage_xls
#' @return \code{create_coverage_xls} stores information on the
#' primer coverage in a spreadsheet.
#' @export
#' @examples
#'
#' data(Ippolito)
#' filename <- tempfile("cvg_overview", fileext = ".xls")
#' # Store coverage of a single primer in an XLS file:
#' my.primers <- primer.df[3,]
#' cvd <- unique(unlist(strsplit(my.primers$Covered_Seqs, split = ",")))
#' m <- match(cvd, template.df$Identifier)
#' my.templates <- template.df[m,]
#' create_coverage_xls(my.primers, my.templates, filename, settings)
create_coverage_xls <- function(primer.df, template.df, fname, settings) {
    if (length(primer.df) == 0 || nrow(primer.df) == 0) {
        return(NULL)
    }
    if (!is(settings, "DesignSettings")) {
        stop("A DesignSettings object is required for settings.")
    }
    if (!is(primer.df, "Primers")) {
        stop("Please input a 'Primers' object.")
    }
    if (!is(template.df, "Templates")) {
        stop("Please input a 'Templates' object.")
    }
    if (!"primer_coverage" %in% colnames(primer.df)) {
        warning("Please compute primer coverage first.")
        return(NULL) 
    }
    cvg.constraints <- cvg_constraints(settings)
    constraints <- names(cvg.constraints)[names(cvg.constraints) %in% c("primer_efficiency", "annealing_DeltaG")]
    constraints <- constraints[constraints %in% colnames(primer.df)]
    if (length(constraints) > 1) {
        warning("Only one constraint possible at a time. Selecting the first one.")
        constraints <- constraints[1]
    }
    # initialize workbook:
    wb <- openxlsx::createWorkbook()
    openxlsx::addWorksheet(wb, primer.df$Run[1])
    # create coverage data for individual mismatch settings:
    mm <- max(unique(unlist(sapply(strsplit(c(primer.df$Nbr_of_mismatches_fw, primer.df$Nbr_of_mismatches_rev), split = ","), function(x) as.numeric(x)))))
    mm <- c(0, seq_len(mm))
    # add mismatch column to worksheet:
    mm.column <- data.frame("Allowed_Mismatches" = unlist(lapply(mm, function(x) rep(x, nrow(primer.df)))))
    openxlsx::writeData(wb, 1, mm.column , startCol = 1, 
              rowNames = FALSE)
    #openxlsx::openXL(wb) # error already here: SET_VECTOR_ELT() can only be applied to a list, not a symbol

    start.row <- 1
    for (i in seq_along(mm)) {
        # create coverage data for one mismatch setting
        allowed.mm <- mm[i]
        fw.mm <- lapply(strsplit(primer.df$Nbr_of_mismatches_fw, split = ","), as.numeric)
        allowed.idx.fw <- lapply(fw.mm, function(x) which(x <= allowed.mm))
        rev.mm <- lapply(strsplit(primer.df$Nbr_of_mismatches_fw, split = ","), as.numeric)
        allowed.idx.rev <- lapply(fw.mm, function(x) which(x <= allowed.mm))
        # select worst case inclusion
        allowed.idx <- lapply(seq_len(nrow(primer.df)), function(x) {
                            if (primer.df$Direction[x] == "fw") {
                                allowed.idx.fw[[x]]
                            } else if (primer.df$Direction[x] == "rev") {
                                allowed.idx.rev[[x]]
                            } else { # both
                                intersect(allowed.idx.fw[[x]],allowed.idx.rev[[x]])
                            }
        })
        # update covered seqs for allowed nbr of mismatches:
        cur.df <- primer.df
        cvd.idx <- covered.seqs.to.idx(primer.df$Covered_Seqs, template.df)
        cvd <- sapply(seq_along(allowed.idx), function(x) paste(template.df$Identifier[cvd.idx[[x]][allowed.idx[[x]]]], collapse = ","))
        cur.t.df <- update_template_cvg(template.df, cur.df)
        cur.df$Covered_Seqs <- cvd
        # get coverage matrix for specific number of mismatches
        if (length(constraints) == 0) {
            cvg.matrix <- t(get.coverage.matrix(cur.df, cur.t.df))
        } else {
            # change entries to cvg constraint values: efficiency / deltaG of annealing
            # re-write constraints: select only the current constraint events
            split.vals <- strsplit(primer.df[, constraints], split = ",")
            vals <- unlist(lapply(seq_along(split.vals), function(x) paste(split.vals[[x]][allowed.idx[[x]]], collapse = ",")))
            cur.df[,constraints] <- vals
            cvg.matrix <- t(get.coverage.matrix(cur.df, cur.t.df, constraints = constraints))
            # overwrite previous 0/1 coverag entries with cvg constraint values
        }
        cnames <- c("Primer", colnames(cvg.matrix))
        cvg.matrix <- cbind(Primer = rownames(cvg.matrix), 
                                data.frame(cvg.matrix, stringsAsFactors = FALSE))
        cvg.matrix$Primer <- as.character(cvg.matrix$Primer)
        colnames(cvg.matrix) <- cnames
        # add entries to notebook
        wb <- add_cvg_to_workbook(cvg.matrix, wb, start.row = start.row, start.col = 2, is.first.entry = (i == 1))
        if (i == 1) {
            # first entry in workbook
            start.row <- start.row + nrow(primer.df) + 1 # plus the header row
        } else {
            start.row <- start.row + nrow(primer.df)
        }
    }
    # conditional formatting is applied when xls is created and based on the final values ...
    # color according to cvg.matrix entries -> 1 covered, 0 not
    coveredStyle <- openxlsx::createStyle(fontColour = "#000000", bgFill = "#00b050")
    uncoveredStyle <- openxlsx::createStyle(fontColour = "#000000", bgFill = "#ff0000")
    # row and column indices for style rules:
    r <- seq_len(nrow(mm.column)) + 1
    c <- seq_len(nrow(template.df)) + 2 # add 2 for primer and mismatch cols
    if (length(constraints) == 0) {
        match.rule <- "==1"
        mismatch.rule <- "==0"
    } else {
        if (constraints == "primer_efficiency") {
            match.rule <- paste0(">=", cvg.constraints[[constraints]]["min"])
            mismatch.rule <- paste0("<", cvg.constraints[[constraints]]["min"])
        } else { # annealing
            match.rule <- paste0("<=", cvg.constraints[[constraints]]["max"])
            mismatch.rule <- paste0(">", cvg.constraints[[constraints]]["max"])
        }
    }
    openxlsx::conditionalFormatting(wb, 1, cols = c, rows = r, rule = match.rule, style = coveredStyle)
    openxlsx::conditionalFormatting(wb, 1, cols = c, rows = r, rule = mismatch.rule, style = uncoveredStyle)
    #openxlsx::openXL(wb) # open temporary version for testing!
    openxlsx::saveWorkbook(wb, fname, overwrite = TRUE) 
}

#' @rdname Output
#' @name Output
#' @aliases create_report
#' @return \code{create_report} Creates a PDF file summarizing the results
#' from analyzing one or multiple sets of primers.
#' @export
#' @include primers.R templates.R
#' @note
#' Creating the report requires the external programs Pandoc (http://pandoc.org)
#' and LaTeX (http://latex-project.org).
#' @examples
#' 
#' setting.xml <- system.file("extdata", "settings", 
#'                  "C_Taq_PCR_high_stringency.xml", package = "openPrimeR")
#' settings <- read_settings(setting.xml)
#' # Creation of a report for a single primer set
#' data(Ippolito)
#' out.file.single <- tempfile("evaluation_report", fileext = ".pdf")
#' create_report(primer.df, template.df, out.file.single, settings)
#' # Creation of a report for multiple primer sets
#' data(Comparison)
#' set.sizes <- sapply(primer.data, nrow)
#' sel.sets <- order(set.sizes)[1:2]
#' out.file.comp <- tempfile("comparison_report", fileext = ".pdf")
#' create_report(primer.data[sel.sets], template.data[sel.sets], out.file.comp, settings)
setGeneric("create_report", 
    function(primers, templates, fname, settings, 
             sample.name = NULL, used.settings = NULL, ...) {
        standardGeneric("create_report")
})
#' Creation of a PDF report.
#' 
#' Creates a PDF report for a set of primers.
#'
#' @param primers An evaluated \code{Primers} object.
#' @param templates A \code{Templates} object.
#' @param fname A character vector giving the file to store the report in.
#' @param settings A \code{DesignSettings} object.
#' @param sample.name An identifier for your analysis.
#' @param used.settings A named list (with fields "fw" and "rev") containing the forward/reverse used design settings.
#' @param required.cvg The required coverage ratio.
#' @return Creates a PDF file reporting on the input primers.
#' @keywords internal
#' @note
#' Creating the report requires the external programs Pandoc (http://pandoc.org)
#' and LaTeX (http://latex-project.org).
setMethod("create_report", c("Primers", "Templates"), function(primers, templates, 
             fname, settings, sample.name, used.settings, required.cvg = 1) {
    mode.directionality <- get.analysis.mode(primers)
    if (length(templates) == 0 || nrow(templates) == 0 ||
        !is(templates, "Templates")) {
        warning("No available templates for creating a report; returning.")
        return()
    }
    if (length(primers) == 0 || nrow(primers) == 0 || 
        !is(primers, "Primers")) {
        warning("No available primers for creating a report; returning.")
        return()
    }
    if (is.null(sample.name)) {
        sample.name <- primers$Run[1]
    }
    if (!is(settings, "DesignSettings") ) {
        stop("Settings are invalid.")
    }
    if (any(sapply(used.settings, function(x) !is(x, "DesignSettings")))) {
        stop("used.settings are invalid.")
    }
    if (!"primer_coverage" %in% colnames(primers)) {
        warning("Please compute primer coverage first.")
        return()
    }
    if (!requireNamespace("pander")) {
        warning("Cannot create the report without 'pander'.")
        return()
    }
    # create param list for comparison report
    params <- list("primers" = primers, "templates" = templates,
                "direction" = mode.directionality, "sample" = sample.name,
                "settings" = settings,
                "used_settings" = used.settings,
                "required_cvg" = required.cvg)
    render_report(params, "report.Rmd", fname)
    return()
})
#####
#' Creation of a PDF Report for Primer Comparison.
#' 
#' Creates a PDF report for comparing multiple primers.
#'
#' @param primers A list with evaluated \code{Primers} objects.
#' @param templates A list with \code{Templates} objects.
#' @param fname A character vector giving the file to store the report in.
#' @param settings A \code{DesignSettings} object.
#' @param sample.name An identifier for your analysis.
#' @param used.settings A named list (with fields "fw" and "rev") containing the forward/reverse used design settings.
#' @return Creates a PDF file giving a report on the comparison of the input primers.
#' @keywords internal
#' @note
#' Creating the report requires the external programs Pandoc (http://pandoc.org)
#' and LateX (http://latex-project.org).
setMethod("create_report", c("list", "list"), function(primers, templates, 
             fname, settings, 
             sample.name = NULL, 
             used.settings = NULL) {
    mode.directionality <- sapply(primers, function(x) get.analysis.mode(x))
    # check type of primer and template data
    if (length(primers) == 0 || length(templates) == 0) {
        stop("No primer/template data was inputted.")
    }
    template.classes <- sapply(templates, function(x) class(x))
    primer.classes <- sapply(primers, function(x) class(x))
    if (any(template.classes != "Templates") || any(primer.classes != "Primers")) {
        stop("Check types of primers/templates.")
    }
    if (is.null(sample.name)) {
        # assume templates relate to the same sample
        sample.name <- templates[[1]]$Run[1]
    }
    if (!is(settings, "DesignSettings") ) {
        stop("Settings are invalid.")
    }
    # cvg available/evaluated set?
    cvg.available <- sapply(primers, function(x) 
                        "primer_coverage" %in% colnames(x))

    if (!all(cvg.available)) {
        warning("Please compute primer coverage first.")
        return()
    }
    params <- list("primers" = primers, "templates" = templates,
                "direction" = mode.directionality, "sample" = sample.name,
                "settings" = settings)
    render_report(params, "report_comparison.Rmd", fname)
    return()
})

##########################
# Primer functionalities
##########################
#' Creation of k-mers of a single sequence.
#' @param seq A character vector.
#' @param k The size of the k-mer.
#' @return A names character vector, where the names 
#' are the relative positions of the k-mers and the values
#' give the character vector of the k-mer.
#' @keywords internal
create.kmer <- function(seq, k) {
    # generates all unique substrings of length k contained in the single sequence 'seq' 
    if (length(seq) != 1) {
        stop("This function is only for single sequences")
    }
    if (k > nchar(seq)) {
        k <- nchar(seq)
        warning("Specified k-mer size was larger than sequence length; ",
                "Reduced k-mer size to: ", k, ".")
    }
    k.mers <- unique(substring(seq, seq_len(nchar(seq) - k + 1), k:nchar(seq)))
    kmer.pos <- -seq(nchar(seq), k)
    names(k.mers) <- kmer.pos
    return(k.mers)
}
#' Creation of k-mers for multiple sequences.
#' @param seq A character vector.
#' @param k The size of the k-mer.
#' @return A list with named character vectors, containing the k-mers.
#' @keywords internal
create.k.mers <- function(seqs, k) {
    # generates all substrings of length k contained in the single sequence 'seq' 
    i <- NULL
    k.mers <- foreach (i = seq_along(seqs), .combine = c) %dopar% {
        k.mer <- list(create.kmer(seqs[i], k))
    }
    return(k.mers)
}
#' Estimation of Primer Coverage.
#'
#' Estimates the possible coverage of primers using
#' probes of size \code{k} and only considering perfect
#' matches without consideration of ambiguities.
#'
#' @param seqs A character vector of sequences to evaluate coverage for.
#' @param k A numeric giving the size of the primers.
#' @param id An optional identifier for the primers.
#' @return A data frame with binding information.
#' @keywords internal
estimate.cvg.dir <- function(seqs, k, id = "") {
    # 1. create all k-mers for input seqs
    kmers <- create.k.mers(seqs, k)
    # since we're creating before the target region, k mer pos is negative
    kmer.df <- do.call(rbind, lapply(seq_along(kmers), function(x) data.frame("Template_ID" = x, "Primer" = kmers[[x]], "Start" = as.numeric(names(kmers[[x]])), "End" = as.numeric(names(kmers[[x]])) + k)))
    # use ddply to determine binding positions / cvg
    binding.df <- ddply(kmer.df, "Primer", summarize, 
                    "Start" = paste(substitute(Start), collapse = ","), 
                    "End" = paste(substitute(End), collapse = ","),
                    "Covered_Seqs" = paste(substitute(Template_ID), collapse = ","),
                    "primer_coverage" = length(substitute(Template_ID)))
    binding.df$Coverage <- unlist(lapply(strsplit(binding.df$Start, split = ","), function(x) length(x)))
    binding.df$Coverage_Ratio <- binding.df$Coverage / length(seqs)
    return(binding.df)
}
#' Estimation of Primer Coverage.
#'
#' Estimates the possible coverage of primers using
#' probes of size \code{k} and only considering perfect
#' matches without consideration of ambiguities.
#'
#' @param seqs A character vector of sequences to evaluate coverage for.
#' @param k A numeric giving the size of the primers.
#' @param mode.directionality Estimation of coverage for forward/reverse/both?
#' @param sample An optional identifier for the sample.
#' @return A list with entries \code{fw} and \code{rev} giving
#' data frames for forward/reverse binding.
#' @keywords internal
estimate.cvg <- function(lex.df, k = 18, mode.directionality, sample = "") {
    cvg.fw <- NULL
    cvg.rev <- NULL
    if (mode.directionality == "fw") {
        # seqs <- lex.df$Allowed_fw
        cvg.fw <- estimate.cvg.dir(lex.df$Allowed_fw, k, id = paste0(sample, "_", "fw"))
    } else if (mode.directionality == "rev") {
        cvg.rev <- estimate.cvg.dir(lex.df$Allowed_rev, k, id = paste0(sample, "_", "rev"))
    } else {
        cvg.fw <- estimate.cvg.dir(lex.df$Allowed_fw, k, id = paste0(sample, "_", "fw"))
        cvg.rev <- estimate.cvg.dir(lex.df$Allowed_rev, k, id = paste0(sample, "_", "rev"))
    }
    return(list("fw" = cvg.fw, "rev" = cvg.rev))
}

#' @rdname PrimerDesign
#' @name PrimerDesign
#' @aliases classify_design_problem
#' @details
#' \code{classify_design_problem} determines the difficulty of a primer design 
#' task by estimating the distribution of coverage ratios per primer
#' by performing exact string matching with 
#' primers of length \code{primer.length}, which are constructed
#' by extracting template subsequences. Next, a beta distribution
#' is fitted to the estimated coverage distribution, which is
#' then compare to reference distributions representing
#' primer design problems of different difficulties via the
#' total variance distance. The difficulty of the input primer design
#' problem is found by selecting the class of the 
#' reference distributions that has the smallest distance
#' to the estimated coverage distribution.
#' An estimate of the required number of primers to reach a given
#' \code{required.cvg} can be computed by setting
#' \code{primer.estimate} to \code{TRUE}. Since this estimate
#' is based solely on perfect matching primers, the number of
#' primers that would actually be required is typically less.
#'
#' @return \code{classify_design_problem} returns a list with the 
#' following fields:
#' \describe{
#' \item{\code{Classification}}{The estimated difficulty of the primer design task.}
#' \item{\code{Class-Distances}}{The total variance distance of the fitted
#' beta distribution to the reference distribution.}
#' \item{\code{Confidence}}{The confidence in the estimate of the
#' design tasks' difficulty as based on the class distances.}
#' \item{\code{Uncertain}}{Whether the classification is highly uncertain, that is
#' low-confidence.}
#' \item{\code{Nbr_primers_fw} and \code{Nbr_primers_rev}}{The respective number of 
#' required forward and reverse primers if \code{primer.estimate} was set to \code{TRUE}.}
#' }
#' @export
classify_design_problem <- function(template.df, 
                                    mode.directionality = c("both", "fw", "rev"),
                                    primer.length = 18, 
                                    primer.estimate = FALSE,
                                    required.cvg = 1) {
    .Defunct("Unavailable due to removal of `distr` from CRAN")
}
greedy.primers <- function(binding.df, template.df, required.cvg = 1) {
    # lower bound on the number of required primers
    o <- order(binding.df$Coverage_Ratio, decreasing = TRUE)
    binding.df <- binding.df[o,]
    selected.primers <- NULL
    cur.cvg <- 0
    while ((cur.cvg / nrow(template.df)) < required.cvg && nrow(binding.df) > 0) {
        best.covered <- as.numeric(unlist(strsplit(binding.df[1, "Covered_Seqs"], split = ",")))
        if (length(best.covered) == 0) {
            break
        }
        selected.primers <- my_rbind(selected.primers, binding.df[1,])
        marginal.cvg.gain <- length(best.covered)
        cur.cvg <- cur.cvg + marginal.cvg.gain
        binding.df <- binding.df[-1, ]  # remove the selected primer and selection candidates that were filtered this round
        binding.df <- evaluate.diff.primer.cvg(binding.df, best.covered, template.df)  
    }
    return(selected.primers)
}
#' Validates a Primers Object.
#'
#' Checks whether a Primers object is valid or not.
#'
#' @param object An input data frame to be checked for being a primer data frame.
#' @return \code{TRUE}, if the object is valid, FALSE otherwise.
#' @keywords internal
validate_primers <-  function(object) {
    # specify minimal set of columns that should be present in a primer data frame:
    required.fields <- list("Identifier" = c("factor"), "ID" = "factor", 
                   "Forward" = "character", "Reverse" = "character",
                   "primer_length_fw" = c("integer", "numeric"), 
                   "primer_length_rev" = c("integer", "numeric"), 
                   "Direction" = "character",
                   "Run" = "character")
    possible.fields <- NULL # don't check for additional fields here
    if (!is(object, "data.frame")) {
        return("Input was no data frame.")
    }
    check.fields <- check_setting(possible.fields, object, required.fields)
    if (check.fields) {
        # Check that "Run" is unique
        if (length(unique(object$Run)) <= 1) {
            return(TRUE)
        } else {
            msg <- paste0("The 'Run' column may only contain a single unique value. Observed unique values were: ", paste0(unique(object$Run), collapse = ","))
            return(msg)
        }
    } else {
        return(check.fields)
    }
}

#' @rdname PrimerEval
#' @name PrimerEval
#' @aliases check_restriction_sites
#' @return \code{check_restriction_sites} returns a data frame 
#' with possible restriction sites found in the primers.
#' @references
#' Roberts, R.J., Vincze, T., Posfai, J., Macelis, D. (2010) REBASE–a database for DNA restriction
#' and modification: enzymes, genes and genomes. Nucl. Acids Res. 38: D234-D236. http://rebase.neb.com
#' @export
#' @examples
#' 
#' data(Ippolito)
#' # Check the first primer for restriction sites with respect to the first 10 templates
#' site.df <- check_restriction_sites(primer.df[1,], template.df[1:10])
check_restriction_sites <- function(primer.df, template.df, 
                            adapter.action = c("warn", "rm"), 
                            selected = NULL, only.confident.calls = TRUE,
                            updateProgress = NULL) {
    if (length(template.df) == 0 || nrow(template.df) == 0 || 
        !is(template.df, "Templates")) {
        stop("Need correct templates to identify restriction sites.") 
    }
    if (length(primer.df) == 0 || nrow(primer.df) == 0 || 
        !is(primer.df, "Primers")) {
            stop("Need valid primers.")
    }
    adapter.action <- match.arg(adapter.action)
    fw.idx <- which(nchar(primer.df$Forward) != 0)
    rev.idx <- which(nchar(primer.df$Reverse) != 0)
    fw.sites <- NULL
    rev.sites <- NULL
    if (length(fw.idx) != 0) {
        seqs <- DNAStringSet(primer.df$Forward[fw.idx])
        names(seqs) <- primer.df$ID[fw.idx]
        primer.seqs <- DNAStringSet(seqs)
        template.seqs <- template.df$Sequence
        names(template.seqs) <- template.df$Identifier
        template.seqs <- DNAStringSet(template.seqs)
        fw.sites <- check_restriction_sites_single(primer.seqs, template.seqs, 
                    adapter.action, 
                    direction = "fw",
                    only.confident.calls = only.confident.calls,
                    updateProgress = updateProgress)
    }
    if (length(rev.idx) != 0) {
        seqs <- primer.df$Reverse[rev.idx]
        names(seqs) <- primer.df$ID[rev.idx]
        primer.seqs <- DNAStringSet(seqs)
        template.seqs <- reverseComplement(template.seqs)
        rev.sites <- check_restriction_sites_single(primer.seqs, template.seqs, 
                    adapter.action, 
                    direction = "rev", 
                    only.confident.calls = only.confident.calls,
                    updateProgress = updateProgress)
    }
    site.df <- rbind(fw.sites, rev.sites)
    return(site.df)
}
#' Update of Primer Binding Regions.
#'
#' Updates the relative primer binding sites in the templates
#' when the template binding regions have changed since the last
#' coverage computation.
#'
#' @param primer.df A \code{Primers} data frame.
#' @param template.df Templates with the new binding regions.
#' @param old.template.df Templates with the old binding regions.
#' @return A \code{Primers} object with updated relative binding positions.
#' @keywords internal
update_primer_binding_regions <- function(primer.df, template.df, old.template.df) {
    if (length(primer.df) == 0 || nrow(primer.df) == 0) {
        return(primer.df)
    }
    if (!is(primer.df, "Primers")) {
        stop("Please input a 'Primers' object.")
    }
    if (!"primer_coverage" %in% colnames(primer.df)) {
        warning("Cannot update primer binding regions as no coverage was available.")
        return(primer.df)
    }
    if (!is(template.df, "Templates") || !is(old.template.df, "Templates")) {
        stop("Need to have 'Templates' objects to update the primer regions.")
    }
    # check that new region annotations are for the same set of templates
    if (nrow(template.df) != nrow(old.template.df) || template.df$ID != old.template.df$ID) {
        # template sets do not match -> don't adjust anything
        return(primer.df)
    }
     # update the current primers object in rv_primers
    fw.region <- list(template.df$Allowed_Start_fw, template.df$Allowed_End_fw)
    rev.region <- list(template.df$Allowed_Start_rev, template.df$Allowed_End_rev)
    fw.region.old <- list(old.template.df$Allowed_Start_fw, old.template.df$Allowed_End_fw)
    rev.region.old <- list(old.template.df$Allowed_Start_rev, old.template.df$Allowed_End_rev)
    # update the relative binding regions of the primers according to the changes from new to old regions
    # for fw: 'end' defines the relatie position of interest
    fw.diff <- fw.region[[2]] - fw.region.old[[2]]
    # for reverse: start defines the position of relative interest
    rev.diff <- rev.region[[1]] - rev.region.old[[1]]
    dirs <- c("_fw", "_rev")
    # cols with relation to forward binding region:
    relative.cols.fw <- unlist(lapply(dirs, function(x) paste0("Relative_Forward_Binding_Position_", c("Start", "End"), x)))
    # this won't work if we have paired primers -> need to know covered_seqs_fw and covered_seqs_rev in this case
    template.ids <- lapply(strsplit(primer.df$Covered_Seqs, split = ","), as.numeric)
    # assume template order hasn't changed:
    template.idx <- lapply(template.ids, function(x) match(x, template.df$Identifier))
    for (i in seq_along(relative.cols.fw)) {
        col <- relative.cols.fw[i]
        vals <- lapply(strsplit(primer.df[, col], split = ","), as.numeric)
        len.ok <- unlist(lapply(seq_along(vals), function(x) length(vals[[x]]) == length(template.idx[[x]])))
        if (any(!len.ok) && any(primer.df[,col] != "")) {
            warning("Can't adjust binding range for paired primers at the moment.")
            return(primer.df)
        }
        # for adjustment, need to get the template-specific difference
        adj.vals <- unlist(lapply(seq_along(vals), function(x) paste(vals[[x]] - fw.diff[template.idx[[x]]], collapse = ",")))
        primer.df[, col] <- adj.vals
    }
    relative.cols.rev <- unlist(lapply(dirs, function(x) paste0("Relative_Reverse_Binding_Position_", c("Start", "End"), x)))
     for (i in seq_along(relative.cols.rev)) {
        col <- relative.cols.rev[i]
        vals <- lapply(strsplit(primer.df[, col], split = ","), as.numeric)
        len.ok <- unlist(lapply(seq_along(vals), function(x) length(vals[[x]]) == length(template.idx[[x]])))
        if (any(!len.ok) && any(primer.df[,col] != "")) {
            warning("Can't adjust binding range for paired primers at the moment.")
            return(primer.df)
        }
        # for adjustment, need to get the template-specific difference
        adj.vals <- unlist(lapply(seq_along(vals), function(x) paste(vals[[x]] + rev.diff[template.idx[[x]]], collapse = ",")))
        primer.df[, col] <- adj.vals
    }
    return(primer.df)
}
#' Identification of Restriction Sites.
#'
#' Identifies restriction sites in a list with putative
#' restriction sites provided by \code{bad.regions} using
#' a data frame of restriction sites given by \code{DB}.

#' @param bad.region IRanges with possible adapter sites.
#' @param DB A data frame with restriction enzyme sites.
#' @return A boolean data frame indicating the presence
#' of adapters for all considered restriction sites.
#' @keywords internal
restriction_hits <- function(bad.regions, DB) {
    bad.seqs <- unlist(lapply(bad.regions, function(x) as.character(unlist(x))))
    names(bad.seqs) <- unlist(lapply(seq_along(bad.regions), function(x) rep(names(bad.regions)[x], length(unlist(bad.regions[[x]])))))
    i <- NULL
    hit.db <- foreach (i = seq_len(nrow(DB)), .combine = cbind) %dopar% {
        enzyme <- as.character(DB$nam[i])
        temp <- vmatchPattern(DB$rep[i], bad.seqs, 
                max.mismatch = 0, with.indels = FALSE)
        hits <- sapply(temp, function(x) length(x) != 0)
        if (length(hits) == 0) {
            NULL
        } else {
            hit.idx <- which(hits)
            result <- data.frame("Hit" = hits, stringsAsFactors = FALSE) 
            colnames(result)[colnames(result) == "Hit"] <- enzyme
            result
        }
    }
    if (length(hit.db) != 0) {
        attr(hit.db, "ID") <- names(bad.seqs)
    }
    return(hit.db)
}

#' Identification of Badly Fitting Regions.
#'
#' Identify regions in the templates where the primers are
#' not very complementary. These regions indicate possible
#' restriction enzyme adapters.
#'
#' @param primer.seqs Primer sequences.
#' @param template.seqs Template sequences.
#' @param search.hits Template substrings that agree well
#' with the input primers.
#' @return A list with putative restriction sites for every primer.
#' @keywords internal
restriction_ali <- function(primer.seqs, template.seqs, search.hits) {
    bad.regions <- vector("list", length(search.hits))
    # define substitution matrix for alignment
    mat <- nucleotideSubstitutionMatrix(match = 1, mismatch = 0, 
                                                    baseOnly = FALSE, type = "DNA")
    for (i in seq_along(search.hits)) {
        # align primer with hit region in the templates
        if (is.na(search.hits[i])) {
            next
        }
        ali <- pairwiseAlignment(primer.seqs[i], search.hits[i],
                    substitutionMatrix = mat,
                     gapOpening = -3, gapExtension = -1, type = "global-local") # open up gaps directly
        # identify indels and non-matching regions
        # minimal number of bases of restriction 
        restrict.size <- 3
        if (length(ali@subject@indel[[1]]) != 0) {
            # there were indels -> look at the indel regions
            indel <- ali@subject@indel[[1]]
            # only select indels larger than cutoff
            sel <- indel@width >= restrict.size
            indel <- indel[sel]
            if (length(indel) != 0) {
                site <- extractAt(primer.seqs[i], indel)
                bad.regions[[i]] <- unname(site)
            }
        } else {
            # no indels -> look at the longest mismatch region
            mm.region <- unlist(mismatch(ali@pattern))
            # discard short mismatch regions:
            if (length(mm.region) > restrict.size) { 
                # more than 3 mismatches -> consider as adapter seq
                site <- extractAt(primer.seqs[i], 
                        IRanges(min(mm.region), max(mm.region)))
                bad.regions[[i]] <- unname(site)
            }
        }
    }
    names(bad.regions) <- names(primer.seqs)
    return(bad.regions)
}
#' Identification of Sequence Matches.
#'
#' Determines the most similar template sequence for every
#' input primer sequence. Used to identify regions for
#' alignment for the identification of restriction sites.
#'
#' @param primer.seqs Primer sequences.
#' @param template.seqs Template sequences.
#' @return A vector with the template regions matching the 
#' \code{primer.seqs} best.
#' @keywords internal
restriction_match <- function(primer.seqs, template.seqs) {
    search.hits <- rep(NA, length(primer.seqs))
    for (i in seq_along(primer.seqs)) {
        scores <- rep(Inf, length(template.seqs))
        best.ali <- NA
        for (j in seq_len(length(template.seqs))) { # n^2 * n^2 = n^4 performance ...
            # vmatchpattern doesn't support indels ...
            # max.mismatch: gives the detection limit for adapter sequences/overhang extent
            d <- matchPattern(as.character(primer.seqs[i]), 
                    as.character(template.seqs[j]), 
                    with.indels = TRUE, 
                    max.mismatch = 12)
            if (length(d@ranges) == 0) {
                # no hits
                next
            }
            # retrieve hits in the templates
            hits <- sapply(d, function(x) as.character(x))
            # determine best hit by computing distance between region of interest and pattern
            d <- stringdist::stringdist(hits, as.character(primer.seqs[i]))
            # select the hit with the smallest edit distance
            sel <- which.min(d)
            scores[j] <- d[sel]
            hit <- hits[sel]
            if (scores[j] <= min(scores)) {
                best.ali <- hit
            }
        }
        search.hits[i] <- best.ali
    }
    return(search.hits)
}
#' Identification of Sequence Restriction Sites.
#'
#' Checks the input sequences \code{seqs} for the presence of
#' restriction sites. By removing the restriction sites from a primer set, 
#' it is possible to identify the coverage of the primers
#' (e.g. using \code{\link{check_constraints}}) discounting for 
#' the impact of the mismatching bases caused by the insert.
#'
#' @param primer.seqs Nucleotide sequences of primers to be checked for restriction sites in terms of a \code{DNAStringSet} object.
#' @param template.seqs A \code{DNAStringSet} object with nucleotide sequences 
#' containing the templates corresponding to \code{seqs}.
#' @param adapter.action The action to be performed when adapter sequences
#' are found. Either "warn" to issue warning about adapter sequences or
#' "rm" to remove identified adapter sequences.
#' @param The primer direction that is checked.
#' @param selected Names of restriction sites that are to be checked.
#' By default \code{selected} is \code{NULL} in which case all REBASE 
#' restriction sites are checked.
#' @param only.confident.calls Only output confident calls of restriction sites.
#' @param updateProgress A Shiny progress callback function.
#' @return A data frame with restriction sites, if any could be found.
#' @references
#' Roberts, R.J., Vincze, T., Posfai, J., Macelis, D. (2010) REBASE–a database for DNA restriction
#' and modification: enzymes, genes and genomes. Nucl. Acids Res. 38: D234-D236. http://rebase.neb.com
#' @keywords internal
check_restriction_sites_single <- function(primer.seqs, template.seqs, adapter.action, 
                            direction = c("fw", "rev"),
                            selected = NULL, 
                            only.confident.calls = TRUE,
                            
                            updateProgress = NULL) {

    if (length(template.seqs) == 0 || length(primer.seqs) == 0) {
        return(NULL)
    }
    if (!is(primer.seqs, "DNAStringSet")) {
        stop("Primers should be a DNAStringSet.")
    }
    if (!is(template.seqs, "DNAStringSet")) {
        stop("Primers should be a DNAStringSet.")
    }
    if (length(direction) == 0) {
        stop("Please provide the 'direction' argument.")
    }
    direction <- match.arg(direction)
    #############
    # n.b.: this approach only works if the primer makes sense at all
    # if the primer isn't really complementary to the templates, 
    # we also won't be able to find restriction sites.
    # sysdata is available within the package automatically: enzdata
    #############
    # use only restriction sites without degeneracies for simplicity..
    counts <- sapply(regmatches(enzdata$site, gregexpr("[a-z]", enzdata$site, perl=TRUE)), length)
    sel <- which(counts == 0)
    DB <- enzdata[sel,]
    if (!is.null(selected)) {
        m <- match(selected, DB$nam)
        if (any(is.na(m))) {
            msg <- paste("Selected enzyme not available: ",
                paste(selected[is.na(m)], collapse = ", "), sep = "")
            stop(msg)
        }
        DB <- DB[m,]
    }
    if (is.function(updateProgress)) {
        detail <- "String matching"
        updateProgress(1/3, detail, "inc")
    }
    ##########
    # Confident calls: check for mismatches in the templates
    ##########
    # 1. Identify matches of primers in templates
    search.hits <- restriction_match(primer.seqs, template.seqs)
    if (is.function(updateProgress)) {
        detail <- "Aligning"
        updateProgress(1/3, detail, "inc")
    }
    # 2. Find the non-matching regions in the templates
    bad.regions <- restriction_ali(primer.seqs, template.seqs, search.hits)
    #####
    # Non-confident call: input primers
    #######
    input.regions <- vector("list", length(primer.seqs))
    names(input.regions) <- names(bad.regions)
    for (i in seq_along(primer.seqs)) {
        input.regions[[i]] <- DNAStringSetList(as.character(primer.seqs[i]))
    }
    all.regions <- list("Confident" = bad.regions, "Unconfident" = input.regions)
    if (is.function(updateProgress)) {
        detail <- "Identifying adapters"
        updateProgress(1/3, detail, "inc")
    }
    # 3. Identify whether the non-matching regions are restriction sites
    all.hits <- vector("list", length(all.regions))
    for (j in seq_along(all.regions)) {
        bad.regions <- all.regions[[j]]
        hit.db <- restriction_hits(bad.regions, DB) 
        if (is.null(hit.db)) {
            next
        }
        # 4. from all hits, only consider the common restriction sites
        common.DB <- DB[DB$common,]
        idx <- lapply(seq_len(nrow(hit.db)), function(x) which(as.logical(hit.db[x,])))
        hit.names <- lapply(idx, function(x) colnames(hit.db)[x])
        found.DB.idx <- lapply(seq_along(hit.names), function(x) idx[[x]][which(hit.names[[x]] %in% common.DB$nam)])
        no.hit.idx <- intersect(which(sapply(found.DB.idx, length) == 0), which(sapply(idx, length) != 0))
        if (length(no.hit.idx) != 0) {
            # no hit in the common restriction sites -> check for the most specific rare restriction site
            found.DB.idx[no.hit.idx] <- lapply(seq_along(no.hit.idx), function(x) idx[[no.hit.idx[x]]][which.max(nchar(DB$rep[idx[[no.hit.idx[x]]]]))])
        }
        if (length(found.DB.idx) == 0) {
            # still no hits? we're done
            next
        }
        names(found.DB.idx) <- attr(hit.db, "ID")
        hit.out <- vector("list", length(found.DB.idx))
        for (i in seq_along(found.DB.idx)) {
            if (length(found.DB.idx[[i]]) == 0) {
                # no hits found
                next
            }
            p.id <- names(found.DB.idx)[i]
            p.idx <- match(p.id, names(primer.seqs@ranges))
            out <- data.frame("Primer_Identifier" = p.id,
                              "Direction" = direction,
                              "Sequence" = unname(tolower(as.character(primer.seqs[p.idx]))),
                        "Enzyme" = as.character(DB$nam)[found.DB.idx[[i]]],
                        "Site" = tolower(DB$rep[found.DB.idx[[i]]]),
                        stringsAsFactors = FALSE)
            hit.out[[i]] <- out
        }
        hit.out <- do.call(rbind, hit.out)
        if (length(hit.out) != 0) {
            hit.out$Confidence <- names(all.regions)[j]
        }
        all.hits[[j]] <- hit.out
    }
    hit.out <- do.call(rbind, all.hits)
    # only return unique hits per primer
    hit.out <- ddply(hit.out, c("Primer_Identifier", "Sequence", "Enzyme", "Site"),
            function(x) arrange(x, substitute(Confidence))[1, ])
    if (only.confident.calls) {
        hit.out <- hit.out[hit.out$Confidence == "Confident",]
    }
    # warn about restriction sites:
    if (adapter.action == "warn") {
        # warn about restriction sites
        enzymes <- ddply(hit.out, c("Enzyme", "Site"), summarize, 
                    Identifiers = paste(substitute(Primer_Identifier), collapse = ","))
        if (nrow(enzymes) != 0) {
            out <- sapply(seq_len(nrow(enzymes)), function(x) 
                    paste("Found ", enzymes[x,"Enzyme"], " (", enzymes[x,"Site"], 
                    ") adapter in primers ", enzymes[x,"Identifiers"], 
                    ". Please check your sequences.", sep = ""))
            warning(paste(out, collapse = "\n"))
        }
    } else if (adapter.action == "rm") {
        # remove restriction sites from primers
        warning("Removal is not implemented yet.")
    }
    return(hit.out)
}
#' @rdname Input
#' @name Input
#' @aliases Primers-class
#' @section Basic columns:
#' In the following you can find a description of the most
#' important columns that can be found in objects of class \code{Primers}. 
#' Note that angular brackets indicate the existence of multiple possibilities.
#' The following columns are present when a set of primers
#' is loaded from a FASTA file using \code{\link{read_primers}}:
#' \describe{
#' \item{\code{ID}}{The identifiers of the primers.}
#' \item{\code{Identifier}}{The internal identifiers of the primers.}
#' \item{\code{Forward}}{The sequences of forward primers.}
#' \item{\code{Reverse}}{The sequences of reverse primers.}
#' \item{\code{primer_length<fw|rev>}}{The lengths of
#' forward and reverse primer sequences, respectively.}
#' \item{\code{Direction}}{Either 'fw' for forward primers,
#' 'rev' for reverse primers, or 'both' for a primer pair.}
#' \item{\code{Degeneracy_<fw|rev>}}{The degeneracy (ambiguity) of
#' forward and reverse primers, respectively.}
#' \item{\code{Run}}{An identifier describing the primer set.}
#' }
#'
#' @section Coverage-related columns:
#' The following columns are only available in an object of
#' class \code{Primers} after primer coverage
#' has been computed, that is after \code{\link{check_constraints}}
#' has been called with the active \code{primer_coverage} constraint. Computed coverage
#' values relating solely to string matching are indicated by the prefix
#' \code{Basic_}, while columns without this prefix relate to the coverage after
#' applying the constraints formulated via \code{CoverageConstraints}.
#' Information on off-target coverage events are indicated by
#' the \code{Off_} prefix, while on-target coverage events do not carry
#' this prefix.
#' 
#' \describe{
#' \item{\code{primer_coverage}}{The number of templates that are
#' covered by the primers. Note that if a primer set contains
#' primers of both directions, a template is only considered covered
#' if it is covered by primers of both directions.}
#' \item{\code{Coverage_Ratio}}{The ratio of templates that are covered by the primers.}
#' \item{\code{Binding_Position_Start_<fw|rev>}}{The upstream position in the 
#' templates where forward and reverse primers respectively bind.}
#' \item{\code{Binding_Position_End_<fw|rev>}}{The downstream position in the templates where forward and reverse primers respectively bind.}
#' \item{\code{Relative_<Forward|Reverse>_Binding_Position_<Start|End>_<fw|rev>}}{
#' The binding upstream (\code{Start}) or downstream (\code{End}) positions 
#' of the primers relative to the forward (\code{Forward})
#' or reverse (\code{Reverse}) binding regions, either for 
#' forward (\code{fw}) or reverse primers (\code{rev}).}
#' \item{\code{Binding_Region_Allowed}}{Whether a coverage event
#' occurred in the target binding region or not. If the allowed
#' off-target ratio was set to 0 only coverage events within the 
#' the target region are reported.}
#' \item{\code{Nbr_of_mismatches_<fw|rev>}}{The number of mismatches
#' of forward and reverse primer coverage events, respectively.}
#' \item{\code{Mismatch_pos_<fw|rev>}}{The position of mismatches
#' for forward and reverse coverage events, respectively. Mismatch
#' positions are reported relative to the 3' end, that is, position
#' 1 indicates a mismatch in the last base of a primer.}
#' \item{\code{primer_specificity}}{The specificity of a primer
#' as determined by its ratio of off-target binding events.}
#' }
#' 
#' @section Constraint-related columns:
#' Each constraint that is considered when calling \code{\link{check_constraints}}
#' gives rise to at least one column in the provided \code{Primers} object.
#' Due to the large number of possible constraints, we will limit our description
#' to the \code{gc_clamp} constraint. Once the GC clamp property has been computed,
#' the \code{gc_clamp_fw} column contains the length of the GC clamp for forward 
#' primers and \code{gc_clamp_rev} the corresponding length for reverse primers.
#' Whether the desired extent of the GC clamp was obtained by a primer
#' is indicated by the \code{EVAL_gc_clamp} column. It contains \code{TRUE} when
#' the GC clamp constraint was fulfilled and \code{FALSE} when it was broken.
#' To identify whether all required constraints were fulfilled by a primer,
#' the \code{constraints_passed} column can be used. It contains \code{TRUE}
#' if all \code{active.constraints} used by \code{\link{check_constraints}} were fulfilled
#' and \code{FALSE} otherwise.
#'
#' @return The \code{Primers} constructor returns an object of 
#' class \code{Primers}.
#' @exportClass Primers
#' @export
#' @examples
#' 
#' # Load a set of primers
#' primer.location <- system.file("extdata", "IMGT_data", "primers", "IGHV", 
#'                      "Ippolito2012.fasta", package = "openPrimeR")
#' primer.df <- read_primers(primer.location, "_fw", "_rev")
setClass("Primers",
   contains = c("data.frame"), validity = validate_primers)


setMethod("initialize", "Primers",
    function(.Object, ...) {
        # just call data frame constructor
        Object <- callNextMethod(.Object, ...)
    }
) 
#' @rdname Input
#' @name Input
#' @aliases Primers-class
#' @export
Primers <- function(...) new("Primers", ...) 

my_format_df <- function (x) {
    if (is.atomic(x) && !is.null(x)) {
        stop("Internal structure doesn't seem to be a list. Possibly corrupt data.frame.")
    }
    classes <- unlist(lapply(x, class))
    factor.idx <- which(classes == "factor")
    for (i in factor.idx) {
        x[, i] <- as.character(x[, i])
    }
    char.trunc <- function(x, trunc.char = 20) {
        trunc.char = max(0L, suppressWarnings(as.integer(trunc.char[1L])), na.rm=TRUE)
        if (!is.character(x) || trunc.char <= 0L) return(x)
        idx = which(nchar(x) > trunc.char)
        x[idx] = paste(substr(x[idx], 1L, as.integer(trunc.char)), "...", sep="")
        x
    }
    # truncate all strings
    classes <- unlist(lapply(x, class))
    string.idx <- which(classes == "character")
    for (i in string.idx) {
        x[, i] <- char.trunc(x[,i])
    }
    return(x)
}

my_show_df <- function(x, topn = 3,
                       nrows = 6) {
    # topn  - print the top topn and bottom topn rows with '---' inbetween 
    # nrows - if x is <= nrows, the whle table is printed
    if (!is.numeric(nrows)) nrows = 10
    if (!is.infinite(nrows)) nrows = as.integer(nrows)
    if (nrows <= 0L) return(invisible())   # ability to turn off printing
    if (!is.numeric(topn)) topn = 5
    topnmiss = missing(topn)
    topn = max(as.integer(topn),1L)
    if (nrow(x) == 0L) {
        if (length(x)==0L)
           cat(paste0("Empty object of class '", class(x), "'.\n"))
        else
           cat(paste0("Object of class '", class(x), "' without any rows.\n"))
        return(invisible())
    }
    if (topn*2<nrow(x) && (nrow(x)>nrows || !topnmiss)) {
        toprint = rbind(head(x, topn), tail(x, topn))
        rn = c(seq_len(topn), seq.int(to=nrow(x), length.out=topn))
        printdots = TRUE
    } else {
        toprint = x
        rn = seq_len(nrow(x))
        printdots = FALSE
    }
    toprint = my_format_df(x) # modify data frame to have only character classes 
    rownames(toprint) <- seq_len(nrow(toprint))
    if (is.null(names(x)) | all(names(x) == "")) colnames(toprint)=rep("", ncol(toprint))
    if (printdots) {
        toprint = rbind(head(toprint,topn),"---" = "",tail(toprint,topn))
        rownames(toprint) = format(rownames(toprint),justify="right")
        print(toprint,right=TRUE)
        return(invisible())
    }
    if (nrow(toprint)>20L)
        # repeat colnames at the bottom if more than 20 rows are printed
        toprint=rbind(toprint,matrix(colnames(toprint),nrow=1))
    print(toprint,right=TRUE)
    invisible()
}
setMethod("show", "Primers", function(object) {
    # overwrite the 'print' function in order to limit the output
    my_show_df(asS3(object))
})
setMethod("summary", "Primers", function(object) {
    stats <- primer.set.parameter.stats(object, get.analysis.mode(object), NULL)
    return(stats)
})

#' cbind for Primers class.
#'
#' Ensures that the cbind result has the appropriate class.
#'
#' @param ... Parameters for cbind function.
#' @return Column binded Primers data frame.
#' @keywords internal
#' @export
#' @examples
#' data(Ippolito)
#' primer.df <- cbind(primer.df, primer.df)
cbind.Primers <- function(...) {
    df <- cbind.data.frame(...)
    df <- Primers(df)
    return(df)
}
#' S4 cbind for Primers.
#'
#' S4 cbind function for Primers.
#'
#' @export
#' @rdname Primers-method
#' @param x The Primers data frame.
#' @param y Another data frame.
#' @return Cbinded primer data frame.
#' @keywords internal
#' @examples
#' data(Ippolito)
#' primer.df <- cbind2(primer.df, seq_len(nrow(primer.df)))
setMethod("cbind2", c(x = "Primers", y = "ANY"),
    # need this in case of S3 dispatch fails.
    function(x, y, ...) {
        df <- cbind.data.frame(x, y, ...) # packages should not call .Internal()?
        df <- Primers(df)
        return(df)
    }
)
#' S4 rbind for Primers.
#'
#' S4 rbind function for Primers.
#'
#' @export
#' @return Rbinded primer data frame.
#' @rdname Primers-method
#' @keywords internal
setMethod("rbind2", c(x = "Primers", y = "ANY"),
    # need this in case of S3 dispatch fails.
    function(x, y, ...) {
        df <- rbind.data.frame(x, y, ...)
        df <- Primers(df)
        return(df)
    }
)
#' rbind for Primers class.
#'
#' Ensures that the rbind result has the appropriate class.
#'
#' @param ... Parameters for rbind function.
#' @return Row-binded Primers data frame.
#' @keywords internal
#' @export
#' @examples
#' data(Ippolito)
#' primer.df <- rbind(primer.df, primer.df)
rbind.Primers <- function(...) {
    df <- rbind.data.frame(...)
    df <- Primers(df)
    return(df)
}

#' Slicing Operator for Primers.
#'
#' Slices a Primers data frame.
#'
#' @param x The Primers data frame.
#' @param i The row index.
#' @param j The column index.
#' @param ... Other arguments to the slice operator.
#' @param drop Simplify data frame?
#' @exportMethod [
#' @rdname Primers-method
#' @return Subset of primer data frame.
#' @keywords internal
#' @aliases [,Primers-method [,Primers,ANY-method
#' @examples
#' data(Ippolito)
#' primer.df <- primer.df[1:2,]
setMethod("[", c("Primers", "ANY", "ANY"),
    function(x, i, j, ..., drop = TRUE) {
        if (missing(drop)) {
            df <- asS3(x)[i, j, ...]
        } else {
            df <- asS3(x)[i, j, ..., drop = drop]
        }
        if (is(df, "data.frame")) {
            # only set my class if we haven't simplified.
            #options("show.error.messages" = FALSE)
            p.df <- suppressWarnings(try(Primers(df), silent = TRUE))
            #options("show.error.messages" = TRUE)
            if (is(p.df, "try-error")) {
                # removed crucial columns -> turn into data frame
                p.df <- asS3(df)
            } 
            df <- p.df
        }
        return(df)
    }
)
#' Dollar Operator for Primers.
#'
#' Stores data in a column of a Primers data frame.
#'
#' @exportMethod $<-
#' @rdname Primers-method
#' @param name The name of the column.
#' @param value The values of the column.
#' @return Primer data frame with replaced column.
#' @keywords internal
#' @examples
#' data(Ippolito)
#' primer.df$Forward[1] <- "ctagcgggaccg"
setMethod("$<-", "Primers", 
    function(x, name, value) {
        df <- asS3(x)
        eval(parse(text = paste("df$", name, " <- value", sep = "")))
        df <- Primers(df)
        return(df)
    }
)
#' Updates the Primer Coverage.
#'
#' Updates the most important columns in a primer data frame according to the selected
#' coverage definition. Only coverage events with less or equal than the allowed number of
#' mismatches according to the selected coverage definition will be retained.
#'
#' @param primer.df A \code{Primers} object.
#' @param template.df A \code{Templates} object.
#' @param allowed.mismatches A numeric giving the maximal number of allowed.mismatches.
#' @param cvg.definition The definition of coverage to be used, either "constrained" or "basic".
#' @return A primer data frame with modified coverage information.
#' @keywords internal
update_primer_cvg <- function(primer.df, template.df, allowed.mismatches, cvg.definition = c("constrained", "basic")) {
    # note: only updates primer_coverage, covered_seqs, not binding positions etc.
    # Consider only coverage events with <= allowed.mismatches and according to the provided cvg.definition
    if (length(primer.df) == 0 || nrow(primer.df) == 0) {
        # nothing to update ..
        return(primer.df)
    }
    mode.directionality <- get.analysis.mode(primer.df)
    cvg.definition <- match.arg(cvg.definition)
    mm.info <- prepare_mm_plot(primer.df, template.df)
    # select events according to selected cvg.definition and number of mismatches only
    df <- mm.info[mm.info$Coverage_Type == cvg.definition & mm.info$Number_of_mismatches <= allowed.mismatches, ]
    ##################
    # just move from mismatch rep to coverage rep
    ddf <-  ddply(df, c("Primer", "Direction", "Template", "Group"), summarize,
                            Position = unique(substitute(Position_3terminus)), 
                            Number_of_mismatches = unique(substitute(Number_of_mismatches)))
    #####
    if (mode.directionality == "both") {
        # remove individual binding events
        # for pairs of primers: presence of any other other direction is fine for cvg
        dir.count.both <- ddply(ddf, c("Template", "Group"), summarize, DirectionCount = length(unique(substitute(Direction))))
        rm.idx <- which(dir.count.both$DirectionCount <= 1)
        if (length(rm.idx) != 0) {
            m <- match(dir.count.both$Template[rm.idx], ddf$Template)
            ddf <- ddf[-m,]
        }
    }
    # for primers of both directions: they were replicated (fw+both) earlier, need to select events were pairs are present and then select the individual primer once again
    both.sel <- which(primer.df$Direction[match(ddf$Primer, primer.df$ID)] == "both")
    dff.both <- ddply(ddf[both.sel,], c("Primer", "Template"),
        summarize, 
        Group = unique(substitute(Group)))
    dff.single <- ddf[setdiff(seq_len(nrow(ddf)), both.sel), c("Primer", "Template", "Group")]
    dff <- rbind(dff.both, dff.single)
    ##########
    cvd <- ddply(dff, "Primer", summarize, Covered_Seqs = paste(substitute(Template), collapse = ","))
    m <- match(cvd$Primer, primer.df$ID)
    new.df <- data.frame(ID = primer.df$ID, Covered_Seqs = "", primer_coverage = 0,
                         stringsAsFactors = FALSE)
    if (length(m) != 0) { # we found coverage events for our selection
        # set corresponding entries in new.df to updated values
        cvd.identifier <- unlist(lapply(strsplit(cvd$Covered_Seqs, split = ","), function(x) paste(template.df$Identifier[match(x, template.df$ID)], collapse = ",")))
        new.df[m, 'Covered_Seqs'] <- cvd.identifier
        new.df[m, 'primer_coverage'] <- sapply(strsplit(cvd.identifier, split = ","), length)
    }
    # update primer data frame with new values
    new.primer.df <- primer.df
    new.primer.df[, colnames(new.df)] <- new.df
    return(new.primer.df)
}

#' @rdname Plots
#' @name Plots
#' @aliases plot_primer_binding_regions
#' @return \code{plot_primer_binding_regions} returns a plot of the primer binding regions.
#' @export
#' @include templates.R
#' @examples
#' 
#' # Primer binding regions of a single primer set
#' data(Ippolito)
#' p <- plot_primer_binding_regions(primer.df, template.df)
#' # Primer binding regions of multiple primer sets
#' data(Comparison)
#' p.comp <- plot_primer_binding_regions(primer.data[1:3], template.data[1:3])
setGeneric("plot_primer_binding_regions", 
    function(primers, templates, direction = c("both", "fw", "rev"), 
        group = NULL, relation = c("fw", "rev"), 
        region.names = c("Binding region", "Amplification region"), ...) {
        standardGeneric("plot_primer_binding_regions")
})
#' Plot of Primer Binding Regions for a Single Primer Set.
#'
#' Plots the primer binding regions in the templates.
#'
#' @param primers An object of class \code{Primers} with annotated
#' primer coverage.
#' @param templates An object of class \code{Templates} providing the
#' template sequences corresponding to \code{primers}. 
#' @param direction Primer direction
#' @param group The template groups for which binding events should
#' be determined. By default, \code{group} is set to \code{NULL} such that
#' all templates are considered.
#' @param relation A character vector specifying whether binding region data shall
#' be plotted relative to the forward (\code{fw}) or reverse (\code{rev}) 
#' target binding regions.
#' @param region.names Names for the primer binding region and the amplified region.
#' @return A histogram of primer binding regions.
#' @keywords internal
setMethod("plot_primer_binding_regions", 
    signature(primers = "Primers", templates = "Templates"),
    function(primers, templates,
        direction = c("both", "fw", "rev"), group = NULL, 
        relation = c("fw", "rev"), 
        region.names = c("Binding region", "Amplification region")) {
        
        relation <- match.arg(relation)
        direction <- match.arg(direction)
        if (length(primers) == 0 && nrow(primers) == 0) {
            return(NULL)
        }
        if (!is(primers, "Primers")) {
            stop("Please input a valid primers object.")
        }
        primer.data <- list(primers)
        names(primer.data) <- unique(primers$Run)
        p <- plot_primer_binding_regions(primer.data, list(templates),
                direction, group, relation, region.names = region.names)
        return(p)
})
create_region_boxes <- function(primers, templates, relation, region.names, ymin, ymax, xmax) {
    regions <- vector("list", length(templates)) # binding regions
    if (relation == "fw") {
        for (i in seq_along(templates)) {
            template.df <- templates[[i]]
            region.extents <- template.df$Allowed_End_fw_initial - template.df$Allowed_Start_fw_initial
            idx <- which.max(region.extents)
            s <- template.df$Allowed_Start_fw_initial[idx]
            e <- template.df$Allowed_End_fw_initial[idx]
            e.now <- template.df$Allowed_End_fw[idx]
            s.now <- template.df$Allowed_Start_fw[idx]
            delta <- e.now - e
            rel.binding.start <- -(e.now - s.now) - 1
            rect.x.leader <- -((e-s) + delta) - 1
            rect.x.leader.end <- rect.x.leader + e - s
            region.info.l <- data.frame("Region" = region.names[1], "xmin" = rect.x.leader, "xmax" = rect.x.leader.end, "ymin" = ymin, "ymax" = ymax)
            region.info.t <- data.frame("Region" = region.names[2], "xmin" = rect.x.leader.end + 1, "xmax" = xmax, "ymin" = ymin, "ymax" = ymax)
            region.df <- rbind(region.info.l, region.info.t)
            region.df$Run <- unique(primers[i]$Run)
            region.df$RelStartPosition <- rel.binding.start
            regions[[i]] <- region.df 
            #binding.starts[[i]] <- data.frame("Run" = names(primer.data)[i], "Position" = rel.binding.start)
        }
    } else {
        for (i in seq_along(templates)) {
            template.df <- templates[[i]]
            region.extents <- template.df$Allowed_End_rev_initial - template.df$Allowed_Start_rev_initial
            idx <- which.max(region.extents)
            seq.len <- nchar(template.df$Sequence[idx])
            s <- seq.len - template.df$Allowed_Start_rev_initial[idx] + 1
            e <- seq.len - template.df$Allowed_End_rev_initial[idx] + 1
            e.now <- seq.len - template.df$Allowed_End_rev[idx] + 1
            s.now <- seq.len - template.df$Allowed_Start_rev[idx] + 1
            # positive delta: binding region has to be shifted back (negative), negative delta: binding region has to be shifted forward (positive)
            delta <- s - s.now
            rel.binding.start <- e.now - s.now - 1
            ########
             rect.x.leader <- -((e-s) + delta) - 1
            rect.x.leader.end <- rect.x.leader + e - s
            #########
            rect.x.leader <- -(s - e + delta) - 1
            rect.x.leader.end <- rect.x.leader - e + s
            region.info.l <- data.frame("Region" = region.names[1], "xmin" = rect.x.leader, "xmax" = rect.x.leader.end, "ymin" = ymin, "ymax" = ymax)
            region.info.t <- data.frame("Region" = region.names[2], "xmin" = rect.x.leader.end + 1, "xmax" = xmax, "ymin" = ymin, "ymax" = ymax)
            region.df <- rbind(region.info.l, region.info.t)
            region.df$Run <- unique(primers[i]$Run)
            region.df$RelStartPosition = rel.binding.start
            regions[[i]] <- region.df 
            #binding.starts[[i]] <- data.frame("Run" = names(primer.data)[i], "RelStartPosition" = rel.binding.start)
        }
    }
    region.df <- do.call(rbind, regions)
    return(region.df)
}
#' Plot of Primer Binding Regions for Multiple Sets.
#'
#' Plots the primer binding regions for every primer set.
#'
#' @param primers List with primer data frames.
#' @param templates List with template data frames.
#' @param direction Direction of primers.
#' @param group Template groups to plot.
#' This defaults to plotting all groups.
#' @param relation Plot binding region relative to forward binding region or reverse?
#' @param region.names Names for the primer binding region and the amplified region.
#' @param highlight.set Primer sets to highlight in the plot.
#' @return A plot for primer binding region comparison.
#' @keywords internal
setMethod("plot_primer_binding_regions", 
    signature(primers = "list", templates = "list"),
    function(primers, templates,
        direction = c("both", "fw", "rev"), group = NULL, 
        relation = c("fw", "rev"), 
        region.names = c("Binding region", "Amplification region"), 
        highlight.set = NULL) {

    if (length(region.names) != 2) {
        stop("Need 2 region names.")
    }
    xlab <- "Location"
    direction <- match.arg(direction)
    relation <- match.arg(relation)
    xlab <- "Binding position"
    plot.data <- lapply(seq_along(primers), function(x) primer.binding.regions.data(primers[[x]], 
        templates[[x]], direction, group, relation))
    # annotate with run info
    runs <- get.run.names(primers)
    plot.data <- lapply(seq_along(plot.data), function(x) if (length(plot.data[[x]]) != 
        0 && nrow(plot.data[[x]]) != 0) {
        cbind(plot.data[[x]], Run = rep(primers[[x]]$Run[1], nrow(plot.data[[x]])))
    })
    plot.df <- do.call(rbind, plot.data)
    if (!is(plot.df, "data.frame")) {
        return(NULL)
    }
    if (length(highlight.set) != 0) {
        # check whether highlight set is specified correctly
        m <- match(highlight.set, plot.df$Run)
        na.idx <- which(is.na(m))
        if (length(na.idx) != 0) {
            msg <- paste("Highlight set not available in data:",
                paste(highlight.set[na.idx], collapse = ","))
            warning(msg)
            highlight.set <- highlight.set[!is.na(m)]
        }
    }
    # new annotation function for rectangle plot
    plot.df <- ddply(plot.df, c("ID", "Start", "End", "Run"), summarize, Count = length(substitute(ID)))
    # cumulate counts for multiple primer IDs -> ymax value for rectangles
    plot.df$ymax <- ave(plot.df$Count, plot.df$Start, plot.df$End, plot.df$Run, FUN = cumsum)
    # ymin: cumulative sum minus the current count
    plot.df$ymin <- plot.df$ymax - plot.df$Count
    colnames(plot.df)[colnames(plot.df) %in% c("Start", "End")] <- c("xmin", "xmax")
    #print(plot.df)
    levels <- unique(plot.df$Run)
    plot.df$Run <- factor(plot.df$Run, levels = levels[order(levels)])
    bwidth <- 10 # cover 10 positions with one bar
    ymin <- min(c(-1, -0.1 * max(plot.df$ymax)))
    ymax <- -0.2 # slightly negative to retain the border of the box
    # consider widest range of allowed regions from all template sets:
    ############################
    # set boundaries of the plot:
    xmin <- min(plot.df$xmin)
    xmax <- max(plot.df$xmax)
    ## require a minimal length for the region segments
    if (xmin > 50) {
        xmin <- -50
    }
    if (xmax < 50) {
        xmax <- 50
    }
    region.df <- create_region_boxes(primers, templates, relation, region.names, ymin, ymax, xmax)
    xmin <- min(xmin, min(region.df$xmin))
    x.ticks <- pretty(seq(xmin, xmax))
    x.labels <- x.ticks
    x.labels[x.labels > 0] <- paste0("+", x.labels[x.labels > 0])
    rect.colors <- c("#e5f4ff", "#ffefe5")
    names(rect.colors) <- region.names
    r.colors <- rep(rect.colors, length(levels(plot.df$Run)))
    plot.df$ID <- abbreviate(plot.df$ID, getOption("openPrimeR.plot_abbrev")) # shorten primer IDs
    pal <- getOption("openPrimeR.plot_colors")["Primer"] # the RColorBrewer palette to use
    primer.colors <- colorRampPalette(brewer.pal(8, pal))(length(unique(plot.df[, "ID"])))
    names(primer.colors) <- unique(plot.df[, "ID"])
    primer.colors <- c(rect.colors, primer.colors)
    p <- ggplot(plot.df) + ylab("Number of coverage events") + 
        xlab(xlab) + 
        ggtitle("Sites of primer binding in the templates")  +
        theme(axis.text.x = element_text(angle = 90, hjust = 0.5, vjust = 0.5)) +
         scale_x_continuous(limits = c(xmin, xmax),
                           breaks = x.ticks,
                           labels = x.labels) + 
        geom_vline(xintercept = -1, colour = "red") + # end of binding region
        geom_vline(data = region.df, aes(xintercept = .data[["RelStartPosition"]]), colour = "red") + # start of binding region
        # rectangles for histogram of binding events:
        geom_rect(data = plot.df, 
                  mapping = aes(xmin = .data[["xmin"]], xmax = .data[["xmax"]],
                                      ymin = .data[["ymin"]], ymax = .data[["ymax"]],
                                      fill = .data[["ID"]]),
                    linetype = "blank", alpha = 0.35) + # rectangles show some overlap due to the transparency!!
       scale_fill_manual(values = primer.colors, breaks = unique(plot.df$ID)) + 
        # x-axis rectangles to annotate binding/amplification region:
        geom_rect(data = region.df, 
            mapping = aes(xmin=.data[["xmin"]], xmax=.data[["xmax"]], 
                        ymin=.data[["ymin"]], ymax=.data[["ymax"]], fill = .data[["Region"]]),
            alpha = 0.5,
            colour = "#3d3835", 
            size = 0.3, show.legend = FALSE)
    if (length(unique(plot.df$Run)) > 1) {
        # show facets and don't show individual primer legend
        p <- p + facet_wrap(~Run, ncol = 2) +
            guides(fill = "none")
    } else {
        # only show rectangle text for single plot
        p <- p + geom_text(data=region.df, 
            aes_string(x = "xmin+(xmax-xmin)/2", 
                       y = "ymin+(ymax-ymin)/2", 
                       label = "Region"), size = 4)

    }
    if (length(unique(plot.df$ID)) > 15) {
        # don't show legend for many primers
        p <- p + guides(fill = "none")
    }
    if (length(highlight.set) != 0) {
        # highlight selected sets
        # highlight of strip.text (facet part) individually doesn't work
        # -> this is the solution i came up with
        highlights <- data.frame(Run = highlight.set)
        p <- p + geom_rect(data=highlights,aes(xmin=-Inf, xmax=Inf, 
                    ymin=-Inf, ymax=Inf), fill='red', alpha=0.1)
    }
    return(p)
})

#' @rdname Input
#' @name Input
#' @aliases read_primers
#' @details
#' When loading primers via \code{read_primers}, the input arguments
#' \code{fw.id}, \code{rev.id}, \code{merge.ambig}, and \code{max.degen} 
#' are only used for loading primers from a FASTA file.
#' In this case, please ensure that \code{fw.id} and 
#' \code{rev.id} are set according to the keywords indicating
#' the primer directionalities in the FASTA file.
#' When loading primers from a CSV file, the format of the file should adhere
#' to the structure defined by the \code{\link{Primers}} class. 
#' 
#' @return \code{read_primers} returns a single object of class \code{Primers}
#' if a single input file is provided or a list of such objects if multiple
#' files are provided.
#' @export
#' @examples
#' 
#' primer.fasta <- system.file("extdata", "IMGT_data", "primers", "IGHV", 
#'                      "Ippolito2012.fasta", package = "openPrimeR")
#' primer.df <- read_primers(primer.fasta, "_fw", "_rev")
#' # Read multiple FASTA files
#' fasta.files <- list.files(system.file("extdata", "IMGT_data", "primers", 
#'                  "IGHV", package = "openPrimeR"), pattern = "*\\.fasta",
#'                  full.names = TRUE)[1:3]
#' primer.data <- read_primers(fasta.files)
#' # Read primers from a CSV file
#' primer.csv <- system.file("extdata", "IMGT_data", "comparison", 
#'              "primer_sets", "IGL", "IGL_openPrimeR2017.csv",  package = "openPrimeR")
#' primer.df <- read_primers(primer.csv)
#' # Read multiple primer CSV files
#' primer.files <- list.files(path = system.file("extdata", "IMGT_data", "comparison", 
#'                          "primer_sets", "IGH", package = "openPrimeR"),
#'                           pattern = "*\\.csv", full.names = TRUE)[1:3]
#' primer.data <- read_primers(primer.files)
#' # Read a mixture of FASTA/CSV files:
#' mixed.primers <- c(primer.fasta, primer.csv)
#' primer.data <- read_primers(mixed.primers)
read_primers <- function(fname, fw.id = "_fw", rev.id = "_rev", 
                    merge.ambig = c("none", "merge", "unmerge"), 
                    max.degen = 16, template.df = NULL, 
                    adapter.action = c("warn", "rm"), sample.name = NULL,
                    updateProgress = NULL) {

    adapter.action <- match.arg(adapter.action)
    if (is.function(updateProgress)) {
        detail <- "Reading primers"
        updateProgress(1/2, detail, "inc")
    }
    if (length(fname) > 1) {
        # load multiple primer FASTA/CSV files
        primers <- read_primers_multiple(fname, fw.id = fw.id, 
                    rev.id = rev.id, merge.ambig = merge.ambig, max.degen = max.degen, 
                    template.df = template.df,
                    adapter.action = adapter.action, sample.name = sample.name,
                    updateProgress = updateProgress) 
    } else {
        # load a single primer set from FASTA/CSV
        primers <- read_primers_single(fname, fw.id = fw.id, rev.id = rev.id,
                    merge.ambig = merge.ambig, max.degen = max.degen, 
                    template.df = template.df,
                    adapter.action = adapter.action, sample.name = sample.name,
                    updateProgress = updateProgress) 
    }
    return(primers)
}
read_primers_single <- function(primer.location, fw.id = "_fw", rev.id = "_rev", 
                    merge.ambig = c("none", "merge", "unmerge"), 
                    max.degen = 16, template.df = NULL, 
                    adapter.action = c("warn", "rm"), sample.name = NULL,
                    updateProgress = NULL) {

    if (!file.exists(primer.location)) {
        stop(paste("No file found at specified location: ",
                primer.location, sep = ""))
    }
    # load the primers and at the same time determine whether it's FASTA/CSV input
    primers <- try(my.read.fasta(primer.location, 
                tolower(names(IUPAC_CODE_MAP))), silent = TRUE)
    if (is(primers, "try-error")) {
        # let's try to read as CSV
        fasta.error <- attr(primers, "condition")
        primers <- try(read_primers_csv(primer.location), silent = TRUE)
        if (is(primers, "try-error")) {
            csv.error <- attr(primers, "condition")
            stop(paste("Could not read the file as FASTA or CSV.",
                "The FASTA error was: ",
                fasta.error, ".\n The CSV error was: ", csv.error))
       } else {
            # it's a CSV:
            ext <- "csv"
       }
    } else {
        # it's a FASTA:
        ext <- "fasta"
    }
    if (ext == "csv") {
        # nothing to change in the input
    } else if (ext == "fasta") {
        # sanitize the input
        if (length(primers) == 0) {
            return(NULL)
        }
        if (is.function(updateProgress)) {
            detail <- "Annotating primers"
            updateProgress(1/2, detail, "inc")
        }
        primer.seqs <- sapply(primers, function(x) paste(tolower(x), collapse = ""))
        headers <- sapply(primers, function(x) attr(x, "Annot"))
        if (length(sample.name) == 0) { 
            # use fasta name without file ending as sample name
            sample.name <- sub("^([^.]*).*", "\\1", basename(primer.location))
        }
        # create Primers object:
        primers <- read_primers.internal(primer.seqs, headers, fw.id, rev.id, 
                        merge.ambig, max.degen, sample.name)
    } else {
        stop("Unknown filetype.")
    }
    if (!is.null(template.df)) {
        # check for restriction sites if template.df was inputted
        sites <- check_restriction_sites(primers, template.df, 
                    adapter.action, updateProgress = updateProgress)
    }
    return(primers)
}
#' Input of Multiple Primer Sets.
#'
#' Reads multiple CSV files representing stored objects of class \code{Primers}.
#'
#' @param filenames The paths to multiple primer CSV/FASTA files.
#' @return A list containing objects of class \code{Primers}.
#' @keywords internal
read_primers_multiple <- function(filenames, fw.id, rev.id, merge.ambig,
                    max.degen, template.df, adapter.action, 
                    sample.name, updateProgress) {
    data <- lapply(filenames, function(x) read_primers_single(x, fw.id = fw.id, 
                    rev.id = rev.id, merge.ambig = merge.ambig,
                    max.degen = max.degen, template.df = template.df, 
                    adapter.action = adapter.action, sample.name = sample.name,
                    updateProgress = updateProgress))
    # annotate names of list/make 'Run' identifiers unique:
    data <- set.run.names(data, filenames)
    return(data)
}
set.run.names <- function(data, filenames = NULL) {
    # annotate names of list using set run identifiers
    if (length(data) == 0) {
        return(data)
    }
    if (length(filenames) != 0) {
        # filenames are available -> use stripped filename for empty files
        run.names <- sub("^([^.]*).*", "\\1", basename(filenames))
    } else {
        run.names <- rep("Unknown", length(data))
    }
    non.empty.idx <- which(unlist(lapply(data, nrow)) != 0)
    # use stored run identifier for non-empty files:
    run.names[non.empty.idx] <- sapply(data[non.empty.idx], function(x) x$Run[1])
    # ensure that run names are unique
    run.names <- uniqtag::uniqtag(run.names, k = Inf, make.unique)
    names(data) <- run.names
    # update run names in the loaded primer sets that aren't empty
    for (i in seq_along(non.empty.idx)) {
        data[[non.empty.idx[i]]]$Run <- run.names[non.empty.idx[i]]
    }
    return(data)
}
#' Read Primer CSV File.
#'
#' Reads a primer data frame stored in a CSV file.
#'
#' @param file The path to a csv file containing the primer data.
#' @return A \code{Primers} object, an instance of a data frame.
#' @keywords internal
read_primers_csv <- function(file) { 
    # turn some columns from integer to char if necessary 
    # (column with a single number will be interpreted as
    # numeric otherwise) ...
    if (!file.exists(file)) {
        stop(paste("Primer CSV file at '", file, "' could not be found."), sep = "")
    }
    # character columns:
    fix.cvg.cols.constrained <- c("Relative_Forward_Binding_Position_Start_fw", "Relative_Forward_Binding_Position_End_fw", 
        "Relative_Forward_Binding_Position_Start_rev", "Relative_Forward_Binding_Position_End_rev", 
        "Relative_Reverse_Binding_Position_Start_fw", "Relative_Reverse_Binding_Position_End_fw", 
        "Relative_Reverse_Binding_Position_Start_rev", "Relative_Reverse_Binding_Position_End_rev", 
        "Covered_Seqs", "Binding_Position_Start_rev", "Binding_Position_End_rev", 
        "Binding_Position_Start_fw", "Binding_Position_End_fw", "Binding_Region_Allowed", 
        "Nbr_of_mismatches_fw", "Nbr_of_mismatches_rev", "Binding_Region_Allowed_fw", 
        "Mismatch_pos_fw", "Mismatch_pos_rev", "Binding_Region_Allowed_rev")
    off.fix.cvg.cols.constrained <- paste0("Off_", fix.cvg.cols.constrained)
    fix.cvg.cols.constrained <- c(fix.cvg.cols.constrained, off.fix.cvg.cols.constrained)
    fix.cvg.cols.basic <- paste0("Basic_", fix.cvg.cols.constrained)
    fix.cvg.cols <- c(fix.cvg.cols.constrained, fix.cvg.cols.basic)
    fix.cols <- c(fix.cvg.cols,
                   # coverage constraints:
                   c("primer_efficiency", "T_EVAL_primer_efficiency", 
                     "terminal_mismatch_pos", "annealing_DeltaG",
                     "stop_codon", "substitution", "coverage_model",
                  "Forward", "Reverse", "constraints_passed_T", "Direction", "Run"))
    numeric.fix.cols <- c("primer_length_fw", "primer_length_rev",
                        "mean_primer_efficiency", "primer_specificity")
    factor.fix.cols <- c("ID", "Identifier")
    p <- try(read.csv(file, stringsAsFactors = FALSE, row.names = NULL), silent = TRUE)
    if (is(p, "try-error")) {
        msg <- paste0("The csv file: '", file, "' does not seem to represent a valid object of class 'Primers'",
                      " Please ensure the correct format of the file!")
        my.error("TemplateFormatIncorrect", msg)
    }
    for (i in seq_len(ncol(p))) {
        if (colnames(p)[i] %in% fix.cols) {
            p[, colnames(p)[i]] <- as.character(p[, colnames(p)[i]])
            na.idx <- which(is.na(p[, colnames(p)[i]]))
            if (length(na.idx) != 0) {
                # replace NA's in char cols with empty string:
                p[na.idx, colnames(p)[i]] <- ""
            }
        } else if (colnames(p)[i] %in% numeric.fix.cols) {
            p[, colnames(p)[i]] <- as.numeric(p[, colnames(p)[i]])
        }  else if (colnames(p)[i] %in% factor.fix.cols) {
            p[, colnames(p)[i]] <- factor(p[, colnames(p)[[i]]], 
                                    levels = p[, colnames(p)[[i]]])
        }
    }
    p <- try(Primers(p), silent = TRUE)
    if (is(p, "try-error")) {
        my.error("TemplateFormatIncorrect", 
            paste0("The loaded primer csv data from the file '",
                   file, "' did not represent a valid 'Primers' object.",
                   " Please check your input!"))
    }
    return(p)
}


#' Internal Function for Reading Primers
#'
#' Reads the given primer sequences into a data frame.
#'
#' @param primer.seqs Primer sequences.
#' @param headers Headers of the primer FASTA file.
#' @param fw.id Identifier of forward primers in the headers.
#' @param rev.id Identifier of reverse primers in the headers.
#' @param merge.ambig Should ambiguous primers be merged?
#' @param max.degen Maximum allowed degeneracy
#' @return A data frame with primer sequences.
#' @keywords internal
read_primers.internal <- function(primer.seqs, headers, fw.id, rev.id, merge.ambig = c("none", "merge", "unmerge"), 
    max.degen, sample.name) {
   
    merge.ambig <- match.arg(merge.ambig)
    mode <- "default"
    if (length(primer.seqs) == 0) {
        return(NULL)
    }
    if (fw.id == "" && rev.id == "") {
        # assume only fw primers are present
        fw.id <- "_fw"
        rev.id <- NA
        headers <- paste(headers, fw.id, sep = "")
        my.warning("NotifyPrimersMissingKeyword", "No keywords provided for primer directionalities. Primers were assumed to be sense primers.")
    } else if (fw.id == "" && rev.id != "") {
        # rev id is given -> assume complement of rev hits are forward primers
        my.warning("NotifyPrimersMissingKeyword", "A keyword was provided only for reverse primers. All primers without the reverse keyword were assumed to be forward primers.")
        mode <- "guess_forward"
        fw.id <- NA
    } else if (fw.id != "" && rev.id == "") {
        my.warning("NotifyPrimersMissingKeyword", "A keyword was provided only for forward primers. All primers without the forward keyword were assumed to be reverse primers.")
        mode <- "guess_reverse"
        rev.id <- NA
    }
    dirs <- rep(NA, length(headers))
    dir.fw.idx <- NULL
    dir.rev.idx <- NULL
    if (mode != "guess_forward") {
        dir.fw.idx <- grep(fw.id, headers, fixed = TRUE)
    } else {
        dir.fw.idx <- setdiff(seq_along(headers), grep(rev.id, headers, fixed = TRUE))
    }    
    if (mode != "guess_reverse") {
        dir.rev.idx <- grep(rev.id, headers, fixed = TRUE)
    } else {
        dir.rev.idx <- setdiff(seq_along(headers), grep(fw.id, headers, fixed = TRUE))
    }
    if (length(intersect(dir.fw.idx, dir.rev.idx) != 0)) {
        msg <- "Direction annotation is not distinct. Please use distinct keywords."
        my.error("NotifyPrimersNoDirection", msg)
    }
    dirs[dir.fw.idx] <- "Forward" # names of columns
    dirs[dir.rev.idx] <- "Reverse"
    use.idx <- seq_along(headers)
    if (any(is.na(dirs))) {
        warning.msg <- "Not all primer directions are known. Check your input primer fasta file for the correct direction keywords. Primers without known direction are ignored. The affected primers are:"
        p <- paste(headers[which(is.na(dirs))], collapse = ",")
        msg <- paste(warning.msg, p, sep = "")
        my.error("NotifyPrimersNoDirection", msg)
    }
    p.df <- data.frame(Header = headers, Direction = dirs, Sequence = primer.seqs, 
        Seq_Length = nchar(primer.seqs), stringsAsFactors = FALSE)[use.idx, ]
    rownames(p.df) <- NULL
    # try to match identifiers of primers
    h <- headers
    if (!is.na(fw.id)) {
        h <- sub(fw.id, "", headers[use.idx], fixed = TRUE)
    }
    if (!is.na(rev.id)) {
        h <- sub(rev.id, "", h, fixed = TRUE)
    }
    h <- sub(">", "", h)
    p.df$ID <- h
    # check that every sample has non-duplicate fw/rev primer
    unique.directions <- ddply(p.df, "ID", summarize, duplicate = any(duplicated(substitute(Direction))))
    if (any(unique.directions$duplicate)) {
        # duplicated directions found
        idx <- which(unique.directions$duplicate)
        warn <- paste("Found primers with duplicate directions! The following primers were affected:\n", 
            paste(unique.directions$ID[idx], collapse = "\n", sep = ""))
        my.error("PrimersDuplicateDirections", warn)
    }
    # n.b.: dcast changes variable order -> change back again
    ori.order <- p.df$ID
    p.df <- dcast(p.df, ID ~ Direction, value.var = "Sequence")  # wide-format :-)
    m <- match(p.df$ID, ori.order)
    p.df <- p.df[order(m),]
    # replace missing primers with ''
    p.df[is.na(p.df)] <- ""
    # if one direction is not present -> set it
    dir.mode <- "both"
    if (!"Forward" %in% colnames(p.df)) {
        dir.mode <- "rev"
        p.df$Forward <- ""
    }
    if (!"Reverse" %in% colnames(p.df)) {
        dir.mode <- "fw"
        p.df$Reverse <- ""
    }
    if (merge.ambig == "merge") {
        p.df <- merge.ambig.primers(p.df, dir.mode, max.degen)
        # message(p.df) disambiguate primers with ambiguities
    } else if (merge.ambig == "unmerge") {
        p.df <- disambiguate.primers(p.df)
    }
    # annotate direction of primers:
    directions <- ifelse(nchar(p.df$Forward) != 0 & nchar(p.df$Reverse) != 0, "both", 
                    ifelse(nchar(p.df$Forward) != 0, "fw", "rev"))
    p.df <- cbind(Identifier = paste(seq_len(nrow(p.df)), directions, sep = ""), 
        ID = p.df$ID, p.df[, c("Forward", "Reverse")], 
        primer_length_fw = nchar(p.df$Forward), 
		primer_length_rev = nchar(p.df$Reverse), 
        Direction = directions,
		Degeneracy_fw =  score_degen(strsplit(p.df$Forward, split = "")),
		Degeneracy_rev = score_degen(strsplit(p.df$Reverse, split = "")),
        Run = sample.name,
        stringsAsFactors = FALSE)
    sel <- which(p.df$Direction == "fw") 
    if (length(sel) != 0 && !is.na(fw.id)) {
        p.df$ID[sel] <- paste0(p.df$ID[sel], fw.id)
    }
    sel <- which(p.df$Direction == "rev") 
    if (length(sel) != 0 && !is.na(rev.id)) {
        p.df$ID[sel] <- paste0(p.df$ID[sel], rev.id)
    }
    p.df$ID <- factor(p.df$ID, levels = p.df$ID)
    p.df$Identifier <- factor(p.df$Identifier, levels = p.df$Identifier)
    p.df <- Primers(p.df) # create Primers object
    return(p.df)
}
#' Disambiguation of Primers.
#'
#' Disambiguates ambiguous primer sequences into all possible sequences.
#'
#' @param p.df Primer data frame.
#' @return Data frame with disambiguated primers.
#' @keywords internal
disambiguate.primers <- function(p.df) {
    f <- convert.from.iupac(p.df$Forward)
    r <- convert.from.iupac(p.df$Reverse)
    # combine
    combis <- lapply(seq_along(f), function(x) expand.grid(f[[x]], r[[x]], stringsAsFactors = FALSE))
    identifiers <- unlist(lapply(seq_along(combis), function(x) paste(p.df$ID[x], 
        "_", seq_len(nrow(combis[[x]])), sep = "")))
    df <- do.call(rbind, combis)
    new.df <- data.frame(ID = identifiers, Forward = df[, 1], Reverse = df[, 2], 
        stringsAsFactors = FALSE)
    # message(new.df) message(class(new.df$Forward))
    return(new.df)
}
#' Pairing of Forward and Reverse Primers.
#'
#' Pairs forward and reverse primers such that coverage is maximized 
#' for every pair.
#'
#' @param primer.df An object of class \code{Primers}.
#' @return An object of class \code{Primers} containing the paired primers.
#' @keywords internal
pair_primers <- function(primer.df, template.df) {
    fw.idx <- which(primer.df$Forward != "")
    rev.idx <- which(primer.df$Reverse != "")
    if (length(fw.idx) == 0 || length(rev.idx) == 0) {
        return(primer.df) # nothing to pair
    }
    if (!"primer_coverage" %in% colnames(primer.df)) {
        stop("pair_primers requires primer coverage.")
    }
    # store already paired primers in 'both.df' for later output
    both.df <- primer.df[primer.df$Direction == "both",]
    # try to pair all single forward/reverse primers
    fw.s.idx <- which(primer.df$Direction == "fw")
    rev.s.idx <- which(primer.df$Direction == "rev")
    # combinations of fw and rev primers
    combis <- expand.grid(fw.s.idx, rev.s.idx)
    if (nrow(combis) != 0) {
        # determine intersection between all combinations
        cvd <- covered.seqs.to.idx(primer.df[, "Covered_Seqs"], template.df)
        cvd.shared <- lapply(seq_len(nrow(combis)), 
            function(x) intersect(cvd[[combis[x,1]]], cvd[[combis[x,2]]])
        )
        cvg.shared <- sapply(cvd.shared, length)
        combis$Coverage <- cvg.shared
        # exclude pairs that don't share any coverage
        sel.combis <- which(combis$Coverage != 0)
        combis <- combis[sel.combis,]
        # construct a new primer data frame using the pairs
        pair.df <- primer.df[combis[,1],]
        pair.df$Reverse <- primer.df$Reverse[combis[,2]]
        pair.df$primer_coverage <- combis$Coverage
        pair.df$Coverage_Ratio <- pair.df$primer_coverage / nrow(template.df)
        pair.df$Covered_Seqs <- unlist(lapply(cvd.shared[sel.combis], function(x) paste(template.df$Identifier[x], collapse = ",")))
        pair.df$Direction <- rep("both", nrow(combis))
        pair.df$ID <- factor(paste0(abbreviate(primer.df$ID[combis[,1]], 5), "+", abbreviate(primer.df$ID[combis[,2]], 5)))
        # select non-redundant pairs of primers:
        cvg.matrix <- get.coverage.matrix(pair.df, template.df)
        sel.cols <- remove.redundant.cols(seq_len(nrow(pair.df)), cvg.matrix)
        pair.df <- pair.df[sel.cols,]
    } else {
        pair.df <- NULL
    }
    # output the already paired primers as well as the new pairs
    out.df <- my_rbind(both.df, pair.df)
    ################
    # when pairing, we exclude primers that weren't paired such that only paired primers remained. otherwise we have problems with the subset coverage!!
    # check whether some primers couldn't be paired
    #missing.idx <- which(unlist(lapply(seq_len(nrow(primer.df)), function(x) all(!grepl(primer.df$ID[x], out.df$ID)))))
    #out.df <- my_rbind(primer.df[missing.idx, ], out.df)
    #################
    return(out.df)
}

#' @rdname Output
#' @name Output
#' @aliases write_primers
#' @return \code{write_primers} stores primers to disk.
#' @export
#' @examples
#' 
#' data(Ippolito)
#' # Store primers as FASTA
#' fname.fasta <- tempfile("my_primers", fileext = ".fasta")
#' write_primers(primer.df, fname.fasta)
#' # Store primers as CSV
#' fname.csv <- tempfile("my_primers", fileext = ".csv")
#' write_primers(primer.df, fname.csv, "CSV")
write_primers <- function(primer.df, fname, ftype = c("FASTA", "CSV")) {
    ftype <- match.arg(ftype)
    if (ftype == "FASTA") {
        fw.idx <- which(primer.df$Forward != "")
        rev.idx <- which(primer.df$Reverse != "")
        seqs <- c(primer.df$Forward[fw.idx], primer.df$Reverse[rev.idx])
        labels <- c(as.character(primer.df$ID[fw.idx]), as.character(primer.df$ID[rev.idx]))
        seqinr::write.fasta(as.list(seqs), as.list(labels), fname)
    } else if (ftype == "CSV") {
        write.csv(primer.df, file = fname, row.names = FALSE)
    } else {
        warning("Unknown filetype: ", ftype)
    }
}
#' Direction of Primers.
#'
#' Identifies the directionality of the input primers.
#'
#' @param primers A primer data frame.
#' @return \code{both} if both, forward and reverse primers exist in \code{primers}.
#' Otherwise, if either only forward primers or reverse primers exist, returns \code{fw}
#' or \code{rev}, respectively.
#' @keywords internal
get.analysis.mode <- function(primers) {
    if (length(primers) == 0) {
        return(NULL)
    }
    mode <- "both"
    if (all(primers$Forward == "" | is.na(primers$Forward))) {
        mode <- "rev"
    } else if (all(primers$Reverse == "" | is.na(primers$Reverse))) {
        mode <- "fw"
    }
    return(mode)
}

#' @rdname AnalysisStats
#' @name AnalysisStats
#' @aliases get_cvg_ratio
#' @details
#' The manner in which \code{get_cvg_ratio} determines the coverage ratio 
#' depends on the directionality of the input primers.
#' If either only forward or reverse primers are inputted, the individual
#' coverage of each primer is used to determine the overall coverage. 
#' If, however, forward and reverse primers are inputted at the same time, 
#' the coverage is defined by the intersection of binding events from both,
#' forward and reverse primers.
#'
#' @return By default, \code{get_cvg_ratio} returns a numeric providing 
#' the expected primer coverage ratio.
#' If \code{as.char} is \code{TRUE}, the output is provided as a
#' percentage-formatted character vector. 
#' The attributes \code{no_covered}, \code{no_templates}, 
#' and \code{covered_templates} provide the number of covered templates, 
#' the total number of templates, and the IDs of covered templates, respectively.
#' @export
#' @examples
#' 
#' data(Ippolito)
#' # Determine the overall coverage
#' cvg.ratio <- get_cvg_ratio(primer.df, template.df)
#' # Determine the identitity coverage ratio 
#' cvg.ratio.0 <- get_cvg_ratio(primer.df, template.df, allowed.mismatches = 0)
get_cvg_ratio <- function(primer.df, template.df, allowed.mismatches = NULL, 
                        cvg.definition = c("constrained", "basic"), 
                        mode.directionality = NULL, as.char = FALSE) {
    if (length(primer.df) == 0 || nrow(primer.df) == 0) {
        # no coverage if there's no input.
        return(0.0)
    }
    if (!is(primer.df, "Primers")) {
        stop("Please input a primer data frame.")
    }
    if (!is(template.df, "Templates")) {
        stop("Please provide a valid template data frame.")
    }
    if (!"Covered_Seqs" %in% colnames(primer.df) || length(template.df) == 0) {
        return(NA)
    }
    cvg.definition <- match.arg(cvg.definition)
    # check whether primers and templates correspond to one another
    if (!check_correspondence(primer.df, template.df)) {
        msg <- paste0("Could not match all coverage events to the input template data frame.\n",
            "Please verify whether the primer coverage was actually computed for the input template data frame.")
        stop(msg)
    }
    if (length(mode.directionality) == 0) {
        # determine directionality
        mode.directionality <- get.analysis.mode(primer.df)
    } else {
        # check input mode
        if (!mode.directionality %in% c("fw", "rev", "both")) {
            stop("Unknown mode directionality for cvg_ratio")
        }
    }
    if (cvg.definition == "constrained" && length(allowed.mismatches) == 0) {
        # use the vanilla coverage definition: faster and doesn't require mismatch information
        cvd <- get_covered.vanilla(primer.df, template.df, mode.directionality)
    } else {
        cvg.data <- get_template_cvg_data(primer.df, template.df)
        # select only coverage events from the selected cvg definition
        cvg.data <- cvg.data[cvg.data$Status == cvg.definition, ]
        # retrieve only the allowed coverage events
        if (length(allowed.mismatches) != 0) {
            # select only events with less or equal the allowed number of mismatches
            cvg.data <- cvg.data[cvg.data$Number_of_mismatches <= allowed.mismatches,]
        } 
        cvd <- unique(cvg.data$Template)
    }
    max.cvg <- length(cvd) / nrow(template.df)  # maximum theoretical coverage
    if (as.char) {
        max.cvg <- paste0(round(max.cvg *100, 2), "%")
    }
    # add number of templates covered and total number of templates as attributes
    attr(max.cvg, "no_covered") <- length(cvd)
    attr(max.cvg, "no_templates") <- nrow(template.df)
    attr(max.cvg, "covered_templates") <- as.character(cvd)
    return(max.cvg)
}

#' Determination of the Covered Sequences.
#'
#' Determines the covered template sequences given by \code{template.df} 
#' that are covered by the primers given by \code{primers}.
#'
#' The manner in which the coverage ratio is evaluated depends on 
#' the directionality of the input primers.
#' If either only forward or reverse primers are inputted, the individual
#' coverage of each primer is used to determine the overall coverage. If, however,
#' forward and reverse primers are inputted at the same time, 
#' the coverage is defined by the intersection of binding events from both,
#' forward and reverse primers.
#'
#' @param primers A \code{Primers} object containing the primers
#' for which the coverage should be evaluated.
#' @param template.df A \code{Templates} object containing
#' the template sequences corresponding to \code{primers}. 
#' @param mode.directionality If \code{mode.directionality} is provided,
#' the coverage of templates is computed for a specific direction of primers.
#' Either "fw" (forward coverage only), "rev" (reverse coverage only), or "both" for both directions. If \code{mode.directionality} is not provided
#' the direction is determined by the input primers.
#' @return The IDs of all covered templates.
#' @keywords internal
get_covered.vanilla <- function(primers, template.df, mode.directionality = NULL) {
    if (length(primers) == 0 || nrow(primers) == 0) {
        # no coverage if there's no input.
        return(0.0)
    }
    if (!is(primers, "Primers")) {
        stop("Please input a primer data frame.")
    }
    if (!is(template.df, "Templates")) {
        stop("Please provide a valid template data frame.")
    }
    if (!"Covered_Seqs" %in% colnames(primers) || length(template.df) == 0) {
        return(NA)
    }
    # check whether primers and templates correspond to one another
    if (!check_correspondence(primers, template.df)) {
        msg <- paste0("Could not match all coverage events to the input template data frame.\n",
            "Please verify whether the primer coverage was actually computed for the input template data frame.")
        stop(msg)
    }
    fw.idx <- which(primers$Forward != "")
    rev.idx <- which(primers$Reverse != "")
    if (length(mode.directionality) == 0) {
        # determine directionality
        mode.directionality <- get.analysis.mode(primers)
    } else {
        # check input mode
        if (!mode.directionality %in% c("fw", "rev", "both")) {
            stop("Unknown mode directionality for cvg_ratio")
        }
    }
    primer.df <- primers
    if (mode.directionality == "both") {
        # fw and rev primers -> intersection of fw and rev cvg events gives the total
        # coverage
        hits.fw <- unique(unlist(covered.seqs.to.idx(primer.df$Covered_Seqs[fw.idx], 
            template.df)))
        hits.rev <- unique(unlist(covered.seqs.to.idx(primer.df$Covered_Seqs[rev.idx], 
            template.df)))
        cvd <- intersect(hits.fw, hits.rev)  # idx of covered templates
    } else if (mode.directionality == "fw") {
        # only fw primers direction
        cvd <- unique(unlist(covered.seqs.to.idx(primer.df$Covered_Seqs[fw.idx], template.df)))
    } else {
        cvd <- unique(unlist(covered.seqs.to.idx(primer.df$Covered_Seqs[rev.idx], template.df)))
    }
    # convert to template IDs
    cvd <- template.df$ID[cvd]
    return(cvd)
}

############
# Evaluation functions for determining whether constraints are fulfilled
#############

#' Apply Constraints to a List.
#'
#' Checks whether the input values are within the specified limits.
#'
#' Applies a constraint to every element in a vector of comma separated strings.
#' Applied when filtering covered seqs according to primer efficiency.
#' 
#' @param gc.ratio.fw Forward values (comma-separated strings).
#' @param gc.ratio.rev Reverse values (comma-separated strings).
#' @param min.GC Minimal allowed value.
#' @param max.GC Maximal allowed value.
#' @param fw.idx Indices of forward values to consider.
#' @param rev.idx Indices of reverse values to consider.
#' @param mode.directionality Direction of primers
#' @return Data frame with \code{TRUE} for values fulfilling the constraints, 
#' \code{FALSE} otherwise.
#' @keywords internal
apply.constraint.list <- function(gc.ratio.fw, gc.ratio.rev, min.GC, max.GC, fw.idx, 
    rev.idx, mode.directionality = c("fw", "rev", "both")) {

    if (length(mode.directionality) == 0) {
        stop("Please supply the 'mode.directionality' argument.")
    }
    mode.directionality <- match.arg(mode.directionality)
    if (is.character(gc.ratio.fw)) {
        gc.ratio.fw <- lapply(gc.ratio.fw, function(x) as.numeric(unlist(strsplit(x, 
            split = ","))))
        gc.ratio.rev <- lapply(gc.ratio.rev, function(x) as.numeric(unlist(strsplit(x, 
            split = ","))))
    }
    if (length(min.GC) != 0 && is.na(min.GC)) {
        min.GC <- numeric(0)
    }
    if (length(max.GC) != 0 && is.na(max.GC)) {
        max.GC <- numeric(0)
    }
    if (length(min.GC) != 0 && length(max.GC) != 0) {
        s <- min.GC
        e <- max.GC
    } else if (length(min.GC) == 0) {
        s <- -Inf
        e <- max.GC
    } else {
        s <- min.GC
        e <- Inf
    }
    if (length(fw.idx) != 0) {
        b.fw <- lapply(gc.ratio.fw[fw.idx], function(x) x >= s & x <= e)
    } else {
        b.fw <- lapply(gc.ratio.fw, function(x) x >= s & x <= e)
    }
    if (length(rev.idx) != 0) {
        b.rev <- lapply(gc.ratio.rev[rev.idx], function(x) x >= s & x <= e)
    } else {
        b.rev <- lapply(gc.ratio.rev, function(x) x >= s & x <= e)
        
    }
    res.fw <- vector("list", length(gc.ratio.fw))
    res.rev <- vector("list", length(gc.ratio.rev))
    if (length(fw.idx) != 0) {
        res.fw[fw.idx] <- b.fw
    } else {
        res.fw <- b.fw
    }
    if (length(rev.idx) != 0) {
        res.rev[rev.idx] <- b.rev
    } else {
        res.rev <- b.rev
    }
    res <- data.frame(Forward = rep("", length(gc.ratio.fw)), Reverse = rep("", length(gc.ratio.fw)))
    if (mode.directionality == "both") {
        res$Forward <- string.rep(res.fw)
        res$Reverse <- string.rep(res.rev)
    } else if (mode.directionality == "fw") {
        res$Forward <- string.rep(res.fw)
    } else {
        res$Reverse <- string.rep(res.rev)
    }
    return(res)
}
#' Application of Constraints
#'
#' Checks whether the input values are within the specified limits.
#'
#' @param gc.ratio.fw Forward values.
#' @param gc.ratio.rev Reverse values.
#' @param min.GC Minimal allowed value.
#' @param max.GC Maximal allowed value.
#' @param fw.idx Indices of forward values to consider.
#' @param rev.idx Indices of reverse values to consider.
#' @param mode.directionality Direction of primers
#' @return Data frame with \code{TRUE} for values fulfilling the constraints, 
#' \code{FALSE} otherwise.  Also returns \code{FALSE} if a data point is not available.
#' @keywords internal
apply.constraint <- function(gc.ratio.fw, gc.ratio.rev, min.GC, max.GC, fw.idx, rev.idx, 
    mode.directionality = c("fw", "rev", "both")) {
    if (length(min.GC) == 0 && length(max.GC) == 0) {
        return(data.frame(rep(TRUE, length(gc.ratio.fw))))
    }
    if (length(min.GC) != 0 && is.na(min.GC)) {
        min.GC <- numeric(0)
    }
    if (length(max.GC) != 0 && is.na(max.GC)) {
        max.GC <- numeric(0)
    }
    if (length(min.GC) != 0 && length(max.GC) != 0) {
        s <- min.GC
        e <- max.GC
    } else if (length(min.GC) == 0) {
        s <- -Inf
        e <- max.GC
    } else {
        s <- min.GC
        e <- Inf
    }
    result <- rep(TRUE, length(gc.ratio.fw))
    r.fw <- gc.ratio.fw[fw.idx] >= s & gc.ratio.fw[fw.idx] <= e
    res.fw <- rep(TRUE, length(gc.ratio.fw))
    if (length(fw.idx) != 0) {
        # only take constraint values if input was given (!= '') as determined by fw.idx
        res.fw[fw.idx] <- r.fw
    }
    r.rev <- gc.ratio.rev[rev.idx] >= s & gc.ratio.rev[rev.idx] <= e
    res.rev <- rep(TRUE, length(gc.ratio.rev))
    if (length(rev.idx) != 0) {
        res.rev[rev.idx] <- r.rev
    }
    # ensure that not avaialable values are FALSE
    res.rev[is.na(res.rev)] <- FALSE
    res.fw [is.na(res.fw)] <- FALSE
    if (mode.directionality == "both") {
        res <- data.frame(Forward = res.fw, Reverse = res.rev)
    } else if (mode.directionality == "fw") {
        res <- data.frame(Forward = res.fw)
    } else {
        res <- data.frame(Reverse = res.rev)
    }
    return(res)
}
string.rep <- function(values) {
    str <- unlist(lapply(values, function(x) {
        if (length(x) != 0 && all(is.na(x))) { 
            ""
        } else {
            paste0(x, collapse = ",")
        }
        }))
    if (length(str) == 0) {
        # ensure we output a string
        str <- ""
    }
    return(str)
}
#' Computation of Constraints.
#'
#' Computes the specified constraints for the input primers.
#'
#' @param primer.df Primer data frame.
#' @param mode.directionality Primer directionality.
#' @param template.df Template data frame.
#' @param settings A \code{DesignSettings} object.
#' @param active.constraints Strings giving the constraints that are to be computed.
#' @param no.structures Whether dimer structures shall be computed.
#' @param for.shiny Whether to format output for HTML.
#' @param updateProgress Progress callback function for shiny.
#' @return A data frame with columns for each constraint in \code{active.constraints}.
#' @keywords internal
compute.constraints <- function(primer.df, mode.directionality = c("fw", "rev", "both"), template.df, settings,
    active.constraints = c("primer_coverage", "primer_length", "primer_specificity",
                            "gc_clamp", "gc_ratio", "no_runs", "no_repeats", "self_dimerization", 
                            "cross_dimerization", "melting_temp_range", 
                            "melting_temp_diff", "secondary_structure", 
                            # coverage conditions:
                            "primer_efficiency", "annealing_DeltaG",
                            "stop_codon", "terminal_mismatch_pos",
                            "substitution", "hexamer_coverage",
                            "coverage_model",
                            # off-target cvg constraints
                            "off_primer_efficiency", "off_annealing_DeltaG",
                            "off_coverage_model"),
    no.structures = FALSE, for.shiny = FALSE, updateProgress = NULL) {
    if (length(active.constraints) == 0) {
        return(NULL)
    }
    ###################
    # retrieve settings
    primer_conc <- PCR(settings)$primer_concentration
    template_conc <- PCR(settings)$template_concentration
    na_salt_conc <- PCR(settings)$Na_concentration
    mg_salt_conc <- PCR(settings)$Mg_concentration
    k_salt_conc <- PCR(settings)$K_concentration
    tris_salt_conc <- PCR(settings)$Tris_concentration
    annealing.temp <- PCR(settings)$annealing_temp
    if (length(annealing.temp) == 1) {
        # ensure that every primer has an individual annealing temperature
        annealing.temp <- rep(annealing.temp, nrow(primer.df))
    }
    taqEfficiency <- PCR(settings)$use_taq_polymerase
    allowed.mismatches <- conOptions(settings)$allowed_mismatches
    allowed.other.binding.ratio <- conOptions(settings)$allowed_other_binding_ratio
    allowed.region.definition <- conOptions(settings)$allowed_region_definition 
    ##################
    if (length(active.constraints) != 0) { # don't match if there are no constraints
        active.constraints <- match.arg(active.constraints, several.ok = TRUE)
    }
    if (length(mode.directionality) == 0) {
        stop("Please supply the 'mode.directionality' argument.")
    }
    mode.directionality <- match.arg(mode.directionality)
    # Note: only place values with different names than original constraints arising from active constraints directly here (e.g.
    # with the same name as constraints. others will be removed.  store all
    # constraint results in 'constraint.values'
    constraint.values <- data.frame(#mean_primer_efficiency = numeric(nrow(primer.df)),
        primer_length_fw = numeric(nrow(primer.df)), primer_length_rev = numeric(nrow(primer.df)), 
        gc_clamp_fw = numeric(nrow(primer.df)), gc_clamp_rev = numeric(nrow(primer.df)), 
        gc_ratio_fw = numeric(nrow(primer.df)), gc_ratio_rev = numeric(nrow(primer.df)), 
        no_repeats_fw = numeric(nrow(primer.df)), no_repeats_rev = numeric(nrow(primer.df)), 
        no_runs_fw = numeric(nrow(primer.df)), no_runs_rev = numeric(nrow(primer.df)))
    original.cols <- colnames(constraint.values)
    melting.temps <- NULL  # required for specificity
    if ("melting_temp_range" %in% active.constraints || "melting_temp_diff" %in% active.constraints) {
        # melting temperature
        if (is.function(updateProgress)) {
            detail <- "melting temp"
            updateProgress(1/length(active.constraints), detail, "inc")
        }
        # melting temperature T_m:
        melting.temps <- compute.melting.temps(primer.df, primer_conc, 
                na_salt_conc, mg_salt_conc, k_salt_conc, 
                tris_salt_conc, mode.directionality)
        # update melting temp
        primer.df$melting_temp <- melting.temps$melting_temp
        # prevent an error due to cbind with data frame is empty
        if (length(constraint.values) == 0) {
            constraint.values <- melting.temps
        } else {
            constraint.values <- cbind(constraint.values, melting.temps)
        }
    }
    # determine annealing temp if required but not available
    annealing.required <- any(active.constraints %in% c("secondary_structure", "primer_efficiency", 
                                                        "annealing_DeltaG", "off_primer_efficiency",
                                                        "off_annealing_DeltaG",
                                                        "self_dimerization", "cross_dimerization"))
    if (length(annealing.temp) == 0 && annealing.required) {
        annealing.temp <- compute_annealing_temp(primer.df, mode.directionality, 
                  template.df, na_salt_conc, mg_salt_conc, k_salt_conc, 
                  tris_salt_conc, primer_conc)
    }
    #print("Ta is:")
    #print(annealing.temp)
    cvg.required <- FALSE
    if ("primer_coverage" %in% active.constraints || "primer_specificity" %in% active.constraints && !"primer_specificity" %in% colnames(primer.df)) {
        # only trigger cvg computation for specificity purposes if coverage is explicitly required by constraints, otherwise use existing entries.
        cvg.required <- TRUE
    }
    if (cvg.required) {
        # covered templates by primers
        if (is.function(updateProgress)) {
            detail <- "primer coverage"
            updateProgress(1/length(active.constraints), detail, "inc")
        }
        # evaluate primer cvg first to have coverage data all constraints (specificity)
        cvg.df <- evaluate.primer.cvg(template.df, primer.df, mode.directionality, settings, updateProgress)
        # update primer df with coverage (might be needed for other constraints).
        primer.df <- update.constraint.values(primer.df, cvg.df)
        constraint.values <- cbind(constraint.values, cvg.df)
    }
    if ("primer_length" %in% active.constraints) {
        # length of primers
        if (is.function(updateProgress)) {
            detail <- "primer length"
            updateProgress(1/length(active.constraints), detail, "inc")
        }
        # should already be computed in the input, no computation necessary
        if (!"primer_length_fw" %in% colnames(primer.df)) {
            stop("WARNING: primer len not computed already")
        }
        constraint.values$primer_length_fw <- primer.df$primer_length_fw
        constraint.values$primer_length_rev <- primer.df$primer_length_rev
    }
    if ("gc_clamp" %in% active.constraints) {
        # count of gc at 3' end
        if (is.function(updateProgress)) {
            detail <- "gc clamp"
            updateProgress(1/length(active.constraints), detail, "inc")
        }
        gc.count.fw <- evaluate.GC.clamp(primer.df$Forward)
        gc.count.rev <- evaluate.GC.clamp(primer.df$Reverse)
        constraint.values$gc_clamp_fw <- gc.count.fw
        constraint.values$gc_clamp_rev <- gc.count.rev
    }
    if ("gc_ratio" %in% active.constraints) {
        # ratio of gc
        if (is.function(updateProgress)) {
            detail <- "gc ratio"
            updateProgress(1/length(active.constraints), detail, "inc")
        }
        gc.ratio.fw <- compute.gc.ratio(primer.df$Forward)
        gc.ratio.rev <- compute.gc.ratio(primer.df$Reverse)
        constraint.values$gc_ratio_fw <- gc.ratio.fw
        constraint.values$gc_ratio_rev <- gc.ratio.rev
    }
    if ("no_repeats" %in% active.constraints) {
        # number of repeats
        if (is.function(updateProgress)) {
            detail <- "no repeats"
            updateProgress(1/length(active.constraints), detail, "inc")
        }
        no.repeats.fw <- nbr.of.repeats(primer.df$Forward)
        no.repeats.rev <- nbr.of.repeats(primer.df$Reverse)
        constraint.values$no_repeats_fw <- no.repeats.fw
        constraint.values$no_repeats_rev
    }
    if ("no_runs" %in% active.constraints) {
        # number of runs
        if (is.function(updateProgress)) {
            detail <- "no runs"
            updateProgress(1/length(active.constraints), detail, "inc")
        }
        nbr.runs.fw <- nbr.of.runs(primer.df$Forward)
        nbr.runs.rev <- nbr.of.runs(primer.df$Reverse)
        constraint.values$no_runs_fw <- nbr.runs.fw
        constraint.values$no_runs_rev <- nbr.runs.rev
    }
    #######
    # COVERAGE CONSTRAINTS: terminal mismatches, stop codons, efficiency, annealingDeltaG
    #######
    if (any(c("stop_codon", "substitution") %in% active.constraints)) {
        if (is.function(updateProgress)) {
            detail <- "Identifying mutations"
            updateProgress(1/length(active.constraints), detail, "inc")
        }
        #######################
        # Stop codons
        ###################
        mutation.types <- c("stop_codon", "substitution")[c("stop_codon", "substitution") %in% active.constraints]
        mutation.check <- mismatch.mutation.check(primer.df, template.df, mutation.types)
        out.strings <- lapply(mutation.check, function(x) 
            if (length(x) == 0) { 
                ""
            } else {
                unlist(lapply(seq_len(ncol(x)), function(y) paste(x[, y], collapse =",")))
        })
        if ("stop_codon" %in% mutation.types) {
            stop.str <- sapply(out.strings, function(x) x[1])
            constraint.values$stop_codon <- stop.str
        }
        if ("substitution" %in% mutation.types) {
            sub.str <- sapply(out.strings, function(x) x[2]) 
            constraint.values$substitution <- sub.str
        }
    }
    if ("terminal_mismatch_pos" %in% active.constraints) {
        ###########################
        # Identify 3' mismatches
        ############################
        if (is.function(updateProgress)) {
            detail <- "Terminal mismatch identification"
            updateProgress(1/length(active.constraints), detail, "inc")
        }
        mm.check <- check.3prime.mismatches(template.df, primer.df, mode.directionality)
        mm.str <- string.rep(mm.check)
        constraint.values$terminal_mismatch_pos <- mm.str
    }
    if ("hexamer_coverage" %in% active.constraints && length(conOptions(settings)$hexamer_coverage) != 0 && conOptions(settings)$hexamer_coverage == "active") {
        # only computed when active because presence of hexamer_coverage is used for ignoring the filtering later! 
        ###########################
        # Identify 3' hexamers with 100% complementarity to template
        ############################
        if (is.function(updateProgress)) {
            detail <- "3' Hexamer detection"
            updateProgress(1/length(active.constraints), detail, "inc")
        }
        hex.check <- check.3prime.hexamers(template.df, primer.df, mode.directionality)
        hex.str <- string.rep(hex.check)
        constraint.values$hexamer_coverage <- hex.str
    }
    if ("coverage_model" %in% active.constraints) {
        #######
        # Logistic regression model for predicting coverage
        ########
        if (is.function(updateProgress)) {
            detail <- "Coverage model"
            updateProgress(1/length(active.constraints), detail, "inc")
        }
        preds <- predict_coverage(primer.df, template.df, settings,
                                  updateProgress = updateProgress)
        pred.string <- string.rep(preds)
        constraint.values$coverage_model <- pred.string
    }
    if ("off_coverage_model" %in% active.constraints) {
        #######
        # Logistic regression model for predicting coverage
        ########
        if (is.function(updateProgress)) {
            detail <- "Coverage model"
            updateProgress(1/length(active.constraints), detail, "inc")
        }
        preds <- predict_coverage(primer.df, template.df, settings, mode = "off_target",
                                  updateProgress = updateProgress)
        pred.string <- string.rep(preds)
        constraint.values$off_coverage_model <- pred.string
    }
    if ("primer_efficiency" %in% active.constraints) {
        ############
        # Thermodynamic model of annealing + 3' terminal mismatch model (for taq) 
        #############
        if (is.function(updateProgress)) {
            detail <- "Primer efficiency"
            updateProgress(1/length(active.constraints), detail, "inc")
        }
        eff_taq <- compute.primer.efficiencies(primer.df, template.df, 
                            annealing.temp, taqEfficiency = TRUE, 
                            PCR(settings)$primer_concentration,
                            PCR(settings)$Na_concentration,
                            PCR(settings)$Mg_concentration,
                            PCR(settings)$K_concentration,
                            PCR(settings)$Tris_concentration,
                            mode = "on_target")
        eff.string <- string.rep(eff_taq)
        eff.mean <- sapply(eff_taq, function(x) ifelse(length(x) == 0, 0, mean(x)))
        constraint.values$primer_efficiency <- eff.string
        constraint.values$mean_primer_efficiency <- eff.mean
    }
    if ("off_primer_efficiency" %in% active.constraints) {
        ############
        # Off-target binding events: Thermodynamic model of annealing + 3' terminal mismatch model (for taq) 
        #############
        if (is.function(updateProgress)) {
            detail <- "Primer efficiency"
            updateProgress(1/length(active.constraints), detail, "inc")
        }
        eff_off <- compute.primer.efficiencies(primer.df, template.df, 
                            annealing.temp, taqEfficiency = TRUE, 
                            PCR(settings)$primer_concentration,
                            PCR(settings)$Na_concentration,
                            PCR(settings)$Mg_concentration,
                            PCR(settings)$K_concentration,
                            PCR(settings)$Tris_concentration,
                            mode = "off_target")
        eff.string <- string.rep(eff_off)
        constraint.values$off_primer_efficiency <- eff.string
    }
    if ("annealing_DeltaG" %in% active.constraints) {
        ######################
        # Duplex binding energies
        ########################
        if (is.function(updateProgress)) {
            detail <- "Free energy of annealing"
            updateProgress(1/length(active.constraints), detail, "inc")
        }
        deltaG_ABs <- get.duplex.energies(primer.df, template.df, annealing.temp, settings, mode = "on_target")
        deltaG.string <- string.rep(deltaG_ABs)
        deltaG.mean <- unlist(lapply(deltaG_ABs, function(x) ifelse(length(x) == 0, 0, mean(x))))
        constraint.values$annealing_DeltaG <- deltaG.string
        constraint.values$mean_annealing_DeltaG <- deltaG.mean
    }
     if ("off_annealing_DeltaG" %in% active.constraints) {
        ######################
        # Off-target duplex binding energies
        ########################
        if (is.function(updateProgress)) {
            detail <- "Off-targets: free energy of annealing"
            updateProgress(1/length(active.constraints), detail, "inc")
        }
        deltaG_ABs <- get.duplex.energies(primer.df, template.df, annealing.temp, settings, mode = "off_target")
        deltaG.string <- string.rep(deltaG_ABs)
        constraint.values$off_annealing_DeltaG <- deltaG.string
    }
    # get secondary structures
    if ("secondary_structure" %in% active.constraints) {
        # primer secondary structures
        if (is.function(updateProgress)) {
            detail <- "secondary structure"
            updateProgress(1/length(active.constraints), detail, "inc")
        }
        structs <- compute.secondary.structures(primer.df, mode.directionality, annealing.temp)
        if (length(constraint.values) == 0) {
            constraint.values <- structs
        } else {
            constraint.values <- cbind(constraint.values, structs)
        }
    }
    if ("primer_specificity" %in% active.constraints) {
        # primer specificity
        if (length(primer.df$primer_specificity) != 0) {
            # look at existing data
            constraint.values$primer_specificity <- primer.df$primer_specificity
        } 
    }
    if ("self_dimerization" %in% active.constraints) {
        if (is.function(updateProgress)) {
            detail <- "self dimerization"
            updateProgress(1/length(active.constraints), detail, "inc")
        }
        self.dimer.df <- compute.all.self.dimers.frontend(primer.df, primer_conc, 
            na_salt_conc, mg_salt_conc, k_salt_conc, tris_salt_conc, annealing.temp, for.shiny = for.shiny, no.structures = no.structures)
        if (length(self.dimer.df) != 0) {
            # we don't have any column entries when there's nothing to report ..
            if (length(constraint.values) == 0 || nrow(constraint.values) == 0) {
                constraint.values <- self.dimer.df
            } else {
                constraint.values <- cbind(constraint.values, self.dimer.df)
            }
        }
    }
    if ("cross_dimerization" %in% active.constraints) {
        if (is.function(updateProgress)) {
            detail <- "cross dimerization"
            updateProgress(1/length(active.constraints), detail, "inc")
        }
        cross.dimer.df <- compute.all.cross.dimers.frontend(primer.df, primer_conc, 
            na_salt_conc, mg_salt_conc, k_salt_conc, tris_salt_conc, annealing.temp, for.shiny = for.shiny, no.structures = no.structures)
        if (length(cross.dimer.df) != 0) {
            # no entries in constraint.values when there's nothing to report ...
            if (length(constraint.values) == 0 || nrow(constraint.values) == 0) {
                constraint.values <- cross.dimer.df
            } else {
                constraint.values <- cbind(constraint.values, cross.dimer.df)
            }
        }
    }
    # remove all 'constraint.values' that that were not actually computed
    keep.idx.a <- which(!(colnames(constraint.values) %in% original.cols))
    keep.idx.b <- unlist(lapply(active.constraints, function(x) grep(x, colnames(constraint.values))))
    keep.idx.c <- unlist(lapply(names(cvg_constraints(settings)), function(x) grep(x, colnames(constraint.values)))) # retain cvg constraint values
    #print("to keep:")
    #print(colnames(constraint.values[keep.idx.a]))
    keep.idx <- union(keep.idx.a, c(keep.idx.b, keep.idx.c))
    #print(colnames(constraint.values)[keep.idx])
    constraint.values <- constraint.values[, keep.idx, drop = FALSE]
    return(constraint.values)
}
#' Check Constraints on Templates
#'
#' Transforms the comma-separated input strings to a boolean representation.
#'
#' @param template.constraints Strings with comma-separated values to be turned to logical.
#' @return List with boolean values
#' @keywords internal
check.template.constraints <- function(template.constraints) {
    # evaluate list constraints (e.g. primer efficiency)
    if (length(template.constraints) == 0) {
        return(NULL)
    }
    all.res <- vector("list", length(template.constraints))
    # message('check.template.constraints:') message(template.constraints) for each
    # template constraint
    for (i in seq_along(template.constraints)) {
        data <- template.constraints[[i]]
        res <- lapply(seq_len(nrow(data)), function(y) {
            # each x is two lists, one for fw, one for rev
            x <- data[y, ]
            # message(x)
            s1 <- as.logical(strsplit(as.character(x[[1]]), split = ",")[[1]])
            s2 <- as.logical(strsplit(as.character(x[[2]]), split = ",")[[1]])
            if (length(s1) == 0) {
                res <- s2
            } else if (length(s2) == 0) {
                res <- s1
            } else {
                res <- unlist(lapply(seq_along(s1), function(y) s1[y] && s2[y]))
            }
        })
        all.res[[i]] <- res
    }
    names(all.res) <- names(template.constraints)
    return(all.res)
}
#' Merge Template Decisions.
#' 
#' Merges the results for multiple template evaluations.
#'
#' @param eval.t List with evaluated template constraints
#' @return List with merged boolean decisions.
#' @keywords internal
merge.template.decisions <- function(eval.t) {
    # get merged result for all template decisions
    final.res <- NULL
    if (length(eval.t) != 0) {
        final.res <- vector("list", length(eval.t[[1]]))
    }
    for (i in seq_along(final.res)) {
        # for every primer
        template.decisions <- rep(TRUE, length(eval.t[[1]][[i]]))
        for (j in seq_along(eval.t)) {
            # for every constraint
            cur.decisions <- eval.t[[j]][[i]]
            f.idx <- which(!cur.decisions)
            if (length(f.idx) != 0) {
                template.decisions[f.idx] <- FALSE
            }
        }
        final.res[[i]] <- template.decisions
    }
    return(final.res)
}
#' Check for Relaxation
#'
#' Determines whether constraints where relaxed or not.
#'
#' @param used.constraints The constraints that were used during the optimization.
#' @param input.constraints The user-specified constraints.
#' @return If \code{input.constraints} was relaxed TRUE is returned, otherwise FALSE.
#' @keywords internal
were.constraints.relaxed <- function(used.constraints, input.constraints) {
    # determines whether used constraints differ from input constraints
    
    m <- match(names(used.constraints), names(input.constraints))
    was.relaxed <- any(sapply(seq_along(m), function(x) any(used.constraints[[x]] != 
        input.constraints[[m[x]]])))
    return(was.relaxed)
}
#' Evaluation of Template Constraints.
#'
#' Evaluates the input template constraints.
#'
#' @param constraint.values Data frame with template constraints
#' @param constraint.settings List specifying the allowed values for constraint evaluation.
#' @param active.constraints Strings specifying the constraints to check.
#' @param mode.directionality Direction of primers.
#' @return List indicating which template constraints where fulfilled or not (TRUE/FALSE).
#' @keywords internal
evaluate.template.constraints <- function(constraint.values, constraint.settings, 
                                          active.constraints, mode.directionality = c("fw", "rev", "both")) {

    if (length(mode.directionality) == 0) {
        stop("Please supply the 'mode.directionality' argument.")
    }
    mode.directionality <- match.arg(mode.directionality)
    template.constraints <- list()  # constraints on template-specific level (lists)
    if ("coverage_model" %in% active.constraints) {
        cvg.model <- apply.constraint.list(constraint.values$coverage_model, 
            constraint.values$coverage_model, constraint.settings[["coverage_model"]]["min"], 
            constraint.settings[["coverage_model"]]["max"], NULL, NULL, mode.directionality)
        template.constraints[["T_EVAL_coverage_model"]] <- cvg.model
    }
    if ("off_coverage_model" %in% active.constraints) {
        model.FPR <- apply.constraint.list(constraint.values$off_coverage_model, 
            constraint.values$off_coverage_model, constraint.settings[["off_coverage_model"]]["min"],
            constraint.settings[["off_coverage_model"]]["max"], NULL, NULL, mode.directionality)
        template.constraints[["T_off_EVAL_coverage_model"]] <- model.FPR
    }

    if ("primer_efficiency" %in% active.constraints) {
        primer.eff <- apply.constraint.list(constraint.values$primer_efficiency, 
            constraint.values$primer_efficiency, constraint.settings[["primer_efficiency"]]["min"], 
            constraint.settings[["primer_efficiency"]]["max"], NULL, NULL, mode.directionality)
        template.constraints[["T_EVAL_primer_efficiency"]] <- primer.eff
    }
    if ("off_primer_efficiency" %in% active.constraints) { # off-target events
        primer.eff <- apply.constraint.list(constraint.values$off_primer_efficiency, 
            constraint.values$off_primer_efficiency, constraint.settings[["off_primer_efficiency"]]["min"], 
            constraint.settings[["off_primer_efficiency"]]["max"], NULL, NULL, mode.directionality)
        template.constraints[["T_off_EVAL_primer_efficiency"]] <- primer.eff
    }
    if ("annealing_DeltaG" %in% active.constraints) {
        annealing.deltaG <- apply.constraint.list(constraint.values$annealing_DeltaG, 
            constraint.values$annealing_DeltaG, constraint.settings[["annealing_DeltaG"]]["min"],
            constraint.settings[["annealing_DeltaG"]]["max"], NULL, NULL, mode.directionality)
        template.constraints[["T_EVAL_annealing_DeltaG"]] <- annealing.deltaG
    }
    if ("off_annealing_DeltaG" %in% active.constraints) {
        annealing.deltaG <- apply.constraint.list(constraint.values$off_annealing_DeltaG, 
            constraint.values$off_annealing_DeltaG, constraint.settings[["off_annealing_DeltaG"]]["min"],
            constraint.settings[["off_annealing_DeltaG"]]["max"], NULL, NULL, mode.directionality)
        template.constraints[["T_off_EVAL_annealing_DeltaG"]] <- annealing.deltaG
    }
    if ("stop_codon" %in% active.constraints) {
        stop.sel <- apply.constraint.list(constraint.values$stop_codon, constraint.values$stop_codon, 
                                        constraint.settings[["stop_codon"]]["min"], constraint.settings[["stop_codon"]]["max"],
                                        NULL, NULL, mode.directionality)
        template.constraints[["T_EVAL_stop_codon"]] <- stop.sel
    }
    if ("substitution" %in% active.constraints) {
        sub.sel <- apply.constraint.list(constraint.values$substitution, constraint.values$substitution, 
                                        constraint.settings[["substitution"]]["min"], constraint.settings[["substitution"]]["max"],
                                        NULL, NULL, mode.directionality)
        template.constraints[["T_EVAL_substitution"]] <- sub.sel
    }

    if ("terminal_mismatch_pos" %in% active.constraints) {
        # 3' Mismatch filtering
        mm.sel <- apply.constraint.list(constraint.values$terminal_mismatch_pos, constraint.values$terminal_mismatch_pos, 
                                        constraint.settings[["terminal_mismatch_pos"]]["min"], constraint.settings[["terminal_mismatch_pos"]]["max"],
                                        NULL, NULL, mode.directionality)

        template.constraints[["T_EVAL_terminal_mismatch_pos"]] <- mm.sel
    } 
    template.con.result <- check.template.constraints(template.constraints)
    return(template.con.result)
}

#' Update of Primer Constraints.
#'
#' Updates the input primer data frame with the computed constraint values.
#'
#' @param constraint.df Primer data frame.
#' @param constraint.values Data frame with computed constraint values.
#' @return A primer data frame with updated columns.
#' @keywords internal
update.constraint.values <- function(constraint.df, constraint.values) {
    # update/insert computed constraint.values into constraint.df
    if (length(constraint.values) != 0) {
        idx.replace <- which(colnames(constraint.values) %in% colnames(constraint.df))
        idx.add <- which(!(colnames(constraint.values) %in% colnames(constraint.df)))
        if (length(idx.replace) != 0) {
            constraint.df[, colnames(constraint.values)[idx.replace]] <- constraint.values[, 
                idx.replace]
        }
        if (length(idx.add) != 0) {
            constraint.df[, colnames(constraint.values)[idx.add]] <- constraint.values[, 
                idx.add]
        }
    }
    return(constraint.df)
}
#' @rdname PrimerEval
#' @name PrimerEval
#' @aliases check_constraints
#' @details
#' When the optional argument
#' \code{active.constraints} is supplied to \code{check_constraints}, 
#' only a subset of the constraints
#' provided in \code{settings} is evaluated. Only constraints that
#' are defined in \code{settings} can be computed. For a detailed
#' description of all possible constraints and their options, please
#' consider the \code{\link{ConstraintSettings}} documentation.
#'
#' @export
#' @examples
#' 
#' data(Ippolito)
#' settings.xml <- system.file("extdata", "settings", 
#'                  "C_Taq_PCR_high_stringency.xml", package = "openPrimeR")
#' settings <- read_settings(settings.xml)
#' # Check GC clamp and number of runs for all primers:
#' constraint.df <- check_constraints(primer.df, template.df, 
#'                      settings, active.constraints = c("gc_clamp", "no_runs"))
#' # Summarize the evaluation results
#' summary(constraint.df)
check_constraints <- function(primer.df, template.df, settings,
    active.constraints = names(constraints(settings)),
    to.compute.constraints = active.constraints, for.shiny = FALSE, updateProgress = NULL) {
    if (length(settings) == 0 || !is(settings, "DesignSettings")) {
        stop("check_constraints: Please provide a 'DesignSettings' object for 'settings'.")
    }
    if (!is(primer.df, "Primers")) {
        stop("Please input a 'Primers' object for 'primer.df'.")
    }
    if (!is(template.df, "Templates")) {
        stop("Please input a valid 'Templates' object for 'template.df'.")
    }
    mode.directionality <- get.analysis.mode(primer.df)
    # only allow active.constraints for which a setting exists
    ok.constraints <- active.constraints %in% names(constraints(settings))
    bad.constraints <- which(!ok.constraints)
    if (length(bad.constraints) != 0) {
        warning("The following constraints were not found in 'settings' and excluded: \n",
                paste0(active.constraints[bad.constraints], collapse = ", ")) 
    }
    active.constraints <- active.constraints[ok.constraints]
    if (is.null(to.compute.constraints)) {
        # if NULL, compute all constraints
        to.compute.constraints <- active.constraints
    }
    to.compute.constraints <- to.compute.constraints[to.compute.constraints %in% active.constraints]
    if (length(active.constraints) == 0) {
        # nothing to check/compute
        warning("No 'active.constraints' available: no constraints checked!")
        primer.df$constraints_passed <- rep(TRUE, nrow(primer.df))
        return(primer.df)
    }
    PCR.settings <- PCR(settings)
    other.settings <- conOptions(settings)
    constraint.values <- compute.constraints(primer.df, mode.directionality, template.df, settings,
        to.compute.constraints, for.shiny = for.shiny, 
        updateProgress = updateProgress)
    constraint.df <- update.constraint.values(primer.df, constraint.values)  # update constraint.df with computed values
    eval <- eval.constraints(constraint.df, constraints(settings), active.constraints, 
        mode.directionality, primer.df)
    idx.replace <- which(colnames(eval) %in% colnames(constraint.df))
    idx.add <- which(!(colnames(eval) %in% colnames(constraint.df)))
    if (length(idx.replace) != 0) {
        constraint.df[, colnames(eval)[idx.replace]] <- eval[, idx.replace]
    }
    if (length(idx.add) != 0) {
        constraint.df[, colnames(eval)[idx.add]] <- eval[, idx.add]
    }
    if (length(eval) != 0) {
        not.fulfilled.idx <- which(apply(eval, 1, function(x) any(!x, na.rm = TRUE)))
    } else {
        not.fulfilled.idx <- NULL  # everything fulfilled if there's nothing to check ..
    }
    constraint.pass <- rep(TRUE, nrow(constraint.df))
    constraint.pass[not.fulfilled.idx] <- FALSE
    constraint.df$constraints_passed <- constraint.pass
    return(constraint.df)
}
#' Batch Procedure for Evaluating Primer Sets.
#' 
#' @param primer.data A list of objects of class \code{Primers}.
#' @param template.data A list of objects of class \code{Templates}
#' corresponding to \code{primer.data}.
#' @param settings An object of class \code{DesignSettings}.
#' @param active.constraints A character vector providing identifiers
#' of constraints to be considered.
#' @param to.compute.constraints A character vector providing identifiers
#' of constraints to be computed.
#' @param for.shiny A logical indicating whether the results are
#' indicated for the Shiny app or not.
#' @param updateProgress A callback function to track progress in the Shiny app.
#' @return A list with objects of class \code{Primers}.
#' @keywords internal
#' @examples
#' \dontrun{
#' data(Comparison)
#' eval.data <- check_constraints_comparison(primer.data[1:2], template.data[1:2], settings)
#' }
check_constraints_comparison <- function(primer.data, template.data, settings,
    active.constraints = names(constraints(settings)),
    to.compute.constraints = active.constraints, 
    for.shiny = FALSE, updateProgress = NULL) {
   
    # TODO: could provide this function in the pkg as an exported function by making 'check_constraints' a generic function
    # check input types
    primers.ok <- sapply(primer.data, function(x) is(x, "Primers"))
    if (any(!primers.ok)) {
        stop("Not all 'primer.data' elements were of class 'Primers'.")
    }
    templates.ok <- sapply(template.data, function(x) is(x, "Templates"))
    if (any(!templates.ok)) {
        stop("Not all 'template.data.data' elements were of class 'Templates'.")
    }
    if (length(primer.data) != length(template.data)) {
        stop("Number of primer and template sets didn't correspond!")
    }
    for (i in seq_along(primer.data)) {
        if (is.function(updateProgress)) {
            detail <- paste0("Evaluating set ", i)
            updateProgress(1/length(primer.data), detail, "inc")
        }        
        primer.df <- primer.data[[i]]
        template.df <- template.data[[i]]
        primer.df <- check_constraints(primer.df, template.df, settings, 
                    active.constraints = names(constraints(settings)),
                    to.compute.constraints = active.constraints, 
                    for.shiny = for.shiny, updateProgress = NULL)
        primer.data[[i]] <- primer.df

    }
    return(primer.data)
}

#' Evaluation of Coverage Constraints.
#'
#' Computes the biochemical properties specified in the
#' \code{settings} object and determines whether the primers
#' fulfill the required constraints. 
#' 
#' @param primer.df A \code{Primers} object containing the primers
#' to be checked.
#' @param template.df A \code{Templates} object containing the 
#' template sequences corresponding to the primers.
#' @param settings A \code{DesignSettings} object containing the 
#' coverage constraints to be checked and their settings.
#' @param active.constraints Identifiers of constraints that are to be checked.
#' @param to.compute.constraints Constraints that are to be computed.
#' @param for.shiny Whether to format output for HTML.
#' @param updateProgress Progress callback function for shiny.
#' @return A \code{Primers} object with with columns for each constraint in \code{active.constraints}. 
#' @note Please note that some constraints can only be computed if additional software is installed,
#' please see \code{\link{DesignSettings}} for an overview.
#' @keywords internal
check_cvg_constraints <- function(primer.df, template.df, settings,
                                  active.constraints = names(cvg_constraints(settings)),
    to.compute.constraints = active.constraints, for.shiny = FALSE, updateProgress = NULL) {
    
    if (length(settings) == 0 || !is(settings, "DesignSettings")) {
        stop("check_constraints: Please check the input settings")
    }
    if (!is(primer.df, "Primers")) {
        stop("Please input a primer data frame.")
    }
    if (!is(template.df, "Templates")) {
        stop("Please input a valid template data frame.")
    }
    if (!"primer_coverage" %in% colnames(primer.df)) {
        stop("Please compute the primer coverage first before checking the coverage constraints.")
    }
    mode.directionality <- get.analysis.mode(primer.df)
    # only allow active constraints for which a setting exists
    active.constraints <- active.constraints[active.constraints %in% names(cvg_constraints(settings))]

    to.compute.constraints <- to.compute.constraints[to.compute.constraints %in% active.constraints]
    if (length(active.constraints) == 0) {
        # nothing to check/compute
        return(primer.df)
    }
    #########
    # consider on-target binding events to determine coverage
    ########
    constraint.values <- compute.constraints(primer.df, mode.directionality, template.df, settings,
        to.compute.constraints, for.shiny = for.shiny, updateProgress = updateProgress)
    constraint.df <- update.constraint.values(primer.df, constraint.values)  # update constraint.df with computed values
    eval.t <- evaluate.template.constraints(constraint.values, cvg_constraints(settings), 
                                            active.constraints, mode.directionality)
    eval.d <- lapply(eval.t, function(x) unlist(lapply(x, function(y) paste(y, collapse = ","))))
    eval.d <- do.call(cbind, eval.d)
    final.res <- merge.template.decisions(eval.t)
    idx.replace <- which(colnames(eval.d) %in% colnames(constraint.df))
    idx.add <- which(!(colnames(eval.d) %in% colnames(constraint.df)))
    if (length(idx.replace) != 0) {
        # don't use drop = TRUE here, we need individual column if we just have a single element
        constraint.df[, colnames(eval.d)[idx.replace]] <- eval.d[, idx.replace]
    }
    if (length(idx.add) != 0) {
        constraint.df[, colnames(eval.d)[idx.add]] <- eval.d[, idx.add]
    }
    # add template constraint result:
    if (length(final.res) != 0) {
        constraint.df$constraints_passed_T <- unlist(lapply(final.res, function(x) paste(x, 
            collapse = ",")))
    } 
    ##########
    # consider off-target binding constraints to update primer specificities
    ########
    # only check efficiency & deltaG constraints for off coverage events
    off.constraints <- active.constraints[active.constraints %in% c("primer_efficiency", "annealing_DeltaG", "coverage_model")]
    off.constraint.settings <- cvg_constraints(settings)[off.constraints]
    if (length(off.constraint.settings) != 0) {
        off.constraints <- paste0("off_", off.constraints) # compute constraints for off-target events 
        names(off.constraint.settings) <- off.constraints
    }
    off.constraint.vals <- compute.constraints(constraint.df, mode.directionality, template.df, settings,
                                                   off.constraints, mode.directionality)
    eval.t <- evaluate.template.constraints(off.constraint.vals, off.constraint.settings,
                                                off.constraints, mode.directionality)
    eval.d <- lapply(eval.t, function(x) unlist(lapply(x, function(y) paste(y, collapse = ","))))
    eval.d <- do.call(cbind, eval.d)
    final.res <- merge.template.decisions(eval.t)
    # add off-target info to result
    idx.replace <- which(colnames(eval.d) %in% colnames(constraint.df))
    idx.add <- which(!(colnames(eval.d) %in% colnames(constraint.df)))
    if (length(idx.replace) != 0) {
        # don't use drop = TRUE here, we need individual column if we just have a single element
        constraint.df[, colnames(eval.d)[idx.replace]] <- eval.d[, idx.replace]
    }
    if (length(idx.add) != 0) {
        constraint.df[, colnames(eval.d)[idx.add]] <- eval.d[, idx.add]
    }
    if (length(final.res) != 0) {
        # update off-target binding event values:
        constraint.df$constraints_passed_off_T <- unlist(lapply(final.res, function(x) paste(x, collapse = ",")))
    }
    return(constraint.df)
}


#' Retrive Constraint Indices.
#'
#' Gets the index of the required constraint columns in the primer data frame.
#' 
#' @param active.constraints The names of the constraints for which to find the indices in \code{constraint.df}.
#' @param constraint.df The primer data frame where the \code{active.constraints} should be found.
#' @return Indices of \code{active.constraints} in \code{constraint.df}.
#' @keywords internal
get.constraint.value.idx <- function(active.constraints, constraint.df) {
    # important: mapping from constraints to entries in constraint.df
    CONSTRAINTS.TO.VALUE.MAPPING <- list(primer_coverage = "primer_coverage", 
        primer_length = c("primer_length_fw", "primer_length_rev"), 
        primer_specificity = "primer_specificity", 
        gc_clamp = c("gc_clamp_fw", "gc_clamp_rev"), 
        gc_ratio = c("gc_ratio_fw", "gc_ratio_rev"), 
        no_runs = c("no_runs_fw", "no_runs_rev"), 
        no_repeats = c("no_repeats_fw", "no_repeats_rev"), 
        self_dimerization = "Self_Dimer_DeltaG", 
        cross_dimerization = "Cross_Dimer_DeltaG", 
        melting_temp_range = "melting_temp", 
        melting_temp_diff = "melting_temp_diff",
        secondary_structure = "Structure_deltaG", 
        # coverage constraints:
        primer_efficiency = "primer_efficiency",
        annealing_DeltaG = "annealing_DeltaG",
        stop_codon = "stop_codon",
        substitution = "substitution",
        terminal_mismatch_pos = "terminal_mismatch_pos",
        coverage_model = "coverage_model")
    if (!all(active.constraints %in% names(CONSTRAINTS.TO.VALUE.MAPPING))) {
        msg <- paste(active.constraints[!active.constraints %in% names(CONSTRAINTS.TO.VALUE.MAPPING)], collapse = ",")
        stop(paste("active.constraints: incorrect values. The following ",
                "constraints were not available: ", msg, sep = ""))
    }
    check.active <- sapply(active.constraints, function(x) any(!CONSTRAINTS.TO.VALUE.MAPPING[[x]] %in% 
        colnames(constraint.df)))
    if (any(check.active)) {
        idx <- which(check.active)
        warning(paste("Some active constraints were not contained in constraint.df according to the constraints to value mapping: ", 
            paste(active.constraints[check.active], collapse = ","), sep = ""))
    }
    value.idx <- lapply(CONSTRAINTS.TO.VALUE.MAPPING[active.constraints[!check.active]], function(x) match(x, 
        colnames(constraint.df)))
    return(value.idx)
}
#' Evaluation of Primers for Comparison
#'
#' Evaluate multiple primer sets according to the input constraint settings.
#'
#' @param primer.data List with primer data frames.
#' @param constraint.settings List with constraint.settings.
#' @return List with evaluated primer data frames.
#' @keywords internal
eval.comparison.primers <- function(primer.data, constraint.settings) {
    
    active.constraints <- names(constraint.settings)
    new.data <- primer.data
    for (i in seq_along(primer.data)) {
        constraint.df <- primer.data[[i]]
        mode.directionality <- get.analysis.mode(constraint.df)
        eval.df <- eval.constraints(constraint.df, constraint.settings, active.constraints, 
            mode.directionality, constraint.df)
        primer.df <- update.constraint.values(constraint.df, eval.df)
        new.data[[i]] <- primer.df
    }
    return(new.data)
}
#' Evaluation of Constraints'
#' 
#' Evaluates whether the given primer data frame fulfills the required conditions.
#'
#' Constraint values should be contained in \code{constraint.df}. 
#' For each constraint in \code{active.constraints}, a boolean column with the name
#' \code{EVAL_<constraint_name>} is generated, which indicates whether 
#' a primer in a given rows fulfills a constraint or not.
#'
#' @param constraint.df Primer data frame with computed constraints.
#' @param constraint.settings List with allowed values pers constraint.
#' @param active.constraints Names of constraints to be evaluated.
#' @param mode.directionality Directionality of primers
#' @param primer.df Primer data frame corresponding to \code{constraint.df}.
#' @return Augments the \code{constraint.df} data frame with evaluation columns.
#' @keywords internal
eval.constraints <- function(constraint.df, constraint.settings, active.constraints, 
    mode.directionality = c("fw", "rev", "both"), primer.df) {
    # evaluates constraints, i.e. determines whether constraints are fulfilled or
    # not. Requires that constraints have been precomputed.  Returns a data frame
    # with EVAL_<constraint_name> column, indicating whether the constraints of
    # individual primers (rows) are fulfilled or not.
   
    if (length(mode.directionality) == 0) {
        stop("Please supply the 'mode.directionality' argument.")
    }
    mode.directionality <- match.arg(mode.directionality)
    # exclude constraints that shouldn't be checked
    template.constraints <- c("primer_efficiency", "annealing_DeltaG", "stop_codon", "substitution", "terminal_mismatch_pos", "hexamer_coverage", "coverage_model")
    active.constraints <- active.constraints[!active.constraints %in% template.constraints]
    if (length(active.constraints) == 0) {
        return(NULL)
    }
    # use the global variable constraints.to.value.mapping to identify the value
    # columns corresponding to the constraints:
    value.idx <- get.constraint.value.idx(active.constraints, constraint.df)  # idx of constraint values in constraint.df
    # determine if we should constrain both fw and rev, or only one direction
    fw.idx <- which(primer.df$Forward != "")
    rev.idx <- which(primer.df$Reverse != "")
    all.constraints <- vector("list", length(active.constraints))
    names(all.constraints) <- paste("EVAL_", active.constraints, sep = "")
    for (i in seq_along(value.idx)) {
        active.constraint <- names(value.idx)[i]
        #message(paste("evaluating constraint: ", active.constraint, sep = ""))
        constraint.idx <- value.idx[[i]]
        constraints <- constraint.settings[[active.constraint]]
        if (any(is.null(names(constraints)) | !names(constraints) %in% c("min", "max"))) {
            msg <- paste("Constraint did not have a proper annotation of min/max setting: ", active.constraint, sep = "")
            stop(msg)
        }
        min.condition <- unname(constraints["min"])
        max.condition <- unname(constraints["max"])
        if (is.na(min.condition)) {
            min.condition <- NULL
        } else if (is.na(max.condition)) {
            max.condition <- NULL
        }
        if (length(constraint.idx) == 2) {
            # fw and rev values are available for current constraint
            eval <- apply.constraint(constraint.df[, constraint.idx[1]], constraint.df[, 
                constraint.idx[2]], min.condition, max.condition, fw.idx, rev.idx, 
                mode.directionality)
        } else {
            eval <- apply.constraint(constraint.df[, constraint.idx[1]], constraint.df[, 
                constraint.idx[1]], min.condition, max.condition, fw.idx, rev.idx, 
                mode.directionality)
        }
        all.constraints[[i]] <- eval
    }
    if (mode.directionality == "both") {
        all.constraints <- lapply(all.constraints, function(y) apply(y, 1, function(x) all(x, 
            na.rm = TRUE)))
    } else if (mode.directionality == "fw") {
        # message(all.constraints)
        all.constraints <- lapply(all.constraints, function(y) y[, "Forward"])
    } else {
        all.constraints <- lapply(all.constraints, function(y) y[, "Reverse"])
    }
    constraint.df <- do.call(cbind, all.constraints)
    return(constraint.df)
}

#' Filter Multiple Primer Sets.
#' 
#' Filters multiple primer sets at once.
#'
#' @param primers List with primer data frames.
#' @param templates List with template data frames.
#' @param active.constraints Strings giving the constraints that are to be checked.
#' @param settings List with settings.
#' @param updateProgress Progress callback function for shiny.
#' @return A list with filtered primer data frames.
#' @keywords internal
filter.comparison.primers <- function(primers, templates, active.constraints, 
    settings, updateProgress = NULL) {

    names <- names(primers)
    no.structures <- TRUE
    for (i in seq_along(primers)) {
        primer.df <- primers[[i]]
        template.df <- templates[[i]]
        mode.directionality <- get.analysis.mode(primer.df)
        filtered.data <- cascaded.filter.quick(primer.df, template.df, settings, 
            active.constraints, mode.directionality,
            active.constraints, no.structures, updateProgress)
        filtered.result <- filtered.data$data
        template.df <- update_template_cvg(template.df, primer.df, mode.directionality)
        primers[[i]] <- filtered.result
        templates[[i]] <- template.df
    }
    names(primers) <- names
    result <- list(primers = primers, templates = templates)
    #print("out len:")
    #print(length(result$primers))
    #print("out len templates:")
    #print(length(result$templates))
    return(result)
}

#' @rdname PrimerEval
#' @name PrimerEval
#' @aliases filter_primers
#' @return \code{filter_primers} returns a \code{Primers} object containing
#' only those primers fulfilling all specified constraints.
#' @export
#' @examples
#' 
#' data(Ippolito)
#' filename <- system.file("extdata", "settings", 
#'              "C_Taq_PCR_high_stringency.xml", package = "openPrimeR")
#' settings <- read_settings(filename)
#' # Only retain the primers fulfilling the GC clamp constraint:
#' filtered.df <- filter_primers(primer.df, template.df, settings,
#'                  active.constraints = c("gc_ratio"))
filter_primers <- function(primer.df, template.df, settings,
                    active.constraints = names(constraints(settings))) {
   
    # nb: only input_filtering_constraint settings are considered
    if (length(settings) == 0 || !is(settings, "DesignSettings")) {
        stop("Please provide a DesignSettings object.")
    }
    if (!is(primer.df, "Primers")) {
        stop("Please input a primer data frame.")
    }
    if (!is(template.df, "Templates")) {
        stop("Please provide a valid template data frame.")
    }
    filter.result <- cascaded.filter.quick(primer.df, template.df, settings, 
        to.compute.constraints = active.constraints, 
        mode.directionality = get.analysis.mode(primer.df), 
        active.constraints = active.constraints, 
        updateProgress = NULL)
    return(filter.result$data) # return remaining primers after filtering
}
#' Filter By Constraints
#' 
#' Remove primers that do not fulfill the current constraints (evaluate all primers).
#'
#' @param filtered.df Primer data frame.
#' @param constraint.df Data frame with constraint values.
#' @param current.constraints List with constraint settings.
#' @param active.constraints Strings giving the names of active constraints.
#' @param mode.directionality Direction of primers
#' @param template.df Template data frame.
#' @return A list containing the filtered primer data frame, as well as a 
#' data frame of the excluded primers and the used filtering settings.
#' @keywords internal
filter.by.constraints <- function(filtered.df, constraint.df, current.constraints, 
    active.constraints, mode.directionality = c("fw", "rev", "both"), template.df) {
    
    if (length(mode.directionality) == 0) {
        stop("Please supply the 'mode.directionality' argument.")
    }
    mode.directionality <- match.arg(mode.directionality)
    #print(colnames(constraint.df))
    # evaluate currently active constraints
    eval.df <- eval.constraints(constraint.df, current.constraints, active.constraints, 
        mode.directionality, filtered.df)
    eval.t <- evaluate.template.constraints(constraint.df, current.constraints,
                                active.constraints, mode.directionality)
    # add hexamer coverage constraint if present
    eval.d <- lapply(eval.t, function(x) unlist(lapply(x, function(y) paste(y, collapse = ","))))
    eval.d <- do.call(cbind, eval.d)
    final.res <- merge.template.decisions(eval.t)
    idx.add <- which(!colnames(eval.df) %in% constraint.df)
    idx.rep <- which(colnames(eval.df) %in% constraint.df)
    add.df <- constraint.df
    if (length(idx.add) != 0) {
        add.df[, colnames(eval.df)[idx.add]] <- eval.df[, idx.add]
    }
    if (length(idx.rep) != 0) {
        add.df[, colnames(eval.df)[idx.rep]] <- eval.df[, idx.rep]
    }
    m <- match(colnames(add.df), colnames(filtered.df))
    # filtered.df becomes matrix here ...
    if (any(!is.na(m))) {
        filtered.df[, m[which(!is.na(m))]] <- add.df[, which(!is.na(m)), drop = FALSE]
    }
    if (any(is.na(m))) {
        filtered.df <- cbind(filtered.df, add.df[, which(is.na(m)), drop = FALSE])
    }
    # consider template-based constraints: remove covered seqs
    if ("hexamer_coverage" %in% active.constraints) {
        # don't filter coverage events with complementary 3' ends
        # adjust selection of coverage events: free-pass for hexa cvg
        hexa <- lapply(strsplit(constraint.df$hexamer_coverage, split = ","), function(x) as.logical(x))
        final.res <- lapply(seq_along(final.res), function(x) hexa[[x]] | final.res[[x]])
    }
    if (length(final.res) != 0) {
        # only update template-specific coverage data if we compute template-coverage specific constraints
        # determine indices of cvg events to be retained:
        sel <- lapply(final.res, function(x) which(x)) # selection of coverage events to retain according to cvg constraints
        filtered.df <- update.cvg.data(filtered.df, sel, template.df, mode = "on_target", active.constraints)
        if ("primer_coverage" %in% colnames(constraint.df)) {
            if ("constraints_passed_off_T" %in% colnames(constraint.df)) {
                # filtering of off-target binding events to update specificity
                off.sel <- lapply(strsplit(constraint.df$constraints_passed_off_T, split = ","), function(x) which(as.logical(x)))
            } else {
                # no off-target events filtered -> select all
                off.sel <- lapply(strsplit(constraint.df$Off_Covered_Seqs, split = ","), function(x) seq_len(length(x)))
            }
            filtered.df <- update.cvg.data(filtered.df, off.sel, template.df, mode = "off_target", active.constraints)
        }
    }
    # filter primers by considering all computed constraints
    passed <- rep(TRUE, nrow(filtered.df))
    if (length(eval.df) != 0) {
        passed <- apply(eval.df, 1, all)
    }
    # filter primer.df
    excluded <- NULL
    passed.idx <- which(passed)
    if (length(which(!passed) != 0)) {
        # primers were excluded
        excluded <- filtered.df[which(!passed), , drop = FALSE]
    }
    if (length(passed.idx) != 0) {
        filtered.df <- filtered.df[which(passed), ]
    } else {
        # nothing passed the constraints
        filtered.df <- filtered.df[FALSE, ]
    }
    result <- list(Filtered = filtered.df, Excluded = excluded, Passed_Idx = passed.idx)
    return(result)
}
#' Cascaded Filter
#'
#' Filter primers in a cascaded fashion. 
#'
#' At each constraint evaluation all primers that do not fulfill the current constraint are removed.
#'
#' Constraints that are specified in \code{to.compute.constraints} are computed on the fly.
#'
#' @param primer.df Primer data frame.
#' @param template.df Template data frame.
#' @param settings Settings object.
#' @param to.compute.constraints Names of constraints that still have to be computed.
#' @param mode.directionality Primer direction.
#' @param active.constraints The constraints that are to be used for filtering.
#' If \code{active.constraints} is \code{NULL}, all filtering constraints are used.
#' @param no.structures Whether dimerization structures shall be computed.
#' @param updateProgress Progress callback function for shiny.
#' @return The filtered primer data frame.
#' @keywords internal
cascaded.filter.quick <- function(primer.df, template.df, settings, 
    to.compute.constraints, 
    mode.directionality = c("fw", "rev", "both"), 
    active.constraints = NULL, no.structures = FALSE, updateProgress = NULL) {

    if (length(primer.df) == 0 || length(template.df) == 0) {
        return(NULL)
    }
    if (!is(primer.df, "Primers")) {
        stop("Please supply a 'Primers' object.")
    }
    if (!is(template.df, "Templates")) {
        stop("Please supply a 'Templates' object.")
    }
    if (!is(settings, "DesignSettings")) {
        stop("Please supply a 'DesignSettings' object.")
    }
    PCR_conditions <- PCR(settings)
    other.settings <- conOptions(settings)
    allowed.region.definition <- other.settings$allowed_region_definition
    if (length(mode.directionality) == 0) {
        stop("Please supply the 'mode.directionality' argument.")
    }
    mode.directionality <- match.arg(mode.directionality)
    filtered.df <- primer.df
    if (is.null(active.constraints)) {
        # no constraints provided -> use all constraints to compute
        active.constraints <- to.compute.constraints
    }
    # use specified constraints 
    current.constraints <- constraints(settings)
    m <- match(active.constraints, names(current.constraints))
    if (any(is.na(m))) {
        # couldn't find an input constraint
        stop(paste("Unknown 'active.constraints' specified.",
            "The constraints should be an available constraint from the input",
            "settings object. Unknown constraints are: ",
            paste(active.constraints[is.na(m)], collapse = ",")))
    }
    current.constraints <- current.constraints[m]
    if (length(current.constraints) == 0) {
        return(list(data = primer.df, stats = NULL, excluded = NULL, used_settings = settings))  # nothing to filter 
    }
    stat.N <- length(current.constraints)  # +1 for initial nbr of primers
    stat.df <- data.frame(Constraint = names(current.constraints), Direction = mode.directionality, Remaining = rep(0, stat.N), Excluded = rep(0, 
        stat.N), Current_Coverage = rep(0, 
        stat.N), Time = rep(0, stat.N), stringsAsFactors = FALSE)  # determine statistics about filtering
    max.cvg <- get_cvg_ratio(filtered.df, template.df)
    # load settings from input data
    excluded.df <- primer.df[FALSE, ]
    last.constraints <- current.constraints
    for (i in seq_along(current.constraints)) {
        active.constraints <- names(current.constraints)[i]
        #print(active.constraints)
        ####### Time measurement for filtering step
        ptm <- proc.time()
        if (active.constraints %in% to.compute.constraints) {
            # value should be computed
            constraint.df <- compute.constraints(filtered.df, mode.directionality, 
                template.df, settings, active.constraints, 
                no.structures = no.structures,
                updateProgress = updateProgress)
        } else {
            constraint.df <- filtered.df  # value isn't recomputed -> should be in filtered.df
        }
        ptm <- proc.time() - ptm
        ####### Time measurement for filtering step ends
        filtered.df.pure <- filtered.df  # wasn't filtered yet :-)
        # evaluate currently active constraints
        filter.result <- filter.by.constraints(filtered.df, constraint.df, current.constraints, 
            active.constraints, mode.directionality, template.df)
        filtered.df <- filter.result$Filtered
        excluded <- filter.result$Excluded
        if (length(excluded) != 0) {
            excluded.df <- my_rbind(excluded.df, excluded)
        }
        max.cvg <- get_cvg_ratio(filtered.df, template.df)
        stat.df[i, "Constraint"] <- active.constraints
        stat.df[i, "Current_Coverage"] <- max.cvg
        nbr.excluded <- ifelse(length(excluded) == 0, 0, nrow(excluded))
        stat.df[i, "Excluded"] <- nbr.excluded
        stat.df[i, "Remaining"] <- nrow(filtered.df)  # included/remaining observations
        stat.df[i, "Time"] <- ptm["elapsed"]
        percentage.remaining <- paste0(round(nrow(filtered.df) / nrow(primer.df) * 100, 2), "%")
        message("\to ", active.constraints, ": ", nrow(filtered.df), " primers remaining (", percentage.remaining, ")")
        if (nrow(filtered.df) == 0) {
            # nothing more to filter.
            break
        }
    }
    s <- 1  # start with first entry now (no more initial)
    e <- length(stat.df$Constraint)
    excluded.df$Exclusion_Reason <- unlist(lapply(s:e, function(x) rep(stat.df$Constraint[x], 
        stat.df$Excluded[x])))
    # does excluded.df have some colnames missing?
    diff <- setdiff(colnames(filtered.df), colnames(excluded.df))
    if (length(diff) != 0 && nrow(excluded.df) != 0) {
        excluded.df[, diff] <- rep(NA, nrow(excluded.df))
    }
    # update melting_temp_diff
    if (length(filtered.df) != 0 && "melting_temp_diff" %in% colnames(filtered.df) && nrow(filtered.df) != 0) {
        temp.diff <- get.melting.temp.diff(filtered.df$Tm_C_fw, filtered.df$Tm_C_rev)
        filtered.df[, "melting_temp_diff"] <- temp.diff
    }
    stat.df$Excluded_Percentage <- stat.df$Excluded/(stat.df$Remaining + stat.df$Excluded)
    result <- list(data = filtered.df, stats = stat.df, excluded = excluded.df)
    return(result)
}

#' Writes Filtering Data Sets to Disk.
#' @param filtered.df A filtered \code{Primers} set.
#' @param excluded.df A set of \code{Primers} that were excluded.
#' @param results.loc The location where to store the data.
#' @param tag A tag for the output files.
#' @param stat.df Data frame with statistics of the filtering procedure.
#' @param settings A \code{DesignSettings} object.
#' @return No return value, writes output to disk.
#' @keywords internal
store.filtering.sets <- function(filtered.df, excluded.df, results.loc, tag = "", stat.df = NULL, settings = NULL) {
    if (length(results.loc) != 0) {
        write.csv(excluded.df, file = file.path(results.loc, paste0("excluded_primers_", tag, ".csv")), 
            row.names = FALSE)
        write.csv(filtered.df, file = file.path(results.loc, paste0("filtered_primers_", tag, ".csv")), 
            row.names = FALSE)
        if (length(stat.df) != 0) {
            write.csv(stat.df, file = file.path(results.loc, paste0("filtered_primers_stats_", tag, ".csv")), 
                row.names = FALSE)
        }
        if (length(settings) != 0) {
            write_settings(settings, file.path(results.loc, paste0("relaxed_settings_", tag, ".xml")))
        }
    }
    invisible()
}
#' Filtering for the Optimization
#' 
#' Filter primers according to constraints and relax constraints if necessary.
#'
#' Constraints are relaxed if the \code{required.cvg} could not be reached with the input constraints.
#' 
#' @param primer.df Primer data frame.
#' @param template.df Template data frame.
#' @param settings Settings object.
#' @param mode.directionality Primer direction.
#' @param required.cvg Required ratio of covered templates.
#' If \code{required.cvg} is set to 0, the constraints are not relaxed.
#' @param target.temps Target melting temperature of the primers in Celsius.
#' This argument is only required if we try to match the melting temperatures of another primer set,
#' e.g. when first optimizing forward and then optimizing reverse primers.
#' @param updateProgress Progress callback function for shiny.
#' @param results.loc Directory where the filtering results should be stored.
#' @return The filtered primer data frame with respect to \code{required.cvg}.
#' @keywords internal
cascaded.filter <- function(primer.df, template.df, settings, mode.directionality = c("fw", "rev", "both"), 
                            required.cvg = 1, target.temps = NULL, 
                            updateProgress = NULL, results.loc = NULL) {
    
    if (length(mode.directionality) == 0) {
        stop("Please supply the 'mode.directionality' argument.")
    }
    mode.directionality <- match.arg(mode.directionality)
    constraint.settings <- filters(settings)
    to.compute.constraints <- names(constraint.settings)  # compute all constraints given as input
    message("a) Filtering of primers")
    # set no.structures to TRUE since this is the design algorithm
    filtered.data <- cascaded.filter.quick(primer.df, template.df, settings, 
        to.compute.constraints, mode.directionality, no.structures = TRUE,
        updateProgress = updateProgress)  # filter without relaxation
    if (length(filtered.data$excluded) == 0) {
        return(list(excluded = filtered.data$excluded, data = filtered.data$data, 
            stats = filtered.data$stats, used_settings = settings, 
            stats_relaxed = NULL))
    }
    excluded.df <- filtered.data$excluded
    filtered.df <- filtered.data$data
    stat.df <- filtered.data$stats
    # relax here:
    stat.df$Excluded_Percentage <- stat.df$Excluded/(stat.df$Remaining + stat.df$Excluded)
    # save results up to now
    store.filtering.sets(filtered.df, excluded.df, results.loc, tag = "pre", stat.df = stat.df)
    message("b) Relaxation of constraints")
    # relax constraints:
    relaxed.data <- relax.constraints(settings, 
        filtered.df, excluded.df, stat.df, template.df, mode.directionality, 
        required.cvg, target.temps = target.temps,
        results.loc = results.loc)
    excluded.df <- relaxed.data$excluded
    filtered.df <- relaxed.data$filtered  # included seqs
    stats.relax <- relaxed.data$stats
    relaxed.settings <- relaxed.data$new_settings
    #max.cvg <- get_cvg_ratio(filtered.df, template.df)
    result <- list(data = filtered.df, stats = stat.df, excluded = excluded.df, 
                   used_settings = relaxed.settings, stats_relax = stats.relax)
    return(result)
}
#' Uncovered Templates.
#'
#' Computes a data frame containing the templates that are
#' not yet covered.
#'
#' @param filtered.df An object of class \code{Primers}.
#' @param template.df An object of class \code{Templates}.
#' @return A Templates data frame containing the missing templates.
#' @keywords internal
get.missing.df <- function(filtered.df, template.df, Tm.brackets, settings, mode.directionality) {
    # NB: this could also be done for individual temperature sets instead
    # of all primers in one go
    if (!"Covered_Seqs" %in% colnames(filtered.df)) {
        covered.idx <- NULL
    } else {
        # get covered templates per temperature-set
        cvg.info <- get_max_set_coverage(filtered.df, template.df, Tm.brackets, settings, mode.directionality, max.only = FALSE)
        covered.idx <- lapply(cvg.info, function(x) match(attr(x,"covered_templates"), template.df$ID))
    }
    # determine primer candidates that can still improve the coverage
    missing.idx <- NULL
    if (length(covered.idx) == 0) {
        missing.idx <- seq_len(nrow(template.df))
    } else {
        to.cover <- seq_len(nrow(template.df))
        # consider all uncovered templates from all temperature sets:
        missing.idx <- unique(unlist(lapply(covered.idx, function(x) setdiff(to.cover, x))))
    }
    missing.df <- template.df[missing.idx, ]
    return(missing.df)
}
#' Computation of Coverage Gain.
#'
#' Computes the coverage gain from \code{covered.seqs}.
#'
#' @param covered.seqs List with covered sequences.
#' @param template.df A \code{Templates} data frame.
#' @param missing.df A \code{Templates} data frame 
#' containing only the templates that still need to be covered.
#' @param candidate.df A \code{Primers} data frame containing
#' candidate primers.
#' @param con.names The constraint to evaluate the coverage gain for
#' upon being relaxed.
#' @param constraint.limits A list with constraint limits.
#' @param feasible.only Whether only feasible coverage gains 
#' are to be outtputed. Here, \emph{feasible} relates to 
#' coverage gains that can be obtained directly with the next relaxation.
#' @return The number of covered sequences to be gained.
#' @keywords internal
get.cvg.gain <- function(covered.seqs, template.df, missing.df, candidate.df, 
                        con.names, constraint.limits, feasible.only = FALSE) {
    # identify the coverage for covered.seqs for a constraint
    # covered seqs: identifiers of covered seqs
    con.name <- unique(con.names)
    cur.limit <- constraint.limits[[con.name]]
    values <- get.constraint.values(unique(con.names), candidate.df, "both")
    fw.idx <- which(candidate.df$Forward != "")
    rev.idx <- which(candidate.df$Reverse != "")
    covered.idx <- covered.seqs.to.idx(covered.seqs, template.df)
    if (feasible.only) {
        feasible <- apply.constraint(values$fw, values$rev, cur.limit["min"], 
                         cur.limit["max"], fw.idx, rev.idx, "both")
        sel <- which(apply(feasible, 1, all))
        covered.idx <- covered.idx[sel]
    }
    unique.covered.idx <- unique(unlist(lapply(covered.idx, function(x) intersect(missing.df$Identifier, template.df[x, "Identifier"]))))
    cvg <- length(unique.covered.idx)
    return(cvg)
}
#' Gain of Coverage by Excluded Primers.
#'
#' Computes a data frame on the excluded sequences per constraint.
#'
#' @param candidate.df An object of class \code{Primers} containing
#' excluded primers that are considered for addition to \code{filtered.df}.
#' @param constraint.settings A list with the current constraint settings.
#' @param constraint.limits The current constraint limits.
#' @param relax.df Data frame with count of relaxations per constraint.
#' @return A data frame with exclusion data.
#' @keywords internal
build.gain.df <- function(candidate.df, constraint.settings, constraint.limits, relax.df) {
    if (length(candidate.df) == 0 || nrow(candidate.df) == 0) {
        # nothing to be added
        return(NULL)
    }
    # determine cvg gain of primer candidates per exclusion reason
    df <- relax.df
    if (nrow(df) != 0) {
        # only consider constraints for relaxation that are associated with an exclusion reason
        idx <- unlist(lapply(df$Constraint, function(x) length(which(candidate.df$Exclusion_Reason == x))))
        sel.idx <- which(idx != 0)
        # select 'unrelaxed' constraints first and then select by recommended order of relaxation (relax unimportant constraints first)
        relaxable.idx <- which(unlist(lapply(df$Constraint, function(x) 
                                   any(constraint.settings[[x]] != constraint.limits[[x]]))))
        idx <- intersect(sel.idx, relaxable.idx)
        df <- df[idx,]
        df <- df[order(df$Relax_Count, match(df$Constraint, getOption("openPrimeR.relax_order"))), ]
    }
    #message("Coverage gain:")
    #message(df)
    return(df)
}
#' Determination of Maximal Coverage.
#'
#' Determines the maximal coverage ratio of a set of primers 
#' for primer subsets valid for a certain temperature range.
#' a certain melting temperature range.
#'
#' @param primer.df An object of class \code{Primers}.
#' @param template.df An objectc of class \code{Templates}.
#' @param Tm.brackets A data frame with temperature information.
#' @param settings A \code{DesignSettings} object.
#' @param mode.directionality The direction of the primers.
#' @param max.only Whether only the maxium coverage shall be returned.
#' If \code{max.only} is \code{FALSE}, the coverage ratios of all 
#' melting temperature sets according to \code{Tm.brackets} are returned.
#' @return The maximal coverage ratio of a primer set if \code{max.only} is \code{TRUE}
#' or the coverages of all melting temperature seets if \code{max.only} is \code{FALSE}.
#' @keywords internal
get_max_set_coverage <- function(primer.df, template.df, Tm.brackets, settings, mode.directionality, max.only = TRUE) {
    # determine coverage for individual melting temperature ranges to ensure that we have a set with max cvg that is temperature compatible!
    #print("get max set coverage")
    PCR.settings <- PCR(settings)
    other.settings <- conOptions(settings)
    allowed.region.definition <- other.settings$allowed_region_definition
    Tm.sets <- compute.Tm.sets(primer.df, template.df, Tm.brackets, settings, 
        mode.directionality, PCR.settings$primer_concentration, PCR.settings$template_concentration,
        PCR.settings$Na_concentration, PCR.settings$Mg_concentration, 
        PCR.settings$K_concentration, PCR.settings$Tris_concentration,
        other.settings$allowed_mismatches, 
        other.settings$allowed_other_binding_ratio, other.settings$allowed_stop_codons, allowed.region.definition, 
        other.settings$disallowed_mismatch_pos, FALSE, NULL, NULL, NULL, NULL)$sets  # get temperature-filtered sets
    cvg.vals <- lapply(Tm.sets, function(x) get_cvg_ratio(x, template.df)) # get cvg of every Tm set
    if (max.only) {
        cvg.vals <- max(unlist(cvg.vals))
    }
    #print("max set coverage done")
    return(cvg.vals)
}
#' Relaxation of Constraint Limits.
#'
#' Relaxes the constraint limits by moving according to the difference
#' between the \code{initial.limits} and \code{initial.constraints}.
#'
#' @param cur.limits List with current constraint settings.
#' @param initial.limits List with initial coonstraint limits.
#' @param initial.constraints List with initial constraint settings before relaxing.
#' @param con.name The constraint for which the settings are to be changed.
#' @return A list with relaxed constraint limits.
#' @keywords internal
set.new.limits <- function(cur.limits, initial.limits, initial.constraints, con.name = NULL) {
    # relaxes the constraint limits
    # cur.limit: current constraint limits
    # initial.limits: limits for constraints initially
    # initial.constraints: settings for constraints initially

    # 1. Determine the differences between the initial.constraints and the initial.limits 
    # -> determines the speed of the relaxation.
    if (length(con.name) == 0) {
        con.name <- names(cur.limits)
    }
    constraint.delta <- lapply(seq_along(con.name), function(x) initial.limits[[con.name[x]]] - initial.constraints[[con.name[x]]])
    names(constraint.delta) <- con.name
    # 2. update all limits according to the shift from the initial constraints
    new.limits <- cur.limits
    m <- match(con.name, names(new.limits))
    new.limits[con.name] <- lapply(seq_along(con.name), function(x) new.limits[[con.name[x]]] + constraint.delta[[con.name[x]]])
    names(new.limits)[m] <- con.name
    return(new.limits)
}
#' Get the Values of a Constraint.
#'
#' @param con.name The name of the constraint.
#' @param cur.candidates The \code{Primers} data frame where the values should be retrieved.
#' @param mode.directionality The direction for which values should be retrieved.
#' @return The constraint values corresponding to \code{con.name} for the primers \code{cur.candidates}.
#' @keywords internal
get.constraint.values <- function(con.name, cur.candidates, mode.directionality) {
    value.idx <- unlist(get.constraint.value.idx(con.name, cur.candidates))  # idx of constraint values in constraint.df
    val.name <- colnames(cur.candidates)[value.idx]
    values <- asS3(cur.candidates)[, val.name, drop = FALSE]
    rev.col <- grep("_rev", colnames(values))  # index of rev col value
    # for both: retrieve fw & rev!
    if (mode.directionality == "both") {
        modes <- c("fw", "rev")
    } else {
        modes <- mode.directionality
    }
    result <- vector("list", 2)
    names(result) <- c("fw", "rev")
    for (i in seq_along(modes)) {
        mode <- modes[i]
        col.id <- grep(paste("_", mode, sep = ""), colnames(values))  # index of fw col value
        if (length(col.id) != 0) {
            # constraint with fw & rev values
            out.values <- as.numeric(values[, col.id])
        } else {
            # constraint with a single value for both directions
            out.values <- values[, 1]
        }
        result[[mode]] <- out.values
    }
    return(result)
}
#' Augmentation of Primer Coverage.
#'
#' Computes the coverage for the primers in \code{primer.df} 
#' that is still missing such that the relaxation procedure can adjust
#' appropriate constraints.
#'
#' @param primer.df A \code{Primers} object for which
#' the primer coverage shall be augmented.
#' @param template.df A \code{Templates} object.
#' @param settings A \code{DesignSettings} object giving
#' the parameters for coverage computations.
#' @param partial Whether all missing primer coverage values should be 
#' computed. If \code{partial} is \code{TRUE}, only 
#' the coverage values of the primers that were excluded due to
#' the specified \code{constraint} are computed.
#' @param constraint A character vector specifying the exclusion
#' reason for which the partial augmentation should take place.
#' @return A \code{Primers} object with augmented coverage entries.
#' @keywords internal
augment.primer.cvg <- function(primer.df, template.df, settings, 
                               partial = FALSE, constraint = NULL) {
    if (!is(primer.df, "Primers")) {
        stop("Please input a valid primer data frame.")
    }
    if (!is(template.df, "Templates")) {
        stop("Please input a valid template data frame.")
    }
    if (!is(settings, "DesignSettings")) {
        stop("Please input a valid settings object.")
    }
    ###
    # NOTE: 'coverage_model' is NA for excluded.df = primer.df initially
    ###
    is.exclusion.df <- "Exclusion_Reason" %in% colnames(primer.df)
    if (!"primer_coverage" %in% colnames(primer.df)) {
        primer.df$primer_coverage <- NA
        primer.df$Covered_Seqs <- NA
    }
    # compute only missing coverage entries
    idx <- which(is.na(primer.df$primer_coverage))
    ori.idx <- idx
    if (partial && is.exclusion.df && length(constraint) != 0) {
        con.idx <- which(primer.df$Exclusion_Reason == constraint)
        idx <- intersect(idx, con.idx) # primers excluded for 'constraint' with missing cvg
    }
    if (length(idx) == 0) {
        # nothing to adjust
        return(primer.df)
    }
    con.s <- ifelse(length(constraint) != 0, paste0(constraint, ": "), "")
    message("\to ", con.s, "Augmenting primer coverage for ", length(idx), " of ", length(ori.idx), " target primers.")
    augment.df <- check_constraints(primer.df[idx,], template.df, settings, "primer_coverage")
    existing.cols <- intersect(colnames(primer.df), colnames(augment.df))
    primer.df[idx,existing.cols] <- augment.df[, existing.cols]
    new.cols <- setdiff(colnames(augment.df), colnames(primer.df))
    for (i in seq_along(new.cols)) {
        col <- new.cols[i]
        primer.df[idx, col] <- augment.df[idx, col]
    }
    return(primer.df)
}
#' Relaxation of Constraints
#'
#' Relax constraints trying to reach the target coverage ratio.
#'
#' @param settings A \code{DesignSettings} object.
#' @param filtered.df Data set of primers that fulfilled all constraints of the filtering procedure.
#' @param excluded.df Data frame with excluded primers from the first filtering round.
#' @param stat.df Data frame with statistics from filtering.
#' @param template.df Template data frame.
#' @param mode.directionality Primer direction.
#' @param required.cvg Required ratio of covered templates.
#' @param target.temps Target melting temperature values.
#' @param results.loc The location where intermediary results should be stored.
#' @return A list containing information about the relaxation as well as the filtered primers.
#' @keywords internal
relax.constraints <- function(settings, filtered.df, excluded.df, stat.df, template.df, 
                              mode.directionality = c("fw", "rev"), required.cvg, 
                              target.temps = NULL,
                              results.loc = NULL) {

    if (length(results.loc) != 0) {
        # create result location
        results.loc <- file.path(results.loc, "relaxation")
        dir.create(results.loc, showWarnings = FALSE)
    }
    if (length(mode.directionality) == 0) {
        stop("Please supply the 'mode.directionality' argument.")
    }
    mode.directionality <- match.arg(mode.directionality)
    initial.constraints <- filters(settings)
    active.constraints <- names(initial.constraints)
    initial.limits <- filterLimits(settings)
    initial.opti.constraints <- opti(settings)
    initial.opti.limits <- optiLimits(settings)
    new.filtered.df <- filtered.df  # updated filtered data from relaxation
    if (nrow(new.filtered.df) != 0) {
        new.filtered.df$Exclusion_Reason <- NA # add exclusion reason column
    }
    new.excluded.df <- excluded.df
    Tm.brackets <- create.Tm.brackets(new.filtered.df, template.df, settings, target.temps)
    new.filtered.df <- Tm.brackets$primers
    max.cvg <- get_max_set_coverage(new.filtered.df, template.df, Tm.brackets, settings, mode.directionality)
    if (is.na(max.cvg)) {
        stop("The maximal coverage is 'NA', which indicates that ",
             "primer coverage was not available after filtering.",
             "This shouldn't have had happened, sorry.")
    } 
    while.count <- 0
    relax.df <- data.frame("Constraint" = names(initial.constraints), "Relax_Count" = 0, stringsAsFactors=FALSE)
    # count the total gained coverage:
    total.gain.df <- data.frame(Exclusion_Reason = active.constraints,
                                Actual_Gain = 0, stringsAsFactors=FALSE)
    target.cvg <- required.cvg
    # Augment primer coverage for all primers (not only for missing templates since runtime would increase for multiple cvg calls, would need to recompute later)
    #t <- proc.time()["elapsed"]
    missing.df <- NULL
    if (max.cvg < target.cvg) {
        new.excluded.df <- augment.primer.cvg(excluded.df, template.df, settings)
        missing.df <- get.missing.df(new.filtered.df, template.df, Tm.brackets, settings, mode.directionality)
    }
    #t <- proc.time()["elapsed"] - t
    #print(t)
    while (max.cvg < target.cvg) {
        if (length(results.loc) != 0) {
            # store relaxation results
            store.filtering.sets(new.filtered.df, new.excluded.df, results.loc, tag = as.character(while.count), stat.df = relax.df, settings = settings)
        }
        while.count <- while.count + 1
        # improve coverage as long as we haven't reached the target
        if ("Covered_Seqs" %in% colnames(new.filtered.df)) {
            prev.cvg <- unique(unlist(strsplit(new.filtered.df$Covered_Seqs, split = ",")))
        } else {
            # everything was filtered before cvg was computed.
            prev.cvg <- NULL
        }
        # 1. Build Exclusion Reason data frame
        #message(paste("Relaxation run: ", while.count, sep = ""))
        gain.df <- build.gain.df(new.excluded.df, filters(settings), filterLimits(settings), relax.df)
        ######################
        #print("#########")
        #print(gain.df)
        #print("#########")
        #print("Melting temperatures of excluded seqs: ")
        #print(new.excluded.df$melting_temp)
        ########################
        if (length(gain.df) == 0 || nrow(gain.df) == 0) {
            # there's no constraint to relax anymore to bring forth an increase in coverage
            # DEBUG:
            #cvg.to.gain <- unique(unlist(strsplit(new.excluded.df$Covered_Seqs, split = ",")))
            #cur.cvg <- unique(unlist(strsplit(new.filtered.df$Covered_Seqs, split = ",")))
            #possible.gain <- setdiff(cvg.to.gain, cur.cvg)
            if ("melting_temp_diff" %in% names(opti(settings)) && "melting_temp" %in% colnames(new.filtered.df) && 
                any(constraints(settings)$melting_temp_diff != constraintLimits(settings)$melting_temp_diff)) {
                    max.steps <- 10 # limit the number of steps, in case it's just not possible (e.g. cannot relax anymore due to user input)
                    step <- 0
                    # this procedure uses only 'new.filtered.df', assuming that we couldn't relax anymore and it's only a matter of 'Tm diff' to reach the target.
                    while (max.cvg < target.cvg && step < max.steps) {
                        step <- step + 1
                        message("Final melting temp diff relaxation ...")
                        # adjust melting temp diff until we reach target cvg
                        constraints(settings)$melting_temp_diff <- relax.opti.constraints(opti(settings), 
                                                                   initial.opti.limits, initial.opti.constraints)$melting_temp_diff
                        constraintLimits(settings)$melting_temp_diff <- set.new.limits(optiLimits(settings), initial.opti.limits, 
                                                                    initial.opti.constraints, "melting_temp_diff")$melting_temp_diff
                        Tm.brackets <- create.Tm.brackets(new.filtered.df, template.df, settings, target.temps)
                        new.filtered.df <- Tm.brackets$primers
                        max.cvg <- get_max_set_coverage(new.filtered.df, template.df, Tm.brackets, settings, mode.directionality)
                    }
            }
            break
        } 
        # select the first constraint in gain.df: most promising exclusion reason for relaxing
        con.name <- gain.df$Constraint[1] # gain.df is sorted in the order we want to relax constraints in
        message("Relaxing constraint: ", con.name)
        con.idx <- which(active.constraints == con.name)
        # determine candidate primers 
        # only re-check constraints for those primers that give us a gain in coverage
        cvd.idx <- covered.seqs.to.idx(new.excluded.df$Covered_Seqs, template.df)
        cvg.idx <- which(sapply(cvd.idx, function(x) length(intersect(missing.df$Identifier, template.df$Identifier[x])) != 0))
        new.excluded.df <- new.excluded.df[cvg.idx, ]
        exclusion.idx <- which(new.excluded.df$Exclusion_Reason == con.name)
        #message("Candidate idx: ", paste(idx, collapse = ","))
        # augment primer coverage for all primers that failed the current constraint
        cur.candidates <- new.excluded.df[exclusion.idx,]
        # determine constraints that still need to be computed and checked
        to.compute.constraints <- active.constraints[tail(seq_along(filters(settings)), 
            length(filters(settings)) - con.idx)]  # constraints that need to be computed (all later constraints than con.idx)
        active.constraints.check <- active.constraints[tail(seq_along(filters(settings)), 
            length(filters(settings)) - con.idx + 1)]  # constraints to be checked for fulfillment (all later constraints including the current constraint)
        # never compute cvg!
        to.compute.constraints <- to.compute.constraints[!to.compute.constraints %in% c("primer_coverage", "primer_specificity")]
        message("Computing constraints: ", paste(to.compute.constraints, collapse = ","))
        # set new constraint value to limits
        new.setting <- constraintLimits(settings)[[con.name]]
        # relax the constraint
        message("\to ", con.name, ": Relaxed constraint from [", 
            paste0(round(filters(settings)[[con.name]], 2), 
                    collapse = ", "), 
             "] to [", 
              paste0(round(new.setting, 2), collapse = ","), 
              "].")
        relax.df[relax.df$Constraint == con.name, "Relax_Count"] <- relax.df[relax.df$Constraint == con.name, "Relax_Count"] + 1
        constraints(settings)[[con.name]] <- new.setting  # update constraints
        # evaluate constraints again
        # set no.structures to TRUE since we are designing
        filter.result <- suppressMessages(cascaded.filter.quick(cur.candidates, template.df, settings, 
            to.compute.constraints = to.compute.constraints, 
            mode.directionality = mode.directionality,
            active.constraints = active.constraints.check, 
            no.structures = TRUE, updateProgress = NULL))
        # update the filtered primers
        new.primers <- filter.result$data
        new.filtered.df <- my_rbind(new.filtered.df, new.primers)
        # update excluded primers:
        if (length(exclusion.idx) != 0) {
            # remove old entries
            new.excluded.df <- new.excluded.df[-exclusion.idx, ]
            new.excluded.df <- my_rbind(new.excluded.df, filter.result$excluded)
        }
        # store actual coverage gain
        if (con.name %in% new.filtered.df$Exclusion_Reason) {
            new.covered <- unique(unlist(strsplit(new.filtered.df$Covered_Seqs[new.filtered.df$Exclusion_Reason == con.name], split = ",")))
        } else {
            new.covered <- NULL
        }
        additional.cvg <- length(setdiff(new.covered, prev.cvg))
        # store feasible and actual cvg gains
        total.gain.entry <- total.gain.df[total.gain.df$Exclusion_Reason == con.name, "Actual_Gain"]
        total.gain.entry <- total.gain.entry + additional.cvg
        total.gain.df[total.gain.df$Exclusion_Reason == con.name, "Actual_Gain"] <- total.gain.entry
        # update tm brackets to make sure that new melting temp data is taken into account
        Tm.brackets <- create.Tm.brackets(new.filtered.df, template.df, settings, target.temps)
        new.filtered.df <- Tm.brackets$primers
        # update cvg for while loop
        max.cvg <- get_max_set_coverage(new.filtered.df, template.df, Tm.brackets, settings, mode.directionality)
        message(paste("Current coverage is: ", max.cvg, sep = ""))
        # modify the limit of the relaxed constraint for possibly further relaxations
        constraintLimits(settings)[con.name] <- set.new.limits(filterLimits(settings), initial.limits, initial.constraints, con.name)[con.name]
        # update missing templates
        missing.df <- get.missing.df(new.filtered.df, template.df, Tm.brackets, settings, mode.directionality)
    }
    if ("melting_temp_diff" %in% names(constraints(settings))) {
        # keep only temperature primer sets with cvg >= target cvg
        Tm.cvg.vals <- unlist(get_max_set_coverage(new.filtered.df, template.df, Tm.brackets, settings, mode.directionality, max.only = FALSE))
        set.keep.idx <- which(Tm.cvg.vals >= target.cvg)
        if (length(set.keep.idx) != 0) {
            #print("Sets to keep:")
            #print(set.keep.idx)
            # get melting temperature range of primers to keep
            Tm.df <- Tm.brackets$df[set.keep.idx, ]
            #print("Tm.df:")
            #print(Tm.df)
            # only remove additional primers if any set fulfills the criterion
            min.Tm <- min(Tm.df[,"min_Tm"])
            max.Tm <- max(Tm.df[,"max_Tm"])
            # only retain primers between min.Tm and max.Tm, the others are not part of a set with neccesary cvg
            excluded.idx <- which(new.filtered.df$melting_temp < min.Tm | new.filtered.df$melting_temp > max.Tm)
            if (length(excluded.idx) != 0) {
                Tm.exclusion <- new.filtered.df[excluded.idx,]
                Tm.exclusion$Exclusion_Reason <- "melting_temp_range"
                new.excluded.df <- my_rbind(new.excluded.df, Tm.exclusion)
                new.filtered.df <- new.filtered.df[-excluded.idx,]
                # update current cvg (should be the same as before ...)
                max.cvg <- get_max_set_coverage(new.filtered.df, template.df, Tm.brackets, settings, mode.directionality)
            }
        }
    }
    # store final set:
    store.filtering.sets(new.filtered.df, new.excluded.df, results.loc, tag = as.character(while.count), stat.df = relax.df, settings = settings)
    # store new statistics:
    excl.counts <- ddply(excluded.df, "Exclusion_Reason", summarize, "Count" = length(substitute(Exclusion_Reason)))
    # for additional cvg: only the last failed exclusion reason is considered
    counts <- ddply(new.filtered.df, "Exclusion_Reason", summarize, "Count" = length(substitute(Exclusion_Reason)), "Covered" = paste(unique(unlist(strsplit(substitute(Covered_Seqs), split = ","))), collapse = ","))
    if (nrow(counts) != 0) {
        # remove count entries that didn't fail a constraint
        na.idx <- which(is.na(counts$Exclusion_Reason ))
        if (length(na.idx) != 0) {
            counts <- counts[-na.idx,]
        }
    }
    new.stat.df <- data.frame(Constraint = active.constraints,
            Direction = mode.directionality, Remaining = 0,
            Time = 0, Excluded = 0, # nb time not stored here
            Current_Coverage = NA, 
            Excluded_Percentage = NA,
            stringsAsFactors = FALSE)
    m.c <- match(counts$Exclusion_Reason, new.stat.df$Constraint)
    new.stat.df$Remaining[m.c] <- counts$Count # Additional inclusions per category
    m.e <- match(excl.counts$Exclusion_Reason, new.stat.df$Constraint)
    new.stat.df$Excluded[m.e] <- excl.counts$Count # Previous exclusion counts
    new.stat.df$Excluded_Percentage <- new.stat.df$Excluded / (new.stat.df$Remaining + new.stat.df$Excluded) 
    # store current coverage: gain of cvg by every constraint
    new.stat.df$Current_Coverage <- total.gain.df$Actual_Gain / nrow(template.df)
    # return result:
    cvg.s <- paste0(round(max.cvg *100, 2), "%")
    message("-> Available number of candidates after relaxation: ", nrow(new.filtered.df), " (", cvg.s, " coverage)")
    out <- list(new_settings = settings, filtered = new.filtered.df, 
        excluded = new.excluded.df, stats = new.stat.df)
    return(out)
}

#' Update Constraint Settings.
#'
#' Updates the constraint settings with new values. Sets to the maximal
#' observed values within the limits or less if \code{relax.speed} is less than 1.
#'
#' @param values Observed constraint values considered for the update.
#' @param relax.speed The speed at which the constraints should be relaxed.
#' This value should be in the interval [0,1].
#' @param cur.limits List with current relaxation limits.
#' @param cur.setting List with current constraint settings.
#' @return Relaxed constraint settings according to the given \code{values} and \code{cur.limits}.
#' @keywords internal
set.new.constraint.value <- function(values, relax.speed, cur.limits, cur.setting) {
     #values <- get.constraint.values(con.name, cur.candidates, mode.directionality)[[mode.directionality]]

    # remove NA's from values
    values <- values[!is.na(values)]
    if (length(values) == 0) {
        return(NULL)        
    }
    new.setting <- cur.setting
    if (length(cur.limits) == 2) {
        sel.idx <- which(values >= cur.limits["min"] & values <= cur.limits["max"])  # only consider values that can be relaxed
        val <- values[sel.idx]
        if (length(val) != 0) {
            values.low <- val[val <= cur.setting["min"]]
            values.high <- val[val >= cur.setting["max"]]
            if (length(values.low) != 0) {
                new.setting["min"] <- unname(quantile(values.low, 1 - relax.speed))
            }
            if (length(values.high) != 0) {
                new.setting["max"] <- unname(quantile(values.high, relax.speed))
            }
            if (length(values.low) == 0 && length(values.high) == 0) {
                # it's crucial that even though there's nothing relaxable that the function returns something in this case for the relaxation procedure to work.
                return(cur.limits)
            }
            # don't set to limits when nothing changes but rather set to min/max of remaining
            # values to keep the constraints tight
            if (new.setting["min"] == cur.setting["min"] && new.setting["max"] == cur.setting["max"]) {
                if (length(values.low) != 0) {
                  # only change when there are still values
                  new.setting["min"] <- min(values.low, na.rm = TRUE)
                }
                if (length(values.high) != 0) {
                  new.setting["max"] <- max(values.high, na.rm = TRUE)
                }
            }
        } else {
            # no values within the current limits -> set constraints to the limits
            return(cur.limits)
        }
    } else if (length(cur.limits) == 1) {
        if (names(cur.limits) == "max") {
            sel.idx <- which(values <= cur.limits)
            val <- values[sel.idx]
            if (length(val) != 0) {
                new.setting["max"] <- unname(quantile(val, relax.speed))
            } else {
                # no values within the current limits -> set constraint to the limit
                return(cur.limits)
            }
            if (new.setting["max"] == cur.setting["max"]) {
                new.setting["max"] <- max(val, na.rm = TRUE)
            }
        } else {
            # min constraint
            sel.idx <- which(values >= cur.limits)
            val <- values[sel.idx]
            if (length(val) != 0) {
                new.setting["min"] <- unname(quantile(val, 1 - relax.speed))
            } else {
                return(cur.limits)  # nothing to relax 
            }
            if (new.setting["min"] == cur.setting["min"]) {
                new.setting["min"] <- min(val, na.rm = TRUE)
            }
        }
    }

    return(new.setting)
}
#' Filtering of Primers
#'
#' Filters a primer set during the optimization procedure.
#' 
#' @param primer.df Primer data frame.
#' @param sample Name of the current template sample.
#' @param template.df Template data frame.
#' @param settings List with settings for the constraints to be used for filtering.
#' @param mode.directionality Primer direction.
#' @param required.cvg Required ratio of covered templates.
#' If \code{required.cvg} is set to 0, the constraints are not relaxed.
#' @param target.temps Target melting temperature of the primers in Celsius.
#' This argument is only required if we try to match the melting temperatures of another primer set,
#' e.g. when first optimizing forward and then optimizing reverse primers.
#' @param results.loc Path to a directory where the results should be written.
#' @return The filtered primer data frame with respect to \code{required.cvg}.
#' @keywords internal
filter.primer.set.opti <- function(primer.df, sample, template.df, settings, 
                                   mode.directionality, required.cvg, results.loc, 
                                   target.temps) {
    
    if (length(primer.df) == 0) {
        return(NULL)
    }
    filtered.result <- cascaded.filter(primer.df, template.df, settings, mode.directionality, 
        required.cvg, target.temps, results.loc = results.loc)
    if (length(filtered.result) == 0) {
        # no feasible solution
        return(NULL)
    }
    if (length(results.loc) != 0) {
        # store data
        filtered.df <- filtered.result$data
        filtered.stats <- filtered.result$stats
        stats.relax <- filtered.result$stats_relax
        template.df <- update_template_cvg(template.df, filtered.df, mode.directionality)
        write.csv(template.df, file.path(results.loc, paste(sample, "_filtered_templates.csv", 
            sep = "")), row.names = FALSE)
        write.csv(filtered.df, file.path(results.loc, paste(sample, "_filtered_primers.csv", 
            sep = "")), row.names = FALSE)
        write.csv(filtered.stats, file.path(results.loc, paste(sample, "_filtered_primers_stats.csv", 
            sep = "")), row.names = FALSE)
        write.csv(stats.relax, file.path(results.loc, paste(sample, "_filtered_primers_stats_relax.csv", 
            sep = "")), row.names = FALSE)
        excluded.df <- filtered.result$excluded
        write.csv(excluded.df, file.path(results.loc, paste(sample, "_excluded_primers.csv", 
            sep = "")), row.names = FALSE)
    }
    return(filtered.result)
}

########
# Secondary structure of templates
########

#' Template Secondary Structures
#' 
#' Computes template secondary structures.
#'
#' @param template.df Template data frame.
#' @param annealing.temperature Temperature [C] at which 
#' to compute secondary structures.
#' @param regions List containing the positional intervals for which the template
#' secondary structure should be computed.
#' @param constraints String giving secondary structure constraints.
#' For example \code{xxxxxx...} would forbid folding in the first 6 bases
#' of a template with length 9 and allow folding in its last 3 bases.
#' @return Data frame with info on template secondary structures.
#' @keywords internal
compute.template.secondary.structures <- function(template.df, annealing.temperature, 
    regions = NULL, constraints = NULL) {
    seqs <- template.df$Sequence
    if (length(regions) != 0) {
        # compute local secondary structures
        if (length(regions) != length(seqs)) {
            stop("Length of regions does not agree with length of seqs for template secondary structure analysis.")
        }
        seqs <- sapply(seq_along(regions), function(x) substring(seqs[x], regions[[x]][1], 
            regions[[x]][2]))  # extract region substrings
    }
    struct <- compute.structure.vienna(seqs, annealing.temperature, constraints)
    return(struct)
}

#' Optimization of Template Binding
#'
#' Optimizes template binding regions according to secondary structures.
#'
#' @param template.df Template data frame.
#' @param annealing.temperature Temperature at which to compute secondary structures.
#' @param primer.lengths Target length of primers that are to be used.
#' @param mode.directionality Direction of primers.
#' @return List with new binding intervals for every template.
#' @keywords internal
optimize.template.binding.regions.single <- function(template.df, annealing.temperature, 
    primer.lengths, mode.directionality = c("fw", "rev")) {
    # determine local template regions for secondary structure analysis for fw/rev
    # direction
    
    # annealing.temperature: annealing temp for which to compute deltaG
    # primer.length: interval of minimal and maximal primer length
    # mode.directionality: adjust fw or rev binding region?
    if (length(mode.directionality) == 0) {
        stop("Please supply the 'mode.directionality' argument.")
    }
    mode.directionality <- match.arg(mode.directionality)
    if (length(annealing.temperature) == 0) {
        # no annealing temp available yet choice of temperature shouldn't matter too much
        # here, since we compute Delta Delta G only
        annealing.temperature <- 50  # set to typical annealing temperature
    }
    region.slack <- 40  # how many additional positions (outside the target region) to consider after the target region has ended. ensures 'correct folding'.
    if (mode.directionality == "fw") {
        sub.s.base <- template.df$Allowed_Start_fw
        sub.e.base <- template.df$Allowed_End_fw
        sub.e <- template.df$Allowed_End_fw + region.slack
        sub.e <- sapply(seq_along(sub.e), function(x) ifelse(sub.e[x] > nchar(template.df$Sequence)[x], 
            nchar(template.df$Sequence)[x], sub.e[x]))  # make sure that we don't overextend
        sub.s <- sub.s.base
    } else {
        sub.s.base <- template.df$Allowed_Start_rev
        sub.e.base <- template.df$Allowed_End_rev
        sub.s <- template.df$Allowed_Start_rev - region.slack
        sub.s <- sapply(seq_along(sub.s), function(x) ifelse(sub.s[x] < 1, 1, sub.s[x]))  # make sure that we don't underextend
        sub.e <- sub.e.base
    }
    if (all(sub.s == 0) || all(sub.e == 0)) {
        # nothing to optimize
        return(NULL)
    }
    regions <- lapply(seq_along(sub.s), function(x) c(sub.s[x], sub.e[x]))  # regions for secondary structure computation (extended region)
    regions.base <- lapply(seq_along(sub.s.base), function(x) c(sub.s.base[x], sub.e.base[x]))  # regions for constraining (only the target region)
    best.structs <- compute.template.secondary.structures(template.df, annealing.temperature, 
        regions, NULL)  # deltaG of unconstrained structures
    best.structs$ID <- template.df$ID
    # determine constrained structures in windows 1. create windows determine windows
    # for constrained folding
    pol.size <- 0  # is it necessary to add some space for polymerase binding without secondary structure in the way?
    win.size <- min(primer.lengths) + pol.size  # window should at least span the minimal primer size
    # very important: assure that all windows have the same size, otherwise there's a
    # biasA
    # a) individual windows -> problem for plotting
    #windows <- lapply(regions.base, function(x) Hmisc::cut2(x[1]:x[2], m = win.size, 
        #onlycuts = TRUE))
    # b) uniform windows
    windows <- replicate(length(sub.s.base), seq(min(sub.s.base), max(sub.e.base), win.size), simplify = FALSE)
    # 2. create constraint strings for selected region
    constraints <- vector("list", length(windows))
    for (i in seq_along(windows)) {
        win <- windows[[i]]
        win.pos <- win - min(win) + 1
        base.condition <- rep(".", max(regions[[i]]) - min(regions[[i]]) + 1)  # dot indicates no constraints on binding, base.condition is for the extended target region
        # message(length(base.condition))
        cons <- rep(NA, length(win) - 1)  # constraints for each window into current seq
        for (j in seq_along(win[-length(win)])) {
            # don't create windows after the target region
            condition <- base.condition
            s <- win.pos[j]
            max.pos <- sub.e.base[i] - sub.s.base[i] + 1
            if (s > length(condition) || s > max.pos) {
                next
            }
            if (win[j+1] > sub.e.base[i]) {
                # always extend the last window until the end
                #e <- win.pos[j+1]
                e <- sub.e.base[i] - sub.s.base[i] + 1
            } else {
                e <- win.pos[j+1] - 1
            }
            # adjust end to the end of the target region if it exceeds it 
            if (e > max.pos) {
                e <- max.pos
            }
            condition[s:e] <- "x"  # x indicates constraints on binding
            cons[j] <- paste(condition, collapse = "")
        }
        if (length(cons) == 0) {
            # window is a single position
            base.condition[win[1]] <- "x"
            cons <- paste(base.condition, collapse = "")
        }
        constraints[[i]] <- cons
    }
    # 3. compute constrained structures deltaG
    constrained.foldings <- NULL
    max.nbr.intervals <- max(sapply(windows, function(x) length(x) - 1))
    if (max.nbr.intervals == 0) {
        return(NULL)
    }
    for (i in seq_len(max.nbr.intervals)) {
        cur.constraints <- sapply(constraints, function(x) x[i])  # constraints for all seqs
        sel <- which(!is.na(cur.constraints))
        if (length(sel) == 0) {
            next
        }
        con.structs <- compute.template.secondary.structures(template.df[sel, ], annealing.temperature, 
            regions[sel], cur.constraints[sel])  # deltaG of constrained structures
        con.structs$Constraint <- cur.constraints[sel]
        con.structs$Constraint_Idx <- i  # the i-th window of constraints on the secondary structure
        con.structs$Constraint_Region_Start <- sapply(windows[sel], function(x) x[i])
        # determine the window ends
        if (i == max.nbr.intervals) {
            con.structs$Constraint_Region_End <- sapply(windows[sel], function(x) x[i + 1])
        } else {
            con.structs$Constraint_Region_End <- sapply(windows[sel], function(x) x[i + 1] - 1)
        }
        # compute effective region length (the constrained region)
        con.structs$Length <- unlist(lapply(strsplit(cur.constraints[sel], split = ""), function(x) length(which(x == "x"))))
        con.structs$Constraint_Region_Actual_End <- sapply(seq_along(sel), function(x) windows[sel][[x]][i] + con.structs$Length[x] - 1)
        con.structs$ID <- template.df[sel, "ID"]
        constrained.foldings <- rbind(constrained.foldings, con.structs)
    }
    colnames(constrained.foldings)[colnames(constrained.foldings) == "deltaG"] <- "DeltaG_constrained"
    colnames(constrained.foldings)[colnames(constrained.foldings) == "Structure"] <- "Structure_constrained"
    constrained.foldings$Structure_best <- best.structs$Structure[match(constrained.foldings$ID, 
        best.structs$ID)]
    
    constrained.foldings$DeltaG_best <- best.structs$deltaG[match(constrained.foldings$ID, 
        best.structs$ID)]
    # Delta Delta G: the difference between the constrained and unconstrained folding: the smaller Delta Delta G, the better (in this case there is little benefit of folding with secondary structure in the current region)
    constrained.foldings$Delta_Delta_G <- constrained.foldings$DeltaG_constrained - 
        constrained.foldings$DeltaG_best
    # annotate with geneGroups
    constrained.foldings$Group <- template.df$Group[match(constrained.foldings$ID, 
        template.df$ID)]
    # plot.Delta.DeltaG(constrained.foldings, stratify = TRUE)
    # my_ggsave(file.path(results.loc,
    # 'template_secondary_structure_delta_delta_G.pdf')) 4. select windows: a)
    # determine standard deviation of Delta Delta G for each template
    deviations <- sapply(seq_len(nrow(template.df)), function(x) 
                    sd(constrained.foldings$Delta_Delta_G[constrained.foldings$ID == template.df[x, "ID"]]))
    myfun <- function(len, deltaG, win.size) {
        return(deltaG[which(len >= win.size)])
    }
    # don't select windows that are too short for defining the minDeltaG
    minG.per.template <- ddply(constrained.foldings, c("ID", "Group"), here(summarize),
        MinG = myfun(substitute(Length), substitute(Delta_Delta_G), win.size))
    m <- match(template.df$ID, minG.per.template$ID)
    minG.per.template <- minG.per.template[m,]
    # determine the maximal DeltaDeltaG of the optimized regions:
    maxG.per.template <- minG.per.template$MinG + deviations
    # b) retrieve windows in each template with DeltaDeltaG <= maxG.per.template
    sel.windows <- lapply(seq_along(template.df$ID), function(x) {
        temp.data <- constrained.foldings[constrained.foldings$ID == template.df$ID[x], ]
        allowed.deltaG <- maxG.per.template[match(template.df$ID[x], minG.per.template$ID)]
        sel <- temp.data$Delta_Delta_G <= allowed.deltaG
        con <- temp.data$Constraint_Idx[sel]
        # c) select longest consecutive windows for implementation reasons
        if (length(con) > 1) {
            # multiple best windows found
            result <- rle(diff(con))
            idx <- which(result$values == 1)
            if (length(idx) == 0) {
                # no consecutive runs -> select window with smallest DeltaDeltaG that has allowed effective length
                data <- temp.data[sel, ]
                len.sel <- which(data$Length >= win.size)
                con <- con[len.sel[which.min(data$Delta_Delta_G[len.sel])]]
            } else {
                longest.runs <- idx[which.max(result$lengths[idx])]
                if (longest.runs == 1) {
                  sel.start <- 1
                } else {
                  sel.start <- sum(result$length[seq_len(longest.runs - 1)]) + 1
                }
                sel.end <- sum(result$length[seq_len(longest.runs)]) + 1
                con <- con[sel.start:sel.end]
            }
        } else {
            # only one best window found
            con
        }
    })
    new.regions <- lapply(seq_along(template.df$ID), function(x) {
        sel <- which(constrained.foldings$ID == template.df$ID[x])  # retrieve start/end of regions
        sel <- sel[sel.windows[[x]]]
        s <- min(constrained.foldings$Constraint_Region_Start[sel])  # retrieve selected consecutive regions
        e <- max((constrained.foldings$Constraint_Region_Actual_End[sel])) # use the actual end of the region here
        return(c(s, e))
    }) 
    constrained.foldings$Direction <- mode.directionality
    # don't output incomplete regions
    constrained.foldings <- constrained.foldings[constrained.foldings$Length >= win.size,]
    result <- list("Intervals" = new.regions, "Foldings" = constrained.foldings)
    return(result)
}
#' Optimization of Binding Regions
#' 
#' Optimizes the template binding regions.
#'
#' @param template.df Template data frame.
#' @param annealing.temperature Temperature at which to compute secondary structures.
#' @param primer.lengths Target length of primers that are to be used.
#' @param mode.directionality Direction of primers.
#' @return List with intervals indicating improved primer binding regions.
#' @keywords internal
optimize.template.binding.regions.dir <- function(template.df, annealing.temperature = NULL, 
    primer.lengths, mode.directionality = c("fw", "rev", "both")) {
    # compute uniform result for all possible direction inputs
    if (length(mode.directionality) == 0) {
        stop("Please supply the 'mode.directionality' argument.")
    }
    mode.directionality <- match.arg(mode.directionality)
    if (mode.directionality == "both") {
        binding.regions.fw <- optimize.template.binding.regions.single(template.df, annealing.temperature, 
            primer.lengths, "fw")
        binding.regions.rev <- optimize.template.binding.regions.single(template.df, annealing.temperature, 
            primer.lengths, "rev")
        intervals <- list(fw = binding.regions.fw$Intervals, rev = binding.regions.rev$Intervals)
        foldings <- rbind.fill(binding.regions.fw$Foldings, binding.regions.rev$Foldings)
    } else {
        binding.regions <- optimize.template.binding.regions.single(template.df, annealing.temperature, 
            primer.lengths, mode.directionality)
        intervals <- list(binding.regions$Intervals)
        names(intervals) <- mode.directionality
        foldings <- binding.regions$Foldings
    }
    result <- list("Intervals" = intervals, "Foldings" = foldings)
    return(result)
}
#' Delta DeltaG Plot
#'
#' Plots the difference between the free energy of constrained 
#' and unconstrained foldings.
#'
#' @param constrained.foldings Data frame with info from constrained foldings.
#' @param stratify Stratify according to template groups?
#' @return Plot of Delta DeltaG.
#' @keywords internal
plot.Delta.DeltaG <- function(constrained.foldings, stratify = FALSE) {
    # plot line plot of DeltaDeltaG (constrained - MFE structure) for all templates
    # at once, stratify by class?
    plot.df <- constrained.foldings
    groups <- max(constrained.foldings$Constraint_Idx)
    plot.df$Binding_Region <- Hmisc::cut2(c(1, constrained.foldings$Constraint_Region_End), 
        g = groups)[-1]
    
    if (!stratify) {
        p <- ggplot(plot.df, aes(x = .data[["Binding_Region"]], y = .data[["Delta_Delta_G"]])) + geom_bar(stat = "summary")
    } else {
        p <- ggplot(plot.df, aes(x = .data[["Binding_Region"]], y = .data[["Delta_Delta_G"]])) + geom_boxplot() + 
            facet_wrap(~Group)
    }
    x.names <- unique(paste(plot.df$Constraint_Region_Start, "-", plot.df$Constraint_Region_End, 
        sep = ""))
    title <- "Tendency for secondary structure formation in template target regions"
    # with geom_bar
    p <- p + xlab("Region constrained to be free of base pairing") + ylab("Delta Delta G") + 
        ggtitle(title)
    return(p)
}

######
# Constraint: annealing temperature
########
#' Annealing temperature.
#'
#' Identifies the optimal annealing temperature of a set of primers.
#' If primers cover template sequences, the annealing temperature
#' is computed using Rychlik's formula.
#' Otherwise, the annealing temperature is determined using the
#' rule of thumb based on the melting temperatures of the primers.
#'
#' @param primer.df Primer data frame.
#' @param mode.directionality Primer directionality.
#' @param template.df Template data frame
#' @param na_salt_conc Sodium ion concentration.
#' @param mg_salt_conc Magensium ion concentration.
#' @param k_salt_conc Potassium ion concentration.
#' @param tris_salt_conc Tris buffer concentration.
#' @param primer_conc Primer concentration.
#' @return The optimal annealing temperature.
#' @keywords internal
compute_annealing_temp <- function(primer.df, mode.directionality, 
                      template.df, na_salt_conc, mg_salt_conc, k_salt_conc, 
                      tris_salt_conc, primer_conc) {
    
    if (length(primer.df) == 0) {
        # can't suggest an annealing temperature
        return(NULL)
    }
    if (length(primer.df) != 0) {
        compute.idx <- seq_len(nrow(primer.df))
    }
    if ("melting_temp" %in% colnames(primer.df)) {
        compute.idx <- which(is.na(primer.df$melting_temp))
    }
    if (length(compute.idx) != 0) {
        # compute melting temps
         melting.temps <- compute.melting.temps(primer.df[compute.idx,], primer_conc, 
            na_salt_conc, mg_salt_conc, k_salt_conc, tris_salt_conc, 
            mode.directionality)
        p.df <- update.constraint.values(primer.df[compute.idx, ], melting.temps)
        if ("melting_temp" %in% colnames(primer.df)) {
            # overwrite existing MELT entries with new rows
            primer.df[compute.idx, ] <- p.df
        } else {
            # add the columns to the data frame
            primer.df <- cbind(primer.df, p.df)
        }
    } 
    Ta <- annealing.temp.rule.of.thumb(primer.df$melting_temp)
    return(Ta)
}

#' Rule of thumb for annealing temperature
#'
#' Computes the annealing temperature using a rule of thumb
#' 
#' @param melting.temp Melting temperatures of primers
#' 
#' @return The annealing temperature corresponding to the input melting temperature.
#' @keywords internal
annealing.temp.rule.of.thumb <- function(melting.temp) {
    if (length(melting.temp) == 0 || all(is.na(melting.temp ))) {
        return(NA)
    }
    Ta <- melting.temp - 5
    return(Ta)
}

#' Annealing temperature
#'
#' Computes the annealing temperature using all binding events.
#'
#' @param primer.df Primer data frame.
#' @param mode.directionality Primer directionality.
#' @param template.df Template data frame
#' @param na_salt_conc Sodium ion concentration.
#' @param mg_salt_conc Magensium ion concentration.
#' @param k_salt_conc Potassium ion concentration.
#' @param tris_salt_conc Tris buffer concentration.
#' @param primer_conc Primer concentration.
#'
#' @return The recommended annealing temperature.
#' @keywords internal
compute.covered.Ta <- function(primer.df, mode.directionality = c("fw", "rev", "both"), template.df, na_salt_conc, 
                               mg_salt_conc, k_salt_conc, tris_salt_conc, primer_conc) {
    if (length(mode.directionality) == 0) {
        stop("'mode.directionality' not supplied.")
    }
    warning("covered.Ta is deprecated: Ta is too low!")
    mode.directionality <- match.arg(mode.directionality)
    if (!"Covered_Seqs" %in% colnames(primer.df) || length(primer.df) == 0) {
        stop("Cannot compute annealing temperature: primer coverage is not available yet!")
    }
    seq.idx <- covered.seqs.to.idx(primer.df$Covered_Seqs, template.df)
    Ta <- na.omit(unlist(lapply(seq_len(nrow(primer.df)), function(x) {
        compute.Ta(primer.df[x, ], template.df[seq.idx[[x]], ], 
                   mode.directionality, na_salt_conc, mg_salt_conc, 
                   k_salt_conc, tris_salt_conc, primer_conc)
        })))
    return(Ta)
}
#' Annealing temperature
#'
#' Computes the annealing temperature using all binding events.
#'
#' @param primer.df Primer data frame.
#' @param template.df Template data frame
#' @param mode.directionality Primer directionality.
#' @param na_salt_conc Sodium ion concentration.
#' @param mg_salt_conc Magensium ion concentration.
#' @param k_salt_conc Potassium ion concentration.
#' @param tris_salt_conc Tris buffer concentration.
#' @param primer_conc Primer concentration.
#'
#' @return All annealing temperatures for given binding events.
#' @references 
#' Rychlik, W. J. S. W., W. J. Spencer, and R. E. Rhoads. "Optimization of the annealing temperature for DNA amplification in vitro." Nucleic acids research 18.21 (1990): 6409-6412.
#' @keywords internal
compute.Ta <- function(primer.df, template.df, mode.directionality = c("fw", "rev", "both"), 
                       na_salt_conc, mg_salt_conc, 
                       k_salt_conc, tris_salt_conc, primer_conc) {
    if (length(mode.directionality) == 0) {
        stop("'mode.directionality' not supplied.")
    }
    mode.directionality <- match.arg(mode.directionality)
    if (length(template.df) == 0 || nrow(template.df) == 0) {
        return(NA)
    }
    if (!"melting_temp" %in% colnames(primer.df)) {
        Tm.p <- compute.melting.temps(primer.df, primer_conc, na_salt_conc, mg_salt_conc, 
                k_salt_conc, tris_salt_conc, mode.directionality)$melting_temp  # primer Tm
    } else {
        Tm.p <- primer.df$melting_temp
    }
    if (!"melting_temp" %in% colnames(template.df)) {
        Tm.s <- compute.Tm.baldino(template.df$Sequence, na_salt_conc, mg_salt_conc, k_salt_conc, 
            tris_salt_conc, primer_conc)  # template Tm
    } else {
        Tm.s <- template.df$melting_temp
    }
    result <- 0.3 * Tm.p + 0.7 * Tm.s - 14.9
    na.idx <- which(is.nan(result))
    if (length(na.idx) != 0) {
        result[na.idx] <- NA
    }
    return(result)
}

## ----vignette_options, echo = FALSE, message = FALSE, warning = FALSE---------
options(rmarkdown.html_vignette.check_title = FALSE)
knitr::opts_chunk$set(collapse = TRUE, comment = "#>")
library(openPrimeR)
ggplot2::theme_set(ggplot2::theme_grey(base_size = 12)) 

## ----check_dependencies, message = FALSE, warning = FALSE, eval = FALSE-------
# library(openPrimeR)

## ----loading_data_table, echo=FALSE, message=FALSE, warnings=FALSE, results='asis'----
tabl <- "
| Task             | Templates | Primers      | Input file format  |
|------------------|-----------|--------------|--------------------|
| Design primers   | &#10003;  |              | FASTA, CSV         |
| Analyze primers  | &#10003;  | &#10003;     | FASTA, CSV         |
| Compare primers  | &#10003;  | &#10003;     | FASTA, CSV         |
"
cat(tabl) # output the table in a format good for HTML/PDF/docx conversion

## ----simple_load_templates, message = FALSE, warning = FALSE------------------
# Specify a FASTA file containing the templates:
fasta.file <- system.file("extdata", "IMGT_data", "templates", 
                "Homo_sapiens_IGH_functional_exon.fasta", package = "openPrimeR")
# Load the template sequences from 'fasta.file'
seq.df.simple <- read_templates(fasta.file)

## ----header_structure, message = FALSE, warning = FALSE-----------------------
seq.df.simple$Header[1]

## ----load_templates, message = FALSE, warning = FALSE-------------------------
hdr.structure <- c("ACCESSION", "GROUP", "SPECIES", "FUNCTION")
seq.df <- read_templates(fasta.file, hdr.structure, delim = "|", id.column = "GROUP")

## ----loaded_header_structure, message = FALSE, warning = FALSE----------------
# Show the loaded metadata for the first template
c(seq.df$Accession[1], seq.df$Group[1], seq.df$Species[1], seq.df$Function[1])
# Show the ID (group) of the first template
seq.df$ID[1]

## ----default_binding_regions, message = FALSE, warning = FALSE----------------
# Show the binding region of the first template for forward primers
seq.df$Allowed_fw[1]
# Show the corresponding interval in the templates
c(seq.df$Allowed_Start_fw[1], seq.df$Allowed_End_fw[1])
# Show the binding region of the first template for reverse primers
seq.df$Allowed_rev[1]
# Show the corresponding interval in the templates
c(seq.df$Allowed_Start_rev[1], seq.df$Allowed_End_rev[1])

## ----assign_uniform_binding, message = FALSE, warning = FALSE-----------------
template.df.uni <- assign_binding_regions(seq.df, fw = c(1,50), rev = c(1,40))

## ----uniform_binding_regions, message = FALSE, warning = FALSE----------------
# Show the new forward binding region (first 50 bases)
template.df.uni$Allowed_fw[1]
# Show the new reverse binding region (last 40 bases)
template.df.uni$Allowed_rev[1]

## ----assign_individual_binding, message = FALSE, warning = FALSE--------------
l.fasta.file <- system.file("extdata", "IMGT_data", "templates", 
                "Homo_sapiens_IGH_functional_leader.fasta", package = "openPrimeR")
template.df <- assign_binding_regions(seq.df, fw = l.fasta.file, rev = NULL)

## ----assign_individual_binding_out, message = FALSE, warning = FALSE----------
# An example of two templates with different binding regions
c(template.df$Allowed_Start_fw[1], template.df$Allowed_End_fw[1])
c(template.df$Allowed_Start_fw[150], template.df$Allowed_End_fw[150])

## ----assign_individual_binding_example_rev, message = FALSE, warning = FALSE----
# Verify that the binding region for reverse primers did not change for the first template:
template.df$Allowed_rev[1]

## ----available_settings, message = FALSE, warning = FALSE---------------------
list.files(system.file("extdata", "settings", package = "openPrimeR"), pattern = "*\\.xml")

## ----load_settings, message = FALSE, warning = FALSE--------------------------
settings.xml <- system.file("extdata", "settings", 
                    "C_Taq_PCR_high_stringency.xml", package = "openPrimeR")
settings <- read_settings(settings.xml)

## ----settings_slots, echo=FALSE, message=FALSE, warnings=FALSE, results='asis'----
tabl <- "
| Slot             | Getter/Setter |  Purpose |
|------------------|---------------|----------|
| `Input_Constraints` | `constraints()` | Desired values for primer properties |
| `Input_Constraint_Boundaries` | `constraintLimits()`  | Limits for relaxing constraints during primer design |
| `Coverage_Constraints` | `cvg_constraints()` | Constraints for estimating the coverage |
| `PCR_conditions` | `PCR()`  | Experimental PCR conditions |
| `constraint_settings` | `conOptions()` | Settings for evaluating the constraints |

"
cat(tabl) # output the table in a format good for HTML/PDF/docx conversion

## ----cvg_constraint_setup, message = FALSE, warning = FALSE-------------------
# No constraints on primer coverage (not recommended!)
cvg_constraints(settings) <- list()
# Instead consider all primers binding with up to 3 mismatches to cover the corresponding templates
conOptions(settings)$allowed_mismatches <- 3

## ----change_settings1, message = FALSE, warning = FALSE-----------------------
design.settings <- settings
constraints(design.settings) <- constraints(design.settings)[!grepl(
                            "gc_clamp", names(constraints(design.settings)))]

## ----change_settings2, message = FALSE, warning = FALSE-----------------------
constraints(design.settings)[["primer_length"]] <- c("min" = 25, "max" = 25)

## ----change_settings3, message = FALSE, warning = FALSE-----------------------
conOptions(design.settings)[["allowed_mismatches"]] <- 0

## ----write_settings, message = FALSE, warning = FALSE, eval = FALSE-----------
# out.file <- tempfile("settings", fileext = ".xml")
# write_settings(design.settings, out.file)

## ----design_primers, message = FALSE, warning = FALSE, eval = TRUE------------
# Design forward primers for the first two templates
optimal.primers <- design_primers(template.df[1:2,], mode.directionality = "fw",
                                  settings = design.settings)

## ----store_primers, message = FALSE, warning = FALSE, eval = FALSE------------
# out.file <- tempfile("my_primers", fileext = ".fasta")
# write_primers(optimal.primers$opti, out.file)

## ----read_primers, message = FALSE, warning = FALSE---------------------------
# Define the FASTA primer file to load
primer.location <- system.file("extdata", "IMGT_data", "primers", "IGHV", 
                               "Ippolito2012.fasta", package = "openPrimeR")
# Load the primers
primer.df <- read_primers(primer.location, fw.id = "_fw")

## ----view_primers, message = FALSE, warning = FALSE---------------------------
print(primer.df[, c("ID", "Forward")]) 

## ----check_constraints, message = FALSE, warning = FALSE----------------------
# Allow off-target coverage
conOptions(settings)$allowed_other_binding_ratio <- c("max" = 1.0)
# Evaluate all constraints found in 'settings'
constraint.df <- check_constraints(primer.df, template.df, 
                 settings, active.constraints = names(constraints(settings)))

## ----view_primer_coverage, message = FALSE, warning = FALSE-------------------
constraint.df$primer_coverage

## ----update_template_cvg, message = FALSE, warning = FALSE--------------------
template.df <- update_template_cvg(template.df, constraint.df)

## ----template_coverage, message = FALSE, warning = FALSE----------------------
template.df$primer_coverage[1:5]

## ----cvg_ratio, message = FALSE, warning = FALSE------------------------------
as.numeric(get_cvg_ratio(constraint.df, template.df))

## ----cvg_stats, message = FALSE, warning = FALSE------------------------------
cvg.stats <- get_cvg_stats(constraint.df, template.df, for.viewing = TRUE)

## ----cvg_table, echo=FALSE, results='asis'------------------------------------
knitr::kable(cvg.stats[, !(grepl("_fw", colnames(cvg.stats)) | grepl("_rev", colnames(cvg.stats)))], row.names = FALSE)

## ----template_cvg_plot, fig.show='hold', fig.width=5, fig.height=5------------
plot_template_cvg(constraint.df, template.df)

## ----primer_subsets, message = FALSE, warning = FALSE-------------------------
primer.subsets <- subset_primer_set(constraint.df, template.df)

## ----cvg_subsets, fig.show='hold', fig.width=5, fig.height=5------------------
plot_primer_subsets(primer.subsets, template.df)

## ----optimal_subsets, message = FALSE, warning = FALSE------------------------
my.primer.subset <- primer.subsets[[3]]

## ----verify_optimal_subset, message = FALSE, warning = FALSE------------------
original.cvg <- as.character(get_cvg_ratio(constraint.df, template.df, as.char = TRUE))
subset.cvg <- as.character(get_cvg_ratio(my.primer.subset, template.df, as.char = TRUE))
print(paste0("Coverage (n=", nrow(constraint.df), "): ", original.cvg, "; Subset Coverage (n=", nrow(my.primer.subset), "): ", subset.cvg))

## ----binding_regions, message = FALSE, warning = FALSE, fig.show='hold', fig.width=5, fig.height=5----
plot_primer_binding_regions(constraint.df, template.df)

## ----cvg_primer_view, message = FALSE, warning = FALSE, fig.show='hold', fig.width=8, fig.height=8----
plot_primer(constraint.df[1,], template.df[1:10,])

## ----plot_constraints, fig.show='hold', fig.width=7, fig.height=7, message = FALSE, warning = FALSE----
plot_constraint_fulfillment(constraint.df, settings)

## ----constraint_eval_gc, message = FALSE, warning = FALSE---------------------
# View the number of terminal GCs for primers failing the GC constraint
constraint.df$gc_clamp_fw[!constraint.df$EVAL_gc_clamp]
# View the desired number of terminal GCs:
constraints(settings)$gc_clamp

## ----plot_constraint_qualitative, fig.show='hold', fig.width=5, fig.height=5, message = FALSE, warning = FALSE----
plot_constraint(constraint.df, settings, "gc_clamp")

## ----primer_filtering, message = FALSE, warning = FALSE-----------------------
filtered.df <- filter_primers(constraint.df, template.df, settings,
               active.constraints = c("gc_clamp", "gc_ratio"))

## ----eval_report, message = FALSE, warning = FALSE, eval = FALSE--------------
# # Define the path to the output file
# my.file <- tempfile(fileext = ".pdf")
# # Store a PDF report for 'constraint.df' in 'my.file'
# create_report(constraint.df, template.df, my.file,
#               settings, sample.name = "My analysis")

## ----writing_comparison_data, message = FALSE, warning = FALSE----------------
primer.xml <- tempfile("my_primers", fileext =".csv")
write_primers(constraint.df, primer.xml, "CSV")
template.xml <- tempfile("my_templates", fileext = ".csv")
write_templates(constraint.df, template.xml, "CSV")

## ----loading_comparison_data, message = FALSE, warning = FALSE----------------
# Define the primer sets we want to load
sel.sets <- c("Glas1999", "Rubinstein1998", "Cardona1995", "Persson1991", "Ippolito2012", "Scheid2011")
# List all available IGH primer sets
primer.files <- list.files(path = system.file("extdata", "IMGT_data", "comparison", 
                           "primer_sets", "IGH", package = "openPrimeR"),
                pattern = "*\\.csv", full.names = TRUE)
# Load all available primer sets
primer.data <- read_primers(primer.files)
# Select only the sets defined via 'sel.sets'
sel.idx <- which(names(primer.data) %in% sel.sets)
primer.data <- primer.data[sel.idx]
# Provide a set of templates for every primer set
template.files <- rep(system.file("extdata", "IMGT_data", "comparison", "templates", 
                              "IGH_templates.csv", package = "openPrimeR"), 
                              length(primer.data))
template.data <- read_templates(template.files)

## ----comparison_plots_overview, fig.show='hold', fig.width=7, fig.height=7, message = FALSE----
plot_constraint_fulfillment(primer.data, settings, plot.p.vals = FALSE)

## ----comparison_plots_details, fig.show='hold', fig.width=7, fig.height=7, message = FALSE----
plot_constraint(primer.data, settings, active.constraints = c("gc_ratio", "melting_temp_range"))

## ----comparison_primer_binding, fig.show='hold', fig.width=7, fig.height=7, message = FALSE----
plot_primer_binding_regions(primer.data, template.data)

########
# Custom error handling functions
#########
#' Custom Error
#' 
#' Creates an error with a custom class.
#'
#' @param subclass String giving the specific type of error.
#' @param message Message to be displayed to the user.
#' @param call Environment where the error ocurred.
#' @param ... Other arguments to be passed to  the condition function.
#' @return Generates a custom error.
#' @keywords internal
my.error <- function(subclass, message, call = sys.call(-1), ...) {
    c <- condition(c(subclass, "openPrimeR_Err", "error"), message, call = call, ...)
    stop(c)
}
#' Custom Warning.
#' 
#' Creates a warning with a custom class.
#'
#' @param subclass String giving the specific type of error.
#' @param message Message to be displayed to the user.
#' @param call Environment where the error ocurred.
#' @param ... Other arguments to the condition function.
#' @return Generates a custom warning.
#' @keywords internal
my.warning <- function(subclass, message, call = sys.call(-1), ...) {
  c <- condition(c(subclass, "openPrimeR_Err", "warning"), message, call = call, ...)
  warning(c)
}
#' Condition Constructor
#'
#' Constructs a condition for custom errors.
#'
#' @param subclass String giving the specific error.
#' @param message String giving the user message.
#' @param call Environment object.
#' @param ... Other arguments for the output structure.
#' @return A condition structure.
#' @keywords internal
condition <- function(subclass, message, call = sys.call(-1), ...) {
    # subclass: inherit from warning/error/message
  structure(
    class = c(subclass, "condition"), # always inherit from condition
    list(message = message, call = call),
    ...
  )
}

context("Primer length")

test_that("Primer length", {
    # test primer lengths
    data(Ippolito)
    expect_that(nchar(primer.df$Forward), equals(primer.df$primer_length_fw))
    # no rev primers: length should be 0
    expect_that(nchar(primer.df$Reverse), equals(rep(0, nrow(primer.df))))
}) 

## Comparing primer sets

At its core, comparing primer sets is based on the same functions as for analyzing individual primer sets. Therefore, we will now simply load several pre-evaluated sets of primers and their corresponding templates:

```{r load_comparison_data, exercise = TRUE}
sel.sets <- c("Glas", "Rubinstein", "Persson", "Ippolito", "Scheid")
primer.files <- list.files(path = system.file("extdata", "IMGT_data", "comparison", 
                           "primer_sets", "IGH", package = "openPrimeR"),
                pattern = "*\\.csv", full.names = TRUE)
primer.data <- read_primers(primer.files)
sel.idx <- unlist(lapply(sel.sets, function(x) grep(x, names(primer.data))[1]))
primer.data <- primer.data[sel.idx]
template.files <- rep(system.file("extdata", "IMGT_data", "comparison", "templates", 
                              "IGH_templates.csv", package = "openPrimeR"), 
                              length(primer.data))
template.data <- read_templates(template.files)
```

### Comparing the coverage

To compare the coverage of the templates that is afforded by each primer set, we can use `plot_template_cvg()`:

```{r comparison_setup}
sel.sets <- c("Glas", "Rubinstein", "Persson", "Ippolito", "Scheid")
primer.files <- list.files(path = system.file("extdata", "IMGT_data", "comparison", 
                           "primer_sets", "IGH", package = "openPrimeR"),
                pattern = "*\\.csv", full.names = TRUE)
primer.data <- read_primers(primer.files)
sel.idx <- unlist(lapply(sel.sets, function(x) grep(x, names(primer.data))[1]))
primer.data <- primer.data[sel.idx]
template.files <- rep(system.file("extdata", "IMGT_data", "comparison", "templates", 
                              "IGH_templates.csv", package = "openPrimeR"), 
                              length(primer.data))
template.data <- read_templates(template.files)
```

```{r comparison_cvg, exercise = TRUE, exercise.setup = "comparison_setup"}
# Plot the template coverage
```

```{r comparison_cvg-solution}
plot_template_cvg(primer.data, template.data)
```

The plot shows that the primer sets from *Glas* and *Ippolito* outperform the other sets with respect to coverage. The sets from *Person*, *Rubinstein*, and *Scheid* seem to lack coverage of IGHV2. To gain a better understanding of the qualities of the individual primer sets, let's analyze the distribution of coverage events with regard to mismatches:

```{r comparison_cvg_mm, exercise = TRUE, exercise.setup = "comparison_setup"}
# Plot the template coverage with mismatches
```

```{r comparison_cvg_mm-solution}
plot_template_cvg(primer.data, template.data, per.mismatch = TRUE)
```

From the distribution of binding events with regard to mismatches, we see that the *Ippolito* has a favorable performance for few mismatches since the set guarantees high coverages already without introducing many mutations into the sequences. While the sets from *Scheid* and *Glas* perform similarly well for few mismatches, the sets from *Persson* and *Rubinstein* achieve high coverages only with many mismatches. 

Apart from the coverage, the number of primers may be of concern since unnecessarily large sets may increase costs and be subject to primer-primer interactions. We can analyze the relationship between the percentage of covered templates and the size of the primer sets through `plot_cvg_vs_set_size()`:

```{r comparison_size, exercise = TRUE, exercise.setup = "comparison_setup"}
# Plot coverage vs size of primer set
```

```{r comparison_size-solution}
plot_cvg_vs_set_size(primer.data, template.data)
```

The plot reveals that the primer sets from *Persson*, *Rubinstein*, and *Ippolito* have acceptable sizes and that the sets from *Scheid* and *Glas* may be prohibitively large. The plot also reveals to which degree the sets fulfill the constraints as the radii of the points indicate the degree of constraint fulfillment. Since the radii of the sets are quite similar there doesn't seem to be a big difference between the sets with respect to constraint fulfillment.

### Comparing primer properties

We'll now investigate which constraints are unsatisfied by using `plot_constraint_fulfillment()` and determine the extent to which the constraints are broken using `plot_constraint_deviation()`:

```{r comparison_setup2}
library(openPrimeR)
xml.file <- system.file("extdata", "settings", "B_Taq_PCR_evaluate.xml",
              package = "openPrimeR")
settings <- read_settings(xml.file)
# don't consider specificity: we don't want to interpret this here ..
constraints(settings) <- constraints(settings)[names(constraints(settings)) != "primer_specificity"]
sel.sets <- c("Glas", "Rubinstein", "Persson", "Ippolito", "Scheid")
primer.files <- list.files(path = system.file("extdata", "IMGT_data", "comparison", 
                           "primer_sets", "IGH", package = "openPrimeR"),
                pattern = "*\\.csv", full.names = TRUE)
primer.data <- read_primers(primer.files)
sel.idx <- unlist(lapply(sel.sets, function(x) grep(x, names(primer.data))[1]))
primer.data <- primer.data[sel.idx]
template.files <- rep(system.file("extdata", "IMGT_data", "comparison", "templates", 
                              "IGH_templates.csv", package = "openPrimeR"), 
                              length(primer.data))
template.data <- read_templates(template.files)
```


```{r comparison_deviation, exercise = TRUE, exercise.setup = "comparison_setup2"}
# Plot the constraint fulfillment and deviation 
```

```{r comparison_deviation-solution}
plot_constraint_fulfillment(primer.data, settings)
plot_constraint_deviation(primer.data, settings)
```

Looking at the results, we can make the following findings:

* Although the extent of the deviation from the cross dimerization constraint is small, all sets exhibit some cross dimerization, which may cause problems.
* All primer sets except that from *Persson* have high melting temperature deviations. The melting temperature deviation for the *Ippolito* set may still be acceptable.
* Some primers from *Glas* may form considerable secondary structures.

In summary, the primers from *Persson* fulfill the constraints extremely well. The sets from *Glas*, *Rubinstein*, and *Scheid* may be problematic due to the high melting temperature deviations. 

### Comparing primer binding regions

Finally, let us consider the regions where the primers bind in the templates. Since we've allowed for off-target binding events it's important to verify whether the primers bind close to the target region or not using `plot_primer_binding_regions()`:

```{r comparison_binding_regions, exercise = TRUE, exercise.setup = "comparison_setup2"}
# Plot the binding positions of the primers
```

```{r comparison_binding_regions-solution}
plot_primer_binding_regions(primer.data, template.data)
```

In this case, the blue region indicates the leader of the immunoglobulins, while the red region shows the variable exon region. Hence, the primers should start amplifying the templates at position 0 (the start of the exon) at the latest. However, this is not the case for all primer sets:

* The primers from *Glas* seem to bind along the full range of the sequence, which would result in shortened amplicons
* For the sets from *Rubinstein*, *Persson*, and *Scheid*, at least one primer seems to bind too far downstream to amplify the full exon
* All primers from *Ippolito* bind directly at the start of the exon ensuring complete amplicons
* The primers from *Scheid* bind to the leader, which can be favorable for amplifying highly mutated immunoglobulins

### What is the best primer set?

Considering all criteria, we would probably want to select the set from *Ippolito* for performing PCRs on the heavy chains of immunoglobulins due to the following reasons:

* The set has one of the highest coverage of templates, also for few mismatches
* The set is quite small
* It fulfills the desirable constraints for primers reasonably well

If you'd like to store the results of comparing the primer sets you can create a PDF report using `create_report()`. 

---
output:
  pdf_document:
    fig_caption: yes
    includes:
        in_header: header.tex 
date: "`r format(Sys.time(), '%d %B, %Y')`"
params:
  sample: NA
  primers: NA
  templates: NA
  direction: NA
  settings: NA
  used_settings: NA
  required_cvg: NA

title: "`r paste0('Report for ', params$sample)`"
---

<!-- Vignette options and disclaimer-->
```{r child = 'disclaimer.Rmd'}
```
<!-- no title for first section, would create a new page ... -->

```{r primer_table}
out.df <- openPrimeR:::view.primers.report(params$primers, params$templates)
#"$\\Delta \\text{T}_\\text{m}$")
knitr::kable(out.df, row.names = FALSE, caption = "Overview of the analyzed primers. If forward and reverse primers are present, primers are paired such as to maximize the coverage. Ambiguous bases are shown in italics. The \\textit{Coverage} column indicates the percentage of template sequences that are covered by each primer. \\textit{Mismatches} gives the percentage of coverage events that are subject to a certain number of mismatches. \\textit{Position} indicates the major binding mode of a primer relative to the end of the target binding region. Negative values indicate binding upstream of the specified region, while positive positions indicate binding downstream of the target region.",
            escape = TRUE, longtable = TRUE)
```

# Coverage

```{r template_cvg_stats}
cvg.stats <- openPrimeR::get_cvg_stats(params$primers, params$templates)
```

```{r cvg_table, results='asis'}
out.stats <- openPrimeR::get_cvg_stats(params$primers, params$templates, for.viewing = TRUE)
knitr::kable(out.stats, row.names = FALSE, caption = "Number of covered template sequences per group. In case that primers for both directions are present, the \\textit{Coverage (fw)} and \\textit{Coverage (rev)} columns refer to the individual coverage of forward and reverse primers, respectively. If primers of both directions are present, the total coverage is determined by the intersection of coverage events from primers of both directions. If, however, only primers of a single direction are present, the coverage solely depends on the primers of the given direction.", longtable = TRUE)
```

```{r template_coverage_plot, fig.width = 4.5, fig.height = 4.5, fig.show='hold', fig.cap="The number of covered and available template sequences per group of templates. \\textit{Identity Coverage} refers to the number of template sequences for which fully complementary primers exist. \\textit{Expected Coverage} indicates the expected number of covered templates, which is determined by applying the coverage constraints to the set of potential coverage events for the specified number of maximal mismatches. \\textit{Available Templates} shows the total number of available template sequences per group."}
openPrimeR::plot_template_cvg(params$primers, params$templates)
```

```{r cvg_subsets, fig.show='hold', fig.width = 5, fig.height = 5, fig.cap = 'The coverage of optimal subsets of the input primer set. Here \\textit{optimal} refers to the fact that the subsets were selected such as to maximize the coverage. The line plot indicates the total percentage of covered template sequences, while the bars indicate the percentage of covered templates for individual primers. The cumulative coverage of the bars can exceed 100 percent because different primers may cover the same template redundantly. The target coverage ratio is indicated by the dashed horizontal line.'}
primer.subsets <- openPrimeR::subset_primer_set(params$primers, params$templates)
openPrimeR::plot_primer_subsets(primer.subsets, params$templates, 
                                required.cvg = params$required_cvg)
```

```{r binding_regions_fw, fig.width = 5, fig.height = 5,fig.show='hold', fig.cap = 'The binding positions of the primers relative to the target binding region of forward primers. Here, \\textit{Binding region} indicates the region where the forward primers should bind, while \\textit{Amplification region} indicates the following region.'}
if (params$direction == "fw" || params$direction == "both") {
    openPrimeR::plot_primer_binding_regions(params$primers, params$templates, relation = "fw")
}
```

# Physicochemical properties

```{r plot_constraints_fulfilled_individual, fig.show='hold', fig.width = 5, fig.height = 5, fig.cap = 'Failed and passed constraints on physicochemical properties for every primer. Constraints that are failed by a primer are colored in red, while constraints that are fulfilled are shown in blue. The provided p-value is an indicator of the overall quality of the primer set: significant p-values indicate primer sets fulfilling more constraints than reference primer sets from the literature.'}
openPrimeR::plot_constraint_fulfillment(params$primers, params$settings)
```

```{r plot_constraints_deviation, fig.show='hold', fig.width = 5, fig.height = 5, fig.cap = 'Deviation of physicochemical properties from the desired value range. Positive deviations indicate that the upper bound of a property was exceeded, while negative deviations indicate values below the lower bound. Each dot corresponds to the deviation of a primer for an individual constraint. The boxes indicate the 1st, 2nd (median), and 3rd quartiles from top to bottom.'}
openPrimeR::plot_constraint_deviation(params$primers, params$settings)
```

<!-- Input Parameters -->
```{r child = 'parameters.Rmd'}
```

```{r vignette_options_end, echo = FALSE}
# reset theme to previous setting
ggplot2::theme_set(old.theme)
```

---
output:
  pdf_document:
    fig_caption: yes
    includes:
        in_header: header.tex 
date: "`r format(Sys.time(), '%d %B, %Y')`"
params:
  sample: NA
  primers: NA
  templates: NA
  direction: NA
  settings: NA
  used_settings: NA

title: "`r paste0('Report for ', params$sample)`"
---

<!-- Vignette options and disclaimer-->

```{r child = 'disclaimer.Rmd'}
``` 

<!--dont show caption heading, this would clear the first page ... Primer Sets-->

```{r primer_table, results = 'asis'}
for (i in seq_along(params$primers)) {
    primer.df <- params$primers[[i]]
    template.df <- params$templates[[i]]
    id <- primer.df$Run[1]
    caption <-  paste0("Overview of the *", id, "* primer set.")
    out.df <- openPrimeR:::view.primers.report(primer.df, template.df)
    tab <- knitr::kable(out.df, row.names = FALSE, caption = caption,
                escape = TRUE, longtable = FALSE)
    print(tab)
    # pandoc doesn't support multi-column/row tables
    #row.idx <- which(colnames(out.df) == "Mismatches")
    #if (length(row.idx) != 0) {
        #tab <- kableExtra::add_header_above(tab, c("" = 1, "aa" = 2)
        #c("Mismatches" = row.idx[1], "" = tail(row.idx, n = 1)+1))
    #}
}
```

# Primer Set Summary

```{r primer_summary, results = 'asis', echo = FALSE}
#summary(cars)
tab <- openPrimeR:::get_comparison_table(params$templates, params$primers, params$sample)
caption <- paste0("Summary of the primer sets. Brackets indicate inter-quartile ranges.")
# show fewer columns to remain in the boundary of the page, don't show template (2nd column)
#knitr::kable(tab[,c(1,seq(3,5))], row.names = FALSE, caption = caption,
             #escape = TRUE, longtable = FALSE)
# use pander to split up the table instead:
pander::pander(tab[,c(1,seq(3,ncol(tab)))], caption = caption,
                split.table = 120, style = 'rmarkdown')

#########
# this is how a long table works (still too long ..)
#cat(paste0(
#"\\begin{table}[ht]\\centering\\rotatebox{90}{",
    #paste0(capture.output(
      #print(xtable(tab), floating=FALSE, comment=FALSE)),
      #collapse="\n"),
    #"}\\caption{Landscape table.}\\end{table}"))
############

```

# Coverage

```{r cvg_table, results = 'asis'}
cvg.stats <- openPrimeR::get_cvg_stats(params$primers, params$templates, for.viewing = TRUE)
knitr::kable(cvg.stats, row.names = FALSE, caption = "Percentage of covered template sequences per primer set and group of template sequences.", longtable = TRUE)
```

```{r coverage_plot, fig.show='hold', fig.width = 5, fig.height = 5, fig.cap = 'Coverage of each primer set.'}
openPrimeR:::plot_template_cvg(params$primers, params$templates)
```


```{r binding_plot_fw, fig.show='hold', fig.width = 6, fig.height = 6, fig.cap = 'The binding position of the primers relative to the target binding region of forward primers. The rectangles stretching along the x-axis indicate the target binding region of forward primers and the following region, respectively.'}
idx <- which(sapply(params$direction, function(x) x %in% c("fw", "both")))
if (length(idx) != 0) {
    openPrimeR::plot_primer_binding_regions(params$primers[idx], 
                params$templates[idx], relation = "fw")
}
```

```{r binding_plot_rev, fig.show='hold', fig.width = 6, fig.height = 6, fig.cap = 'The binding position of the primers relative to the target binding region of reverse primers. The rectangles stretching along the x-axis indicate the target binding region of reverse primers and the following region, respectively.'}
idx <- which(sapply(params$direction, function(x) x %in% c("rev", "both")))
if (length(idx) != 0) {
    openPrimeR::plot_primer_binding_regions(params$primers[idx], 
                    params$templates[idx], relation = "rev")
}
```

```{r cvg_vs_size, fig.show='hold', fig.width = 7, fig.height = 7, fig.cap = 'Coverage versus number of primers. Primer sets offering a good trade-off between coverage and set size will appear in the top left corner of the plot. The radius of each dot is correlated with the ratio at which the corresponding primer set fulfills the constraints.'}
openPrimeR::plot_cvg_vs_set_size(params$primers, params$templates)
```

# Physicochemical properties

```{r plot_constraints_fulfilled_overall, fig.show='hold', fig.width = 5, fig.height = 5, fig.cap = 'Percentage of constraints that are fulfilled by each primer set.'}
# not enough space for p-vals
#openPrimeR::plot_constraint_signature(params$primers, params$templates,  # TODO: does not work in package yet.
openPrimeR::plot_constraint_fulfillment(params$primers,
                                params$settings, 
                                plot.p.vals = FALSE) 
```

```{r plot_constraints_binding, fig.show='hold', fig.width = 5, fig.height = 5, fig.cap = 'Binding properties of the primers. The horizontal dashed lines indicate the desired property value ranges.'}
constraints <- c("primer_coverage", "primer_efficiency", "primer_specificity")
constraints <- constraints[constraints %in% names(openPrimeR::constraints(params$settings))]
openPrimeR::plot_constraint(params$primers, params$settings,
    active.constraints = c("primer_coverage", "primer_efficiency", 
                    "primer_specificity"))
```

```{r plot_constraints_structure, fig.show='hold', fig.cap = 'Structural properties of the primers. The horizontal dashed lines indicate the desired property value ranges.'}
constraints <- c("self_dimerization", "cross_dimerization", "secondary_structure")
constraints <- constraints[constraints %in% names(openPrimeR::constraints(params$settings))]
openPrimeR::plot_constraint(params$primers, params$settings,
      active.constraints = constraints)
```

```{r plot_constraints_composition, fig.show='hold', fig.cap = 'Properties relating to the nucleotide composition of the primers. The horizontal dashed lines indicate the desired property value ranges.'}
constraints <- c("primer_length", "no_runs",
                "no_repeats", "gc_ratio", 
                "gc_clamp", "melting_temp_range")
constraints <- constraints[constraints %in% names(openPrimeR::constraints(params$settings))]
openPrimeR::plot_constraint(params$primers, params$settings, 
    active.constraints = constraints)
```

<!-- Input parameters -->

```{r child = 'parameters.Rmd'}
```

```{r vignette_options_end, echo = FALSE}
# reset theme to default setting
ggplot2::theme_set(old.theme)
```

## Loading primers

Existing sets of primers can also be loaded from FASTA files. Why don't you try loading the following primer FASTA file using `read_primers()` into the `primer.df` variable?

```{r primer_FASTA_setup}
fasta.file <- system.file("extdata", "IMGT_data", "primers", 
        "IGHV", "Tiller2008_1st.fasta", package = "openPrimeR")
```

```{r primer_FASTA_1, exercise = TRUE}
# Load the prmiers from 'fasta.file' into 'primer.df' and explore the data
fasta.file <- system.file("extdata", "IMGT_data", "primers", 
        "IGHV", "Tiller2008_1st.fasta", package = "openPrimeR")
```

```{r primer_FASTA_1-solution}
# Load the prmiers from 'fasta.file' into 'primer.df' and explore the data
fasta.file <- system.file("extdata", "IMGT_data", "primers", 
        "IGHV", "Tiller2008_1st.fasta", package = "openPrimeR")
primer.df <- read_primers(fasta.file)
print(primer.df)
```

As we can see, the data frame only contains entries in the `Forward`, but not the `Reverse` column indicating that this is a set of forward primers only. Moreover, we obtain some basic information about the primers such as their length via `primer_length_fw` and their degeneracy via `Degeneracy_fw`. As a careful reader, you may now wonder how it was possible to annotate the primers with their directionalities without specifying any identifiers. 

This was possible because the input FASTA file already provided the default keywords used by `read_primers` for identifying the directions of the primers, which you verify by finding that the tag `_fw` is shared by all entries in the `ID` column. For your own primers, it is crucial that you tag the directionality of the primers in the FASTA file correctly and then specify the respective keywords using the `fw.id` and `rev.id` arguments in case that you are deviating from the defaults, for
example:

```{r primer_FASTA_example}
fasta.file <- system.file("extdata", "IMGT_data", "primers", 
        "IGHV", "Tiller2008_1st.fasta", package = "openPrimeR")
primer.df <- read_primers(fasta.file, fw.id = "_fw", rev.id = "_rev")
```

## Loading templates

Let's try to load the sequences from the following FASTA file containing germline cDNA from human immunoglobulin heavy chains :

```{r setup_template_FASTA}
fasta.file <- system.file("extdata", "IMGT_data", "templates", 
        "Homo_sapiens_IGH_functional_exon.fasta", package = "openPrimeR")
```

Please use an editor to get acquainted with the format of the file and then take a look at the template sequences using `read_templates()`.

```{r load_templates_FASTA_1, exercise = TRUE}
# Load the template sequences from 'fasta.file' and view the results
fasta.file <- system.file("extdata", "IMGT_data", "templates", 
        "Homo_sapiens_IGH_functional_exon.fasta", package = "openPrimeR")
``` 

```{r load_templates_FASTA_1-solution}
# Load the template sequences from 'fasta.file'
fasta.file <- system.file("extdata", "IMGT_data", "templates", 
        "Homo_sapiens_IGH_functional_exon.fasta", package = "openPrimeR")
template.df <- read_templates(fasta.file)
asS3(template.df)
``` 

As you can see, the templates were successfully loaded; Please explore the loaded templates. You can change columns by using the top right and left arrows. Comparing the loaded templates with the original FASTA file, we can see that the metadata contained in the headers of the FASTA file were not annotated by `read_templates()`.

### Annotating templates with FASTA metadata

To store the metadata correctly, you can use the `hdr.structure` and `delim` arguments of `read_templates()`, where `hdr.structure` provides a character vector of identifiers for information contained in the header and `delim` is a single character that is used in the header of the FASTA file to delimit individual metadata. When supplying the keyword 'GROUP' in the `hdr.structure` vector, template groups are annotated with their groups, which is relevant for visualizing properties of the templates later. Now, it's your turn: Try to load the the template sequences with their annotated IGHV groups:

```{r load_templates_FASTA_2, exercise = TRUE}
# Load the template sequences such that their groups are annotated correctly
fasta.file <- system.file("extdata", "IMGT_data", "templates", 
        "Homo_sapiens_IGH_functional_exon.fasta", package = "openPrimeR")
hdr.structure <- NULL # adjust according to the FASTA header
delim <- NULL # adjust according to the FASTA header
template.df <- read_templates(fasta.file, hdr.structure = hdr.structure, 
                               delim = delim)
```

```{r load_templates_FASTA_2-solution}
# Load the template sequences such that their groups are annotated correctly
fasta.file <- system.file("extdata", "IMGT_data", "templates", 
        "Homo_sapiens_IGH_functional_exon.fasta", package = "openPrimeR")
hdr.structure <- c("ACCESSION", "GROUP", "SPECIES", "FUNCTION")
delim <- "|"
template.df <- read_templates(fasta.file, hdr.structure = hdr.structure, 
                                delim = "|")
```

Note that, if not specified otherwise via the `id.column` argument, the first entry of `hdr.structure` is used as the template identifier. We  can verify that the template groups are available now via:


```{r load_templates_FASTA_3-setup}
fasta.file <- system.file("extdata", "IMGT_data", "templates", 
        "Homo_sapiens_IGH_functional_exon.fasta", package = "openPrimeR")
hdr.structure <- c("ACCESSION", "GROUP", "SPECIES", "FUNCTION")
delim <- "|"
template.df <- read_templates(fasta.file, hdr.structure = hdr.structure, 
                                delim = "|")
```


```{r load_templates_FASTA_3, exercise = TRUE}
print(template.df$Group)
```

### Defining the primer binding regions

Next, we will move on to defining the binding region of the primers in the templates. Since we deal with immunological templates, we want to annotate the leader region for each sequence individually. For this purpose, we will use a FASTA file containing the leader sequences corresponding to the templates:

```{r leader_file, exercise = TRUE}
leader.fasta <- system.file("extdata", "IMGT_data", "templates", 
        "Homo_sapiens_IGH_functional_leader.fasta", package = "openPrimeR")
```

Again, you may want to take a look at the structure of the file before continuing. The entries in the file containing the individual binding regions should match those in the template FASTA file and the provided regions should be subsequences of the loaded sequences.

We can define the binding region in the templates using `assign_binding_regions()`; since we only care about binding of the forward primers, we will only adjust the forward binding region.

```{r assign_leader, exercise = TRUE}
# Assign the forward binding regions from 'leader.fasta' to 'template.df':
leader.fasta <- system.file("extdata", "IMGT_data", "templates", 
        "Homo_sapiens_IGH_functional_leader.fasta", package = "openPrimeR")
```

```{r assign_leader-solution}
# Assign the forward binding regions from 'leader.fasta' to 'template.df':
leader.fasta <- system.file("extdata", "IMGT_data", "templates", 
        "Homo_sapiens_IGH_functional_leader.fasta", package = "openPrimeR")
template.df <- assign_binding_regions(template.df, fw = leader.fasta, rev = NULL)
```

We can verify that the binding regions for forward primers were annotated successfully in the following way:

```{r leader_template_setup}
fasta.file <- system.file("extdata", "IMGT_data", "templates", 
        "Homo_sapiens_IGH_functional_exon.fasta", package = "openPrimeR")
hdr.structure <- c("ACCESSION", "GROUP", "SPECIES", "FUNCTION")
delim <- "|"
template.df <- read_templates(fasta.file, hdr.structure = hdr.structure, 
                                delim = "|")
leader.fasta <- system.file("extdata", "IMGT_data", "templates", 
        "Homo_sapiens_IGH_functional_leader.fasta", package = "openPrimeR")
template.df <- assign_binding_regions(template.df, fw = leader.fasta, rev = NULL)

```

```{r verify_target_regions, exercise.setup = "leader_template_setup"}
utils::head(template.df$Allowed_fw)
cbind(utils::head(template.df$Allowed_Start_fw), utils::head(template.df$Allowed_End_fw))
```

where `Allowed_fw` contains the sequence of the leader region (the region before the exon) and `Allowed_Start_fw` and `Allowed_End_fw` provide the interval of allowed binding positions in the templates for forward primers.

### Adjusting the template-specific binding regions

As a final step, we need to modify the binding region for forward primer such that the binding region includes the first position of the exon, which lies outside the leader. Hence, we will extend the current binding region by one position. For this purpose, `adjust_binding_regions` can be used. This function requires a modified binding range, relative to the previously annotated binding region. The relative positions are defined such that position 0 is the first position after the end of the annotated binding region. For example, assigning the interval [0, 30] would allow binding within the first 30 positions of the exon. Next please extend the binding regions annotation of `template.df` by one position via `adjust_binding_regions()`:

```{r final_regions, exercise = TRUE, exercise.setup = "leader_template_setup"}
# Extend the binding regions of 'template.df' by one position:
```

```{r final_regions-solution}
# Extend the binding regions of 'template.df' by one position:
template.df <- adjust_binding_regions(template.df, 
        c(-max(template.df$Allowed_End_fw), 0), NULL)
# Verify the new annotation:
head(cbind(template.df$Allowed_Start_fw, template.df$Allowed_End_fw))
```

Note that we have chosen the maximum relative position for the start of the binding region to not affect any change, while we have extended the end of the binding region to the 0-th position, where 0 indicates the start of the exon. 

## Primer constraints

The physicochemical properties of the primers is facilitated by `check_cvg_constraints()`. Since we've already analyzed the primer coverage, we will compute all remaining constraints now by passing the corresponding `active.constraints` character vector to `check_cvg_constraints()`:

```{r my_constraint_setup}
xml.file <- system.file("extdata", "settings", "B_Taq_PCR_evaluate.xml",
              package = "openPrimeR")
settings <- read_settings(xml.file)
constraints(settings) <- constraints(settings)[names(constraints(settings)) != "gc_clamp"]
constraints(settings)$primer_coverage <- c("min" = 5)
conOptions(settings)$allowed_mismatches <- 7
conOptions(settings)$allowed_other_binding_ratio <- 0
conOptions(settings)$allowed_region_definition <- "any"

fasta.file <- system.file("extdata", "IMGT_data", "templates", 
        "Homo_sapiens_IGH_functional_exon.fasta", package = "openPrimeR")
hdr.structure <- c("ACCESSION", "GROUP", "SPECIES", "FUNCTION")
delim <- "|"
template.df <- read_templates(fasta.file, hdr.structure = hdr.structure, 
                                delim = "|")
leader.fasta <- system.file("extdata", "IMGT_data", "templates", 
        "Homo_sapiens_IGH_functional_leader.fasta", package = "openPrimeR")
template.df <- assign_binding_regions(template.df, fw = leader.fasta, rev = NULL)
template.df <- openPrimeR:::adjust_binding_regions(template.df, 
        c(-max(template.df$Allowed_End_fw), 0), NULL)

fasta.file <- system.file("extdata", "IMGT_data", "primers", 
        "IGHV", "Tiller2008_1st.fasta", package = "openPrimeR")
primer.df <- read_primers(fasta.file)
constraint.df <- check_constraints(primer.df, template.df, settings, active.constraints = "primer_coverage")
```

```{r check_constraints, exercise = TRUE, exercise.setup = "my_constraint_setup"}
# Check all constraints except for 'primer_coverage'
```

```{r check_constraints-solution}
constraint.df <- check_constraints(constraint.df, template.df, settings,
                 active.constraints = setdiff(names(constraints(settings)), "primer_coverage"))
asS3(constraint.df)

```

You can access the values of the computed properties in `constraint.df`. Fields containing the keyword `EVAL` indicate whether a primer passed or failed the constraints defined in `settings`. Let's try to access the computed melting temperatures and verify that the temperatures are in the desired range:

```{r constraint_setup}
xml.file <- system.file("extdata", "settings", "B_Taq_PCR_evaluate.xml",
              package = "openPrimeR")
settings <- read_settings(xml.file)
constraints(settings) <- constraints(settings)[names(constraints(settings)) != "gc_clamp"]
constraints(settings)$primer_coverage <- c("min" = 5)
conOptions(settings)$allowed_mismatches <- 7
conOptions(settings)$allowed_other_binding_ratio <- 0
conOptions(settings)$allowed_region_definition <- "any"

fasta.file <- system.file("extdata", "IMGT_data", "templates", 
        "Homo_sapiens_IGH_functional_exon.fasta", package = "openPrimeR")
hdr.structure <- c("ACCESSION", "GROUP", "SPECIES", "FUNCTION")
delim <- "|"
template.df <- read_templates(fasta.file, hdr.structure = hdr.structure, 
                                delim = "|")
leader.fasta <- system.file("extdata", "IMGT_data", "templates", 
        "Homo_sapiens_IGH_functional_leader.fasta", package = "openPrimeR")
template.df <- assign_binding_regions(template.df, fw = leader.fasta, rev = NULL)
template.df <- openPrimeR:::adjust_binding_regions(template.df, 
        c(-max(template.df$Allowed_End_fw), 0), NULL)

fasta.file <- system.file("extdata", "IMGT_data", "primers", 
        "IGHV", "Tiller2008_1st.fasta", package = "openPrimeR")
primer.df <- read_primers(fasta.file)
constraint.df <- check_constraints(primer.df, template.df, settings)
```

```{r constraint_manual, exercise = TRUE, exercise.setup = "constraint_setup"}
# Verify whether the melting temperatures are in the desired range
```

```{r constraint_manual-solution}
tm <- constraint.df$melting_temp
print(paste0("Melting temperatures: ", paste0(tm, collapse = ",")))
print(constraints(settings)$melting_temp_range)
idx <- which(constraint.df$EVAL_melting_temp_range)
print(tm[idx])
print(constraint.df[idx, "melting_temp"] <= constraints(settings)$melting_temp_range["max"] & 
      constraint.df[idx, "melting_temp"] >= constraints(settings)$melting_temp_range["min"])
```

Since analyzing the properties in this way is cumbersome, we'll create a visualization showing which primers fulfill the constraints using `plot_constraint_fulfillment()`:

```{r constraint_setup2}
xml.file <- system.file("extdata", "settings", "B_Taq_PCR_evaluate.xml",
              package = "openPrimeR")
settings <- read_settings(xml.file)
constraints(settings) <- constraints(settings)[names(constraints(settings)) != "gc_clamp"]
constraints(settings)$primer_coverage <- c("min" = 5)
conOptions(settings)$allowed_mismatches <- 7
conOptions(settings)$allowed_other_binding_ratio <- 0
conOptions(settings)$allowed_region_definition <- "any"

fasta.file <- system.file("extdata", "IMGT_data", "templates", 
        "Homo_sapiens_IGH_functional_exon.fasta", package = "openPrimeR")
hdr.structure <- c("ACCESSION", "GROUP", "SPECIES", "FUNCTION")
delim <- "|"
template.df <- read_templates(fasta.file, hdr.structure = hdr.structure, 
                                delim = "|")
leader.fasta <- system.file("extdata", "IMGT_data", "templates", 
        "Homo_sapiens_IGH_functional_leader.fasta", package = "openPrimeR")
template.df <- assign_binding_regions(template.df, fw = leader.fasta, rev = NULL)
template.df <- openPrimeR:::adjust_binding_regions(template.df, 
        c(-max(template.df$Allowed_End_fw), 0), NULL)

fasta.file <- system.file("extdata", "IMGT_data", "primers", 
        "IGHV", "Tiller2008_1st.fasta", package = "openPrimeR")
primer.df <- read_primers(fasta.file)
constraint.df <- check_constraints(primer.df, template.df, settings)
```


```{r constraint_plot, exercise = TRUE, exercise.setup = "constraint_setup"}
# Plot the constraint fulfillment
```


```{r constraint_plot-solution} 
plot_constraint_fulfillment(constraint.df, settings)
```

The plot quickly reveals that the main flaw of the primer set is associated with a high spread of primer melting temperatures. To quantify the deviations of the primer properties from the target constraints , we can use `plot_constraint_deviation`:


```{r constraint_deviation, exercise = TRUE, exercise.setup = "constraint_setup2"}
# Plot the constraint deviation
```

```{r constraint_deviation-solution}
plot_constraint_deviation(constraint.df, settings)
```

The plot shows that the constraint on the melting temperature deviations was violated most strongly, while some other constraints were violated only marginally. From these results we can conclude that, overall, the properties of the primers are reasonable. However, the high deviations in melting temperatures may require a relative low PCR annealing temperature and therefore lead to unspecific binding of the primers exhibiting high melting temperatures. 

Note that we haven't determined whether the GC clamp criterion is fulfilled since we've removed it at the beginning of the tutorial. Let's determine the values for the GC clamp and redraw the plots for the constraints:

```{r constraint_gc_clamp, exercise = TRUE, exercise.setup = "constraint_setup2"}
# Require 1 to 3 terminal GCs. Then determine the 'gc_clamp' property and plot the constraint fulfillment and deviation once again.
```

```{r constraint_gc_clamp-solution}
# require 1 to 3 terminal GCs
constraints(settings)$gc_clamp <- c("min" = 1, "max" = 3)
# update the evaluated primers
constraint.df <- check_constraints(constraint.df, template.df, settings, active.constraints = "gc_clamp")
# visualize the constraints
plot_constraint_fulfillment(constraint.df, settings)
plot_constraint_deviation(constraint.df, settings)
```

Great! Finally, we've really evaluated all constraints and there were no bad surprises when including the GC clamp condition. 

Having finished our analysis of the constraints, we may want to store the results in some way. For this purpose, there are two options: storing evaluated data as a CSV file or creating a PDF report. Storing the results as a CSV file can be time-saving if you want to re-analyze the same data at a later point in time, while the PDF report can be used for filing the results. You can store primers as a CSV with `write_primers()` and create a report with `create_report()`.

# Analysis settings

\begin{center}
\textbf{\textcolor{blue}{Please verify that the analysis was performed with the desired settings!}}
\end{center}

```{r analysis_settings, results = 'asis'}
# CONSTRAINTS TABLE
used.settings.fw <- NULL
used.settings.rev <- NULL
if (length(params$used_settings) == 0 || params$used_settings[1] == "NA") {
    used.settings.fw <- NULL
    used.settings.rev <- NULL
} else if (length(params$used_settings) == 1) {
    if (names(params$used_settings) == "fw") {
        used.settings.fw <- openPrimeR::constraints(params$used_settings[[1]])
        used.settings.rev <- NULL
    } else {
        used.settings.rev <- openPrimeR::constraints(params$used_settings[[1]])
        used.settings.fw <- NULL
    }
} else if (length(params$used_settings) == 2) {
   used.settings.fw <- openPrimeR::constraints(params$used_settings$fw)
   used.settings.rev <- openPrimeR::constraints(params$used_settings$rev)
} else {
    stop("Unsupported used settings length. Please input a list with the settings object for forward and reverse design.")
}
constraint.tab <- openPrimeR:::create.constraint.table(openPrimeR::constraints(params$settings), 
                            constraints.used.fw = used.settings.fw,
                            constraints.used.rev = used.settings.rev, format.type = "report")
caption.A <- "Constraints on the physicochemical properties of primers. The column \\textit{Target range} summarizes the user-specified constraints."
caption.B <- "The \\textit{Used range} columns indicate the actually applied constraints during the optimization procedure for forward (\\textit{fw}) and reverse (\\textit{rev}) primers, respectively."
if (length(used.settings.fw) == 0 && length(used.settings.rev) == 0) {
    # no primer design (used) settings available
    caption <- caption.A
} else {
    # used settings (design) available
    caption <- paste(caption.A, caption.B)
}
knitr::kable(constraint.tab, row.names = FALSE, caption = caption, longtable = TRUE)
# OPTIONS TABLE
tab <- openPrimeR:::create.options.table(openPrimeR::conOptions(params$settings), "report")
knitr::kable(tab, row.names = FALSE, caption = "Options for the active constraints. \\textit{Allowed mismatches} refers to the maximal allowed number of mismatches between a primer and a template. \\textit{Allowed off-target binding ratio} indicates the ratio of primers that are allowed to bind to non-target regions. \\textit{Binding region definition} defines whether primers shall bind within the target region (\\textit{within}) or are only required to overlap with the target region (\\textit{any}).", longtable = TRUE)
# COVERAGE SETTINGS TABLE 
tab <- openPrimeR:::create.constraint.table(openPrimeR::cvg_constraints(params$settings), format.type = "report")
knitr::kable(tab, row.names = FALSE, caption = "The conditions for primer coverage. \\textit{Stop codons} indicates whether mismatch binding events are allowed to induce stop codons (\\textit{1}) or not (\\textit{0}). \\textit{Efficiency} indicates the required amplification efficiency of the primers. \\textit{Annealing} gives the required free energy of annealing. \\textit{3\' Mismatch Position} indicates the positions in the primers (from the 3\' end such that \\textit{1} indicates the last position) for which mismatches are allowed.", longtable = TRUE)
# PCR TABLE
tab <- openPrimeR:::create.PCR.table(openPrimeR::PCR(params$settings), "report")
knitr::kable(tab, row.names = FALSE, caption = "The experimental conditions that were used to evaluate the primers.", longtable = TRUE)
```

```{r vignette_options, echo = FALSE, include = FALSE}
# can't disable longtable, it's a pandoc default -> ugly table breaks are the reuslt!
options(knitr.kable.NA = '') # ignore NAs in tables
knitr::opts_chunk$set(message = FALSE, warning = FALSE, echo = FALSE,
                      fig.width = 7, fig.height = 7, fig.pos = 'H',
                      fig.align='center') # out.extra: force latex figures
#options(knitr.table.format = "pandoc") 
# set figure size
old.theme <- ggplot2::theme_get()
ggplot2::theme_set(ggplot2::theme_grey(base_size = 10)) 
```
\begin{center}
\textbf{\textcolor{blue}{openPrimeR is a research tool - we do not guarantee for any result!}}
\end{center} 

## Preliminaries

This tutorial will introduce you to the core functionalities of the openPrimeR package:

* Loading primers and templates
* Analyzing existing sets of primers
* Designing novel primer sets

If you're already accustomed with these functions, you can skip this tutorial. Otherwise, let's start with attaching the openPrimeR package:


```{r eval = FALSE}
library(openPrimeR)
```

## Defining settings
There's a lot of experience with performing PCRs and therefore we know which physicochemical properties of primers are favorable for amplification and which are not. In openPrimeR, the desired ranges of values of these properties are called `constraints` and they are stored in a `DesignSettings` object. 

Let's define some constraints by loading one of the XML settings files that are shipped with the package by storing the result of `read_settings()` in the `settings` variable:

```{r read_settings, exercise = TRUE}
# Load the specified XML file into the 'settings' variable:
xml.file <- system.file("extdata", "settings", "B_Taq_PCR_evaluate.xml",
              package = "openPrimeR")
```

```{r read_settings-solution}
# Load the specified XML file into the 'settings' variable:
xml.file <- system.file("extdata", "settings", "B_Taq_PCR_evaluate.xml",
              package = "openPrimeR")
settings <- read_settings(xml.file)
```

```{r settings_setup}
xml.file <- system.file("extdata", "settings", "B_Taq_PCR_evaluate.xml",
              package = "openPrimeR")
settings <- read_settings(xml.file)
```

Let's take a look at the structure of `settings`:

```{r settings_structure, exercise = TRUE, exercise.setup = "settings_setup"}
print(settings)
```

As you can see, the `DesignSettings` object contains far more information than just the constraint settings, but let's focus on the constraints first. 

### Defining constraints on physicochemical properties

You can change which physicochemical properties are considered and the desired range of property values by using the `constraints()` function:

```{r settings_constraints, exercise = TRUE, exercise.setup = "settings_setup"}
# View the constraints specified in 'settings':
```

```{r settings_constraints-solution}
constraints(settings)
```

The `constraints()` function returns a named list where the names provide the identifiers of the active constraints and the constraints are specified as named vectors containing the entries `min` and `max` where `min` indicates the smallest allowed value and `max` indicates the maximal allowed value of a property. If either `min` or `max` is missing, this indicates that the corresponding range is unlimited. For example, the extent of the GC clamp should be between `r constraints(settings)$gc_clamp["min"]` and `r constraints(settings)$gc_clamp["max"]`, while the coverage of every primer should be at least `r constraints(settings)$primer_coverage["min"]` (there's no upper limit). 

You can customize the constraint settings by using `constraints()` as a setter. Why don't you try to exclude the GC clamp property from consideration and increase the required number of coverage events per primer to 5? Note that you always have to provide named (`min` and/or `max`) numeric vectors when modifying constraints.

```{r modify_constraints, exercise = TRUE}
# Remove the GC clamp constraint and set the minimal primer coverage to 5
```

```{r modify_constraints-solution}
# Remove the GC clamp constraint and set the minimal primer coverage to 5
constraints(settings) <- constraints(settings)[names(constraints(settings)) != "gc_clamp"]
constraints(settings)$primer_coverage <- c("min" = 5)
constraints(settings)
```

### Modifying the coverage constraints

```{r settings_config}
xml.file <- system.file("extdata", "settings", "B_Taq_PCR_evaluate.xml",
              package = "openPrimeR")
settings <- read_settings(xml.file)
constraints(settings) <- constraints(settings)[names(constraints(settings)) != "gc_clamp"]
constraints(settings)$primer_coverage <- c("min" = 5)
```

The coverage constraints determine under which circumstances a primer is considered to cover a template, which means that the primer is likely to successfully amplify the corresponding template. Depending on the selected coverage constraints, the estimate of amplified templates can be more sensitive (e.g. free energy of annealing) or more specific (modeling terminal mismatches). You can modify the coverage conditions using `cvg_constraints()`:

```{r cvg_constraints, exercise = TRUE, exercise.setup = "settings_config"}
# View the active coverage constraints
```


```{r cvg_constraints-solution}
# View the active coverage constraints
cvg_constraints(settings)
```

We see that the maximal false positive rate for calling coverage events is set to `r paste0(round(cvg_constraints(settings)$coverage_model * 100, 2), "%")`. This means that only coverage events whose estimated probability of being false positives is below `r paste0(round(cvg_constraints(settings)$coverage_model * 100, 2), "%")` are retained, while all other events are removed. `stop_codon` and `substitution` are codon design constraints, where 1 denominates coverage events inducing stop codons or substitutions and 0 indicates the absence of such events. Since the maxima of both constraints are set to 1, we do not filter coverage events according to these events at the moment. To learn about other possible coverage constraints, we refer to the openPrimeR manual or the output of viewing the `DesignSettings` object, which indicates the inactive constraints.

### Modifying the constraint options

The constraint options define additional options for the computation of some constraints. Let us now review the active constraint options via `conOptions()`:


```{r con_options, exercise = TRUE, exercise.setup = "settings_config"}
# View the active constraint options
```

```{r con_options-solution}
# View the active constraint options
conOptions(settings)
```

In the constraint options, `allowed_mismatches` provides the maximal number of mismatches between primers and templates, `allowed_other_binding_ratio` sets the maximal ratio of off-target binding events, and `allowed_region_definition` determines whether primers are required to bind within the specified binding region or may also only overlap with the target region. 

Please specify the following settings now:

* Set `allowed_mismatches` to 0 to ensure that only fully complementary primers are considered to cover the templates
* Set `allowed_other_binding_ratio` to 0 to ensure that only primers binding to the target region are considered to cover the templates
* Set `allowed_region_definition` to `any` to consider the coverage of primers that only overlap with the target region

```{r con_options_mod, exercise = TRUE, exercise.setup = "settings_config"}
# Modify the constraint options as suggested
```

```{r con_options_mod-solution}
conOptions(settings)$allowed_mismatches <- 0
conOptions(settings)$allowed_other_binding_ratio <- 0
conOptions(settings)$allowed_region_definition <- "any"
conOptions(settings)
```

```{r settings_config2}
xml.file <- system.file("extdata", "settings", "B_Taq_PCR_evaluate.xml",
              package = "openPrimeR")
settings <- read_settings(xml.file)
constraints(settings) <- constraints(settings)[names(constraints(settings)) != "gc_clamp"]
constraints(settings)$primer_coverage <- c("min" = 5)
conOptions(settings)$allowed_mismatches <- 0
conOptions(settings)$allowed_other_binding_ratio <- 0
conOptions(settings)$allowed_region_definition <- "any"
```

### Adjusting the PCR conditions

Last, we should take a look at the current PCR conditions using `PCR()`:

```{r settings_PCR,  exercise = TRUE, exercise.setup = "settings_config2"}
# Investigate the PCR conditions
```

```{r settings_PCR-solution}
# Investigate the PCR conditions
PCR(settings)
```

When performing an analysis, please ensure that all ion concentrations (molar) and other PCR conditions (e.g. the annealing temperature) reflect your experimental setup correctly. We will retain the default settings and just note that the PCR settings can be adjusted analogously to the other settings.

## Designing primers
You can design novel primer sets via `design_primers()`. Before you can start designing primers, you should perform the following steps:

1. Load the template sequences
2. Annotate the target binding regions in the templates
3. Define the primer design settings
4. Learn how to customize the primer design function

You should already be familiar with steps 1 to 3 from the previous tutorial sections, so we will only quickly recapitulate these steps and then focus on using the priemr design function. In this part of the tutorial, we will design forward primers for the leaders of the heavy chain of human germline immunoglobulins and will set up the template sequences accordingly in the following.

### Loading templates
We will load the functional human immunologublin heavy chain variable segments from a FASTA file with `read_templates()`:

```{r load_templates_design, exercise = TRUE}
fasta.file <- system.file("extdata", "IMGT_data", "templates", 
        "Homo_sapiens_IGH_functional_exon.fasta", package = "openPrimeR")
# Load the template data
```

```{r load_templates_design-solution}
fasta.file <- system.file("extdata", "IMGT_data", "templates", 
        "Homo_sapiens_IGH_functional_exon.fasta", package = "openPrimeR")
# Load the template data
hdr.structure <- c("ACCESSION", "GROUP", "SPECIES", "FUNCTION")
template.df <- read_templates(fasta.file, hdr.structure = hdr.structure, 
                              delim = "|")
```

```{r loaded_templates_design}
fasta.file <- system.file("extdata", "IMGT_data", "templates", 
        "Homo_sapiens_IGH_functional_exon.fasta", package = "openPrimeR")
# Load the template data
hdr.structure <- c("ACCESSION", "GROUP", "SPECIES", "FUNCTION")
template.df <- read_templates(fasta.file, hdr.structure = hdr.structure, 
                              delim = "|")
```

### Annotating binding regions
We will adjust the target binding regions uniformly across all templates  with `assign_binding_regions()`. This ensures that the primers are designed for the first positions of the leader:

```{r binding_uniform, exercise = TRUE, exercise.setup = "loaded_templates_design"}
# Specify the first 30 bases as the binding region for forward primers:
```

```{r binding_uniform-solution}
# Specify the first 30 bases as the binding region for forward primers:
template.df <- assign_binding_regions(template.df, fw = c(1,30))
```

### Defining the design settings
Let's load the default settings for designing primers via `read_settings()`:

```{r read_settings_design, exercise = TRUE}
# Load default settings from a supplied XML file:
xml.file <- system.file("extdata", "settings", "A_Taq_PCR_design.xml",
              package = "openPrimeR")
```

```{r read_settings_design-solution, exercise = TRUE}
# Load default settings from a supplied XML file:
xml.file <- system.file("extdata", "settings", "A_Taq_PCR_design.xml",
              package = "openPrimeR")
settings <- read_settings(xml.file)
```

```{r loaded_settings_design_pre}
# Load default settings from a supplied XML file:
xml.file <- system.file("extdata", "settings", "A_Taq_PCR_design.xml",
              package = "openPrimeR")
settings <- read_settings(xml.file)
```

Having loaded the supplied settings, it is your job to verify the settings for your primer design task. Here is a list of considerations that may be relevant for you:

* Are the specified constraints appropriate? Primers that do not fulfill the specified constraints may be removed during the filtering procedure. Hence, if the constraints are too narrow, the designed primer set may be too large. You can check/modify the constraints via `constraints()`.
* Are the specified constraint limits appropriate? The constraint limits define the speed at which the constraints are relaxed in case that no suitable set of primers could be found. The constraint limits can be checked/modified with `constraintLimits()`.
* Are the coverage constraints appropriate? The coverage constraints determine when a template is expected to be amplified by a primer. For designing primers, a more stringent coverage definition should be used to ensure that the designed set of primers is likely to amplify all target templates. The coverage conditions can be accessed via `cvg_constraints()`.
* Are the specified constraint options appropriate? Most importantly, ensure that the maximal number of mismatches is in accordance with your requirements. The constraint options can be modified with `conOptions()`.
* Do the specified PCR conditions match your experimental conditions? It is particularly important to ensure that the ion concentrations match your experimental setup to guarantee that the thermodynamic calculations are accurate. The PCR conditions can be accessed with `PCR()`. 

Possible fields for each slot of the settings object are described in the openPrimeR manual. Inactive options are shown when printing the settings object:


```{r print_settings, exercise = TRUE, exercise.setup = "loaded_settings_design_pre"}
# Print the settings object to view the currently active settings, as well as the inactive settings
```

```{r print_settings-solution}
# Print the settings object to view the currently active settings, as well as other possible, inactive settings
print(settings)
```

### Overview of the primer design procedure

The `design_primers()` function consists of three phases:

1. Initialization of primer candidates.
2. Filtering of primer candidates according to the constraints.
3. Selection of an optimized set of primers.

### Initialization of primer candidates

The procedure for initializing a set of primers is controlled by the `init.algo` argument. When set to `naive`, primers are initialized by extracting substrings from the input templates. If `init.algo` is set to `tree`, degenerate primeres are created. The maximal degeneration can be controlled via the `max.degen` argument.

### Filtering of primers

The filtering procedure is most affected by the constraints that are provided in the `settings` object passed to `design_primers`. However, there are also other arguments that influence the filtering procedure. Most importantly, `required.cvg` provides a numeric in the range [0,1] which indicates the desired coverage ratio of the templates. If the desired coverage ratio can't be achieved because too many primers have been filtered, a relaxation procedure is initialized. This procedure adjusts the constraints such that more primer candidates are selected in order to reach the target coverage.

### Selection of optimal primers

The selection of an optimal set of primers (i.e. solving the set cover problem) can be performed either by a greedy algorithm or an integer linear program. While the worst-case runtime of the greedy algorithm is less than that of the integer linear program, the integer linear program may be able to find a smaller set of optimal primers than the greedy algorithm. To use the greedy algorithm, you can set the `opti.algo` argument to `Greedy` and to use an integer linear program instead, you can set `opti.algo` to `ILP`.

### Using the primer design function

Let's design some primers targeting the leaders of the heavy chain immunologublins. Note that the primer design procedure needs some time to finish.

```{r design_settings, exercise = TRUE, message = FALSE}
# Load templates
fasta.file <- system.file("extdata", "IMGT_data", "templates", 
        "Homo_sapiens_IGH_functional_exon.fasta", package = "openPrimeR")
hdr.structure <- c("ACCESSION", "GROUP", "SPECIES", "FUNCTION")
template.df <- read_templates(fasta.file, hdr.structure = hdr.structure, 
                              delim = "|")
# Define binding regions
template.df <- assign_binding_regions(template.df, fw = c(1,30))
# Load settings
xml.file <- system.file("extdata", "settings", "A_Taq_PCR_design.xml",
              package = "openPrimeR")
settings <- read_settings(xml.file)
# Modify settings
constraints(settings)$primer_length <- c("min" = 18, "max" = 18)
# Design forward primers (only for 'target.templates' to save time) with naive initialization, and greedy optimization; store in 'design.data'
set.seed(1)
target.templates <- template.df[sample(nrow(template.df), 10),]
design.data <- design_primers(target.templates, "fw", settings,
                init.algo = "naive", opti.algo = "Greedy")
```

```{r design_settings-solution, message = FALSE}
# Load templates
fasta.file <- system.file("extdata", "IMGT_data", "templates", 
        "Homo_sapiens_IGH_functional_exon.fasta", package = "openPrimeR")
hdr.structure <- c("ACCESSION", "GROUP", "SPECIES", "FUNCTION")
template.df <- read_templates(fasta.file, hdr.structure = hdr.structure, 
                              delim = "|")
# Define binding regions
template.df <- assign_binding_regions(template.df, fw = c(1,30))
# Load settings
xml.file <- system.file("extdata", "settings", "A_Taq_PCR_design.xml",
              package = "openPrimeR")
settings <- read_settings(xml.file)
# Modify settings
constraints(settings)$primer_length <- c("min" = 18, "max" = 18)
# Design forward primers (only for 'target.templates' to save time) with naive initialization, and greedy optimization:
set.seed(1)
target.templates <- template.df[sample(nrow(template.df), 10),]
design.data <- design_primers(target.templates, "fw", settings,
                init.algo = "naive", opti.algo = "Greedy")
```

Excellent! Let's verify the quality of the designed primers:

```{r design_settings_setup}
# Load templates
fasta.file <- system.file("extdata", "IMGT_data", "templates", 
        "Homo_sapiens_IGH_functional_exon.fasta", package = "openPrimeR")
hdr.structure <- c("ACCESSION", "GROUP", "SPECIES", "FUNCTION")
template.df <- read_templates(fasta.file, hdr.structure = hdr.structure, 
                              delim = "|")
# Define binding regions
template.df <- assign_binding_regions(template.df, fw = c(1,30))
# Load settings
xml.file <- system.file("extdata", "settings", "A_Taq_PCR_design.xml",
              package = "openPrimeR")
settings <- read_settings(xml.file)
# Modify settings
constraints(settings)$primer_length <- c("min" = 18, "max" = 18)
# Design forward primers (only for the first 5 templates) with naive initialization, and greedy optimization:
set.seed(1)
target.templates <- template.df[sample(nrow(template.df), 10),]
design.data <- design_primers(target.templates, "fw", settings,
                init.algo = "naive", opti.algo = "Greedy")
```

```{r design_analysis1, exercise = TRUE, exercise.setup="design_settings_setup", message = FALSE}
# Determine the size of the optimized set:

# Determine the coverage of the optimized set

# Visualize the constraints

# Which constraints were used for the design?

# Which primers passed the filtering procedure?
```

```{r design_analysis1-solution}
# Determine the size of the optimized set:
primer.df <- design.data$opti
asS3(primer.df)
# Determine the coverage of the optimized set
plot_template_cvg(primer.df, target.templates)
# Visualize the constraints
plot_constraint_deviation(primer.df, settings)
# Which constraints were used for the design?
design.data$used_constraints
# Which primers passed the filtering procedure?
design.data$filtered$data
```

The results demonstrate that the designed primers fulfill the constraints very well and that all templates could be covered.

---
title: "openPrimeR Tutorial"
output: 
  learnr::tutorial:
    progressive: true
    allow_skip: true
runtime: shiny_prerendered
---

```{r setup, include=FALSE}
# global variables don't work for some reason ...
library(learnr)
library(openPrimeR)
tutorial_options(exercise.timelimit = 500)
```

```{r child = 'preliminaries.Rmd'}
```

```{r child = 'loading_templates.Rmd'}
```

```{r child = 'loading_primers.Rmd'}
```

```{r child = 'loading_settings.Rmd'}
```

```{r child = 'primer_coverage.Rmd'}
```

```{r child = 'evaluate_constraints.Rmd'}
```

```{r child = 'compare_primers.Rmd'}
```

```{r child = 'designing_primers.Rmd'}
```

## Primer coverage

```{r cvg_setup1}
xml.file <- system.file("extdata", "settings", "B_Taq_PCR_evaluate.xml",
              package = "openPrimeR")
settings <- read_settings(xml.file)
constraints(settings) <- constraints(settings)[names(constraints(settings)) != "gc_clamp"]
constraints(settings)$primer_coverage <- c("min" = 5)
conOptions(settings)$allowed_mismatches <- 0
conOptions(settings)$allowed_other_binding_ratio <- 0
conOptions(settings)$allowed_region_definition <- "any"

fasta.file <- system.file("extdata", "IMGT_data", "templates", 
        "Homo_sapiens_IGH_functional_exon.fasta", package = "openPrimeR")
hdr.structure <- c("ACCESSION", "GROUP", "SPECIES", "FUNCTION")
delim <- "|"
template.df <- read_templates(fasta.file, hdr.structure = hdr.structure, 
                                delim = "|")
leader.fasta <- system.file("extdata", "IMGT_data", "templates", 
        "Homo_sapiens_IGH_functional_leader.fasta", package = "openPrimeR")
template.df <- assign_binding_regions(template.df, fw = leader.fasta, rev = NULL)
template.df <- adjust_binding_regions(template.df, 
        c(-max(template.df$Allowed_End_fw), 0), NULL)

fasta.file <- system.file("extdata", "IMGT_data", "primers", 
        "IGHV", "Tiller2008_1st.fasta", package = "openPrimeR")
primer.df <- read_primers(fasta.file)
```

To learn more about the properties of the primers, we can use `check_constraints()`. In this part of the tutorial, we will specifically deal with the coverage of the templates that is afforded by the primers. To analyze only the coverage, you can supply `primer_coverage` to the `active.constraints` argument of `check_constraints()`:

```{r check1, exercise = TRUE, exercise.setup = "cvg_setup1", message = FALSE}
# Evaluate the primer coverage and store the results in 'constraint.df'
```

```{r check1-solution}
# Evaluate the primer coverage and store the results in 'constraint.df'
constraint.df <- check_constraints(primer.df, template.df, settings, active.constraints = "primer_coverage")
```

```{r cvg_setup2}
xml.file <- system.file("extdata", "settings", "B_Taq_PCR_evaluate.xml",
              package = "openPrimeR")
settings <- read_settings(xml.file)
constraints(settings) <- constraints(settings)[names(constraints(settings)) != "gc_clamp"]
constraints(settings)$primer_coverage <- c("min" = 5)
conOptions(settings)$allowed_mismatches <- 0
conOptions(settings)$allowed_other_binding_ratio <- 0
conOptions(settings)$allowed_region_definition <- "any"

fasta.file <- system.file("extdata", "IMGT_data", "templates", 
        "Homo_sapiens_IGH_functional_exon.fasta", package = "openPrimeR")
hdr.structure <- c("ACCESSION", "GROUP", "SPECIES", "FUNCTION")
delim <- "|"
template.df <- read_templates(fasta.file, hdr.structure = hdr.structure, 
                                delim = "|")
leader.fasta <- system.file("extdata", "IMGT_data", "templates", 
        "Homo_sapiens_IGH_functional_leader.fasta", package = "openPrimeR")
template.df <- assign_binding_regions(template.df, fw = leader.fasta, rev = NULL)
template.df <- adjust_binding_regions(template.df, 
        c(-max(template.df$Allowed_End_fw), 0), NULL)

fasta.file <- system.file("extdata", "IMGT_data", "primers", 
        "IGHV", "Tiller2008_1st.fasta", package = "openPrimeR")
primer.df <- read_primers(fasta.file)
constraint.df <- check_constraints(primer.df, template.df, settings, active.constraints = "primer_coverage")
```

Please investigate the structure of `constraint.df`. The column `primer_coverage` provides the number of covered templates and the column `Covered_Seqs` gives comma-separated strings with the identifiers of the covered templates. Let's try to find the primer with the highest coverage as well as the template sequences that are covered by the primer:

```{r highest_cvg, exercise = TRUE, exercise.setup = "cvg_setup2"}
# Investigate the structure of the primers and then find the primer with the highest coverage and the templates that it covers
```

```{r highest_cvg-solution}
asS3(constraint.df)
max.idx <- which.max(constraint.df$primer_coverage)
max.ID <- primer.df$ID[max.idx]
print(max.ID)
covered.templates.id <- strsplit(constraint.df$Covered_Seqs[max.idx], split = ",")[[1]]
covered.templates <- template.df$ID[match(covered.templates.id, template.df$Identifier)]
print(covered.templates)
```

Great! Let's visualize which templates are covered by the primers using `plot_template_cvg()`. 

```{r template_cvg_plot, exercise = TRUE, exercise.setup = "cvg_setup2"}
plot_template_cvg(constraint.df, template.df)
```

In the plot, `Identity Coverage` indicates the coverage when requiring full complementarity, while `Expected Coverage` provides the coverage when applying the coverage constraints. Since we didn't allow for any mismatches, both coverage values are basically identical. `Available Templates` provides the number of template sequences per group. From the plot we can see that, when we don't allow for any mismatches, about 47% of the templates are covered by the primers and that IGHV3, IGHV4, IGHV5, and IGHV7 are (partially) covered. To find out which primer amplifies which template groups, we can use `plot_primer_cvg`:

```{r primer_cvg_plot, exercise = TRUE, exercise.setup = "cvg_setup2"}
# Plot the primer coverage
```


```{r primer_cvg_plot-solution}
plot_primer_cvg(constraint.df, template.df)
```

This plot reveals that each primer binds with 100% complementarity only to individual groups of templates; note that the primer `VH_1` targets only IGHV7, which consists of a single template sequence. 

Our analysis was extremely conservative, since we didn't consider mismatch binding events. To more accurately estimate the coverage of the primers, let's ramp up the number of considered template-primer binding events by allowing to 7 mismatches using `conOptions()`:


```{r mod_settings_cvg, exercise = TRUE, exercise.setup = "cvg_setup"}
# Increase the number of allowed mismatches to 7
```


```{r mod_settings_cvg-solution}
conOptions(settings)$allowed_mismatches <- 7
```

```{r cvg_setup3}
xml.file <- system.file("extdata", "settings", "B_Taq_PCR_evaluate.xml",
              package = "openPrimeR")
settings <- read_settings(xml.file)
constraints(settings) <- constraints(settings)[names(constraints(settings)) != "gc_clamp"]
constraints(settings)$primer_coverage <- c("min" = 5)
conOptions(settings)$allowed_mismatches <- 7
conOptions(settings)$allowed_other_binding_ratio <- 0
conOptions(settings)$allowed_region_definition <- "any"

fasta.file <- system.file("extdata", "IMGT_data", "templates", 
        "Homo_sapiens_IGH_functional_exon.fasta", package = "openPrimeR")
hdr.structure <- c("ACCESSION", "GROUP", "SPECIES", "FUNCTION")
delim <- "|"
template.df <- read_templates(fasta.file, hdr.structure = hdr.structure, 
                                delim = "|")
leader.fasta <- system.file("extdata", "IMGT_data", "templates", 
        "Homo_sapiens_IGH_functional_leader.fasta", package = "openPrimeR")
template.df <- assign_binding_regions(template.df, fw = leader.fasta, rev = NULL)
template.df <- adjust_binding_regions(template.df, 
        c(-max(template.df$Allowed_End_fw), 0), NULL)

fasta.file <- system.file("extdata", "IMGT_data", "primers", 
        "IGHV", "Tiller2008_1st.fasta", package = "openPrimeR")
primer.df <- read_primers(fasta.file)
constraint.df <- check_constraints(primer.df, template.df, settings, active.constraints = "primer_coverage")
```

Let's re-analyze the coverage with the changed settings using `check_constraints()`:


```{r cvg2, exercise = TRUE, exercise.setup = "cvg_setup3"}
# Compute the coverage again
```

```{r cvg2-solution}
constraint.df <- check_constraints(primer.df, template.df, settings, active.constraints = "primer_coverage")
```

Let's visualize the template and primer coverage again to identify how the coverage has changed:


```{r cvg_plots2, exercise = TRUE, exercise.setup = "cvg_setup3"}
# Plot the template coverage and the primer coverage
```

```{r cvg_plots2-solution}
plot_template_cvg(constraint.df, template.df)
plot_primer_cvg(constraint.df, template.df)
```

The new results are impressively different to the previous ones. When we allow for more mismatches, quite a large percentage of templates are estimated to be covered and even the `VH_1` primer is revealed to cover multiple template groups at the same time. We can take a closer look at the distribution of coverage events occur ING for different numbers of mismatches between primers and templates by supplying the boolean `per.mismatch` argument to the two plotting functions:


```{r cvg_plots3, exercise = TRUE, exercise.setup = "cvg_setup3"}
# Plot the template coverage and the primer coverage, stratified by mismatches
```

```{r cvg_plots3-solution}
plot_template_cvg(constraint.df, template.df, per.mismatch = TRUE)
plot_primer_cvg(constraint.df, template.df, per.mismatch = TRUE)
```

These plots reveal that allowing for only 1 mismatch already provides more than 50% coverage and that most coverage events of `VH_1` occur with at least 6 mismatches.

Note that the estimated coverage takes into account only the properties of the primers that are directly associated with binding to a template (e.g. free energy and mismatches). Of course, there are many more properties that can influence whether an amplification is successful or not. 

For example, if a primer forms a complex with another primer in a multiplex reaction, this may greatly reduce product yields. Therefore, the estimated coverage should be used as an indicator for the coverage can be achieved if the other properties of the primers are reasonable. To determine if this is the case, we will analyze these properties in the next part of the tutorial.

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/initialize_primers_tree.R
\name{create.primers.tree}
\alias{create.primers.tree}
\title{Tree-based Initialization of Primers.}
\usage{
create.primers.tree(
  seqs,
  seq.IDs,
  seq.groups,
  start,
  end,
  primer.lengths,
  allowed.region.definition,
  max.degen,
  conservation,
  sample = "",
  identifier = "",
  updateProgress = NULL
)
}
\arguments{
\item{seqs}{Template sequences.}

\item{seq.IDs}{Identifiers of template sequences.}

\item{seq.groups}{Group identifiers of template sequences.}

\item{start}{For each template the start of the interval where primers 
should be created.}

\item{end}{For each template the end of the interval where primers
should be created.}

\item{primer.lengths}{Vector of desired primer lengths.}

\item{allowed.region.definition}{Definition of allowed regions.}

\item{max.degen}{Maximal degeneracy of primers.}

\item{conservation}{Required conservation of template regions considered
for the generation of primers. Conservation identifies the top conserved 
percentile of possible primers.}

\item{sample}{Sample name for the analysis.}

\item{identifier}{Identifier (e.g. for directionality).}

\item{updateProgress}{Shiny progress object.}
}
\value{
A vector with initialized primers.
}
\description{
Creates a set of candidate primers using a tree-based algorithm.
}
\details{
First, primers are aligned and their sequence similarity is determined to
compute a phylogenetic tree using hierarchical clustering.
Next, the tree is traversed from leaves to top in order to identify
groups of primers that can be merged (consensus) without exceeding the maximum
degeneracy of primers.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/templates.R
\name{add.uniform.leaders.to.seqs}
\alias{add.uniform.leaders.to.seqs}
\title{Add Uniform Binding Regions.}
\usage{
add.uniform.leaders.to.seqs(lex.seq, leaders)
}
\arguments{
\item{lex.seq}{Template data frame}

\item{leaders}{Data frame with uniform binding regions.}
}
\value{
Template data frame with updated binding regions.
}
\description{
Augments a template data frame with uniform binding regions.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/settings.R
\name{parse.constraints}
\alias{parse.constraints}
\title{Parse XML Constraint Data.}
\usage{
parse.constraints(xml_data)
}
\arguments{
\item{xml_data}{XML object from a parsed XML file.}
}
\value{
List with constraint settings.
}
\description{
Parses the constraint settings contained in an XML object.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/con_melting_temperature.R
\name{compute.melting.temps}
\alias{compute.melting.temps}
\title{Computation of Melting Temperatures.}
\usage{
compute.melting.temps(
  primer.df,
  primer_conc,
  na_salt_conc,
  mg_salt_conc,
  k_salt_conc,
  tris_salt_conc,
  mode.directionality = c("fw", "rev", "both")
)
}
\arguments{
\item{primer.df}{Primer data frame.}

\item{primer_conc}{Primer concentration.}

\item{na_salt_conc}{Sodium ion concentration.}

\item{mg_salt_conc}{Magensium ion concentration.}

\item{k_salt_conc}{Potassium ion concentration.}

\item{tris_salt_conc}{Tris buffer concentration.}

\item{mode.directionality}{Direction of primers}
}
\value{
Data frame with melting temperature info for the input primers.
}
\description{
Use nearest-neighbor thermodynamic computations to find melting temperatures.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/con_melting_temperature.R
\name{get.melting.temp.diff}
\alias{get.melting.temp.diff}
\title{Computation of Maximal Melting Temperature Differences.}
\usage{
get.melting.temp.diff(Tm.fw, Tm.rev)
}
\arguments{
\item{Tm.fw}{The melting temperatures of forward primers.}

\item{Tmr.rev}{The melting temperatures of reverse primers.}
}
\value{
The worst-case melting temperature difference, for every primer.
}
\description{
Computation of Maximal Melting Temperature Differences.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/filters.R
\name{cascaded.filter}
\alias{cascaded.filter}
\title{Filtering for the Optimization}
\usage{
cascaded.filter(
  primer.df,
  template.df,
  settings,
  mode.directionality = c("fw", "rev", "both"),
  required.cvg = 1,
  target.temps = NULL,
  updateProgress = NULL,
  results.loc = NULL
)
}
\arguments{
\item{primer.df}{Primer data frame.}

\item{template.df}{Template data frame.}

\item{settings}{Settings object.}

\item{mode.directionality}{Primer direction.}

\item{required.cvg}{Required ratio of covered templates.
If \code{required.cvg} is set to 0, the constraints are not relaxed.}

\item{target.temps}{Target melting temperature of the primers in Celsius.
This argument is only required if we try to match the melting temperatures of another primer set,
e.g. when first optimizing forward and then optimizing reverse primers.}

\item{updateProgress}{Progress callback function for shiny.}

\item{results.loc}{Directory where the filtering results should be stored.}
}
\value{
The filtered primer data frame with respect to \code{required.cvg}.
}
\description{
Filter primers according to constraints and relax constraints if necessary.
}
\details{
Constraints are relaxed if the \code{required.cvg} could not be reached with the input constraints.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ambiguity.R
\name{merge.primer.entries.single}
\alias{merge.primer.entries.single}
\title{Merge input sequences}
\usage{
\method{merge}{primer.entries.single}(seqs, merge.idx)
}
\arguments{
\item{seqs}{The input sequences.}

\item{merge.idx}{List of list with merge indices.}
}
\value{
Merged input sequences according to the input merge indices.
}
\description{
Merges the input sequences given a list with merge indices.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/settings.R
\name{filters}
\alias{filters}
\alias{filters,DesignSettings-method}
\title{Getter for Filtering Constraints.}
\usage{
filters(x)

\S4method{filters}{DesignSettings}(x)
}
\arguments{
\item{x}{A \code{DesignSettings} object.}
}
\value{
Gets the list of filtering constraints.
}
\description{
Gets the constraints on the physicochemical properties
that are used for the filtering procedure when designing primers
using the \code{Input_Constraints} slot of the provided
\code{DesignSettings} object \code{x}.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plots_constraints.R
\name{primer.set.parameter.stats}
\alias{primer.set.parameter.stats}
\title{Primer Set Statistics}
\usage{
primer.set.parameter.stats(primer.df, mode.directionality, lex.seq)
}
\arguments{
\item{primer.df}{Primer data frame.}

\item{mode.directionality}{Direction of primers.}

\item{lex.seq}{Template data frame.}
}
\value{
A data frame with statistics.
}
\description{
Creates an overview of all primer set constraint values.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/con_primer_secondary_structures.R
\name{read.secondary.structure.raw}
\alias{read.secondary.structure.raw}
\title{Read a Secondary Structure}
\usage{
read.secondary.structure.raw(fw.out)
}
\arguments{
\item{fw.out}{Path to a ViennaRNA secondary structure output file.}
}
\value{
Data frame with information on secondary structures.
}
\description{
Reads the secondary structure output of ViennaRNA.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/initialize_primers_tree.R
\name{score.conservation}
\alias{score.conservation}
\title{Conservation Scores}
\usage{
score.conservation(primer.range, ali.entropy)
}
\arguments{
\item{primer.range}{A data frame with starts/ends of primers.}

\item{ali.entropy}{Entropies corresponding to the alignment}
}
\value{
Entropies indicating conservation (similarity) of regions.
}
\description{
Scores the conservation of alignment regions.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/con_dimerization.R
\name{create.G.matrix}
\alias{create.G.matrix}
\title{Create free energy matrix}
\usage{
create.G.matrix(primer.df, G.df, primer.df.2 = NULL)
}
\arguments{
\item{primer.df}{Primer data frame.}

\item{G.df}{Free energy data for the primers.}

\item{primer.df.2}{Optional second primer data frame}
}
\value{
Matrix with the smallest free dimerization energy for every primer pair.
}
\description{
Creates a matrix giving the deltaG values of all primers.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/constraints_eval.R
\name{evaluate.template.constraints}
\alias{evaluate.template.constraints}
\title{Evaluation of Template Constraints.}
\usage{
evaluate.template.constraints(
  constraint.values,
  constraint.settings,
  active.constraints,
  mode.directionality = c("fw", "rev", "both")
)
}
\arguments{
\item{constraint.values}{Data frame with template constraints}

\item{constraint.settings}{List specifying the allowed values for constraint evaluation.}

\item{active.constraints}{Strings specifying the constraints to check.}

\item{mode.directionality}{Direction of primers.}
}
\value{
List indicating which template constraints where fulfilled or not (TRUE/FALSE).
}
\description{
Evaluates the input template constraints.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/settings.R
\name{get.other.constraint.settings}
\alias{get.other.constraint.settings}
\title{Gather all Other Constraints (for Shiny frontend).}
\usage{
get.other.constraint.settings(
  allowed_mismatches,
  allowed_other_binding_ratio,
  allowed_region_definition
)
}
\arguments{
\item{allowed_mismatches}{Allowed mismatches for primers binding events.}

\item{allowed_other_binding_ratio}{Ratio of primers allowed to bind to non-target regions.}

\item{allowed_region_definition}{The definition of the allowed region.}
}
\value{
List with all other constraint settings.
}
\description{
Constructor for other constraint settings (non-PCR, non-filtering, non-optimization).
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/con_melting_temperature.R
\name{compute.Tm.baldino}
\alias{compute.Tm.baldino}
\title{Baldino Formula}
\usage{
compute.Tm.baldino(
  sequences,
  na_salt_conc,
  mg_salt_conc,
  k_salt_conc,
  tris_salt_conc,
  primer_conc
)
}
\arguments{
\item{sequences}{Input sequence strings.}

\item{na_salt_conc}{Sodium ion concentration.}

\item{mg_salt_conc}{Magensium ion concentration.}

\item{k_salt_conc}{Potassium ion concentration.}

\item{tris_salt_conc}{Tris buffer concentration.}

\item{primer_conc}{Primer concentration.}
}
\value{
The melting temperature for the input sequences.
}
\description{
Computes the melting temperature using the formulation by Baldino.
}
\references{
Rychlik, W. J. S. W., W. J. Spencer, and R. E. Rhoads. "Optimization of the annealing temperature for DNA amplification in vitro." Nucleic acids research 18.21 (1990): 6409-6412.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/constraints_eval.R
\name{merge.template.decisions}
\alias{merge.template.decisions}
\title{Merge Template Decisions.}
\usage{
\method{merge}{template.decisions}(eval.t)
}
\arguments{
\item{eval.t}{List with evaluated template constraints}
}
\value{
List with merged boolean decisions.
}
\description{
Merges the results for multiple template evaluations.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/constraints_eval.R
\name{get.constraint.value.idx}
\alias{get.constraint.value.idx}
\title{Retrive Constraint Indices.}
\usage{
get.constraint.value.idx(active.constraints, constraint.df)
}
\arguments{
\item{active.constraints}{The names of the constraints for which to find the indices in \code{constraint.df}.}

\item{constraint.df}{The primer data frame where the \code{active.constraints} should be found.}
}
\value{
Indices of \code{active.constraints} in \code{constraint.df}.
}
\description{
Gets the index of the required constraint columns in the primer data frame.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/IO_view.R
\name{view.input.primers}
\alias{view.input.primers}
\title{View the Input Primers.}
\usage{
view.input.primers(primer.df, mode.directionality, for.shiny = TRUE)
}
\arguments{
\item{primer.df}{A \code{Primers} object.}

\item{mode.directionality}{The direction of the primers.}

\item{for.shiny}{Whether output is intended for Shiny.}
}
\value{
A formatted primer table.
}
\description{
Creates a formatted primers table.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/con_primer_coverage.R
\name{get.3prime.mismatch.pos}
\alias{get.3prime.mismatch.pos}
\title{Identification of 3' Mismatches.}
\usage{
get.3prime.mismatch.pos(primers, mismatches)
}
\arguments{
\item{primers}{Primer sequence strings.}

\item{mismatches}{Comma-separated strings containing the primer mismatch positions.}
}
\value{
The closest position of a mismatch relative to the 3' end of the primer.
Here, 1 indicates the terminal position, 2 the penultimate position, and so on.
No mismatch is indicated by an infinite value.
}
\description{
Computes the lastmost position of a 3' mismatches of a primer with a template.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/optimization_greedy.R
\name{select.best.primer.idx}
\alias{select.best.primer.idx}
\title{Greedy Choice}
\usage{
select.best.primer.idx(
  result,
  primers,
  deltaG.cutoff,
  target.temp,
  primer_conc,
  na_salt_conc,
  mg_salt_conc,
  k_salt_conc,
  tris_salt_conc
)
}
\arguments{
\item{result}{Data frame of current optimized primer data set that is to be augmented.}

\item{primers}{Data frame of candidate primers for addition to \code{result}.}

\item{deltaG.cutoff}{Free energy cutoff for cross-dimerization.}

\item{target.temp}{Target annealing temperature in Celsius.}

\item{primer_conc}{Primer concentration.}

\item{na_salt_conc}{Sodium ion concentration.}

\item{mg_salt_conc}{Magensium ion concentration.}

\item{k_salt_conc}{Potassium ion concentration.}

\item{tris_salt_conc}{Tris ion concentration.}
}
\value{
The index of a suitable primer according to Greedy selection.
}
\description{
Selects the currently best primer for Greedy primer selection.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/con_primer_coverage.R
\name{get.relative.binding.pos}
\alias{get.relative.binding.pos}
\title{Retrieval of Relative Binding Positions.}
\usage{
get.relative.binding.pos(allowed, primer.pos, direction, covered.seqs.idx)
}
\arguments{
\item{allowed}{Positions where binding is allowed in the templates.}

\item{primer.pos}{Binding position of primer (absolute).}

\item{direction}{Direction (either fw/rev).}

\item{covered.seqs.idx}{Indices of covered templates.}
}
\value{
Numeric of relative binding position to allowed region.
}
\description{
Retrieves primer binding position relative to allowed regions of either
forward or reverse primers, as specified by \code{direction}.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/con_primer_coverage.R
\name{update.cvg.data}
\alias{update.cvg.data}
\title{Update Coverage Information.}
\usage{
\method{update}{cvg.data}(
  filtered.df,
  sel,
  template.df,
  mode = c("on_target", "off_target"),
  active.constraints
)
}
\arguments{
\item{filtered.df}{Primer data frame.}

\item{sel}{List with indices of covered templates to be retained, one list with template indices to keep per primer.}

\item{template.df}{Template data frame.}

\item{mode}{Either \code{on_target} to filter on-target binding events
or \code{off_target} to filter off-target binding events. The corresponding
\code{sel} argument should be different.}

\item{active.constraints}{The active coverage constraints.}
}
\value{
A primer data frame with updated coverage information.
}
\description{
Updates the coverage-related columns in the input primer data frame.
Does not modify the entries of template-specific coverage columns
such as primer efficiency (comma-separated values).
}
\details{
Removes all coverage events of templates whose index is not in \code{sel}.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/con_template_secondary_structures.R
\name{optimize.template.binding.regions.single}
\alias{optimize.template.binding.regions.single}
\title{Optimization of Template Binding}
\usage{
optimize.template.binding.regions.single(
  template.df,
  annealing.temperature,
  primer.lengths,
  mode.directionality = c("fw", "rev")
)
}
\arguments{
\item{template.df}{Template data frame.}

\item{annealing.temperature}{Temperature at which to compute secondary structures.}

\item{primer.lengths}{Target length of primers that are to be used.}

\item{mode.directionality}{Direction of primers.}
}
\value{
List with new binding intervals for every template.
}
\description{
Optimizes template binding regions according to secondary structures.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/templates.R, R/TemplatesDoc.R
\name{TemplatesFunctions}
\alias{TemplatesFunctions}
\alias{update_template_cvg}
\alias{adjust_binding_regions}
\alias{assign_binding_regions}
\alias{select_regions_by_conservation}
\title{Template Functionalities.}
\usage{
update_template_cvg(template.df, primer.df, mode.directionality = NULL)

adjust_binding_regions(template.df, region.fw, region.rev)

assign_binding_regions(
  template.df,
  fw = NULL,
  rev = NULL,
  optimize.region = FALSE,
  primer.length = 20,
  gap.char = "-"
)

select_regions_by_conservation(
  template.df,
  gap.char = "-",
  win.len = 30,
  by.group = TRUE,
  mode.directionality = c("both", "fw", "rev")
)
}
\arguments{
\item{template.df}{An object of class \code{Templates}.}

\item{primer.df}{An object of class \code{Primers} containing
primers with annotated coverage that are to be used to update 
the template coverage in \code{template.df}.}

\item{mode.directionality}{The directionality of primers/templates.}

\item{region.fw}{Interval of new binding regions relative to the forward binding region defined in \code{template.df}.}

\item{region.rev}{Interval of new binding regions relative to the reverse binding region defined in \code{template.df}}

\item{fw}{Binding regions for forward primers. Either a numeric interval indicating a uniform
binding range relative to the template 5' end or a path to a FASTA file providing
binding sequences for every template. If \code{fw} is missing, only
\code{rev} is considered.}

\item{rev}{Binding regions for reverse primers. Either a numeric interval indicating a uniform
binding range relative to the template 3' end or the path to a FASTA file providing
binding sequences for every template. If \code{rev} is missing,
only \code{fw} is considered.}

\item{optimize.region}{If \code{TRUE}, the binding regions
specified via \code{fw} and \code{rev} are 
adjusted such that binding regions that may form secondary structures are 
avoided. This feature requires ViennaRNA (see notes). If \code{FALSE}
(the default), the input binding regions are not modified.}

\item{primer.length}{A numeric scalar providing the probe length that is used for
adjusting the primer binding regions when \code{optimize.region} is \code{TRUE}.}

\item{gap.char}{The character in the input file representing gaps.}

\item{win.len}{The extent of the desired primer binding region.
This should be smaller than the \code{allowed.region}. The default is 30.}

\item{by.group}{Shall the determination of binding regions be stratified
according to the groups defined in \code{template.df}. By default,
this is set to \code{TRUE}.}
}
\value{
\code{update_template_cvg} returns an object of class 
\code{Templates} with updated coverage columns.

\code{adjust_binding_regions} returns a \code{Templates} object with
updated binding regions.

\code{assign_binding_regions} returns an object of class \code{Templates} with newly assigned binding regions.

\code{select_regions_by_conservation} returns a
\code{Templates} object with adjusted binding regions.
The attribute \code{entropies} gives a data frame with positional entropies
and the attribute \code{alignments} gives the alignments of the templates.
}
\description{
\describe{
\item{\code{adjust_binding_regions}}{Adjusts the existing annotation 
of binding regions by specifying
a new binding interval relative to the existing binding region.}
\item{\code{assign_binding_regions}}{Assigns the primer target binding 
regions to a set of template sequences.}
\item{\code{update_template_cvg}}{Annotates the template coverage.}
\item{\code{select_regions_by_conservation}}{Computes Shannon entropy
for the defined binding regions and determines the most conserved regions.}
}
}
\details{
When modifying binding regions with \code{adjust_binding_regions}, new
binding intervals can be specified via \code{fw} and \code{rev}
for forward and reverse primers, respectively. The new regions should
be provided relative to the existing definition of binding regions 
in \code{template.df}.
For specifying the new binding regions, position \code{0} refers to 
the first position after the end of the existing
binding region. Hence, negative positions relate to regions within the
existing binding region, while non-negative values relate to positions
outside the defined binding region.

Binding regions are defined using \code{assign_binding_regions}, where
the arguments \code{fw} and \code{rev} provide data describing
the binding regions of the forward and reverse primers, respectively.
To specify binding regions for each template
individually, \code{fw} and \code{rev} should provide the paths to FASTA files. The headers of these
FASTA file should match the headers of the loaded \code{template.df} 
and the sequences in the files specified by \code{fw} and \code{rev} should indicate
the target binding regions. 

To specify uniform binding regions,
\code{fw} and \code{rev} should be numeric intervals indicating the allowed
binding range for primers in the templates. Setting the forward interval
to (1,30) indicates that the first 30 bases should be used for forward primers
and specifying the reverse interval to (1,30) indicates that the last
30 bases should be used for reverse primer binding.

If \code{optimize.region} is \code{TRUE}, the input binding region is
adjusted such that regions forming secondary structures are avoided.
}
\note{
\code{assign_binding_regions} requires the program ViennaRNA
(https://www.tbi.univie.ac.at/RNA/)
for adjusting the binding regions when \code{optimize.region} 
is set to \code{TRUE}.

\code{select_regions_by_conservation} requires the MAFFT software
for multiple alignments 
(http://mafft.cbrc.jp/alignment/software/).
}
\examples{

# Annotate the coverage of the templates
data(Ippolito)
template.df <- update_template_cvg(template.df, primer.df)
data(Ippolito)
# Extend the binding region by one position
relative.interval <- c(-max(template.df$Allowed_End_fw), 0)
template.df.adj <- adjust_binding_regions(template.df, relative.interval)
# compare old and new annotations:
head(cbind(template.df$Allowed_Start_fw, template.df$Allowed_End_fw))
head(cbind(template.df.adj$Allowed_Start_fw, template.df.adj$Allowed_End_fw))
data(Ippolito)
# Assignment of individual binding regions
l.fasta.file <- system.file("extdata", "IMGT_data", "templates", 
     "Homo_sapiens_IGH_functional_leader.fasta", package = "openPrimeR")
template.df.individual <- assign_binding_regions(template.df, l.fasta.file, NULL)
# Assign the first/last 30 bases as forward/reverse binding regions
template.df.uniform <- assign_binding_regions(template.df, c(1,30), c(1,30))
# Optimization of binding regions (requires ViennaRNA)
\dontrun{template.df.opti <- assign_binding_regions(template.df, c(1,30), c(1,30),
                     optimize.region = TRUE, primer.length = 20)}
data(Ippolito)
new.template.df <- select_regions_by_conservation(template.df)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plots_constraints.R
\name{plot_constraint.histogram.primer.efficiencies}
\alias{plot_constraint.histogram.primer.efficiencies}
\title{Histogram of Efficiencies}
\usage{
plot_constraint.histogram.primer.efficiencies(primer.df, opti.constraints)
}
\arguments{
\item{primer.df}{Primer data frame.}

\item{opti.constraints}{List with constraint settings.}
}
\value{
A plot of primer efficiencies.
}
\description{
Plots a histogram of primer efficiencies.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/filters.R
\name{filter.by.constraints}
\alias{filter.by.constraints}
\title{Filter By Constraints}
\usage{
filter.by.constraints(
  filtered.df,
  constraint.df,
  current.constraints,
  active.constraints,
  mode.directionality = c("fw", "rev", "both"),
  template.df
)
}
\arguments{
\item{filtered.df}{Primer data frame.}

\item{constraint.df}{Data frame with constraint values.}

\item{current.constraints}{List with constraint settings.}

\item{active.constraints}{Strings giving the names of active constraints.}

\item{mode.directionality}{Direction of primers}

\item{template.df}{Template data frame.}
}
\value{
A list containing the filtered primer data frame, as well as a 
data frame of the excluded primers and the used filtering settings.
}
\description{
Remove primers that do not fulfill the current constraints (evaluate all primers).
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/optimization_algo.R
\name{visualize.all.results}
\alias{visualize.all.results}
\title{Visualization of Design Results.}
\usage{
visualize.all.results(
  sample,
  filtering.results.loc,
  opti.results.loc,
  primer_conc,
  na_salt_conc,
  mg_salt_conc,
  k_salt_conc,
  tris_salt_conc,
  settings,
  mode.directionality,
  used.settings,
  required.cvg
)
}
\arguments{
\item{sample}{Identifier of the design run.}

\item{filtering.results.loc}{Location of filtering results.}

\item{opti.results.loc}{Location of optimization results.}

\item{primer_conc}{Primer concentration.}

\item{na_salt_conc}{Sodium ion concentration.}

\item{mg_salt_conc}{Magensium ion concentration.}

\item{k_salt_conc}{Potassium ion concentration.}

\item{tris_salt_conc}{Tris ion concentration.}

\item{settings}{The \code{DesignSettings} object.}

\item{mode.directionality}{Strand direction for which primers were designed.}

\item{used.settings}{A list with the used settings for optimization (fields "fw" and "rev").}

\item{required.cvg}{The required coverage.}

\item{template_conc}{Template concentration.}
}
\value{
Writes visualizations to files in.
}
\description{
Visualizes all results from designing primers.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/settings.R
\name{rename.constraint.options}
\alias{rename.constraint.options}
\title{Renaming of Constraint Options.}
\usage{
rename.constraint.options(constraint.options)
}
\arguments{
\item{constraint.options}{A list with constraint options.}
}
\value{
A list with renamed constraint options.
}
\description{
Renames the input list with constraint options.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/settings.R
\name{listToXml}
\alias{listToXml}
\title{List to XML}
\usage{
listToXml(item, tag)
}
\arguments{
\item{item}{}

\item{tag}{xml tag}
}
\value{
xmlNode
}
\description{
Convert List to XML.
}
\details{
Can convert list or other object to an xml object using xmlNode.
}
\author{
David LeBauer, Carl Davidson, Rob Kooper
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/settings.R
\name{convert.PCR.units}
\alias{convert.PCR.units}
\title{Conversion of PCR Units}
\usage{
convert.PCR.units(pcr.settings, to.mol = TRUE)
}
\arguments{
\item{pcr.settings}{List with several PCR settings (concentrations).}

\item{to.mol}{If \code{TRUE}, convert to the molar concentration.
If \code{FALSE} convert to the unit representation in the XML.}
}
\value{
List with concentrations for usage in the backend.
}
\description{
Converts frontend PCR concentration units to the units used for the backend.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ambiguity.R
\name{rev.sequence}
\alias{rev.sequence}
\title{Reversion of a sequence}
\usage{
\method{rev}{sequence}(seq)
}
\arguments{
\item{seq}{the input sequence.}
}
\value{
The input sequence in reverse order.
}
\description{
Reverses the input sequences.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plots_constraints.R
\name{dimerization.table}
\alias{dimerization.table}
\title{Dimerization Table.}
\usage{
dimerization.table(
  dimer.data,
  deltaG.cutoff,
  dimer.type = c("Self-Dimerization", "Cross-Dimerization")
)
}
\arguments{
\item{dimer.data}{Data frame with dimerization data.}

\item{deltaG.cutoff}{Free energy cutoff for dimerization.}

\item{dimer.type}{String identifying whether \code{dimer.data} refers to cross-dimers or self-dimers?}
}
\value{
Data frame with dimer counts.
}
\description{
Summarizes how often individual primers dimerize according to the \code{deltaG.cutoff}.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/zzz.R
\name{selenium.installed}
\alias{selenium.installed}
\title{Determination if Selenium is installed.}
\usage{
selenium.installed()
}
\value{
\code{TRUE} is selenium for python is available,
\code{FALSE} otherwise.
}
\description{
Checks whether selenium module for python is installed on the system.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/primers.R
\name{read_primers.internal}
\alias{read_primers.internal}
\title{Internal Function for Reading Primers}
\usage{
read_primers.internal(
  primer.seqs,
  headers,
  fw.id,
  rev.id,
  merge.ambig = c("none", "merge", "unmerge"),
  max.degen,
  sample.name
)
}
\arguments{
\item{primer.seqs}{Primer sequences.}

\item{headers}{Headers of the primer FASTA file.}

\item{fw.id}{Identifier of forward primers in the headers.}

\item{rev.id}{Identifier of reverse primers in the headers.}

\item{merge.ambig}{Should ambiguous primers be merged?}

\item{max.degen}{Maximum allowed degeneracy}
}
\value{
A data frame with primer sequences.
}
\description{
Reads the given primer sequences into a data frame.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/optimization_algo.R
\name{create.Tm.brackets}
\alias{create.Tm.brackets}
\title{Creation of Melting Temperature Groups}
\usage{
create.Tm.brackets(primers, template.df, settings, target.temps = NULL)
}
\arguments{
\item{primers}{An object of class \code{Primers} for which to create groups based on melting temperatures.}

\item{template.df}{An object of class \code{Templates} corresponding to the primers.}

\item{settings}{A \code{DesignSettings} objects.}

\item{target.temps}{Pre-defined target melting temperatures to use instead
of automatically determining groups from the \code{primers}.}
}
\value{
Data frame with target melting temperatures for individual primer sets.
}
\description{
Creates a data frame identifying target melting temperatures of individual primer sets.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/templates.R
\name{parse.header}
\alias{parse.header}
\title{Parse FASTA headers}
\usage{
parse.header(hdr, delim, hdr.str, id.column)
}
\arguments{
\item{hdr}{The headers (> entries) of a FASTA File.}

\item{delim}{The delimiter used to separate distinct fields in the headers.
For example, \code{|} for headers such as > E.coli | GeneX | ...}

\item{hdr.str}{Names of the fields appearing in the header.}

\item{id.column}{Field in the header to be used used as an identifier for the sequences.}
}
\value{
Data frame with structured information from the headers.
}
\description{
Parses the headers of a FASTA file.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/templates.R
\name{ungap_sequence}
\alias{ungap_sequence}
\title{Ungapping of Sequences.}
\usage{
ungap_sequence(seqs, gap.char = "-")
}
\arguments{
\item{seqs}{The input character vector with sequences}

\item{gap.char}{The character used to represent gaps.}
}
\value{
\code{seqs} with gaps removed.
}
\description{
Removes gaps from the input sequences.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/optimization_algo.R
\name{get.init.file.name}
\alias{get.init.file.name}
\title{File Name for Initialized Primers.}
\usage{
get.init.file.name(
  cur.results.loc,
  GROUP,
  primer.lengths,
  mode.directionality,
  allowed.region.definition,
  init.algo,
  max.degen,
  conservation
)
}
\arguments{
\item{cur.results.loc}{Directory where the file should be stored.}

\item{GROUP}{Sample name of templates.}

\item{primer.lengths}{Interval of desired primer lengths.}

\item{mode.directionality}{Directionality of the primers}

\item{allowed.region.definition}{Definition of the allowed region.}

\item{init.algo}{Initialization algorithm identifier.}

\item{max.degen}{Maximum degeneracy of primers.}

\item{conservation}{Required ratio of primer conservation.}
}
\value{
A filename for the initialized primers.
}
\description{
Constructs a filename for initialized primers.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/optimization_ILP.R
\name{get.ILP.vars}
\alias{get.ILP.vars}
\title{Retrieval of ILP Decisions}
\usage{
get.ILP.vars(ILP, original.dim = NULL)
}
\arguments{
\item{ILP}{A solved ILP instance.}

\item{original.dim}{Dimension of \code{ILP} before using presolve.}
}
\value{
The ILP decision variables.
}
\description{
Retrieves ILP decision variables.
}
\details{
The original dimension of the ILP is required to determine the correct decisions
when presolve has been active and dimensions of the ILP might have changed.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/filters.R
\name{cascaded.filter.quick}
\alias{cascaded.filter.quick}
\title{Cascaded Filter}
\usage{
cascaded.filter.quick(
  primer.df,
  template.df,
  settings,
  to.compute.constraints,
  mode.directionality = c("fw", "rev", "both"),
  active.constraints = NULL,
  no.structures = FALSE,
  updateProgress = NULL
)
}
\arguments{
\item{primer.df}{Primer data frame.}

\item{template.df}{Template data frame.}

\item{settings}{Settings object.}

\item{to.compute.constraints}{Names of constraints that still have to be computed.}

\item{mode.directionality}{Primer direction.}

\item{active.constraints}{The constraints that are to be used for filtering.
If \code{active.constraints} is \code{NULL}, all filtering constraints are used.}

\item{no.structures}{Whether dimerization structures shall be computed.}

\item{updateProgress}{Progress callback function for shiny.}
}
\value{
The filtered primer data frame.
}
\description{
Filter primers in a cascaded fashion.
}
\details{
At each constraint evaluation all primers that do not fulfill the current constraint are removed.

Constraints that are specified in \code{to.compute.constraints} are computed on the fly.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/con_annealing_temperature.R
\name{compute.Ta}
\alias{compute.Ta}
\title{Annealing temperature}
\usage{
compute.Ta(
  primer.df,
  template.df,
  mode.directionality = c("fw", "rev", "both"),
  na_salt_conc,
  mg_salt_conc,
  k_salt_conc,
  tris_salt_conc,
  primer_conc
)
}
\arguments{
\item{primer.df}{Primer data frame.}

\item{template.df}{Template data frame}

\item{mode.directionality}{Primer directionality.}

\item{na_salt_conc}{Sodium ion concentration.}

\item{mg_salt_conc}{Magensium ion concentration.}

\item{k_salt_conc}{Potassium ion concentration.}

\item{tris_salt_conc}{Tris buffer concentration.}

\item{primer_conc}{Primer concentration.}
}
\value{
All annealing temperatures for given binding events.
}
\description{
Computes the annealing temperature using all binding events.
}
\references{
Rychlik, W. J. S. W., W. J. Spencer, and R. E. Rhoads. "Optimization of the annealing temperature for DNA amplification in vitro." Nucleic acids research 18.21 (1990): 6409-6412.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/con_primer_coverage.R
\name{select.allowed.binding.events}
\alias{select.allowed.binding.events}
\title{Selection of Binding Events}
\usage{
select.allowed.binding.events(
  bound.fw,
  bound.to.allowed.region.fw,
  allowed.other.binding.ratio
)
}
\arguments{
\item{bound.fw}{Indices of covered templates of a single primer.}

\item{bound.to.allowed.region.fw}{Corresponding allowed binding regions.}

\item{allowed.other.binding.ratio}{The ratio of other binding events. If
this is different from 0, disallowed binding events will also be reported.}
}
\value{
The indices of the allowed binding events.
}
\description{
Selects primer binding events that are within the allowed binding regions.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plots_coverage.R
\name{plot_template_cvg,list,list-method}
\alias{plot_template_cvg,list,list-method}
\title{Templates Coverage for Multiple Primer Sets.}
\usage{
\S4method{plot_template_cvg}{list,list}(primers, templates, per.mismatch, highlight.set = NULL)
}
\arguments{
\item{primers}{List with primer data frames.}

\item{templates}{List with template data frames.}

\item{highlight.set}{Primer sets to be highlighted.}

\item{colors}{Color for every primer set.}
}
\value{
A plot for comparing primer coverage.
}
\description{
Plots the coverage of multiple primer sets.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/primers.R
\name{read_primers_multiple}
\alias{read_primers_multiple}
\title{Input of Multiple Primer Sets.}
\usage{
read_primers_multiple(
  filenames,
  fw.id,
  rev.id,
  merge.ambig,
  max.degen,
  template.df,
  adapter.action,
  sample.name,
  updateProgress
)
}
\arguments{
\item{filenames}{The paths to multiple primer CSV/FASTA files.}
}
\value{
A list containing objects of class \code{Primers}.
}
\description{
Reads multiple CSV files representing stored objects of class \code{Primers}.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/con_dimerization.R
\name{compute.all.cross.dimers.unfiltered}
\alias{compute.all.cross.dimers.unfiltered}
\title{Cross dimerization}
\usage{
compute.all.cross.dimers.unfiltered(
  primer.df,
  primer_conc,
  na_salt_conc,
  mg_salt_conc,
  k_salt_conc,
  tris_salt_conc,
  annealing.temp,
  check.idx = NULL,
  for.shiny = FALSE,
  no.structures = FALSE
)
}
\arguments{
\item{primer.df}{Input primers data frame.}

\item{primer_conc}{Primer concentration.}

\item{na_salt_conc}{Sodium ion concentration.}

\item{mg_salt_conc}{Magensium ion concentration.}

\item{k_salt_conc}{Potassium ion concentration.}

\item{tris_salt_conc}{Tris buffer concentration.}

\item{annealing.temp}{The PCR annealing temperature.}

\item{check.idx}{Indices of primers for checking cross-dimerization.}

\item{for.shiny}{Whether to format for HTML output.}

\item{no.structures}{Whether dimer structures shall not be determined.
If \code{TRUE}, structures are not computed resulting in faster runtimes.}
}
\value{
All cross dimers.
}
\description{
Compute DeltaG data frame for possible primer cross-dimers.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/con_primer_coverage.R
\name{annotate.binding.events}
\alias{annotate.binding.events}
\title{Annotation of Primer Binding Events.}
\usage{
annotate.binding.events(
  fw.binding,
  allowed.range,
  nbr.primers,
  allowed.region.definition = c("within", "any")
)
}
\arguments{
\item{fw.binding}{IRanges with coverage information.}

\item{allowed.range}{IRanges of the allowed binding ranges in the templates.}

\item{nbr.primers}{Number of primers to consider.}

\item{allowed.region.definition}{Definition of the allowed binding region}
}
\value{
IRanges with annotations of (preliminary) specificity and allowed binding. 
The field \code{all_binding} contains all binding regions, \code{on_target} contains
all events in the target region, and \code{off_target} contains all off-target
binding events.
}
\description{
Annotates whether primer binding events are in the allowed binding region or not.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/templates.R
\name{assign_binding_regions.character}
\alias{assign_binding_regions.character}
\title{Character Assignment of Binding Regions}
\usage{
\method{assign_binding_regions}{character}(
  template.df,
  fw,
  rev,
  optimize.region = FALSE,
  primer.length = 20,
  gap.char = "-"
)
}
\arguments{
\item{template.df}{Template data frame.}

\item{fw}{FASTA file specifying the forward binding regions.}

\item{rev}{FASTA file specifying the reverse binding regions.}

\item{optimize.region}{Should the primer binding region be optimized using secondary structure prediction?}

\item{primer.length}{Probe length for optimizing template secondary structure.}

\item{gap.char}{The gap character for aligned sequences.}
}
\value{
Template data frame with assigned binding regions.
}
\description{
Generic method for assigning the binding region using individual binding regions.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/initialize_primers_tree.R
\name{hclust.tree}
\alias{hclust.tree}
\title{Hierarchical Clustering.}
\usage{
hclust.tree(primer.candidates)
}
\arguments{
\item{primer.candidates}{Alignment of primer candidates.}
}
\value{
Phylogeny of the input \code{primer.candidates}.
}
\description{
Performs hierarchical clustering on aligned primer sequences.
}
\details{
The clustering is performed to identify similar groups of primer candidates
that can be merged to form degenerate primers.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/optimization_ILP.R
\name{get.sets.from.decisions}
\alias{get.sets.from.decisions}
\title{Optimal Sets from Decision Variables}
\usage{
get.sets.from.decisions(ILP.df, Tm.sets)
}
\arguments{
\item{ILP.df}{Data frame with ILP optimization results.}

\item{Tm.sets}{List with primer data frames for every target melting temperature.}
}
\value{
A list with optimal primer data sets for every target temperature.
}
\description{
Determines primer sets from decision variables from ILP.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/filters.R
\name{augment.primer.cvg}
\alias{augment.primer.cvg}
\title{Augmentation of Primer Coverage.}
\usage{
augment.primer.cvg(
  primer.df,
  template.df,
  settings,
  partial = FALSE,
  constraint = NULL
)
}
\arguments{
\item{primer.df}{A \code{Primers} object for which
the primer coverage shall be augmented.}

\item{template.df}{A \code{Templates} object.}

\item{settings}{A \code{DesignSettings} object giving
the parameters for coverage computations.}

\item{partial}{Whether all missing primer coverage values should be 
computed. If \code{partial} is \code{TRUE}, only 
the coverage values of the primers that were excluded due to
the specified \code{constraint} are computed.}

\item{constraint}{A character vector specifying the exclusion
reason for which the partial augmentation should take place.}
}
\value{
A \code{Primers} object with augmented coverage entries.
}
\description{
Computes the coverage for the primers in \code{primer.df} 
that is still missing such that the relaxation procedure can adjust
appropriate constraints.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/filters.R
\name{get.cvg.gain}
\alias{get.cvg.gain}
\title{Computation of Coverage Gain.}
\usage{
get.cvg.gain(
  covered.seqs,
  template.df,
  missing.df,
  candidate.df,
  con.names,
  constraint.limits,
  feasible.only = FALSE
)
}
\arguments{
\item{covered.seqs}{List with covered sequences.}

\item{template.df}{A \code{Templates} data frame.}

\item{missing.df}{A \code{Templates} data frame 
containing only the templates that still need to be covered.}

\item{candidate.df}{A \code{Primers} data frame containing
candidate primers.}

\item{con.names}{The constraint to evaluate the coverage gain for
upon being relaxed.}

\item{constraint.limits}{A list with constraint limits.}

\item{feasible.only}{Whether only feasible coverage gains 
are to be outtputed. Here, \emph{feasible} relates to 
coverage gains that can be obtained directly with the next relaxation.}
}
\value{
The number of covered sequences to be gained.
}
\description{
Computes the coverage gain from \code{covered.seqs}.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/settings.R
\name{check_setting}
\alias{check_setting}
\title{Check Setting Validity.}
\usage{
check_setting(known.options, options, mandatory.options = NULL)
}
\arguments{
\item{known.options}{Vector with names and classes of allowed options.}

\item{options}{Active options to be checked.}

\item{mandatory.options}{Fields that have to be present.}
}
\value{
\code{TRUE} if the setting is valid, \code{FALSE} otherwise.
}
\description{
Checks whether the input settings are valid or not.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/templates.R
\name{unify.leaders}
\alias{unify.leaders}
\title{Unification of Leaders}
\usage{
unify.leaders(l.seq.fw, l.seq.rev, lex.seq, gap.char)
}
\arguments{
\item{l.seq.fw}{Data frame with binding information for forward primers.}

\item{l.seq.rev}{Data frame with binding information for reverse primers.}

\item{lex.seq}{Template data frame.}

\item{gap.char}{The character for indicating alignment gaps.}
}
\value{
Template data frame with annotated binding regions.
}
\description{
Unifies individual binding regions for forward and reverse primers.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/openPrimeR.R
\docType{package}
\name{openPrimeR-package}
\alias{openPrimeR}
\alias{openPrimeR-package}
\title{Multiplex PCR Primer Design and Analysis.}
\description{
With openPrimeR you can evaluate existing
primers or design novel primers for multiplex
polymerase chain reaction that are optimized with respect
to the coverage of template sequences and the physicochemical
properties of the primers.
}
\details{
For designing primers, you just need the function \code{\link{design_primers}}
from \pkg{openPrimeR}. As a minimal input, this function requires:
\describe{
\item{A set of template sequences}{You an load a \code{\link{Templates}} object with \code{\link{read_templates}}.}
\item{Settings for primer design}{You can load a \code{\link{DesignSettings}} object from a (supplied) XML file with \code{\link{read_settings}}.
The settings can be easily customized using the setters \code{\link{constraints}}, \code{\link{constraintLimits}}, \code{\link{cvg_constraints}}, \code{\link{conOptions}}, and \code{\link{PCR}}.}
}
For evaluating existing primers you can load a FASTA or CSV file 
containing the primers and templates of of interest using
\code{\link{read_primers}} and \code{\link{read_templates}}, respectively.
After evaluating the properties of the primers
using \code{\link{check_constraints}}, you can interpret the results 
with several functions. For example, you can analyze the coverage of 
the template sequences using \code{\link{get_cvg_stats}}, determine 
the deviation from the target constraints using \code{\link{plot_constraint_deviation}}, or create a comprehensive report on the analyzed
primers using \code{\link{create_report}}.
In order to compare several primer sets with each other,
you can create a table of the 
properties of the primer sets using \code{get_comparison_table}
or create a full report, again using \code{\link{create_report}}.
}
\section{Package options}{

openPrimeR uses the following options:
\describe{
  \item{\code{openPrimeR.constraint_order}}{The identifiers of constraints 
in the order they are applied during the filtering procedure. 
This order is maintained when loading a \code{DesignSettings} object.}
  \item{\code{openPrimeR.relax_order}}{The identifiers of constraints in the order in which they shall be relaxed during the relaxation procedure when designing primers.}
  \item{\code{openPrimeR.plot_abbrev}}{The maximal number of allowed characters for tick labels in plots.}
  \item{\code{openPrimeR.plot_colors}}{A named vector providing the identifiers of RColorBrewer palettes. Each vector entry provides the plotting colors for a specific type of stratification (i.e. by run, constraint, or primer). The palettes should provide at least eight colors.}
}
}

\author{
\strong{Maintainer}: Matthias Döring \email{matthias-doering@gmx.de}

Authors:
\itemize{
  \item Nico Pfeifer \email{pfeifer@informatik.uni-tuebingen.de}
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/settings.R
\name{fix_constraint_boundaries}
\alias{fix_constraint_boundaries}
\title{Correction of Constraint Boundaries.}
\usage{
fix_constraint_boundaries(constraints, constraint.limits, fix.limit = TRUE)
}
\arguments{
\item{constraints}{A list with constraint settings.}

\item{constraint.limits}{A list with constraint limits.}

\item{fix.limit}{Whether the constraint limits should be
adjusted. If \code{FALSE}, the constraint settings are adjusted.}
}
\value{
The corrected constraint limits.
}
\description{
Fixes the constraint boundaries if they are more narrow than the current settings.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/settings.R
\name{check_limit_value}
\alias{check_limit_value}
\title{Check of limit correctness.}
\usage{
check_limit_value(setting, limit)
}
\arguments{
\item{setting}{A single constraint setting.}

\item{limit}{A single constraint limit.}
}
\value{
A vector containing \code{TRUE} if the limit is more general
than the constraint setting and \code{FALSE} otherwise.
}
\description{
Checks whether a constraint limit is more general than the setting.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/settings.R
\name{create.constraint.table}
\alias{create.constraint.table}
\title{Output a Constraint Overview Table}
\usage{
create.constraint.table(
  constraints,
  constraint.limits = NULL,
  constraints.used.fw = NULL,
  constraints.used.rev = NULL,
  format.type = c("backend", "shiny", "report")
)
}
\arguments{
\item{constraints}{List with constraint settings.}

\item{constraint.limits}{List with constraint limits.}

\item{constraints.used.fw}{Constraints used for forward primer design.}

\item{constraints.used.rev}{Constraints used for reverse primer design.}

\item{format.type}{The type of formatting to be performed on entries.}
}
\value{
Data frame with summary of constraints.
}
\description{
Outputs a table showing the values of constraints.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plots_coverage.R
\name{plot_primer_cvg_mismatches}
\alias{plot_primer_cvg_mismatches}
\title{Plot of Individual Primer Coverage and Mismatches.}
\usage{
plot_primer_cvg_mismatches(
  primer.df,
  template.df,
  groups = NULL,
  nfacets = NULL
)
}
\arguments{
\item{primer.df}{A \code{Primers} object.}

\item{template.df}{A \code{Templates} object.}

\item{groups}{Optional identifiers of template groups to be considered.
If not provided, all template groups are considered.}

\item{nfacets}{A numeric providing the number of facet columns to use. 
By default, \code{nfacets} is set to \code{NULL} such that a
suitable number of columns is chosen automatically.}
}
\value{
A bar plot showing the coverage of individual primers for different mismatch settings.
}
\description{
Plots the coverage of individual primers for different mismatch settings.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ambiguity.R
\name{convert.to.iupac}
\alias{convert.to.iupac}
\title{Merge sequences.}
\usage{
convert.to.iupac(seqs)
}
\arguments{
\item{seqs}{Vector of strings.}
}
\value{
Consensus sequence of seqs.
}
\description{
Merges the input sequences to one sequence containing IUPAC ambiguity codes.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/IO.R
\name{sanitize_path}
\alias{sanitize_path}
\title{Sanitiziation of Filename.}
\usage{
sanitize_path(path, suffix = "", sub.char = "_")
}
\arguments{
\item{path}{The path to the file to be sanitized, without file extension.}

\item{suffix}{The suffix (e.g. ".png") of a file.}

\item{sub_char}{The character used to replacing disallowed chars.}
}
\value{
The sanitized filename
}
\description{
Ensures that a filename is valid for the file system.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/settings.R
\name{create.PCR.table}
\alias{create.PCR.table}
\title{Creation of a Table for PCR Conditions}
\usage{
create.PCR.table(other.settings, format.type = c("backend", "shiny", "report"))
}
\arguments{
\item{other.settings}{List with PCR settings.}

\item{format.tyep}{How the table shall be formatted.}
}
\value{
A data frame.
}
\description{
Creation of a Table for PCR Conditions
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/con_dimerization.R
\name{combine.strings}
\alias{combine.strings}
\title{Combination of OligoArrayAux Structure Sequences.}
\usage{
combine.strings(s1, s2)
}
\arguments{
\item{s1}{A character vector to be combined with \code{s2}.}

\item{s2}{A character vector to be included into \code{s1}.}
}
\value{
A character vector.
}
\description{
Combines the input strings.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/con_dimerization.R
\name{select.min.cross.idx}
\alias{select.min.cross.idx}
\title{Selection of cross dimer index}
\usage{
select.min.cross.idx(deltaG, primers)
}
\arguments{
\item{deltaG}{Data frame with thermodynamic info.}

\item{primers}{The corresponding primers.}
}
\value{
The indices with smallest DeltaG for every primer.
}
\description{
Select the index with the smallest DeltaG value.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/templates.R
\name{assign_binding_regions.numeric}
\alias{assign_binding_regions.numeric}
\title{Numeric Assignment of Binding Regions.}
\usage{
\method{assign_binding_regions}{numeric}(
  template.df,
  fw,
  rev,
  optimize.region = FALSE,
  primer.length = 20,
  gap.char = "-"
)
}
\arguments{
\item{template.df}{Template data frame.}

\item{fw}{Binding region data forward primers.}

\item{rev}{Binding region data for reverse primers.}

\item{optimize.region}{Should the primer binding region be optimized using secondary structure prediction?}

\item{primer.length}{Probe length for optimizing template secondary structure.}
}
\value{
The template data frame with assigned binding regions.
}
\description{
Numeric S3 generic case for assigning binding regions.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/con_primer_coverage.R
\name{evaluate.diff.primer.cvg}
\alias{evaluate.diff.primer.cvg}
\title{Evaluation of Coverage.}
\usage{
evaluate.diff.primer.cvg(primers, excluded.seqs, template.df)
}
\arguments{
\item{primers}{Primer data frame.}

\item{excluded.seqs}{Identifiers of templates to be excluded.}

\item{template.df}{Template data frame}
}
\value{
Primer data frame with updated coverage under the exclusion of \code{excluded.seeqs}.
}
\description{
Re-evaluates the coverage of primers under exclusion of certain templates.
}
\details{
This function requires that \code{primers} was already annotated with primer coverage before.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/settings.R
\name{check_names}
\alias{check_names}
\title{Check Setting Names.}
\usage{
check_names(known.options, input.options)
}
\arguments{
\item{known.options}{Allowed setting names.}

\item{input.options}{Input setting names}
}
\value{
Mapping of \code{input.options} to \code{known.options} or \code{NULL} if invalid.
}
\description{
Checks whether the specified settings hvae the correct names.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/optimization_ILP.R
\name{remove.redundant.cols}
\alias{remove.redundant.cols}
\title{Removal of Redundant Primers.}
\usage{
remove.redundant.cols(S, cvg.matrix)
}
\arguments{
\item{S}{Indices of primers that are selected in an optimal solution.}

\item{cvg.matrix}{Binary matrix of coverage events.}
}
\value{
Updated indices of selected primers \code{S} such that indices
representing primers with redundant coverage are removed.
}
\description{
Removes redundant primers from an optimal solution.
}
\details{
An optimal solution can contain primers with redundant coverage when using presolve or greedy optimization.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/primers.R
\name{create.k.mers}
\alias{create.k.mers}
\title{Creation of k-mers for multiple sequences.}
\usage{
create.k.mers(seqs, k)
}
\arguments{
\item{k}{The size of the k-mer.}

\item{seq}{A character vector.}
}
\value{
A list with named character vectors, containing the k-mers.
}
\description{
Creation of k-mers for multiple sequences.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/initialize_primers.R
\name{create.primers.naive}
\alias{create.primers.naive}
\title{Naive Initialization of Primers.}
\usage{
create.primers.naive(
  seqs,
  seq.IDs,
  seq.groups,
  l.s,
  e.s,
  primer.lengths,
  allowed.region.definition,
  max.degen,
  sample = "",
  identifier = "",
  updateProgress = NULL
)
}
\arguments{
\item{seqs}{The template sequence strings.}

\item{seq.IDs}{The identifiers of the templates.}

\item{seq.groups}{The group identifiers of the templates.}

\item{l.s}{The positions where the allowed region starts for each template.}

\item{e.s}{The positions where the allowed reigon ends for each template.}

\item{primer.lengths}{Vector of desired primer lengths.}

\item{allowed.region.definition}{Definition of the allowed region.}

\item{max.degen}{Maximum allowed degeneracy of primers.}

\item{sample}{Template sample identifier.}

\item{updatProgress}{Shiny progress object.}
}
\value{
Data frame with initialized primer candidates.
}
\description{
Initialize primers by extracting substrings from all templates.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/initialize_primers.R
\name{check.init.primer.length.single}
\alias{check.init.primer.length.single}
\title{Primer Length Check.}
\usage{
check.init.primer.length.single(
  allowed,
  allowed.region.definition = c("within", "any"),
  min.len
)
}
\arguments{
\item{allowed}{String containing the allowed binding sequence.}

\item{allowed.region.definition}{Definition of allowed binding regions.}

\item{min.len}{Minimal desired primer lengths.}
}
\value{
TRUE if primers of the desired length can be constructed,
FALSE otherwise.
}
\description{
Checks whether it is possible to construct primers of the desired length.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/constraints_eval.R
\name{update.constraint.values}
\alias{update.constraint.values}
\title{Update of Primer Constraints.}
\usage{
\method{update}{constraint.values}(constraint.df, constraint.values)
}
\arguments{
\item{constraint.df}{Primer data frame.}

\item{constraint.values}{Data frame with computed constraint values.}
}
\value{
A primer data frame with updated columns.
}
\description{
Updates the input primer data frame with the computed constraint values.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/optimization_algo.R
\name{compute.all.primer.subsets.ILP}
\alias{compute.all.primer.subsets.ILP}
\title{Computation of Primer Subsets}
\usage{
compute.all.primer.subsets.ILP(
  primer.df,
  template.df,
  k,
  groups,
  cur.results.loc,
  required.cvg = 1
)
}
\arguments{
\item{primer.df}{Primer data frame.}

\item{template.df}{Template data frame.}

\item{k}{Subset size-increment.}

\item{groups}{Identifiers of template groups in order to limit coverage to
certain groups of template sequences.}

\item{cur.results.loc}{Location for storing the results.}

\item{required.cvg}{The required coverage ratio.}
}
\value{
Write-out of results.
}
\description{
Computes all optimal primer subsets and stores their plots.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plots_constraints.R
\name{plot_constraint.histogram.nbr.mismatches}
\alias{plot_constraint.histogram.nbr.mismatches}
\title{Histogram of Number of Mismatches.}
\usage{
plot_constraint.histogram.nbr.mismatches(primer.df, allowed.mismatches)
}
\arguments{
\item{primer.df}{Primer data frame.}

\item{allowed.mismatches}{Number of allowed mismatches.}
}
\value{
A plot of the number of primer mismatches.
}
\description{
Plots a histogram of mismatches.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/con_primer_coverage.R
\name{get.primer.binding.idx}
\alias{get.primer.binding.idx}
\title{Retrieval of Allowed Binding Indices.}
\usage{
get.primer.binding.idx(
  binding,
  primer.type = c("fw", "rev", "both"),
  x,
  allowed.other.binding.ratio
)
}
\arguments{
\item{binding}{IRanges binding information.}

\item{primer.type}{Direction of primer.}

\item{x}{Index of primer in the primer data frame.}

\item{allowed.other.binding.ratio}{The ratio of allowed off-target binding events.}
}
\value{
Indices in binding for primer with index code{x} that are allowed.
}
\description{
Retrieves the indices of allowed binding events in \code{binding}
for the primer with index \code{x} and type \code{primer.type}.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/con_primer_coverage.R
\name{covered.primers.to.ID.string}
\alias{covered.primers.to.ID.string}
\title{Conversion of Primer Indices to ID string}
\usage{
covered.primers.to.ID.string(covered.primers, primer.df)
}
\arguments{
\item{covered.primers}{Identifiers of primers covering sequences.}

\item{primer.df}{Primer data frame.}
}
\value{
String containing the covered template IDs.

A string containing the IDs of covering primers.
}
\description{
Converts the input coverage indices to a comma-separated string with the template IDs.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/IO.R
\name{my.read.fasta}
\alias{my.read.fasta}
\title{Read FASTA File.}
\usage{
my.read.fasta(fasta.file, NTs)
}
\arguments{
\item{fasta.file}{The path to a FASTA file.}

\item{NTs}{The allowed set of nucleotides.}
}
\value{
List with vectors of chars.
}
\description{
Reads the input FASTA file.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/check_stop_codons.R
\name{get.ORFs}
\alias{get.ORFs}
\title{Identification of ORFs.}
\usage{
get.ORFs(template.df)
}
\arguments{
\item{template.df}{template data frame.}
}
\value{
Returns a data frame containing the shift of the ORF (either 0,1, or 2) 
for every sequence, as well as a comment in case of problems.
}
\description{
Given a template data frame, identify the exon reading frames in the sequences.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/con_primer_coverage.R
\name{combine.binding.events}
\alias{combine.binding.events}
\title{Combination of Binding Events.}
\usage{
combine.binding.events(my.binding.fw, my.binding.rev, fw.m, rev.m)
}
\arguments{
\item{my.binding.fw}{Forward binding events of individual primers.}

\item{my.binding.rev}{Reverse binding events of individual primers.}

\item{fw.m}{Forward binding events of paired primers.}

\item{rev.m}{Reverse binding events of paired primers.}
}
\value{
IRanges of all binding events.
}
\description{
Appends all binding events.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/templates.R
\name{build_leader_df}
\alias{build_leader_df}
\title{Building of Leader Data Frame.}
\usage{
build_leader_df(
  direction = c("fw", "rev"),
  leader,
  start,
  end,
  ali.start,
  ali.end
)
}
\arguments{
\item{direction}{The primer direction for which we are annotating binding regions.}

\item{leader}{The binding region sequence.}

\item{start}{The start positions of the binding region.}

\item{end}{The end positions of the binding region.}

\item{ali.start}{The start positions of the binding region in the aligned input.}

\item{ali.end}{The end positions of the binding region in the aligned input.}
}
\value{
A data frame with binding region information.
}
\description{
Constructs the leader data frame.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/errors.R
\name{my.error}
\alias{my.error}
\title{Custom Error}
\usage{
my.error(subclass, message, call = sys.call(-1), ...)
}
\arguments{
\item{subclass}{String giving the specific type of error.}

\item{message}{Message to be displayed to the user.}

\item{call}{Environment where the error ocurred.}

\item{...}{Other arguments to be passed to  the condition function.}
}
\value{
Generates a custom error.
}
\description{
Creates an error with a custom class.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/con_melting_temperature.R
\name{get.delta.G}
\alias{get.delta.G}
\title{Change in Free Energy.}
\usage{
get.delta.G(delta.H, delta.S, temp = 37)
}
\arguments{
\item{delta.H}{Change in enthalpy in cal/mol.}

\item{delta.S}{Change in entropy cal/mol*K.}

\item{temp}{Temperature in Celsius for which to compute free energy change.}
}
\value{
The change in free energy in kcal/mol.
}
\description{
Computes the change in free energy.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/con_primer_coverage.R
\name{covered.seqs.to.idx}
\alias{covered.seqs.to.idx}
\title{Conversion of Coverage Strings to Indices.}
\usage{
covered.seqs.to.idx(covered.seqs, template.df)
}
\arguments{
\item{covered.seqs}{Strings of covered sequences to be converted.}

\item{template.df}{Template data frame containing the identifiers of templates}
}
\value{
Indices of covered templates.
}
\description{
Converts the input coverage strings (comma separated template Identifiers) into indices.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/primers.R
\name{estimate.cvg}
\alias{estimate.cvg}
\title{Estimation of Primer Coverage.}
\usage{
estimate.cvg(lex.df, k = 18, mode.directionality, sample = "")
}
\arguments{
\item{k}{A numeric giving the size of the primers.}

\item{mode.directionality}{Estimation of coverage for forward/reverse/both?}

\item{sample}{An optional identifier for the sample.}

\item{seqs}{A character vector of sequences to evaluate coverage for.}
}
\value{
A list with entries \code{fw} and \code{rev} giving
data frames for forward/reverse binding.
}
\description{
Estimates the possible coverage of primers using
probes of size \code{k} and only considering perfect
matches without consideration of ambiguities.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/constraints_eval.R
\name{check_constraints_comparison}
\alias{check_constraints_comparison}
\title{Batch Procedure for Evaluating Primer Sets.}
\usage{
check_constraints_comparison(
  primer.data,
  template.data,
  settings,
  active.constraints = names(constraints(settings)),
  to.compute.constraints = active.constraints,
  for.shiny = FALSE,
  updateProgress = NULL
)
}
\arguments{
\item{primer.data}{A list of objects of class \code{Primers}.}

\item{template.data}{A list of objects of class \code{Templates}
corresponding to \code{primer.data}.}

\item{settings}{An object of class \code{DesignSettings}.}

\item{active.constraints}{A character vector providing identifiers
of constraints to be considered.}

\item{to.compute.constraints}{A character vector providing identifiers
of constraints to be computed.}

\item{for.shiny}{A logical indicating whether the results are
indicated for the Shiny app or not.}

\item{updateProgress}{A callback function to track progress in the Shiny app.}
}
\value{
A list with objects of class \code{Primers}.
}
\description{
Batch Procedure for Evaluating Primer Sets.
}
\examples{
\dontrun{
data(Comparison)
eval.data <- check_constraints_comparison(primer.data[1:2], template.data[1:2], settings)
}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/con_melting_temperature.R
\name{joule.to.cal}
\alias{joule.to.cal}
\title{Conversion from J to cal}
\usage{
joule.to.cal(val.J)
}
\arguments{
\item{val.J}{Numeric Joule value.}
}
\value{
The value correspdoning to \code{val.J} in calories.
}
\description{
Converts the input from Joule to calories.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/con_primer_coverage.R
\name{merge.binding.information}
\alias{merge.binding.information}
\title{Merge of Forward/Reverse Binding Information.}
\usage{
\method{merge}{binding.information}(
  primers,
  fw.binding.filtered,
  rev.binding.filtered,
  mode.directionality = c("fw", "rev", "both"),
  idx.fw,
  idx.rev
)
}
\arguments{
\item{primers}{The primer data frame.}

\item{mode.directionality}{Primer directionality.}

\item{idx.fw}{Index of fw primers.}

\item{idx.rev}{Index of rev primers.}

\item{fw.binding}{IRanges object with binding events of fw primers.}

\item{rev.binding}{IRanges object with binding events of rev primers.}
}
\value{
IRanges with correct binding events.
}
\description{
Determines binding events of individual and pairs of primers.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/primers.R
\name{read_primers_csv}
\alias{read_primers_csv}
\title{Read Primer CSV File.}
\usage{
read_primers_csv(file)
}
\arguments{
\item{file}{The path to a csv file containing the primer data.}
}
\value{
A \code{Primers} object, an instance of a data frame.
}
\description{
Reads a primer data frame stored in a CSV file.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/IO_view.R
\name{view.primers.report}
\alias{view.primers.report}
\title{View the Evaluated Primers in the Report.}
\usage{
view.primers.report(primer.df, template.df)
}
\arguments{
\item{primer.df}{A \code{Primers} object.}

\item{template.df}{A \code{Templates} object.}
}
\value{
A formatted primer table.
}
\description{
Creates a formatted primers table for the report PDF.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/check_stop_codons.R
\name{compute.mismatch.table}
\alias{compute.mismatch.table}
\title{Mismatch overview table}
\usage{
compute.mismatch.table(
  primer.df,
  template.df,
  mode.directionality = c("fw", "rev")
)
}
\arguments{
\item{primer.df}{Primer data frame.}

\item{template.df}{Template data.}

\item{mode.directionality}{Direction of primers.}
}
\value{
: A data frame summarizing all mismatches of the input primers with the input templates.
}
\description{
Computes a table summarizing all of the mismatches caused by the primers in the
input data frame.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/settings.R
\name{get.static.tool.info}
\alias{get.static.tool.info}
\title{Retrieval of Tool Information.}
\usage{
get.static.tool.info()
}
\value{
A data frame with information about the required tools.
}
\description{
Constructs a data frame containing information about the tools.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/SettingsDoc.R, R/constraints.R, R/settings.R
\docType{class}
\name{Settings}
\alias{Settings}
\alias{ConstraintSettings}
\alias{ConstraintSettings-class}
\alias{CoverageConstraints}
\alias{CoverageConstraints-class}
\alias{ConstraintOptions}
\alias{ConstraintOptions-class}
\alias{PCR_Conditions}
\alias{PCR_Conditions-class}
\alias{DesignSettings}
\alias{DesignSettings-class}
\alias{constraints}
\alias{constraints,DesignSettings-method}
\alias{constraints,AbstractConstraintSettings-method}
\alias{cvg_constraints}
\alias{cvg_constraints,DesignSettings-method}
\alias{PCR}
\alias{PCR,DesignSettings-method}
\alias{conOptions}
\alias{conOptions,DesignSettings-method}
\alias{constraintLimits}
\alias{constraintLimits,DesignSettings-method}
\alias{constraints<-}
\alias{constraints<-,DesignSettings,list-method}
\alias{constraints<-,AbstractConstraintSettings,list-method}
\alias{cvg_constraints<-}
\alias{cvg_constraints<-,DesignSettings-method}
\alias{constraintLimits<-}
\alias{constraintLimits<-,DesignSettings-method}
\alias{PCR<-}
\alias{PCR<-,DesignSettings-method}
\alias{conOptions<-}
\alias{conOptions<-,DesignSettings-method}
\alias{parallel_setup}
\title{Settings Functionalities.}
\usage{
constraints(x)

\S4method{constraints}{DesignSettings}(x)

\S4method{constraints}{AbstractConstraintSettings}(x)

cvg_constraints(x)

\S4method{cvg_constraints}{DesignSettings}(x)

PCR(x)

\S4method{PCR}{DesignSettings}(x)

conOptions(x)

\S4method{conOptions}{DesignSettings}(x)

constraintLimits(x)

\S4method{constraintLimits}{DesignSettings}(x)

constraints(x) <- value

\S4method{constraints}{DesignSettings,list}(x) <- value

\S4method{constraints}{AbstractConstraintSettings,list}(x) <- value

cvg_constraints(x) <- value

\S4method{cvg_constraints}{DesignSettings}(x) <- value

constraintLimits(x) <- value

\S4method{constraintLimits}{DesignSettings}(x) <- value

PCR(x) <- value

\S4method{PCR}{DesignSettings}(x) <- value

conOptions(x) <- value

\S4method{conOptions}{DesignSettings}(x) <- value

parallel_setup(cores = NULL)
}
\arguments{
\item{x}{A \code{DesignSettings} object.}

\item{value}{An object to be used in one of the setters.
For \code{constraints<-} and \code{constraintLimits<-}, a list with constraint settings or boundaries. Each list entry
should have a permissible name and consist of at most two
values providing the minimal and/or maximal allowed values, which
have to be denominated via \code{min} and \code{max}.

For \code{conOptions<-}, a list with constraint options. The permissible
fields of the list and their types are documented in the 
\code{\link{ConstraintOptions}} class.

For \code{cvg_constraints<-}, a list with coverage constraints. Each
list entry must have a permissible name and
contain a numeric vector with at most two components
describing the minimal and/or maximal required values that
are to be indicated via \code{min} and \code{max}.
The permissible contraint identifiers are documented in the
\code{\link{CoverageConstraints}} class.

For \code{PCR<-}, a named list providing PCR conditions 
The permissible fields of the list and their types
are documented in the \code{\link{PCR_Conditions}} class.}

\item{cores}{A numeric providing the number of cores to use. The default is \code{NULL}
such that half the number of available cores are used.}
}
\value{
The \code{ConstraintSettings} constructor defines a new
\code{ConstraintSettings} object.

The \code{CoverageConstraints} constructor initializes a new
\code{CoverageConstraints} object.

The \code{ConstraintOptions} constructor returns
a new \code{ConstraintOptions} object.

The \code{PCR_Conditions} constructor defines a new \code{PCR_Conditions} object.

The \code{DesignSettings} constructor defines a \code{DesignSettings} object.

\code{constraints} gets a list with the active constraint settings.

\code{cvg_constraints} returns the list of active coverage constraints.

\code{PCR} gets the list of PCR conditions defined in the
provided \code{DesignSettings} object.

\code{conOptions} returns a list with constraint options.

\code{constraintLimits} gets the list of constraint limits.

\code{constraints<-} sets the list of constraints in a \code{DesignSettings} object.

\code{cvg_constraints<-} sets the list of coverage constraints in the provided \code{DesignSettings} object.

\code{constraintLimits<-} sets the list of constraint limits in 
the provided \code{DesignSettings} object.

\code{PCR<-} sets the constraint options in the provided
\code{DesignSettings} object.

\code{conOptions<-} sets the specified list of constraint options in the provided \code{DesignSettings} object.

\code{parallel_setup} returns \code{NULL}.
}
\description{
\describe{
\item{\code{DesignSettings}}{The \code{DesignSettings} class encapsulates all settings 
for designing and evaluating primer sets. 
Upon loading an XML file, the \code{DesignSettings} class checks whether
the defined constraints can be applied by identifying whether 
the requirements for external programs are fulfilled. 
If the requirements are not fulfilled, the affected constraints 
are removed from the loaded \code{DesignSettings} object
and a warning is issued.
The loaded constraints are automatically ordered according to
the option \code{openPrimeR.constraint_order} such that 
the runtime of the \code{\link{design_primers}} and \code{\link{filter_primers}}
functions is optimized.}
\item{\code{constraints}}{Gets the active constraints of the provided
\code{DesignSettings} object.}
\item{\code{constraints<-}}{Sets the active constraints of the provided
\code{DesignSettings} object.}
\item{\code{cvg_constraints}}{Gets the coverage constraints of the provided
\code{DesignSettings} object.}
\item{\code{cvg_constraints<-}}{Sets the coverage constraints of the provided
\code{DesignSettings} object.}
\item{\code{conOptions}}{Gets the constraint settings of the provided 
\code{DesignSettings} object.}
\item{\code{conOptions<-}}{Sets the constraint settings of the provided
\code{DesignSettings} object.}
\item{\code{constraintLimits}}{Gets the constraint limits that are defined in the provided 
\code{DesignSettings} object.}
\item{\code{constraintLimits<-}}{Sets the constraint limits of the provided 
\code{DesignSettings} object.}
\item{\code{PCR}}{Gets the PCR conditions that are defined in the provided
\code{DesignSettings} object.}
\item{\code{PCR<-}}{Sets the PCR conditions that are defined in the provided
\code{DesignSettings} object.}
\item{\code{ConstraintSettings}}{The \code{ConstraintSettings} class encapsulates the constraints
on the physicochemical properties of primers.}
\item{\code{CoverageConstraints}}{The \code{CoverageConstraints} class encapsulates the conditions
under which the coverage of primers is evaluated.}
\item{\code{PCR_Conditions}}{The \code{PCR_Conditions} class encapsulates the PCR conditions
for the computation of primer properties.}
\item{\code{ConstraintOptions}}{The \code{ConstraintOptions} class encapsulates the options
for constraint computations.}
\item{\code{parallel_setup}}{Registers the specified number of cores
with the parallel backend.}
}
}
\details{
Note that for the \code{DesignSettings} class, the fields \code{Input_Constraints}, \code{Input_Constraint_Boundaries}, and \code{Coverage_Constraints} should 
contain entries with at most two components using the fields \code{min} and/or \code{max}.
The \code{Input_Constraint_Boundaries} should always be at least as general as the
specified \code{Input_Constraints}.

For an overview of permissible constraints,
please consider the \code{\link{ConstraintSettings}} documentation.
}
\section{Slots}{

\describe{
\item{\code{Input_Constraints}}{A \code{\link{ConstraintSettings}} object specifying the 
desired target value ranges for primer properties.}

\item{\code{Input_Constraint_Boundaries}}{A \code{\link{ConstraintSettings}} object specifying
the limits for relaxing the constraints during the primer design procedure.
This slot may contain the same fields as the \code{Input_Constraints} slot,
but the specified desired ranges should be at least as general as
those specified in the \code{Input_Constraints} slot.}

\item{\code{Coverage_Constraints}}{A \code{\link{CoverageConstraints}} object specifying
the constraints for computing the primer coverage.}

\item{\code{PCR_conditions}}{A \code{\link{PCR_Conditions}} object specifying the PCR-related settings.}

\item{\code{constraint_settings}}{A \code{\link{ConstraintSettings}} object providing options 
for the computation of individual physicochemical properties.}

\item{\code{status}}{Named boolean vector indicating 
which of the possible constraints are active (\code{TRUE})
and which are not (\code{FALSE}).}

\item{\code{settings}}{For \code{ConstraintSettings}, a named list containing the settings for
the active constraints.
The list may contain the following fields:
\describe{
\item{primer_coverage:}{The required number of covered template sequences per primer.}
\item{primer_specificity:}{The required required specificity of 
                           primers in terms of a ratio in the interval [0,1].}
\item{primer_length:}{The required lengths of primer sequences.}
\item{gc_clamp:}{The desired number of GCs at primer 3' termini.}
\item{gc_ratio:}{The desired ratio of GCs in primers 
                 in terms of numbers in the interval [0,1].}
\item{no_runs:}{The accepted length homopolymer runs in a primer.}
\item{no_repeats:}{The accepted length of dinucleotide repeats in a primer.}
\item{self_dimerization:}{The lowest acceptable free energy [kcal/mol] for the 
                          interaction of a primer with itself. The identification
                           of self dimers requires the software \emph{OligoArrayAux} (see notes).}
\item{melting_temp_range:}{The desired melting temperature (Celsius) of primers.
                          The accurate computation of melting temperatures requires the software \emph{MELTING} (see notes).}
\item{melting_temp_diff:}{The maximal allowed difference between the melting temperatures (Celsius)
                          of primers contained in the same set. The accurate computation of melting temperatures
                          requires the software \emph{MELTING} (see notes).}
\item{cross_dimerization:}{The lowest acceptable free energy [kcal/mol] for the
                           interaction of a primer with another primer. The identification
                           of cross dimers requires the software \emph{OligoArrayAux} (see notes).}
\item{secondary_structure:}{The lowest acceptable free energy [kcal/mol] for the
                            formation of primer secondary structures. Secondary structures are determined
                            using the software \emph{ViennaRNA} (see notes).}
}

For \code{PCR_Conditions}, a named list with PCR conditions.
The following fields are possible: 
\describe{
\item{\code{use_taq_polymerase}:}{A logical identifying whether you are performing PCR with a Taq polymerase (\code{TRUE}) or not (\code{FALSE}).}
\item{\code{annealing_temp}:}{The annealing temperature in Celsius that is to be used for evaluating the
constraints defined in the \code{\link{ConstraintSettings}} object.
If the annealing temperature field
is not provided, a suitable annealing temperature is automatically computed using a rule of thumb (i.e. subtracting 5 from the melting temperature).}
\item{\code{Na_concentration}:}{The molar concentration of monovalent sodium ions.}
\item{\code{Mg_concentration}:}{The molar concentration of divalent magnesium ions.}
\item{\code{K_concentration}:}{The molar concentration of monovalent potassium ions.}
\item{\code{Tris_concentration}:}{The molar concentration of the Tris(hydroxymethyl)-aminomethan buffer.}
\item{\code{primer_concentration}:}{The molar concentration of the PCR primers.}
\item{\code{template_concentration}:}{The molar concentration of the PCR templates.}
}

For \code{CoverageConstraints}, a named list with constraint options. Each 
list entry should have an entry \code{min} and/or \code{max}
in order to indicate the minimal and maximal allowed values,
respectively. 
The following identifiers can be used as coverage constraints:
\describe{
\item{\code{primer_efficiency}:}{The desired efficiencies of primer-template amplification events 
in order to be considered as \emph{covered}. \code{primer_efficiency} provides a value in the interval [0,1],
which is based on \pkg{DECIPHER}'s thermodynamic model, which considers the impact of 3' terminal mismatches.}
\item{\code{annealing_DeltaG}:}{The desired free energies of annealing for putative
coverage events between primers and templates. Typically, one would 
limit the maximally allowed free energy.}
\item{\code{stop_codon}:}{Whether coverage events introducing
stop codons into the amplicons should be allowed or discarded. 
Here, a value of \code{1} indicates coverage events that induce stop codons.
As such, setting both minimum and maximium to zero will disregard
coverage events inducing stop codons, while setting the minimum to zero
and the maximum to 1 will allow coverage events that induce stop codons.}
\item{\code{substitution}:}{Whether coverage events introducing
substitutions into the amino acid sequence are considered or discarded.
The same encoding as for \code{stop_codon} is used, that is,
the value \code{1} indicates coverage events
inducing substitutions. Hence, to prevent substitutions,
the maximal value of \code{substitution} can be set to zero.}
\item{\code{terminal_mismatch_pos}:}{The position relative to 
the primer 3' terminal end for which mismatch binding events should be allowed,
where the last base in a primer is indicated by position \code{1}.
For example, setting the minimal value of \code{terminal_mismatch_pos} 
to \code{7} means that only coverage events that do not have a terminal mismatch
within the last 6 bases of the primer are allowed.}
\item{\code{coverage_model}:}{Use a logistic regression model combining the free energy of annealing and 3' terminal mismatch positions
to determine the expected rate of false positive coverage calls. 
Using \code{coverage_model}, you can specify the allowed ratio of falsely predicted coverage events.
Typically, one would limit the maximal allowed rate of false positives. Note that setting a
small false positive rate will reduce the sensitivity of the coverage calls (i.e. true positives will be missed).}
}

For \code{ConstraintOptions}, a named list with constraint options.
The following fields are permissible:
\describe{
\item{allowed_mismatches:}{The maximal number of allowed mismatches between
a primer and a template sequence. If the number of mismatches of a primer
with a template exceeds the specified value, the primer is not considered
to cover the corresponding template when the coverage is being computed.}
\item{allowed_other_binding_ratio:}{Ratio of allowed binding events
outside the target binding ratio. This value should be in the interval
[0,1]. If the specified value is greater than zero, all coverage events
outside the primer binding region are reported. If, however, the
identified ratio of off-target events should exceed the allowed ratio,
a warning is issued. If \code{allowed_other_binding_ratio} is set to \code{0},
only on-target primer binding events are reported.
The setting of \code{allowed_other_binding_ratio} is ignored when designing primers, 
which always uses a value of 0.}
\item{allowed_region_definition:}{The definition of the target
binding regions that is used for evaluating the coverage.
In case that \code{allowed_region_definition} is \code{within}, primers have to lie within the allowed binding region.
If \code{allowed_region_definition} is \code{any}, primers only have to overlap with the target binding region.}
\item{hexamer_coverage:}{If \code{hexamer_coverage} is set to "active", primers whose 3' hexamer (the last 6 bases) is fully complementary to the corresponding
template region are automatically considered to cover the template. 
If \code{hexamer_coverage} is set to \code{inactive}, 
hexamer complementarity does not guarantee template coverage.}
}}
}}

\note{
The following external programs are required for constraint computations:
\describe{
\item{MELTING (http://www.ebi.ac.uk/biomodels/tools/melting/):}{Thermodynamic computations (optional) for determining melting temperatures for the constraints \code{melting_temp_diff} and \code{melting_temp_range}}
\item{OligoArrayAux (http://unafold.rna.albany.edu/OligoArrayAux.php):}{Thermodynamic computations used for computing \code{self_dimerization} and \code{cross_dimerization}.
Also required for computing \code{primer_coverage} when a constraint based on the free energy of annealing is active.}
\item{ViennaRNA (http://www.tbi.univie.ac.at/RNA/):}{Secondary structure predictions used for the constraint \code{secondary_structure}}
}

The following external programs are required for computing the coverage constraints:
\describe{
\item{OligoArrayAux (http://unafold.rna.albany.edu/OligoArrayAux.php):}{Thermodynamic computations used for 
computing the coverage constraints \code{annealing_DeltaG}, \code{primer_efficiency}, and \code{coverage_model}}
}
}
\section{primer_coverage}{

Computing the primer coverage involves identifying
which templates are expected to be amplified (covered) by which primers.
The \code{primer_coverage} constraint
determines the minimal and maximal number of coverage events per primer
that are required. The computation of primer coverage is governed
by the coverage constraints postulated via \code{\link{CoverageConstraints}}
and the constraint options defined via \code{\link{ConstraintOptions}}.
}

\section{primer_specificity}{

Primer specificity is automatically determined during the primer coverage
computations but the constraint is only checked when the \code{primer_specificity}
field is available. The specificity of a primer is defined as its ratio of
on-target vs total coverage events (including off-target coverage). Low-specificity
primers should be excluded as they may not amplify the target region effectively.
}

\section{primer_length}{

The length of a primer is defined by its number of bases. Typical
primers have lengths between 18 and 22. Longer primers may guarantee higher
specificities.
}

\section{gc_clamp}{

The GC clamp refers to the presence of GCs at the 3' end of a primer.
For the \code{gc_clamp} constraint, we consider the number of 3' terminal GCs.
For example, the primer \emph{actgaaatttcaccg} has a GC clamp of length 3.
The presence of a GC clamp is supposed to aid the stability of the
polymerase complex. At the same time, long GC clamps should be avoided.
}

\section{no_runs}{

Homopolymer runs (e.g. the primer \emph{aaaaa} has a run of 5 A's)
may lead to secondary structure formation and unspecific binding
and should therefore be avoided.
}

\section{no_repeats}{

Dinucleotide repeats (e.g. the primer \emph{tatata} has 3 TA repeats)
should be avoided for the same reason a long homopolymer runs.
}

\section{self_dimerization}{

Self dimerization refers to a primer that binds to itself rather than
to one of the templates. Primers exhibiting self dimers should
be avoided as they may prevent the primer from amplifying the templates.
Therefore primers with small free energies of dimerization should 
be avoided.
}

\section{melting_temp_range}{

The melting temperature is the temperature at which 50% of the primers
are in duplex with templates and 50% are still present as random coils.
Hence, primers exhibiting high melting temperatures have high affinities
to the templates, while primers with small melting temperatures
have small affinities. The melting temperatures of the primers
determine the annealing temperature of the PCR, which is why
the melting temperatures of the primers should not deviate too much (see
\code{melting_temp_diff}).
}

\section{melting_temp_diff}{

The differences between the melting temperatures of primers in a
set of primers should not deviate too much as the annealing temperaturte
of a PCR should be based on the smallest melting temperature
of a primer in the set. If there are other primers in the set 
exhibiting considerably higher melting temperatures,
these primers may bind inspecifically due to the low annealing temperature.
}

\section{cross_dimerization}{

When two different primers bind to each each other rather than
to the templates, this is called cross dimerization.
Cross dimerization should be prevent at all costs because
such primers cannot effectively amplify their target templates.
Cross dimerizing primers can be excluding primers
exhibiting small free energies of cross dimerization.
}

\section{secondary_structure}{

When a primer exhibits secondary structure, this may prevent it from
binding to the templates. To prevent this, 
primers with low free energies of secondary structure formation
can be excluded.
}

\examples{

# Initializing a new 'ConstraintSettings' object:
constraint.settings <- new("ConstraintSettings")
# Retrieving the constraint settings from a 'DesignSettings' object:
data(Ippolito) # loads a 'DesignSettings' object into 'settings'
constraints(settings)
# Modifying the constraint settings:
constraints(settings)$no_runs["max"] <- 10
constraints(settings) <- constraints(settings)[names(constraints(settings)) != "gc_clamp"]

# Initialize a new 'CoverageConstraints' object:
cvg.constraints <- new("CoverageConstraints")
# Retrieving the coverage constraints from a 'DesignSettings' object:
data(Ippolito) # loads a 'DesignSettings' object into 'settings'
cvg_constraints(settings)
# Modifying the coverage constraints
cvg_constraints(settings)$primer_efficiency["min"] <- 0.001

# Initialize a new 'ConstraintOptions' object:
constraint.options <- new("ConstraintOptions")
# Retrieve the constraint options from a 'DesignSettings' object:
data(Ippolito) # loads a 'DesignSettings' object into 'settings'
conOptions(settings)
# Prevent off-target binding:
conOptions(settings)$allowed_other_binding_ratio <- 0

# Initialize a new 'PCR_Conditions' object:
PCR.conditions <- new("PCR_Conditions")
# Retrieving the PCR conditions from a 'DesignSettings' object:
data(Ippolito) # loads a 'DesignSettings' object into 'settings'
PCR(settings)
# Modifying the PCR conditions:
PCR(settings)$use_taq_polymerase <- FALSE

# Load a settings object
filename <- system.file("extdata", "settings", 
                 "C_Taq_PCR_high_stringency.xml", package = "openPrimeR")
settings <- read_settings(filename)
# Modify the constraints
constraints(settings)$gc_clamp["min"] <- 0
# Modify the constraint limits for designing primers
constraintLimits(settings)$gc_clamp["max"] <- 6
# Modify the coverage constraints
cvg_constraints(settings)$primer_efficiency["min"] <- 0.001
# Modify the PCR conditions
PCR(settings)$Na_concentration <- 0.0001
# Modify the constraint options
conOptions(settings)$allowed_mismatches <- 0

# Load some settings
data(Ippolito)
# View the active constraints
constraints(settings)
# Require a minimal GC clamp extent of 0
constraints(settings)$gc_clamp["min"] <- 0
# View available constraints
settings

# Load some settings
data(Ippolito)
# View all active coverage constraints
cvg_constraints(settings)
# Increase the maximal false positive rate to increase the sensitiviity of coverage predictions
cvg_constraints(settings)$coverage_model <- c("max" = 0.1) 
# View available coverage constraints:
settings

# Load some settings
data(Ippolito)
# View the active constraint limits
constraintLimits(settings)
# Extend the GC relaxation limit
constraintLimits(settings)$gc_clamp <- c("min" = 0, "max" = 6)
# View available constraints
settings

# Load some settings
data(Ippolito)
# View the active PCR conditions
PCR(settings)
# Evaluate primers with a fixed annealing temperature
PCR(settings)$annealing_temperature <- 50 # celsius
# View available PCR conditions
settings

# Load some settings
data(Ippolito)
# View the active constraint options
conOptions(settings)
# Prevent mismatch binding events
conOptions(settings)$allowed_mismatches <- 0
# View available constraint options
settings
# Use two cores for parallel processing:
parallel_setup(2)
}
\seealso{
\code{\link{read_settings}} for reading settings from XML files,
\code{\link{write_settings}} for storing settings as XML files,
\code{\link{constraints}} for accessing constraints,
\code{\link{constraintLimits}} for accessing constraint boundaries,
\code{\link{cvg_constraints}} for accessing coverage constraints,
\code{\link{conOptions}} for accessing constraint options,
\code{\link{PCR}} for accessing the PCR conditions.
}
\concept{settings functions}
\keyword{Settings}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/optimization_ILP.R
\name{select.best.ILP}
\alias{select.best.ILP}
\title{Selection of Best ILP}
\usage{
select.best.ILP(ILP.df)
}
\arguments{
\item{ILP.df}{Data frame with ILP result properties.}
}
\value{
The index of the best solution.
}
\description{
Selects the best solution from multiple solved ILP instances.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/con_template_secondary_structures.R
\name{compute.template.secondary.structures}
\alias{compute.template.secondary.structures}
\title{Template Secondary Structures}
\usage{
compute.template.secondary.structures(
  template.df,
  annealing.temperature,
  regions = NULL,
  constraints = NULL
)
}
\arguments{
\item{template.df}{Template data frame.}

\item{annealing.temperature}{Temperature [C] at which 
to compute secondary structures.}

\item{regions}{List containing the positional intervals for which the template
secondary structure should be computed.}

\item{constraints}{String giving secondary structure constraints.
For example \code{xxxxxx...} would forbid folding in the first 6 bases
of a template with length 9 and allow folding in its last 3 bases.}
}
\value{
Data frame with info on template secondary structures.
}
\description{
Computes template secondary structures.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/con_repeats.R
\name{nbr.of.repeats}
\alias{nbr.of.repeats}
\title{Number of Repeats}
\usage{
nbr.of.repeats(x)
}
\arguments{
\item{x}{Input sequence strings.}
}
\value{
The maximal number of dinucleotide repeats in \code{x}.
}
\description{
Computes the number of dinucleotide repeats in the input sequences.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/IO_view.R
\name{modify.col.rep}
\alias{modify.col.rep}
\title{Modification of Column Names.}
\usage{
modify.col.rep(template.df, for.shiny = TRUE)
}
\arguments{
\item{template.df}{The data frame whose column names are to be modified.}

\item{for.shiny}{Whether formatting should be for shiny.}
}
\value{
\code{template.df} with modified column names.
}
\description{
Modifies column names for frontend output.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plots_filtering.R
\name{stats_plot_data}
\alias{stats_plot_data}
\title{Combination of Filtering Stats.}
\usage{
stats_plot_data(stats, stats.relax)
}
\arguments{
\item{stats}{Statistics of the filtering procedure.}

\item{stats.relax}{Statistic of the relaxation procedure.}
}
\value{
A data frame combinin filtering/relaxation stats.
}
\description{
Summarizes filtering/relaxation statistics for plotting.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/templates.R
\name{update.binding.regions}
\alias{update.binding.regions}
\title{Update of Binding Regions.}
\usage{
\method{update}{binding.regions}(template.df, opti.regions)
}
\arguments{
\item{template.df}{An object of class \code{Templates}.}

\item{opti.regions}{List with new binding intervals.
The list can contain the components \code{fw} and \code{rev} providing
numeric vectors of length 2 providing the start and end
of the binding regions in the templates, for forward
and reverse binding regions, respectively.}
}
\value{
A \code{Templates} object with updated binding regions.
}
\description{
Updates the binding regions in the templates by providing
new intervals for forward and reverse binding regions.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/helper_functions.R
\name{get.unlist.idx}
\alias{get.unlist.idx}
\title{Index for Unlisting.}
\usage{
get.unlist.idx(primer.start, primer.data.idx)
}
\arguments{
\item{primer.start}{Numeric vector.}

\item{primer.data.idx}{Selection indices.}
}
\value{
Indices.
}
\description{
Determines indices for unlisting.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/con_primer_coverage.R
\name{compute.basic.details}
\alias{compute.basic.details}
\title{Computation of Coverage Details}
\usage{
compute.basic.details(
  binding,
  mode = c("on_target", "off_target"),
  template.df,
  primers,
  mode.directionality = c("fw", "rev", "both"),
  allowed.mismatches,
  allowed.other.binding.ratio,
  allowed.region.definition = c("within", "any"),
  updateProgress = NULL
)
}
\arguments{
\item{binding}{An \code{IRanges} object with primer binding information.}

\item{mode}{Either \code{on_target} for on-target binding or \code{off_target} for off-target binding.}

\item{template.df}{Template data frame.}

\item{primers}{Primer data frame.}

\item{mode.directionality}{Primer directionality.}

\item{allowed.mismatches}{The number of allowed mismatches per binding event.}

\item{allowed.other.binding.ratio}{Ratio of primers that are allowed to bind to non-allowed regions.
If \code{allowed.other.binding.ratio} >0 primers are allowed to bind at any location within the templates.
However, a warning is given if the ratio of primers binding to non-target regions exceeds the \code{allowed.other.binding.ratio}.}

\item{allowed.region.definition}{Definition of the target binding sites used for evaluating the coverage.
If \code{allowed.region.definition} is \code{within}, primers have to lie within the allowed binding region.
If \code{allowed.region.definition} is \code{any}, primers have to overlap with the allowed binding region.
The default is that primers have to bind within the target binding region.}

\item{updateProgress}{Progress callback function for shiny.}
}
\value{
Primer data frame with information on the covered template sequences.
}
\description{
Determines binding properties of primers.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/primers.R
\name{create.kmer}
\alias{create.kmer}
\title{Creation of k-mers of a single sequence.}
\usage{
create.kmer(seq, k)
}
\arguments{
\item{seq}{A character vector.}

\item{k}{The size of the k-mer.}
}
\value{
A names character vector, where the names 
are the relative positions of the k-mers and the values
give the character vector of the k-mer.
}
\description{
Creation of k-mers of a single sequence.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/initialize_primers.R
\name{check.init.primer.length}
\alias{check.init.primer.length}
\title{Primer Length Check.}
\usage{
check.init.primer.length(
  template.df,
  allowed.region.definition = c("within", "any"),
  primer.lengths,
  mode.directionality = c("fw", "rev", "both")
)
}
\arguments{
\item{template.df}{Template data frame.}

\item{allowed.region.definition}{Definition of allowed binding regions.}

\item{primer.lengths}{The desired lengths of the priemrs.}

\item{mode.directionality}{The primer directionality.}
}
\value{
TRUE, if primers of the desired length can be constructed,
}
\description{
Checks whether it is possible to construct primers of the desired length.
}
\keyword{FALSE}
\keyword{internal}
\keyword{otherwise.}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/con_primer_efficiency.R
\name{compute.primer.efficiencies}
\alias{compute.primer.efficiencies}
\title{Primer Efficiency.}
\usage{
compute.primer.efficiencies(
  primer.df,
  template.df,
  annealing.temp,
  taqEfficiency,
  primer_conc,
  na_salt_conc,
  mg_salt_conc,
  k_salt_conc,
  tris_salt_conc,
  mode = c("on_target", "off_target")
)
}
\arguments{
\item{primer.df}{Primer data frame.}

\item{template.df}{Template data frame.}

\item{annealing.temp}{Annealing temperature for which to evaluate efficiency.}

\item{taqEfficiency}{Whether the efficiency shall be computed
using a mismatch-model developed for Taq polymerases. The default setting
is \code{TRUE}. Set \code{taqEfficiency} to \code{FALSE} if you are using
another polymerase than Taq.}

\item{primer_conc}{Primer concentration.}

\item{na_salt_conc}{Sodium ion concentration.}

\item{mg_salt_conc}{Magensium ion concentration.}

\item{k_salt_conc}{Potassium ion concentration.}

\item{tris_salt_conc}{Tris ion concentration.}

\item{mode}{Compute efficiencies for on-target coverage events (\code{on_target})
or off-target coverage events (\code{off_target}).}
}
\value{
A list with the efficiency of every primer binding event.
}
\description{
Computes the efficiency of primer binding events for Taq polymerase.
}
\details{
This function uses DECIPHER's \code{\link[DECIPHER]{CalculateEfficiencyPCR}}.
}
\examples{
data(Ippolito)
p <- PCR(settings)
# Requires OligoArrayAux software:
\dontrun{
eff.df <- compute.primer.efficiencies(primer.df, template.df, 55, 
             p$primer_concentration, p$Na_concentration,
             p$Mg_concentration, p$K_concentration, p$Tris_concentration)
} 
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plots_constraints.R
\name{plot_constraint_fulfillment,list-method}
\alias{plot_constraint_fulfillment,list-method}
\title{Comparison of Evaluation Results.}
\usage{
\S4method{plot_constraint_fulfillment}{list}(
  primers,
  settings,
  active.constraints,
  plot.p.vals = FALSE,
  ncol = 2,
  highlight.set = NULL
)
}
\arguments{
\item{primers}{A list of \code{Primers} objects.}

\item{settings}{A \code{DesignSettings} object.}

\item{active.constraints}{The identifiers of constarints to be plotted for fulfillment.}

\item{plot.p.vals}{Whether p-values from Fisher's exact test
should be annotated for every primer set.}

\item{ncol}{The number of columns for facet wrap.}

\item{highlight.set}{Identifiers of primer sets to be highlighted.}
}
\value{
Plot indicating the ratio of primers
fulfilling the constraints specified in \code{constraint.settings}
for each primer set in \code{primers}.
}
\description{
Plots the percentage of primers fulfilling the specified
constraints for multiple primer sets.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/primers.R
\name{plot_primer_binding_regions,Primers,Templates-method}
\alias{plot_primer_binding_regions,Primers,Templates-method}
\title{Plot of Primer Binding Regions for a Single Primer Set.}
\usage{
\S4method{plot_primer_binding_regions}{Primers,Templates}(
  primers,
  templates,
  direction = c("both", "fw", "rev"),
  group = NULL,
  relation = c("fw", "rev"),
  region.names = c("Binding region", "Amplification region")
)
}
\arguments{
\item{primers}{An object of class \code{Primers} with annotated
primer coverage.}

\item{templates}{An object of class \code{Templates} providing the
template sequences corresponding to \code{primers}.}

\item{direction}{Primer direction}

\item{group}{The template groups for which binding events should
be determined. By default, \code{group} is set to \code{NULL} such that
all templates are considered.}

\item{relation}{A character vector specifying whether binding region data shall
be plotted relative to the forward (\code{fw}) or reverse (\code{rev}) 
target binding regions.}

\item{region.names}{Names for the primer binding region and the amplified region.}
}
\value{
A histogram of primer binding regions.
}
\description{
Plots the primer binding regions in the templates.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/constraints.R
\name{check_constraint_settings_validity}
\alias{check_constraint_settings_validity}
\title{Check the Validity of the Constraint Settings.}
\usage{
check_constraint_settings_validity(object)
}
\arguments{
\item{object}{An \code{AbstractConstraintSettings} object.}
}
\value{
\code{TRUE} if the constraint settings are valid, \code{FALSE} otherwise.
}
\description{
Checks whether the status and the active constraints match.
Determines whether the constraints are allowed/known.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/IO_view.R
\name{exclude.cols}
\alias{exclude.cols}
\title{Exclusion of Columns}
\usage{
exclude.cols(excl.col, template.df)
}
\arguments{
\item{excl.col}{Names of columns in \code{template.df} to be removed.}

\item{template.df}{Data frame for which columns in \code{excl.col} should be removed.}
}
\value{
\code{template.df} with removed columns as specified in \code{excl.col}.
}
\description{
Removes columns from a data frame.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/filters.R
\name{get_max_set_coverage}
\alias{get_max_set_coverage}
\title{Determination of Maximal Coverage.}
\usage{
get_max_set_coverage(
  primer.df,
  template.df,
  Tm.brackets,
  settings,
  mode.directionality,
  max.only = TRUE
)
}
\arguments{
\item{primer.df}{An object of class \code{Primers}.}

\item{template.df}{An objectc of class \code{Templates}.}

\item{Tm.brackets}{A data frame with temperature information.}

\item{settings}{A \code{DesignSettings} object.}

\item{mode.directionality}{The direction of the primers.}

\item{max.only}{Whether only the maxium coverage shall be returned.
If \code{max.only} is \code{FALSE}, the coverage ratios of all 
melting temperature sets according to \code{Tm.brackets} are returned.}
}
\value{
The maximal coverage ratio of a primer set if \code{max.only} is \code{TRUE}
or the coverages of all melting temperature seets if \code{max.only} is \code{FALSE}.
}
\description{
Determines the maximal coverage ratio of a set of primers 
for primer subsets valid for a certain temperature range.
a certain melting temperature range.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/templates.R
\name{cbind2,Templates,ANY-method}
\alias{cbind2,Templates,ANY-method}
\alias{rbind2,Templates,ANY-method}
\alias{[,Templates,ANY-method}
\alias{[,Templates-method}
\alias{$<-,Templates-method}
\title{S4 cbind for Templates.}
\usage{
\S4method{cbind2}{Templates,ANY}(x, y, ...)

\S4method{rbind2}{Templates,ANY}(x, y, ...)

\S4method{[}{Templates,ANY}(x, i, j, ..., drop = TRUE)

\S4method{$}{Templates}(x, name) <- value
}
\arguments{
\item{x}{The Template data frame.}

\item{y}{Another data frame.}

\item{...}{Other arguments to the slice operator.}

\item{i}{The row index.}

\item{j}{The column index.}

\item{drop}{Simplify data frame?}

\item{name}{The name of the column.}

\item{value}{The values of the column.}
}
\value{
Cbinded template data frame.

Rbinded template data frame.

Subsetted template data frame.

Templates with replaced column.
}
\description{
S4 cbind function for Templates data frame.

S4 rbind function for templates.

Slicing of Templates data frame object.

Set a column in a Templates data frame.
}
\examples{
data(Ippolito)
template.df <- cbind2(template.df, seq_len(nrow(template.df)))
data(Ippolito)
template.df <- rbind2(template.df, template.df)
data(Ippolito)
template.df <- template.df[1:2,]
data(Ippolito)
template.df$ID[1] <- "newID"
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/helper_functions.R
\name{my_rbind}
\alias{my_rbind}
\title{Smartbind preserving classes.}
\usage{
my_rbind(...)
}
\arguments{
\item{...}{Data frames.}
}
\value{
A data frame resulting from row binding of \code{...}.
}
\description{
Rbind allowing for column mismatch, retains the classes of the data frames.
Motivation: smartbind/rbind.fill only keep the data.frame class
but not additional classes.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/check_stop_codons.R
\name{format.seq.ali}
\alias{format.seq.ali}
\title{Format mismatches}
\usage{
\method{format}{seq.ali}(seq, pos, format.type)
}
\arguments{
\item{seq}{The input sequence.}

\item{pos}{The mismatch positions to be formatted.}

\item{format.type}{Vector of giving the style (bold/italics) for each pos.}
}
\value{
The input sequence with highlighted mismatch positions.
}
\description{
Formats a sequence for highlighting mismatches in an alignment.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/primers.R, R/PrimerDesign.R,
%   R/initialize_primers.R, R/optimization_algo.R
\name{PrimerDesign}
\alias{PrimerDesign}
\alias{classify_design_problem}
\alias{get_initial_primers}
\alias{design_primers}
\title{Primer Design Functionalities.}
\usage{
classify_design_problem(
  template.df,
  mode.directionality = c("both", "fw", "rev"),
  primer.length = 18,
  primer.estimate = FALSE,
  required.cvg = 1
)

get_initial_primers(
  sample,
  template.df,
  primer.lengths,
  mode.directionality = c("fw", "rev"),
  allowed.region.definition = c("within", "any"),
  init.algo = c("naive", "tree"),
  max.degen = 16,
  conservation = 1,
  updateProgress = NULL
)

design_primers(
  template.df,
  mode.directionality = c("both", "fw", "rev"),
  settings,
  init.algo = c("naive", "tree"),
  opti.algo = c("Greedy", "ILP"),
  required.cvg = 1,
  timeout = Inf,
  max.degen = 16,
  conservation = 1,
  sample.name = NULL,
  cur.results.loc = NULL,
  primer.df = NULL,
  updateProgress = NULL
)
}
\arguments{
\item{template.df}{A \code{Templates} object containing the template
sequences with annotated primer target binding regions.}

\item{mode.directionality}{The template strand for which primers shall be designed.
Primers can be designed either for forward strands ("fw"), 
for reverse strands ("rev"), or for both strands ("both"). The default setting
is "both".}

\item{primer.length}{A scalar numeric providing the 
target length of the designed primers. The default length 
of generated primers is set to \code{18}.}

\item{primer.estimate}{Whether the number of required primers shall be estimated. By default (\code{FALSE}), the number of required primers is not estimated.}

\item{required.cvg}{The desired ratio of of covered template sequences. 
If the target coverage ratio cannot be reached, the constraint settings
are relaxed according to the the constraint limits in order to reach the target coverage. 
The default \code{required.cvg} is set to 1, indicating that 100\% of the templates are to be covered.}

\item{sample}{Character vector providing an identifier for the templates.}

\item{primer.lengths}{Numeric vector of length 2 providing the 
minimal and maximal allowed lengths for generated primers.}

\item{allowed.region.definition}{A character vector providing the definition
of region where primers are to be constructed.
If \code{allowed.region.definition} is "within", constructed primers lie within the allowed binding region.
If \code{allowed.region.definition} is "any", primers overlap with the allowed binding region.
The default is "within".}

\item{init.algo}{The algorithm to be used for initializing primers.
If \code{init.algo} is "naive", then primers are constructed from substrings of the input template sequences.
If \code{init.algo} is "tree", phylogenetic trees are used to form degenerate primers whose degeneracy is bounded by \code{max.degen}.
This option requires an installation of MAFFT (see notes). The default \code{init.algo} is "naive".}

\item{max.degen}{The maximal degeneracy of primer candidates. This setting is particularly
relevant when \code{init.algo} is set to "tree". The default setting is \code{16}, which means
that at most 4 maximally degenerate positions are allowed per primer.}

\item{conservation}{Restrict the percentile of considered regions according to their conservation.
Only applicable for the tree-based primer initialization. At the 
default of 1, all available binding regions are considered.}

\item{updateProgress}{Shiny progress callback function. The default is \code{NULL}
such that no progress is logged.}

\item{settings}{A \code{DesignSettings} object specifying the constraint settings for designing primers.}

\item{opti.algo}{The algorithm to be used for solving the primer set covering problem. 
If \code{opti.algo} is "Greedy" a greedy algorithm is used to solve the 
set cover problem. If \code{opti.algo} is "ILP" an integer linear 
programming formulation is used. The default \code{opti.algo} is "Greedy".}

\item{timeout}{Timeout in seconds. Only applicable when \code{opti.algo} is "ILP".
The default is \code{Inf}, which does not limit the runtime.}

\item{sample.name}{An identifier for the primer design task. The default setting is
\code{NULL}, which means that the run identifier provided in \code{template.df} is used.}

\item{cur.results.loc}{Directory for storing the results of the primer design procedure.
The default setting is \code{NULL} such that no output is stored.}

\item{primer.df}{An optional \code{Primers} object. If an evaluated \code{primer.df} is provided,
the primer design procedure only optimizes \code{primer.df} and does not perform
the initialization and filtering steps. The default is \code{NULL} such that
primers are initialized and filtered from scratch.}
}
\value{
\code{classify_design_problem} returns a list with the 
following fields:
\describe{
\item{\code{Classification}}{The estimated difficulty of the primer design task.}
\item{\code{Class-Distances}}{The total variance distance of the fitted
beta distribution to the reference distribution.}
\item{\code{Confidence}}{The confidence in the estimate of the
design tasks' difficulty as based on the class distances.}
\item{\code{Uncertain}}{Whether the classification is highly uncertain, that is
low-confidence.}
\item{\code{Nbr_primers_fw} and \code{Nbr_primers_rev}}{The respective number of 
required forward and reverse primers if \code{primer.estimate} was set to \code{TRUE}.}
}

\code{get_initial_primers} returns a data frame 
with candidate primers for optimization.

\code{design_primers} returns a list with the following fields:
\describe{
\item{\code{opti}:}{A \code{Primers} object providing the designed primer set.}
\item{\code{used_constraints}:}{A list with \code{DesignSettings} objects
for each primer direction providing the (possibly relaxed) constraints used
for designing the optimal primers.}
\item{\code{all_results}:}{A list containing objects of class \code{Primers}.
Each list entry corresponds to an optimal primer set for a given
melting temperature.}
\item{\code{all_used_constraints}:}{A list containing \code{DesignSettings} object for each optimized set in \code{all_results}.}
\item{\code{filtered}:}{A list with data providing information on the results
of the filtering procedure.}
}
}
\description{
\describe{
\item{\code{design_primers}}{Designs a primer set maximizing the number
of covered templates using the smallest possible number of primers. 
The algorithm tries to ensure that the designed set of primers 
achieves a coverage ratio not lower than \code{required.cvg}.
To this end, the constraints for designing primers may be relaxed.}
\item{\code{get_initial_primers}}{Creates a set of primer candidates 
based on the input template sequences. This set of primers can 
be used to create custom primer design algorithms.}
}
}
\details{
\code{classify_design_problem} determines the difficulty of a primer design 
task by estimating the distribution of coverage ratios per primer
by performing exact string matching with 
primers of length \code{primer.length}, which are constructed
by extracting template subsequences. Next, a beta distribution
is fitted to the estimated coverage distribution, which is
then compare to reference distributions representing
primer design problems of different difficulties via the
total variance distance. The difficulty of the input primer design
problem is found by selecting the class of the 
reference distributions that has the smallest distance
to the estimated coverage distribution.
An estimate of the required number of primers to reach a given
\code{required.cvg} can be computed by setting
\code{primer.estimate} to \code{TRUE}. Since this estimate
is based solely on perfect matching primers, the number of
primers that would actually be required is typically less.

The primer design algorithm used by \code{design_primers} consists
of three steps: primer initialization, filtering, and optimization.
The method for initializing a set of candidate primers is determined
via \code{init.algo}. If \code{init.algo} is set to \emph{naive}, primers
are created by extracting substrings from all input template sequences.
If \code{init.algo} is set to \emph{tree}, degenerate primers are created by
merging similar subsequences by forming their consensus sequence up to
a degeneracy of at most \code{max.degen}. The tree-based initialization
is recommended for related sequences.

The candidate primer set is filtered according to the constraints
specified in the \code{settings} object. In some cases, it is necessary
to relax the constraints in order to reach the desired \code{required.cvg}.
In these cases, primers that fail the input constraints may be selected. 
If you would like to skip the initialization and filtering stages,
you can provide an evaluated \code{Primers} object via \code{primer.df}.

Optimizing a primer set entails finding the smallest subset of primers
maximizing the coverage, which is done by solving the set cover problem.
If melting temperature differences are a constraint,
the optimization procedure automatically samples ranges of melting
temperatures to find optimal sets for all possible temperatures.
You can select the used optimization algorithm via \code{optia.algo}, where
you can set "Greedy" for a greedy algorithm or "ILP for 
an integer linear program formulation (ILP).
While the worst-case runtime of the
greedy algorithm is shorter than the worst-case runtime of the ILP, 
the greedy solution may yield larger primer sets than the ILP solution.
}
\note{
Some constraints can only be computed if additional software is installed,
please see the documentation of \code{\link{DesignSettings}} for more information.
The usage of \code{init.algo = "tree"} requires an installation of
the multiple alignment program MAFFT 
(http://mafft.cbrc.jp/alignment/software/).
}
\examples{

data(Ippolito)
# Naive primer initialization
init.primers <- get_initial_primers("InitialPrimers", template.df, 
                         c(18,18), "fw", init.algo = "naive")
# Tree-based primer initialization (requires MAFFT)
\dontrun{
init.primers <- get_initial_primers("InitialPrimers", template.df, 
                         c(18,18), "fw", init.algo = "tree")
}

# Define PCR settings and primer criteria
data(Ippolito)
# design only with minimal set of constraints
constraints(settings)$primer_length <- c("min" = 18, "max" = 18)
constraints(settings) <- constraints(settings)[c("primer_length", "primer_coverage")]
# Design only forward primers using a greedy algorithm
optimal.primers.greedy <- design_primers(template.df[1:2,], "both", settings, init.algo = "naive")
# Usage of the tree-based initialization strategy (requires MAFFT)
\dontrun{
out.dir <- tempdir()
optimal.primers.tree <- design_primers(template.df[1:2,], "both", settings,
                         init.algo = "tree", opti.algo = "ILP",
                         max.degen = 16,
                         cur.results.loc = out.dir)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/optimization_ILP.R
\name{ILPConstrained}
\alias{ILPConstrained}
\title{Construct Coverage ILP.}
\usage{
ILPConstrained(D, cvg.matrix, time.limit = NULL, presolve.active = FALSE)
}
\arguments{
\item{D}{Binary dimerization matrix.}

\item{cvg.matrix}{Binary coverage matrix.}

\item{time.limit}{Time limit for ILP optimization in seconds.}

\item{presolve.active}{Whether the ILP presolver should be used.
This is set to \code{FALSE} by default, since presolving may lead
to inferior solutions. However, for large problems presolving
might be useful.}
}
\value{
An instance of the set cover ILP.
}
\description{
Constructs an ILP modeling the primer set cover problem.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/optimization_greedy.R
\name{optimize.primer.cvg}
\alias{optimize.primer.cvg}
\title{Greedy Optimization}
\usage{
optimize.primer.cvg(
  primers,
  template.df,
  mode.directionality,
  cur.opti.constraints,
  target.temp,
  allowed.mismatches,
  opti.limits,
  primer_conc,
  na_salt_conc,
  mg_salt_conc,
  k_salt_conc,
  tris_salt_conc,
  updateProgress = NULL
)
}
\arguments{
\item{primers}{Primer data frame to be optimized.}

\item{template.df}{Template data frame.}

\item{mode.directionality}{Primer direction.}

\item{cur.opti.constraints}{List with optimization constraint settings.}

\item{target.temp}{Target annealing temperature of the optimized primer set in Celsius.}

\item{allowed.mismatches}{The number of mismatches primers are allowed to have with the templates.}

\item{opti.limits}{List with optimization limits.}

\item{primer_conc}{Primer concentration.}

\item{na_salt_conc}{Sodium ion concentration.}

\item{mg_salt_conc}{Magensium ion concentration.}

\item{k_salt_conc}{Potassium ion concentration.}

\item{tris_salt_conc}{Tris ion concentration.}

\item{updateProgress}{Shiny progress callback function.}

\item{template_conc}{Template concentration.}
}
\value{
List with optimization data.
}
\description{
Greedy approach for solving the primer set coverage problem.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/initialize_primers_tree.R
\name{align.seqs}
\alias{align.seqs}
\title{Multiple Sequence Alignment}
\usage{
align.seqs(seqs, names)
}
\arguments{
\item{seqs}{The sequences to be aligned.}

\item{names}{The identifiers of the sequences.}
}
\value{
An alignment object as created from seqinr's read.alignment method.
}
\description{
Computes a multiple sequence alignment using MAFFT.
}
\references{
Katoh, Misawa, Kuma, Miyata 2002 (Nucleic Acids Res. 30:3059-3066) 
MAFFT: a novel method for rapid multiple sequence alignment based
on fast Fourier transform.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plots_constraints.R
\name{plot_constraint_deviation,list-method}
\alias{plot_constraint_deviation,list-method}
\title{Plot of Constraint Deviations for Multiple Primer Sets.}
\usage{
\S4method{plot_constraint_deviation}{list}(
  primers,
  settings,
  active.constraints,
  deviation.per.primer = FALSE
)
}
\arguments{
\item{settings}{A \code{DesignSettings} object
containing the target ranges for the primer properties.}

\item{active.constraints}{Constraint identifiers to be plotted.}

\item{deviation.per.primer}{Whether to show the deviation per primer or per constraint.}

\item{constraint.df}{An evaluated object of class \code{Primers}.}
}
\value{
A boxplot of deviations
}
\description{
Plots a box plot of the absolute mean deviation of each primer 
for comparing multiple primer sets.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/settings.R
\name{filterLimits}
\alias{filterLimits}
\alias{filterLimits,DesignSettings-method}
\title{Getter for Filtering Constraint Limits.}
\usage{
filterLimits(x)

\S4method{filterLimits}{DesignSettings}(x)
}
\arguments{
\item{x}{A \code{DesignSettings} object.}
}
\value{
Gets the list of filtering limits.
}
\description{
Gets the limits on the constraints
that are used for the filtering procedure when designing primers
using the \code{Input_Constraint_Boundaries} slot of the provided
\code{DesignSettings} object \code{x}.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/filters.R
\name{filter.comparison.primers}
\alias{filter.comparison.primers}
\title{Filter Multiple Primer Sets.}
\usage{
filter.comparison.primers(
  primers,
  templates,
  active.constraints,
  settings,
  updateProgress = NULL
)
}
\arguments{
\item{primers}{List with primer data frames.}

\item{templates}{List with template data frames.}

\item{active.constraints}{Strings giving the constraints that are to be checked.}

\item{settings}{List with settings.}

\item{updateProgress}{Progress callback function for shiny.}
}
\value{
A list with filtered primer data frames.
}
\description{
Filters multiple primer sets at once.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/templates.R
\name{get.leader.exon.regions.single}
\alias{get.leader.exon.regions.single}
\title{Individual Binding Annotation}
\usage{
get.leader.exon.regions.single(
  l.seq,
  lex.seq,
  direction = c("fw", "rev"),
  gap.char
)
}
\arguments{
\item{l.seq}{Data frame with individual binding regions.}

\item{lex.seq}{Template data frame.}

\item{direction}{The primer direction for which the binding info is valid.}

\item{gap.char}{The character for gaps in alignments.}
}
\value{
Template data  frame with annotated binding regions.
}
\description{
Annotate individual binding regions.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/con_dimerization.R
\name{batchify.temp}
\alias{batchify.temp}
\title{Batchification by Temperature.}
\usage{
batchify.temp(tasks, annealing.temp)
}
\arguments{
\item{tasks}{The tasks to assign to individual batches.}

\item{annealing.temp}{The annealing temperatures according to which batches are to be created.}
}
\value{
A list of lists containing indices corresponding to \code{tasks}, each list gives a batch.
}
\description{
Batchification by Temperature.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/primers.R
\name{rbind.Primers}
\alias{rbind.Primers}
\title{rbind for Primers class.}
\usage{
\method{rbind}{Primers}(...)
}
\arguments{
\item{...}{Parameters for rbind function.}
}
\value{
Row-binded Primers data frame.
}
\description{
Ensures that the rbind result has the appropriate class.
}
\examples{
data(Ippolito)
primer.df <- rbind(primer.df, primer.df)
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/errors.R
\name{my.warning}
\alias{my.warning}
\title{Custom Warning.}
\usage{
my.warning(subclass, message, call = sys.call(-1), ...)
}
\arguments{
\item{subclass}{String giving the specific type of error.}

\item{message}{Message to be displayed to the user.}

\item{call}{Environment where the error ocurred.}

\item{...}{Other arguments to the condition function.}
}
\value{
Generates a custom warning.
}
\description{
Creates a warning with a custom class.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/IO.R
\name{get_report_fname}
\alias{get_report_fname}
\title{Creation of a Filename for Reports.}
\usage{
get_report_fname(report.name, sample.name)
}
\arguments{
\item{report.name}{The identifier for the report type.}

\item{sample.name}{The identifier of the sample that was analyzed.}
}
\value{
A character vector.
}
\description{
Creates the filename for reports.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/check_stop_codons.R
\name{split_str_by_index}
\alias{split_str_by_index}
\title{Split a sequence}
\usage{
split_str_by_index(target, index)
}
\arguments{
\item{target}{The target string.}

\item{index}{The position for the split.}
}
\value{
List with splitted strings
}
\description{
Splits a sequence at a specified positions
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/initialize_primers.R
\name{create.initial.primer.set}
\alias{create.initial.primer.set}
\title{Creation of an Initial Primer Set.}
\usage{
create.initial.primer.set(
  template.df,
  primer.lengths,
  mode.directionality = c("fw", "rev"),
  sample,
  allowed.region.definition = c("within", "any"),
  init.algo = c("naive", "tree"),
  max.degen,
  conservation,
  updateProgress = NULL
)
}
\arguments{
\item{template.df}{Template data frame.}

\item{primer.lengths}{Vector containing the permissible primer lengths.}

\item{mode.directionality}{Direction of primers to be created.}

\item{sample}{Name of the template sample.}

\item{allowed.region.definition}{Definition of the allowed binding region.}

\item{init.algo}{Algorithm for initializing primers.}

\item{max.degen}{Maximal allowed degeneration of created primers.}

\item{conservation}{Required conservation of primers.
The value of \code{conservation} should be in the range[0,1].}

\item{updateProgress}{Shiny progress object.}
}
\value{
An initialized data frame of candidate primers.
}
\description{
Creates an initial set of candidate primers for primer design.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/templates.R
\name{cbind.Templates}
\alias{cbind.Templates}
\title{cbind for Template class.}
\usage{
\method{cbind}{Templates}(...)
}
\arguments{
\item{...}{Parameters for cbind function.}
}
\value{
Column binded Templates data frame.
}
\description{
Ensures that the cbind result has the appropriate class.
}
\examples{
data(Ippolito)
template.df <- cbind(template.df, seq_len(nrow(template.df)))
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/templates.R
\name{validate_templates}
\alias{validate_templates}
\title{Validates a Template Object.}
\usage{
validate_templates(object)
}
\arguments{
\item{object}{An input data frame to be checked for being a template data frame.}
}
\value{
\code{TRUE}, if the object is valid, \code{FALSE} otherwise.
}
\description{
Checks whether a Templates object is valid or not.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plots_coverage.R
\name{get_cvg_stats,list-method}
\alias{get_cvg_stats,list-method}
\title{Coverage Statistics for Multiple Primer Sets.}
\usage{
\S4method{get_cvg_stats}{list}(
  primers,
  templates,
  for.viewing = FALSE,
  total.percentages = FALSE,
  allowed.mismatches = Inf,
  cvg.definition = c("constrained", "basic")
)
}
\arguments{
\item{primers}{A list with objects of class \code{Primers} containing
primers with evaluated coverage.}

\item{templates}{A list with objects of class \code{Templates} containing
templates with evaluated coverage.}

\item{for.viewing}{Whether the table should be formatted
for viewing rather than processing.}

\item{total.percentages}{Whether group coverage percentages
should relate to all template sequences or just those templates
belonging to a specific group.}

\item{allowed.mismatches}{The maximal allowed number of mismatches.
By default, the number of mismatches is not restricted.}

\item{cvg.definition}{If \code{cvg.definition} is set to
"constrained", the statistics for the expected
coverage (after applying the coverage constraints) are retrieved.
If \code{cvg.definition} is set to "basic", the coverage is determined 
solely by string matching (i.e. without applying the coverage constraints).
By default, \code{cvg.definition} is set to "constrained".}
}
\value{
Data frame with coverage statistics.
}
\description{
Retrieve statistics on covered templates for multiple primer sets.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/settings.R
\name{xmlToChar}
\alias{xmlToChar}
\title{Conversion of XML to Character.}
\usage{
xmlToChar(xml)
}
\arguments{
\item{xml}{An xml object to be converted to character.}
}
\value{
A character vector.
}
\description{
Converts an XML object to a character string.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ambiguity.R
\name{merge.primer.entries}
\alias{merge.primer.entries}
\title{Merge similar primers}
\usage{
\method{merge}{primer.entries}(opti.result, merge.idx, mode.directionality = c("fw", "rev", "both"))
}
\arguments{
\item{opti.result}{Input primer data frame.}

\item{merge.idx}{List of lists with merge indices (get.merge.idx).}

\item{mode.directionality}{Direction of primers.}
}
\value{
A primer data frame where entries of similar primers are merged.
}
\description{
Merges the entries of similar entries in the input primer data frame, given a list with merge indices.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/con_melting_temperature.R
\name{call.melt}
\alias{call.melt}
\title{Thermodynamic melting temperature computations.}
\usage{
call.melt(
  primers,
  complements,
  primer_conc,
  na_salt_conc,
  mg_salt_conc,
  k_salt_conc,
  tris_salt_conc
)
}
\arguments{
\item{primers}{Character vector of primer strings.}

\item{complements}{Character vector with complement sequences
corresponding to \code{primers}.}

\item{primer_conc}{Primer concentration.}

\item{na_salt_conc}{Sodium ion concentration.}

\item{mg_salt_conc}{Magensium ion concentration.}

\item{k_salt_conc}{Potassium ion concentration.}

\item{tris_salt_conc}{Tris buffer concentration.}
}
\value{
Melting temperature data frame.
}
\description{
Computes the melting temperature for the input primers.
}
\references{
Le Novère N. (2001). MELTING, computing the 
melting temperature of nucleic acid duplex. Bioinformatics, 17: 1226-1227. 
Dumousseau M., Rodriguez N., Juty N., Le Novère N. (2012) MELTING, 
a flexible platform to predict the melting temperatures of nucleic acids. 
BMC Bioinformatics, 13: 101.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plots_comparison.R
\name{plot_primer.comparison.mismatches}
\alias{plot_primer.comparison.mismatches}
\title{Plot Primer Mismatches}
\usage{
plot_primer.comparison.mismatches(
  primer.data,
  template.data,
  allowed.mismatches,
  highlight.set = NULL
)
}
\arguments{
\item{primer.data}{List with primer data frames.}

\item{template.data}{List with template data frames.}

\item{allowed.mismatches}{Allowed mismatches.}

\item{highlight.set}{Primer sets to highlight in the plot.}
}
\value{
Plot of mismatches for comparison.
}
\description{
Plots primer mismatches for every set.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/constraints_eval.R
\name{were.constraints.relaxed}
\alias{were.constraints.relaxed}
\title{Check for Relaxation}
\usage{
were.constraints.relaxed(used.constraints, input.constraints)
}
\arguments{
\item{used.constraints}{The constraints that were used during the optimization.}

\item{input.constraints}{The user-specified constraints.}
}
\value{
If \code{input.constraints} was relaxed TRUE is returned, otherwise FALSE.
}
\description{
Determines whether constraints where relaxed or not.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/check_stop_codons.R
\name{interleave}
\alias{interleave}
\title{Interleave strings
Combines the input vectors in an interleaved fashion.}
\usage{
interleave(v1, v2)
}
\arguments{
\item{v1}{Input string.}

\item{v2}{Input string.}
}
\value{
The interleaved combination of \code{v1} and \code{v2}.
}
\description{
Interleave strings
Combines the input vectors in an interleaved fashion.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/zzz.R
\name{check.tool.function}
\alias{check.tool.function}
\title{Check Functionality of Third-Party Tools.}
\usage{
check.tool.function(frontend = FALSE)
}
\arguments{
\item{frontend}{Whether tool functionality shall be checked for the frontend.}
}
\value{
\code{TRUE} for each functioning tool, \code{FALSE} for non-functioning tools.
}
\description{
Checks whether all required tools should work.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/settings.R
\name{check_settings_validity}
\alias{check_settings_validity}
\title{Validity Check for DesignSettings.}
\usage{
check_settings_validity(object)
}
\arguments{
\item{object}{A DesignSettings object to be checked for validity.}
}
\value{
\code{TRUE} if \code{object} is valid, FALSE otherwise.
}
\description{
Validates whether a DesignSettings object has the correct structure.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/templates.R
\name{get.leader.exon.regions}
\alias{get.leader.exon.regions}
\title{Assign Binding Regions}
\usage{
get.leader.exon.regions(lex.seqs, uni.leaders)
}
\arguments{
\item{lex.seqs}{Data frame with template sequences.}

\item{uni.leaders}{Data frame with individual allowed binding regions.}
}
\value{
Template data frame with annotated binding regions.
}
\description{
Augments a template data frame with individual binding regions.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/optimization_greedy.R
\name{filter_primers.by.Tm.delta}
\alias{filter_primers.by.Tm.delta}
\title{Filter by Melting Temperature Difference}
\usage{
filter_primers.by.Tm.delta(target.temp, selected.primers, max.Tm.delta)
}
\arguments{
\item{target.temp}{Target melting temperature in Celsius.}

\item{selected.primers}{Current candidate primer data frame.}

\item{max.Tm.delta}{Maximum allowed difference of primer melting temperatures to target temperature.}
}
\value{
Filtered primer data frame.
}
\description{
Filters primers by melting temperature differences.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ambiguity.R
\name{merge.ambig.primers}
\alias{merge.ambig.primers}
\title{Merge similar primers}
\usage{
\method{merge}{ambig.primers}(primer.df, mode.directionality = c("fw", "rev", "both"), max.degeneracy)
}
\arguments{
\item{primer.df}{Primer data frame.}

\item{mode.directionality}{Analysis direction.}

\item{max.degeneracy}{Maximal degeneracy of merged primers.}
}
\value{
A primer data frame where similar primers are merged into one entry.
}
\description{
Merges similar primers contained in the input primer data frame.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/optimization_greedy.R
\name{select.primers.by.cvg}
\alias{select.primers.by.cvg}
\title{Greedy Optimization.}
\usage{
select.primers.by.cvg(
  primers,
  settings,
  template.df,
  mode.directionality = c("fw", "rev"),
  required.cvg = 1,
  allowed.mismatches,
  primer_conc,
  na_salt_conc,
  mg_salt_conc,
  k_salt_conc,
  tris_salt_conc,
  template_conc,
  allowed.other.binding.ratio,
  allowed.stop.codons,
  allowed.region.definition = c("within", "any"),
  disallowed.mismatch.pos,
  target.temps = NULL,
  fw.primers = NULL,
  updateProgress = NULL
)
}
\arguments{
\item{primers}{Primer data frame to be optimized.}

\item{settings}{A \code{DesignSettings} object.}

\item{template.df}{Template data frame.}

\item{mode.directionality}{Primer direction.}

\item{required.cvg}{Target coverage ratio.}

\item{allowed.mismatches}{The number of mismatches primers are allowed to have with the templates.}

\item{primer_conc}{Primer concentration.}

\item{na_salt_conc}{Sodium ion concentration.}

\item{mg_salt_conc}{Magensium ion concentration.}

\item{k_salt_conc}{Potassium ion concentration.}

\item{tris_salt_conc}{Tris ion concentration.}

\item{template_conc}{Template data frame.}

\item{allowed.other.binding.ratio}{Ratio of primers allowed to bind to non-target regions.}

\item{allowed.stop.codons}{Consider mismatch binding events that induce stop codons.}

\item{allowed.region.definition}{Definition of the target binding sites used for evaluating the coverage.
If \code{allowed.region.definition} is \code{within}, primers have to lie within the allowed binding region.
If \code{allowed.region.definition} is \code{any}, primers have to overlap with the allowed binding region.
The default is that primers have to bind within the target binding region.}

\item{disallowed.mismatch.pos}{The number of positions from the primer 3' end where mismatches should not be allowed.
All primers binding templates with mismatches within \code{disallowed.mismatch.pos} from the 3' end are disregarded.}

\item{target.temps}{Target melting temperatures for optimized sets in Celsius.}

\item{fw.primers}{List with already optimized primer data frames corresponding to \code{target.temps}.}

\item{updateProgress}{Shiny progress callback function.}
}
\value{
List with optimization data.
}
\description{
Greedy approach for solving the primer set coverage problem.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/constraints_eval.R
\name{apply.constraint}
\alias{apply.constraint}
\title{Application of Constraints}
\usage{
apply.constraint(
  gc.ratio.fw,
  gc.ratio.rev,
  min.GC,
  max.GC,
  fw.idx,
  rev.idx,
  mode.directionality = c("fw", "rev", "both")
)
}
\arguments{
\item{gc.ratio.fw}{Forward values.}

\item{gc.ratio.rev}{Reverse values.}

\item{min.GC}{Minimal allowed value.}

\item{max.GC}{Maximal allowed value.}

\item{fw.idx}{Indices of forward values to consider.}

\item{rev.idx}{Indices of reverse values to consider.}

\item{mode.directionality}{Direction of primers}
}
\value{
Data frame with \code{TRUE} for values fulfilling the constraints, 
\code{FALSE} otherwise.  Also returns \code{FALSE} if a data point is not available.
}
\description{
Checks whether the input values are within the specified limits.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/filters.R
\name{build.gain.df}
\alias{build.gain.df}
\title{Gain of Coverage by Excluded Primers.}
\usage{
build.gain.df(candidate.df, constraint.settings, constraint.limits, relax.df)
}
\arguments{
\item{candidate.df}{An object of class \code{Primers} containing
excluded primers that are considered for addition to \code{filtered.df}.}

\item{constraint.settings}{A list with the current constraint settings.}

\item{constraint.limits}{The current constraint limits.}

\item{relax.df}{Data frame with count of relaxations per constraint.}
}
\value{
A data frame with exclusion data.
}
\description{
Computes a data frame on the excluded sequences per constraint.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ambiguity.R
\name{rev.comp.sequence}
\alias{rev.comp.sequence}
\title{Reverse complement of a sequence}
\usage{
\method{rev}{comp.sequence}(seq)
}
\arguments{
\item{seq}{the input strings}
}
\value{
The reverse complement of the input sequences.
}
\description{
Computes the reverse complement of the input sequences.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/con_primer_coverage.R
\name{get.duplex.energies}
\alias{get.duplex.energies}
\title{Determination of the Free Binding Energy.}
\usage{
get.duplex.energies(
  primer.df,
  template.df,
  annealing.temp,
  settings,
  mode = c("on_target", "off_target")
)
}
\arguments{
\item{primer.df}{A \code{Primers} object.}

\item{template.df}{A \code{Templates} object.}

\item{annealing.temp}{The vector of optimal annealing temperatures of the primers.}

\item{settings}{A \code{DesignSettings} object.}

\item{mode}{If the \code{mode} is set to "on_target", 
the free energies of binding events in the allowed region are computed, while if the \code{mode} is set to
"off_target", the free energies of off-target events are computed.}
}
\value{
A list of lists containing the numeric free energies of the annealing events for every primer.
}
\description{
Computest the free energy of annealing between primers and templates. If the \code{mode} is set to "on_target", 
the free energies of binding events in the allowed region are computed, while if the \code{mode} is set to
"off_target", the free energies of off-target events are computed.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/con_primer_coverage.R
\name{check.mutations}
\alias{check.mutations}
\title{Identification of Mismatch Mutations.}
\usage{
check.mutations(
  primer.seq,
  pos.start,
  pos.end,
  template.df,
  covered.seqs,
  ORF.data,
  mode.directionality = c("fw", "rev"),
  mutation.types = c("stop_codon", "substitution")
)
}
\arguments{
\item{primer.seq}{Primer sequence string.}

\item{pos.start}{Binding position of primer (start).}

\item{pos.end}{Binding position of primer (end).}

\item{template.df}{Template data frame.}

\item{covered.seqs}{Identifiers of covered templates.}

\item{ORF.data}{Reading frame information of templates.}

\item{mode.directionality}{Directionality of primers.}

\item{mutation.types}{Character vector of the mutation types to be checked for.}
}
\value{
TRUE if the \code{primer.seq} induces a mutation that is
forbidden according to the provided \code{mutation.types}.
}
\description{
Identifies primers that induce mutations due to mismatch binding.
}
\details{
Checks for one primer and all covered templates whether any templates are bound
with mismatches such that a forbidden mutation is induced. A boolean vector indicating 
which binding events induce a forbidden mutation is returned.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/startApp.R
\name{runTutorial}
\alias{runTutorial}
\title{The openPrimeR Tutorial.}
\usage{
runTutorial(dev = FALSE)
}
\arguments{
\item{dev}{A logical indicating whether to start the development version of the tutorial (default: \code{FALSE}).}
}
\value{
Opens the openPrimeR tutorial in a web browser.
}
\description{
Starts a Shiny app containing the openPrimeR tutorial, which
was built using the learnr package. The application
starts locally and should open a new tab in your default
browser. If no browser is opened, please consider the console
output to identify the local port on which the server is running.
}
\note{
The Shiny app can be started only if you fulfill all 
of the suggested package dependencies for the Shiny framework,
so please ensure that you've installed openPrimeR including
all suggested dependencies.
}
\examples{
\dontrun{
# Open the tutorial
if (interactive()) {
runTutorial()
}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/constraints.R
\docType{class}
\name{AbstractConstraintSettings-class}
\alias{AbstractConstraintSettings-class}
\alias{AbstractConstraintSettings}
\title{AbstractClass for Constraint Settings.}
\value{
An \code{AbstractConstraintSettings} object.
}
\description{
The \code{ConstraintSettings} class encapsulates the constraints
on the physicochemical properties of primers.
}
\section{Slots}{

\describe{
\item{\code{status}}{Named boolean vector indicating 
which of the possible constraints are active (\code{TRUE})
and which are not (\code{FALSE}).}

\item{\code{settings}}{Named list containing the 
settings of the active constraints}
}}

\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/errors.R
\name{condition}
\alias{condition}
\title{Condition Constructor}
\usage{
condition(subclass, message, call = sys.call(-1), ...)
}
\arguments{
\item{subclass}{String giving the specific error.}

\item{message}{String giving the user message.}

\item{call}{Environment object.}

\item{...}{Other arguments for the output structure.}
}
\value{
A condition structure.
}
\description{
Constructs a condition for custom errors.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/check_stop_codons.R
\name{adjust.ORF.start}
\alias{adjust.ORF.start}
\title{Adjust ORF position}
\usage{
adjust.ORF.start(ORFs, seq.start)
}
\arguments{
\item{ORFs}{the reading frames (0,1,2).}

\item{seq.start}{the position where a sequence is extracted .}
}
\value{
The adjusted reading frames for the given start positions.
}
\description{
Adjusts the reading frame according to the position at which we consider a
subsequence.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plots_constraints.R
\name{get_constraint_deviation_data}
\alias{get_constraint_deviation_data}
\title{Retrieve data for Constraint Deviations.}
\usage{
get_constraint_deviation_data(constraint.df, constraint.settings)
}
\arguments{
\item{constraint.df}{An evaluated object of class \code{Primers}.}

\item{constraint.settings}{A list with settings for the constraints
that are to be evalated.}
}
\value{
A data frame providing primer-specific 
information on deviations of primer properties from
the desired properties.
}
\description{
Retrieve data for Constraint Deviations.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/primer_significance.R
\name{create_fulfilled_counts}
\alias{create_fulfilled_counts}
\title{Creation of Fulfilled/Failed Constraint Counts.}
\usage{
create_fulfilled_counts(primer.df, eval.cols = NULL)
}
\arguments{
\item{primer.df}{An evaluated \code{Primers} object.}

\item{eval.cols}{Evaluation columns in \code{primer.df} to consider.
By default (\code{NULL}) all evaluation columns are considered.}
}
\value{
A data frame with the number of fulfilled/failed constraints
for \code{primer.df}.
}
\description{
Creates counts of fullfilled/failed constraints.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/constraints_eval.R
\name{apply.constraint.list}
\alias{apply.constraint.list}
\title{Apply Constraints to a List.}
\usage{
apply.constraint.list(
  gc.ratio.fw,
  gc.ratio.rev,
  min.GC,
  max.GC,
  fw.idx,
  rev.idx,
  mode.directionality = c("fw", "rev", "both")
)
}
\arguments{
\item{gc.ratio.fw}{Forward values (comma-separated strings).}

\item{gc.ratio.rev}{Reverse values (comma-separated strings).}

\item{min.GC}{Minimal allowed value.}

\item{max.GC}{Maximal allowed value.}

\item{fw.idx}{Indices of forward values to consider.}

\item{rev.idx}{Indices of reverse values to consider.}

\item{mode.directionality}{Direction of primers}
}
\value{
Data frame with \code{TRUE} for values fulfilling the constraints, 
\code{FALSE} otherwise.
}
\description{
Checks whether the input values are within the specified limits.
}
\details{
Applies a constraint to every element in a vector of comma separated strings.
Applied when filtering covered seqs according to primer efficiency.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/constraints_eval.R
\name{eval.comparison.primers}
\alias{eval.comparison.primers}
\title{Evaluation of Primers for Comparison}
\usage{
eval.comparison.primers(primer.data, constraint.settings)
}
\arguments{
\item{primer.data}{List with primer data frames.}

\item{constraint.settings}{List with constraint.settings.}
}
\value{
List with evaluated primer data frames.
}
\description{
Evaluate multiple primer sets according to the input constraint settings.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plots_constraints.R
\name{plot_constraint.histogram}
\alias{plot_constraint.histogram}
\title{Histogram of Constraints.}
\usage{
plot_constraint.histogram(
  primer.df,
  con.cols,
  con.identifier,
  boundaries = NULL,
  x.limits = NULL
)
}
\arguments{
\item{primer.df}{Primer data frame, not necessarily a \code{Primers} object.}

\item{con.cols}{Constraint identifiers in \code{primer.df} to plot.}

\item{con.identifier}{Name of the constraint to plot.}

\item{boundaries}{List with constraint settings.}

\item{x.limits}{Interval limiting the extent of the x-axis.}
}
\value{
A constraint histogram plot.
}
\description{
Plots a histogram of constraint values.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plots_filtering.R
\name{plot.filtering.stats}
\alias{plot.filtering.stats}
\title{Plot of Overall Filtering Stats.}
\usage{
\method{plot}{filtering.stats}(stats, stats.relax = NULL)
}
\arguments{
\item{stats}{Statistics on the filtering procedure}

\item{stats.relax}{Statistic on the filtering procedure after relaxation.}
}
\value{
A plot for the number of primers after filtering.
}
\description{
Plots the number of primers remaining after each filtering step.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/optimization_ILP.R
\name{get.coverage.matrix}
\alias{get.coverage.matrix}
\title{Coverage Matrix}
\usage{
get.coverage.matrix(primer.df, template.df, constraints = NULL)
}
\arguments{
\item{primer.df}{Primer data frame.}

\item{template.df}{Template data frame.}

\item{constraints}{A character vector of coverage constraints
to be used as entries for the coverage matrix instead of the 0/1 encoding.
At its default setting (\code{NULL}), the 0/1 encoding is used.}
}
\value{
The binary coverage matrix.
}
\description{
Constructs a coverage matrix where rows indicate templates and columns indicate primers.
}
\details{
Entry (i,j) in the matrix is equal to 1 if primer j covers template i and otherwise 0.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/IO.R
\name{check_report_deps}
\alias{check_report_deps}
\title{Check for Report Dependencies.}
\usage{
check_report_deps()
}
\value{
A logical vector giving the dependency availability status.
}
\description{
Checks whether the dependencies for rmarkdown::render() are fulfilled.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/settings.R
\name{constraints.xml.format}
\alias{constraints.xml.format}
\title{Constraint XML Format.}
\usage{
constraints.xml.format(constraints, set.name)
}
\arguments{
\item{constraints}{List with constraint settings.}

\item{set.name}{Identifier for the constraint settings.}
}
\value{
XML string containing the constraint settings.
}
\description{
Format constraint settings for XML output.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/initialize_primers_tree.R
\name{select.primer.region.by.conservation}
\alias{select.primer.region.by.conservation}
\title{Selection by Conservation}
\usage{
select.primer.region.by.conservation(
  primer.range,
  ali.entropy,
  conservation,
  bin,
  gap.char = "-"
)
}
\arguments{
\item{primer.range}{Data frame with primer starts/stops.}

\item{ali.entropy}{Entropy values for the alignment.}

\item{conservation}{Desired ratio of primer conservation.
Only regions with a conservation of at least \code{conservation}
are considered for the initialization of primers.}

\item{bin}{\code{DNABin} alignment of templates.}

\item{gap.char}{The character for alignment gaps.}
}
\value{
Updated primer regions according to the desired conservation.
}
\description{
Selects primer regions for initialization of primers according
to their conservation scores.
}
\details{
The conservation scores are computed using the entropies computed from
the alignment of the template sequence regions of interests.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/templates.R
\name{parse.IMGT.gene.groups}
\alias{parse.IMGT.gene.groups}
\title{Parser for IMGT Groups.}
\usage{
parse.IMGT.gene.groups(IDs)
}
\arguments{
\item{IDs}{Group information strings to be parsed.}
}
\value{
Data frame with structured group information.
}
\description{
Parses IMGT group information contained in FASTA headers.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/con_primer_secondary_structures.R
\name{compute.structure.vienna}
\alias{compute.structure.vienna}
\title{Computation of Secondary Structures with ViennaRNA.}
\usage{
compute.structure.vienna(
  seqs,
  annealing.temperature,
  folding.constraints = NULL,
  id = ""
)
}
\arguments{
\item{seqs}{The input sequences for which structures shall be computed.}

\item{annealing.temperature}{The temperature in degree Celsius 
at which to compute secondary structures.}

\item{folding.constraints}{Character vector specifying the folding
conditions for every input sequence.
For example the constraint \code{xxxxxx...} would forbid 
folding in the first 6 bases and allow folding in the last 3 bases.}

\item{id}{An identifier for storing the files}

\item{constraints}{If provide}
}
\value{
A data frame with secondary structures.
}
\description{
Computes secondary structures using ViennaRNA.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/constraints_eval.R
\name{check.template.constraints}
\alias{check.template.constraints}
\title{Check Constraints on Templates}
\usage{
check.template.constraints(template.constraints)
}
\arguments{
\item{template.constraints}{Strings with comma-separated values to be turned to logical.}
}
\value{
List with boolean values
}
\description{
Transforms the comma-separated input strings to a boolean representation.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/primers.R
\name{validate_primers}
\alias{validate_primers}
\title{Validates a Primers Object.}
\usage{
validate_primers(object)
}
\arguments{
\item{object}{An input data frame to be checked for being a primer data frame.}
}
\value{
\code{TRUE}, if the object is valid, FALSE otherwise.
}
\description{
Checks whether a Primers object is valid or not.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/filters.R
\name{get.constraint.values}
\alias{get.constraint.values}
\title{Get the Values of a Constraint.}
\usage{
get.constraint.values(con.name, cur.candidates, mode.directionality)
}
\arguments{
\item{con.name}{The name of the constraint.}

\item{cur.candidates}{The \code{Primers} data frame where the values should be retrieved.}

\item{mode.directionality}{The direction for which values should be retrieved.}
}
\value{
The constraint values corresponding to \code{con.name} for the primers \code{cur.candidates}.
}
\description{
Get the Values of a Constraint.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plots_coverage.R
\name{plot.excluded.hist}
\alias{plot.excluded.hist}
\title{Plot of Excluded Primers}
\usage{
\method{plot}{excluded.hist}(excluded.df, filtered.stats, template.df)
}
\arguments{
\item{excluded.df}{Data frame with excluded primers.}

\item{filtered.stats}{Data frame with statistics of the filtering procedure.}

\item{template.df}{Template data frame.}
}
\value{
A plot of excluded primers.
}
\description{
Plots histogram of excluded primers.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/constraints_eval.R
\name{eval.constraints}
\alias{eval.constraints}
\title{Evaluation of Constraints'}
\usage{
eval.constraints(
  constraint.df,
  constraint.settings,
  active.constraints,
  mode.directionality = c("fw", "rev", "both"),
  primer.df
)
}
\arguments{
\item{constraint.df}{Primer data frame with computed constraints.}

\item{constraint.settings}{List with allowed values pers constraint.}

\item{active.constraints}{Names of constraints to be evaluated.}

\item{mode.directionality}{Directionality of primers}

\item{primer.df}{Primer data frame corresponding to \code{constraint.df}.}
}
\value{
Augments the \code{constraint.df} data frame with evaluation columns.
}
\description{
Evaluates whether the given primer data frame fulfills the required conditions.
}
\details{
Constraint values should be contained in \code{constraint.df}. 
For each constraint in \code{active.constraints}, a boolean column with the name
\code{EVAL_<constraint_name>} is generated, which indicates whether 
a primer in a given rows fulfills a constraint or not.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/initialize_primers_tree.R
\name{get.tree.seqs}
\alias{get.tree.seqs}
\title{Determine Tree Consensus Sequences}
\usage{
get.tree.seqs(tree, max.degen, primer.candidates)
}
\arguments{
\item{tree}{The phylogenetic tree.}

\item{max.degen}{The maximal degeneration of consensus primers.}

\item{primer.candidates}{Alignment of primers.}
}
\value{
Data frame with consensus primers extracted from the tree.
}
\description{
Creates all possible consensus sequences from a phylogenetic tree.
}
\details{
Ambiguous sequences are only generated with a degeneracy of at most \code{max.degen}. 
The tree is iterated from leaves to the top, i.e., starting from least degeneracy to most degeneracy. 
Merges only take place when the degeneracy of the resulting sequence would
be at most \code{max.degen}. Gaps are removed from the alignments.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plots_coverage.R
\name{get.run.names}
\alias{get.run.names}
\title{Getter for Run Names.}
\usage{
get.run.names(primer.data)
}
\arguments{
\item{primer.data}{A list with \code{Primers} or \code{Templates}.}
}
\value{
A vector with identifiers for every set.
}
\description{
Retrieves the run names of the input data.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/settings.R
\name{create.other.table}
\alias{create.other.table}
\title{Creation of a Table for Other Constraint Settings.}
\usage{
create.other.table(other.settings, col.names, format.type)
}
\arguments{
\item{other.settings}{List with other constraint settings.}

\item{format.type}{How the table shall be formatted.}
}
\value{
A data frame.
}
\description{
Creation of a Table for Other Constraint Settings.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/con_melting_temperature.R
\name{call.melt.single}
\alias{call.melt.single}
\title{Thermodynamic melting temperature computations.}
\usage{
call.melt.single(
  primers,
  complements,
  out.file,
  primer_conc,
  na_salt_conc,
  mg_salt_conc,
  k_salt_conc,
  tris_salt_conc,
  ID
)
}
\arguments{
\item{primers}{List of primer strings.}

\item{complements}{List with corresponding complements.}

\item{out.file}{Path to the file where MELTING will write the results.}

\item{primer_conc}{Primer concentration.}

\item{na_salt_conc}{Sodium ion concentration.}

\item{mg_salt_conc}{Magensium ion concentration.}

\item{k_salt_conc}{Potassium ion concentration.}

\item{tris_salt_conc}{Tris buffer concentration.}

\item{ID}{identifiers of the input primers}
}
\value{
Melting temperature data frame.
}
\description{
Computes the melting temperature for the input primers.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plots_comparison.R
\name{comparison.stats.raw}
\alias{comparison.stats.raw}
\title{Computation of Raw Stats for Primer Comparison}
\usage{
comparison.stats.raw(primer.data, template.data)
}
\arguments{
\item{primer.data}{List with primer data frames.}

\item{template.data}{List with template data frames.}
}
\value{
Raw statistics for primer comparison.
}
\description{
Computes raw stats for primer comparison.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/filters.R
\name{filter.primer.set.opti}
\alias{filter.primer.set.opti}
\title{Filtering of Primers}
\usage{
filter.primer.set.opti(
  primer.df,
  sample,
  template.df,
  settings,
  mode.directionality,
  required.cvg,
  results.loc,
  target.temps
)
}
\arguments{
\item{primer.df}{Primer data frame.}

\item{sample}{Name of the current template sample.}

\item{template.df}{Template data frame.}

\item{settings}{List with settings for the constraints to be used for filtering.}

\item{mode.directionality}{Primer direction.}

\item{required.cvg}{Required ratio of covered templates.
If \code{required.cvg} is set to 0, the constraints are not relaxed.}

\item{results.loc}{Path to a directory where the results should be written.}

\item{target.temps}{Target melting temperature of the primers in Celsius.
This argument is only required if we try to match the melting temperatures of another primer set,
e.g. when first optimizing forward and then optimizing reverse primers.}
}
\value{
The filtered primer data frame with respect to \code{required.cvg}.
}
\description{
Filters a primer set during the optimization procedure.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/IO_view.R
\name{view.cvg.primers}
\alias{view.cvg.primers}
\title{View the Evaluated Primers.}
\usage{
view.cvg.primers(
  primer.df,
  template.df,
  mode.directionality,
  view.cvg.individual = c("active", "inactive"),
  for.shiny = TRUE
)
}
\arguments{
\item{primer.df}{A \code{Primers} object.}

\item{template.df}{A \code{Templates} object.}

\item{mode.directionality}{The direction of the primers.}

\item{view.cvg.individual}{Whether information on individual coverage events should be retrained.}

\item{for.shiny}{Whether the table is intended for Shiny (HTML) or not.}
}
\value{
A formatted primer table.
}
\description{
Creates a formatted primers table.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/templates.R
\name{update.binding.ranges.by.conservation}
\alias{update.binding.ranges.by.conservation}
\title{Updates Binding Region in the Alignment by conservation.}
\usage{
\method{update}{binding.ranges.by.conservation}(
  template.df,
  bins,
  entropy.df,
  gap.char = "-",
  win.len = 30,
  direction = c("fw", "rev")
)
}
\arguments{
\item{template.df}{A \code{Templates} object.}

\item{bins}{A list with \code{DNAbin} alignments, one for each group of template sequences.}

\item{entropy.df}{A data frame with entropy information.}

\item{gap.char}{The gap character for alignments.}

\item{win.len}{The desired length of the new binding region.}

\item{direction}{The direction for which the binding range shall be adjusted.}
}
\value{
A \code{Templates} object with modified binding regions.
}
\description{
Updates Binding Region in the Alignment by conservation.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/templates.R
\name{retrieve.leader.region}
\alias{retrieve.leader.region}
\title{Retrieval of Binding Regions}
\usage{
retrieve.leader.region(
  template.df,
  direction = c("fw", "rev"),
  start,
  end,
  gap.char,
  init.from.leader
)
}
\arguments{
\item{template.df}{Template data frame.}

\item{direction}{Identify forward and reverse.}

\item{start}{Start positions.}

\item{end}{End positions.}

\item{gap.char}{The character for gaps in alignments.}

\item{init.from.leader}{Whether the binding regions are initialized
from leader sequences.}
}
\value{
Data frame with information on allowed binding regions.
}
\description{
Retrives information about individual binding regions.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/templates.R, R/primers.R, R/Plots.R,
%   R/plots_constraints.R, R/plots_coverage.R
\name{Plots}
\alias{Plots}
\alias{plot_conservation}
\alias{plot_primer_binding_regions}
\alias{plot_constraint}
\alias{plot_constraint_fulfillment}
\alias{plot_cvg_constraints}
\alias{plot_constraint_deviation}
\alias{plot_cvg_vs_set_size}
\alias{plot_penalty_vs_set_size}
\alias{plot_primer_subsets}
\alias{plot_primer}
\alias{plot_template_cvg}
\alias{plot_primer_cvg}
\title{Plotting Functions.}
\usage{
plot_conservation(entropy.df, alignments, template.df, gap.char = "-")

plot_primer_binding_regions(
  primers,
  templates,
  direction = c("both", "fw", "rev"),
  group = NULL,
  relation = c("fw", "rev"),
  region.names = c("Binding region", "Amplification region"),
  ...
)

plot_constraint(
  primers,
  settings,
  active.constraints = names(constraints(settings)),
  ...
)

plot_constraint_fulfillment(
  primers,
  settings,
  active.constraints = names(constraints(settings)),
  plot.p.vals = FALSE,
  ...
)

plot_cvg_constraints(
  primers,
  settings,
  active.constraints = names(cvg_constraints(settings)),
  ...
)

plot_constraint_deviation(
  primers,
  settings,
  active.constraints = names(constraints(settings)),
  ...
)

plot_cvg_vs_set_size(
  primer.data,
  template.data,
  show.labels = TRUE,
  highlight.set = NULL
)

plot_penalty_vs_set_size(
  primer.data,
  settings,
  active.constraints = names(constraints(settings)),
  alpha = 0
)

plot_primer_subsets(primer.subsets, template.df, required.cvg = 1)

plot_primer(
  primer.df,
  template.df,
  identifier = NULL,
  relation = c("fw", "rev"),
  region.names = c("Binding region", "Amplification region")
)

plot_template_cvg(primers, templates, per.mismatch = FALSE, ...)

plot_primer_cvg(primers, templates, per.mismatch = FALSE, ...)
}
\arguments{
\item{entropy.df}{A data frame with entropies.
Each row gives the entropies of a group of related template
sequences for all columns of the alignment.}

\item{alignments}{A list with \code{DNABin} alignment objects
corresponding to the groups (rows) in the alignment.}

\item{template.df}{An object of class \code{Templates} containing the
template sequences.}

\item{gap.char}{The gap char in the alignments. By default,
\code{gap.char} is set to "-".}

\item{primers}{Either a single \code{Primers} object with evaluated primer coverage
or a list containing such \code{Primers} objects.}

\item{templates}{If \code{primers} is a \code{Primers} object, \code{templates} should be a \code{Templates} object.
If \code{primers} is a list, then \code{templates} should be a list of \code{Templates} objects.}

\item{direction}{The directionality of primers to be plotted. This can either
be "both" to plot primers of both directions (the default), "fw" to plot
only forward primers, or "rev" to plot only reverse primers.}

\item{group}{Optional identifiers of template groups for which binding events should
be determined. By default, \code{group} is set to \code{NULL} such that
all templates are considered.}

\item{relation}{Whether binding positions are computed relative to forward ("fw")
or reverse ("rev") binding regions. The default is "fw".}

\item{region.names}{Character vector of length 2 providing the names
of the binding and amplification region.}

\item{...}{Optional arguments \code{groups} (a character vector of groups to be plotted when \code{primers} is a single primer set), 
\code{highlight.set} (the identifier of a primer set to be highlighted when \code{primers} is a list), 
\code{ncol} (a numeric indicating the number of facet columns if \code{primers} is a list),
\code{deviation.per.primer} (a boolean indicating whether constraint deviations
should be plotted per primer rather than per constraint if \code{primers} is a list)}

\item{settings}{An object of class \code{DesignSettings} containing
the constraints to be considered.}

\item{active.constraints}{A character vector containing the identifiers
to be considered for plotting. By default, \code{active.constraints} is
\code{NULL} such that all computed constraints found in \code{settings} are plotted.}

\item{plot.p.vals}{An optional logical argument indicating whether
p-values computed via \code{\link{primer_significance}} should be annotated in the plot. The default is \code{FALSE}.}

\item{primer.data}{List with objects of class \code{Primers} containing
the primer sets that are to be compared.}

\item{template.data}{List with objects of class \code{Templates} containing
the templates corresponding to \code{primer.data}.}

\item{show.labels}{Whether the identifiers of the primer sets
should be annotated in the plot. The default is \code{TRUE}.}

\item{highlight.set}{A character vector providing the identifiers
of primer sets to highlight. By default, \code{highlight.set} is \code{NULL}
such that no highlighting takes place.}

\item{alpha}{A numeric in the range [0,1] defining the trade-off between
the maximal deviation of a constraint (large \code{alpha}) and
all constraint deviations (large \code{alpha}).
By default, \code{alpha} is set to 0 such that the absolute
deviation across all constraints is considered.}

\item{primer.subsets}{A list with optimal primer subsets, each of class \code{Primers}. The \code{k}-th list entry should correspond to an object of class \code{Primers}
representing the primer subset of size \code{k} whose coverage ratio
is the largest among all possible subsets of size \code{k}.}

\item{required.cvg}{The required coverage ratio.
The default is 100\%; this value is plotted as a horizontal line.}

\item{primer.df}{An object of class \code{Primers} containing
primers with evaluated primer coverage.}

\item{identifier}{Identifiers of primers that are to be considered.
If \code{identifier} is set to \code{NULL} (the default), all primers are considered.}

\item{per.mismatch}{A logical specifying whether the visualization should be stratified
according to the allowed number of mismatches. By default,
\code{per.mismatch} is set to \code{FALSE} such that the overall coverage
is plotted.}
}
\value{
\code{plot_conseration} returns a plot showing the degree of sequence conservation in the templates.

\code{plot_primer_binding_regions} returns a plot of the primer binding regions.

\code{plot_constraint} returns a plot showing the distribution of primer properties.

\code{plot_constraint_fulfillment} returns a plot indicating the constraints that are fulfilled by the input primers.

\code{plot_cvg_constraints} returns a plot showing the distribution of the coverage constraint values.

\code{plot_constraint_deviation} returns a plot showing the deviations of the primer properties from the target constraints.

\code{plot_cvg_vs_set_size} returns a plot of coverage vs set size.

\code{plot_penalty_vs_set_size} returns a plot of constraint penalties
vs primer set sizes.

\code{plot_primer_subsets} plots the coverages of the primer subsets provided via \code{primer.subsets}.

\code{plot_primer} plots the primer binding sites in the templates.

\code{plot_template_cvg} creates a plot showing the number of covered template sequences.

\code{plot_primer_cvg} creates a plot showing the coverage of individual primers.
}
\description{
\describe{
\item{\code{plot_cvg_vs_set_size}}{Plots the coverage ratios of the input primer sets 
against the size of the sets.}
\item{\code{plot_penalty_vs_set_size}}{Plots the penalties of the input primer sets 
against the number of primers contained in each set.
The penalties are computed using \code{\link{score_primers}}
where more information is provided on how to set \code{alpha}.}
\item{\code{plot_primer_subsets}}{Visualizes the coverage of optimized primer subsets.}
\item{\code{plot_primer}}{Visualizes the binding positions of every primer relative to
the target binding region in the corresponding template sequences.}
\item{\code{plot_template_cvg}}{Creates a bar plot visualizing the covered templates.}
\item{\code{plot_primer_cvg}}{Shows which groups of templates are covered by individual primers.}
\item{\code{plot_constraint}}{Shows the distribution of the primer properties.
The current constraint settings are indicated with dashed lines.}
\item{\code{plot_constraint_fulfillment}}{Visualizes which primers pass the constraints 
and which primers break the constraints}
\item{\code{plot_cvg_constraints}}{Plots the distribution of the coverage constraint values.}
\item{\code{plot_constraint_deviation}}{Plots the deviation of primer properties from the target ranges.}
\item{\code{plot_primer_binding_regions}}{Visualizes the number of binding events of the primers
with respect to the allowed binding regions in the templates.}
\item{\code{plot_conservation}}{Plots the template sequence conservation (range [0,1]) according to
the Shannon entropy of the sequences.}
}
}
\details{
The deviations for \code{plot_constraint_deviation} are computed in the following way. Let the
minimum and maximum allowed constraint values be given by
the interval \eqn{[s, e]} and the observed value be \eqn{p}. Then,
if \eqn{p < s}, we output \eqn{-p/|s|}, if \eqn{p > e} we output \eqn{p/|e|},
and otherwise, i.e. if \eqn{s <= p <= e}, we output 0.

The \code{primer.subsets} argument for \code{plot_primer_subsets} can be computed using
\code{\link{subset_primer_set}}. 
The line plot indicates the ratio of covered templates when considering
all primers in a primer set of a given size. The bar plots indicate
the coverage ratios of individual primers in a set. The target coverage
ratio is indicated by a horizontal line. Bars exceeding the target ratio
possibly indicate the existence of redundant coverage events.
}
\note{
Computing the conservation scores for using \code{plot_conservation} requires
the MAFFT software for multiple alignments (http://mafft.cbrc.jp/alignment/software/).
}
\examples{

data(Ippolito)
# Select binding regions for every group of templates and plot:
template.df <- select_regions_by_conservation(template.df, win.len = 30)
if (length(template.df) != 0) {
     p1 <- plot_conservation(attr(template.df, "entropies"), 
                             attr(template.df, "alignments"), template.df)
}
# Select binding regions for all templates and plot:
data(Ippolito)
template.df <- select_regions_by_conservation(template.df, by.group = FALSE)
if (length(template.df) != 0) {
     p2 <- plot_conservation(attr(template.df, "entropies"), 
                             attr(template.df, "alignments"), template.df)
} 

# Primer binding regions of a single primer set
data(Ippolito)
p <- plot_primer_binding_regions(primer.df, template.df)
# Primer binding regions of multiple primer sets
data(Comparison)
p.comp <- plot_primer_binding_regions(primer.data[1:3], template.data[1:3])

# Plot histogram of constraints for a single primer set
data(Ippolito)
p <- plot_constraint(primer.df, settings, 
                     active.constraints = c("gc_clamp", "gc_ratio"))
# Compare constraints across multiple primer sets
data(Comparison)
p.cmp <- plot_constraint(primer.data[1:3], settings, 
                         active.constraints = c("gc_clamp", "gc_ratio"))

# Plot fulfillment for a single primer set:
data(Ippolito)
p <- plot_constraint_fulfillment(primer.df, settings)
# Plot fulfillment for multiple primer sets:
data(Comparison)
p.cmp <- plot_constraint_fulfillment(primer.data[1:5], settings)

# Plot coverage constraints of a single primer set
data(Ippolito)
p <- plot_cvg_constraints(primer.df, settings)
# Plot coverage constraints for mulitple primer sets
data(Comparison)
p.cmp <- plot_cvg_constraints(primer.data[1:2], settings)

# Deviations for a single primer set
data(Ippolito)
p.dev <- plot_constraint_deviation(primer.df, settings)
# Deviations for multiple primer sets
data(Comparison)
p.dev.cmp <- plot_constraint_deviation(primer.data, settings)

# Plot coverage vs primer set size
data(Comparison)
p <- plot_cvg_vs_set_size(primer.data, template.data)

# Plot penalties vs number of primers
data(Comparison)
p <- plot_penalty_vs_set_size(primer.data, settings)

# Plot the coverage of optimal primer subsets
data(Ippolito)
primer.subsets <- subset_primer_set(primer.df, template.df, k = 3)
p <- plot_primer_subsets(primer.subsets, template.df)

# Plot of individual primer binding positions
data(Ippolito)
p <- plot_primer(primer.df[1,], template.df[1:30,])

# Visualize the template coverage of a single primer set
data(Ippolito)
p.cvg <- plot_template_cvg(primer.df, template.df)
# Stratify by allowed mismatches:
p.mm.cvg <- plot_template_cvg(primer.df, template.df, per.mismatch = TRUE)
# Compare the coverage of multiple primer sets
data(Comparison)
p.cmp.cvg <- plot_template_cvg(primer.data[1:2], template.data[1:2])
# Stratify by allowed mismatches:
p.cmp.cvg.mm <- plot_template_cvg(primer.data[1:2], template.data[1:2], 
                                 per.mismatch = TRUE)

# Plot expected coverage per primer
data(Ippolito)
p.cvg <- plot_primer_cvg(primer.df, template.df)
# Plot coverage stratified by allowed mismatches:
p.cvg.mm <- plot_primer_cvg(primer.df, template.df, per.mismatch = TRUE)
# Plot coverage of multiple primer sets
data(Comparison)
p.cvg.cmp <- plot_primer_cvg(primer.data[1:3], template.data[1:3])
}
\concept{templates}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/zzz.R
\name{check.tool.installation}
\alias{check.tool.installation}
\title{Check Tool Installation}
\usage{
check.tool.installation(frontend = FALSE)
}
\arguments{
\item{frontend}{Whether tool installation shall be checked for the frontend.
If \code{TRUE}, dependencies that are required only by the frontend are considered additionally.}
}
\value{
\code{TRUE} for each installed tool, \code{FALSE} otherwise.
}
\description{
Checks whether all required tools are installed.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/con_template_secondary_structures.R
\name{plot.Delta.DeltaG}
\alias{plot.Delta.DeltaG}
\title{Delta DeltaG Plot}
\usage{
\method{plot}{Delta.DeltaG}(constrained.foldings, stratify = FALSE)
}
\arguments{
\item{constrained.foldings}{Data frame with info from constrained foldings.}

\item{stratify}{Stratify according to template groups?}
}
\value{
Plot of Delta DeltaG.
}
\description{
Plots the difference between the free energy of constrained 
and unconstrained foldings.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/con_primer_coverage.R
\name{compute.unique.covered.idx}
\alias{compute.unique.covered.idx}
\title{Unique Coverage Indices
Computes the indices of templates that are covered uniquely 
covered by an individual primer.}
\usage{
compute.unique.covered.idx(primer.df, template.df)
}
\arguments{
\item{primer.df}{Primer data frame.}

\item{template.df}{Template data frame.}
}
\value{
Index of templates uniquely covered per input primer.
}
\description{
Unique Coverage Indices
Computes the indices of templates that are covered uniquely 
covered by an individual primer.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/filters.R
\name{set.new.limits}
\alias{set.new.limits}
\title{Relaxation of Constraint Limits.}
\usage{
set.new.limits(
  cur.limits,
  initial.limits,
  initial.constraints,
  con.name = NULL
)
}
\arguments{
\item{cur.limits}{List with current constraint settings.}

\item{initial.limits}{List with initial coonstraint limits.}

\item{initial.constraints}{List with initial constraint settings before relaxing.}

\item{con.name}{The constraint for which the settings are to be changed.}
}
\value{
A list with relaxed constraint limits.
}
\description{
Relaxes the constraint limits by moving according to the difference
between the \code{initial.limits} and \code{initial.constraints}.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/primers.R
\name{get_covered.vanilla}
\alias{get_covered.vanilla}
\title{Determination of the Covered Sequences.}
\usage{
get_covered.vanilla(primers, template.df, mode.directionality = NULL)
}
\arguments{
\item{primers}{A \code{Primers} object containing the primers
for which the coverage should be evaluated.}

\item{template.df}{A \code{Templates} object containing
the template sequences corresponding to \code{primers}.}

\item{mode.directionality}{If \code{mode.directionality} is provided,
the coverage of templates is computed for a specific direction of primers.
Either "fw" (forward coverage only), "rev" (reverse coverage only), or "both" for both directions. If \code{mode.directionality} is not provided
the direction is determined by the input primers.}
}
\value{
The IDs of all covered templates.
}
\description{
Determines the covered template sequences given by \code{template.df} 
that are covered by the primers given by \code{primers}.
}
\details{
The manner in which the coverage ratio is evaluated depends on 
the directionality of the input primers.
If either only forward or reverse primers are inputted, the individual
coverage of each primer is used to determine the overall coverage. If, however,
forward and reverse primers are inputted at the same time, 
the coverage is defined by the intersection of binding events from both,
forward and reverse primers.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ambiguity.R
\name{comp}
\alias{comp}
\title{Sequence complement}
\usage{
comp(seq, forceToLower = TRUE, ambiguous = FALSE)
}
\arguments{
\item{seq}{Input char vector.}

\item{forceToLower}{if TRUE the input is transformed to lower case.}

\item{ambiguous}{if TRUE ambiguous IUPAC nucleotides are complemented.}
}
\value{
The complement of \code{seq}.
}
\description{
Complements the input sequence (re-write of seqinr comp function for gap support)
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plots_constraints.R
\name{plot_constraint,list-method}
\alias{plot_constraint,list-method}
\title{Boxplot for Comparing Constraints.}
\usage{
\S4method{plot_constraint}{list}(
  primers,
  settings,
  active.constraints,
  highlight.set = NULL,
  nfacets = NULL
)
}
\arguments{
\item{primers}{List with evaluated objects of class \code{Primers}. 
Each list element corresponds to a single primer set.}

\item{settings}{A \code{DesignSettings} object containing
the constraints to be plotted.}

\item{active.constraints}{The names of the constraints to be plotted.}

\item{highlight.set}{Identifiers of primer sets to be highlighted.}

\item{nfacets}{A numeric providing the number of facet columns to show.
By default \code{nfacets} is \code{NULL} such that the number of 
facet columns is chosen automatically.}

\item{constraint.settings}{List with settings for each constraint.}
}
\value{
Boxplot comparing the values of the properties specified 
by \code{constraints}.
}
\description{
Creates a boxplot visualizing the physicochemical properties
of multiple primer sets.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/check_stop_codons.R
\name{highlight.mismatch}
\alias{highlight.mismatch}
\title{Highlight mismatches}
\usage{
highlight.mismatch(seq, mm.seq)
}
\arguments{
\item{seq}{character vector of the original sequence}

\item{mm.seq}{character vector of the mutated sequence}
}
\value{
A list highlighting the mutations and additional information 
(mutation type, number, etc.)
}
\description{
Collects information on the mutations present in the input and highlights the
mutations in the sequence.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/filters.R
\name{store.filtering.sets}
\alias{store.filtering.sets}
\title{Writes Filtering Data Sets to Disk.}
\usage{
store.filtering.sets(
  filtered.df,
  excluded.df,
  results.loc,
  tag = "",
  stat.df = NULL,
  settings = NULL
)
}
\arguments{
\item{filtered.df}{A filtered \code{Primers} set.}

\item{excluded.df}{A set of \code{Primers} that were excluded.}

\item{results.loc}{The location where to store the data.}

\item{tag}{A tag for the output files.}

\item{stat.df}{Data frame with statistics of the filtering procedure.}

\item{settings}{A \code{DesignSettings} object.}
}
\value{
No return value, writes output to disk.
}
\description{
Writes Filtering Data Sets to Disk.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plots_comparison.R
\name{prepare.constraint.plot}
\alias{prepare.constraint.plot}
\title{Preparation of Comparison Plot for Evaluation.}
\usage{
prepare.constraint.plot(primer.data, constraint.settings, plot.p.vals = FALSE)
}
\arguments{
\item{primer.data}{List with objects of class \code{Primers}. Each 
list entry corresponds to a single primer set.}

\item{constraint.settings}{List with settings for each constraint.
If \code{NULL} (the default), use the available evaluation results
in \code{primer.data}.}

\item{plot.p.vals}{Whether p-values from Fisher's exact test
should be annotated for every primer set.}
}
\value{
Plot indicating the ratio of primers
fulfilling the constraints specified in \code{constraint.settings}
for each primer set in \code{primer.data}.
}
\description{
Preparation of Comparison Plot for Evaluation.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/con_dimerization.R
\name{view.dimer.df}
\alias{view.dimer.df}
\title{Formatted dimerization data.}
\usage{
view.dimer.df(dimers, type = c("Self", "Cross"))
}
\arguments{
\item{dimers}{Dimerization data frame.}

\item{type}{Type of dimerization.}
}
\value{
A data frame whose columns are formatted in a user-readable way.
}
\description{
Format a dimerization data frame for frontend output.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/primers.R
\name{restriction_ali}
\alias{restriction_ali}
\title{Identification of Badly Fitting Regions.}
\usage{
restriction_ali(primer.seqs, template.seqs, search.hits)
}
\arguments{
\item{primer.seqs}{Primer sequences.}

\item{template.seqs}{Template sequences.}

\item{search.hits}{Template substrings that agree well
with the input primers.}
}
\value{
A list with putative restriction sites for every primer.
}
\description{
Identify regions in the templates where the primers are
not very complementary. These regions indicate possible
restriction enzyme adapters.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Comparison.R, R/Data.R, R/Ippolito.R,
%   R/RefCoverage.R, R/Tiller.R
\docType{data}
\name{Data}
\alias{Data}
\alias{primer.data}
\alias{template.data}
\alias{template.df}
\alias{primer.df}
\alias{settings}
\alias{RefCoverage}
\alias{feature.matrix}
\alias{ref.data}
\alias{Tiller}
\alias{tiller.template.df}
\alias{tiller.primer.df}
\alias{tiller.settings}
\title{Data Sets.}
\format{
For the \code{RefCoverage} data set, the \code{feature.matrix} data frame contains the properties
of the primer set from Tiller et al. as well as a primer set
that was designed by openPrimeR. The column \code{Experimental_Coverage}
indicates the experimentally determined coverage, while the other
columns relate to properties of the primers that were computed
with openPrimeR. 
The \code{ref.data} list contains the raw experimental coverage
of individual primers from the primer sets from Tiller and openPrimeR, which
both target templates from the IGH locus.
The rows of the data frames indicate primers and the columns indicate
IGH templates for which experimental coverage was determined.
The cell entries are hex codes. Each hex code represents a color 
indicating a certain experimental coverage status. 
Hex codes representing red shades indicate no or little amplification, 
while hex codes for green shades indicate high yields.

For the \code{Ippolito} data set, 
\code{primer.df} provides a \code{Primers} object containing
the evaluated set of primers from Tiller et al.
\code{template.df} provides
a \code{Templates} object containing functional, human IGHV
templates for, and \code{settings} provides a 
\code{DesignSettings} object providing the used analysis settings.

For the \code{Comparison} data set, \code{primer.data} and \code{template.data} are
lists of \code{Primers} and \code{Templates} objects, respectively.

For the \code{Tiller} data set, \code{tiller.primer.df} 
provides a \code{Primers} object, \code{tiller.template.df} provides 
the corresponding \code{Templates} object, and \code{tiller.settings} 
provides the \code{DesignSettings} object that was used
for evaluating \code{tiller.primer.df}.
}
\usage{
data(Comparison)

data(Ippolito)

data(RefCoverage)

data(Tiller)
}
\description{
\describe{
\item{\code{Ippolito}}{IGHV primer data from Ippolito et al.}
\item{\code{Tiller}}{IGHV primer data from Tiller et al.}
\item{\code{Comparison}}{Evaluated primer sets targeting the 
functional human IGH immunoglobulin
genes. The sets were generated using the default evaluation settings
of openPrimeR. The primer sets were gathered from
IMGT and the literature.}
\item{\code{RefCoverage}}{Experimental results of multiplex PCR.}
}
}
\examples{

# Load the comparison data
data(Comparison)
# Explore the first entry of the primer and template data:
primer.data[[1]]
template.data[[1]]
# Summarize the primer properties:
get_comparison_table(template.data, primer.data)

# Load the data from Ippolito et al.
data(Ippolito)
primer.df
template.df
constraints(settings)

# Load experimental PCR results
data(RefCoverage)

# Load the data from Tiller et al.
data(Tiller)
tiller.primer.df
tiller.template.df
constraints(tiller.settings)
}
\references{
IMGT®, the international ImMunoGeneTics information system® http://www.imgt.org (founder and director: Marie-Paule Lefranc, Montpellier, France).

Ippolito GC, Hoi KH, Reddy ST, Carroll SM, Ge X, Rogosch T, 
Zemlin M, Shultz LD, Ellington AD, VanDenBerg CL, Georgiou G. 2012. 
Antibody Repertoires in Humanized NOD-scid-IL2R gamma null Mice and Human B
Cells Reveals Human-Like Diversification and Tolerance Checkpoints 
in the Mouse. PLoS One 7:e35497.

Tiller, Thomas, et al. "Efficient generation of monoclonal antibodies from single human B cells by single cell RT-PCR and expression vector cloning." Journal of immunological methods 329.1 (2008): 112-124.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/initialize_primers_tree.R
\name{prefilter.primer.candidates}
\alias{prefilter.primer.candidates}
\title{Identication of Short Primers.}
\usage{
prefilter.primer.candidates(primer.candidates, min.len)
}
\arguments{
\item{primer.candidates}{Primer alignment.}

\item{min.len}{Minimal primer length.}
}
\value{
The index of proposed primers that are shorter than \code{min.len}.
}
\description{
Identify initial primers that are too short.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/templates.R
\name{read.leaders}
\alias{read.leaders}
\title{Read Individual Binding Regions}
\usage{
read.leaders(
  fasta.file,
  direction = c("fw", "rev"),
  rm.keywords = NULL,
  gap.character
)
}
\arguments{
\item{fasta.file}{Path to a FASTA file with binding regions.}

\item{direction}{String identifying whether the FASTA file
contains information pertaining to the binding region of forward
or reverse primers.}

\item{rm.keywords}{A vector of keywords that are used to remove templates whose headers contain any of the keywords.}

\item{gap.character}{The character for indicating gaps in sequences.}
}
\value{
A data frame with individual binding regions.
}
\description{
Reads individual binding regions into a data frame.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/initialize_primers.R
\name{get.primer.identifier.string}
\alias{get.primer.identifier.string}
\title{Primer Identifier Creation.}
\usage{
get.primer.identifier.string(
  sample,
  seq.IDs,
  seq.identifier,
  all.starts,
  all.ends,
  identifier,
  seq.primers
)
}
\arguments{
\item{sample}{Sample name of the templates.}

\item{seq.IDs}{Identifiers of the templates.}

\item{all.starts}{Primer positions (start).}

\item{all.ends}{Primer positions (end).}

\item{identifier}{Direction keyword.}

\item{seq.primers}{The primer sequences as strings.}

\item{seq.identifiers}{The index of the seq.}
}
\value{
Identifiers for each primer.
}
\description{
Creates identifiers for generated primers.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/helper_functions.R
\name{dir.copy}
\alias{dir.copy}
\title{Copy Directories.}
\usage{
dir.copy(src.dir, dest.dir, overwrite)
}
\arguments{
\item{src.dir}{The directory to be copied.}

\item{dest.dir}{The target directory.}

\item{overwrite}{Overwrite existing files in \code{dest.dir}.}
}
\value{
TRUE if copying was successful, FALSE otherwise.
}
\description{
Copies a directory to another location.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/con_dimerization.R
\name{compute.all.self.dimers}
\alias{compute.all.self.dimers}
\title{Self dimers}
\usage{
compute.all.self.dimers(
  primer.df,
  primer_conc,
  na_salt_conc,
  mg_salt_conc,
  k_salt_conc,
  tris_salt_conc,
  annealing.temp,
  for.shiny = FALSE,
  no.structures = FALSE
)
}
\arguments{
\item{primer.df}{Input primer data frame.}

\item{primer_conc}{Primer concentration.}

\item{na_salt_conc}{Sodium ion concentration.}

\item{mg_salt_conc}{Magensium ion concentration.}

\item{k_salt_conc}{Potassium ion concentration.}

\item{tris_salt_conc}{Tris buffer concentration.}

\item{annealing.temp}{The PCR annealing temperature.}

\item{for.shiny}{Whether the output is to be formatted for HTML.}

\item{no.structures}{Whether dimerization structures shall be outputted.}
}
\value{
Data frame with thermodynamic information on all self dimers.
}
\description{
Computes all possible self dimers for the primers in the input data frame.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plots_coverage.R
\name{plot_primer_cvg,list,list-method}
\alias{plot_primer_cvg,list,list-method}
\title{Plot Multiple Primer Coverages.}
\usage{
\S4method{plot_primer_cvg}{list,list}(primers, templates, per.mismatch = FALSE)
}
\arguments{
\item{primers}{List with \code{Primers} objects.}

\item{templates}{List with \code{Templates} objects.}

\item{per.mismatch}{Whether the coverage should
be broken down for individual settings of allowed mismatches.}
}
\value{
A bar plot showing the coverage of individual primers.
}
\description{
Plots the coverage of individual primers for multiple sets.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/con_gc_ratio.R
\name{compute.gc.ratio}
\alias{compute.gc.ratio}
\title{GC ratio}
\usage{
compute.gc.ratio(x)
}
\arguments{
\item{x}{Input sequence.}
}
\value{
The fraction of G/Cs in \code{x}.
}
\description{
Computes the ratio of G/Cs in a sequence.
}
\details{
In case of ambiguities, the mean GC ratio of all possible sequences is computed.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plots_constraints.R
\name{plot.dimer.dist}
\alias{plot.dimer.dist}
\title{Plot Dimer DeltaG}
\usage{
\method{plot}{dimer.dist}(dimer.data, deltaG.cutoff)
}
\arguments{
\item{dimer.data}{Data frame with dimerization information.}

\item{deltaG.cutoff}{Free energy cutoff for dimerization.}
}
\value{
A plot of dimerization free energies.
}
\description{
Plot the distribution of dimerization free energies.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/optimization_ILP.R
\name{J.cvg}
\alias{J.cvg}
\title{Template Coverage.}
\usage{
J.cvg(cvg.matrix)
}
\arguments{
\item{cvg.matrix}{Binary matrix of covering events.}
}
\value{
A list with covering primers for every template.
}
\description{
Determines the indices of covering primers for every template.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/optimization_ILP.R
\name{get.redundant.cols}
\alias{get.redundant.cols}
\title{Identification of Redudant Primers.}
\usage{
get.redundant.cols(cvg.matrix)
}
\arguments{
\item{cvg.matrix}{Binary matrix of coverage events.}
}
\value{
TRUE for redundant primers, FALSE otherwise.
}
\description{
Identifies primers that are redundant.
}
\details{
Redundant primers do not reduce the coverage when removed.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/con_melting_temperature.R
\name{convert.temperature}
\alias{convert.temperature}
\title{Conversion between Celsius and Kelvin}
\usage{
convert.temperature(temp, temp.scale = c("K", "C"))
}
\arguments{
\item{temp}{The input temperature.}

\item{temp.scale}{The desired unit of the output temperature.}
}
\value{
Transforms the input temperature to the specified \code{temp.scale}.
}
\description{
Converts the input from Kelvin to Celsius or from Celsius to Kelvin.
}
\details{
If \code{temp.scale} is 'K', \code{T_m} is transformed from Celsius
to Kelvin. If \code{temp.scale} is 'C', \code{T_m} is transformed from
Kelvin to Celsius. The default is to transform from Celsius to Kelvin.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/con_dimerization.R
\name{align.structures}
\alias{align.structures}
\title{Formatting of Dimerization Structures.}
\usage{
align.structures(structs)
}
\arguments{
\item{s1}{strutcs A character vector, where a block of 4 elements
contains: sequence 1 (with removed overlaps), part of sequence 1
overlappign with sequence 2, part of sequence 2 
overapping with sequence 1, and sequence 2 (with removed overlaps).}
}
\value{
A list of two elements givin the conformation of the first
and the second sequence, respectively.
}
\description{
Formats the given dimerization structures nicely.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plots_coverage.R
\name{primer.binding.regions.data}
\alias{primer.binding.regions.data}
\title{Primer Binding Region Data}
\usage{
primer.binding.regions.data(
  primer.df,
  template.df,
  direction = c("both", "fw", "rev"),
  group = NULL,
  relation = c("fw", "rev")
)
}
\arguments{
\item{primer.df}{Primer data frame.}

\item{template.df}{Template data frame.}

\item{direction}{Primer direction}

\item{group}{The groups for which binding data shall be retrieved.}

\item{relation}{Binding region data relative to forward/reverse binding region?}
}
\value{
Data frame with primer binding data.
}
\description{
Collects all data concerning primer binding regions.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/con_template_secondary_structures.R
\name{optimize.template.binding.regions.dir}
\alias{optimize.template.binding.regions.dir}
\title{Optimization of Binding Regions}
\usage{
optimize.template.binding.regions.dir(
  template.df,
  annealing.temperature = NULL,
  primer.lengths,
  mode.directionality = c("fw", "rev", "both")
)
}
\arguments{
\item{template.df}{Template data frame.}

\item{annealing.temperature}{Temperature at which to compute secondary structures.}

\item{primer.lengths}{Target length of primers that are to be used.}

\item{mode.directionality}{Direction of primers.}
}
\value{
List with intervals indicating improved primer binding regions.
}
\description{
Optimizes the template binding regions.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ambiguity.R
\name{merge.select}
\alias{merge.select}
\title{Select merge indices}
\usage{
\method{merge}{select}(merge.idx)
}
\arguments{
\item{merge.idx}{list of lists containing the indices of possible merges}
}
\value{
The smallest number of possible merge operations as an index list.
}
\description{
Greedily identifies the smallest number of possible sequences merges that can be performed.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/optimization_ILP.R
\name{subset.ILP}
\alias{subset.ILP}
\title{Subset ILP Constructor}
\usage{
\method{subset}{ILP}(primer.df, template.df, k)
}
\arguments{
\item{primer.df}{Primer data frame to be subsetted.}

\item{template.df}{Template data frame.}

\item{k}{Required number of primers to be selected.}
}
\value{
An ILP for choosing the primer subset of size \code{k} with the largest coverage.
}
\description{
Constructs an ILP for selecting optimal primer subsets.
}
\details{
Here, "optimal" refers to a subset of a certain size that maximizes the coverage.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/templates.R
\name{read_templates_single}
\alias{read_templates_single}
\title{Input of a Single Template File.}
\usage{
read_templates_single(
  template.file,
  hdr.structure = NULL,
  delim = NULL,
  id.column = NULL,
  rm.keywords = NULL,
  remove.duplicates = FALSE,
  fw.region = c(1, 30),
  rev.region = c(1, 30),
  gap.character = "-",
  run = NULL
)
}
\arguments{
\item{template.file}{Path to a FASTA or CSV file containing the template sequences.}

\item{hdr.structure}{A character vector describing the information contained in the FASTA headers. In case that the headers of \code{fasta.file} contain
template group information, please include the keyword "GROUP" in
\code{hdr.structure}.}

\item{delim}{Delimiter for the information in the FASTA headers.}

\item{id.column}{Field in the header to be used as the identifier.}

\item{rm.keywords}{A vector of keywords that are used to remove templates whose headers contain any of the keywords.}

\item{remove.duplicates}{Whether duplicate sequence shall be removed.}

\item{fw.region}{The positional interval from the template 5' end specifying the 
binding sites for forward primers.}

\item{rev.region}{The positional interval from the template 3' end specifying
the binding sites for reverse primers.}

\item{gap.character}{The character in the input file representing gaps.
Gaps are automatically removed upon input.}

\item{run}{An identifier for the template sequences.}
}
\value{
An object of class \code{Templates}.
}
\description{
Read template sequences from a FASTA or CSV file.
}
\details{
When supplying a FASTA file with template sequences,
the input arguments \code{hdr.structure}, \code{delim}, \code{id.column}, 
\code{rm.keywords}, \code{remove.duplicates}, \code{fw.region}, 
\code{rev.region}, \code{gap.character}, and \code{run} are utilized.
Most importantly, \code{hdr.structure} and \code{delim} should
match the FASTA header structure.
When supplying a CSV file with template sequences,
the data are loaded without any modification and the CSV file
should represent an object of class \code{\link{Templates}}, which 
can be stored using the \code{\link{write_templates}} function.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/initialize_primers_tree.R
\name{ancestor_of}
\alias{ancestor_of}
\title{Tree Ancestry}
\usage{
ancestor_of(tree, ancestor.node, test.node)
}
\arguments{
\item{tree}{The phylogenetic tree to be tested.}

\item{ancestor.node}{A node to be checked for being an ancestor to \code{test.node}.}

\item{test.node}{Possible descendants of \code{ancestor.node}.}
}
\value{
TRUE, if \code{ancestor.node} is an ancestor to any node in \code{test.node}.
}
\description{
Checks whether \code{ancestor.node} is an ancestor to the nodes 
specified in \code{test.node}.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/con_primer_coverage.R
\name{get.covered.templates}
\alias{get.covered.templates}
\title{Covered Templates}
\usage{
get.covered.templates(Tm.set, template.df)
}
\arguments{
\item{Tm.set}{Primer data frame.}

\item{template.df}{Template data set.}
}
\value{
Index of templates that are covered by the primers in \code{Tm.set}.
}
\description{
Get the indices of covered templates.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/con_dimerization.R
\name{get.cross.dimers}
\alias{get.cross.dimers}
\title{Cross dimers}
\usage{
get.cross.dimers(
  primers.1,
  primers.2,
  ions,
  annealing.temp,
  check.idx = NULL,
  no.structures = FALSE,
  mode = c("symmetric", "asymmetric")
)
}
\arguments{
\item{primers.1}{Input primers.}

\item{primers.2}{Input primers.}

\item{ions}{Sodium-equivalent ionic concentration.}

\item{annealing.temp}{The PCR annealing temperature.}

\item{check.idx}{indices of primers for checking cross-dimerization}

\item{no.structures}{Whether to compute structures of dimers.}

\item{mode}{'symmetric', if \code{primers.1} and \code{primers.2}
carry the same information (i.e. fw-fw, rev-rev, fw-rev), 'asymetric' else.}
}
\value{
Data frame with potential cross dimers.
}
\description{
Computes all possible primer cross-dimers.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plots_constraints.R
\name{plot_cvg_constraints,Primers-method}
\alias{plot_cvg_constraints,Primers-method}
\title{Histogram of Coverage Constraints.}
\usage{
\S4method{plot_cvg_constraints}{Primers}(primers, settings, active.constraints)
}
\arguments{
\item{primers}{A \code{Primers} object.}

\item{settings}{A \code{DesignSettings} object.}

\item{active.constraints}{Names of coverage constraints to be plotted.}
}
\value{
A plot of coverage constraints.
}
\description{
Plots a histogram of coverage constraint values.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/templates.R
\name{read_templates_fasta}
\alias{read_templates_fasta}
\title{Input of Template Sequences.}
\usage{
read_templates_fasta(
  fasta.file,
  hdr.structure = NULL,
  delim = NULL,
  id.column = NULL,
  rm.keywords = NULL,
  remove.duplicates = FALSE,
  fw.region = c(1, 30),
  rev.region = c(1, 30),
  gap.character = "-",
  run = NULL
)
}
\arguments{
\item{fasta.file}{Path to a FASTA file containing the template sequences.}

\item{hdr.structure}{Names describing the information contained in the FASTA headers. In case that the headers of \code{fasta.file} contain
template group information, please include the keyword "GROUP" in
\code{hdr.structure}.}

\item{delim}{Delimiter for the information in the FASTA headers.}

\item{id.column}{Field in the header to be used as the identifier.}

\item{rm.keywords}{A vector of keywords that are used to remove templates whose headers contain any of the keywords.}

\item{remove.duplicates}{Whether duplicate sequence shall be removed.}

\item{fw.region}{The positional interval from the template 5' end specifying the 
binding sites for forward primers.}

\item{rev.region}{The positional interval from the template 3' end specifying
the binding sites for reverse primers.}

\item{gap.character}{The character in the input file representing gaps.
Gaps are automatically removed upon input.}

\item{run}{An identifier for the template sequences.}
}
\value{
An object of class \code{Templates}.
}
\description{
Read template sequences from a FASTA file.
}
\examples{
fasta.file <- system.file("extdata", "IMGT_data", "templates", 
              "Homo_sapiens_IGH_functional_exon.fasta", package = "openPrimeR")
hdr.structure <- c("ACCESSION", "GROUP", "SPECIES", "FUNCTION")
template.df <- read_templates(fasta.file, hdr.structure, "|", "GROUP")
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/constraints_eval.R
\name{check_cvg_constraints}
\alias{check_cvg_constraints}
\title{Evaluation of Coverage Constraints.}
\usage{
check_cvg_constraints(
  primer.df,
  template.df,
  settings,
  active.constraints = names(cvg_constraints(settings)),
  to.compute.constraints = active.constraints,
  for.shiny = FALSE,
  updateProgress = NULL
)
}
\arguments{
\item{primer.df}{A \code{Primers} object containing the primers
to be checked.}

\item{template.df}{A \code{Templates} object containing the 
template sequences corresponding to the primers.}

\item{settings}{A \code{DesignSettings} object containing the 
coverage constraints to be checked and their settings.}

\item{active.constraints}{Identifiers of constraints that are to be checked.}

\item{to.compute.constraints}{Constraints that are to be computed.}

\item{for.shiny}{Whether to format output for HTML.}

\item{updateProgress}{Progress callback function for shiny.}
}
\value{
A \code{Primers} object with with columns for each constraint in \code{active.constraints}.
}
\description{
Computes the biochemical properties specified in the
\code{settings} object and determines whether the primers
fulfill the required constraints.
}
\note{
Please note that some constraints can only be computed if additional software is installed,
please see \code{\link{DesignSettings}} for an overview.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plots_coverage.R
\name{plot_primer_cvg,Primers,Templates-method}
\alias{plot_primer_cvg,Primers,Templates-method}
\title{Plot Individual Primer Coverage.}
\usage{
\S4method{plot_primer_cvg}{Primers,Templates}(primers, templates, per.mismatch = FALSE, groups = NULL)
}
\arguments{
\item{per.mismatch}{Whether the coverage should
be broken down for individual settings of allowed mismatches.}

\item{p.df}{Primer data frame.}

\item{template.df}{Template data frame.}

\item{excluded.seqs}{Identifiers of templates that should not be considered.}
}
\value{
A bar plot showing the coverage of individual primers.
}
\description{
Shows which templates are covered by individual primers.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/initialize_primers_tree.R
\name{create.primer.ranges}
\alias{create.primer.ranges}
\title{Ranges for Initial Primers.}
\usage{
create.primer.ranges(
  end.position,
  p.lens,
  start.position,
  step.size = 1,
  groups = NULL
)
}
\arguments{
\item{end.position}{End positions of primers.}

\item{p.lens}{Desired primer lengths.}

\item{step.size}{A numeric giving the steps with which start
positions are cycled. Should be 1 for primer design (evaluate all positions)
and higher values can be used for windowing.}

\item{groups}{Character vector with group annotation.}

\item{start.posiion}{The start positions of primers.}
}
\value{
Data frame with ranges for initial primers.
}
\description{
Creates a data frame indicating primer starts and ends.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/settings.R
\name{get.cvg.constraint.settings}
\alias{get.cvg.constraint.settings}
\title{Gather all Coverage Constraints.}
\usage{
get.cvg.constraint.settings(
  allowed.stop.codons,
  allowed.efficiency,
  disallowed.mismatch.pos,
  allowed.anneal.deltaG,
  allowed.substitutions,
  allowed.coverage.model
)
}
\arguments{
\item{allowed.stop.codons}{Whether mismatch binding events inducing
stop codons in the amino acid sequence are allowed.}

\item{allowed.efficiency}{Min/max for primer efficiency.}

\item{disallowed.mismatch.pos}{The positions from the 3' terminal
end of primers where mismatches shall be prevented.}

\item{allowed.anneal.deltaG}{Maximal allowed free energy of template-primer annealing.}

\item{allowed.substitutions}{Whether mismatch binding events inducing substitutions
in the amino acid sequence are allowed.}
}
\value{
List with all coverage constraint settings.
}
\description{
Constructor for coverage constraint settings.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/primers.R
\name{update_primer_cvg}
\alias{update_primer_cvg}
\title{Updates the Primer Coverage.}
\usage{
update_primer_cvg(
  primer.df,
  template.df,
  allowed.mismatches,
  cvg.definition = c("constrained", "basic")
)
}
\arguments{
\item{primer.df}{A \code{Primers} object.}

\item{template.df}{A \code{Templates} object.}

\item{allowed.mismatches}{A numeric giving the maximal number of allowed.mismatches.}

\item{cvg.definition}{The definition of coverage to be used, either "constrained" or "basic".}
}
\value{
A primer data frame with modified coverage information.
}
\description{
Updates the most important columns in a primer data frame according to the selected
coverage definition. Only coverage events with less or equal than the allowed number of
mismatches according to the selected coverage definition will be retained.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plots_coverage.R
\name{get_template_cvg_data}
\alias{get_template_cvg_data}
\title{Retrieval of Template Coverage Data.}
\usage{
get_template_cvg_data(primer.df, template.df)
}
\arguments{
\item{primer.df}{A \code{Primers} object.}

\item{template.df}{A \code{Templates} object.}
}
\value{
Computes a data frame providing the coverage of the 
templates for the basic as well as expected (constrained) coverage.
}
\description{
Determines the coverage of the templates for individual
allowed mismatch settings and coverage definitions.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plots_comparison.R
\name{plot_primer.comparison.box}
\alias{plot_primer.comparison.box}
\title{Boxplot for Primer Comparison}
\usage{
plot_primer.comparison.box(
  primer.data,
  con.identifier,
  con.cols,
  boundaries,
  y.limits = NULL,
  show.points = TRUE,
  highlight.set = NULL,
  nfacets = NULL
)
}
\arguments{
\item{primer.data}{List with primer data frames.}

\item{con.identifier}{Identifier of constraint to be plotted.}

\item{con.cols}{Column names with the constraint values in the primer data frames.}

\item{boundaries}{List with constraint settings.}

\item{y.limits}{Limits for the extent of the y-axis.}

\item{show.points}{If \code{TRUE} (the default), individual data points
are visualized in the boxplot, otherwise they are not shown.}

\item{highlight.set}{The identifier of a primer set to highlight in the plot.}

\item{nfacets}{A numeric providing the number of facet columns to show.
By default \code{nfacets} is \code{NULL} such that the number of 
facet columns is chosen automatically.}
}
\value{
A boxplot for primer comparison.
}
\description{
Constructs a box plot showing constraint values for each primer set.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/primers.R
\name{plot_primer_binding_regions,list,list-method}
\alias{plot_primer_binding_regions,list,list-method}
\title{Plot of Primer Binding Regions for Multiple Sets.}
\usage{
\S4method{plot_primer_binding_regions}{list,list}(
  primers,
  templates,
  direction = c("both", "fw", "rev"),
  group = NULL,
  relation = c("fw", "rev"),
  region.names = c("Binding region", "Amplification region"),
  highlight.set = NULL
)
}
\arguments{
\item{primers}{List with primer data frames.}

\item{templates}{List with template data frames.}

\item{direction}{Direction of primers.}

\item{group}{Template groups to plot.
This defaults to plotting all groups.}

\item{relation}{Plot binding region relative to forward binding region or reverse?}

\item{region.names}{Names for the primer binding region and the amplified region.}

\item{highlight.set}{Primer sets to highlight in the plot.}
}
\value{
A plot for primer binding region comparison.
}
\description{
Plots the primer binding regions for every primer set.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/optimization_algo.R
\name{select.best.primer.set}
\alias{select.best.primer.set}
\title{Selection of Best Primer Set.}
\usage{
select.best.primer.set(stats)
}
\arguments{
\item{stats}{Statistics of the primer sets to be evaluated.}
}
\value{
The index of the best primer set.
}
\description{
Selects the best primer set according to coverage and melting temperature differences among primers in the set.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/con_primer_coverage.R
\name{check.3prime.hexamers}
\alias{check.3prime.hexamers}
\title{3' Hexamer Check.}
\usage{
check.3prime.hexamers(
  template.df,
  primer.df,
  mode.directionality = c("fw", "rev", "both")
)
}
\arguments{
\item{template.df}{Template data frame.}

\item{primer.df}{Primer data frame.}

\item{mode.directionality}{Primer directionality.}
}
\value{
Returns \code{TRUE} if the 3' hexamer of a primer is fully complementary
to the corresponding template region and \code{FALSE} otherwise.
}
\description{
Check whether the 3' hexamer of a primer is fully complementary
to the corresponding region in the template.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/optimization_greedy.R
\name{select.best.opti.result}
\alias{select.best.opti.result}
\title{Selection of Best Greedy Result}
\usage{
select.best.opti.result(opti.results, template.df)
}
\arguments{
\item{opti.results}{List with primer data frames for different target melting temperatures.}

\item{template.df}{Template data frame.}
}
\value{
The index of the best primer data set.
}
\description{
Selects best greedy primer data set.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/con_primer_efficiency.R
\name{compute.efficiency}
\alias{compute.efficiency}
\title{Primer Efficiency.}
\usage{
compute.efficiency(
  fw.primers,
  fw.start,
  fw.end,
  covered,
  taqEfficiency,
  annealing.temp,
  primer_conc,
  sodium.eq.concentration,
  mode.directionality = c("fw", "rev"),
  seqs
)
}
\arguments{
\item{fw.primers}{Primer sequence strings.}

\item{fw.start}{Binding position (start).}

\item{fw.end}{Binding position (end).}

\item{covered}{List of covered template indices per primer.}

\item{taqEfficiency}{Whether the efficiency shall be computed
using a mismatch-model developed for Taq polymerases. The default setting
is \code{TRUE}. Set \code{taqEfficiency} to \code{FALSE} if you are using
another polymerase than Taq.}

\item{annealing.temp}{Annealing temperature for which to evaluate efficiency.}

\item{primer_conc}{Primer concentration.}

\item{sodium.eq.concentration}{The sodium-equivalent concentration of ions.}

\item{mode.directionality}{Primer directionality.}

\item{seqs}{Template sequence strings.}
}
\value{
The efficiencies of primer binding events.
}
\description{
Computes the efficiency of primer binding events for Taq polymerase.
}
\details{
This function uses DECIPHER's \link[DECIPHER]{CalculateEfficiencyPCR}.
}
\references{
Wright, Erik S., et al. 
"Exploiting extension bias in polymerase chain reaction to improve 
primer specificity in ensembles of nearly identical DNA templates." 
Environmental microbiology 16.5 (2014): 1354-1365.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/con_dimerization.R
\name{get.dimer.data}
\alias{get.dimer.data}
\title{Retrieval of dimerization energies.}
\usage{
get.dimer.data(s1, s2, annealing.temp, ions, no.structures)
}
\arguments{
\item{s1}{Nucleotide character vectors (5' to 3')}

\item{s2}{Nucleotide character vectors  (5' to 3')}

\item{annealing.temp}{The PCR annealing temperature in Celsius.}

\item{ions}{The sodium-equivalent ions used in the PCR.}

\item{no.structures}{Whether to compute structures of dimers.}
}
\value{
A data frame containing free energies in the field
\code{DeltaG} and the dimerization structure in \code{Structure}.
}
\description{
Uses OligoArrayAux to compute dimerization candidates.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plots_filtering.R
\name{plot.filtering.runtime}
\alias{plot.filtering.runtime}
\title{Plot Filtering Runtimes}
\usage{
\method{plot}{filtering.runtime}(filtered.stats)
}
\arguments{
\item{filered.stats}{Stats from filtering.}
}
\value{
A plot of the runtime for each filtering step.
}
\description{
Plots the runtimes of individual evaluation steps in the filtering procedure.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/optimization_algo.R
\name{design_primers.single}
\alias{design_primers.single}
\title{Design Primers for a Single Direction}
\usage{
design_primers.single(
  template.df,
  sample.name,
  mode.directionality = c("fw", "rev"),
  settings,
  timeout,
  opti.algo,
  allowed.region.definition,
  init.algo,
  max.degen,
  conservation,
  target.temps,
  required.cvg,
  fw.primers = NULL,
  cur.results.loc = NULL,
  primer.df = NULL,
  updateProgress = NULL
)
}
\arguments{
\item{template.df}{Template data frame with sequences for which primers shall be designed.}

\item{sample.name}{Identifier for the templates.}

\item{mode.directionality}{Template strands for which primers shall be designed.
Primers can be designed either only for forward strands, only for reverse strands, or
both strand directions.}

\item{settings}{A \code{DesignSettings} object specifying the 
criteria for designing primers.}

\item{timeout}{Timeout in seconds for the optimization with ILPs.}

\item{opti.algo}{The algorithm to be used for solving the primer set covering problem.}

\item{allowed.region.definition}{Definition of the target binding sites used for evaluating the coverage.
If \code{allowed.region.definition} is "within", primers have to lie within the allowed binding region.
If \code{allowed.region.definition} is "any", primers have to overlap with the allowed binding region.
The default is that primers have to bind within the target binding region.}

\item{init.algo}{The algorithm to be used for initializing primers.
If \code{init.algo} is \code{naive}, then primers are constructed from substrings of the input template sequences.
If \code{init.algo} is \code{tree}, phylogenetic trees are used to form degenerate primers whose degeneracy is
bounded by \code{max.degen}.}

\item{max.degen}{Maximal degeneracy of merged primers.}

\item{conservation}{When using the tree-based primer initialization, consider only the \code{conservation}
percentile of regions with the highest conservation.}

\item{target.temps}{Target melting temperatures for optimized primer sets in Celsius.
Only required when optimizing primers for both strand directions and one optimization was already performed.}

\item{required.cvg}{The target ratio of covered template sequences. 
If the target ratio cannot be reached, the constraint settings are relaxed up to the relaxation limits.}

\item{fw.primers}{List with optimized primer data frames corresponding to \code{target.temps}. 
Only required for optimizing both strand directions and only 
in the second optimization run in order to check for cross dimerization.}

\item{cur.results.loc}{Directory for storing results of the primer design procedure.}

\item{primer.df}{A data frame of evaluated primer candidates that can be optimized directly.}

\item{updateProgress}{Shiny progress callback function.}
}
\value{
A list containing the results of the primer design procedure:
\describe{
\item{\code{opti}:}{A \code{Primers} object representing the set of optimized primers.}
\item{\code{all_results}:}{A list containing the optimal results for each sampled melting temperature range
                           in terms of a \code{Primers} object 
                           in case that the \code{melting_temp_diff} constraint was active. Otherwise,
                           \code{all_results} only has a single entry representing a primer set relating to an undefined melting temperature.}
\item{\code{used_constraints}:}{A \code{DesignSettings} object with the (adjusted) analysis settings.}
\item{\code{filtered}:}{A \code{Primers} object containing the primer candidates that passed the filtering procedure
                        and which gave rise to the final optimal set.}
}
}
\description{
Designs primers for a single direction.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/con_primer_coverage.R
\name{predict_coverage}
\alias{predict_coverage}
\title{Prediction of Primer Coveragee.}
\usage{
predict_coverage(
  primer.df,
  template.df,
  settings,
  mode = c("on_target", "off_target"),
  updateProgress = NULL
)
}
\arguments{
\item{primer.df}{A \code{Primers} data frame.}

\item{template.df}{A \code{Templates} data frame.}

\item{settings}{A \code{DesignSettings} object.}

\item{mode}{Whether on-target or off-target events shall be considered.}
}
\value{
The predictions for primer coverage
}
\description{
Predicts primer coverage using a logistic regression model.
Converts coverage probabilities to expected false positive rate
for a given probability.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plots_filtering.R
\name{plot.filtering.stats.cvg}
\alias{plot.filtering.stats.cvg}
\title{Plot of Filtering Stats for Coverage.}
\usage{
\method{plot}{filtering.stats.cvg}(stats, stats.relax = NULL)
}
\arguments{
\item{stats}{Statistics of the filtering procedure.}

\item{stats.relax}{Statistic of the relaxation procedure.}
}
\value{
A plot showing the possible coverage after each filtering step.
}
\description{
Plots the remaining coverage after each filtering step.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/con_melting_temperature.R
\name{compute.sodium.equivalent.conc}
\alias{compute.sodium.equivalent.conc}
\title{Sodium-equivalent Concentration}
\usage{
compute.sodium.equivalent.conc(
  na_salt_conc,
  mg_salt_conc,
  k_salt_conc,
  tris_salt_conc
)
}
\arguments{
\item{na_salt_conc}{Sodium ion concentration.}

\item{mg_salt_conc}{Magensium ion concentration.}

\item{k_salt_conc}{Potassium ion concentration.}

\item{tris_salt_conc}{Tris buffer concentration.}
}
\value{
The sodium-equivalent concentration of the input ion concentrations.
}
\description{
Computes the sodium-equivalent concentration for the input ion concentrations.
}
\references{
Record, M. Thomas. "Effects of Na+ and Mg++ ions on the helix–coil transition of DNA." 
Biopolymers 14.10 (1975): 2137-2158.

Owczarzy, Richard, et al. "Predicting stability of DNA duplexes in solutions containing magnesium and monovalent cations." 
Biochemistry 47.19 (2008): 5336-5353.

Peyret, Nicolas. Prediction of nucleic acid hybridization: parameters and algorithms. 
Detroit: Wayne State University, 2000.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/primers.R
\name{update_primer_binding_regions}
\alias{update_primer_binding_regions}
\title{Update of Primer Binding Regions.}
\usage{
update_primer_binding_regions(primer.df, template.df, old.template.df)
}
\arguments{
\item{primer.df}{A \code{Primers} data frame.}

\item{template.df}{Templates with the new binding regions.}

\item{old.template.df}{Templates with the old binding regions.}
}
\value{
A \code{Primers} object with updated relative binding positions.
}
\description{
Updates the relative primer binding sites in the templates
when the template binding regions have changed since the last
coverage computation.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/filters.R
\name{relax.constraints}
\alias{relax.constraints}
\title{Relaxation of Constraints}
\usage{
relax.constraints(
  settings,
  filtered.df,
  excluded.df,
  stat.df,
  template.df,
  mode.directionality = c("fw", "rev"),
  required.cvg,
  target.temps = NULL,
  results.loc = NULL
)
}
\arguments{
\item{settings}{A \code{DesignSettings} object.}

\item{filtered.df}{Data set of primers that fulfilled all constraints of the filtering procedure.}

\item{excluded.df}{Data frame with excluded primers from the first filtering round.}

\item{stat.df}{Data frame with statistics from filtering.}

\item{template.df}{Template data frame.}

\item{mode.directionality}{Primer direction.}

\item{required.cvg}{Required ratio of covered templates.}

\item{target.temps}{Target melting temperature values.}

\item{results.loc}{The location where intermediary results should be stored.}
}
\value{
A list containing information about the relaxation as well as the filtered primers.
}
\description{
Relax constraints trying to reach the target coverage ratio.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/settings.R
\name{compare.constraints}
\alias{compare.constraints}
\title{Constraint list comparison}
\usage{
compare.constraints(A, B)
}
\arguments{
\item{A}{First constraint list.}

\item{B}{Second constraint list.}
}
\value{
TRUE if the constraints are identical, FALSE else.
}
\description{
Determines whether two list with constraints are identical.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plots_constraints.R
\name{plot_cvg_constraints,list-method}
\alias{plot_cvg_constraints,list-method}
\title{Plot for Comparing Primer Coverage Constraints.}
\usage{
\S4method{plot_cvg_constraints}{list}(
  primers,
  settings,
  active.constraints,
  highlight.set = NULL
)
}
\arguments{
\item{primers}{List with objects of class \code{Primers}.}

\item{settings}{A \code{DesignSettings} object.}

\item{active.constraints}{Names of the coverage constraints to be plotted.}

\item{highlight.set}{Primer sets to highlight in the plot.}
}
\value{
Plot of primer coverage constraints for multiple sets.
}
\description{
Plot for Comparing Primer Coverage Constraints.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/con_primer_coverage.R
\name{check.3prime.mismatches}
\alias{check.3prime.mismatches}
\title{3' Mismatch Check.}
\usage{
check.3prime.mismatches(
  template.df,
  primer.df,
  mode.directionality = c("fw", "rev", "both")
)
}
\arguments{
\item{template.df}{Template data frame.}

\item{primer.df}{Primer data frame.}

\item{mode.directionality}{Primer directionality.}
}
\value{
Returns the distance of mismatches from the 3' terminal end of primers.
}
\description{
Check for mismatches at primer 3' ends.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/con_primer_coverage.R
\name{mismatch.info}
\alias{mismatch.info}
\title{Information about Mismatches.}
\usage{
mismatch.info(primer, seqs)
}
\arguments{
\item{primer}{Primer character vector.}

\item{seqs}{Template binding sequences of primers as a \code{XStringsView} object.}
}
\value{
List with positions and number of mismatches of the \code{primer} in the 
\code{seqs}. The list contains the field \code{mm.pos} containing a list
with the positions of the mismatches and the field \code{Nbr} containing
a numeric vector with the number of mismatches per template binding event.
}
\description{
Computes information about mismatch binding events.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/optimization_algo.R
\name{evaluate.fw.rev.combinations}
\alias{evaluate.fw.rev.combinations}
\title{Evaluation of Set Combinations}
\usage{
evaluate.fw.rev.combinations(opti.fw, opti.rev, compatible.df, template.df)
}
\arguments{
\item{opti.fw}{List with forward optimal primer sets.}

\item{opti.rev}{List with reverse optimal primer sets.}

\item{compatible.df}{Data frame containing the indicices
of temperature-compatible forward and reverse primers sets.}

\item{template.df}{Template data frame for which primers were designed.}

\item{opti.rev.indices}{Indices for accessing \code{opti.rev}.}
}
\value{
List with information on the combinations of forward and reverse primers
as well as the combined data frames themselves.
}
\description{
Evaluates the combinations of forward and reverse primer sets.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ambiguity.R
\name{get.merge.idx}
\alias{get.merge.idx}
\title{Indices for merging sequences}
\usage{
get.merge.idx(seqs, max.degeneracy)
}
\arguments{
\item{seqs}{The input sequence strings.}

\item{max.degeneracy}{The maximal allowed degeneracy of a merged seq.}
}
\value{
A list of lists containing the indices of seqs to be merged. 
  For example [[1,2,3]] would indicate to merge primers 1, 2, and 3.
}
\description{
Identifies the indices of similar input sequences to be merged.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/templates.R, R/Scoring.R,
%   R/initialize_primers_tree.R, R/optimization_global.R
\name{Scoring}
\alias{Scoring}
\alias{score_conservation}
\alias{score_degen}
\alias{score_primers}
\title{Scoring Functions.}
\usage{
score_conservation(template.df, gap.char = "-", win.len = 30, by.group = TRUE)

score_degen(seq, gap.char = "-")

score_primers(
  primer.df,
  settings,
  active.constraints = names(constraints(settings)),
  alpha = 0.5
)
}
\arguments{
\item{template.df}{A \code{Templates} object providing the set of templates.}

\item{gap.char}{The gap character in the sequences.
The default is "-".}

\item{win.len}{The size of a window for evaluating conservation.
The default window size is set to 30.}

\item{by.group}{Whether the determination of binding regions 
should be stratified according to the groups defined in \code{template.df}.
The default is \code{TRUE}.}

\item{seq}{A list of vectors containing individual characters of a nucleotide sequence.}

\item{primer.df}{A \code{Primers} object containing the primers.}

\item{settings}{A \code{DesignSettings} object containing the
analysis settings.}

\item{active.constraints}{A character vector of constraint identifiers
that are considered for scoring the primers.}

\item{alpha}{A numeric that is used to determine the trade-off
between the impact of the maximal observed deviation and the total
deviation. At its default \code{alpha} is set to 0.5 such that
the maximal deviation and the total deviation have an equal weight
when computing the penalties.}
}
\value{
A list containing \code{Entropies} and \code{Alignments}.
\code{Entropies} is a data frame with conservation scores. 
Each column indicates a position in the alignment of template sequences
and each row gives the entropies of the sequences 
belonging to a specific group of template sequences.
\code{Alignments} is a list of \code{DNABin} objects, where each
object gives the alignment corresponding to one group of template sequences.

\code{score_degen} finds the number of unambiguous sequences
that are represented by \code{seq}.

\code{score_primers} returns a data frame containing
scores for individual primers.
}
\description{
\describe{
\item{\code{score_degen}}{Determines the degeneration score of a sequence.}
\item{\code{score_conservation}}{Determines the sequence conservation
scores of a set of templates using Shannon entropy.}
\item{\code{score_primers}}{Computes scores for a set of primers
based on the deviations of the primers from the constraints.}
}
}
\details{
\code{score_degen} computes the degeneration of an ambiguous sequence
by considering the number of unambiguous sequences that
are represented by the the ambiguous sequence.
Let a sequence \code{S} of length \code{n} be represented by a collection of sets such that
\deqn{S = {s_1, s_2, \ldots, s_n}}
where \eqn{s_i} indicates the set of unambiguous bases found
at position \eqn{i} of the primer. Then the degeneracy \eqn{D} of a primer
can be defined as 
\deqn{D = \prod_i{|s_i|}}
where \eqn{|s_i|} provides the number of disambiguated bases at position \eqn{i}.

\code{score_primers} determines the penalty of a primer in the following way.
Let \code{d} be a vector indicating the absolute deviations from 
individual constraints and let \code{p} be the scalar penalty that
is assigned to a primer. We define
\deqn{p = \alpha \cdot \max_i d_i + \sum_i (1 - \alpha) \cdot d_i}
such that for large values of \code{alpha} the maximal deviation 
dominates giving rise to a local penalty (reflecting the largest
absolute deviation) and for small \code{alpha} the total deviation
dominates giving rise to a global penalty 
(reflecting the sum of constraint deviations). 
When \code{alpha} is 1 only the most extreme absolute deviation is
considered and when \code{alpha} is 0 the sum of all absolute 
deviations is computed.
}
\note{
\code{score_conservation} requires the MAFFT software 
for multiple alignments (http://mafft.cbrc.jp/alignment/software/).
}
\examples{
\dontrun{
data(Ippolito)
entropy.data <- score_conservation(template.df, gap.char = "-", win.len = 18, by.group = TRUE)
}
# Compute degeneration for sequences with differing number of ambiguous bases
seq <- strsplit(c("ctggaattacggtacc", "taggaaccggrtaagc", "rtaaasrygtar"), split = "")
degen <- score_degen(seq)

# Score the primers
data(Ippolito)
primer.scores <- score_primers(primer.df, settings)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plots_coverage.R
\name{plot_template_cvg_unstratified}
\alias{plot_template_cvg_unstratified}
\title{Bar Plot of Template Coverage.}
\usage{
plot_template_cvg_unstratified(primers, templates, groups = NULL)
}
\arguments{
\item{primers}{A \code{Primers} object with evaluated primer coverage.}

\item{templates}{A \code{Templates} object containing the template sequences.}

\item{groups}{Identifiers of template groups for which plot should be created. By default, \code{groups} is set to \code{NULL} such that all 
templates are considered.
according to the number of mismatches between primer-template pairs.}
}
\value{
A plot showing the number of covered template sequences.
}
\description{
Creates a bar plot showing the coverage for every group of template sequences.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/con_gc_clamp.R
\name{evaluate.GC.clamp}
\alias{evaluate.GC.clamp}
\title{GC clamp}
\usage{
evaluate.GC.clamp(y)
}
\arguments{
\item{y}{Pirmer sequence from 5' to 3'.}
}
\value{
The length of the GC clamp.
}
\description{
Determines the number of consecutive G/Cs at the 3' end.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/IO.R
\name{create_report,Primers,Templates-method}
\alias{create_report,Primers,Templates-method}
\title{Creation of a PDF report.}
\usage{
\S4method{create_report}{Primers,Templates}(
  primers,
  templates,
  fname,
  settings,
  sample.name,
  used.settings,
  required.cvg = 1
)
}
\arguments{
\item{primers}{An evaluated \code{Primers} object.}

\item{templates}{A \code{Templates} object.}

\item{fname}{A character vector giving the file to store the report in.}

\item{settings}{A \code{DesignSettings} object.}

\item{sample.name}{An identifier for your analysis.}

\item{used.settings}{A named list (with fields "fw" and "rev") containing the forward/reverse used design settings.}

\item{required.cvg}{The required coverage ratio.}
}
\value{
Creates a PDF file reporting on the input primers.
}
\description{
Creates a PDF report for a set of primers.
}
\note{
Creating the report requires the external programs Pandoc (http://pandoc.org)
and LaTeX (http://latex-project.org).
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/con_runs.R
\name{nbr.of.runs}
\alias{nbr.of.runs}
\title{Number of Runs}
\usage{
nbr.of.runs(x)
}
\arguments{
\item{x}{Primer character sequences.}
}
\value{
The longest repeat of a single character in \code{x}.
}
\description{
Computes the longest run of a single character in the input sequence.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/con_dimerization.R
\name{batchify}
\alias{batchify}
\title{Creates multiple Batches for Parallelization.}
\usage{
batchify(tasks, annealing.temps = NULL)
}
\arguments{
\item{tasks}{An integer vector with indices representing individual computations.}

\item{annealing.temps}{Temperatures according to which to batchify.}
}
\value{
A list of lists containing indices corresponding to \code{tasks}, each list gives a batch.
}
\description{
Creates multiple Batches for Parallelization.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/settings.R
\name{select.constraints}
\alias{select.constraints}
\title{Selection of Constraints.}
\usage{
select.constraints(active.constraints)
}
\arguments{
\item{active.constraints}{A vector whose names give the constraints to be checked.}
}
\value{
A vector of useable constraint identifiers.
}
\description{
Selects constraints that can be computed according to installed third-party software.
This function is only used for initializing the 'constraint_order' option.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plots_constraints.R
\name{plot_constraint,Primers-method}
\alias{plot_constraint,Primers-method}
\title{Histogram of Constraint Values.}
\usage{
\S4method{plot_constraint}{Primers}(primers, settings, active.constraints)
}
\arguments{
\item{primers}{An evaluated object of class \code{Primers}.}

\item{settings}{A \code{DesignSettings} object containing the
settings for the constraints to be plotted.}

\item{active.constraints}{Identifiers of constraints to be plotted.
provided settings are used to visualize the desired ranges of constraints.
If \code{active.constraints} is not provided, the plotting method
will automatically try to plot all constraints defined in \code{settings}.}
}
\value{
A histogram of constraint values for the properties specified by
\code{constraints}.
}
\description{
Plots a histogram of constraint values.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/optimization_algo.R
\name{write.out.primer.info}
\alias{write.out.primer.info}
\title{Write Out Optimization Data}
\usage{
write.out.primer.info(
  opti.results.loc,
  optimal.primers.data,
  mode.directionality,
  settings,
  sample.name,
  template.df,
  max.degen
)
}
\arguments{
\item{opti.results.loc}{Folder where optimization data reside.}

\item{optimal.primers.data}{List with optimization results.}

\item{mode.directionality}{Direction of primers.}

\item{settings}{Settings used in the optimization procedure.
List containing fw, rev settings.}

\item{sample.name}{Name of template sample.}

\item{template.df}{Template data frame.}

\item{max.degen}{Maximal degeneracy of primers.}
}
\value{
Write-out of primer information to \code{opti.results.loc}.
}
\description{
Writes out all data relating to the optimization of primers.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/con_primer_coverage.R
\name{select.binding.events}
\alias{select.binding.events}
\title{Selection of Individual Binding Events}
\usage{
select.binding.events(fw.binding.filtered, p.idx)
}
\arguments{
\item{fw.binding.filtered}{IRanges with binding events.}

\item{p.idx}{Index of binding events to keep.}
}
\value{
An IRanges object containing only the selected binding events.
}
\description{
Selects only binding events of interest.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plots_constraints.R
\name{plot_constraint_deviation,Primers-method}
\alias{plot_constraint_deviation,Primers-method}
\title{Plot of Constraint Deviations for a Single Primer Set.}
\usage{
\S4method{plot_constraint_deviation}{Primers}(primers, settings, active.constraints)
}
\arguments{
\item{primers}{An evaluated object of class \code{Primers}.}

\item{settings}{A \code{DesignSettings} object
containing the target ranges for the primer properties.}

\item{active.constraints}{Constraint identifiers to be plotted.}
}
\value{
A boxplot of deviations
}
\description{
Plots a box plot of deviations of 
primer properties from the target ranges.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/con_annealing_temperature.R
\name{annealing.temp.rule.of.thumb}
\alias{annealing.temp.rule.of.thumb}
\title{Rule of thumb for annealing temperature}
\usage{
annealing.temp.rule.of.thumb(melting.temp)
}
\arguments{
\item{melting.temp}{Melting temperatures of primers}
}
\value{
The annealing temperature corresponding to the input melting temperature.
}
\description{
Computes the annealing temperature using a rule of thumb
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/helper_functions.R
\name{get.extension}
\alias{get.extension}
\title{Identification of File extension.}
\usage{
get.extension(x)
}
\arguments{
\item{x}{A string for a filename.}
}
\value{
The extension of \code{x}.
}
\description{
Identifies the file extension of \code{x}.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plots_comparison.R
\name{comparison.cvg}
\alias{comparison.cvg}
\title{Comparison Coverage Stats.}
\usage{
comparison.cvg(primer.data, template.data)
}
\arguments{
\item{primer.data}{List with primer data frames.}

\item{template.data}{List with template data frames.}
}
\value{
Coverage statistics for comparing primers.
}
\description{
Computes coverage stats for primer comparison.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/templates.R
\name{read.sequences}
\alias{read.sequences}
\title{Read Sequences.}
\usage{
read.sequences(fasta.file, gap.character)
}
\arguments{
\item{fasta.file}{The path to a FASTA file.}

\item{The}{character indicating gaps in sequences.}
}
\value{
A data frame with sequences.
}
\description{
Reads an input FASTA file.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/optimization_algo.R
\name{relax.opti.constraints}
\alias{relax.opti.constraints}
\title{Relaxation of Optimization Constraints}
\usage{
relax.opti.constraints(
  cur.opti.constraints,
  initial.opti.limits,
  initial.opti.constraints
)
}
\arguments{
\item{cur.opti.constraints}{List with optimization constraint settings.}

\item{initial.opti.limits}{Initial optimization limits.}

\item{initial.opti.constraints}{Initial optimization constraints.}
}
\value{
Relaxed optimization constraints.
}
\description{
Relax optimization constraints.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/primers.R
\name{get.analysis.mode}
\alias{get.analysis.mode}
\title{Direction of Primers.}
\usage{
get.analysis.mode(primers)
}
\arguments{
\item{primers}{A primer data frame.}
}
\value{
\code{both} if both, forward and reverse primers exist in \code{primers}.
Otherwise, if either only forward primers or reverse primers exist, returns \code{fw}
or \code{rev}, respectively.
}
\description{
Identifies the directionality of the input primers.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ambiguity.R
\name{convert.from.iupac}
\alias{convert.from.iupac}
\title{Conversion from IUPAC nucleotides}
\usage{
convert.from.iupac(seqs)
}
\arguments{
\item{seqs}{A vector of strings.}
}
\value{
A list containing the disambiguated input sequences.
}
\description{
Convert sequences with IUPAC ambiguity codes to all possible sequences without ambiguities.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/con_gc_clamp.R
\name{consecutive.GC.count}
\alias{consecutive.GC.count}
\title{Consecutive GCs}
\usage{
consecutive.GC.count(y, len)
}
\arguments{
\item{y}{Positions where G/C occurs.
Positions are numbered from 1 to 5 where 5 is the end of the primer.}

\item{len}{Is the number of bases from the primer end considered.}
}
\value{
The maximal number of consecutive G/Cs.
}
\description{
Determines the maximum number of consecutive G/Cs
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/settings.R
\name{create.constraint.XML}
\alias{create.constraint.XML}
\title{XML Output of Constraints}
\usage{
create.constraint.XML(
  filtering.constraints,
  c.f.lim,
  cvg.constraints,
  PCR.settings,
  constraint.settings
)
}
\arguments{
\item{filtering.constraints}{List with constraint settings for filtering.}

\item{c.f.lim}{Relaxation limits for the filtering constraints.}

\item{cvg.constraints}{List with constraints for coverage computations.}

\item{PCR.settings}{Settings for the PCR.}

\item{constraint.settings}{Other settings of constraints (e.g. coverage).}
}
\value{
String in XML format containing all constraint settings.
}
\description{
Creates an XML summarizing all settings.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/con_dimerization.R
\name{prepare.dimer.seqs}
\alias{prepare.dimer.seqs}
\title{Preparation of Input for Dimerization.}
\usage{
prepare.dimer.seqs(s1, s2)
}
\arguments{
\item{s1}{Nucleotide character vectors (5' to 3')}

\item{s2}{Nucleotide character vectors  (5' to 3')}
}
\value{
A list with two fields containing character vectors.
}
\description{
Preparation of Input for Dimerization.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/con_primer_coverage.R
\name{select_best_binding}
\alias{select_best_binding}
\title{Selection of Best (smallest number of mismatches) Binding Event per Template Coverage Event.}
\usage{
select_best_binding(binding, fw.mm.info)
}
\arguments{
\item{binding}{Binding information.}

\item{fw.mm.info}{Info about mismatches.}
}
\value{
A list with entries 'fw' and 'rev' giving the best indices of primers.
}
\description{
Selection of Best (smallest number of mismatches) Binding Event per Template Coverage Event.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/con_melting_temperature.R
\name{compute.empiric.melting.temp}
\alias{compute.empiric.melting.temp}
\title{Non-Thermodynamic Computation of Melting Temperatures.}
\usage{
compute.empiric.melting.temp(primer.df)
}
\arguments{
\item{primer.df}{A \code{Primers} object.}
}
\value{
A data frame with melting temperature information for the primers.
}
\description{
Computes the melting temperature of primers from an empiric formula.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/optimization_algo.R
\name{plot.all.filtering.stats}
\alias{plot.all.filtering.stats}
\title{Plot Filtering Stats.}
\usage{
\method{plot}{all.filtering.stats}(
  results.loc,
  sample,
  excluded.df,
  filtered.stats,
  stats.relax,
  template.df
)
}
\arguments{
\item{results.loc}{Location where the filtering results are stored.}

\item{sample}{Primer design run identifier.}

\item{excluded.df}{Data frame with excluded primers.}

\item{filtered.stats}{Filtering statistics data frame.}

\item{stats.relax}{Filtering statistics after relaxation.}

\item{template.df}{Template data frame.}
}
\value{
Write-out of filtering results.
}
\description{
Plots filtering statistics.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/initialize_primers.R
\name{initialize.primer.set}
\alias{initialize.primer.set}
\title{Creation of Initial Primers}
\usage{
initialize.primer.set(
  template.df,
  sample.name,
  primer.lengths,
  allowed.region.definition,
  mode.directionality,
  init.algo,
  max.degen,
  conservation,
  cur.results.loc
)
}
\arguments{
\item{template.df}{Template data frame.}

\item{sample.name}{Name of the template sample.}

\item{primer.lengths}{Interval of minimal and maximal desired primer length.}

\item{allowed.region.definition}{Definition of the allowed binding region.}

\item{mode.directionality}{Direction of primers to be created.}

\item{init.algo}{Algorithm for initializing primers.}

\item{max.degen}{Maximal allowed degeneration of created primers.}

\item{conservation}{Required conservation of primers.
The value of \code{conservation} should be in the range[0,1].}

\item{cur.results.loc}{Location for writing the primers as csv.}
}
\value{
An initial primer data frame.
}
\description{
Creates a set of candidate primers.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/con_annealing_temperature.R
\name{compute_annealing_temp}
\alias{compute_annealing_temp}
\title{Annealing temperature.}
\usage{
compute_annealing_temp(
  primer.df,
  mode.directionality,
  template.df,
  na_salt_conc,
  mg_salt_conc,
  k_salt_conc,
  tris_salt_conc,
  primer_conc
)
}
\arguments{
\item{primer.df}{Primer data frame.}

\item{mode.directionality}{Primer directionality.}

\item{template.df}{Template data frame}

\item{na_salt_conc}{Sodium ion concentration.}

\item{mg_salt_conc}{Magensium ion concentration.}

\item{k_salt_conc}{Potassium ion concentration.}

\item{tris_salt_conc}{Tris buffer concentration.}

\item{primer_conc}{Primer concentration.}
}
\value{
The optimal annealing temperature.
}
\description{
Identifies the optimal annealing temperature of a set of primers.
If primers cover template sequences, the annealing temperature
is computed using Rychlik's formula.
Otherwise, the annealing temperature is determined using the
rule of thumb based on the melting temperatures of the primers.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plots_coverage.R
\name{plot_template_cvg_mismatches}
\alias{plot_template_cvg_mismatches}
\title{Bar Plot of Template Coverage for Mismatches.}
\usage{
plot_template_cvg_mismatches(
  primer.df,
  template.df,
  groups = NULL,
  nfacets = 2
)
}
\arguments{
\item{primer.df}{A \code{Primers} object.}

\item{template.df}{A \code{Templates} object.}

\item{groups}{Identifiers of template groups for which plot should be created. By default, \code{groups} is set to \code{NULL} such that all 
templates are considered.}

\item{nfacets}{The number of facets columns to plot. By default,
\code{nfacets} is set to 2.}
}
\value{
A plot showing the number of covered template sequences.
}
\description{
Creates a bar plot showing the coverage for every group of template sequences.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/settings.R
\name{con_select}
\alias{con_select}
\title{Quick Selection of Constraints.}
\usage{
con_select(active.constraints)
}
\arguments{
\item{active.constraints}{Identifiers of constraints.}
}
\value{
The identifiers of constraints that can be computed.
}
\description{
Select constraints that can be used according to third-party tools quickly.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/optimization_ILP.R
\name{optimize.ILP}
\alias{optimize.ILP}
\title{Solver for ILP Set Cover}
\usage{
optimize.ILP(
  primer.df,
  template.df,
  settings,
  primer_conc,
  template_conc,
  na_salt_conc,
  mg_salt_conc,
  k_salt_conc,
  tris_salt_conc,
  allowed.mismatches,
  allowed.other.binding.ratio,
  allowed.stop.codons,
  allowed.region.definition,
  disallowed.mismatch.pos,
  target.temps,
  required.cvg,
  fw.primers = NULL,
  diagnostic.location = NULL,
  timeout = Inf,
  updateProgress = NULL
)
}
\arguments{
\item{primer.df}{Primer data frame to be optimized.}

\item{template.df}{Template data frame with sequences.}

\item{settings}{A \code{DesignSettings} object.}

\item{primer_conc}{Primer concentration.}

\item{template_conc}{Template concentration.}

\item{na_salt_conc}{Sodium ion concentration.}

\item{mg_salt_conc}{Magensium ion concentration.}

\item{k_salt_conc}{Potassium ion concentration.}

\item{tris_salt_conc}{Tris ion concentration.}

\item{allowed.mismatches}{The number of mismatches primers are allowed to have with the templates.}

\item{allowed.other.binding.ratio}{Ratio of primers allowed to bind to non-target regions.}

\item{allowed.stop.codons}{Consider mismatch binding events that induce stop codons.}

\item{allowed.region.definition}{Definition of the target binding sites used for evaluating the coverage.
If \code{allowed.region.definition} is \code{within}, primers have to lie within the allowed binding region.
If \code{allowed.region.definition} is \code{any}, primers have to overlap with the allowed binding region.
The default is that primers have to bind within the target binding region.}

\item{disallowed.mismatch.pos}{The number of positions from the primer 3' end where mismatches should not be allowed.
All primers binding templates with mismatches within \code{disallowed.mismatch.pos} from the 3' end are disregarded.}

\item{target.temps}{Target melting temperatures for primer sets in Celsius.}

\item{required.cvg}{Target coverage ratio of the templates by the primers.}

\item{fw.primers}{List with optimized primer data frames corresponding to \code{target.temps}. 
Only required for optimizing both strand directions and only 
in the second optimization run in order to check for cross dimerization.}

\item{diagnostic.location}{Directory for storing results.}

\item{timeout}{Timeout in seconds for the optimization with ILPs.}

\item{updateProgress}{Shiny progress callback function.}
}
\value{
List with optimization results.
}
\description{
Solves the primer set cover problem using an ILP formulation.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/templates.R
\name{read_templates_csv}
\alias{read_templates_csv}
\title{Read Template CSV File}
\usage{
read_templates_csv(fname)
}
\arguments{
\item{fname}{The filename of the input template CSV file.}
}
\value{
A template data frame.
}
\description{
Reads an input template CSV file.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/primers.R
\name{estimate.cvg.dir}
\alias{estimate.cvg.dir}
\title{Estimation of Primer Coverage.}
\usage{
estimate.cvg.dir(seqs, k, id = "")
}
\arguments{
\item{seqs}{A character vector of sequences to evaluate coverage for.}

\item{k}{A numeric giving the size of the primers.}

\item{id}{An optional identifier for the primers.}
}
\value{
A data frame with binding information.
}
\description{
Estimates the possible coverage of primers using
probes of size \code{k} and only considering perfect
matches without consideration of ambiguities.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/IO_view.R
\name{string.list.format}
\alias{string.list.format}
\title{Format a String List.}
\usage{
string.list.format(values, order.mode = c("percentage", "value"))
}
\arguments{
\item{values}{The string list to format.}

\item{order.mode}{How the result should be ordered.
For "percentage", the strings are ordered by their percentages, while
for "value", the strings are ordered by their values.}
}
\value{
A formatted string with percentage annotations.
}
\description{
Formats a string list, summarizing values with percentages.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/optimization_greedy.R
\name{check_correspondence}
\alias{check_correspondence}
\title{Check of Primer and Template Correspondence.}
\usage{
check_correspondence(primer.df, template.df)
}
\arguments{
\item{primer.df}{An object of class \code{Primers}.}

\item{template.df}{An object of class \code{Templates}.}
}
\value{
\code{TRUE} if the primers and templates seem to correspond,
\code{FALSE} otherwise.
}
\description{
Checks whether the primers relate to the correct templates.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/con_primer_coverage.R
\name{evaluate.cvg}
\alias{evaluate.cvg}
\title{Evaluation of Coverage.}
\usage{
evaluate.cvg(
  template.seqs,
  primers,
  mode.directionality = c("fw", "rev"),
  allowed.mismatches,
  updateProgress = NULL
)
}
\arguments{
\item{template.seqs}{Template sequences as a \code{DNAStringSet}.}

\item{primers}{Primer sequences as a \code{DNAStringSet}.}

\item{mode.directionality}{Directionality of primres}

\item{allowed.mismatches}{Allowed number of mismatches between a primer and a template.}

\item{updateProgress}{Progress function for shiny}
}
\value{
IRanges object with primer coverage information.
}
\description{
Evaluates primer coverage.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/optimization_ILP.R
\name{I.cvg}
\alias{I.cvg}
\title{Primer Coverage.}
\usage{
I.cvg(cvg.matrix)
}
\arguments{
\item{cvg.matrix}{Binary matrix of covering events.}
}
\value{
A list with covered templates for every primer.
}
\description{
Determines the indices of covered templates for every primer.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/templates.R, R/primers.R, R/IO.R, R/Output.R,
%   R/settings.R
\name{Output}
\alias{Output}
\alias{write_templates}
\alias{write_primers}
\alias{create_coverage_xls}
\alias{create_report}
\alias{write_settings}
\title{Output Functionalities.}
\usage{
write_templates(template.df, fname, ftype = c("FASTA", "CSV"))

write_primers(primer.df, fname, ftype = c("FASTA", "CSV"))

create_coverage_xls(primer.df, template.df, fname, settings)

create_report(
  primers,
  templates,
  fname,
  settings,
  sample.name = NULL,
  used.settings = NULL,
  ...
)

write_settings(settings, fname)
}
\arguments{
\item{template.df}{An object of class \code{Templates}.}

\item{fname}{The path to the output file.}

\item{ftype}{A character vector giving the type of the file.
This can either be "FASTA" or "CSV" (default: "FASTA").}

\item{primer.df}{An object of class \code{Primers}.}

\item{settings}{A \code{DesignSettings} object to be stored to disk.}

\item{primers}{To create a report for a single primer set, please provide
an evaluated \code{Primers} object.
For creating a report comparing multiple primer sets, please provide
a list of \code{Primers} objects.}

\item{templates}{If \code{primers} is a \code{Primers} object, \code{templates} should be a \code{Templates} object.
If \code{primers} is a list of \code{Primers} objects, \code{templates}
should be a list of \code{Templates} objects of the same length as \code{primers}.}

\item{sample.name}{An identifier for your analysis. By default ( 
\code{NULL}), the sample identifier is selected from the
\code{Run} column of the input templates.}

\item{used.settings}{A named list (with fields \code{fw} and \code{rev}) containing the relaxed settings
for designing forward/reverse primers. By default (\code{NULL}), 
the relaxed settings are not shown in the report.}

\item{...}{\code{required.cvg} (optional, default: 1), the desired coverage ratio if \code{primers} is a single primer set.}
}
\value{
\code{write_templates} stores templates to \code{fname}.

\code{write_primers} stores primers to disk.

\code{create_coverage_xls} stores information on the
primer coverage in a spreadsheet.

\code{create_report} Creates a PDF file summarizing the results
from analyzing one or multiple sets of primers.

\code{write_settings} returns the status from closing the connection to the output file.
}
\description{
\describe{
\item{\code{write_primers}}{Writes a set of primers to disk, either as a FASTA or CSV file.}
\item{\code{write_settings}}{ Stores primer analysis settings to a file in XML format.}
\item{\code{write_templates}}{Stores a set of templates as a FASTA or CSV file.}
\item{\code{create_report}}{Creates a PDF report for analyzed primer sets.}
\item{\code{create_coverage_xls}}{Creation of an XLS spreadsheet 
providing an overview of the covered
template sequences for each primer. Each cell in the spreadsheet
indicates a coverage event between a primer and template using
color codes. Identified coverage events are indicated by green, while
primer-template pairs without coverage are indicated by red.
In case that a primer binding condition (see \code{\link{CoverageConstraints}})
was active when computing the coverage, the numeric value of the
coverage condition is annotated for each cell.}
}
}
\note{
Creating the report requires the external programs Pandoc (http://pandoc.org)
and LaTeX (http://latex-project.org).
}
\examples{

data(Ippolito)
# Store templates as FASTA
fname.fasta <- tempfile("my_templates", fileext = ".fasta")
write_templates(template.df, fname.fasta)
# Store templates as CSV
fname.csv <- tempfile("my_templates", fileext = ".csv")
write_templates(template.df, fname.csv, "CSV")

data(Ippolito)
# Store primers as FASTA
fname.fasta <- tempfile("my_primers", fileext = ".fasta")
write_primers(primer.df, fname.fasta)
# Store primers as CSV
fname.csv <- tempfile("my_primers", fileext = ".csv")
write_primers(primer.df, fname.csv, "CSV")

data(Ippolito)
filename <- tempfile("cvg_overview", fileext = ".xls")
# Store coverage of a single primer in an XLS file:
my.primers <- primer.df[3,]
cvd <- unique(unlist(strsplit(my.primers$Covered_Seqs, split = ",")))
m <- match(cvd, template.df$Identifier)
my.templates <- template.df[m,]
create_coverage_xls(my.primers, my.templates, filename, settings)

setting.xml <- system.file("extdata", "settings", 
                 "C_Taq_PCR_high_stringency.xml", package = "openPrimeR")
settings <- read_settings(setting.xml)
# Creation of a report for a single primer set
data(Ippolito)
out.file.single <- tempfile("evaluation_report", fileext = ".pdf")
create_report(primer.df, template.df, out.file.single, settings)
# Creation of a report for multiple primer sets
data(Comparison)
set.sizes <- sapply(primer.data, nrow)
sel.sets <- order(set.sizes)[1:2]
out.file.comp <- tempfile("comparison_report", fileext = ".pdf")
create_report(primer.data[sel.sets], template.data[sel.sets], out.file.comp, settings)

# Store settings to disk
xml <- system.file("extdata", "settings", 
       "C_Taq_PCR_high_stringency.xml", package = "openPrimeR")
settings <- read_settings(xml)
out.file <- tempfile("my_settings", fileext = ".xml")
write_settings(settings, out.file)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/primers.R, R/PrimerEvaluation.R,
%   R/constraints_eval.R, R/filters.R, R/optimization_ILP.R,
%   R/primer_significance.R
\name{PrimerEval}
\alias{PrimerEval}
\alias{check_restriction_sites}
\alias{check_constraints}
\alias{filter_primers}
\alias{subset_primer_set}
\alias{primer_significance}
\title{Primer Evaluation.}
\usage{
check_restriction_sites(
  primer.df,
  template.df,
  adapter.action = c("warn", "rm"),
  selected = NULL,
  only.confident.calls = TRUE,
  updateProgress = NULL
)

check_constraints(
  primer.df,
  template.df,
  settings,
  active.constraints = names(constraints(settings)),
  to.compute.constraints = active.constraints,
  for.shiny = FALSE,
  updateProgress = NULL
)

filter_primers(
  primer.df,
  template.df,
  settings,
  active.constraints = names(constraints(settings))
)

subset_primer_set(
  primer.df,
  template.df,
  k = 1,
  groups = NULL,
  identifier = NULL,
  cur.results.loc = NULL
)

primer_significance(primer.df, set.name = NULL, active.constraints = NULL)
}
\arguments{
\item{primer.df}{A \code{Primers} object containing the primers
whose properties are to be checked.}

\item{template.df}{A \code{Templates} object containing the 
template sequences corresponding to \code{primer.df}.}

\item{adapter.action}{The action to be performed when adapter sequences
are found. Either "warn" to issue a warning about adapter sequences or
"rm" to remove identified adapter sequences. Currently, only
the default setting ("warn") is supported.}

\item{selected}{Names of restriction sites that are to be checked.
By default \code{selected} is \code{NULL} in which case all REBASE 
restriction sites are taken into account.}

\item{only.confident.calls}{Whether only confident calls
of restriction sites are returned.
All restriction site call is considered \emph{confident} if the restriction site
is located in a region that does not match the template sequences.
Note that this classification requires that the provided primers
are somehow complementary to the provided templates.
In contrast, non-confident restriction site calls are 
based solely on the primer sequences and do not take the templates
into account, resulting in more false positive calls of restriction sites.}

\item{updateProgress}{Progress callback function for shiny. The defaut is
\code{NULL} meaning that no progress is monitored via the Shiny interface.}

\item{settings}{A \code{DesignSettings} object containing the 
constraints that are to be considered.}

\item{active.constraints}{A subset of the constraint identifiers 
provided by \code{settings} that are to be checked
for fulfillment. By default \code{active.constraints} is \code{NULL} such that
all constraints found in \code{settings} are evaluated. Otherwise,
only the constraints specified via \code{active.constraints} 
that are available in \code{settings} are considered.}

\item{to.compute.constraints}{Constraints that are to be computed.
By default, \code{to.compute.constraints} is set to \code{NULL} such that
all \code{active.constraints} are computed. If \code{to.compute.constraints}
is a subset of \code{active.constraints}, all constraints specified
via \code{active.constraints} are evaluated for fulfillment,
but only the constraints in \code{to.compute.constraints} are newly calculated.}

\item{for.shiny}{Whether the output of the function shall be
formatted as HTML. The default setting is \code{FALSE}.}

\item{k}{The spacing between generated primer subset sizes. By default,
\code{k} is set to 1 such that all primer subsets are constructed.}

\item{groups}{The identifiers of template groups according to which 
coverage should be determined. By default, \code{groups} is set to 
\code{NULL} such that all all covered templates are considered.}

\item{identifier}{An identifier for storing the primer set. By default,
\code{identifier} is set to  \code{NULL}.}

\item{cur.results.loc}{Directory for storing the results. By default,
\code{cur.results.loc} is set to \code{NULL}, which means that
no results are stored.}

\item{set.name}{An identifier for the input primers. If \code{NULL},
the run identifier is used.}
}
\value{
\code{check_restriction_sites} returns a data frame 
with possible restriction sites found in the primers.

\code{check_constraints} returns a \code{Primers} object 
that is augmented with columns providing the results for the evaluated 
constraints.
The \code{constraints_passed} column indicates whether all 
\code{active.constraints} were fulfilled.
The \code{EVAL_*} columns indicate the fulfillment of primer-specific constraints.
The \code{T_EVAL_*} columns indicate the fulfillment of template-specific
(e.g. coverage-based) constraints.
For the coverage computations, columns prefixed by \code{Basic_},
indicate the results from string matching, while all other results
(e.g. \code{primer_coverage}) indicate the expected coverage
after applying the coverage constraints specified in \code{settings}.
Columns prefixed by \code{Off_} indicate off-target binding results.

\code{filter_primers} returns a \code{Primers} object containing
only those primers fulfilling all specified constraints.

\code{subset_primer_set} returns a list with optimal primer subsets,
each of class \code{Primers}.

\code{primer_significance} returns a numeric providing
the p-value of the primer set according to Fisher's exact test.
The returned value has the following attributes: 
\describe{
\item{\code{test}}{The results of the significance test}
\item{\code{tab}}{The confusion matrix for Fisher's exact test}
\item{\code{constraints}}{The names of the considered constraints}
}
}
\description{
\describe{
\item{\code{check_constraints}}{Determines whether a set of primers
fulfills the constraints on the properties of the primers.}
\item{\code{check_restriction_sites}}{Checks a set of primers 
for the presence of restriction sites. 
To reduce the number of possible restriction sites,
only unambiguous restriction sites are taken into account and 
only common (typically used) restriction sites are checked if a common
restriction site can be found in a sequence.}
\item{\code{filter_primers}}{Filters a primer set according to 
the specified constraints such that all primers
that do not fulfill the constraints are removed from the primer set.}
\item{\code{primer_significance}}{Uses Fisher's exact test to determine 
the significance of a primer set according to its ratio of fulfilled 
constraints.}
\item{\code{subset_primer_set}}{Determines subsets of the input primer set
that are optimal with regard to the number of covered template sequences.}
}
}
\details{
When the optional argument
\code{active.constraints} is supplied to \code{check_constraints}, 
only a subset of the constraints
provided in \code{settings} is evaluated. Only constraints that
are defined in \code{settings} can be computed. For a detailed
description of all possible constraints and their options, please
consider the \code{\link{ConstraintSettings}} documentation.

\code{subset_primer_set} determines optimal subsets of the input primer set
by solving an integer-linear program.
Since the quality of the primers (in terms of properties) is not taken into
account when creating the subsets, this method should only be used
for primer sets that are already of high quality.

\code{primer_significance} computes the significance by comparing
the total count of fulfilled and failed constraints
with the corresponding counts of primer sets from the literature.
Significant p-values indicate primer sets whose rate of constraint 
fulfillment is higher compared to the reference sets.
}
\note{
Please note that some constraint computations 
may require the installation of additional programs; for more information
please view the documentation of \code{\link{DesignSettings}}.
}
\examples{

data(Ippolito)
# Check the first primer for restriction sites with respect to the first 10 templates
site.df <- check_restriction_sites(primer.df[1,], template.df[1:10])

data(Ippolito)
settings.xml <- system.file("extdata", "settings", 
                 "C_Taq_PCR_high_stringency.xml", package = "openPrimeR")
settings <- read_settings(settings.xml)
# Check GC clamp and number of runs for all primers:
constraint.df <- check_constraints(primer.df, template.df, 
                     settings, active.constraints = c("gc_clamp", "no_runs"))
# Summarize the evaluation results
summary(constraint.df)

data(Ippolito)
filename <- system.file("extdata", "settings", 
             "C_Taq_PCR_high_stringency.xml", package = "openPrimeR")
settings <- read_settings(filename)
# Only retain the primers fulfilling the GC clamp constraint:
filtered.df <- filter_primers(primer.df, template.df, settings,
                 active.constraints = c("gc_ratio"))

# Determine optimal primer subsets
data(Ippolito)
primer.subsets <- subset_primer_set(primer.df, template.df, k = 3)

# Determine the significance of a primer set
data(Ippolito)
p.data <- primer_significance(primer.df, "Ippolito")
attr(p.data,"tab") # the confusion matrix
attr(p.data, "test") # results from Fisher's test
attr(p.data, "constraints") # considered constraints for the test
}
\references{
Roberts, R.J., Vincze, T., Posfai, J., Macelis, D. (2010) REBASE–a database for DNA restriction
and modification: enzymes, genes and genomes. Nucl. Acids Res. 38: D234-D236. http://rebase.neb.com
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/IO_view.R
\name{format.seqs.tex}
\alias{format.seqs.tex}
\title{Format a Sequence for LateX output.}
\usage{
\method{format}{seqs.tex}(seqs)
}
\arguments{
\item{seqs}{Character vector of sequences.}
}
\value{
Formatted sequences.
}
\description{
Formats a sequence for LateX report output.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/settings.R
\name{create.options.table}
\alias{create.options.table}
\title{Creation of a Table for Constraint Options.}
\usage{
create.options.table(
  other.settings,
  format.type = c("backend", "shiny", "report")
)
}
\arguments{
\item{other.settings}{List with constraint options}

\item{format.type}{How the table shall be formatted.}
}
\value{
A data frame.
}
\description{
Creation of a Table for Constraint Options.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/con_primer_secondary_structures.R
\name{compute.secondary.structures}
\alias{compute.secondary.structures}
\title{Secondary Structure Computations.}
\usage{
compute.secondary.structures(
  primer.df,
  mode.directionality = c("fw", "rev", "both"),
  annealing.temperature
)
}
\arguments{
\item{primer.df}{Primer data frame.}

\item{mode.directionality}{Direction of primers.}

\item{annealing.temperature}{Temperatures at which to compute secondary structures for every primer}
}
\value{
Data frame with secondary structure information.
}
\description{
Computes the secondary structures of the input primers using ViennaRNA.
}
\references{
Lorenz, Ronny and Bernhart, Stephan H. and Höner zu Siederdissen, 
Christian and Tafer, Hakim and Flamm, Christoph and Stadler, Peter F. and Hofacker, Ivo L.
ViennaRNA Package 2.0
Algorithms for Molecular Biology, 6:1 26, 2011, doi:10.1186/1748-7188-6-26
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/primers.R
\name{cbind.Primers}
\alias{cbind.Primers}
\title{cbind for Primers class.}
\usage{
\method{cbind}{Primers}(...)
}
\arguments{
\item{...}{Parameters for cbind function.}
}
\value{
Column binded Primers data frame.
}
\description{
Ensures that the cbind result has the appropriate class.
}
\examples{
data(Ippolito)
primer.df <- cbind(primer.df, primer.df)
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/settings.R
\name{get.PCR.settings}
\alias{get.PCR.settings}
\title{Gather all PCR settings.}
\usage{
get.PCR.settings(
  use_taq_polymerase,
  annealing_temp,
  Na_concentration,
  Mg_concentration,
  K_concentration,
  Tris_concentration,
  primer_concentration,
  template_concentration,
  nbr.cycles
)
}
\arguments{
\item{annealing_temp}{Annealing temperature in Celsius.}

\item{Na_concentration}{Sodium ion concentration.}

\item{Mg_concentration}{Magensium ion concentration.}

\item{K_concentration}{Potassium ion concentration.}

\item{Tris_concentration}{Tris buffer concentration.}

\item{primer_concentration}{Primer concentration.}

\item{template_concentration}{Template concentration.}
}
\value{
List with all PCR settings.
}
\description{
Gathers all PCR settings (e.g. for XML output).
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/optimization_ILP.R
\name{primer.coverage.for.groups}
\alias{primer.coverage.for.groups}
\title{Determination of Primer Coverage for Groups.}
\usage{
primer.coverage.for.groups(primer.df, template.df, groups)
}
\arguments{
\item{primer.df}{Primer data frame.}

\item{template.df}{Template data frame.}

\item{groups}{Template groups for which coverage should be determined.}
}
\value{
\code{primer.df} with coverage considered only for \code{groups}.
}
\description{
Modifies a primer data frame to retain only coverage events
relating to the selected groups of templates.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/optimization_algo.R
\name{visualize.filtering.results}
\alias{visualize.filtering.results}
\title{Visualization of Filtering Results.}
\usage{
visualize.filtering.results(
  sample,
  results.loc,
  mode.directionality,
  excluded.df,
  template.df,
  filtered.df,
  filtered.stats,
  stats.relax,
  primer_conc,
  na_salt_conc,
  mg_salt_conc,
  k_salt_conc,
  tris_salt_conc,
  settings,
  required.cvg
)
}
\arguments{
\item{sample}{Primer design run identifier.}

\item{results.loc}{Location where the filtering results are stored.}

\item{mode.directionality}{Design direction.}

\item{excluded.df}{Data frame with excluded primers.}

\item{template.df}{Template data frame.}

\item{filtered.df}{Primer data frame containing the primers that passed the constraints.}

\item{stats.relax}{Filtering statistics after relaxation.}

\item{primer_conc}{Primer concentration.}

\item{na_salt_conc}{Sodium ion concentration.}

\item{mg_salt_conc}{Magensium ion concentration.}

\item{k_salt_conc}{Potassium ion concentration.}

\item{tris_salt_conc}{Tris ion concentration.}

\item{settings}{A \code{DesignSettings} object.}

\item{required.cvg}{The required coverage.}

\item{template_conc}{Template concentration.}
}
\value{
Write-out of filtering results.
}
\description{
Visualizes the filtering results.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/initialize_primers_tree.R
\name{filter.primer.candidates}
\alias{filter.primer.candidates}
\title{Filtering of Primer Candidates}
\usage{
filter.primer.candidates(primer.candidates, min.len)
}
\arguments{
\item{primer.candidates}{Alignment of candidate primers.}

\item{min.len}{Minimal required length of primers.}
}
\value{
Filtered alignment of candidate primers.
}
\description{
Filters primer candidates according to length and duplications.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/optimization_greedy.R
\name{reorder.primer.table}
\alias{reorder.primer.table}
\title{Reorder Primers}
\usage{
\method{reorder}{primer.table}(filtered.primers, primer.ID.order)
}
\arguments{
\item{filtered.primers}{Primer data frame.}

\item{primer.ID.order}{new ordering of IDs in the data frame.}
}
\value{
Reordered primer data frame.
}
\description{
Reorders a primer set according to the IDs of primers.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/con_primer_coverage.R
\name{covered.seqs.to.ID.string}
\alias{covered.seqs.to.ID.string}
\title{Conversion of Template Coverage Indices to ID string}
\usage{
covered.seqs.to.ID.string(covered.seqs, template.df)
}
\arguments{
\item{covered.seqs}{Indices of covered template sequences.}

\item{template.df}{Template data frame.}
}
\value{
A string containing the covered template IDs.
}
\description{
Converts the input coverage indices to a comma-separated string with the template IDs.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/con_dimerization.R
\name{compute.dimer.matrix}
\alias{compute.dimer.matrix}
\title{Dimerization matrix}
\usage{
compute.dimer.matrix(G, deltaG.cutoff = -7)
}
\arguments{
\item{G}{Matrix with free energies of all considered primer pairs.}

\item{deltaG.cutoff}{Primers with free energies below the cutoff are considered dimerizing.}
}
\value{
Binary matrix with dimerization events according to the \code{deltaG.cutoff}.
Contains '1' if primers (i,j) dimerize and '0' else.
}
\description{
Computes a matrix indicating all dimerizing primers according to a DeltaG cutoff.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/primers.R
\name{disambiguate.primers}
\alias{disambiguate.primers}
\title{Disambiguation of Primers.}
\usage{
disambiguate.primers(p.df)
}
\arguments{
\item{p.df}{Primer data frame.}
}
\value{
Data frame with disambiguated primers.
}
\description{
Disambiguates ambiguous primer sequences into all possible sequences.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/IO.R
\name{create.cvg.text}
\alias{create.cvg.text}
\title{Coverage Info Text}
\usage{
create.cvg.text(stats, selected.group = NULL, ident = NULL)
}
\arguments{
\item{stats}{Data frame with coverage statistics.}

\item{selected.group}{Retrieve information for a subgroup of templates only.}

\item{ident}{An identifier for the coverage.}
}
\value{
A string with information on the coverage.
}
\description{
Creates a string with information on the coverage
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/optimization_algo.R
\name{plot.all.cvg.info}
\alias{plot.all.cvg.info}
\title{Plots Coverage Information}
\usage{
\method{plot}{all.cvg.info}(
  sample,
  results.loc,
  primers,
  template.df,
  mode.directionality,
  identifier = c("filtering", "optimized"),
  primer_conc,
  na_salt_conc,
  mg_salt_conc,
  k_salt_conc,
  tris_salt_conc,
  settings,
  required.cvg,
  used.settings = NULL
)
}
\arguments{
\item{sample}{Primer design run identifier.}

\item{results.loc}{Location where the filtering results are stored.}

\item{primers}{Primer data frame.}

\item{template.df}{Template data frame.}

\item{mode.directionality}{Design direction.}

\item{identifier}{Identifies whether filtering or optimization info should be displayed.}

\item{primer_conc}{Primer concentration.}

\item{na_salt_conc}{Sodium ion concentration.}

\item{mg_salt_conc}{Magensium ion concentration.}

\item{k_salt_conc}{Potassium ion concentration.}

\item{tris_salt_conc}{Tris ion concentration.}

\item{settings}{The \code{DesignSettings} object.}

\item{required.cvg}{The required coverage.}

\item{used.settings}{A list containing \code{DesignSettings} objects for the 'fw' and 'rev' optimization.}
}
\value{
Writes plots to files.
}
\description{
Visualizes all coverage-related data.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/primers.R
\name{check_restriction_sites_single}
\alias{check_restriction_sites_single}
\title{Identification of Sequence Restriction Sites.}
\usage{
check_restriction_sites_single(
  primer.seqs,
  template.seqs,
  adapter.action,
  direction = c("fw", "rev"),
  selected = NULL,
  only.confident.calls = TRUE,
  updateProgress = NULL
)
}
\arguments{
\item{primer.seqs}{Nucleotide sequences of primers to be checked for restriction sites in terms of a \code{DNAStringSet} object.}

\item{template.seqs}{A \code{DNAStringSet} object with nucleotide sequences 
containing the templates corresponding to \code{seqs}.}

\item{adapter.action}{The action to be performed when adapter sequences
are found. Either "warn" to issue warning about adapter sequences or
"rm" to remove identified adapter sequences.}

\item{selected}{Names of restriction sites that are to be checked.
By default \code{selected} is \code{NULL} in which case all REBASE 
restriction sites are checked.}

\item{only.confident.calls}{Only output confident calls of restriction sites.}

\item{updateProgress}{A Shiny progress callback function.}

\item{The}{primer direction that is checked.}
}
\value{
A data frame with restriction sites, if any could be found.
}
\description{
Checks the input sequences \code{seqs} for the presence of
restriction sites. By removing the restriction sites from a primer set, 
it is possible to identify the coverage of the primers
(e.g. using \code{\link{check_constraints}}) discounting for 
the impact of the mismatching bases caused by the insert.
}
\references{
Roberts, R.J., Vincze, T., Posfai, J., Macelis, D. (2010) REBASE–a database for DNA restriction
and modification: enzymes, genes and genomes. Nucl. Acids Res. 38: D234-D236. http://rebase.neb.com
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/settings.R
\name{build.tool.overview}
\alias{build.tool.overview}
\title{Creation of an Overview of Third-Party Tools.}
\usage{
build.tool.overview(AVAILABLE.TOOLS, for.shiny = FALSE)
}
\arguments{
\item{AVAILBLE.TOOLS}{A vector whose names give the required tools and whose
entries give their installation status as logicals.}

\item{If}{\code{for.shiny} is \code{TRUE}, provide the URLs for the tool using HTML.}
}
\value{
A data frame with information on third-part tools.
}
\description{
Creates a table of required third-party tools and their installation status.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plots_constraints.R
\name{string.to.IQR}
\alias{string.to.IQR}
\title{Conversion of Comma-Separated String to IQR String}
\usage{
string.to.IQR(string.values)
}
\arguments{
\item{string.values}{A vector of comma-separated numeric strings.}
}
\value{
The IQR corresponding to the input string.
}
\description{
Conversion of Comma-Separated String to IQR String
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plots_coverage.R
\name{get_plot_primer_data}
\alias{get_plot_primer_data}
\title{Data for Primer Plot.}
\usage{
get_plot_primer_data(
  primer.df,
  template.df,
  identifier = NULL,
  relation = c("fw", "rev")
)
}
\arguments{
\item{primer.df}{An object of class \code{Primers} containing
primers with evaluated primer coverage.}

\item{template.df}{An object of class \code{Templates} with template sequences
corresponding to \code{primer.df}.}

\item{identifier}{Identifiers of primers that are to be considered.
If \code{identifier} is set to \code{NULL} (the default), all primers are considered.}

\item{relation}{Compute binding positions relative to forward (\code{fw}) or reverse (\code{rev}) binding regions.
The default is "fw".}
}
\value{
Data frame with primer binding data.
}
\description{
Constructs a data frame containing information
about primer binding events.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/helper_functions.R
\name{get.plot.height}
\alias{get.plot.height}
\title{Plot Extent}
\usage{
get.plot.height(N, px.per.n = 50, min.size = 300, max.size = 1500000)
}
\arguments{
\item{N}{Number of observations to plot.}

\item{px.per.n}{Pixels required per observations.}

\item{min.size}{Minimal extent of plot in pixels.}

\item{max.size}{Maximal extent of plot in pixels.}
}
\value{
The extent of the plot.
}
\description{
Returns the extent of a plot.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/settings.R
\name{optiLimits}
\alias{optiLimits}
\alias{optiLimits,DesignSettings-method}
\title{Getter for Optimization Constraint Limits.}
\usage{
optiLimits(x)

\S4method{optiLimits}{DesignSettings}(x)
}
\arguments{
\item{x}{A \code{DesignSettings} object.}
}
\value{
Gets the optimization constraint limits.
}
\description{
Gets the limits for the constraints
that are applied just before the optimization procedure
using the \code{Input_Constraint_Boundaries} slot of the provided
\code{DesignSettings} object \code{x}.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/initialize_primers_tree.R
\name{get.consensus.seq}
\alias{get.consensus.seq}
\title{Computation of Consensus.}
\usage{
get.consensus.seq(ali)
}
\arguments{
\item{ali}{An alignment object.}
}
\value{
A consensus sequence without gap characters.
}
\description{
Computes the consensus of the sequences in the input alignment.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/optimization_algo.R
\name{compute.Tm.sets}
\alias{compute.Tm.sets}
\title{Cross-Dimerization Filtering}
\usage{
compute.Tm.sets(
  primer.df,
  template.df,
  Tm.brackets,
  settings,
  mode.directionality = c("fw", "rev"),
  primer_conc,
  template_conc,
  na_salt_conc,
  mg_salt_conc,
  k_salt_conc,
  tris_salt_conc,
  allowed.mismatches,
  allowed.other.binding.ratio,
  allowed.stop.codons,
  allowed.region.definition,
  disallowed.mismatch.pos,
  opti.mode = FALSE,
  required.cvg = NULL,
  primers.fw = NULL,
  diagnostic.location = NULL,
  updateProgress = NULL
)
}
\arguments{
\item{primer.df}{Primer data frame.}

\item{template.df}{Template data frame.}

\item{Tm.brackets}{Data frame with target primer melting temperatures.}

\item{settings}{A \code{DesignSettings} object.}

\item{mode.directionality}{Identifier of strand for which primers shall be designed.}

\item{primer_conc}{Primer concentration.}

\item{template_conc}{Template concentration.}

\item{na_salt_conc}{Sodium ion concentration.}

\item{mg_salt_conc}{Magensium ion concentration.}

\item{k_salt_conc}{Potassium ion concentration.}

\item{tris_salt_conc}{Tris ion concentration.}

\item{allowed.mismatches}{The number of mismatches primers are allowed to have with the templates.}

\item{allowed.other.binding.ratio}{Ratio of primers allowed to bind to non-target regions.}

\item{allowed.stop.codons}{Consider mismatch binding events that induce stop codons.}

\item{allowed.region.definition}{Definition of the allowed region.}

\item{disallowed.mismatch.pos}{The number of positions from the primer 3' end where mismatches should not be allowed.
All primers binding templates with mismatches within \code{disallowed.mismatch.pos} from the 3' end are disregarded.}

\item{opti.mode}{Compute optimization constraints and relax delta Tm if necessary.}

\item{required.cvg}{Target coverage ratio.}

\item{primers.fw}{Already designed primer sets for the target temperatures given in \code{Tm.brackets}.
Used to determine cross-dimerization.}

\item{diagnostic.location}{Directory for storing results.}

\item{updateProgress}{Shiny progress callback function.}

\item{primers.rev}{The primer data set to be filtered for cross-dimers.}

\item{opti.constraints}{List with optimization constraint settings.}

\item{annealing.temp}{The PCR annealing temperature.}
}
\value{
\code{primers.rev} with removed cross-dimerizing primers.

Primer data frames for every target temperature.
}
\description{
Removes cross-dimerizing primers from the input data.
}
\keyword{Creation}
\keyword{Cross-dimerization}
\keyword{Melting}
\keyword{Sets.}
\keyword{Stratifies}
\keyword{Temperature}
\keyword{\code{primers.fw}}
\keyword{a}
\keyword{according}
\keyword{and}
\keyword{between}
\keyword{case}
\keyword{checked}
\keyword{checks}
\keyword{constraints.}
\keyword{data}
\keyword{forward}
\keyword{frames}
\keyword{here,}
\keyword{in}
\keyword{internal}
\keyword{is}
\keyword{list}
\keyword{melting}
\keyword{of}
\keyword{optimization}
\keyword{primer}
\keyword{primers}
\keyword{provided.}
\keyword{reverse}
\keyword{run}
\keyword{second}
\keyword{temperature-dependent}
\keyword{temperatures}
\keyword{their}
\keyword{this}
\keyword{to}
\keyword{with}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/optimization_ILP.R
\name{add.coverage.constraints}
\alias{add.coverage.constraints}
\title{Addition of Coverage Constraints.}
\usage{
add.coverage.constraints(lprec, covered.templates, template.coverage)
}
\arguments{
\item{lprec}{An ILP instance.}

\item{covered.templates}{Indices of covered template sequences.}

\item{template.coverage}{List containing the indices of covering primers for each template.}
}
\value{
\code{lprec} with coverage constraints.
}
\description{
Adds coverage constraints to ILP instance.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/templates.R
\name{detect.gap.columns}
\alias{detect.gap.columns}
\title{Identification of Gappy Columns in Alignments.}
\usage{
detect.gap.columns(bins, gap.cutoff = 0.95, gap.char = "-")
}
\arguments{
\item{bins}{A list of \code{DNABin} alignments.}

\item{gap.cutoff}{The required percentage of gaps
for consideration as a gap column.}

\item{gap.char}{The gap character in the alignments.}
}
\value{
A list with indices giving the gap columns
for every alignment in \code{bins}.
}
\description{
Identification of Gappy Columns in Alignments.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plots_constraints.R
\name{get.eval.cols}
\alias{get.eval.cols}
\title{Retrieval of Evaluation Columns.}
\usage{
get.eval.cols(primer.data, constraint.settings)
}
\arguments{
\item{primer.data}{A list with \code{Primers} objects.}

\item{constraint.settings}{A list with constraint settings.}
}
\value{
A character vector with EVAL-columns.
}
\description{
Retrieves the evaluation columns by intersecting
the already evaluated constraints in \code{primer.data}
as well as the constraints specified in input constraint settings.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/zzz.R
\name{copy.melt.config}
\alias{copy.melt.config}
\title{Copy MELTING Config File}
\usage{
copy.melt.config(melt.bin = NULL)
}
\value{
TRUE if the file is available in the MELTING folder, FALSE otherwise.
}
\description{
Copies modified MELTING tandem mismatch file to the MELTING data folder.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plots_coverage.R
\name{prepare_template_cvg_mm_data}
\alias{prepare_template_cvg_mm_data}
\title{Preparation of Data for Plotting Mismatch Template Coverage.}
\usage{
prepare_template_cvg_mm_data(primer.df, template.df, allowed.mismatches = NULL)
}
\arguments{
\item{primer.df}{A \code{Primers} object.}

\item{template.df}{A \code{Templates} object.}

\item{allowed.mismatches}{An optional numeric specifying the number of mismatches to be considered at most for plotting.
If not provided, the maximal number of mismatches found for the input primer 
set is used.}
}
\value{
A data frame for creating a plot.
}
\description{
Creates a data frame for plotting a bar plot for the covered templates per allowed mismatches.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plots_coverage.R
\name{plot_template_cvg_comparison_mismatch}
\alias{plot_template_cvg_comparison_mismatch}
\title{Templates Coverage for Multiple Primer Sets.}
\usage{
plot_template_cvg_comparison_mismatch(primers, templates, highlight.set = NULL)
}
\arguments{
\item{primers}{List with primer data frames.}

\item{templates}{List with template data frames.}

\item{highlight.set}{Primer sets to be highlighted.}
}
\value{
A plot for comparing primer coverage.
}
\description{
Plots the coverage of multiple primer sets.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ambiguity.R
\name{my.disambiguate}
\alias{my.disambiguate}
\title{Disambiguation of Sequences.}
\usage{
my.disambiguate(template.seqs, gap.char = "-", degen.cutoff = 2^10)
}
\arguments{
\item{template.seqs}{A \code{DNAStringSet} object with sequences to disambiguate.}

\item{gap.char}{The character indicating gaps in alignments.}

\item{degen.cutoff}{The maximal degeneration of sequences to be disambiguated.}
}
\value{
A \code{DNAStringSetList} object with disambiguated sequences.
}
\description{
Disambiguates the input sequences, but does not disambiguate highly generate sequences.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/optimization_algo.R
\name{update.opti.results}
\alias{update.opti.results}
\title{Augmentation of Optimized Primer Data.}
\usage{
\method{update}{opti.results}(primer.df, settings, template.df)
}
\arguments{
\item{primer.df}{A primer data frame.}

\item{settings}{A \code{DesignSettings} object.}
}
\value{
An updated primer data frame.
}
\description{
Adds melting_temp_diff and cross_dimerization info to optimized sets.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/con_dimerization.R
\name{html.format.structure}
\alias{html.format.structure}
\title{Formats a Dimerization Structure for HTML.}
\usage{
html.format.structure(structures)
}
\arguments{
\item{structures}{A character vector of dimerization structures.}
}
\value{
HTML-formatted character vectors.
}
\description{
Formats a Dimerization Structure for HTML.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/primers.R
\name{cbind2,Primers,ANY-method}
\alias{cbind2,Primers,ANY-method}
\alias{rbind2,Primers,ANY-method}
\alias{[,Primers,ANY-method}
\alias{[,Primers-method}
\alias{$<-,Primers-method}
\title{S4 cbind for Primers.}
\usage{
\S4method{cbind2}{Primers,ANY}(x, y, ...)

\S4method{rbind2}{Primers,ANY}(x, y, ...)

\S4method{[}{Primers,ANY}(x, i, j, ..., drop = TRUE)

\S4method{$}{Primers}(x, name) <- value
}
\arguments{
\item{x}{The Primers data frame.}

\item{y}{Another data frame.}

\item{...}{Other arguments to the slice operator.}

\item{i}{The row index.}

\item{j}{The column index.}

\item{drop}{Simplify data frame?}

\item{name}{The name of the column.}

\item{value}{The values of the column.}
}
\value{
Cbinded primer data frame.

Rbinded primer data frame.

Subset of primer data frame.

Primer data frame with replaced column.
}
\description{
S4 cbind function for Primers.

S4 rbind function for Primers.

Slices a Primers data frame.

Stores data in a column of a Primers data frame.
}
\examples{
data(Ippolito)
primer.df <- cbind2(primer.df, seq_len(nrow(primer.df)))
data(Ippolito)
primer.df <- primer.df[1:2,]
data(Ippolito)
primer.df$Forward[1] <- "ctagcgggaccg"
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/helper_functions.R
\name{string.list.format.total}
\alias{string.list.format.total}
\title{Format Strings}
\usage{
string.list.format.total(values)
}
\arguments{
\item{values}{A comma-separated string with values.}
}
\value{
A percentage-formatted representation of the input string.
}
\description{
Changes the representation of the comma-separated string input.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/AnalysisStats.R, R/primers.R,
%   R/plots_comparison.R, R/plots_coverage.R
\name{AnalysisStats}
\alias{AnalysisStats}
\alias{get_cvg_ratio}
\alias{get_comparison_table}
\alias{get_cvg_stats_primer}
\alias{get_cvg_stats}
\title{Primer Analysis Statistics.}
\usage{
get_cvg_ratio(
  primer.df,
  template.df,
  allowed.mismatches = NULL,
  cvg.definition = c("constrained", "basic"),
  mode.directionality = NULL,
  as.char = FALSE
)

get_comparison_table(templates, primers, sample.name = NULL)

get_cvg_stats_primer(
  primer.df,
  template.df,
  cvg.definition = c("constrained", "basic")
)

get_cvg_stats(
  primers,
  templates,
  for.viewing = FALSE,
  total.percentages = FALSE,
  allowed.mismatches = Inf,
  cvg.definition = c("constrained", "basic")
)
}
\arguments{
\item{primer.df}{A \code{Primers} object containing the primers.}

\item{template.df}{A \code{Templates} object containing
the template sequences corresponding to \code{primer.df}.}

\item{allowed.mismatches}{The number of allowed mismatches
for determining the coverage of the templates. By default,
all annotated coverage events are considered.}

\item{cvg.definition}{If \code{cvg.definition} is set to
"constrained", the statistics for the expected
coverage (after applying the coverage constraints) are retrieved.
If \code{cvg.definition} is set to "basic", the coverage is determined 
solely by string matching (i.e. without applying the coverage constraints).
By default, \code{cvg.definition} is set to "constrained".}

\item{mode.directionality}{If \code{mode.directionality} is provided,
the coverage of templates is computed for a specific direction of primers.
Either "fw" (forward coverage only), "rev" (reverse coverage only), or "both" for both directions. By default, \code{mode.directionality} is \code{NULL}
such that the directionality of the primers is determined automatically.}

\item{as.char}{Whether the coverage ratio should
be outputted as a percentage-formatted character vector. By default,
\code{as.char} is set to \code{FALSE} such that a numeric is returned.}

\item{templates}{If \code{primers} is an object of class \code{Primers},
please provide an object of class \code{Templates} containing the
template sequences targeted by \code{primers}. If \code{primers} is a list,
\code{templates} should be a list of \code{Template} objects.}

\item{primers}{To retrieve statistics for a single primer set, 
please provide an object of class \code{Primers} containing a set of
evaluated primers.
To retrieve statistics for multiple primer sets, pelase provide
a list with evaluated \code{Primers} objects.}

\item{sample.name}{Either a single identifier or a character vector 
of identifiers for every
\code{Templates} object in \code{templates}.
By default, \code{sample.name} is \code{NULL} such that the
\code{Run} annotations in the provided \code{Templates} objects are used.}

\item{for.viewing}{Whether the table should be formatted
to be human-readable. By default, \code{for.viewing} is \code{FALSE}.}

\item{total.percentages}{Whether group coverage percentages
should be computed in relation to the total number of template sequences
or in relation to the number of templates belonging to a specific group.
By default, \code{total.percentages} is \code{FALSE} suc that the
percentages are group-specific.}
}
\value{
By default, \code{get_cvg_ratio} returns a numeric providing 
the expected primer coverage ratio.
If \code{as.char} is \code{TRUE}, the output is provided as a
percentage-formatted character vector. 
The attributes \code{no_covered}, \code{no_templates}, 
and \code{covered_templates} provide the number of covered templates, 
the total number of templates, and the IDs of covered templates, respectively.

\code{get_comparison_table} returns a data frame summarizing
the properties of the provided primer data sets.

\code{get_cvg_stats_primer} returns a list with the following entries. \code{cvg_per_nbr_mismatches} contains a data frame listing 
the number of binding events broken down according to the number
of expected mismatches between primers and templates.
\code{cvg_per_group} contains a data frame listing the the coverage
of individual primers per group of templates.

\code{get_cvg_stats} returns a data frame whose entries provide 
the coverage of templates per group of templates.
}
\description{
\describe{
\item{\code{get_cvg_ratio}}{Determines the ratio of template sequences 
that are covered by the evaluated input primers. The ratio
is in the interval [0,1] where 0 indicates 0\% coverage (no templates
covered) and 1 indicates 100\% coverage (all templates covered).}
\item{\code{get_cvg_stats}}{Retrieve statistics on covered templates, 
either for a single or multiple primer sets.}
\item{\code{get_cvg_stats_primer}}{Creates a table summarizing the 
coverage events of individual primers.}
\item{\code{get_comparison_table}}{Creates an overview of the
properties of multiple primer sets by providing
the inter-quartile range of primer properties in bracket notation.}
}
}
\details{
The manner in which \code{get_cvg_ratio} determines the coverage ratio 
depends on the directionality of the input primers.
If either only forward or reverse primers are inputted, the individual
coverage of each primer is used to determine the overall coverage. 
If, however, forward and reverse primers are inputted at the same time, 
the coverage is defined by the intersection of binding events from both,
forward and reverse primers.

For \code{get_cvg_stats_primer}, the cells corresponding
to columns with numeric identifiers 
indicate the percentage of coverage events occurring
with a certain number of mismatches. For example
column \code{3} provides the number of coverage events
where there are exactly three mismatches between primers and templates.
The column \code{Group_Coverage} provides a listing
of the percentage of covered templates per group.
}
\examples{

data(Ippolito)
# Determine the overall coverage
cvg.ratio <- get_cvg_ratio(primer.df, template.df)
# Determine the identitity coverage ratio 
cvg.ratio.0 <- get_cvg_ratio(primer.df, template.df, allowed.mismatches = 0)

# Summarize the properties of multiple primer sets
data(Comparison)
tab <- get_comparison_table(template.data[1:3], primer.data[1:3], "IGH")

data(Ippolito)
# Determine coverage stats per primer
primer.cvg.stats <- get_cvg_stats_primer(primer.df, template.df)

# Coverage statistics for a single primer set
data(Ippolito)
cvg.stats <- get_cvg_stats(primer.df, template.df)
# Coverage statistics for multiple primer sets
data(Comparison)
cvg.stats.comp <- get_cvg_stats(primer.data[1:2], template.data[1:2])
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/settings.R
\name{opti}
\alias{opti}
\alias{opti,DesignSettings-method}
\title{Getter for Optimization Constraints.}
\usage{
opti(x)

\S4method{opti}{DesignSettings}(x)
}
\arguments{
\item{x}{A \code{DesignSettings} object.}
}
\value{
Gets the list of optimization constraints.
}
\description{
Gets the constraints on the physicochemical properties
that are applied just before the optimization procedure
using the \code{Input_Constraints} slot of the provided
\code{DesignSettings} object \code{x}.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/templates.R
\name{remove.seqs.by.keyword}
\alias{remove.seqs.by.keyword}
\title{Removal of Partial Sequences.}
\usage{
remove.seqs.by.keyword(template.df, keyword = "partial")
}
\arguments{
\item{template.df}{Template data frame.}

\item{Header}{keywords indiating templates that should be excluded.}
}
\value{
Template data frame with partial sequences removed.
}
\description{
Removes partial template sequences.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plots_constraints.R
\name{plot_template_structure}
\alias{plot_template_structure}
\title{Plot of Template Folding Energies.}
\usage{
plot_template_structure(fold.df)
}
\arguments{
\item{fold.df}{A data frame with free energies for the
template regions.}
}
\value{
A plot of DeltaDeltaG.
}
\description{
Plots the DeltaDeltaG of template folding, which is 
the difference between the free energy change of
the unconstrained folding and the free energy change
of the constrained folding.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plots_constraints.R
\name{plot_constraint_fulfillment,Primers-method}
\alias{plot_constraint_fulfillment,Primers-method}
\title{Overview of Constraint Fulfillment.}
\usage{
\S4method{plot_constraint_fulfillment}{Primers}(
  primers,
  settings,
  active.constraints,
  plot.p.vals = TRUE
)
}
\arguments{
\item{primers}{A \code{Primers} object.}

\item{settings}{A \code{DesignSettings} object.}

\item{active.constraints}{The identifiers of constarints to be plotted for fulfillment.}

\item{plot.p.vals}{Show p-value from Fisher's exact test 
for the significance
of primer constraint fulfillment in comparison to reference
primer sets.}
}
\value{
A data frame with statistics on fulfilled constraints.
}
\description{
Plots an overview of which primers passed the filtering constraints and which
primers did not.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/con_dimerization.R
\name{parse.oligo.results}
\alias{parse.oligo.results}
\title{Parser for OligoArrayAux Dimerization Data.}
\usage{
parse.oligo.results(deltaG.file, struct.file)
}
\arguments{
\item{deltaG.file}{A path to a file with OligoArrayAux energies.}

\item{struct.file}{A path to a file with OligoArrayAux structures.}
}
\value{
A data frame with structures and free energies.
}
\description{
Parses the free energies and structures of OligoArrayAux.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/primers.R
\name{restriction_match}
\alias{restriction_match}
\title{Identification of Sequence Matches.}
\usage{
restriction_match(primer.seqs, template.seqs)
}
\arguments{
\item{primer.seqs}{Primer sequences.}

\item{template.seqs}{Template sequences.}
}
\value{
A vector with the template regions matching the 
\code{primer.seqs} best.
}
\description{
Determines the most similar template sequence for every
input primer sequence. Used to identify regions for
alignment for the identification of restriction sites.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/filters.R
\name{set.new.constraint.value}
\alias{set.new.constraint.value}
\title{Update Constraint Settings.}
\usage{
set.new.constraint.value(values, relax.speed, cur.limits, cur.setting)
}
\arguments{
\item{values}{Observed constraint values considered for the update.}

\item{relax.speed}{The speed at which the constraints should be relaxed.
This value should be in the interval [0,1].}

\item{cur.limits}{List with current relaxation limits.}

\item{cur.setting}{List with current constraint settings.}
}
\value{
Relaxed constraint settings according to the given \code{values} and \code{cur.limits}.
}
\description{
Updates the constraint settings with new values. Sets to the maximal
observed values within the limits or less if \code{relax.speed} is less than 1.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/con_primer_coverage.R
\name{mismatch.string.to.list}
\alias{mismatch.string.to.list}
\title{Conversion of Mismatch Postions String to List.}
\usage{
mismatch.string.to.list(mismatches)
}
\arguments{
\item{mismatches}{A character vector where parenthesis give mismatches for a template binding event.}
}
\value{
A list with the mismatches for every template for every primer.
}
\description{
Conversion of Mismatch Postions String to List.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/templates.R
\name{update.individual.binding.region}
\alias{update.individual.binding.region}
\title{Adjustment of Existing Binding Regions for one Direction.}
\usage{
\method{update}{individual.binding.region}(min, max, template.df, mode.directionality)
}
\arguments{
\item{min}{Position where binding should start.}

\item{max}{End position of binding.}

\item{mode.directionality}{Directionality of primers.}

\item{Template}{data frame.}
}
\value{
Template data frame with updated binding regions.
}
\description{
Adjusts the existing annotation of binding regions by specifying
an interval relative to the existing binding region.
}
\details{
Position \code{0} indicates the first position after the existing
binding region. Hence, negative positions adjust the binding region towards
the existing binding regions and non-negative positionis extend
the existing binding region definition aways from the existing target region.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/primers.R
\name{restriction_hits}
\alias{restriction_hits}
\title{Identification of Restriction Sites.}
\usage{
restriction_hits(bad.regions, DB)
}
\arguments{
\item{DB}{A data frame with restriction enzyme sites.}

\item{bad.region}{IRanges with possible adapter sites.}
}
\value{
A boolean data frame indicating the presence
of adapters for all considered restriction sites.
}
\description{
Identifies restriction sites in a list with putative
restriction sites provided by \code{bad.regions} using
a data frame of restriction sites given by \code{DB}.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/templates.R
\name{read_templates_multiple}
\alias{read_templates_multiple}
\title{Input of Multiple Templates.}
\usage{
read_templates_multiple(
  filenames,
  hdr.structure = NULL,
  delim = NULL,
  id.column = NULL,
  rm.keywords = NULL,
  remove.duplicates = FALSE,
  fw.region = c(1, 30),
  rev.region = c(1, 30),
  gap.character = "-",
  run = NULL
)
}
\arguments{
\item{filenames}{Names of FASTA/CSV files containing template data.}

\item{hdr.structure}{A character vector describing the information contained in the FASTA headers. In case that the headers of \code{fasta.file} contain
template group information, please include the keyword "GROUP" in
\code{hdr.structure}.}

\item{delim}{Delimiter for the information in the FASTA headers.}

\item{id.column}{Field in the header to be used as the identifier.}

\item{rm.keywords}{A vector of keywords that are used to remove templates whose headers contain any of the keywords.}

\item{remove.duplicates}{Whether duplicate sequence shall be removed.}

\item{fw.region}{The positional interval from the template 5' end specifying the 
binding sites for forward primers.}

\item{rev.region}{The positional interval from the template 3' end specifying
the binding sites for reverse primers.}

\item{gap.character}{The character in the input file representing gaps.
Gaps are automatically removed upon input.}

\item{run}{An identifier for the template sequences.}
}
\value{
A list containing objects of class \code{Templates}.
}
\description{
Reads multiple template CSV/FASTA files.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/initialize_primers_tree.R
\name{shannon.entropy}
\alias{shannon.entropy}
\title{Shannon Entropy}
\usage{
shannon.entropy(ali)
}
\arguments{
\item{ali}{An alignment of primer sequences.}
}
\value{
The Shannon entropy for the alignment.
}
\description{
Computation of Shannon entropy for an alignment.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/con_dimerization.R
\name{get.self.dimers}
\alias{get.self.dimers}
\title{Self dimerization}
\usage{
get.self.dimers(
  primers.1,
  primers.2,
  ions,
  annealing.temp,
  no.structures = FALSE
)
}
\arguments{
\item{primers.1}{Input primers}

\item{primers.2}{(Copy/reverse) of the input primers}

\item{ions}{Sodium-equivalent ionic concentration.}

\item{annealing.temp}{The annealing temperature.}

\item{no.structures}{Whether the dimerization structure shall be computed.}
}
\value{
Possible self-dimer conformations.
}
\description{
Computes possible self-dimers.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plots_coverage.R
\name{prepare_mm_plot}
\alias{prepare_mm_plot}
\title{Data Preparation for Mismatch Plot.}
\usage{
prepare_mm_plot(primer.df, template.df, mode = c("on_target", "off_target"))
}
\arguments{
\item{primer.df}{A \code{Primers} object.}

\item{template.df}{A \code{Templates} object.}

\item{mode}{Whether to compute for on-target or off-target events.}
}
\value{
A data frame with binding information for every primer.
}
\description{
Data Preparation for Mismatch Plot.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/con_primer_coverage.R
\name{evaluate.constrained.cvg}
\alias{evaluate.constrained.cvg}
\title{Evaluation of Primer Coverage.}
\usage{
evaluate.constrained.cvg(
  template.df,
  primer.df,
  cvg.df,
  mode.directionality = c("fw", "rev", "both"),
  settings,
  updateProgress = NULL
)
}
\arguments{
\item{template.df}{Template data frame.}

\item{primer.df}{Primer data frame.}

\item{cvg.df}{Data frame with basic coverage entries.}

\item{mode.directionality}{Primer directionality.}

\item{settings}{A \code{DesignSettings} object.}

\item{updateProgress}{Progress callback function for shiny.}
}
\value{
Primer data frame with information on the covered template sequences.
}
\description{
Evaluates the coverage of a set of primers.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/settings.R
\name{constraints.to.df}
\alias{constraints.to.df}
\title{Conversion of Constraints List to Data Frame.}
\usage{
constraints.to.df(
  limit.constraints,
  out.names,
  format.type = c("backend", "shiny", "report")
)
}
\arguments{
\item{limit.constraints}{A list with constraints.}

\item{out.names}{The desired column names.}

\item{format.type}{The type of formatting to be performed on the table}
}
\value{
A data frame giving an overview of the constraints.
}
\description{
Converts the input constraints to a data frame representation.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/IO.R
\name{render_report}
\alias{render_report}
\title{Renders an rmarkdown file using Pandoc.}
\usage{
render_report(params, report_template, out.file)
}
\arguments{
\item{params}{A list with parameters for the R markdown parser.}

\item{report_template}{A character vector giving the basename of the
Rmarkdown template to use for report creation.}

\item{out.file}{The filename of the report PDF to be created.}
}
\value{
Creates a PDF in \code{out.file} if successful.
}
\description{
Creates a PDF report using rmarkdown and Pandoc by passing the specified
\code{params} to the markdown file given by \code{report_template} and 
storing the PDF in \code{out.file}.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/con_primer_coverage.R
\name{mismatch.mutation.check}
\alias{mismatch.mutation.check}
\title{Identification of Mutations Induced by Mismatch Binding Events.}
\usage{
mismatch.mutation.check(
  primer.df,
  template.df,
  mutation.types = c("stop_codon", "substitution")
)
}
\arguments{
\item{primer.df}{A \code{Primers} object.}

\item{template.df}{A \code{Template} object.}

\item{mutation.types}{Character vector of the mutation types to be checked for.}
}
\value{
A list containing data frames where an entry of 1 is present if the \code{primer.seq} induces a mutation that is
forbidden according to the provided \code{mutation.types}, otherwise 0.
}
\description{
Identifies whether mutations are induced by mismatch binding events.
}
\details{
Checks for one primer and all covered templates whether any templates are bound
with mismatches such that mismatches are induced. A numeric vector indicating 
which binding events induce a forbidden mismatch according to 
\code{mutation.types} is returned such that \code{1} indicates forbidden events
and \code{0} allowed events.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ambiguity.R
\name{complement.sequence}
\alias{complement.sequence}
\title{Sequence complement}
\usage{
complement.sequence(seq)
}
\arguments{
\item{seq}{The input sequence strings.}
}
\value{
The complements of the input sequences.
}
\description{
Computes the complement of the input sequence.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/templates.R
\name{rbind.Templates}
\alias{rbind.Templates}
\title{rbind for Template class.}
\usage{
\method{rbind}{Templates}(...)
}
\arguments{
\item{...}{Parameters for rbind function.}
}
\value{
Row-binded Templates data frame.
}
\description{
Ensures that the rbind result has the appropriate class.
}
\examples{
data(Ippolito)
template.df <- rbind(template.df, template.df)
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/con_dimerization.R
\name{compute.all.cross.dimers}
\alias{compute.all.cross.dimers}
\title{Cross dimerization}
\usage{
compute.all.cross.dimers(
  primer.df,
  primer_conc,
  na_salt_conc,
  mg_salt_conc,
  k_salt_conc,
  tris_salt_conc,
  annealing.temp,
  results = NULL,
  check.idx = NULL,
  for.shiny = FALSE,
  no.structures = FALSE
)
}
\arguments{
\item{primer.df}{Input primers data frame.}

\item{primer_conc}{Primer concentration.}

\item{na_salt_conc}{Sodium ion concentration.}

\item{mg_salt_conc}{Magensium ion concentration.}

\item{k_salt_conc}{Potassium ion concentration.}

\item{tris_salt_conc}{Tris buffer concentration.}

\item{annealing.temp}{The PCR annealing temperature.}

\item{results}{(optional) Cross dimer data frame (unfiltered)}

\item{check.idx}{Indices of primers for checking cross-dimerization.}

\item{for.shiny}{Whether the table is inteded for HTML display.}

\item{no.structures}{Whether dimerization structures shall not be outputted.}
}
\value{
Worst-case cross dimers.
}
\description{
Compute worst-case DeltaG data frame with all possible primer cross-dimers.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/settings.R
\name{constraints_to_unit}
\alias{constraints_to_unit}
\title{Mapping of Constraints to Units.}
\usage{
constraints_to_unit(
  constraint,
  use.unit = TRUE,
  format.type = c("backend", "HTML", "report")
)
}
\arguments{
\item{constraint}{The names of the constraints to convert to their plot identifiers (units).}

\item{use.unit}{Whether constraint names should be annotated with their units.}

\item{use.HTML}{Whether constraint units should be annotated with HTML units.}
}
\value{
A list of constraint names.
}
\description{
Maps constraints to units for plotting.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/optimization_ILP.R
\name{solve.ILP}
\alias{solve.ILP}
\title{Solve an ILP}
\usage{
\method{solve}{ILP}(
  cur.D,
  cur.G,
  cur.settings,
  cur.cvg.matrix,
  time.limit,
  required.cvg,
  primer.df,
  template.df
)
}
\arguments{
\item{cur.D}{Binary dimerization matrix.}

\item{cur.G}{Free energy matrix for cross-dimerization.}

\item{cur.settings}{Current \code{DesignSettings} object.}

\item{cur.cvg.matrix}{Binary coverage matrix.}

\item{time.limit}{Time limit for solving the ILP in seconds.}

\item{required.cvg}{The target coverage of the designed primer set.}

\item{primer.df}{A \code{Primers} object.}

\item{template.df}{A \code{Templates} object.}

\item{deltaG.cutoff}{Cutoff for dimerization free energy.}

\item{deltaG.limit}{Relaxation limit for free energy cutoff.}
}
\value{
List with ILP solution data.
}
\description{
Constructs and solves an ILP and outputs a list with the reuslts.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/check_stop_codons.R
\name{insert_str}
\alias{insert_str}
\title{String Insertion.}
\usage{
insert_str(target, insert, index)
}
\arguments{
\item{target}{The string to be modified.}

\item{insert}{The string to be inserted.}

\item{index}{The position where the insertion should take place.}
}
\value{
A string where \code{insert} is inserted into \code{target} at position \code{index}.
}
\description{
Inserts a string into another string at the speficied position.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plots_coverage.R
\name{get_primer_cvg_mm_plot_df}
\alias{get_primer_cvg_mm_plot_df}
\title{Data for Mismatch Primer Coverage Plot.}
\usage{
get_primer_cvg_mm_plot_df(primer.df, template.df)
}
\arguments{
\item{primer.df}{A \code{Primers} object.}

\item{template.df}{A \code{Templates} object.}
}
\value{
A data frame for plotting mismatch primer coverage.
}
\description{
Ensures that there's an entry for every possible mismatch setting.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/templates.R
\name{create.uniform.leaders}
\alias{create.uniform.leaders}
\title{Uniform Binding Ranges.}
\usage{
create.uniform.leaders(fw.interval, rev.interval, template.df, gap.char)
}
\arguments{
\item{fw.interval}{Binding region for forward templates.}

\item{rev.interval}{Binding region for reverse templates.}

\item{template.df}{Template data frame.}

\item{gap.char}{The character for gaps in alignments.}
}
\value{
Data frame with binding region information.
}
\description{
Creates uniform binding regions for all templates.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/constraints_eval.R
\name{compute.constraints}
\alias{compute.constraints}
\title{Computation of Constraints.}
\usage{
compute.constraints(
  primer.df,
  mode.directionality = c("fw", "rev", "both"),
  template.df,
  settings,
  active.constraints = c("primer_coverage", "primer_length", "primer_specificity",
    "gc_clamp", "gc_ratio", "no_runs", "no_repeats", "self_dimerization",
    "cross_dimerization", "melting_temp_range", "melting_temp_diff",
    "secondary_structure", "primer_efficiency", "annealing_DeltaG", "stop_codon",
    "terminal_mismatch_pos", "substitution", "hexamer_coverage", "coverage_model",
    "off_primer_efficiency", "off_annealing_DeltaG", "off_coverage_model"),
  no.structures = FALSE,
  for.shiny = FALSE,
  updateProgress = NULL
)
}
\arguments{
\item{primer.df}{Primer data frame.}

\item{mode.directionality}{Primer directionality.}

\item{template.df}{Template data frame.}

\item{settings}{A \code{DesignSettings} object.}

\item{active.constraints}{Strings giving the constraints that are to be computed.}

\item{no.structures}{Whether dimer structures shall be computed.}

\item{for.shiny}{Whether to format output for HTML.}

\item{updateProgress}{Progress callback function for shiny.}
}
\value{
A data frame with columns for each constraint in \code{active.constraints}.
}
\description{
Computes the specified constraints for the input primers.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/templates.R, R/primers.R, R/Input.R,
%   R/settings.R
\docType{class}
\name{Input}
\alias{Input}
\alias{Templates-class}
\alias{Templates}
\alias{read_templates}
\alias{Primers-class}
\alias{Primers}
\alias{read_primers}
\alias{read_settings}
\title{Input Functionalities.}
\usage{
Templates(...)

read_templates(
  fname,
  hdr.structure = NULL,
  delim = NULL,
  id.column = NULL,
  rm.keywords = NULL,
  remove.duplicates = FALSE,
  fw.region = c(1, 30),
  rev.region = c(1, 30),
  gap.char = "-",
  run = NULL
)

Primers(...)

read_primers(
  fname,
  fw.id = "_fw",
  rev.id = "_rev",
  merge.ambig = c("none", "merge", "unmerge"),
  max.degen = 16,
  template.df = NULL,
  adapter.action = c("warn", "rm"),
  sample.name = NULL,
  updateProgress = NULL
)

read_settings(
  filename = list.files(system.file("extdata", "settings", package = "openPrimeR"),
    pattern = "*.xml", full.names = TRUE),
  frontend = FALSE
)
}
\arguments{
\item{...}{A data frame fulfilling the structural requirements
for initializing a \code{Templates} or \code{Primers} object.}

\item{fname}{Character vector providing either a single or multiple
paths to FASTA or CSV files.}

\item{hdr.structure}{A character vector describing the information contained in the FASTA headers. In case that the headers of \code{fasta.file} contain
template group information, please include the keyword "GROUP" in
\code{hdr.structure}. If the numer of elements provided via \code{hdr.structure}
is shorter than the actual header structure, the missing fields are ignored.}

\item{delim}{Delimiter for the information in the FASTA headers.}

\item{id.column}{Field in the header to be used as the identifier of individual
template sequences.}

\item{rm.keywords}{A vector of keywords that are used to remove templates whose headers contain any of the keywords.}

\item{remove.duplicates}{Whether duplicate sequence shall be removed.}

\item{fw.region}{The positional interval from the template 5' end specifying the
binding sites for forward primers. The default \code{fw.region} is set to
the first 30 bases of the templates.}

\item{rev.region}{The positional interval from the template 3' end specifying
the binding sites for reverse primers. The default \code{rev.region}
is set to the last 30 bases of the templates.}

\item{gap.char}{The character in the input file representing gaps. 
Gaps are automatically removed upon input and the default character is "-".}

\item{run}{An identifier for the set of template sequences. By default,
\code{run} is \code{NULL} and its value is set via \code{template.file}.}

\item{fw.id}{For FASTA input, the identifier for forward primers in the FASTA headers.}

\item{rev.id}{For FASTA input, the identifier for reverse primers in the FASTA headers.}

\item{merge.ambig}{Indicates whether similar primers should be merged
("merge") using IUPAC ambiguity codes or whether primers should 
be disambiguated ("unmerge").
By default \code{merge.ambig} is set to "none", leaving primers as they are.}

\item{max.degen}{A scalar numeric providing the maximum
allowed degeneracy for merging primers if
\code{merge.ambig} is set to "merge".
Degeneracy is defined by the number of disambiguated sequences 
that are represented by a degenerate primer.}

\item{template.df}{An object of class \code{Templates}.
If \code{template.df} is provided for \code{read_primers} then the primers are checked
for restriction sites upon input; otherwhise they are not checked.}

\item{adapter.action}{The action to be performed when \code{template.df} is
provided for identifying adapter sequences.
Either "warn" to issue warning about adapter sequences or
"rm" to remove identified adapter sequences. The default is "warn".}

\item{sample.name}{An identifier for the input primers.}

\item{updateProgress}{A Shiny progress callback function. This is
\code{NULL} by default such that no progress is tracked.}

\item{filename}{Path to a valid XML file containing the 
primer analysis settings. By default, \code{filename} is set
to all settings that are shipped with openPrimeR and the lexicographically
first file is loaded.}

\item{frontend}{Indicates whether settings shall be loaded for the Shiny frontend. 
In this case no unit conversions for the PCR settings are performed.
The default setting is \code{FALSE} such that the correct units are used.}
}
\value{
The \code{Templates} constructor returns 
a \code{Templates} object, an instance of a data frame.

\code{read_templates} returns a single object of class
\code{Templates} if a single filename was provided or a list
of such objects if multiple file names were provided.

The \code{Primers} constructor returns an object of 
class \code{Primers}.

\code{read_primers} returns a single object of class \code{Primers}
if a single input file is provided or a list of such objects if multiple
files are provided.

\code{read_settings} returns an object of class \code{DesignSettings}.
}
\description{
\describe{
\item{\code{read_primers}}{Reads one or multiple input files with 
primer sequences. The input can either be in FASTA or in CSV format.}
\item{\code{read_templates}}{Read one or multiple files with template sequences in FASTA or CSV format.}
\item{\code{read_settings}}{Loads primer analysis settings from an XML file.}
\item{\code{Templates}}{The \code{Templates} class encapsulates a data frame
containing the sequencs of the templates, their binding regions,
as well as additional information (e.g. template coverage).}
\item{\code{Primers}}{The \code{Primers} class encapsulates a data frame
representing a set of primers. Objects of this class
store all properties associated with a set of primers,
for example the results from evaluating the properties
of a primer set or from determining its coverage.}
}
}
\details{
In the following you can find a description
of the most important columns that can be found
in an object of class \code{Templates}. Note that
angle brackets in the column names 
indicate the existence of multiple possibilities.
\describe{
\item{\code{ID}}{The identifiers of the templates.}
\item{\code{Identifier}}{The internal identifiers of the templates.}
\item{\code{Group}}{The identifiers of the groups that the templates belong to.}
\item{\code{Allowed_Start_<fw|rev>}}{The start of the interval in the templates
where binding is allowed for forward and reverse primers, respectively.}
\item{\code{Allowed_End_<fw|rev>}}{The end of the interval in the templates
where binding is allowed for forward and reverse primers, respectively.}
\item{\code{Allowed_<fw|rev>}}{The template sequence where binding
is allowed for forward and reverse primers, respectively.}
\item{\code{Run}}{An identifier for the set of template sequences.}
\item{\code{Covered_By_Primers}}{The identifiers of primers covering the templates,
when the template coverage has been annotated.}
\item{\code{primer_coverage}}{The number of primers covering the templates,
when the template coverage has been annotated.}
}

When loading a FASTA file with \code{read_templates},
the input arguments \code{hdr.structure}, \code{delim}, \code{id.column}, 
\code{rm.keywords}, \code{remove.duplicates}, \code{fw.region}, 
\code{rev.region}, \code{gap.character}, and \code{run} are utilized.
Most importantly, \code{hdr.structure} and \code{delim} should
match the FASTA header structure.
To learn more about setting the primer binding regions, consider the
\code{\link{assign_binding_regions}} function.
In contrast, when a CSV file is loaded with \code{read_templates},
the data are loaded without performing any modifications because the CSV file
should represent an object of class \code{\link{Templates}}, which 
can be stored using the \code{\link{write_templates}} function.

When loading primers via \code{read_primers}, the input arguments
\code{fw.id}, \code{rev.id}, \code{merge.ambig}, and \code{max.degen} 
are only used for loading primers from a FASTA file.
In this case, please ensure that \code{fw.id} and 
\code{rev.id} are set according to the keywords indicating
the primer directionalities in the FASTA file.
When loading primers from a CSV file, the format of the file should adhere
to the structure defined by the \code{\link{Primers}} class.

When loading a settings file with \code{read_settings}, 
if \code{filename} is not provided,
a default XMl settings file is loaded. Please review the 
function's examples to learn more about the default settings. If you want
to load custom settings, you can store a modified \code{DesignSettings}
object as an XML file using \code{\link{write_settings}}.
}
\section{Basic columns}{

In the following you can find a description of the most
important columns that can be found in objects of class \code{Primers}. 
Note that angular brackets indicate the existence of multiple possibilities.
The following columns are present when a set of primers
is loaded from a FASTA file using \code{\link{read_primers}}:
\describe{
\item{\code{ID}}{The identifiers of the primers.}
\item{\code{Identifier}}{The internal identifiers of the primers.}
\item{\code{Forward}}{The sequences of forward primers.}
\item{\code{Reverse}}{The sequences of reverse primers.}
\item{\code{primer_length<fw|rev>}}{The lengths of
forward and reverse primer sequences, respectively.}
\item{\code{Direction}}{Either 'fw' for forward primers,
'rev' for reverse primers, or 'both' for a primer pair.}
\item{\code{Degeneracy_<fw|rev>}}{The degeneracy (ambiguity) of
forward and reverse primers, respectively.}
\item{\code{Run}}{An identifier describing the primer set.}
}
}

\section{Coverage-related columns}{

The following columns are only available in an object of
class \code{Primers} after primer coverage
has been computed, that is after \code{\link{check_constraints}}
has been called with the active \code{primer_coverage} constraint. Computed coverage
values relating solely to string matching are indicated by the prefix
\code{Basic_}, while columns without this prefix relate to the coverage after
applying the constraints formulated via \code{CoverageConstraints}.
Information on off-target coverage events are indicated by
the \code{Off_} prefix, while on-target coverage events do not carry
this prefix.

\describe{
\item{\code{primer_coverage}}{The number of templates that are
covered by the primers. Note that if a primer set contains
primers of both directions, a template is only considered covered
if it is covered by primers of both directions.}
\item{\code{Coverage_Ratio}}{The ratio of templates that are covered by the primers.}
\item{\code{Binding_Position_Start_<fw|rev>}}{The upstream position in the 
templates where forward and reverse primers respectively bind.}
\item{\code{Binding_Position_End_<fw|rev>}}{The downstream position in the templates where forward and reverse primers respectively bind.}
\item{\code{Relative_<Forward|Reverse>_Binding_Position_<Start|End>_<fw|rev>}}{
The binding upstream (\code{Start}) or downstream (\code{End}) positions 
of the primers relative to the forward (\code{Forward})
or reverse (\code{Reverse}) binding regions, either for 
forward (\code{fw}) or reverse primers (\code{rev}).}
\item{\code{Binding_Region_Allowed}}{Whether a coverage event
occurred in the target binding region or not. If the allowed
off-target ratio was set to 0 only coverage events within the 
the target region are reported.}
\item{\code{Nbr_of_mismatches_<fw|rev>}}{The number of mismatches
of forward and reverse primer coverage events, respectively.}
\item{\code{Mismatch_pos_<fw|rev>}}{The position of mismatches
for forward and reverse coverage events, respectively. Mismatch
positions are reported relative to the 3' end, that is, position
1 indicates a mismatch in the last base of a primer.}
\item{\code{primer_specificity}}{The specificity of a primer
as determined by its ratio of off-target binding events.}
}
}

\section{Constraint-related columns}{

Each constraint that is considered when calling \code{\link{check_constraints}}
gives rise to at least one column in the provided \code{Primers} object.
Due to the large number of possible constraints, we will limit our description
to the \code{gc_clamp} constraint. Once the GC clamp property has been computed,
the \code{gc_clamp_fw} column contains the length of the GC clamp for forward 
primers and \code{gc_clamp_rev} the corresponding length for reverse primers.
Whether the desired extent of the GC clamp was obtained by a primer
is indicated by the \code{EVAL_gc_clamp} column. It contains \code{TRUE} when
the GC clamp constraint was fulfilled and \code{FALSE} when it was broken.
To identify whether all required constraints were fulfilled by a primer,
the \code{constraints_passed} column can be used. It contains \code{TRUE}
if all \code{active.constraints} used by \code{\link{check_constraints}} were fulfilled
and \code{FALSE} otherwise.
}

\examples{

# Load a set of templates:
fasta.file <- system.file("extdata", "IMGT_data", "templates", 
     "Homo_sapiens_IGH_functional_exon.fasta", package = "openPrimeR")
hdr.structure <- c("ACCESSION", "GROUP", "SPECIES", "FUNCTION")
template.df <- read_templates(fasta.file, hdr.structure, "|", "GROUP")
# Load templates from a FASTA file
fasta.file <- system.file("extdata", "IMGT_data", "templates", 
          "Homo_sapiens_IGH_functional_exon.fasta", package = "openPrimeR")
hdr.structure <- c("ACCESSION", "GROUP", "SPECIES", "FUNCTION")
template.df.fasta <- read_templates(fasta.file, hdr.structure, "|", "GROUP")
# Load mutliple FASTA files
fasta.files <- c(fasta.file, fasta.file)
template.df.fastas <- read_templates(fasta.files, hdr.structure, "|", "GROUP")
# Load templates from a previously stored CSV file
csv.file <- system.file("extdata", "IMGT_data", "comparison", 
               "templates", "IGH_templates.csv", package = "openPrimeR")
template.df.csv <- read_templates(csv.file)
# Load multiple CSV files:
csv.files <- c(csv.file, csv.file)
template.df.csvs <- read_templates(csv.files)
# Load a mixture of FASTA/CSV files:
mixed.files <- c(csv.file, fasta.file)
template.data <- read_templates(mixed.files)

# Load a set of primers
primer.location <- system.file("extdata", "IMGT_data", "primers", "IGHV", 
                     "Ippolito2012.fasta", package = "openPrimeR")
primer.df <- read_primers(primer.location, "_fw", "_rev")

primer.fasta <- system.file("extdata", "IMGT_data", "primers", "IGHV", 
                     "Ippolito2012.fasta", package = "openPrimeR")
primer.df <- read_primers(primer.fasta, "_fw", "_rev")
# Read multiple FASTA files
fasta.files <- list.files(system.file("extdata", "IMGT_data", "primers", 
                 "IGHV", package = "openPrimeR"), pattern = "*\\\\.fasta",
                 full.names = TRUE)[1:3]
primer.data <- read_primers(fasta.files)
# Read primers from a CSV file
primer.csv <- system.file("extdata", "IMGT_data", "comparison", 
             "primer_sets", "IGL", "IGL_openPrimeR2017.csv",  package = "openPrimeR")
primer.df <- read_primers(primer.csv)
# Read multiple primer CSV files
primer.files <- list.files(path = system.file("extdata", "IMGT_data", "comparison", 
                         "primer_sets", "IGH", package = "openPrimeR"),
                          pattern = "*\\\\.csv", full.names = TRUE)[1:3]
primer.data <- read_primers(primer.files)
# Read a mixture of FASTA/CSV files:
mixed.primers <- c(primer.fasta, primer.csv)
primer.data <- read_primers(mixed.primers)

# Select available settings
available.settings <- list.files(
     system.file("extdata", "settings", package = "openPrimeR"), 
     pattern = "*.xml", full.names = TRUE)
# Select one of the settings and load them
filename <- available.settings[1]
settings <- read_settings(filename)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/filters.R
\name{get.missing.df}
\alias{get.missing.df}
\title{Uncovered Templates.}
\usage{
get.missing.df(
  filtered.df,
  template.df,
  Tm.brackets,
  settings,
  mode.directionality
)
}
\arguments{
\item{filtered.df}{An object of class \code{Primers}.}

\item{template.df}{An object of class \code{Templates}.}
}
\value{
A Templates data frame containing the missing templates.
}
\description{
Computes a data frame containing the templates that are
not yet covered.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/helper_functions.R
\name{rbind.primer.data}
\alias{rbind.primer.data}
\title{Rbind for Primer Data Frames.}
\usage{
\method{rbind}{primer.data}(primer.data)
}
\arguments{
\item{primer.data}{List with primer data frames.}
}
\value{
A data frame containing all data in \code{primer.data}.
}
\description{
Merges all primer data frames in \code{primer.data} into one data frame.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/settings.R
\name{check_interval}
\alias{check_interval}
\title{Check Constraint Intervals}
\usage{
check_interval(constraints)
}
\arguments{
\item{constraints}{A list with constraint settings.}
}
\value{
\code{TRUE}, if all constraints specificy valid intervals, 
\code{FALSE} otherwise.
}
\description{
Checks the validity of constraint intervals.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/settings.R
\name{format.constraints}
\alias{format.constraints}
\title{Format Constraint Names.}
\usage{
\method{format}{constraints}(constraints)
}
\arguments{
\item{constraints}{The character vector of constraints to transform.}
}
\value{
A character vector with formatted constraint names.
}
\description{
Formats constraint names for frontend output.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/settings.R
\name{check_limits}
\alias{check_limits}
\title{Validity Check for Limits.}
\usage{
check_limits(constraint.settings, constraint.limits)
}
\arguments{
\item{constraint.settings}{A list with the constraint settings.}

\item{constraint.limits}{A list with the constraint relaxation limits.}
}
\value{
\code{TRUE} if the limits are at least as wide as the constraints, 
\code{FALSE} otherwise.
}
\description{
Checks whether the constraint limits are at least as
general as the constraint settings. This ensures that
the relaxation works in the proper direction.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/con_dimerization.R
\name{batchify.simple}
\alias{batchify.simple}
\title{Simple Batchification}
\usage{
batchify.simple(tasks)
}
\arguments{
\item{tasks}{The tasks to assign to individual batches.}
}
\value{
A list of lists containing indices corresponding to \code{tasks}, each list gives a batch.
}
\description{
Simple Batchification
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/IO.R
\name{my_ggsave}
\alias{my_ggsave}
\title{Wrapper for the ggplot2::ggsave function.}
\usage{
my_ggsave(filename, plot = ggplot2::last_plot(), ...)
}
\arguments{
\item{filename}{The filename to store the plot.}

\item{plot}{The ggplot object.}

\item{...}{Further arguments to the ggplot2 ggsave function.}
}
\value{
Stores \code{p} in \code{fname}.
}
\description{
Saves a plot using ggplot2's ggsave function.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/con_annealing_temperature.R
\name{compute.covered.Ta}
\alias{compute.covered.Ta}
\title{Annealing temperature}
\usage{
compute.covered.Ta(
  primer.df,
  mode.directionality = c("fw", "rev", "both"),
  template.df,
  na_salt_conc,
  mg_salt_conc,
  k_salt_conc,
  tris_salt_conc,
  primer_conc
)
}
\arguments{
\item{primer.df}{Primer data frame.}

\item{mode.directionality}{Primer directionality.}

\item{template.df}{Template data frame}

\item{na_salt_conc}{Sodium ion concentration.}

\item{mg_salt_conc}{Magensium ion concentration.}

\item{k_salt_conc}{Potassium ion concentration.}

\item{tris_salt_conc}{Tris buffer concentration.}

\item{primer_conc}{Primer concentration.}
}
\value{
The recommended annealing temperature.
}
\description{
Computes the annealing temperature using all binding events.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/optimization_ILP.R
\name{add.dimerization.constraints}
\alias{add.dimerization.constraints}
\title{Addition of Dimerization Constraints}
\usage{
add.dimerization.constraints(lprec, D.idx, indices)
}
\arguments{
\item{lprec}{An ILP instance.}

\item{D.idx}{Data frame giving the indices of dimerizing primer pairs.}

\item{indices}{Row indices for setting the dimerization constraints in \code{lprec}.}
}
\value{
\code{lprec} with added dimerization constraints.
}
\description{
Updates ILP formulation with dimerization constraints.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/optimization_ILP.R
\name{build.ILP.df}
\alias{build.ILP.df}
\title{Construction of ILP Results.}
\usage{
build.ILP.df(
  ILP,
  vars,
  primer.df,
  template.df,
  i,
  target.temp,
  time = NA,
  deltaG_Cutoff = NA,
  deltaG_Limit = NA
)
}
\arguments{
\item{ILP}{A solved ILP instance.}

\item{vars}{The ILP decision variables.}

\item{primer.df}{The primer data frame correspdong to the \code{ILP}.}

\item{template.df}{The template data frame.}

\item{i}{Index for the ILP.}

\item{target.temp}{Target melting temperature in Celsius.}

\item{time}{Runtime of the ILP.}

\item{deltaG_Cutoff}{Free energy cutoff used for the dimerization constraint.}

\item{deltaG_Limit}{The free energy boundary for dimerization.}
}
\value{
Data frame summarizing the ILP solution.
}
\description{
Constructs a data frame summarizing the properties of an ILP solution.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/primers.R
\name{pair_primers}
\alias{pair_primers}
\title{Pairing of Forward and Reverse Primers.}
\usage{
pair_primers(primer.df, template.df)
}
\arguments{
\item{primer.df}{An object of class \code{Primers}.}
}
\value{
An object of class \code{Primers} containing the paired primers.
}
\description{
Pairs forward and reverse primers such that coverage is maximized 
for every pair.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/helper_functions.R
\name{pos.to.range}
\alias{pos.to.range}
\title{Conversion of Positions to Ranges.}
\usage{
pos.to.range(pos1, pos2)
}
\arguments{
\item{pos1}{The first value.}

\item{pos2}{The second value.}
}
\value{
A character vector range.
}
\description{
Converts two numeric values to a range.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plots_coverage.R
\name{plot_primer_cvg_unstratified}
\alias{plot_primer_cvg_unstratified}
\title{Plot Individual Primer Coverage.}
\usage{
plot_primer_cvg_unstratified(p.df, template.df, groups = NULL)
}
\arguments{
\item{p.df}{Primer data frame.}

\item{template.df}{Template data frame.}

\item{groups}{Optional identifiers of template groups to be considered.
If not provided, all template groups are considered.}
}
\value{
A bar plot showing the coverage of individual primers.
}
\description{
Plots the coverage of individual primers.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/con_dimerization.R
\name{compute.all.cross.dimers.frontend}
\alias{compute.all.cross.dimers.frontend}
\title{Cross Dimerization.}
\usage{
compute.all.cross.dimers.frontend(
  primer.df,
  primer_conc,
  na_salt_conc,
  mg_salt_conc,
  k_salt_conc,
  tris_salt_conc,
  annealing.temp,
  for.shiny = FALSE,
  no.structures = FALSE
)
}
\arguments{
\item{primer.df}{Input primer data frame.}

\item{primer_conc}{Primer concentration.}

\item{na_salt_conc}{Sodium ion concentration.}

\item{mg_salt_conc}{Magensium ion concentration.}

\item{k_salt_conc}{Potassium ion concentration.}

\item{tris_salt_conc}{Tris buffer concentration.}

\item{annealing.temp}{The PCR annealing temperature.}

\item{for.shiny}{Whether to format the table for HTML output.}

\item{no.structures}{Whether to compute structures of dimers.}
}
\value{
A formatted data frame with cross-dimerization infos
}
\description{
Computes all cross dimers in a user-formatted way.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/IO.R
\name{create_report,list,list-method}
\alias{create_report,list,list-method}
\title{Creation of a PDF Report for Primer Comparison.}
\usage{
\S4method{create_report}{list,list}(
  primers,
  templates,
  fname,
  settings,
  sample.name = NULL,
  used.settings = NULL
)
}
\arguments{
\item{primers}{A list with evaluated \code{Primers} objects.}

\item{templates}{A list with \code{Templates} objects.}

\item{fname}{A character vector giving the file to store the report in.}

\item{settings}{A \code{DesignSettings} object.}

\item{sample.name}{An identifier for your analysis.}

\item{used.settings}{A named list (with fields "fw" and "rev") containing the forward/reverse used design settings.}
}
\value{
Creates a PDF file giving a report on the comparison of the input primers.
}
\description{
Creates a PDF report for comparing multiple primers.
}
\note{
Creating the report requires the external programs Pandoc (http://pandoc.org)
and LateX (http://latex-project.org).
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/con_dimerization.R
\name{get.matches}
\alias{get.matches}
\title{Identification of Sequence Matches.}
\usage{
get.matches(s1, s2)
}
\arguments{
\item{s1}{The aligned nucleotide sequence character vector.}

\item{s2}{The aligned, matching substring of \code{s1}.}
}
\value{
A match vector (\code{M} for matches, \code{X} for mismatches).
}
\description{
Identifies matches between two strings
provided by OligoArrayAux.
}
\keyword{internal}
